/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             AVL AIM
 *
 *      Copyright 2014-2016, Massachusetts Institute of Technology
 *      Licensed under The GNU Lesser General Public License, version 2.1
 *      See http://www.opensource.org/licenses/lgpl-2.1.php
 *
 *      Modified by Ryan Durscher AFLR/RQVC May 10, 2016
 */

#include <string.h>
#include <math.h>
#include "aimUtil.h"

#include "miscUtils.h" // Bring in miscellaneous utilities
#include "vlmUtils.h"

#ifdef WIN32
#define getcwd   _getcwd
#define PATH_MAX _MAX_PATH

#define strcasecmp  stricmp

#else
#include <unistd.h>
#include <limits.h>
#endif

#define NUMINPUT  9
#define NUMOUT    24
#define NUMPTS   47
#define PI        3.1415926535897931159979635
#define NINT(A)         (((A) < 0)   ? (int)(A-0.5) : (int)(A+0.5))

//#define DEBUG

/*!\mainpage Introduction
 *
 * \section overviewAVL AVL AIM Overview
 * The use of lower-dimensional design tools is clearly desirable in a multidisciplinary/multi-fidelity aero
 * design optimization setting. This is the crux of the Computational Aircraft Prototype Syntheses (CAPS) program.
 * In many ways describing geometry appropriate for AVL (the Athena Vortex Lattice) code is more cumbersome than
 * higher fidelity codes that require an Outer Mold Line. The goal is to make a CAPS AIM (Analysis Input Module)
 * that directly feeds input to AVL and extracts the output quantities of interest from AVL's execution. This
 * needs to be consistent with a build description that is hierarchical and multi-fidelity. That is, the build
 * description that generates the geometric data at this level can be further enhanced to produce the complete
 * OML of the aircraft design under consideration. As for the geometric description, AVL requires airfoil section
 * data specified at the appropriate locations that describe the <em> skeleton </em> of the aircraft. These
 * sections when <em> lofted </em> as groups and finally <em> unioned </em> together builds the OML. Clearly,
 * intercepting the state of the geometry before these higher-level operations are applied provides the data
 * appropriate for AVL. This naturally constructs a hierarchical geometric view where a design can progress
 * into higher fidelities and feedback can be achieved where we can go back to this level of description when need be.
 *
 * \section assumptionsAVL Assumptions
 * The AVL coordinate system assumption (X -- downstream, Y -- out the right wing, Z -- up) needs to be followed.
 *
 * Within <b> OpenCSM</b> there are a number of airfoil generation UDPs (User Defined Primitives). These include NACA 4
 * series, a more general NACA 4/5/6 series generator, Sobieczky's PARSEC parameterization and Kulfan's CST
 * parameterization. All of these UDPs generate <b> EGADS</b> <em> FaceBodies</em> where the <em>Face</em>'s underlying
 * <em>Surface</em>  is planar and the bounds of the <em>Face</em> is a closed set of <em>Edges</em> whose
 * underlying <em>Curves</em> contain the airfoil shape. In all cases there is a <em>Node</em> that represents
 * the <em>Leading Edge</em> point and one or two <em>Nodes</em> at the <em>Trailing Edge</em> -- one if the
 * representation is for a sharp TE and the other if the definition is open or blunt. If there are 2 <em>Nodes</em>
 * at the back, then there are 3 <em>Edges</em> all together and closed, even though the airfoil definition
 * was left open at the TE. All of this information will be used to automatically fill in the AVL geometry description.
 *
 * It should be noted that general construction in either <b> OpenCSM</b> or even <b> EGADS</b> will be supported
 * as long as the topology described above is used. But care should be taken when constructing the airfoil shape
 * so that a discontinuity (i.e.,  simply <em>C<sup>0</sup></em>) is not generated at the <em>Node</em> representing
 * the <em>Leading Edge</em>. This can be done by splining the entire shape as one and then intersecting the single
 *  <em>Edge</em> to place the LE <em>Node</em>.
 *
 *  The rest of the information and options required to fill out the AVL geometry input file (<b> xxx.avl</b>)
 *  will be found in the attributes attached to the <em>FaceBody</em> itself. The conventions used will be
 *  described in the next section.
 *
 * Also note that this first implementation is not intended to provide complete control over AVL. In particular,
 * there is no mention above of the <b> BODY</b>, <b> DESIGN</b>, <b> CLAF</b>, or <b> CDCL</b> AVL keywords.

 * Related pages:
 * - \ref attributeAVL
 * - \ref aimInputsAVL
 * - \ref aimOutputsAVL
 */

 /*! \page attributeAVL Attribute Use for the construction of xxx.avl
 * The following list of attributes drives the AVL geometric definition. Each <em>FaceBody</em> which relates to AVL
 * <b> Sections</b> will be marked up in an appropriate manner to drive the input file construction. Many attributes
 * are required and those that are optional are marked so in the description:
 *
 * - <b> capsFidelity</b> This attribute is a CAPS requirement to indicate the analysis fidelity the geometry
 * representation supports. Options are:  ALL, LINEARAERO
 *
 * - <b> capsReferenceArea</b>  [Optional: Default 1.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the SREF entry in the AVL input.
 *
 * - <b> capsReferenceChord</b>  [Optional: Default 1.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the CREF entry in the AVL input.
 *
 * - <b> capsReferenceSpan</b>  [Optional: Default 1.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the BREF entry in the AVL input.
 *
 * - <b> capsReferenceX</b>  [Optional: Default 0.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the XREF entry in the AVL input.
 *
 * - <b> capsReferenceY</b>  [Optional: Default 0.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the YREF entry in the AVL input.
 *
 * - <b> capsReferenceZ</b>  [Optional: Default 0.0] This attribute may exist on any <em> Body</em>.  Its
 *  value will be used as the ZREF entry in the AVL input.
 *
 * - <b> capsGroup</b> This string attribute labels the <em> FaceBody</em> as to which AVL Surface the section
 *  is assigned. This should be something like: <em> Main\_Wing</em>, <em> Horizontal\_Tail</em>, and etc. This informs
 *  the AVL AIM to collect all <em> FaceBodies</em> that match this attribute into a single AVL Surface.
 *
 *
 * The AVL Sections are automatically generated, one from each <em> FaceBody</em> and the details extracted from the geometry.
 * <b> Xle</b>, <b> Yle</b>, and <b> Zle</b>, are taken from the <em> Node</em> Associated with the <em> Leading Edge</em>.
 * The <b> Chord</b> is computed by getting the distance between the LE and TE (if there are 3 <em> Edges</em> in the
 * <em> FaceBody</em> the TE point is considered the mid-position on that third <em> Edge</em>). <b> Ainc</b> is computed by
 * registering the chordal direction of the <em> FaceBody</em> against the X-Z plane. The airfoil shapes are generated by
 * sampling the <em> Curves</em> and put directly in the input file via the <b> AIRFOIL</b> keyword after being normalized.
 *
 */


// To be changed
 /*  - <b> avlComponent</em>  [optional]. This integer (or real) attribute is examined for the AVL Component index.
 *  This may be on any or all <em> FaceBodies</em> associated with the AVL Surface.

- <b> avlKeyword</em>  [optional]. This string attribute may contain the word (or words): {\tt NOWAKE</em>, {\tt NOALBE</em> and/or {\tt NOLOAD</em>. Multiple keywords need
to be delimited by a colon (`:'). This is associated with the AVL Surface.


\item {\bf avlSecValues.} This real valued attribute must be of at least 5 {\it doubles} in length and contain:
   \begin{enumerate}
   \item {\tt Nchord} -- The number of chordwise horseshoe vortices placed on the surface.
   \item {\tt Cspace} -- The chordwise vortex spacing parameter.
   \item {\tt Nspan} -- The number of spanwise horseshoe vortices placed on the surface.
   \item {\tt Sspace} -- The spanwise vortex spacing parameter.
   \item {\tt iYdup} -- The Y symmetry for this Surface (0 -- do nothing, 1 -- duplicate about the Y=0.0 plane).
   \end{enumerate}
   It should be noted that the first 2 (and last) values refer to the entire AVL Surface/capsGroup and therefore the first occurrence of the \textbf{\textit {capsGroup}} attribute in the list of {\it bodies} sets these values in the AVL input file (though {\tt iYdup} can be set in any Section).

\item {\bf avlCntrl}\textbf{\textit {Name}}  [optional]. This optional numerical attribute indicates that this {\it FaceBody} Section is a member of the control surface named as part of the attribute name (for example:  {\tt avlCntrlAileron} specifies that this Section is part of the {\it Aileron} control surface). This real valued attribute must be of at least 6 {\it doubles} in length and contain:
   \begin{enumerate}
   \item {\tt gain} -- The control deflection gain.
   \item {\tt Xhinge} -- The x/c location of hinge.
   \item {\tt Xhvec} -- The first component of the vector giving hinge axis about which surface rotates.
   \item {\tt Yhvec} -- The second component.
   \item {\tt Zhvec} -- The third component.
   \item {\tt SgnDup } -- The signed magnitude (or simply the scale factor) for the deflection of the duplicated surface.
   \end{enumerate}
\end{itemize}
*/


/*
  typedef struct {
    char   *name;
    double Cspace;
    int    Nchord;
    int    nowake;
    int    noalbe;
    int    noload;
    int    compon;
    int    iYdup;
    int    nsec;
    double xyz[3];        per body
    int    le;
    int    index;         per section
  } aimSurface;
*/



  static int nInstance = 0;

  extern /*@null@*/ /*@only@*/
         char       *EG_strdup(/*@null@*/ const char *str);


#ifdef WIN32
static int
getline(char **linep, size_t *linecapp, FILE *stream)
{
  int  i, cnt = -1;
  char buffer[1026];

  *linep    = buffer;
  *linecapp = 1024;
  while (cnt < 1024) {
    i = fread(&buffer[cnt+1], sizeof(char), 1, stream);
    if (i == 0) return cnt;
    if ((buffer[cnt+1] == 12) || (buffer[cnt+1] == 15) ||
        (buffer[cnt+1] == 0)) break;
    cnt++;
  }
  buffer[cnt+1] = 0;
  if (cnt == -1) cnt++;
  return cnt;
}
#endif

/* ********************** AVL AIM Helper Functions ************************** */

static int findLeadingEdge(ego body, int *nodeIndexLE, double *xyzLE )
{
	int status; // Function return
	int  i; // Indexing

	// Node variables
	ego    *nodes = NULL;
	int    numNode;
	double xmin, xyz[3];

	//EGADS returns (unused)
	int oclass, mtype, *sens = NULL, numChildren;
	ego ref, *children = NULL;
  
	*nodeIndexLE = 0;
  
	// Assume the LE position is the most forward Node in X
	status = EG_getBodyTopos(body, NULL, NODE, &numNode, &nodes);
	if (status != EGADS_SUCCESS) goto cleanup;


	for (i = 0; i < numNode; i++) {
		status = EG_getTopology(nodes[i], &ref, &oclass, &mtype, xyz,
							  	&numChildren, &children, &sens);
		if (status != EGADS_SUCCESS) goto cleanup;

		if (*nodeIndexLE == 0) {
			*nodeIndexLE = i+1;
			xmin = xyz[0];
		} else {

			if (xyz[0] < xmin) {
				*nodeIndexLE = i+1;
				xmin = xyz[0];
			}
		}
	}

	if (*nodeIndexLE == 0) return CAPS_SUCCESS;
  
	status = EG_getTopology(nodes[*nodeIndexLE-1], &ref, &oclass, &mtype, xyzLE, &numChildren, &children, &sens);

	goto cleanup;

	cleanup:
		if (status != CAPS_SUCCESS) printf("Error in findLeadingEdge - status %d\n", status);
		if (nodes != NULL) EG_free(nodes);

		return status;
}

static int findTrailingEdge(ego body, int *nodeIndexTE, double *xyzTE )
{
	int status; // Function return
	int  i; // Indexing

	// Node variables
	ego    *nodes = NULL;
	int    numNode;
	double xmax, xyz[3];

	//EGADS returns (unused)
	int oclass, mtype, *sens = NULL, numChildren;
	ego ref, *children = NULL;

	*nodeIndexTE = 0;

	// Assume the TE position is the most rear Node in X
	status = EG_getBodyTopos(body, NULL, NODE, &numNode, &nodes);
	if (status != EGADS_SUCCESS) goto cleanup;


	for (i = 0; i < numNode; i++) {
		status = EG_getTopology(nodes[i], &ref, &oclass, &mtype, xyz,
							  	&numChildren, &children, &sens);
		if (status != EGADS_SUCCESS) goto cleanup;

		if (*nodeIndexTE == 0) {
			*nodeIndexTE = i+1;
			xmax = xyz[0];
		} else {

			if (xyz[0] > xmax) {
				*nodeIndexTE = i+1;
				xmax = xyz[0];
			}
		}
	}

	if (*nodeIndexTE == 0) return CAPS_SUCCESS;

	status = EG_getTopology(nodes[*nodeIndexTE-1], &ref, &oclass, &mtype, xyzTE, &numChildren, &children, &sens);

	goto cleanup;

	cleanup:
		if (status != CAPS_SUCCESS) printf("Error in findTrailingEdge - status %d\n", status);
		if (nodes != NULL) EG_free(nodes);

		return status;
}

/*
// Original
static void
orderSections(int i, aimSurface *surfaces)
{
  int    i1, i2, j, k, hit;
  double box[6], lScale;
  
  j = surfaces[0].index;
  box[0] = box[3] = surfaces[j].xyz[0];
  box[1] = box[4] = surfaces[j].xyz[1];
  box[2] = box[5] = surfaces[j].xyz[2];
  for (k = 0; k < surfaces[i].nsec; k++) {
    j = surfaces[k].index;
    if (surfaces[j].xyz[0] < box[0]) box[0] = surfaces[j].xyz[0];
    if (surfaces[j].xyz[0] > box[3]) box[3] = surfaces[j].xyz[0];
    if (surfaces[j].xyz[1] < box[1]) box[1] = surfaces[j].xyz[1];
    if (surfaces[j].xyz[1] > box[4]) box[4] = surfaces[j].xyz[1];
    if (surfaces[j].xyz[2] < box[2]) box[2] = surfaces[j].xyz[2];
    if (surfaces[j].xyz[2] > box[5]) box[5] = surfaces[j].xyz[2];
  }
  lScale = box[3] - box[0];
  if (box[4] - box[1] > lScale) lScale = box[4] - box[1];
  if (box[5] - box[2] > lScale) lScale = box[5] - box[2];
  
  if ((box[4]-box[1])/lScale > 1.e-5) {
     Y ordering
    do {
      for (hit = k = 0; k < surfaces[i].nsec-1; k++) {
        i1 = surfaces[k  ].index;
        i2 = surfaces[k+1].index;
        if (surfaces[i1].xyz[1] <= surfaces[i2].xyz[1]) continue;
        surfaces[k  ].index = i2;
        surfaces[k+1].index = i1;
        hit++;
      }
    } while (hit != 0);
  } else if ((box[5]-box[2])/lScale > 1.e-5) {
     assume Vertical Tail and do Z ordering
    do {
      for (hit = k = 0; k < surfaces[i].nsec-1; k++) {
        i1 = surfaces[k  ].index;
        i2 = surfaces[k+1].index;
        if (surfaces[i1].xyz[2] <= surfaces[i2].xyz[2]) continue;
        surfaces[k  ].index = i2;
        surfaces[k+1].index = i1;
        hit++;
      }
    } while (hit != 0);
  } else {
    printf(" AIM Warning: No Y or Z Variation in Ssurface LE Bounding Box!\n");
    printf("              %lf %lf   %lf %lf   %lf %lf  - lScale = %lf\n",
           box[0], box[3], box[1], box[4], box[2], box[5], lScale);
  }

   printf(" Sections =");
   for (j = 0; j < surfaces[i].nsec; j++) printf(" %d", surfaces[j].index);
   printf("\n");

}
*/

static void orderSections(int numSection, vlmSectionStruct section[])
{
    int    i1, i2, j, k, hit;
    double box[6], lScale;

    j = section[0].sectionIndex;

    box[0] = box[3] = section[j].xyzLE[0];
    box[1] = box[4] = section[j].xyzLE[1];
    box[2] = box[5] = section[j].xyzLE[2];

    for (k = 0; k < numSection; k++) {
        j = section[k].sectionIndex;
        if (section[j].xyzLE[0] < box[0]) box[0] = section[j].xyzLE[0];
        if (section[j].xyzLE[0] > box[3]) box[3] = section[j].xyzLE[0];
        if (section[j].xyzLE[1] < box[1]) box[1] = section[j].xyzLE[1];
        if (section[j].xyzLE[1] > box[4]) box[4] = section[j].xyzLE[1];
        if (section[j].xyzLE[2] < box[2]) box[2] = section[j].xyzLE[2];
        if (section[j].xyzLE[2] > box[5]) box[5] = section[j].xyzLE[2];
    }

    lScale = box[3] - box[0];
    if (box[4] - box[1] > lScale) lScale = box[4] - box[1];
    if (box[5] - box[2] > lScale) lScale = box[5] - box[2];

    if ((box[4]-box[1])/lScale > 1.e-5) {
        /* Y ordering */
        do {
            for (hit = k = 0; k < numSection-1; k++) {
                i1 = section[k  ].sectionIndex;
                i2 = section[k+1].sectionIndex;

                if (section[i1].xyzLE[1] <= section[i2].xyzLE[1]) continue;

                section[k  ].sectionIndex = i2;
                section[k+1].sectionIndex = i1;
                hit++;
            }
        } while (hit != 0);

    } else if ((box[5]-box[2])/lScale > 1.e-5) {
        /* assume Vertical Tail and do Z ordering */
        do {
            for (hit = k = 0; k < numSection-1; k++) {
                i1 = section[k  ].sectionIndex;
                i2 = section[k+1].sectionIndex;

                if (section[i1].xyzLE[2] <= section[i2].xyzLE[2]) continue;

                section[k  ].sectionIndex = i2;
                section[k+1].sectionIndex = i1;
                hit++;
            }
        } while (hit != 0);

    } else {
        printf(" AIM Warning: No Y or Z Variation in Surface LE Bounding Box!\n");
        printf("              %lf %lf   %lf %lf   %lf %lf  - lScale = %lf\n",box[0],
                                                                             box[3],
                                                                             box[1],
                                                                             box[4],
                                                                             box[2],
                                                                             box[5],
                                                                             lScale);
    }

    /*
     printf(" Sections =");
     for (j = 0; j < surfaces[i].nsec; j++) printf(" %d", surfaces[j].sectionIndex);
     printf("\n");
    */
}


static void writeSection(FILE *fp, ego body, int le, double *xle, int Nspan, double Sspace)
{
  int          i, j, n, stat, atype, alen, nNode, nEdge, oclass, mtype;
  int          *sens, cnt, sense;
  double       chord, ainc, xyz[3], xmin, trange[2], result[9];
  double       xdot[3], ydot[3];
  const int    *ints;
  const char   *string;
  const double *reals;
  ego          *nodes = NULL, *edges = NULL, ref, te, obj, *objs;
  
  
  /* get LE position, Chord and Ainc from the Body */
  stat = EG_getBodyTopos(body, NULL, NODE, &nNode, &nodes);
  if (stat != EGADS_SUCCESS) {
    printf(" AIM Warning: getBodyTopos Nodes = %d\n", stat);
    goto cleanup;
  }
  stat = EG_getBodyTopos(body, NULL, EDGE, &nEdge, &edges);
  if (stat != EGADS_SUCCESS) {
    printf(" AIM Warning: getBodyTopos Edges = %d\n", stat);
    goto cleanup;
  }
  if ((nEdge != nNode) || (nNode < 2) || (nNode > 3)) {
    printf(" AIM Warning: Body has %d nodes and %d edges!\n", nNode, nEdge);
    goto cleanup;
  }
  
  /* find TE Edge object, if any */
  te = NULL;
  for (i = 0; i < nEdge; i++) {
    stat = EG_getTopology(edges[i], &ref, &oclass, &mtype, trange,
                          &n, &objs, &sens);
    if (stat != EGADS_SUCCESS) {
      printf(" AIM Warning: Edge %d getTopology = %d!\n", i, stat);
      continue;
    }
    if (n == 1) {
      printf(" AIM Warning: Edge %d has 1 Node!\n", i);
      continue;
    }
    if ((objs[0] != nodes[le-1]) && (objs[1] != nodes[le-1])) {
      te = edges[i];
/*    printf(" TE Edge = %d\n", i);  */
      break;
    }
  }
  if ((nEdge == 3) && (te == NULL)) {
    printf(" AIM Warning: No TE Edge with nEdge = 3!\n");
    goto cleanup;
  }
  if (te == NULL) {
    i = le;
    if (i == 2) i = 0;
    EG_getTopology(nodes[i], &ref, &oclass, &mtype, result, &n, &objs, &sens);
  } else {
    EG_getTopology(te, &ref, &oclass, &mtype, trange, &n, &objs, &sens);
    xmin = 0.5*(trange[0]+trange[1]);
    stat = EG_evaluate(te, &xmin, result);
    if (stat != EGADS_SUCCESS) {
      printf(" AIM Warning: evaluate = %d!\n", stat);
      goto cleanup;
    }
  }
  xdot[0]  =  result[0] - xle[0];
  xdot[1]  =  result[1] - xle[1];
  xdot[2]  =  result[2] - xle[2];
  chord    =  sqrt(xdot[0]*xdot[0] + xdot[1]*xdot[1] + xdot[2]*xdot[2]);
  xdot[0] /=  chord;
  xdot[1] /=  chord;
  xdot[2] /=  chord;
  ainc     = -acos(xdot[0])*180./PI;
  ydot[0]  = -xdot[2];
  ydot[1]  =  xdot[1];
  ydot[2]  =  xdot[0];
  if (fabs(xdot[1]) > 1.e-5)
    printf(" AIM Warning: Section not in Y-plane!\n");
  
  fprintf(fp, "SECTION\n%lf %lf %lf  %lf %lf  %d %lf\n\n",
          xle[0], xle[1], xle[2], chord, ainc, Nspan, Sspace);
  
  fprintf(fp, "AIRFOIL\n");
  for (cnt = i = 0; i < nEdge; i++) {
    if (edges[i] == te) continue;
    EG_getTopology(edges[i], &ref, &oclass, &mtype, trange, &n, &objs, &sens);
    if (objs[0] == nodes[le-1]) {
      sense = -1;
      obj   = objs[1];
    } else {
      sense =  1;
      obj   = objs[0];
    }
    if (cnt == 1) sense = -sense;
    
    EG_getTopology(obj, &ref, &oclass, &mtype, xyz, &n, &objs, &sens);
    xyz[0] -= xle[0];
    xyz[1] -= xle[1];
    xyz[2] -= xle[2];
    if (cnt == 0) {
      fprintf(fp, "1.000000 %lf\n",
              (ydot[0]*xyz[0]+ydot[1]*xyz[1]+ydot[2]*xyz[2])/chord);
    } else {
      fprintf(fp, "0.000000 0.000000\n");
    }
    for (j = 1; j < NUMPTS; j++) {
      if (sense == 1) {
        xmin = trange[0] + j*(trange[1]-trange[0])/NUMPTS;
      } else {
        xmin = trange[1] - j*(trange[1]-trange[0])/NUMPTS;
      }
      EG_evaluate(edges[i], &xmin, result);
      result[0] -= xle[0];
      result[1] -= xle[1];
      result[2] -= xle[2];
      fprintf(fp, "%lf %lf\n",
              (xdot[0]*result[0]+xdot[1]*result[1]+xdot[2]*result[2])/chord,
              (ydot[0]*result[0]+ydot[1]*result[1]+ydot[2]*result[2])/chord);
    }
    cnt++;
    if (cnt == 2)
      fprintf(fp, "1.000000 %lf\n",
              (ydot[0]*xyz[0]+ydot[1]*xyz[1]+ydot[2]*xyz[2])/chord);
  }
  
  fprintf(fp, "\n");

  goto cleanup;

  cleanup:

  	  if (nodes != NULL) EG_free(nodes);
  	  if (edges != NULL) EG_free(edges);

  	  return;
}



/* ********************** Exposed AIM Functions ***************************** */

int
aimInitialize(/*@unused@*/ int ngIn, /*@unused@*/ /*@null@*/ capsValue *gIn,
              int *nFid, int **fids, int *nIn, int *nOut,
              int *nFields, char ***fnames, int **ranks)
{
  int *ints, ret;

#ifdef DEBUG
  printf("\n avlAIM/aimInitialize   ngIn = %d!\n", ngIn);
#endif

  // Number of geometry fidelities that are required by the AIM and what they are
  *nFid    = 4;
  ints     = (int *) EG_alloc(*nFid*sizeof(int));
  if (ints == NULL) return EGADS_MALLOC;
  ints[0]  = ALL | ALL;
  ints[1]  = ALL | LINEARAERO;
  ints[2]  = FACEBODY | ALL;
  ints[3]  = FACEBODY | LINEARAERO;
  *fids    = ints;

  /*! \page geomFidelityAVL Geometry Fidelity
   * The geometric fidelity for the AVL AIM requires the sections, though the global attribute
   * capsFidelity maybe set to ALL or LINEARAERO.
   */

  /* specify the number of analysis input and out "parameters" */
  *nIn     = NUMINPUT;
  *nOut    = NUMOUT;

  /* specify the field variables this analysis can generate */
  *nFields = 0;
  *ranks   = NULL;
  *fnames  = NULL;
  
  /* return an index for the instance generated */
  ret = nInstance;
  nInstance++;
  return ret;
}


int
aimInputs(/*@unused@*/ int inst, /*@unused@*/ void *aimInfo, int index,
          char **ainame, capsValue *defval)
{
	/*! \page aimInputsAVL AIM Inputs
	 * The following list outlines the AVL inputs along with their default value available
	 * through the AIM interface.
	 */

	#ifdef DEBUG
		printf(" avlAIM/aimInputs instance = %d  index = %d!\n", inst, index);
	#endif
  
	if (index == 1) {
		*ainame           = EG_strdup("Mach");
		defval->type      = Double;
		defval->vals.real = 0.0;
		defval->units     = NULL;

		/*! \page aimInputsAVL
		 * - <B> Mach = 0.0 </B> <br>
		 *  Mach number.
		 */

	} else if (index == 2) {
		*ainame           = EG_strdup("Alpha");
		defval->type      = Double;
		defval->vals.real = 0.0;
		defval->units     = NULL;

		/*! \page aimInputsAVL
		 * - <B> Alpha = 0.0 </B> <br>
		 *  Angle of attack.
		 */

	} else if (index == 3) {
		*ainame           = EG_strdup("Beta");
		defval->type      = Double;
		defval->vals.real = 0.0;
		defval->units     = NULL;

		/*! \page aimInputsAVL
		 * - <B> Beta = 0.0 </B> <br>
		 *  Sideslip angle.
		 */

	} else if (index == 4) {
		*ainame           = EG_strdup("RollRate");
		defval->type      = Double;
		defval->vals.real = 0.0;
		defval->units     = NULL;

		/*! \page aimInputsAVL
		 * - <B> RollRate = 0.0 </B> <br>
		 *  Non-dimensional roll rate.
		 */

	} else if (index == 5) {
		*ainame           = EG_strdup("PitchRate");
		defval->type      = Double;
		defval->vals.real = 0.0;
		defval->units     = NULL;

		/*! \page aimInputsAVL
		 * - <B> PitchRate = 0.0 </B> <br>
		 *  Non-dimensional pitch rate.
		 */

	} else if (index == 6) {
		*ainame           = EG_strdup("YawRate");
		defval->type      = Double;
		defval->vals.real = 0.0;
		defval->units     = NULL;

		/*! \page aimInputsAVL
		 * - <B> YawRate = 0.0 </B> <br>
		 *  Non-dimensional yaw rate.
		 */

	} else if (index == 7) {
		*ainame           = EG_strdup("CDp");
		defval->type      = Double;
		defval->vals.real = 0.0;
		defval->units     = NULL;

		/*! \page aimInputsAVL
		 * - <B> CDp = 0.0 </B> <br>
		 *  A fixed value of profile drag to be added to all simulations.
		 */

	} else if (index == 8) {
		*ainame              = EG_strdup("AVL_Surface");
		defval->type         = Tuple;
		defval->nullVal      = IsNull;
		//defval->units        = NULL;
		defval->dim          = 1;
		defval->lfixed       = Change;
		defval->vals.tuple   = NULL;

		/*! \page aimInputsAVL
		 * - <B>AVL_Surface = NULL </B> <br>
		 * See \ref vlmSurface for additional details.
		 */

	} else if (index == 9) {
		*ainame              = EG_strdup("AVL_Control");
		defval->type         = Tuple;
		defval->nullVal      = IsNull;
		//defval->units        = NULL;
		defval->dim          = 1;
		defval->lfixed       = Change;
		defval->vals.tuple   = NULL;

		/*! \page aimInputsAVL
		 * - <B>AVL_Control = NULL </B> <br>
		 * See \ref vlmControl for additional details.
		 */
	}

	return CAPS_SUCCESS;
}


int aimPreAnalysis(/*@unused@*/ int inst, /*@unused@*/ void *aimInfo,
                   const char *analysisPath, /*@null@*/ capsValue *aimInputs, capsErrs **errs)
{
    int status; // Function return status

    int i, j, body, surf, section, control; // Indexing
    int numBody, attrIndex;

    int found = (int) false; // Boolean tester

    int atype, alen;

    /*@-unrecog@*/
    char currentPath[PATH_MAX];
    /*@+unrecog@*/

    double       Cspace;
    double 	     Sref, Cref, Bref, Xref, Yref, Zref;
    const int    *ints;
    const char   *string, *aname;
    const double *reals;
    ego          *bodies;

    int stringLength = 0; // String manipulation

    // File I/O
    FILE *fp = NULL;
    char inputFilename[] = "avlInput.txt";
    char avlFilename[] = "caps.avl";

    // AVL surface information
    int numAVLSurface = 0;
    vlmSurfaceStruct *avlSurface = NULL;

    // AVL control surface information
    int numAVLControl = 0;
    vlmControlStruct *avlControl = NULL;

    const char *groupName = NULL, *attrName = NULL;

    int numControlName = 0;
    char **controlName = NULL;

    mapAttrToIndexStruct attrMap; // Attribute to index map

    // Control related variables
    //double xyzHingeMag, xyzHingeVec[3];

    #ifdef DEBUG
        printf(" avlAIM/aimPreAnalysis instance = %d\n", inst);
    #endif
  
    // Initialize reference values
    Sref = Cref = Bref = 1.0;
    Xref = Yref = Zref = 0.0;

    // NULL out errs
    *errs = NULL;

    if (aimInputs == NULL) {
        #ifdef DEBUG
            printf(" avlAIM/aimPreAnalysis aimInputs == NULL!\n");
        #endif

        return CAPS_NULLVALUE;
    }

    status = aim_getBodies(aimInfo, &numBody, &bodies);
    if (status != CAPS_SUCCESS) {
        #ifdef DEBUG
            printf(" avlAIM/aimPreAnalysis getBodies = %d!\n", status);
        #endif
        return status;
    }

    if (numBody == 0 || bodies == NULL) {
        printf(" avlAIM/aimPreAnalysis No Bodies!\n");
        return CAPS_SOURCEERR;
    }

    // Container for attribute to index map
    status = initiate_mapAttrToIndexStruct(&attrMap);
    if (status != CAPS_SUCCESS) return status;

    // Get capsGroup name and index mapping to make sure all bodies have a capsGroup value
    status = create_CAPSGroupAttrToIndexMap(numBody,
                                            bodies,
                                            0, // Only search down to the body level of the EGADS body
                                            &attrMap);
    if (status != CAPS_SUCCESS) goto cleanup;

    // Get AVL surface information
    if (aimInputs[aim_getIndex(aimInfo, "AVL_Surface", ANALYSISIN)-1].nullVal == NotNull) {

        status = get_vlmSurface(aimInputs[aim_getIndex(aimInfo, "AVL_Surface", ANALYSISIN)-1].length,
                                aimInputs[aim_getIndex(aimInfo, "AVL_Surface", ANALYSISIN)-1].vals.tuple,
                                &attrMap,
                                &numAVLSurface,
                                &avlSurface);

    } else {
        printf("No AVL_SURFACE tuple specified\n");
        status = CAPS_NOTFOUND;
    }
    if (status != CAPS_SUCCESS) goto cleanup;

    // Get AVL control surface information
    if (aimInputs[aim_getIndex(aimInfo, "AVL_Control", ANALYSISIN)-1].nullVal == NotNull) {

        status = get_vlmControl(aimInputs[aim_getIndex(aimInfo, "AVL_Control", ANALYSISIN)-1].length,
                                aimInputs[aim_getIndex(aimInfo, "AVL_Control", ANALYSISIN)-1].vals.tuple,
                                &numAVLControl,
                                &avlControl);

        if (status != CAPS_SUCCESS) goto cleanup;
    }

    // Accumulate section data
    for (body = 0; body < numBody; body++) {

        // Loop through surfaces
        for (surf = 0; surf < numAVLSurface; surf++) {

            status = retrieve_CAPSGroupAttr(bodies[body], &groupName);
            if (status != CAPS_SUCCESS) {
                printf("No capsGroup value found on body %d, body will not be used\n", body+1);
                continue;
            }

            status = get_mapAttrToIndexIndex(&attrMap, groupName, &attrIndex);
            if (status == CAPS_NOTFOUND) {
                printf("AVL Surface name \"%s\" not found in attrMap\n", groupName);
                goto cleanup;
            }

            found = (int) false;

            // See if attrIndex is in the attrIndex array for the surface
            for (i = 0; i < avlSurface[surf].numAttr; i++) {

                if (attrIndex == avlSurface[surf].attrIndex[i]) {
                    found = (int) true;
                    break;
                }
            }

            // If attrIndex isn't in the array the; body doesn't belong in this surface
            if (found == (int) false) continue;

            // Increment the number of sections
            avlSurface[surf].numSection += 1;

            // Get section index
            section = avlSurface[surf].numSection -1;

            if (avlSurface[surf].numSection == 1) {
                avlSurface[surf].vlmSection = (vlmSectionStruct *) EG_alloc(1*sizeof(vlmSectionStruct));
            } else {

            	avlSurface[surf].vlmSection = (vlmSectionStruct *) EG_reall(avlSurface[surf].vlmSection,
                                                                            avlSurface[surf].numSection*
                                                                            sizeof(vlmSectionStruct));
            }
            if (avlSurface[surf].vlmSection == NULL) {
                status = EGADS_MALLOC;
                goto cleanup;
            }

            status = initiate_vlmSection(&avlSurface[surf].vlmSection[section]);
            if (status != CAPS_SUCCESS) goto cleanup;

            // Store the section index
            avlSurface[surf].vlmSection[section].sectionIndex = section;

            // Store the body index that the section is coming from
            avlSurface[surf].vlmSection[section].bodyIndex = body;

            // Get leading edge coordinates
            status = findLeadingEdge(bodies[body],
            						 &avlSurface[surf].vlmSection[section].nodeIndexLE,
									 avlSurface[surf].vlmSection[section].xyzLE);
            if (status != CAPS_SUCCESS) goto cleanup;

            if (avlSurface[surf].vlmSection[section].nodeIndexLE == 0) {
                printf(" avlAIM/aimPreAnalysis Body %d has no LE!\n", body +1);
                status = CAPS_NOTFOUND;
                goto cleanup;
            }

            // Get trailing edge coordinates
			status = findTrailingEdge(bodies[body],
								   	  &avlSurface[surf].vlmSection[section].nodeIndexTE,
									  avlSurface[surf].vlmSection[section].xyzTE);
			if (status != CAPS_SUCCESS) goto cleanup;

			if (avlSurface[surf].vlmSection[section].nodeIndexTE == 0) {
				printf(" avlAIM/aimPreAnalysis Body %d has no TE!\n", body +1);
				status = CAPS_NOTFOUND;
				goto cleanup;
			}

        }
    }

    // Loop through surfaces and transfer control surface data to sections
    for (surf = 0; surf < numAVLSurface; surf++) {

    	status = get_ControlSurface(bodies,
    								numAVLControl,
									avlControl,
									&avlSurface[surf]);
		if (status != CAPS_SUCCESS) goto cleanup;
    }

    // Order cross sections for each surface
    for (surf = 0; surf < numAVLSurface; surf++) orderSections(avlSurface[surf].numSection, avlSurface[surf].vlmSection);

    /*
   	// Determine the hinge line for any flaps
    for (surf = 0; surf < numAVLSurface; surf++) {

    	for (section = 0; section < avlSurface[surf].numSection; section++) {

    		for (control = 0; control < avlSurface[surf].vlmSection[section].numControl; control++) {

    			for (i = 0; i < avlSurface[surf].numSection; i++) {

    				if (i == section) continue;

    				for (j = 0; j < avlSurface[surf].vlmSection[i].numControl; j++) {

    					if( strcmp(avlSurface[surf].vlmSection[section].vlmControl[control].name,
    							   avlSurface[surf].vlmSection[i].vlmControl[j].name) == 0) {

    						xyzHingeVec[0] = avlSurface[surf].vlmSection[section].vlmControl[control].xyzHinge[0] -
											 avlSurface[surf].vlmSection[i].vlmControl[j].xyzHinge[0];

    						xyzHingeVec[1] = avlSurface[surf].vlmSection[section].vlmControl[control].xyzHinge[1] -
    										 avlSurface[surf].vlmSection[i].vlmControl[j].xyzHinge[1];

    						xyzHingeVec[2] = avlSurface[surf].vlmSection[section].vlmControl[control].xyzHinge[2] -
    										 avlSurface[surf].vlmSection[i].vlmControl[j].xyzHinge[2];

    						xyzHingeMag = sqrt(pow(xyzHingeVec[0], 2) +
    										   pow(xyzHingeVec[1], 2) +
											   pow(xyzHingeVec[2], 2));

    						avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[0] = xyzHingeVec[0]/xyzHingeMag;
    						avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[1] = xyzHingeVec[1]/xyzHingeMag;
    						avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[2] = xyzHingeVec[2]/xyzHingeMag;

    					}
    				}
    			}
    		}
    	}
    }
    */

    // Get where we are and set the path to our input
    (void) getcwd(currentPath, PATH_MAX);
    if (chdir(analysisPath) != 0) {
        #ifdef DEBUG
            printf(" avlAIM/aimPreAnalysis Cannot chdir to %s!\n", analysisPath);
        #endif

        status = CAPS_DIRERR;
        goto cleanup;
    }

    // Open and write the input to control the AVL session
    fp = fopen(inputFilename, "w");
    if (fp == NULL) {
        printf("Unable to open file %s\n!", inputFilename);

        chdir(currentPath);
        status = CAPS_IOERR;
        goto cleanup;
    }

    // Print the session file for the AVL run
    fprintf(fp, "OPER\n");

    fprintf(fp, "A A ");//Alpha
    fprintf(fp, "%lf\n", aimInputs[1].vals.real);

    fprintf(fp, "B B ");//Beta
    fprintf(fp, "%lf\n", aimInputs[2].vals.real);

    fprintf(fp, "R R ");//Roll Rate pb/2V
    fprintf(fp, "%lf\n", aimInputs[3].vals.real);

    fprintf(fp, "P P ");//Pitch Rate qc/2v
    fprintf(fp, "%lf\n", aimInputs[4].vals.real);

    fprintf(fp, "Y Y ");//Yaw Rate rb/2v
    fprintf(fp, "%lf\n", aimInputs[5].vals.real);

    // Check for control surface information
    j = 1;
    for (surf = 0; surf < numAVLSurface; surf++) {

    	for (i = 0; i < avlSurface[surf].numSection; i++) {

    		section = avlSurface[surf].vlmSection[i].sectionIndex;

    		for (control = 0; control < avlSurface[surf].vlmSection[section].numControl; control++) {

    			fprintf(fp, "D%d D%d %f\n",j, j,avlSurface[surf].vlmSection[section].vlmControl[control].deflectionAngle);
    			j += 1;
    		}
    	}
    }

    /*if(aimInputs[7].nullVal != IsNull) {
        // write control surface settings
        if (aimInputs[7].length == 1){
            fprintf(fp, "D1 D1 ");//Pitch Rate qc/2v
            fprintf(fp, "%lf\n", aimInputs[7].vals.real);
        } else {
            for (i=0; i < aimInputs[7].length; i++) {
                fprintf(fp,"D%d D%d %lf\n",i+1,i+1,aimInputs[7].vals.reals[i]);
            }
        }
    }*/

    fprintf(fp, "M\n"); // Modify parameters
    fprintf(fp, "MN\n");//Mach
    fprintf(fp, "%lf\n", aimInputs[0].vals.real);

    fprintf(fp, "\nX\n\n"); // Execute the calculation
    fprintf(fp, "Quit\n"); // Quit AVL
    fclose(fp);
    fp = NULL;

    /* open the file and write the avl data */
    fp = fopen(avlFilename, "w");
    if (fp == NULL) {
        printf("Unable to open file %s\n!", avlFilename);

        chdir(currentPath);

        return CAPS_IOERR;
    }

    // Loop over bodies and look for reference quantity attributes
    for (i=0; i < numBody; i++) {
	  status = EG_attributeRet(bodies[i], "capsReferenceArea", &atype, &alen, &ints, &reals, &string);
	  if (status == EGADS_SUCCESS) Sref = (double) reals[0];

	  status = EG_attributeRet(bodies[i], "capsReferenceChord", &atype, &alen, &ints, &reals, &string);
	  if (status == EGADS_SUCCESS) Cref = (double) reals[0];

	  status = EG_attributeRet(bodies[i], "capsReferenceSpan", &atype, &alen, &ints, &reals, &string);
	  if (status == EGADS_SUCCESS) Bref = (double) reals[0];

	  status = EG_attributeRet(bodies[i], "capsReferenceX", &atype, &alen, &ints, &reals, &string);
	  if (status == EGADS_SUCCESS) Xref = (double) reals[0];

	  status = EG_attributeRet(bodies[i], "capsReferenceY", &atype, &alen, &ints, &reals, &string);
	  if (status == EGADS_SUCCESS) Yref = (double) reals[0];

	  status = EG_attributeRet(bodies[i], "capsReferenceZ", &atype, &alen, &ints, &reals, &string);
	  if (status == EGADS_SUCCESS) Zref = (double) reals[0];
    }

    fprintf(fp, "CAPS generated Configuration\n");
    fprintf(fp, "0.0\n");                              /* Mach */
    fprintf(fp, "0 0 0.0\n");                          /* IYsym   IZsym   Zsym */
    fprintf(fp, "%lf %lf %lf\n", Sref, Cref, Bref); /* Sref    Cref    Bref */
    fprintf(fp, "%lf %lf %lf\n", Xref, Yref, Zref); /* Xref    Yref    Zref */
    fprintf(fp, "%lf\n", aimInputs[6].vals.real);         /* CDp */
  
    // Write out the Surfaces, one at a time
    for (surf = 0; surf < numAVLSurface; surf++) {

        printf("Writing surface - %s (id = %d)\n", avlSurface[surf].name, surf);

        if (avlSurface[surf].numSection < 2) {
            printf("Surface %s only has %d Sections - it will be skipped!\n", avlSurface[surf].name,
                                                                              avlSurface[surf].numSection);
            continue;
        }

        fprintf(fp, "#\nSURFACE\n%s\n%d %lf\n\n", avlSurface[surf].name,
                                                  avlSurface[surf].Nchord,
                                                  avlSurface[surf].Cspace);

        if  (avlSurface[surf].compon != 0)
            fprintf(fp, "COMPONENT\n%d\n\n", avlSurface[surf].compon);

        if  (avlSurface[surf].iYdup  != 0)
            fprintf(fp, "YDUPLICATE\n0.0\n\n");

        if  (avlSurface[surf].nowake == (int) true) fprintf(fp, "NOWAKE\n");
        if  (avlSurface[surf].noalbe == (int) true) fprintf(fp, "NOALBE\n");
        if  (avlSurface[surf].noload == (int) true) fprintf(fp, "NOLOAD\n");

        if ((avlSurface[surf].nowake == (int) true) ||
            (avlSurface[surf].noalbe == (int) true) ||
            (avlSurface[surf].noload == (int) true)) {

            fprintf(fp,"\n");
        }
  
        // Write the sections for each surface
        for (i = 0; i < avlSurface[surf].numSection; i++) {

            section = avlSurface[surf].vlmSection[i].sectionIndex;

            printf("\t Section %d (id = %d)\n", i, section);

            //
            writeSection(fp, bodies[avlSurface[surf].vlmSection[section].bodyIndex],
                             avlSurface[surf].vlmSection[section].nodeIndexLE,
                             avlSurface[surf].vlmSection[section].xyzLE,
							 avlSurface[surf].Nspan,
							 avlSurface[surf].Sspace);
      
            // Write control information for each section
            for (control = 0; control < avlSurface[surf].vlmSection[section].numControl; control++) {

            	printf("\t\t Control %d\n", control);

            	fprintf(fp, "CONTROL\n");
            	fprintf(fp, "%s ", avlSurface[surf].vlmSection[section].vlmControl[control].name);

            	fprintf(fp, "%f ", avlSurface[surf].vlmSection[section].vlmControl[control].controlGain);

            	if (avlSurface[surf].vlmSection[section].vlmControl[control].leOrTe == 0) { // Leading edge (-)

            		fprintf(fp, "%f ", -avlSurface[surf].vlmSection[section].vlmControl[control].percentChord);

            	} else { // Trailing edge (+)

            		fprintf(fp, "%f ", avlSurface[surf].vlmSection[section].vlmControl[control].percentChord);
            	}

            	fprintf(fp, "%f %f %f ", avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[0],
            							 avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[1],
										 avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[2]);

            	fprintf(fp, "%f\n", (double) avlSurface[surf].vlmSection[section].vlmControl[control].deflectionDup);
            }
            fprintf(fp, "\n");
        }
    }

    fclose(fp);
    fp = NULL;
  
    // Restore the path we came in with and get out
    chdir(currentPath);

    status = CAPS_SUCCESS;

    goto cleanup;

    cleanup:

		if (status != CAPS_SUCCESS) printf("avl/preAnalysis status = %d", status);
        if (fp != NULL) fclose(fp);

        // Free array of control name strings
        if (numControlName != 0 && controlName != NULL){
            (void) string_freeArray(numControlName, &controlName);
        }

        // Attribute to index map
        (void ) destroy_mapAttrToIndexStruct(&attrMap);

        // Destroy avlSurfaces
        if (avlSurface != NULL) {
            for (i = 0; i < numAVLSurface; i++) {
                status = destroy_vlmSurface(&avlSurface[i]);
                if (status != CAPS_SUCCESS) printf("destroy_vlmSurface status = %d\n", status);
            }
        }

        if (avlSurface != NULL) EG_free(avlSurface);
        numAVLSurface = 0;

        // Destroy avlControl
        if (avlControl != NULL) {
        	for (i = 0; i < numAVLControl; i++) {
        		status = destroy_vlmControl(&avlControl[i]);
        		if (status != CAPS_SUCCESS) printf("destroy_vlmControl status = %d\n", status);
        	}
        }

        if (avlControl != NULL) EG_free(avlControl);
        numAVLControl = 0;

        return status;
}


int
aimOutputs(/*@unused@*/ int inst, int index, char **aoname, capsValue *form)
{

	/*! \page aimOutputsAVL AIM Outputs
	 * Optional outputs that echo the inputs.  These are parsed from the resulting output and can be used as a sanity check.
	 */

#ifdef DEBUG
  printf(" avlAIM/aimOutputs instance = %d  index = %d!\n", inst, index);
#endif

  // ECHO AVL FLOW CONDITIONS
  if (index == 1) {
      *aoname = EG_strdup("Alpha");
      form->type = Double;

      /*! \page aimOutputsAVL
       * - <B> Alpha = </B> Angle of attack.
       */
  } else if (index == 2) {
      *aoname = EG_strdup("Beta");
      form->type = Double;

      /*! \page aimOutputsAVL
       * - <B> Beta = </B> Sideslip angle.
       */
  } else if (index == 3) {
      *aoname = EG_strdup("Mach");
      form->type = Double;

      /*! \page aimOutputsAVL
       * - <B> Mach = </B> Mach number.
       */

  } else if (index == 4) {
      *aoname = EG_strdup("pb/2V");
      form->type = Double;

      /*! \page aimOutputsAVL
       * - <B> pb/2V = </B> Non-dimensional roll rate.
       */
  } else if (index == 5) {
      *aoname = EG_strdup("qc/2V");
      form->type = Double;

      /*! \page aimOutputsAVL
       * - <B> qc/2V =</B> Non-dimensional pitch rate.
       */
  } else if (index == 6) {
      *aoname = EG_strdup("rb/2V");
      form->type = Double;

      /*! \page aimOutputsAVL
       * - <B> rb/2V = </B> Non-dimensional yaw rate.
       */

  } else if (index == 7) {
      *aoname = EG_strdup("p'b/2V");
      form->type = Double;

      /*! \page aimOutputsAVL
       * - <B> p'b/2V = </B> Non-dimensional roll acceleration.
       */

  } else if (index == 8) {
      *aoname = EG_strdup("r'b/2V");
      form->type = Double;

      /*! \page aimOutputsAVL
       * - <B> r'b/2V = </B> Non-dimensional yaw acceleration.
       */

  // BODY CONTROL DERIVATIVES
  } else if (index == 9) {
	  *aoname = EG_strdup("CXtot");
	  form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * Calculated outputs.  See AVL documentation for a description of each quantity.
       * - <B> CXtot </B>
       */

  } else if (index == 10) {
	  *aoname = EG_strdup("CYtot");
	  form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> CYtot </B>
	   */
  } else if (index == 11) {
      *aoname = EG_strdup("CZtot");
      form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> CZtot </B>
	   */
  } else if (index == 12) {
      *aoname = EG_strdup("Cltot");
      form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> Cltot </B>
	   */
  } else if (index == 13) {
      *aoname = EG_strdup("Cmtot");
      form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> Cmtot </B>
	   */
  } else if (index == 14) {
      *aoname = EG_strdup("Cntot");
      form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> Cntot </B>
	   */
  } else if (index == 15) {
      *aoname = EG_strdup("Cl'tot");
      form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> Cl'tot </B>
	   */
  } else if (index == 16) {
      *aoname = EG_strdup("Cn'tot");
      form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> Cn'tot </B>
	   */
  } else if (index == 17) {
	  *aoname = EG_strdup("CLtot");
	  form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> CLtot </B>
	   */
  } else if (index == 18) {
	  *aoname = EG_strdup("CDtot");
	  form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> CDtot </B>
	   */
  } else if (index == 19) {
	  *aoname = EG_strdup("CDvis");
	  form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> CDvis </B>
	   */
  } else if (index == 20) {
	  *aoname = EG_strdup("CLff");
	  form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> CLff </B>
	   */
  } else if (index == 21) {
	  *aoname = EG_strdup("CYff");
	  form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> CYff </B>
	   */
  } else if (index == 22) {
	  *aoname = EG_strdup("CDind");
	  form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> CDind </B>
	   */
  } else if (index == 23) {
	  *aoname = EG_strdup("CDff");
	  form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> CDff </B>
	   */
  } else if (index == 24) {
	  *aoname = EG_strdup("e");
	  form->type = Double;

	  /*! \page aimOutputsAVL
	   *
	   * - <B> e </B>
	   */
  }


  return CAPS_SUCCESS;
}


int aimCalcOutput(/*@unused@*/ int inst, /*@unused@*/ void *aimInfo,
              const char *analysisPath, int index, capsValue *val, capsErrs **errors)
{
	size_t     linecap = 0;
	char       currentPath[PATH_MAX], *str, *valstr, *line = NULL;
	FILE       *fp;

	#ifdef DEBUG
		int        stat;
		const char *name;
  
		stat = aim_getName(aimInfo, index, ANALYSISOUT, &name);
		printf(" avlAIM/aimCalcOutput instance = %d  index = %d  %s %d!\n", inst, index, name, stat);
	#endif

	*errors        = NULL;
	val->vals.real = 0.0;

	switch (index) {
	case 1:
		str = "Alpha =";
		break;
	case 2:
		str = "Beta  =";
		break;
	case 3:
		str = "Mach  =";
		break;
	case 4:
		str = "pb/2V =";
		break;
	case 5:
		str = "qc/2V =";
		break;
	case 6:
		str = "rb/2V =";
		break;
	case 7:
		str = "p'b/2V =";
		break;
	case 8:
		str = "r'b/2V =";
		break;
	case 9:
		str = "CXtot =";
		break;
	case 10:
		str = "CYtot =";
		break;
	case 11:
		str = "CZtot =";
		break;
	case 12:
		str = "Cltot =";
		break;
	case 13:
		str = "Cmtot =";
		break;
	case 14:
		str = "Cntot =";
		break;
	case 15:
		str = "Cl'tot =";
		break;
	case 16:
		str = "Cn'tot =";
		break;
	case 17:
		str = "CLtot =";
		break;
	case 18:
		str = "CDtot =";
		break;
	case 19:
		str = "CDvis =";
		break;
	case 20:
		str = "CLff  =";
		break;
	case 21:
		str = "CYff  =";
		break;
	case 22:
		str = "CDind =";
		break;
	case 23:
		str = "CDff  =";
		break;
	case 24:
		str = "e =";
		break;
	}

	/* get where we are and set the path to our output */
	(void) getcwd(currentPath, PATH_MAX);
	if (chdir(analysisPath) != 0) {

		#ifdef DEBUG
			printf(" avlAIM/aimCalcOutput Cannot chdir to %s!\n", analysisPath);
		#endif

		return CAPS_DIRERR;
	}

	/* open the AVL output file */
	fp = fopen("avlOutput.txt", "r");
	if (fp == NULL) {
		#ifdef DEBUG
			printf(" avlAIM/aimCalcOutput Cannot open Output file!\n");
		#endif

		chdir(currentPath);

		return CAPS_DIRERR;
	}

	/* scan the file for the string */
	valstr = NULL;
	/*@-unrecog@*/
	while (getline(&line, &linecap, fp) >= 0) {
		/*@+unrecog@*/
		if (line == NULL) continue;
		valstr = strstr(line, str);
		if (valstr != NULL) {
			/* found it -- get the value */
			sscanf(&valstr[7], "%lf", &val->vals.real);
			break;
		}
	}

	/* restore the path we came in with */
	chdir(currentPath);
	fclose(fp);
	if (line != NULL) free(line);
  
	if (valstr == NULL) {
		#ifdef DEBUG
			printf(" avlAIM/aimCalcOutput Cannot find %s in Output file!\n", str);
		#endif
  
		return CAPS_NOTFOUND;
	}

	return CAPS_SUCCESS;
}

/* 
 * since this AIM does not support field variables or CAPS bounds, the other
 * AIM functions are not required to be filled in except for aimFreeDiscr
 * that does some cleanup when CAPS terminates
 */

int aimFreeDiscr(/*@null@*/ capsDiscr *discr)
{
  
	#ifdef DEBUG
		if (discr == NULL) {
			printf(" avlAIM/aimFreeDiscr w/ NULL!\n");
		} else {
			printf(" avlAIM/aimFreeDiscr instance = %d!\n", discr->instance);
		}
	#endif
  
	if (discr == NULL) {
    
		/* signal for complete AIM cleanup */
		nInstance = 0;
    }
  
	return CAPS_SUCCESS;
}
