/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             FRICTION AIM
 *
 *      Copyright 2014-2016, Massachusetts Institute of Technology
 *      Licensed under The GNU Lesser General Public License, version 2.1
 *      See http://www.opensource.org/licenses/lgpl-2.1.php
 *
 */

#include <string.h>
#include <math.h>

#include "aimUtil.h"
#include "miscUtils.h"

#ifdef WIN32
#define getcwd   _getcwd
#define PATH_MAX _MAX_PATH
#else
#include <unistd.h>
#include <limits.h>
#endif

#define NUMINPUT  3
#define NUMOUT    3
#define PI        3.1415926535897931159979635
#define NINT(A)         (((A) < 0)   ? (int)(A-0.5) : (int)(A+0.5))

//#define DEBUG

/*!\mainpage Introduction
 *
 * \section overviewFRICTION FRICTION AIM Overview
 * FRICTION provides an estimate of laminar and turbulent skin friction and form
 * drag suitable for use in aircraft preliminary design. The program has its roots
 * in a program by Ron Hendrickson at Grumman. It runs on any computer. The input
 * requires geometric information and either the Mach and altitude combination, or
 * the Mach and Reynolds number at which the results are desired. It uses standard
 * flat plate skin friction formulas. The compressibility effects on skin friction
 * are found using the Eckert Reference Temperature method for laminar flow and the
 * van Driest II formula for turbulent flow. The basic formulas are valid from
 * subsonic to hypersonic speeds, but the implementation makes assumptions that limit
 * the validity to moderate supersonic speeds (about Mach 3). The key assumption is
 * that the vehicle surface is at the adiabatic wall temperature (the user can easily
 * modify this assumption). Form factors are used to estimate the effect of thickness
 * on drag, and a composite formula is used to include the effect of a partial run of
 * laminar flow.
 *
 * \section assumptionsFRICTION Assumptions
 *
 *
 * Related pages:
 * - \ref attributeFRICTION
 * - \ref aimInputsFRICTION
 * - \ref aimOutputsFRICTION
 *
 */

/*! \page attributeFRICTION Attribute Use for the construction of frictionInput.txt
 * The following list of attributes drives the FRICTION geometric definition. Aircraft components are defined as cross sections
 * in the low fidelity geometry definition. To be able to logically group the cross sections into wings, tails, fuselage, etc
 * they must be given a grouping attribute. This attribute defines a logical group along with identifying a set of cross sections
 * as a lifting surface or a body of revolution. The format is as follows.
 *
 *  - <b> capsGroup</b> This string attribute labels the <em> FaceBody</em> as to which type the section
 *  is assigned. This information is also used to logically group sections together to create wings, tails, stores, etc. Because
 *  AWAVE is relatively rigid <b> capsGroup </b> attributes must be on of the following items:
 *
 *    <em> Lifting Surfaces: </em> Wing, Tail, HTail, VTail, Horizontal_Tail, Vertical_Tail, Cannard
 *
 *    <em> Body of Revolution: </em> Fuselage, Fuse, Store
 *
 *
 *
 *  - <b> capsFidelity</b> This attribute is a CAPS requirement to indicate the analysis fidelity the geometry
 *  representation supports. Options are:  ALL, LINEARAERO
 *
 *  - <b> capsReferenceArea</b>  [Optional: Default 1.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the SREF entry in the AVL input.
 *
 */


typedef struct {
    char *name;
    const char *attribute;
    int type; // 0 lifting surface, 1 body of revolution
    double chordLength;
    double arcLength;
    double thickOverChord;
    double xyz[3]; // leading edge
    double xte[3]; // trailing edge
} aimSurface;

typedef struct {
    const char *name;
    double swet;
    double refLength;
    double thickOverChord; // thickness to chord ratio
    double type; // (0.0) planar surface (1.0) body
    double turbTrans; // 0.0 fully turbulent (default) 1.0 laminar
} frictionSec;

static int        nInstance  = 0;

extern /*@null@*/ /*@only@*/
    char       *EG_strdup(/*@null@*/ const char *str);



/* ****************** FRICTION AIM Helper Functions ************************ */

#ifdef WIN32
static int
getline(char **linep, size_t *linecapp, FILE *stream)
{
  int  i, cnt = -1;
  char buffer[1026];
  
  *linep    = buffer;
  *linecapp = 1024;
  while (cnt < 1024) {
    i = fread(&buffer[cnt+1], sizeof(char), 1, stream);
    if (i == 0) return cnt;
    if ((buffer[cnt+1] == 12) || (buffer[cnt+1] == 15) ||
        (buffer[cnt+1] == 0)) break;
    cnt++;
  }
  buffer[cnt+1] = 0;
  if (cnt == -1) cnt++;
  return cnt;
}
#endif

/*@+bufferoverflowhigh@*/

// ********************** AIM Function Break *****************************
static void
cross(double *A, double *B, double *C)
{
    
    C[0] = A[1]*B[2] - A[2]*B[1];
    C[1] = A[2]*B[0] - A[0]*B[2];
    C[2] = A[0]*B[1] - A[1]*B[0];
    
}
// ********************** AIM Function Break *****************************
static void
distance(double *x1, double *x2, double *x0, double* D)
{
    
    // (x1)-------------(x2)
    //             |D
    //             (x0)
    
    double a[3], b[3], tmp[3], length;

    length = sqrt((x1[0]-x2[0])*(x1[0]-x2[0]) + (x1[1]-x2[1])*(x1[1]-x2[1]) + (x1[2]-x2[2])*(x1[2]-x2[2]));
    if (length < 1.0e-8) {
    	x1[0] = x1[0] - 1.0;
    }
    
    a[0] = x2[0] - x1[0];
    b[0] = x1[0] - x0[0];
    a[1] = x2[1] - x1[1];
    b[1] = x1[1] - x0[1];
    a[2] = x2[2] - x1[2];
    b[2] = x1[2] - x0[2];
    cross(a,b,tmp);
    *D = sqrt(tmp[0]*tmp[0] + tmp[1]*tmp[1] + tmp[2]*tmp[2]) / sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2]);
    
}

// ********************** AIM Function Break *****************************
static int
findSectionData(ego body, int *nle, double *xle, int *nte, double *xte,
                double *chordLength, double *arcLength, double *thickOverChord)
{
    //Input to this function is an egads body object
    //Find leading edge location x,y,z
    //Find trailing edge location x,y,z
    //Find chord length of section (distance between two points),
    //      both should have same y value if an egads body representing
    //      a cross section || to the yplane is entered
    //Find arc length around the body
    
    int    i, n, stat, nNode, nEdge, oclass, mtype, *sens,index;
    double xmin, xyz[3], xmax, box[6], data[4];
    double massData[14];
    // double xyzm1[3], xyz0[3];
    double thickness;
    ego    ref, *nodes, *objs, *edges;
    
    *nle = 0;
    *nte = 0;
    *arcLength = 0.0;
    
    // check body type, looking for a NODE
    stat = EG_getTopology(body, &ref, &oclass, &mtype, data, &n, &objs, &sens);
    if (stat != EGADS_SUCCESS) {
        printf(" FRICTION AIM Warning: EG_getTopology Failure in findSectionData Code :: %d\n", stat);
        return CAPS_IOERR;
    } else {
        printf(" EGADS Body EG_getTopology Class :: %d :: ",oclass);
        switch (oclass) {
            case NODE:
                printf("NODE\n");
                break;
            case EDGE:
                printf("EDGE\n");
                break;
            case LOOP:
                printf("LOOP\n");
                break;
            case FACE:
                printf("FACE\n");
                break;
            case SHELL:
                printf("SHELL\n");
                break;
            case BODY:
                printf("BODY\n");
                break;
            default:
                break;
        }
    }
    
    if (oclass == NODE) {
        *nle = 0;
        *nte = 0;
        *chordLength = 0.0;
        *arcLength = 0.0;
        *thickOverChord = 0.0;
        xle[0] = xte[0] = data[0];
        xle[1] = xte[1] = data[1];
        xle[2] = xte[2] = data[2];
        return CAPS_SUCCESS;
    }
    
    // assume the LE position is the most forward Node in X
    // assume the TE position is the most rearward Node in X
    
    // Get Nodes from EGADS body, input to this function
    stat = EG_getBodyTopos(body, NULL, NODE, &nNode, &nodes);
    if (stat != EGADS_SUCCESS) {
        printf(" FRICTION AIM Warning: LE getBodyTopos Nodes = %d\n", stat);
        return CAPS_IOERR;
    }
    
    stat = EG_getBoundingBox(body, box);
    if (stat != EGADS_SUCCESS) {
        printf(" FRICTION AIM Warning: LE getBoundingBox Nodes = %d\n", stat);
        EG_free(nodes);
        return CAPS_IOERR;
    }
    // Estimate thickness value
    thickness = sqrt((box[1]-box[4])*(box[1]-box[4]) + (box[2]-box[5])*(box[2]-box[5]));
    // Sort throught the list of nodes to determine the LE and TE nodes
    //printf("Nodes %d\n",nNode);
    index = 0;

    for (i = 0; i < nNode; i++) {
        stat = EG_getTopology(nodes[i], &ref, &oclass, &mtype, xyz,
                              &n, &objs, &sens);
        if (stat != EGADS_SUCCESS) continue;

        // LE search
        if (*nle == 0) {
            *nle = i+1;
            xmin = xyz[index];
        } else {
            if (xyz[index] < xmin) {
                *nle = i+1;
                xmin = xyz[index];
            }
        }
        // TE search
        if (*nte == 0) {
            *nte = i+1;
            xmax = xyz[index];
        } else {
            if (xyz[index] > xmax) {
                *nte = i+1;
                xmax = xyz[index];
            }
        }
    }
    
    if (*nle == 0) {
        EG_free(nodes);
        return CAPS_IOERR;
    }
    if (*nte == 0) {
      EG_free(nodes);
      return CAPS_IOERR;
    }

    // assign xyz location variables xle and xte
    EG_getTopology(nodes[*nle-1], &ref, &oclass, &mtype, xle, &n, &objs, &sens);
    EG_getTopology(nodes[*nte-1], &ref, &oclass, &mtype, xte, &n, &objs, &sens);
    // determine distance between LT and TE points
    *chordLength = sqrt((xle[0]-xte[0])*(xle[0]-xte[0]) +
                        (xle[1]-xte[1])*(xle[1]-xte[1]) +
                        (xle[2]-xte[2])*(xle[2]-xte[2]));

    if ( fabs(*chordLength) < 1.0e-8 ) {
    	*chordLength = box[3] - box[0];
    }
    
    *thickOverChord = thickness / *chordLength;
    EG_free(nodes);
    
    // determine the arc length around the body
    stat = EG_getBodyTopos(body, NULL, EDGE, &nEdge, &edges);
    if (stat != EGADS_SUCCESS) {
        printf(" FRICTION AIM Warning: LE getBodyTopos Edges = %d\n", stat);
        return CAPS_IOERR;
    }
    
    for (i = 0; i < nEdge; i++) {
        
        stat = EG_getMassProperties(edges[i], massData);
        
        if (stat != EGADS_SUCCESS) continue;
        // massData array population
        // volume, surface area (length), cg(3), iniria(9)
        *arcLength = *arcLength + massData[1];
        // printf("EDGE %d :: %8.6f Length\n",i,massData[1]);
        
    }
    EG_free(edges);
    
    return CAPS_SUCCESS;
    
}

/* ********************** Exposed AIM Functions ***************************** */


// ********************** AIM Function Break *****************************
int
aimInitialize(/*@unused@*/ int ngIn,  /*@unused@*/ /*@null@*/ capsValue *gIn,
              int *nFid, int **fids, int *nIn, int *nOut,
              int *nFields, char ***fnames, int **ranks)
{
    int *ints, ret;
    
#ifdef DEBUG
    printf("\n frictionAIM/aimInitialize   ngIn = %d!\n", ngIn);
#endif
    
    // Number of geometry fidelities that are required by the AIM and what they are
    *nFid    = 4;
    ints     = (int *) EG_alloc(*nFid*sizeof(int));
    if (ints == NULL) return EGADS_MALLOC;
    ints[0]  = ALL | ALL;
    ints[1]  = ALL | LINEARAERO;
    ints[2]  = FACEBODY | ALL;
    ints[3]  = FACEBODY | LINEARAERO;
    *fids    = ints;
    
    // specify the number of analysis input and out "parameters"
    *nIn     = NUMINPUT;    // Mach, Altitude
    *nOut    = NUMOUT;      // CDtotal, CDform, CDfric
    
    // specify the field variables this analysis can generate
    *nFields = 0;
    *ranks   = NULL;
    *fnames  = NULL;
    
    /* get the instance storage
    if (instances == NULL) {
        instances = (aimStorage *) EG_alloc(sizeof(aimStorage));
        if (instances == NULL) {
            EG_free(*fids);
            *fids   = NULL;
            return EGADS_MALLOC;
        }
    } else {
        tmp = (aimStorage *) EG_reall(instances, (nInstance+1)*sizeof(aimStorage));
        if (tmp == NULL) {
            EG_free(*fids);
            *fids   = NULL;
            return EGADS_MALLOC;
        }
        instances = tmp;
    }
    instances[nInstance].nbody  = 0;
    instances[nInstance].bodies = NULL;  */
    
    /* return an index for the instance generated */
    ret = nInstance;
    nInstance++;
    return ret;
}

// ********************** AIM Function Break *****************************
int
aimInputs(/*@unused@*/ int inst, /*@unused@*/ void *aimStruc, int index,
          char **ainame, capsValue *defval)
{

	/*! \page aimInputsFRICTION AIM Inputs
	 * The following list outlines the FRICTION inputs along with their default value available
	 * through the AIM interface.  All inputs to the FRICTION AIM are variable length arrays.
	 * <B> All inputs must be the same length </B>.
	 */


#ifdef DEBUG
  printf(" frictionAIM/aimInputs instance = %d  index = %d!\n", inst, index);
#endif
    
    printf("READING INPUTS\n");
    
    if (index == 1) {
        *ainame               = EG_strdup("Mach");
        defval->type          = Double;
        defval->lfixed        = Change;
        defval->sfixed        = Change;
        //defval->sfixed        = Fixed;
        defval->nullVal       = IsNull;
        defval->dim           = 1;

        /*! \page aimInputsFRICTION
		 * - <B> Mach = double </B> <br> OR
		 * - <B> Mach = [double, ... , double] </B> <br>
		 *  Mach number.
		 */

    } else if (index == 2) {
        *ainame               = EG_strdup("Altitude");
        defval->type          = Double;
        defval->lfixed        = Change;
        defval->sfixed        = Change;
        //defval->sfixed        = Fixed;
        defval->nullVal       = IsNull;
        defval->dim           = 1;
        defval->units         = EG_strdup("kft");

        /*! \page aimInputsFRICTION
		 * - <B> Altitude = double </B> <br> OR
		 * - <B> Altitude = [double, ... , double] </B> <br>
		 *  Altitude in units of kft number.
		 */

    } else if (index == 3) {
		*ainame               = EG_strdup("BoundaryLayerTransistionPercent");
		defval->type          = Double;
		defval->dim           = 1;
		defval->length        = 1;
		defval->nrow          = 1;
		defval->ncol          = 1;
		defval->units         = NULL;
		defval->lfixed        = Fixed;
		defval->vals.real 	  = 0.10;
    }
    
    /*! \page aimInputsFRICTION
	 * - <B> BoundaryLayerTransistionPercent = double [0.1 default] </B> <br>
	 * Laminar to Turbulent transition [0.0 turbulent to 1.0 laminar] location on all sections.
	 */

    printf("FINISHED INPUTS\n");
    
    return CAPS_SUCCESS;
}

// ********************** AIM Function Break *****************************
int
//aimPreAnalysis(int inst, void *aimStruc, int nbody, /*@null@*/ ego *bodies,
//               char *apath, /*@null@*/ capsValue *inputs, capsErrs **errs)
aimPreAnalysis(/*@unused@*/ int inst, /*@unused@*/ void *aimStruc,
               const char *apath, /*@null@*/ capsValue *inputs, capsErrs **errs)
{
    
    int         nbody, nsecrevfound;
    ego         *bodies;
    int         tmp, nle, nte, i, j, stat, atype, alen, nsec, nsecrev, *nsecrevid; //, *surfid;
    int         *nsecrevidreal = NULL, *nsecrevNewSec= NULL;
    double      xte[3], SREF, Sref, dist, refLength, refArea, cfid;
    //char        tmpSref[10],tmpInt[12];
    char        tmpInt[12];
    char * tmpSref;
    double      massData[14];
    int         firstEntry;
    aimSurface  *surfaces; // aimSurface - structure defined global
    frictionSec *secLift, *secBody = NULL; // structure defined global
    FILE        *fp;
/*@-unrecog@*/
    char         cpath[PATH_MAX];
/*@+unrecog@*/
    const int       *ints;
    const char      *lengthUnitsIn, *string;
    const double    *reals;
  
#ifdef DEBUG
    printf(" frictionAIM/aimPreAnalysis\n");
#endif
    
    *errs = NULL;
    Sref = 1.0;
    
    // get EGADS bodies
    stat = aim_getBodies(aimStruc, &nbody, &bodies);
    if (stat != CAPS_SUCCESS) {
        printf(" avlAIM/aimPreAnalysis getBodies = %d!\n", stat);
        return stat;
    }
    
    // save away the current geometry for the instance
    // instances - aimStorage structure defined in global
    // inst - function input
    // inputs - caps value object
/*
    if (instances[inst].bodies != NULL) EG_free(instances[inst].bodies);
    instances[inst].bodies = NULL;
    instances[inst].nbody  = 0;
    if ((nbody > 0) && (bodies != NULL)) {
        instances[inst].bodies = (ego *) EG_alloc(nbody*sizeof(ego));
        if (instances[inst].bodies == NULL) {
#ifdef DEBUG
            printf(" frictionAIM/aimPreAnalysis Cannot allocate %d bodies!\n", nbody);
#endif
            return EGADS_MALLOC;
        }
        for (i = 0; i < nbody; i++) instances[inst].bodies[i] = bodies[i];
        instances[inst].nbody = nbody;
    }
*/
    if (inputs == NULL) {
#ifdef DEBUG
        printf(" frictionAIM/aimPreAnalysis inputs == NULL!\n");
#endif
        return CAPS_NULLVALUE;
    }
    if ((nbody <= 0) || (bodies == NULL)) {
#ifdef DEBUG
        printf(" frictionAIM/aimPreAnalysis No Bodies!\n");
#endif
        return CAPS_SOURCEERR;
    }
    
    // alocate memory to surfaces structure
    surfaces = (aimSurface *) malloc(nbody*sizeof(aimSurface));
    if (surfaces == NULL) {
#ifdef DEBUG
        printf(" frictionAIM/aimPreAnalysis Cannot allocate %d surfaces!\n", nbody);
#endif
        return EGADS_MALLOC;
    }
    
    //surfid = (int *) malloc(nbody*sizeof(int));
    
    // define default inputs for each structure created
    for (i = 0; i < nbody; i++) {
        surfaces[i].name        = NULL;
        //surfaces[i].attribute   = NULL;
        surfaces[i].type = 0;
        surfaces[i].chordLength = 0.0;
        surfaces[i].arcLength   = 0.0;
        surfaces[i].thickOverChord = 0.0;
        surfaces[i].xyz[0]      = surfaces[i].xyz[1] = surfaces[i].xyz[2] = 0.0;
        surfaces[i].xte[0]      = surfaces[i].xte[1] = surfaces[i].xte[2] = 0.0;
    }
    
    // populate the surfaces structure with real information
    /* accumulate the Surface data */
    // determine number of friction sections
    nsec = 0;
    for (i = 0; i < nbody; i++) {
        
        // search for the "capsFidelity" attribute
        stat = EG_attributeRet(bodies[i], "capsFidelity", &atype, &alen, &ints, &reals, &string);
        if (stat != EGADS_SUCCESS) {
            cfid = 0;
        } else {
            cfid = NINT(reals[0]);
        }
        
        if (cfid != 13) {
            printf(" *** WARNING: frictionAIM: capsFidelity attribute is missing or not defined as 13\n");
        }
        
        // search for the "capsReferenceArea" attribute
        stat = EG_attributeRet(bodies[i], "capsReferenceArea", &atype, &alen, &ints, &reals, &string);
        if (stat == EGADS_SUCCESS) {
            printf(" awaveAIM/aimPreAnalysis EG_attributeRet capsReferenceArea = %8.6f\n", reals[0]);
            Sref = (double) reals[0];
        }
        
        // Determine the length units of the input
        stat = EG_attributeRet(bodies[i], "capsLength", &atype, &alen, &ints, &reals, &lengthUnitsIn);
        if (stat != EGADS_SUCCESS) {
            printf(" *** WARNING: frictionAIM: NO UNITS ASSIGNED *** capsLength is not set in csm input!\n");
        }
        
        // Determine type of body
        // Look for "capsGroup" attribute
        stat = EG_attributeRet(bodies[i], "capsGroup", &atype, &alen, &ints, &reals, &surfaces[i].attribute);
        if (stat != CAPS_SUCCESS) {
            printf(" *** WARNING: frictionAIM: EG_attributeRet(capsGroup) = %d for Body %d!\n",
                   stat, i);
            continue;
        }
        // Wing, Tail, VTail, HTail, Cannard are all lifting surfaces
        // Fuse, Fuselage, Store are all bodies of revolution
        
        if ((strcmp(surfaces[i].attribute,"Wing")    == 0) ||
            (strcmp(surfaces[i].attribute,"Tail")    == 0) ||
            (strcmp(surfaces[i].attribute,"HTail")   == 0) ||
            (strcmp(surfaces[i].attribute,"VTail")   == 0) ||
			(strcmp(surfaces[i].attribute,"Horizontal_Tail")   == 0) ||
			(strcmp(surfaces[i].attribute,"Vertical_Tail")   == 0) ||
            (strcmp(surfaces[i].attribute,"Cannard") == 0)) {
            stat = findSectionData(bodies[i], &nle, surfaces[i].xyz, &nte, xte, &surfaces[i].chordLength,
                        &surfaces[i].arcLength, &surfaces[i].thickOverChord);
            if (stat != CAPS_SUCCESS) {
                printf(" *** WARNING: frictionAIM: findSectionData = %d for Body %d!\n",
                       stat, i);
            } else {
                surfaces[i].xte[0] = xte[0];
                surfaces[i].xte[1] = xte[1];
                surfaces[i].xte[2] = xte[2];
                printf("LIFTING SURFACE, BODY ID %d IN INPUT UNITS %s\n",i, lengthUnitsIn);
                printf("\tXLE:   %8.6f %8.6f %8.6f\n",surfaces[i].xyz[0],surfaces[i].xyz[1],surfaces[i].xyz[2]);
                printf("\tXTE:   %8.6f %8.6f %8.6f\n",xte[0],xte[1],xte[2]);
                printf("\tCHORD: %8.6f\n",surfaces[i].chordLength);
                printf("\tARC:   %8.6f\n",surfaces[i].arcLength);
                printf("\tT/C:   %8.6f\n",surfaces[i].thickOverChord);
                printf("\tATTRIB:%s\n",surfaces[i].attribute);
            }
        } else if ((strcmp(surfaces[i].attribute,"Fuse")     == 0) ||
                   (strcmp(surfaces[i].attribute,"Fuselage") == 0) ||
                   (strcmp(surfaces[i].attribute,"Store")    == 0)) {
            surfaces[i].type = 1;
            stat = findSectionData(bodies[i], &nle, surfaces[i].xyz, &nte, xte, &surfaces[i].chordLength,
                            &surfaces[i].arcLength, &surfaces[i].thickOverChord);
            if (stat != CAPS_SUCCESS) {
                printf(" *** WARNING: frictionAIM: findSectionData = %d for Body %d!\n",
                       stat, i);
            } else {
                // volume, surface area (length), cg(3), iniria(9)
                stat = EG_getMassProperties(bodies[i], massData);
                if (stat != EGADS_SUCCESS) continue;
                surfaces[i].chordLength = 2.0 * sqrt(massData[1] / PI); // ~diameter of cross section
                // printf("MASSDATA:  %8.6f, %8.6f\n",massData[0], massData[1]);
                printf("BODY OF REVOLUTION, BODY ID %d IN INPUT UNITS %s\n",i, lengthUnitsIn);
                printf("\tARC:   %8.6f\n",surfaces[i].arcLength);
                printf("\tDIAM:  %8.6f\n",surfaces[i].chordLength);
                printf("\tATTRIB:%s\n",surfaces[i].attribute);
            }
        } else {
            printf(" *** WARNING: frictionAIM: capsGroup attribute missing for BODY %d\n",i);
            printf("    Options:  Wing, Tail, VTail, HTail, Cannard, Vertical_Tail,Horizontal_Tail  are all lifting surfaces\n");
            printf("              Fuse, Fuselage, Store are all bodies of revolution\n");
        }

        if ((i > 0) && (surfaces[i].type == 0)) {
            if (!strcmp(surfaces[i].attribute,surfaces[i-1].attribute)) nsec++;
        }
    }
    if (nsec == 0) {
        printf(" *** WARNING: frictionAIM: nsec = 0!\n");
        free(surfaces);
        return EGADS_NOTFOUND;
    }

    // alocate memory to surfaces structure
    secLift = (frictionSec *) malloc(nsec*sizeof(frictionSec));
    if (secLift == NULL) {
#ifdef DEBUG
        printf(" frictionAIM/aimPreAnalysis Cannot allocate %d sections!\n", nsec);
#endif
        free(surfaces);
        return EGADS_MALLOC;
    }
    
    for (i = 0; i < nsec; i++) {
        secLift[i].name        = NULL; //surfaces[surfid[i]].attribute;
        secLift[i].swet        = 0.0;
        secLift[i].refLength   = 0.0;
        secLift[i].thickOverChord         = 0.0;
        secLift[i].type        = 0.0;
        secLift[i].turbTrans   = inputs[2].vals.real;
    }
    
    nsec      = 0;
    nsecrev   = 0;
    nsecrevfound = 1;
    nsecrevid = (int *) malloc(nbody*sizeof(int));
    if (nsecrevid == NULL) {
#ifdef DEBUG
        printf(" frictionAIM/aimPreAnalysis Cannot allocate %d secrev!\n", nbody);
#endif
        free(secLift);
        free(surfaces);
        return EGADS_MALLOC;
    }
    // Bodies of revolution may have multiple bodies in a single section.  This method only puts two bodies together into a section
    // The counter nsecrev follows the revolution sections put together
    for (i = 0; i < nbody; i++) {
        nsecrevid[i] = 0;
        if (i > 0) {
            if (!strcmp(surfaces[i].attribute,surfaces[i-1].attribute)) {
                // LIFTING SURFACES
                if (surfaces[i].type == 0) {
                    secLift[nsec].name = surfaces[i].attribute;
                    secLift[nsec].thickOverChord = (surfaces[i].thickOverChord + surfaces[i-1].thickOverChord) / 2.0;
                    distance(surfaces[i-1].xyz, surfaces[i-1].xte, surfaces[i].xyz, &dist);
                    stat = aim_convert(aimStruc, lengthUnitsIn, dist, "ft", &dist);
                    if (stat != CAPS_SUCCESS) printf(" *** WARNING : frictionAIM : dist Unit conversion error\n");
                    refLength = (surfaces[i].chordLength + surfaces[i-1].chordLength) / 2.0;
                    refArea   = (dist * (surfaces[i].arcLength + surfaces[i-1].arcLength)) / 2.0;
                    stat = aim_convert(aimStruc, lengthUnitsIn, refLength, "ft", &secLift[nsec].refLength);
                    if (stat != CAPS_SUCCESS) printf(" *** WARNING : frictionAIM : refLength Unit conversion error\n");
                    stat = aim_convert(aimStruc, lengthUnitsIn, refArea, "ft", &secLift[nsec].swet);
                    if (stat != CAPS_SUCCESS) printf(" *** WARNING : frictionAIM : swet Unit conversion error\n");
                    nsec++;
                // BODY OF REVOLUTION
                } else if (surfaces[i].type == 1) {
                    // count the number of surfaces and note their location
                    nsecrev++;
                    nsecrevfound = 0;
                    nsecrevid[i] = 1;
                }
            }
        }
    }
    
    // determine location of body of revolution sections in the body list
    if (nsecrev > 0) {
        nsecrevidreal = (int *) malloc(nsecrev*sizeof(int));
        nsecrevNewSec = (int *) malloc(nsecrev*sizeof(int));
        if ((nsecrevidreal == NULL) || (nsecrevNewSec == NULL)) {
            if (nsecrevidreal != NULL) free(nsecrevidreal);
            if (nsecrevNewSec != NULL) free(nsecrevNewSec);
#ifdef DEBUG
            printf(" frictionAIM/aimPreAnalysis Cannot allocate %d nsecrev!\n", nsecrev);
#endif
            free(nsecrevid);
            free(secLift);
            free(surfaces);
            return EGADS_MALLOC;
        }
        nsecrev = 1;
        tmp = 0;
        firstEntry = 1;
        for (i = 0; i < nbody; i++) {
            if (nsecrevid[i] == 1) {
                nsecrevidreal[tmp] = i;
                nsecrevNewSec[tmp] = 0;
                if (firstEntry==0 & strcmp(surfaces[i].attribute,surfaces[i-1].attribute) != 0) {
                    // if the sections don't have the same attribute then another body has been found
                    nsecrev++;
                    nsecrevNewSec[tmp] = 1;
                }
                tmp++;
                firstEntry = 0;
            } else {
                firstEntry = 1;
            }
        }
        nsecrevNewSec[0] = 1;
        
        // allocate memory for revolution bodies
        secBody = (frictionSec *) malloc(nsecrev*sizeof(frictionSec));
        if (secBody == NULL) {
#ifdef DEBUG
            printf(" frictionAIM/aimPreAnalysis Cannot allocate %d secBody sections!\n", nsecrev);
#endif
            free(nsecrevidreal);
            free(nsecrevNewSec);
            free(nsecrevid);
            free(secLift);
            free(surfaces);
            return EGADS_MALLOC;
        }
        
        for (i = 0; i < nsecrev; i++) {
            secBody[i].name        = NULL; //surfaces[surfid[i]].attribute;
            secBody[i].swet        = 0.0;
            secBody[i].refLength   = 0.0;
            secBody[i].thickOverChord         = 0.0;
            secBody[i].type        = 1.0;
            secBody[i].turbTrans   = inputs[2].vals.real;
        }
        
        // tmp is the total number of surfaces that make up ALL of the body sections
        nsecrev = -1;
        for (i = 0; i < tmp; i++) {
            firstEntry = 0;
            if (nsecrevNewSec[i] == 1) {
                if (i > 0) { // last body section is finished, onto the next one
                    stat = aim_convert(aimStruc, lengthUnitsIn, SREF, "ft", &SREF);
                    if (stat != CAPS_SUCCESS) printf(" *** WARNING : frictionAIM : Unit conversion error\n");
                    secBody[nsecrev].thickOverChord = secBody[nsecrev].refLength / SREF;
                }
                nsecrev++;
                secBody[nsecrev].name = surfaces[nsecrevidreal[i]].attribute;
                SREF = surfaces[nsecrevidreal[i]].chordLength; // used as a placehold variable for max diameter
                firstEntry = 1;
            }
            if (firstEntry == 0) {
                dist = fabs(surfaces[nsecrevidreal[i]].xyz[0] - surfaces[nsecrevidreal[i-1]].xyz[0]); // aligned with flow direction X - global axis
                stat = aim_convert(aimStruc, lengthUnitsIn, dist, "ft", &dist);
                if (stat != CAPS_SUCCESS) printf(" *** WARNING : frictionAIM : Unit conversion error\n");
                secBody[nsecrev].refLength = secBody[nsecrev].refLength + dist; // keep adding on each length component to the body ref
                refArea   = (surfaces[nsecrevidreal[i]].arcLength + surfaces[nsecrevidreal[i-1]].arcLength);
                stat = aim_convert(aimStruc, lengthUnitsIn, refArea, "ft", &refArea);
                 if (stat != CAPS_SUCCESS) printf(" *** WARNING : frictionAIM : Unit conversion error\n");
                secBody[nsecrev].swet = secBody[nsecrev].swet + (dist * refArea) / 2.0;
                refLength = (surfaces[nsecrevidreal[i]].chordLength + surfaces[nsecrevidreal[i-1]].chordLength) / 2.0; // ref diameter
                stat = aim_convert(aimStruc, lengthUnitsIn, refLength, "ft", &refLength);
                if (stat != CAPS_SUCCESS) printf(" *** WARNING : frictionAIM : Unit conversion error\n");
                if (surfaces[nsecrevidreal[i]].chordLength > SREF) {
                    SREF = surfaces[nsecrevidreal[i]].chordLength;
                }
            }
            
            if (i==tmp-1) {
                stat = aim_convert(aimStruc, lengthUnitsIn, SREF, "ft", &SREF);
                if (stat != CAPS_SUCCESS) printf(" *** WARNING : frictionAIM : Unit conversion error\n");
                secBody[nsecrev].thickOverChord = SREF / secBody[nsecrev].refLength;
            }
            
        }
      
    }
    
    
    /* get where we are and set the path to our input */
    (void) getcwd(cpath, PATH_MAX);
    
    printf("\nCWD :: %s\n",cpath);
    printf("APATH :: %s\n",apath);
    
    if (chdir(apath) != 0) {
#ifdef DEBUG
        printf(" frictionAIM/aimPreAnalysis Cannot chdir to %s!\n", apath);
#endif
        if (secBody       != NULL) free(secBody);
        if (nsecrevidreal != NULL) free(nsecrevidreal);
        if (nsecrevNewSec != NULL) free(nsecrevNewSec);
        free(nsecrevid);
        free(secLift);
        for (i = 0; i < nbody; i++) EG_free(surfaces[i].name);
        free(surfaces);
        return CAPS_DIRERR;
    }

    //calc_Sref(bodies, nbody, &SREF);
    if (Sref != 1.0) {
    	stat = aim_convert(aimStruc, lengthUnitsIn, Sref, "ft", &Sref); // Area is length squared so do twice
    	stat = aim_convert(aimStruc, lengthUnitsIn, Sref, "ft", &Sref);
    	if (stat != CAPS_SUCCESS) printf(" *** WARNING : frictionAIM : SREF Unit conversion error\n");
    }
    
    //d2s8fld(SREF,0,tmpSref); // create char for SREF
    tmpSref = convert_doubleToString(Sref,8,0); // 8 spaces, left justified (0)
    
    printf("**************************\n");
    printf("NSEC %d NSECREV %d\n",nsec,nsecrev);
    printf("**************************\n");



    // create char for number of sections in the friction input (lifting + bodyOfRev)
/*@-bufferoverflowhigh@*/
    if (nsec+nsecrev+1-nsecrevfound >= 10) {
        sprintf(tmpInt,"%d.       ",nsec+nsecrev+1-nsecrevfound);
    } else {
        sprintf(tmpInt,"%d.        ",nsec+nsecrev+1-nsecrevfound);
    }
/*@+bufferoverflowhigh@*/
    
    // create input file for friction
    fp = fopen("frictionInput.txt","w");
    if (fp == NULL) {
      printf(" frictionAIM/aimPreAnalysis Cannot open frictionInput.txt!\n");
      if (secBody       != NULL) free(secBody);
      if (nsecrevidreal != NULL) free(nsecrevidreal);
      if (nsecrevNewSec != NULL) free(nsecrevNewSec);
      free(nsecrevid);
      free(secLift);
      for (i = 0; i < nbody; i++) EG_free(surfaces[i].name);
      free(surfaces);
      return CAPS_IOERR;
    }
    fprintf(fp,"CAPS Generated Friction Input File\n");
    fprintf(fp,"%s  1.0       %s0.0\n",tmpSref,tmpInt);
    if(tmpSref != NULL) EG_free(tmpSref);
    //fprintf(fp,"1234567890123456789012345678901234567890123456789012345678901234567890\n");
    
    // LIFTING SURFACES
    for (i = 0; i < nsec; i++) {
        // Component Name columns 1-16 w/ 4 spaces to 20
        fprintf(fp,"%s",secLift[i].name);
        stat = strlen(secLift[i].name);
        for (j = 0; j < 20-stat; j++) {
            fprintf(fp," ");
        }
        // SWET spaces 21-30
        //d2s8fld(secLift[i].swet,0,tmpSref);
        tmpSref = convert_doubleToString(secLift[i].swet,8,0); // 8 spaces, left justified (0)
        fprintf(fp,"%s  ",tmpSref);
        if(tmpSref != NULL) EG_free(tmpSref);
        // RefL spaces 31-40
        //d2s8fld(secLift[i].refLength,0,tmpSref);
        tmpSref = convert_doubleToString(secLift[i].refLength,8,0);
        fprintf(fp,"%s  ",tmpSref);
        if(tmpSref != NULL) EG_free(tmpSref);
        // ToC spaces 41-50
        //d2s8fld(secLift[i].thickOverChord,0,tmpSref);
        tmpSref = convert_doubleToString(secLift[i].thickOverChord,8,0);
        fprintf(fp,"%s  ",tmpSref);
        if(tmpSref != NULL) EG_free(tmpSref);
        // Component type 51-60
        //d2s8fld(secLift[i].type,0,tmpSref);
        tmpSref = convert_doubleToString(secLift[i].type,8,0);
        fprintf(fp,"%s  ",tmpSref);
        if(tmpSref != NULL) EG_free(tmpSref);
        // FTrans 61-70
        //d2s8fld(secLift[i].turbTrans,0,tmpSref);
        tmpSref = convert_doubleToString(secLift[i].turbTrans,8,0);
        fprintf(fp,"%s  ",tmpSref);
        if(tmpSref != NULL) EG_free(tmpSref);
        //
        fprintf(fp,"\n");
    }
    
    // BODIES OF REVOLUTION
    if (secBody != NULL)
        for (i = 0; i <= nsecrev; i++) {
            // Component Name columns 1-16 w/ 4 spaces to 20
            fprintf(fp,"%s",secBody[i].name);
            stat = strlen(secBody[i].name);
            for (j = 0; j < 20-stat; j++) {
                fprintf(fp," ");
            }
            // SWET spaces 21-30
            //d2s8fld(secBody[i].swet,0,tmpSref);
            tmpSref = convert_doubleToString(secBody[i].swet,8,0);
            fprintf(fp,"%s  ",tmpSref);
            if(tmpSref != NULL) EG_free(tmpSref);
            // RefL spaces 31-40
            //d2s8fld(secBody[i].refLength,0,tmpSref);
            tmpSref = convert_doubleToString(secBody[i].refLength,8,0);
            fprintf(fp,"%s  ",tmpSref);
            if(tmpSref != NULL) EG_free(tmpSref);
            // ToC spaces 41-50
            //d2s8fld(secBody[i].thickOverChord,0,tmpSref);
            tmpSref = convert_doubleToString(secBody[i].thickOverChord,8,0);
            fprintf(fp,"%s  ",tmpSref);
            if(tmpSref != NULL) EG_free(tmpSref);
            // Component type 51-60
            //d2s8fld(secBody[i].type,0,tmpSref);
            tmpSref = convert_doubleToString(secBody[i].type,8,0);
            fprintf(fp,"%s  ",tmpSref);
            if(tmpSref != NULL) EG_free(tmpSref);
            // FTrans 61-70
            //d2s8fld(secBody[i].turbTrans,0,tmpSref);
            tmpSref = convert_doubleToString(secBody[i].turbTrans,8,0);
            fprintf(fp,"%s  ",tmpSref);
            if(tmpSref != NULL) EG_free(tmpSref);
            //
            fprintf(fp,"\n");
        }
    
    if (inputs[0].length != inputs[1].length) {
        printf("WARNING :: frictionAIM PreAnalysis, Inputs Mach and Altitude must be the same length\n");
    }
    
    printf(" INPUT LENGTH %d\n",inputs[0].length);
    
    if (inputs[0].length == 1) {
        // MACH
        //d2s8fld(inputs[0].vals.real,0,tmpSref);
        tmpSref = convert_doubleToString(inputs[0].vals.real,8,0);
        fprintf(fp,"%s  ",tmpSref);
        if(tmpSref != NULL) EG_free(tmpSref);
        // ALTITUDE
        stat = aim_convert(aimStruc, inputs[1].units, inputs[1].vals.real, "kft", &SREF);
        if (stat != CAPS_SUCCESS) printf(" *** WARNING : frictionAIM : Unit conversion error\n");
        //d2s8fld(SREF,0,tmpSref);
        tmpSref = convert_doubleToString(SREF,8,0);
        fprintf(fp,"%s\n",tmpSref);
        if(tmpSref != NULL) EG_free(tmpSref);
    } else {
        for (i = 0; i < inputs[0].length; i++) { // Multiple Mach, Altitude pairs
            // MACH
            //d2s8fld(inputs[0].vals.reals[i],0,tmpSref);
            tmpSref = convert_doubleToString(inputs[0].vals.reals[i],8,0);
            fprintf(fp,"%s  ",tmpSref);
            if(tmpSref != NULL) EG_free(tmpSref);
            // ALTITUDE
            stat = aim_convert(aimStruc, inputs[1].units, inputs[1].vals.reals[i], "kft", &SREF);
            if (stat != CAPS_SUCCESS) printf(" *** WARNING : frictionAIM : Unit conversion error\n");
            //d2s8fld(SREF,0,tmpSref);
            tmpSref = convert_doubleToString(SREF,8,0);
            fprintf(fp,"%s\n",tmpSref);
            if(tmpSref != NULL) EG_free(tmpSref);
        }
    }
    
    
    fprintf(fp,"0.00      0.00\n");
    fclose(fp);
    
    
    chdir(cpath);

    if (secBody       != NULL) free(secBody);
    if (nsecrevidreal != NULL) free(nsecrevidreal);
    if (nsecrevNewSec != NULL) free(nsecrevNewSec);
    free(nsecrevid);
    free(secLift);
    free(surfaces);

    /*
        for (i = 0; i < nbody; i++) EG_free(surfaces[i].name);
        free(surfaces);
        for (i = 0; i <= nsecrev; i++) EG_free(secBody[i].name);
        free(secBody);
        for (i = 0; i < nsec; i++) EG_free(secLift[i].name);
        free(secLift);
     */
    return CAPS_SUCCESS;
}

// ********************** AIM Function Break *****************************
int
aimOutputs( int inst, int index, char **aoname, capsValue *form)
{
#ifdef DEBUG
  printf(" frictionAIM/aimOutputs instance = %d  index = %d!\n", inst, index);
#endif

/*! \page aimOutputsFRICTION AIM Outputs
* Total, Form, and Friction drag components
*/

    if (index == 1) {
        
        *aoname             = EG_strdup("CDtotal");
        form->type          = Double;
        form->units         = NULL;
        form->lfixed        = Change;
        form->sfixed        = Change;
        /*! \page aimOutputsFRICTION
         * - <B> CDtotal = </B> Drag Coefficient [CDform + CDfric].
         */
    } else if (index == 2) {
        
        *aoname             = EG_strdup("CDform");
        form->type          = Double;
        form->units         = NULL;
        form->lfixed        = Change;
        form->sfixed        = Change;
        /*! \page aimOutputsFRICTION
         * - <B> CDform = </B> Form Drag Coefficient.
         */
    } else if (index == 3) {
        
        *aoname             = EG_strdup("CDfric");
        form->type          = Double;
        form->units         = NULL;
        form->lfixed        = Change;
        form->sfixed        = Change;
    }
    /*! \page aimOutputsFRICTION
     * - <B> CDfric = </B> Friction Drag Coefficient.
     */
    return CAPS_SUCCESS;
}

// ********************** AIM Function Break *****************************
int
aimCalcOutput(int inst, /*@unused@*/ void *aimStruc, char *apath, int index,
              capsValue *val, capsErrs **errors)
{

    size_t linecap = 0;
    char   cpath[PATH_MAX], *str, *valstr, *line = NULL;
    double tmp, Fric[100], Prof[100], Tot[100];
    int stat, i, cnt = 0;
    FILE   *fp;
    
#ifdef DEBUG
    printf(" frictionAIM/aimCalcOutput instance = %d  index = %d!\n", inst, index);
#endif
    
    *errors        = NULL;
    //val->vals.real = 0.0;
    
    if (index == 1) {
        str = "SUMMARY";
    }
    
    /* get where we are and set the path to our output */
    (void) getcwd(cpath, PATH_MAX);
    if (chdir(apath) != 0) {
#ifdef DEBUG
        printf(" frictionAIM/aimCalcOutput Cannot chdir to %s!\n", apath);
#endif
        return CAPS_DIRERR;
    }
    
    /* open the friction output file */
    fp = fopen("frictionOutput.txt", "r");
    if (fp == NULL) {
#ifdef DEBUG
        printf(" frictionAIM/aimCalcOutput Cannot open Output file!\n");
#endif
        chdir(cpath);
        return CAPS_DIRERR;
    }
    
    /* scan the file for the string */
    valstr = NULL;
    /*@-unrecog@*/
    while (getline(&line, &linecap, fp) >= 0) {
        /*@+unrecog@*/
        if (line == NULL) continue;
        valstr = strstr(line, str);
        if (valstr != NULL) {
            /* found it -- get the value */
            printf("\n *** FRICTION SUMMARY FOUND ***\n");
            //move down three lines
            getline(&line, &linecap, fp);
            getline(&line, &linecap, fp);
            printf("HEAD :: %s\n",line);
            
            while (getline(&line, &linecap, fp) >= 0) {
                stat = sscanf(line,"%*d %*lf %*lf %*lf %*lf %*lf %lf",&tmp);
                if (stat == -1) break; // using the break to exit the while loop, not a read error
                Tot[cnt] = tmp;
                stat = sscanf(line,"%*d %*lf %*lf %*lf %*lf %lf %*lf",&tmp);
                Prof[cnt] = tmp;
                stat = sscanf(line,"%*d %*lf %*lf %*lf %lf %*lf %*lf",&tmp);
                Fric[cnt] = tmp;
                cnt++;
                printf("DATA :: %s\n",line);
            }
            
            break;
            
        }
    }

    if (cnt == 1) {
        val->length = 1;
        switch (index) {
            case 1:
                val->vals.real = Tot[0];
                printf("CAPS VALUE CDtotal %8.6f\n",Tot[0]);
                break;
            case 2:
                val->vals.real = Prof[0];
                printf("CAPS VALUE  CDform %8.6f\n",Prof[0]);
                break;
            case 3:
                val->vals.real = Fric[0];
                printf("CAPS VALUE  CDfric %8.6f\n",Fric[0]);
                break;
            default:
                return CAPS_NOTFOUND;
        }
    } else if (cnt > 1) {
        val->length = cnt;
        val->dim    = 1;
        val->nrow   = 1;
        val->ncol   = cnt;
        val->vals.reals = (double *) EG_alloc(val->length*sizeof(double));
        if (val->vals.reals == NULL) return EGADS_MALLOC;
        for (i = 0; i < cnt; i++) {
            switch (index) {
                case 1:
                    val->vals.reals[i] = Tot[i];
                    printf("CAPS VALUE CDtotal %8.6f\n", Tot[i]);
                    break;
                case 2:
                    val->vals.reals[i] = Prof[i];
                    printf("CAPS VALUE  CDform %8.6f\n", Prof[i]);
                    break;
                case 3:
                    val->vals.reals[i] = Fric[i];
                    printf("CAPS VALUE  CDfric %8.6f\n", Fric[i]);
                    break;
                default:
                    return CAPS_NOTFOUND;
            }
        }
    }
    
    /* restore the path we came in with */
    chdir(cpath);
    fclose(fp);
    if (line != NULL) free(line);
    
    if (cnt == 0) {
#ifdef DEBUG
        printf(" frictionAIM/aimCalcOutput Cannot find %s in Output file!\n", str);
#endif
        return CAPS_NOTFOUND;
    }

    return CAPS_SUCCESS;
}


// ********************** AIM Function Break *****************************
int
aimFreeDiscr(/*@null@*/ capsDiscr *discr)
{
#ifdef DEBUG
  if (discr == NULL) {
    printf(" frictionAIM/aimFreeDiscr w/ NULL!\n");
  } else {
    printf(" frictionAIM/aimFreeDiscr instance = %d!\n", discr->instance);
  }
#endif
  
    return CAPS_SUCCESS;
}
