/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             Friction AIM tester
 *
 *      Copyright 2014-2016, Massachusetts Institute of Technology
 *      Licensed under The GNU Lesser General Public License, version 2.1
 *      See http://www.opensource.org/licenses/lgpl-2.1.php
 *
 */

#include "caps.h"
#include "capsTypes.h"
#include <string.h>

#ifdef WIN32
#define unlink   _unlink
#define getcwd   _getcwd
#define PATH_MAX _MAX_PATH
#else
#include <limits.h>
#include <unistd.h>
#endif

// ******* VARIABLE LENGTH INPUT DEF *************
#define Ninp  2



int main(int argc, char *argv[])
{
  int            i, j, n, stat, fidelity, nFields, *ranks, dirty, nErr, nLines, dim;
  int            nPar;
  char           *apath, **fnames, **lines, *name, *lunits, endf[13];
  char           check[12] = "END OF CASE";
/*@-unrecog@*/
  char           cpath[PATH_MAX];
/*@+unrecog@*/
  capsObj        pobj, aobj, obj, link, parent, mobj, *par;
  capsErrs       *errors;
  capsOwn        current;
  ego            body;
  enum capsoType type;
  enum capssType subtype;
  FILE           *fp;
  double         Mach[Ninp], Alt[Ninp];


  if ((argc < 2) || (argc > 4)) {
    printf(" usage: frictionTest *.csm scratchdir\n");
    return 1;
  }
  
  // open csm file in CAPS
  stat = caps_open(argv[1], "frictionTest", &pobj);
  if (stat != CAPS_SUCCESS) {
    printf(" caps_open = %d\n", stat);
    return 1;
  }
  
  /* look at the bodies and report units */
  stat = caps_size(pobj, BODIES, NONE, &n);
  if (stat != CAPS_SUCCESS) {
    printf(" caps_size on Bodies = %d\n", stat);
  } else {
    for (i = 1; i <= n; i++) {
      stat = caps_bodyByIndex(pobj, i, &body, &lunits);
      if (stat != CAPS_SUCCESS) {
        printf(" caps_bodyByIndex = %d for Body %d!\n", stat, i);
      } else {
        printf(" Body %d has length units = %s\n", i, lunits);
      }
    }
  }
  
  stat = caps_load(pobj, "frictionAIM", argv[2], 200, 0, NULL, &aobj);
  if (stat != CAPS_SUCCESS) {
    printf(" caps_load = %d\n", stat);
    caps_close(pobj);
    return 1;
  }
    
  stat = caps_info(pobj, &name, &type, &subtype, &link, &parent, &current);
  if (stat != CAPS_SUCCESS) {
    printf(" caps_info on Problem = %d\n", stat);
    caps_close(pobj);
    return 1;
  }
  
    Mach[0] = 0.5;
    Mach[1] = 0.8;
    Alt[0] = 30.;
    Alt[1] = 60.;
 
    dim = 0;
    if (Ninp > 1) {
        dim = 1;
    }

    // Set Mach Inputs
    stat = caps_childByName(aobj, VALUE, ANALYSISIN, "Mach", &mobj);
    if (stat != CAPS_SUCCESS) {
        printf(" caps_childByName = %d\n", stat);
    } else {
        stat = caps_setValue(mobj, Ninp, 1, (void *) &Mach);
        if (stat != CAPS_SUCCESS) printf(" caps_setValue = %d\n", stat);
    }
    
    // Set Altitude Inputs
    stat = caps_childByName(aobj, VALUE, ANALYSISIN, "Altitude", &mobj);
    if (stat != CAPS_SUCCESS) {
        printf(" caps_childByName = %d\n", stat);
    } else {
        stat = caps_setValue(mobj, Ninp, 1, (void *) &Alt);
        if (stat != CAPS_SUCCESS) printf(" caps_setValue = %d\n", stat);
    }
      
  stat = caps_analysisInfo(aobj, &apath, &fidelity, &nPar, &par,
                           &nFields, &fnames, &ranks, &dirty);
  if (stat != CAPS_SUCCESS) {
    printf(" caps_analysisInfo = %d\n", stat);
    caps_close(pobj);
    return 1;
  }
  printf("\n Fidelity =  %d", fidelity);
  printf("\n APath    = %s", apath);
  printf("\n Fields   =");
  for (i = 0; i < nFields; i++) printf("  %s (%d)", fnames[i], ranks[i]);
  printf("\n Dirty    =  %d\n", dirty);

  /* do the analysis */
  if (dirty != 0) {
    stat = caps_preAnalysis(aobj, &nErr, &errors);
    if (stat != CAPS_SUCCESS)
      printf(" caps_preAnalysis = %d\n", stat);
    if (errors != NULL) {
      for (i = 0; i < nErr; i++) {
        stat = caps_errorInfo(errors, i+1, &obj, &nLines, &lines);
        if (stat != CAPS_SUCCESS) {
          printf(" CAPS Error: caps_errorInfo[%d] = %d!\n", i+1, stat);
          continue;
        }
        for (j = 0; j < nLines; j++) printf(" %s\n", lines[j]);
      }
      caps_freeError(errors);
    }
    
    dirty = 0;
    stat = caps_analysisInfo(aobj, &apath, &fidelity, &nPar, &par,
                               &nFields, &fnames, &ranks, &dirty);
    if (stat == CAPS_SUCCESS) printf("\n Dirty    =  %d\n", dirty);
  
    
       
    // execute friction and run the post
    if (dirty == 5) {
      
      (void) getcwd(cpath, PATH_MAX);
      if (chdir(apath) != 0) {
        printf(" ERROR(frictionTest.c): Cannot change directory to -> %s\n", apath);
        caps_close(pobj);
        return 1;
      }
      unlink("frictionOutput.txt");                     // remove any old output
      printf(" Running FRICTION!\n");
      // note: frictionInput.txt was generated by preAnalysis
      system("friction frictionInput.txt frictionOutput.txt");
      // check if ran correctly before invoking postAnalysis
      fp = fopen("frictionOutput.txt", "r");
      if (fp == NULL) {
        printf(" ERROR: Cannot open frictionOutput.txt!\n");
        dirty = -1;
      } else {
        fseek(fp, -14, SEEK_END);
        (void) fread(endf, sizeof(char), 11, fp);
        fclose(fp);
        endf[12] = 0;
          
          printf("endf == %s\n",endf);
          
        if (check[0] != endf[0]) {
        //if (strcmp(check,endf) != 0) {
          printf(" ERROR: End of frictionOutput.txt not correct!\n");
          dirty = -1;
        }
      }
      chdir(cpath);

      if (dirty != -1) {
        // run post if we are OK
        stat = caps_postAnalysis(aobj, current, &nErr, &errors);
        if (stat != CAPS_SUCCESS)
          printf(" caps_postAnalysis = %d\n", stat);
        if (errors != NULL) {
          for (i = 0; i < nErr; i++) {
            stat = caps_errorInfo(errors, i+1, &obj, &nLines, &lines);
            if (stat != CAPS_SUCCESS) {
              printf(" CAPS Error: caps_errorInfo[%d] = %d!\n", i+1, stat);
              continue;
            }
            for (j = 0; j < nLines; j++) printf(" %s\n", lines[j]);
          }
          caps_freeError(errors);
        }
      }
      dirty = 0;
      stat = caps_analysisInfo(aobj, &apath, &fidelity, &nPar, &par,
                                 &nFields, &fnames, &ranks, &dirty);
      if (stat == CAPS_SUCCESS) printf(" Dirty    =  %d\n", dirty);
    }
  }
  
  /* output what we have */
  printf("\n");
  printObjects(pobj, 0);
  printf("\n");
  
  /* checkpoint the Problem */
  if (argc == 4) {
    stat = caps_save(pobj, argv[3]);
    if (stat != CAPS_SUCCESS) printf(" caps_save = %d\n", stat);
  }

  stat = caps_close(pobj);
  if (stat != CAPS_SUCCESS) printf(" caps_close = %d\n", stat);
  
  return 0;
}
