/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             FUN3D AIM
 *
 *     Written by Dr. Ryan Durscher AFRL/RQVC
 *
 */


/*!\mainpage Introduction
 *
 * \section overviewFUN3D FUN3D AIM Overview
 * A module in the Computational Aircraft Prototype Syntheses (CAPS) has been developed to interact (primarily
 * through input files) with NASA LaRC's unstructured flow solver FUN3D. FUN3D is a parallelized flow analysis
 * and design suite capable of addressing a wide variety of complex aerodynamic configurations by utilizing a
 * mixed-element, node-based, finite volume discretization.
 * The suite can simulate perfect gas (both incompressible and compressible), as well as multi-species
 * equilibrium and non-equilibrium flows. Turbulence effects may be represented through a wide variety
 * of models. Currently only a subset of FUN3D's input options have been exposed in the analysis interface
 * module (AIM), but features can easily be included as future needs arise.
 *
 * Current issues include:
 *  - A thorough bug testing needs to be undertaken.
 *  - Not all parameters/variables in fun3d.nml are currently available.
 *  .
 *
 * \section fun3dNML Generating fun3d.nml
 * FUN3D's primarily input file is a master FORTRAN namelist, fun3d.nml. Currently, the FUN3D AIM relies on a
 * Python script to read and write this file. As such the AIM user needs Python installed so that the AIM may
 * be complied with the Python API library (and header file - Python.h). The logic behind using Python to
 * interact with the namelist file is simple, flexibility. Python allows dynamic creation of input variables,
 * while each option available in C would have to be explicit defined. This would essentially lock the AIM
 * into a particular version of FUN3D, as changes due to the addition and deprecation of variables would become
 * cumbersome. Currently, the FUN3D AIM will first try to read an existing fun3d.nml file in the specified
 * directory. If the file does not exist one will be created. Only modified input variables that have been
 * specified as AIM inputs (currently supported variables are outlined in \ref aimInputsFUN3D are
 * updated in the namelist file.
 *
 * Alternatively, if the AIM is not built against the Python API a bare-bones fun3d.nml is created using only
 * the variables set in \ref aimInputsFUN3D. For the fun3d.nml file to be created it is essential to set
 * the Overwrite\_NML input variable to True. This gives the user ample warning that their fun3d.nml (if it exist)
 * will be over written.
 *
 * Related pages:
 * - \ref cfdBoundaryConditions
 * - \ref geomFidelityFUN3D
 * - \ref aimInputsFUN3D
 * - \ref aimOutputsFUN3D
 * - \ref sharableDataFUN3D
 * - \ref dataTransferFUN3D
 */

#include <string.h>
#include <math.h>
#include "capsTypes.h"
#include "aimUtil.h"

#include "meshUtils.h"
#include "cfdUtils.h"
#include "miscUtils.h"
#include "fun3dUtils.h"


#ifdef HAVE_PYTHON
#include "Python.h" // Bring in Python API
#endif


#ifdef WIN32
#define getcwd     _getcwd
#define snprintf   _snprintf
#define strcasecmp stricmp
#define PATH_MAX   _MAX_PATH
#else
#include <unistd.h>
#include <limits.h>
#endif

/*
#define CROSS(a,b,c)      a[0] = (b[1]*c[2]) - (b[2]*c[1]);\
                          a[1] = (b[2]*c[0]) - (b[0]*c[2]);\
                          a[2] = (b[0]*c[1]) - (b[1]*c[0])
#define DOT(a,b)         (a[0]*b[0] + a[1]*b[1] + a[2]*b[2])
*/

#define NUMINPUT   29
#define NUMOUTPUT  24

#define MXCHAR  255

//#define DEBUG

typedef struct {

    // Fun3d project name
    char *projectName;

    // Attribute to index map
    mapAttrToIndexStruct attrMap;

    // Inherited surface mesh mapping
   // genUnstrSurfMeshStruct *surfaceMesh;

	// Flag to see if we need to get a CFD mesh
   int haveVolumeMesh;

    // Check to make sure data transfer is ok
    int dataTransferCheck;

    //int dataTransferBodyIndex; // This aim will only currently work if one body is set to being transfered

    // Analysis file path/directory
    char *analysisPath;

    // Pointer to caps input value for scaling pressure during data transfer
    capsValue *pressureScaleFactor;

    // Pointer to caps input value for offset pressure during data transfer
    capsValue *pressureScaleOffset;

} aimStorage;

static aimStorage *fun3dInstance = NULL;
static int         numInstance  = 0;


extern /*@null@*/ /*@only@*/
	char       *EG_strdup(/*@null@*/ const char *str);
extern  int         EG_inTriExact(double *t1, double *t2, double *t3,
                                  double *p,  double *w );



/* ********************** Exposed AIM Functions ***************************** */

int aimInitialize(/*@unused@*/ int ngIn, /*@unused@*/ /*@null@*/ capsValue *gIn,
                  int *nFid, int **fids, int *nIn, int *nOut,
                  int *nFields, char ***fnames, int **ranks)
{

    int status; // Function status return

    int *ints;

    aimStorage *tmp;

    #ifdef DEBUG
        printf("\n fun3dAIM/aimInitialize   ngIn = %d!\n", ngIn);
    #endif

    *nFid    = 5;
    ints     = (int *) EG_alloc(((int) *nFid)*sizeof(int));
    if (ints == NULL) return EGADS_MALLOC;

    /*! \page geomFidelityFUN3D Geometry Fidelity
    * The geometric fidelity for the FUN3D AIM requires that the body(ies) be either a face body(ies) (FACEBODY), solid
    * body(ies) (SOLIDBODY) or non-manifold sheet body(ies) (SHEETBODY). For 2D simulations the attribute capsFidelity should
    * be set to CFD with the geometric fidelity being either a planar SHEETBODY or FACEBODY, while for 3D simulations
    * it should again be set to CFD but the both SOLIDBODY(ies) or SHEETBODY(ies) are acceptable.
    */
    ints[0]  = ALL | ALL;
    ints[1]  = ALL | CFD;
    ints[2]  = SOLIDBODY | CFD;
    ints[3]  = SHEETBODY | CFD;
    ints[4]  = FACEBODY | CFD;

    *fids    = ints;

    /* specify the number of analysis input and out "parameters" */
    *nIn     = NUMINPUT;
    *nOut    = NUMOUTPUT;

    /* specify the field variables this analysis can generate */
    *nFields = 0;
    *ranks   = NULL;
    *fnames  = NULL;

    // Allocate fun3dInstance
    if (fun3dInstance == NULL) {

    	fun3dInstance = (aimStorage *) EG_alloc(sizeof(aimStorage));

    	if (fun3dInstance == NULL) {
    		EG_free(*fids);
    		*fids   = NULL;
    		return EGADS_MALLOC;
    	}

    } else {
    	tmp = (aimStorage *) EG_reall(fun3dInstance, (numInstance+1)*sizeof(aimStorage));
    	if (tmp == NULL) {
    		EG_free(*fids);
    		*fids   = NULL;
    		return EGADS_MALLOC;
    	}

    	fun3dInstance = tmp;
    }

    // Set initial values for fun3dInstance
    fun3dInstance[numInstance].projectName = NULL;

    // Analysis file path/directory
    fun3dInstance[numInstance].analysisPath = NULL;

    fun3dInstance[numInstance].haveVolumeMesh = (int) false;

    // Container for surface mesh pointer
    //fun3dInstance[numInstance].surfaceMesh = NULL;

    // Container for attribute to index map
    status = initiate_mapAttrToIndexStruct(&fun3dInstance[numInstance].attrMap);
    if (status != CAPS_SUCCESS) return status;

    // Check to make sure data transfer is ok
    fun3dInstance[numInstance].dataTransferCheck = (int) true;

    // Pointer to caps input value for scaling pressure during data transfer
    fun3dInstance[numInstance].pressureScaleFactor = NULL;

    // Pointer to caps input value for off setting pressure during data transfer
    fun3dInstance[numInstance].pressureScaleOffset = NULL;

    // Increment number of instances
    numInstance += 1;

    return (numInstance -1);
}

int aimInputs(int iIndex, void *aimInfo, int index, char **ainame, capsValue *defval)
{
    /*! \page aimInputsFUN3D AIM Inputs
     * The following list outlines the FUN3D inputs along with their default value available
     * through the AIM interface. One will note most of the FUN3D parameters have a NULL value as their default.
     * This is done since a parameter in the FUN3D input deck (fun3d.nml) is only changed if the value has
     * been changed in CAPS (i.e. set to something other than NULL).
     */

	int status;

	#ifdef DEBUG
        printf(" fun3dAIM/aimInputs instance = %d  index = %d!\n", iIndex, index);
	#endif

    *ainame = NULL;
  
	// FUN3D Inputs
	if (index == 1) {
	    *ainame              = EG_strdup("Proj_Name");
		defval->type         = String;
		defval->nullVal      = NotNull;
		defval->vals.string  = EG_strdup("fun3d_CAPS");
		defval->lfixed       = Change;

		/*! \page aimInputsFUN3D
		 * - <B> Proj_Name = "fun3d_CAPS"</B> <br>
		 * This corresponds to the project\_rootname variable in the \&project namelist of fun3d.nml.
		 * This should be consistent with the meshing AIM.
		*/
	} else if (index == 2) {
		*ainame              = EG_strdup("Mach"); // Mach number
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;

		/*! \page aimInputsFUN3D
		 * - <B> Mach = NULL </B> <br>
		 *  This corresponds to the mach\_number variable in the \&reference\_physical\_properties
		 *  namelist of fun3d.nml.
		 */
	} else if (index == 3) {
		*ainame              = EG_strdup("Re"); // Reynolds number
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;

		/*! \page aimInputsFUN3D
		 * - <B> Re = NULL </B> <br>
		 *  This corresponds to the reynolds\_number variable in the \&reference\_physical\_properties
		 *  namelist of fun3d.nml.
		 */
	} else if (index == 4) {
		*ainame              = EG_strdup("Viscous"); // Viscous term
		defval->type         = String;
		defval->vals.string  = NULL;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;

		/*! \page aimInputsFUN3D
		 * - <B> Viscous = NULL </B> <br>
		 *  This corresponds to the viscous\_terms variable in the \&governing\_equation namelist
		 *  of fun3d.nml.
		 */
	} else if (index == 5) {
		*ainame              = EG_strdup("Eqn"); // Equation type
		defval->type         = String;
		defval->vals.string  = NULL;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;

		/*! \page aimInputsFUN3D
		 * - <B> Eqn = NULL </B> <br>
		 * This corresponds to the eqn\_type variable in the \&governing\_equation namelist of
		 * fun3d.nml. This is chosen automatically based on the capsFidelity attribute.
		 */
	} else if (index == 6) {
		*ainame              = EG_strdup("Alpha");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;

		/*! \page aimInputsFUN3D
		 * - <B> Alpha = NULL </B> <br>
		 *  This corresponds to the angle\_of\_attack variable in the \&reference\_physical\_properties
		 *  namelist of fun3d.nml.
		 */
	} else if (index == 7) {
		*ainame              = EG_strdup("Beta");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;

        /*! \page aimInputsFUN3D
         * - <B> Eqn = NULL </B> <br>
         *  This corresponds to the angle\_of\_yaw variable in the \&reference\_physical\_properties
         *  namelist of fun3d.nml.
         */
	} else if (index == 8) {
        *ainame              = EG_strdup("Overwrite_NML");
        defval->type         = Boolean;
        defval->vals.integer = false;
        defval->nullVal      = NotNull;

        /*! \page aimInputsFUN3D
         * - <B> Overwrite_NML = NULL</B> <br>
         *  If Python is not linked with the FUN3d AIM at compile time this flag gives the
         *  AIM permission to overwrite fun3d.nml if present. The namelist produced will solely
         *  consist of input variables present and set in the AIM.
         */
	} else if (index == 9) {
	    *ainame              = EG_strdup("Mesh_Format");
        defval->type         = String;
        defval->vals.string  = EG_strdup("AFLR3");
        defval->lfixed       = Change;

        /*! \page aimInputsFUN3D
         * - <B>Mesh_Format = "AFLR3"</B> <br>
         *  Mesh output format. By default an AFLR3 mesh will be used.
         */
	} else if (index == 10) {
        *ainame              = EG_strdup("Mesh_ASCII_Flag");
        defval->type         = Boolean;
        defval->vals.integer = true;

        /*! \page aimInputsFUN3D
         * - <B>Mesh_ASCII_Flag = True</B> <br>
         *  Output mesh in ASCII format, otherwise write a binary file if applicable.
         */
	} else if (index == 11) {
        *ainame              = EG_strdup("Num_Iter");
        defval->type         = Integer;
        defval->nullVal      = IsNull;
        defval->units        = NULL;
        defval->lfixed       = Change;
        defval->dim          = 0;

        /*! \page aimInputsFUN3D
         * - <B>Num_Iter = NULL</B> <br>
         *  This corresponds to the steps variable in the \&code\_run\_control
         *  namelist of fun3d.nml.
         */
	} else if (index == 12) {
        *ainame               = EG_strdup("CFL_Schedule");
        defval->type          = Double;
        defval->dim           = 1;
        defval->length        = 2;
        defval->nrow          = 2;
        defval->ncol          = 1;
        defval->units         = NULL;
        defval->vals.reals    = (double *) EG_alloc(defval->length*sizeof(double));
        defval->nullVal       = IsNull;
        defval->lfixed        = Fixed;

        /*! \page aimInputsFUN3D
         * - <B>CFL_Schedule = NULL</B> <br>
         *  This corresponds to the schedule\_cfl variable in the \&nonlinear\_solver\_parameters
         *   namelist of fun3d.nml.
         */
	} else if (index == 13) {
        *ainame               = EG_strdup("CFL_Schedule_Iter");
        defval->type          = Integer;
        defval->dim           = 1;
        defval->length        = 2;
        defval->nrow          = 2;
        defval->ncol          = 1;
        defval->units         = NULL;
        defval->vals.integers = (int *) EG_alloc(defval->length*sizeof(int));
        defval->nullVal       = IsNull;
        defval->lfixed        = Fixed;

        /*! \page aimInputsFUN3D
         * - <B>CFL_Schedule_Inter = NULL</B> <br>
         *  This corresponds to the schedule\_iteration variable in the \&nonlinear\_solver\_parameters
         *  namelist of fun3d.nml.
         */
    } else if (index == 14) {
	    *ainame              = EG_strdup("Restart_Read");
		defval->type         = String;
		defval->vals.string  = NULL;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;

        /*! \page aimInputsFUN3D
         * - <B>Restart_Read = NULL</B> <br>
         * This corresponds to the restart_read variable in the \&code_run_control namelist of fun3d.nml.
         */
    } else if (index == 15) {
        *ainame              = EG_strdup("Boundary_Condition");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->dim          = 1;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;

        /*! \page aimInputsFUN3D
         * - <B>Boundary_Condition = NULL </B> <br>
         * See \ref cfdBoundaryConditions for additional details.
         */
    } else if (index == 16) {
        *ainame              = EG_strdup("Use_Python_NML");
        defval->type         = Boolean;
        defval->vals.integer = (int) false;

        /*! \page aimInputsFUN3D
         * - <B>Use_Python_NML = False </B> <br>
         * By default even if Python has been linked to the FUN3D AIM it is not used unless the this value is set to True.
         */
    } else if (index == 17) {
        *ainame              = EG_strdup("Pressure_Scale_Factor");
        defval->type         = Double;
        defval->vals.real    = 1.0;
        defval->units      	 = NULL;

        fun3dInstance[iIndex].pressureScaleFactor = defval;

        /*! \page aimInputsFUN3D
         * - <B>Pressure_Scale_Factor = 1.0</B> <br>
         * Value to scale Cp data during when transferring data. Data is scaled based on Pressure = Pressure_Scale_Factor*Cp + Pressure_Scale_Offset.
         */
    } else if (index == 18) {
    	*ainame              = EG_strdup("Pressure_Scale_Offset");
    	defval->type         = Double;
    	defval->vals.real    = 0.0;
    	defval->units      	 = NULL;

    	fun3dInstance[iIndex].pressureScaleOffset = defval;

        /*! \page aimInputsFUN3D
         * - <B>Pressure_Scale_Offset = 0.0</B> <br>
         * Value to offset Cp data during when transferring data. Data is scaled based on Pressure = Pressure_Scale_Factor*Cp + Pressure_Scale_Offset.
         */
    } else if (index == 19) {
    	*ainame              = EG_strdup("NonInertial_Rotation_Rate");
    	defval->type          = Double;
    	defval->dim           = 1;
    	defval->length        = 3;
    	defval->nrow          = 3;
    	defval->ncol          = 1;
    	defval->units         = NULL;
    	defval->vals.reals    = (double *) EG_alloc(defval->length*sizeof(double));
    	if (defval->vals.reals == NULL) {
    		return EGADS_MALLOC;
    	} else {
    		defval->vals.reals[0] = 0.0;
    		defval->vals.reals[1] = 0.0;
    		defval->vals.reals[2] = 0.0;
    	}
    	defval->nullVal       = IsNull;
    	defval->lfixed        = Fixed;

        /*! \page aimInputsFUN3D
         * - <B>NonInertial_Rotation_Rate = NULL [0.0, 0.0, 0.0]</B> <br>
         * Array values correspond to the rotation\_rate\_x, rotation\_rate\_y, rotation\_rate\_z variables, respectively,
         * in the \&noninertial\_reference\_frame namelist of fun3d.nml.
         */
    } else if (index == 20) {
    	*ainame              = EG_strdup("NonInertial_Rotation_Center");
    	defval->type          = Double;
    	defval->dim           = 1;
    	defval->length        = 3;
    	defval->nrow          = 3;
    	defval->ncol          = 1;
    	defval->units         = NULL;
    	defval->vals.reals    = (double *) EG_alloc(defval->length*sizeof(double));
    	if (defval->vals.reals == NULL) {
    		return EGADS_MALLOC;
    	} else {
    		defval->vals.reals[0] = 0.0;
    		defval->vals.reals[1] = 0.0;
    		defval->vals.reals[2] = 0.0;
    	}
    	defval->nullVal       = IsNull;
    	defval->lfixed        = Fixed;

        /*! \page aimInputsFUN3D
         * - <B>NonInertial_Rotation_Center = NULL, [0.0, 0.0, 0.0]</B> <br>
         * Array values correspond to the rotation\_center\_x, rotation\_center\_y, rotation\_center\_z variables, respectively,
         * in the \&noninertial\_reference\_frame namelist of fun3d.nml.
         */
    } else if (index == 21) {
    	*ainame              = EG_strdup("Two_Dimensional");
    	defval->type          = Boolean;
    	defval->type         = Boolean;
    	defval->vals.integer = (int) false;

    	/*! \page aimInputsFUN3D
    	 * - <B>Two_Dimensional = False</B> <br>
    	 * Run FUN3D in 2D mode. If set to True, body must be a single "sheet" body in the x-z plane. A 3D mesh will be written out,
    	 * where the body is extruded a length of 1 in the y-direction.
    	 */
    }  else if (index == 22) {
        *ainame              = EG_strdup("Modal_Aeroelastic");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->dim          = 1;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;

        /*! \page aimInputsFUN3D
         * - <B>Modal_Aeroelastic = NULL </B> <br>
         * See \ref cfdModalAeroelastic for additional details.
         */
    } else if (index == 23) {
        *ainame              = EG_strdup("Modal_Ref_Velocity");
        defval->type         = Double;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->dim          = 1;
        defval->lfixed       = Change;

        /*! \page aimInputsFUN3D
         * - <B>Modal_Ref_Velocity = NULL </B> <br>
         * The freestream velocity in structural dynamics equation units; used for scaling during modal
         * aeroelastic simulations. This corresponds to the uinf variable in the \&aeroelastic\_modal\_data
         *  namelist of movingbody.input.
         */
    }  else if (index == 24) {
        *ainame              = EG_strdup("Modal_Ref_Length");
        defval->type         = Double;
        //defval->units        = NULL;
        defval->dim          = 1;
        defval->lfixed       = Change;
        defval->vals.real   = 1.0;

        /*! \page aimInputsFUN3D
         * - <B>Modal_Ref_Length = 1.0 </B> <br>
         * The scaling factor between CFD and the structural dynamics equation units; used for scaling during modal
         * aeroelastic simulations. This corresponds to the grefl variable in the \&aeroelastic\_modal\_data
         *  namelist of movingbody.input.
         */
    }  else if (index == 25) {
        *ainame              = EG_strdup("Modal_Ref_Dynamic_Pressure");
        defval->type         = Double;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->dim          = 1;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;

        /*! \page aimInputsFUN3D
         * - <B>Modal_Ref_Dynamic_Pressure = NULL </B> <br>
         * The freestream dynamic pressure in structural dynamics equation units; used for scaling during modal
         * aeroelastic simulations. This corresponds to the qinf variable in the \&aeroelastic\_modal\_data
         *  namelist of movingbody.input.
         */
    } else if (index == 26) {
        *ainame              = EG_strdup("Time_Accuracy");
        defval->type         = String;
        defval->vals.string  = NULL;
        defval->nullVal      = IsNull;
        defval->units        = NULL;
        defval->lfixed       = Change;

        /*! \page aimInputsFUN3D
         * - <B>Time_Accuracy = NULL </B> <br>
         * Defines the temporal scheme to use. This corresponds to the time_accuracy variable
         *  in the \&nonlinear_solver_parameters namelist of fun3d.nml.
         */
    } else if (index == 27) {
        *ainame              = EG_strdup("Time_Step");
        defval->type         = Double;
        defval->nullVal      = IsNull;
        defval->units        = NULL;
        defval->lfixed       = Change;
        defval->dim          = 0;

        /*! \page aimInputsFUN3D
         * - <B>Time_Step = NULL </B> <br>
         * Non-dimenional time step during time accurate simulations. This corresponds to the time_step_nondim
         * variable in the \&nonlinear_solver_parameters namelist of fun3d.nml.
         */
    } else if (index == 28) {
        *ainame              = EG_strdup("Num_Subiter");
        defval->type         = Integer;
        defval->nullVal      = IsNull;
        defval->units        = NULL;
        defval->lfixed       = Change;
        defval->dim          = 0;

        /*! \page aimInputsFUN3D
         * - <B>Num_Subiter = NULL </B> <br>
         * Number of sub-interations used during a time step in a time accurate simulations. This
         * corresponds to the subiterations variable in the \&nonlinear_solver_parameters namelist of fun3d.nml.
         */
    } else if (index == 29) {
        *ainame              = EG_strdup("Temporal_Error");
        defval->type         = Double;
        defval->nullVal      = IsNull;
        defval->units        = NULL;
        defval->lfixed       = Change;
        defval->dim          = 0;

        /*! \page aimInputsFUN3D
         * - <B>Temporal_Error = NULL </B> <br>
         * This sets the tolerance for which subiterations are stopped during time accurate simulations. This
         * corresponds to the temporal_err_floor variable in the \&nonlinear_solver_parameters namelist of fun3d.nml.
         */
    }

	// Link variable(s) to parent(s) if available
	/*if ((index != 1) && (*ainame != NULL) && (index !=15)) {
            status = aim_link(aimInfo, *ainame, ANALYSISIN, defval);
	    printf("Status = %d: Var Index = %d, Type = %d, link = %lX\n",
                   status, index, defval->type, defval->link);
	}*/

	return CAPS_SUCCESS;
}

int aimData(/*@unused@*/ int iIndex,
            /*@unused@*/ const char *name,
            /*@unused@*/ enum capsvType *vtype,
            /*@unused@*/ int *rank,
            /*@unused@*/ int *nrow,
            /*@unused@*/ int *ncol,
            /*@unused@*/ void **data,
            /*@unused@*/ char **units)
{

    /*! \page sharableDataFUN3D AIM Shareable Data
     * Currently the FUN3D AIM does not have any shareable data types or values. It will try, however, to inherit a
     * "Volume_Mesh" (for 3D simulations) or a "Surface_Mesh" (for 2D simulations) from any parent AIMs.
     * Note that the inheritance of the volume/surface mesh variable is required
     * if the FUN3D aim is to write a suitable grid and <sup>*</sup>.mapbc (boundary condition file for the AFLR3 mesh format) file.
     */

    #ifdef DEBUG
        printf(" fun3dAIM/aimData instance = %d  name = %s!\n", iIndex, name);
    #endif
   return CAPS_NOTFOUND;
}

int aimPreAnalysis(int iIndex, void *aimInfo, char *analysisPath,
                   capsValue *aimInputs, capsErrs **errs)
{

    // Function return flag
    int status;

    // Python linking
    int pythonLinked = (int) false;
    int usePython = (int) false;

    // Indexing
    int i, j, body;

    // Data transfer
    int  nrow, ncol, rank;
    void *dataTransfer = NULL;
    enum capsvType vtype;
    char *units = NULL;

    // EGADS return values
    int          atype, alen;
    const int    *ints;
    const char   *string;
    const double *reals;

    // Output filename
    char *filename;

    // Current CAPS fidelity
    int currentFidelity = 0;

    // AIM input bodies
    int  numBody;
    ego *bodies = NULL;
    int  bodySubType = 0, bodyOClass, bodyNumChild, *bodySense;
    ego  bodyRef, *bodyChild;
	double bodyData[4];
  
    // Boundary/surface properties
    cfdBCsStruct   bcProps;
  
    // Modal Aeroelastic properties
    modalAeroelasticStruct   modalAeroelastic;

    // Boundary conditions container - for writing .mapbc file
    bndCondStruct bndConds;
  
    // Volume Mesh obtained from meshing AIM
    genUnstrMeshStruct volumeMesh;
    int numNodeCheck, numTriFaceCheck, numQuadFaceCheck; // Consistency checkers for volume-surface meshes

    // Discrete data transfer variables
	capsDiscr *dataTransferDiscreteObj;
	char **transferName = NULL;
	int numTransferName;

    // NULL out errors
    *errs = NULL;
  
    // Store away the analysis path/directory
    fun3dInstance[iIndex].analysisPath = analysisPath;

    // Get AIM bodies
    status = aim_getBodies(aimInfo, &numBody, &bodies);
    if (status != CAPS_SUCCESS) printf("aim_getBodies status = %d!!\n", status);
  
     #ifdef DEBUG
        printf(" fun3dAIM/aimPreAnalysis instance = %d  numBody = %d!\n", iIndex, numBody);
    #endif
  
    if ((numBody <= 0) || (bodies == NULL)) {
        #ifdef DEBUG
            printf(" fun3dAIM/aimPreAnalysis No Bodies!\n");
        #endif
        return CAPS_SOURCEERR;
    }

    // Check to see if python was linked
    #ifdef HAVE_PYTHON
        pythonLinked = (int) true;
    #else
        pythonLinked = (int) false;
    #endif

    // Should we use python even if it was linked?
    usePython = aimInputs[aim_getIndex(aimInfo, "Use_Python_NML", ANALYSISIN)-1].vals.integer;
    if (usePython == (int) true && pythonLinked == (int) false) {

        printf("Use of Python library requested but not linked!\n");
        usePython = (int) false;

    } else if (usePython == (int) false && pythonLinked == (int) true) {

        printf("Python library was linked, but will not be used!\n");
    }

    // Get project name
    fun3dInstance[iIndex].projectName = aimInputs[aim_getIndex(aimInfo, "Proj_Name", ANALYSISIN)-1].vals.string;

    // Initiate structures variables - will be destroyed during cleanup
    status = initiate_cfdBCsStruct(&bcProps);
    if (status != CAPS_SUCCESS) return status;

    status = initiate_modalAeroelasticStruct(&modalAeroelastic);
    if (status != CAPS_SUCCESS) return status;

    status = initiate_genUnstrMeshStruct(&volumeMesh);
    if (status != CAPS_SUCCESS) return status;

    status = initiate_bndCondStruct(&bndConds);
    if (status != CAPS_SUCCESS) return status;

    // Get fidelity
    for (body = 0; body < numBody; body++) {
        status = EG_attributeRet(bodies[body], "capsFidelity", &atype, &alen, &ints, &reals, &string);
        if (status   != EGADS_SUCCESS) continue; // Need to add an error code

        if (body == 0){
        	currentFidelity = (int) reals[0];

        } else {

        	if (currentFidelity != (int) reals[0]) {
        		printf("All bodies must have the same capsFidelity value - Body 1 = %d, Body %d = %d!\n", currentFidelity,
        																								  body+1,
																										  (int) reals[0]);
        		status = CAPS_BADVALUE;
        		goto cleanup;

        	}
        }
    }

    // Get attribute to index mapping
    status = aim_getData(aimInfo, "Attribute_Map", &vtype, &rank, &nrow, &ncol, &dataTransfer, &units);
    if (status == CAPS_SUCCESS) {

        printf("Found link for attrMap (Attribute_Map) from parent\n");

        //fun3dInstance[iIndex].attrMap = *(mapAttrToIndexStruct *) dataTransfer;

        status = copy_mapAttrToIndexStruct( (mapAttrToIndexStruct *) dataTransfer, &fun3dInstance[iIndex].attrMap);
        if (status != CAPS_SUCCESS) goto cleanup;

    } else {

        if (status == CAPS_DIRTY) printf("Parent AIMS are dirty\n");
        else printf("Linking status during attrMap (Attribute_Map) = %d\n",status);

        printf("Didn't find a link to attrMap from parent - getting it ourselves\n");

        if (aimInputs[aim_getIndex(aimInfo, "Two_Dimensional", ANALYSISIN)-1].vals.integer == (int) true ) {

            // Get capsGroup name and index mapping to make sure all faces and edges have a capsGroup value
        	status = create_CAPSGroupAttrToIndexMap(numBody,
        											bodies,
													2, // Only search down to the edge level of the EGADS body
													&fun3dInstance[iIndex].attrMap);
        	if (status != CAPS_SUCCESS) goto cleanup;

        } else {

        	// Get capsGroup name and index mapping to make sure all faces have a capsGroup value
        	status = create_CAPSGroupAttrToIndexMap(numBody,
        											bodies,
													1, // Only search down to the face level of the EGADS body
													&fun3dInstance[iIndex].attrMap);
        	if (status != CAPS_SUCCESS) goto cleanup;

        }
    }
    // Get boundary conditions - Only if the boundary condition has been set
    if (aimInputs[aim_getIndex(aimInfo, "Boundary_Condition", ANALYSISIN)-1].nullVal ==  NotNull) {

        status = cfd_getBoundaryCondition( aimInputs[aim_getIndex(aimInfo, "Boundary_Condition", ANALYSISIN)-1].length,
                                           aimInputs[aim_getIndex(aimInfo, "Boundary_Condition", ANALYSISIN)-1].vals.tuple,
                                           &fun3dInstance[iIndex].attrMap,
                                           &bcProps);
        if (status != CAPS_SUCCESS) goto cleanup;

    } else {
        printf("Warning: No boundary conditions provided !!!!\n");
    }

	// Get modal aeroelastic information - only get modal aeroelastic inputs if they have be set
	if (aimInputs[aim_getIndex(aimInfo, "Modal_Aeroelastic", ANALYSISIN)-1].nullVal ==  NotNull) {

		status = cfd_getModalAeroelastic( aimInputs[aim_getIndex(aimInfo, "Modal_Aeroelastic", ANALYSISIN)-1].length,
										  aimInputs[aim_getIndex(aimInfo, "Modal_Aeroelastic", ANALYSISIN)-1].vals.tuple,
										  &modalAeroelastic);
		if (status != CAPS_SUCCESS) goto cleanup;


		modalAeroelastic.freestreamVelocity = aimInputs[aim_getIndex(aimInfo, "Modal_Ref_Velocity", ANALYSISIN)-1].vals.real;
		modalAeroelastic.freestreamDynamicPressure = aimInputs[aim_getIndex(aimInfo, "Modal_Ref_Dynamic_Pressure", ANALYSISIN)-1].vals.real;
		modalAeroelastic.lengthScaling = aimInputs[aim_getIndex(aimInfo, "Modal_Ref_Length", ANALYSISIN)-1].vals.real;
	}

    // Are we running in two-mode
    if (aimInputs[aim_getIndex(aimInfo, "Two_Dimensional", ANALYSISIN)-1].vals.integer == (int) true ) {

    	if (numBody > 1) {
    		printf("Only 1 body may be provided when running in two mode!!\n");
    		status = CAPS_BADVALUE;
    		goto cleanup;
    	}

    	for (body = 0; body < numBody; body++) {
			// What type of BODY do we have?
			status = EG_getTopology(bodies[body], &bodyRef, &bodyOClass, &bodySubType, bodyData, &bodyNumChild, &bodyChild, &bodySense);
			if (status != EGADS_SUCCESS) goto cleanup;

			if (bodySubType != FACEBODY && bodySubType != SHEETBODY) {
				printf("Body type must be either FACEBODY (%d) or a SHEETBODY (%d) when running in two mode!\n", FACEBODY, SHEETBODY);
				status = CAPS_BADTYPE;
				goto cleanup;
			}
    	}


    	// Add extruded plane boundary condition
		bcProps.numBCID += 1;
		bcProps.surfaceProps = (cfdSurfaceStruct *) EG_reall(bcProps.surfaceProps, bcProps.numBCID * sizeof(cfdSurfaceStruct));
		if (bcProps.surfaceProps == NULL) return EGADS_MALLOC;

		status = intiate_cfdSurfaceStruct(&bcProps.surfaceProps[bcProps.numBCID-1]);
		if (status != CAPS_SUCCESS) goto cleanup;

		bcProps.surfaceProps[bcProps.numBCID-1].surfaceType = Symmetry;
		bcProps.surfaceProps[bcProps.numBCID-1].symmetryPlane = 2;

		// Find largest index value for bcID and set it plus 1 to the new surfaceProp
		for (i = 0; i < bcProps.numBCID-1; i++) {

			if (bcProps.surfaceProps[i].bcID >= bcProps.surfaceProps[bcProps.numBCID-1].bcID) {

				bcProps.surfaceProps[bcProps.numBCID-1].bcID = bcProps.surfaceProps[i].bcID + 1;
			}
		}

		// Get Surface mesh
    	status = aim_getData(aimInfo, "Surface_Mesh", &vtype, &rank, &nrow, &ncol, &dataTransfer, &units);
    	if (status != CAPS_SUCCESS){

    		if (status == CAPS_DIRTY) printf("Parent AIMS are dirty\n");
    		else printf("Linking status = %d\n",status);

    		printf("Didn't find a link to a surface mesh (Surface_Mesh) from parent - a mesh will NOT be created.\n");

    	} else {

    		printf("Found link for surface mesh (Surface_Mesh) from parent\n");

    		status = fun3d_2DMesh(*(genUnstrSurfMeshStruct *) dataTransfer,
    							  &volumeMesh,
								  &bcProps.surfaceProps[bcProps.numBCID-1].bcID);
    		if (status != CAPS_SUCCESS) goto cleanup;

    	}

    	// Can't currently do data transfer in 2D-mode
    	fun3dInstance[iIndex].dataTransferCheck = (int) false;

    } else {
        // Get Volume mesh
        status = aim_getData(aimInfo, "Volume_Mesh", &vtype, &rank, &nrow, &ncol, &dataTransfer, &units);
        if (status != CAPS_SUCCESS){

            if (status == CAPS_DIRTY) printf("Parent AIMS are dirty\n");
            else printf("Linking status = %d\n",status);

            printf("Didn't find a link to a volume mesh (Volume_Mesh) from parent - a volume mesh will NOT be created.\n");

            // No volume currently means we can't do datatransfer
            fun3dInstance[iIndex].dataTransferCheck = (int) false;

        } else {

        	printf("Found link for volume mesh (Volume_Mesh) from parent\n");
        	volumeMesh = *(genUnstrMeshStruct *) dataTransfer;
        }
    }

    if (status == CAPS_SUCCESS) {

        status = populate_bndCondStruct_from_bcPropsStruct(&bcProps, &bndConds);
        if (status != CAPS_SUCCESS) goto cleanup;

        // Replace dummy values in bcVal with FUN3D specific values
        for (i = 0; i < bcProps.numBCID ; i++) {

            // {UnknownBoundary, Inviscid, Viscous, Farfield, Extrapolate, Freestream,
            //  BackPressure, Symmetry, SubsonicInflow, SubsonicOutflow,
            //  MassflowIn, MassflowOut, FixedInflow, FixedOutflow}

            if      (bcProps.surfaceProps[i].surfaceType == Inviscid) bndConds.bcVal[i] = 3000;
            else if (bcProps.surfaceProps[i].surfaceType == Viscous)  bndConds.bcVal[i] = 4000;
            else if (bcProps.surfaceProps[i].surfaceType == Farfield) bndConds.bcVal[i] = 5000;
            else if (bcProps.surfaceProps[i].surfaceType == Extrapolate) bndConds.bcVal[i] = 5026;
            else if (bcProps.surfaceProps[i].surfaceType == Freestream)  bndConds.bcVal[i] = 5050;
            else if (bcProps.surfaceProps[i].surfaceType == BackPressure)    bndConds.bcVal[i] = 5051;
            else if (bcProps.surfaceProps[i].surfaceType == SubsonicInflow)  bndConds.bcVal[i] = 7011;
            else if (bcProps.surfaceProps[i].surfaceType == SubsonicOutflow) bndConds.bcVal[i] = 7012;
            else if (bcProps.surfaceProps[i].surfaceType == MassflowIn)      bndConds.bcVal[i] = 7036;
            else if (bcProps.surfaceProps[i].surfaceType == MassflowOut)     bndConds.bcVal[i] = 7031;
            else if (bcProps.surfaceProps[i].surfaceType == FixedInflow)     bndConds.bcVal[i] = 7100;
            else if (bcProps.surfaceProps[i].surfaceType == FixedOutflow)    bndConds.bcVal[i] = 7105;
            else if (bcProps.surfaceProps[i].surfaceType == Symmetry) {

                if      (bcProps.surfaceProps[i].symmetryPlane == 1) bndConds.bcVal[i] = 6661;
                else if (bcProps.surfaceProps[i].symmetryPlane == 2) bndConds.bcVal[i] = 6662;
                else if (bcProps.surfaceProps[i].symmetryPlane == 3) bndConds.bcVal[i] = 6663;
                else {
                	printf("Unknown symmetryPlane for boundary %d - Defaulting to y-Symmetry\n", bcProps.surfaceProps[i].bcID);
                	 bndConds.bcVal[i] = 6662;
                }
            }
        }

    	filename = EG_alloc(MXCHAR +1);
    	if (filename == NULL) {
    		status = EGADS_MALLOC;
    		goto cleanup;
    	}
    	strcpy(filename, analysisPath);
    	strcat(filename, "/");
    	strcat(filename, fun3dInstance[iIndex].projectName);

        if (fun3dInstance[iIndex].haveVolumeMesh == (int) false) {

        	// Write AFLR3
        	status = write_AFLR3(filename,
        						 aimInputs[aim_getIndex(aimInfo, "Mesh_ASCII_Flag",  ANALYSISIN)-1].vals.integer,
								 &volumeMesh);

        	if (status != CAPS_SUCCESS) {
        		goto cleanup;
        	}
        }

        // Write *.mapbc file
        status = write_MAPBC(filename,
                             bndConds.numBND,
                             bndConds.bndID,
                             bndConds.bcVal);
        if (filename != NULL) EG_free(filename);
        filename = NULL;

        if (status != CAPS_SUCCESS) goto cleanup;

        // Lets check the volume mesh

        // Do we have an individual surface mesh for each body
        if (volumeMesh.surfaceMesh.numSurface != numBody) {
        	printf("Number of inherited surface meshes in the volume mesh does not match the number "
        			"of bodies - data transfer will NOT be possible.\n\n");
        	fun3dInstance[iIndex].dataTransferCheck = (int) false;
        }

        // Check to make sure the volume mesher didn't add any unaccounted for points/faces
        numNodeCheck = numTriFaceCheck = numQuadFaceCheck = 0;
        for (i = 0; i < volumeMesh.surfaceMesh.numSurface; i++) {
        	numNodeCheck += volumeMesh.surfaceMesh.surfNodesLump[i];
        	numTriFaceCheck += volumeMesh.surfaceMesh.surfTriFaceLump[i];
        	numQuadFaceCheck+= volumeMesh.surfaceMesh.surfQuadFaceLump[i];
        }

        if (numNodeCheck != volumeMesh.surfaceMesh.numNodes) {

        	fun3dInstance[iIndex].dataTransferCheck = (int) false;
        	printf("Volume mesher added nodes - data transfer will NOT be possible.\n");

        } else if (numTriFaceCheck != volumeMesh.surfaceMesh.numTriFace) {

        	fun3dInstance[iIndex].dataTransferCheck = (int) false;
        	printf("Volume mesher added trifaces - data transfer will NOT be possible.\n");

        } else if (numQuadFaceCheck != volumeMesh.surfaceMesh.numQuadFace) {

        	fun3dInstance[iIndex].dataTransferCheck = (int) false;
        	printf("Volume mesher added quadfaces - data transfer will NOT be possible.\n");

        } else { // Data transfer appears to be ok
        	fun3dInstance[iIndex].dataTransferCheck = (int) true;
        }

        // To this point is doesn't appear that the volume mesh has done anything bad to our surface mesh(es)
        if (fun3dInstance[iIndex].dataTransferCheck == (int) true) {

        	// Lets store away our tessellation now
        	for (i = 0; i < volumeMesh.surfaceMesh.numSurface; i++) {

        		if (fun3dInstance[iIndex].haveVolumeMesh == (int) false) {
                	#ifdef DEBUG
        	        	printf(" fun3dAIM/aimPreAnalysis instance = %d Setting body tessellation number %d!\n", iIndex, i);
                	#endif

        	        status = aim_setTess(aimInfo, volumeMesh.surfaceMesh.bodyTessMappingLump[i]->egadsTess);
        	        if (status != CAPS_SUCCESS) {
        	        	printf(" aim_setTess return = %d\n", status);
        	        	goto cleanup;
        	        }

        	        fun3dInstance[iIndex].haveVolumeMesh = (int) true;
        		}
        	}
        }
    }

    //////////////////////////////////////////////////////////
    // Open and write the fun3d.nml input file using Python //
    //////////////////////////////////////////////////////////
    if (usePython == (int) true && pythonLinked == (int) true) {
        #ifdef HAVE_PYTHON
        {
            int        temperature_found = false;
            char       temperature_string[MXCHAR], temp[MXCHAR];
            const char *aname;

            FILE *fpy;

            // Inputs*2 (- option) + python file name + -dir and directory + -mesh and mesh type
            // + -temperature and temperature boundaries
            int	   python_argc = (NUMINPUT)*2 +1+2+2+2;

            char  *python_namelist_helper = (char *) "fun3d_namelist_aim.py";
            char **python_argv;
      
            // Flag to see if Python was already initialized - i.e. the AIM was called from within Python
            int initPy = (int) false;
        
            // Python thread state variable for new interpreter if needed
            PyThreadState *newThread = NULL;

            // Allocate memory for python_argv
            python_argv = (char **) EG_alloc(python_argc * sizeof(char *));
            if (python_argv == NULL) return EGADS_MALLOC;

            // Initiate python_argv elements
            for (i = 0; i < python_argc; i++) {
                //python_argv[i] = NULL;
                python_argv[i] = (char *) EG_alloc((MXCHAR+1)*sizeof(char));
                if (python_argv[i] == NULL) {

                    for (j = 0; j <  i; j++) EG_free(python_argv[j]);

                    EG_free(python_argv);
                    return EGADS_MALLOC;
                }
            }

            // Initialize python
            if (Py_IsInitialized() == 0) {
                printf("\n Initializing Python for FUN3D AIM\n");
                Py_Initialize();
            } else {
                // If python is already initialized - create a new interpreter to run the namelist creator
                //  otherwise you get seg-faults when SetArgv and SimpleFile are used
                PyEval_InitThreads(); // This function call is probably not necessary

                newThread = Py_NewInterpreter();
                (void) PyThreadState_Swap(newThread); // This function call is probably not necessary;

                initPy = (int) true;
            }

            // Open python file
            fpy = fopen(python_namelist_helper,"r");
            if (fpy == NULL) {
                printf("\n\tError opening FUN3D namelist python file - %s\n", python_namelist_helper);
                for (i = 0; i < python_argc; i++) EG_free(python_argv[i]);
                EG_free(python_argv);
                return CAPS_IOERR;
            } else {
                    /*printf("\nPrinting contents of file:\n");
                    while (fgets( str, 100, fpy ) != NULL) {
                        puts(str);
                    }
                    rewind(fpy);*/
            }

            // Parse aimInputs to put in argv string array //
            sprintf(python_argv[0], "%s", python_namelist_helper);
            sprintf(python_argv[1], "%s", "-dir");
            sprintf(python_argv[2], "%s", analysisPath);

            sprintf(python_argv[3], "%s", "-mesh");
            if (strcasecmp(aimInputs[aim_getIndex(aimInfo, "Mesh_Format",ANALYSISIN)-1].vals.string,"AFLR3") != 0) {
                printf("Unsupported mesh format for FUN3D\n");
                for (i = 0; i < python_argc; i++) EG_free(python_argv[i]);
                EG_free(python_argv);
                return CAPS_BADVALUE; // Add CAPS error
            }

            if (aimInputs[aim_getIndex(aimInfo, "Mesh_ASCII_Flag",  ANALYSISIN)-1].vals.integer == (int) true) {

                sprintf(python_argv[4], "%s", "AFLR3 1");
            } else {

                sprintf(python_argv[4], "%s", "AFLR3 0");
            }

            // Parse boundary condition aimInputs
            // Temperature
            temperature_string[0] = 0;
            temp[0] = 0;
            for (i = 0; i < bcProps.numBCID ; i++) {
                if (bcProps.surfaceProps[i].wallTemperatureFlag == (int) true) {

                    sprintf(python_argv[5], "%s", "-temperature_bc");

                    sprintf(temp, "%d ", bcProps.surfaceProps[i].bcID);
                    strcat(temperature_string, temp);
                    sprintf(temp, "%f ", bcProps.surfaceProps[i].wallTemperature);
                    strcat(temperature_string, temp);

                    temperature_found = (int) true;
                }
            }

            if (temperature_found == (int) false) {
                sprintf(python_argv[5], "%s", "-dummy");
                sprintf(python_argv[6], "%s", "False");

            } else {
                sprintf(python_argv[6], "%s", temperature_string);
            }

            j = 7; // First 7 indexes of argv already filled above
            for (i = 0; i < NUMINPUT; i++) {
                status = aim_getName(aimInfo, i+1, ANALYSISIN, &aname); // i+1 since aim_getName is 1 bias

                if (aimInputs[i].nullVal ==  NotNull) {
                    status = aim_getName(aimInfo, i+1, ANALYSISIN, &aname); // i+1 since aim_getName is 1 bias
                    if (status != EGADS_SUCCESS) continue;

                    // Set python argv based on aname
                    if      (strcmp(aname, "Proj_Name") == 0) sprintf(python_argv[j], "%s", "-proj");
                    else if (strcmp(aname, "Mach")      == 0) sprintf(python_argv[j], "%s", "-mach");
                    else if (strcmp(aname, "Re")        == 0) sprintf(python_argv[j], "%s", "-re");
                    else if (strcmp(aname, "Viscous")   == 0) sprintf(python_argv[j], "%s", "-visc");
                    else if (strcmp(aname, "Eqn")       == 0) sprintf(python_argv[j], "%s", "-eqn");
                    else if (strcmp(aname, "Alpha")     == 0) sprintf(python_argv[j], "%s", "-alpha");
                    else if (strcmp(aname, "Beta")      == 0) sprintf(python_argv[j], "%s", "-beta");
                    else if (strcmp(aname, "Num_Iter")  == 0) sprintf(python_argv[j], "%s", "-iter");
                    else if (strcmp(aname, "CFL_Schedule")  == 0) {

                        sprintf(python_argv[j], "%s", "-cfl_schedule");
                        sprintf(python_argv[j+1], "%f %f", aimInputs[i].vals.reals[0], aimInputs[i].vals.reals[1]);

                        j = j+2;
                        continue;
                    }
                    else if (strcmp(aname, "CFL_Schedule_Iter")  == 0) {

                        sprintf(python_argv[j], "%s", "-cfl_schedule_iter");
                        sprintf(python_argv[j+1], "%d %d", aimInputs[i].vals.integers[0], aimInputs[i].vals.integers[1]);
                        j = j+2;
                        continue;
                    }
                    else if (strcmp(aname, "NonInertial_Rotation_Center")  == 0) {

                        sprintf(python_argv[j], "%s", "-noninertial_rot_center");
                        sprintf(python_argv[j+1], "%f %f %f", aimInputs[i].vals.reals[0],
                        								      aimInputs[i].vals.reals[1],
														      aimInputs[i].vals.reals[2]);
                        j = j+2;
                        continue;
                    }
                    else if (strcmp(aname, "NonInertial_Rotation_Rate")  == 0) {

                        sprintf(python_argv[j], "%s", "-noninertial_rot_rate");
                        sprintf(python_argv[j+1], "%f %f %f", aimInputs[i].vals.reals[0],
                                                			  aimInputs[i].vals.reals[1],
															  aimInputs[i].vals.reals[2]);
                        j = j+2;
                        continue;
                    }
                    else if(strcmp(aname, "Restart_Read") == 0) sprintf(python_argv[j], "%s", "-restart");
                    else {

                        sprintf(python_argv[j], "%s", "-dummy");
                        sprintf(python_argv[j+1], "%s", "False");
                        j = j+2;
                        continue;
                    }


                    // Do we have double, string ?
                    if (aimInputs[i].type == Double) {

                        sprintf(python_argv[j+1], "%f", aimInputs[i].vals.real);

                    } else if (aimInputs[i].type == Integer) {

                        sprintf(python_argv[j+1], "%d",aimInputs[i].vals.integer);

                    } else if (aimInputs[i].type == String) {

                        sprintf(python_argv[j+1], "%s", aimInputs[i].vals.string);
                    }

                } else {
                    sprintf(python_argv[j], "%s", "-dummy");
                    sprintf(python_argv[j+1], "%s", "False");
                }

                j = j + 2;
            }

            printf("\n\tRunning Python\n");

            // Regurgitate python arguments
            //for (i = 0; i < python_argc; i++) printf("\nAgrv i = %d, %s\n", i, python_argv[i]);

            // Set python args
            PySys_SetArgv(python_argc, python_argv);

            // Run python file
            PyRun_SimpleFile(fpy, python_namelist_helper);

            // Close down python
            printf("\n\tClosing Python\n\n");

            if (initPy == (int) true) {

              Py_EndInterpreter(newThread);
              (void) PyThreadState_Swap(NULL); // This function call is probably not necessary;

            } else Py_Finalize(); // Do not finalize if the AIM was called from Python

            fclose(fpy);

            // Cleanup python_argv
            //printf("Cleaning up Python arguements - python_argc = %d\n", python_argc);
            for (i = 0; i < python_argc; i++) EG_free(python_argv[i]);
            EG_free(python_argv);

        }
        #endif

    } else {

        if (aimInputs[aim_getIndex(aimInfo, "Overwrite_NML",ANALYSISIN)-1].vals.integer == (int) false) {

            printf("Since Python was not linked and/or being used, the \"Overwrite_NML\" input needs to be set to \"True\" to give");
            printf(" permission to create a new fun3d.nml. fun3d.nml will NOT be updated!!\n");

        } else {

        	printf("Warning: The fun3d.nml file will be overwritten!\n");

        	status = fun3d_writeNML(aimInfo, analysisPath, aimInputs, bcProps);
        	if (status != CAPS_SUCCESS) goto cleanup;

        }
    }

    // If data transfer is ok ....
    if (fun3dInstance[iIndex].dataTransferCheck == (int) true) {

    	//See if we have data transfer information
    	status = aim_getBounds(aimInfo, &numTransferName, &transferName);
    	if (status == CAPS_SUCCESS) {

    		if (aimInputs[aim_getIndex(aimInfo, "Modal_Aeroelastic", ANALYSISIN)-1].nullVal ==  NotNull) {
				status = fun3D_dataTransfer(aimInfo,
											analysisPath,
											fun3dInstance[iIndex].projectName,
											volumeMesh,
											&modalAeroelastic);
				if (status == CAPS_SUCCESS) {

					status = fun3d_writeMovingBody(analysisPath, bcProps, &modalAeroelastic);
				}

    		} else{
    			status = fun3D_dataTransfer(aimInfo,
											analysisPath,
											fun3dInstance[iIndex].projectName,
											volumeMesh,
											NULL);
    		}

    		if (status != CAPS_SUCCESS && status != CAPS_NOTFOUND) goto cleanup;
    	}
    } // End if data transfer ok


    // Sensitivity
    if (fun3dInstance[iIndex].dataTransferCheck == (int) true) {


    	//status = aim_getSensitivity(aimInfo, ego tess, ttype, index, npts, dxyz)


    }

    status = CAPS_SUCCESS;

    goto cleanup;

    cleanup:

		if (transferName != NULL) EG_free(transferName);
		if (filename != NULL) EG_free(filename);

		(void) destroy_cfdBCsStruct(&bcProps);
		(void) destroy_modalAeroelasticStruct(&modalAeroelastic);

		(void) destroy_bndCondStruct(&bndConds);

		// Clean up the volume mesh that was created for 2D mode
		if (aimInputs[aim_getIndex(aimInfo, "Two_Dimensional", ANALYSISIN)-1].vals.integer == (int) true ) {

			// Destroy volume mesh since we created here instead of inheriting it
			(void) destroy_genUnstrMeshStruct(&volumeMesh);
		}

		return status;
}


int aimOutputs(int iIndex, int index, char **aoname, capsValue *form)
{

    /*! \page aimOutputsFUN3D AIM Outputs
     * The following list outlines the FUN3D outputs available through the AIM interface. All variables currently
     * correspond to values for all boundaries (total) found in the *.forces file
     */

	int numOutVars = 8; // Grouped

	#ifdef DEBUG
        printf(" fun3dAIM/aimOutputs instance = %d  index = %d!\n", iIndex, index);
    #endif

    // Total Forces - Pressure + Viscous
    if 		(index == 1) *aoname = EG_strdup("CLtot");
    else if (index == 2) *aoname = EG_strdup("CDtot");
    else if (index == 3) *aoname = EG_strdup("CMXtot");
    else if (index == 4) *aoname = EG_strdup("CMYtot");
    else if (index == 5) *aoname = EG_strdup("CMZtot");
    else if (index == 6) *aoname = EG_strdup("CXtot");
    else if (index == 7) *aoname = EG_strdup("CYtot");
    else if (index == 8) *aoname = EG_strdup("CZtot");

    /*! \page aimOutputsFUN3D
     * Net Forces - Pressure + Viscous:
     * - <B>CLtot</B> = The lift coefficient.
     * - <B>CDtot</B> = The drag coefficient.
     * - <B>CMXtot</B> = The moment coefficient about the x-axis.
     * - <B>CMYtot</B> = The moment coefficient about the y-axis.
     * - <B>CMZtot</B> = The moment coefficient about the z-axis.
     * - <B>CXtot</B> = The force coefficient about the x-axis.
     * - <B>CYtot</B> = The force coefficient about the y-axis.
     * - <B>CZtot</B> = The force coefficient about the z-axis.
     * .
     *
     */

    // Pressure Forces
    else if (index == 1 + numOutVars) *aoname = EG_strdup("CLtot_p");
    else if (index == 2 + numOutVars) *aoname = EG_strdup("CDtot_p");
    else if (index == 3 + numOutVars) *aoname = EG_strdup("CMXtot_p");
    else if (index == 4 + numOutVars) *aoname = EG_strdup("CMYtot_p");
    else if (index == 5 + numOutVars) *aoname = EG_strdup("CMZtot_p");
    else if (index == 6 + numOutVars) *aoname = EG_strdup("CXtot_p");
    else if (index == 7 + numOutVars) *aoname = EG_strdup("CYtot_p");
    else if (index == 8 + numOutVars) *aoname = EG_strdup("CZtot_p");

    /*! \page aimOutputsFUN3D
     * Pressure Forces:
     * - <B>CLtot_p</B> = The lift coefficient - pressure contribution only.
     * - <B>CDtot_p</B> = The drag coefficient - pressure contribution only.
     * - <B>CMXtot_p</B> = The moment coefficient about the x-axis - pressure contribution only.
     * - <B>CMYtot_p</B> = The moment coefficient about the y-axis - pressure contribution only.
     * - <B>CMZtot_p</B> = The moment coefficient about the z-axis - pressure contribution only.
     * - <B>CXtot_p</B> = The force coefficient about the x-axis - pressure contribution only.
     * - <B>CYtot_p</B> = The force coefficient about the y-axis - pressure contribution only.
     * - <B>CZtot_p</B> = The force coefficient about the z-axis - pressure contribution only.
     * .
     *
     */

    // Viscous Forces
    else if (index == 1 + 2*numOutVars) *aoname = EG_strdup("CLtot_v");
    else if (index == 2 + 2*numOutVars) *aoname = EG_strdup("CDtot_v");
    else if (index == 3 + 2*numOutVars) *aoname = EG_strdup("CMXtot_v");
    else if (index == 4 + 2*numOutVars) *aoname = EG_strdup("CMYtot_v");
    else if (index == 5 + 2*numOutVars) *aoname = EG_strdup("CMZtot_v");
    else if (index == 6 + 2*numOutVars) *aoname = EG_strdup("CXtot_v");
    else if (index == 7 + 2*numOutVars) *aoname = EG_strdup("CYtot_v");
    else if (index == 8 + 2*numOutVars) *aoname = EG_strdup("CZtot_v");
    else {

        printf(" fun3dAIM/aimOutputs instance = %d  index = %d NOT Found!\n", iIndex, index);
        return CAPS_NOTFOUND;
    }

    /*! \page aimOutputsFUN3D
     * Viscous Forces:
     * - <B>CLtot_v</B> = The lift coefficient - viscous contribution only.
     * - <B>CDtot_v</B> = The drag coefficient - viscous contribution only.
     * - <B>CMXtot_v</B> = The moment coefficient about the x-axis - viscous contribution only.
     * - <B>CMYtot_v</B> = The moment coefficient about the y-axis - viscous contribution only.
     * - <B>CMZtot_v</B> = The moment coefficient about the z-axis - viscous contribution only.
     * - <B>CXtot_v</B> = The force coefficient about the x-axis - viscous contribution only.
     * - <B>CYtot_v</B> = The force coefficient about the y-axis - viscous contribution only.
     * - <B>CZtot_v</B> = The force coefficient about the z-axis - viscous contribution only.
     */

    //if (index <= 3*numOutVars) {
    form->type   = Double;
    form->dim    = 1;
    form->length = 1;
    form->nrow   = 1;
    form->ncol   = 1;
    form->units  = NULL;
    //}

    return CAPS_SUCCESS;
}

// Calculate FUN3D output
int aimCalcOutput(int iIndex, /*@unused@*/ void *aimInfo, char *analysisPath,
              	  int index, capsValue *val, capsErrs **errors)
{
	size_t linecap = 0;

	char currentPath[PATH_MAX]; // Current directory path
	char *strKeyword = NULL, *bndSectionKeyword = NULL, *bndSubSectionKeyword = NULL; // Keyword strings

	char *strValue;    // Holder string for value of keyword
	char *line = NULL; // Temporary line holder

	char *filename = NULL; // File to open
	char fileExtension[] = ".forces";

	int bndSectionFound = (int) false;
	int bndSubSectionFound = (int) false;

	FILE *fp; // File pointer

	int numOutVars = 8; // Grouped

	#ifdef DEBUG
		printf(" fun3dAIM/aimCalcOutput instance = %d  index = %d!\n", iIndex, index);
	#endif

	*errors        = NULL;
	val->vals.real = 0.0; // Set default value

	//printf("index = %d\n",index);

	// Set the "search" string(s)
	if (index <= numOutVars) {

		bndSectionKeyword  = (char *) " FORCE TOTALS FOR ALL BOUNDARIES\n";
		bndSubSectionKeyword = (char *) " Total forces\n";

	} else if (index > numOutVars && index <= 2*numOutVars) {

		bndSectionKeyword  = (char *) " FORCE TOTALS FOR ALL BOUNDARIES\n";
		bndSubSectionKeyword = (char *) " Pressure forces\n";

	} else if (index > 2*numOutVars && index <= 3*numOutVars) {

		bndSectionKeyword  = (char *) " FORCE TOTALS FOR ALL BOUNDARIES\n";
		bndSubSectionKeyword = (char *) " Viscous forces\n";
	}

	if      (index == 1 || index == 1 + numOutVars || index == 1 + 2*numOutVars) strKeyword = (char *) "Cl  =";
	else if (index == 2 || index == 2 + numOutVars || index == 2 + 2*numOutVars) strKeyword = (char *) "Cd  =";
	else if (index == 3 || index == 3 + numOutVars || index == 3 + 2*numOutVars) strKeyword = (char *) "Cmx =";
	else if (index == 4 || index == 4 + numOutVars || index == 4 + 2*numOutVars) strKeyword = (char *) "Cmy =";
	else if (index == 5 || index == 5 + numOutVars || index == 5 + 2*numOutVars) strKeyword = (char *) "Cmz =";
	else if (index == 6 || index == 6 + numOutVars || index == 6 + 2*numOutVars) strKeyword = (char *) "Cx  =";
	else if (index == 7 || index == 7 + numOutVars || index == 7 + 2*numOutVars) strKeyword = (char *) "Cy  =";
	else if (index == 8 || index == 8 + numOutVars || index == 8 + 2*numOutVars) strKeyword = (char *) "Cz  =";
	else {
		printf("Unreconized output variable index - %d\n", index);
		return CAPS_BADINDEX;
	}

	(void) getcwd(currentPath, PATH_MAX); // Get our current path

	// Set path to analysis directory
	if (chdir(analysisPath) != 0) {
		#ifdef DEBUG
			printf(" fun3dAIM/aimCalcOutput Cannot chdir to %s!\n", analysisPath);
		#endif

		return CAPS_DIRERR;
	}

	// Open fun3d *.force file
	filename = (char *) EG_alloc((strlen(fun3dInstance[iIndex].projectName) + strlen(fileExtension) +1)*sizeof(char));
	if (filename == NULL) return EGADS_MALLOC;

	sprintf(filename, "%s%s", fun3dInstance[iIndex].projectName, fileExtension);

	fp = fopen(filename, "r");

	if (filename != NULL) EG_free(filename); // Free filename allocation

	if (fp == NULL) {
		#ifdef DEBUG
			printf(" fun3dAIM/aimCalcOutput Cannot open Output file!\n");
		#endif

		chdir(currentPath);

		return CAPS_IOERR;
	}

	// Scan the file for the string
	strValue = NULL;
	while (getline(&line, &linecap, fp) >= 0) {

		if (line == NULL) continue;

		if (strcmp(line,bndSectionKeyword) == 0 && bndSectionFound == (int) false) {

			//printf("FOUND section\n");
			bndSectionFound = (int) true;
			continue; // Skip to next line
		}

		if (strcmp(line, bndSubSectionKeyword) == 0 && bndSectionFound    == (int) true
													&& bndSubSectionFound == (int) false) {

			//printf("Found sub-sections\n");
			bndSubSectionFound = (int) true;
			continue; // Skipe to next line
		}

		if (bndSectionFound == (int) true && bndSubSectionFound == (int) true){

			strValue = strstr(line, strKeyword);

			if (strValue != NULL) {

				//printf("String value %s\n", strValue);

				// Found it -- get the value
				sscanf(&strValue[6], "%lf", &val->vals.real);
				break;
			}
		}
	}

	// Restore the path we came in with
	chdir(currentPath);
	fclose(fp);
	if (line != NULL) free(line);

	if (strValue == NULL) {
		#ifdef DEBUG
			printf(" fun3dAIM/aimCalcOutput Cannot find %s in Output file!\n", strKeyword);
		#endif

		return CAPS_NOTFOUND;
	}

	return CAPS_SUCCESS;
}

/************************************************************************/
// CAPS transferring functions

int aimFreeDiscr(capsDiscr *discr)
{
    int iIndex, i;

    int status; // Function return

    #ifdef DEBUG
        if (discr == NULL) {
            printf(" fun3dAIM/aimFreeDiscr w/ NULL!\n");
        } else {
            printf(" fun3dAIM/aimFreeDiscr instance = %d!\n", discr->instance);
        }
    #endif

    if (discr == NULL) {

        // Clean up fun3dInstance data
        for ( iIndex = 0; iIndex < numInstance; iIndex++) {

            printf(" Cleaning up fun3dInstance - %d\n", iIndex);

            // Attribute to index map
            status = destroy_mapAttrToIndexStruct(&fun3dInstance[iIndex].attrMap);
            if (status != CAPS_SUCCESS) return status;

            // FUN3D project name
            fun3dInstance[iIndex].projectName = NULL;

            // Analysis file path/directory
            fun3dInstance[iIndex].analysisPath = NULL;

            // Pointer to caps input value for scaling pressure during data transfer
            fun3dInstance[iIndex].pressureScaleFactor = NULL;

            // Pointer to caps input value for offset pressure during data transfer
            fun3dInstance[iIndex].pressureScaleOffset = NULL;

        }

        if (fun3dInstance != NULL) EG_free(fun3dInstance);
        fun3dInstance = NULL;
        numInstance = 0;

    } else {

        // Free up this capsDiscr

    	discr->nPoints = 0; // Points

    	if (discr->mapping != NULL) EG_free(discr->mapping);
    	discr->mapping = NULL;

        if (discr->types != NULL) { // Element types
        	for (i = 0; i < discr->nTypes; i++) {
        		if (discr->types[i].gst  != NULL) EG_free(discr->types[i].gst);
        		if (discr->types[i].tris != NULL) EG_free(discr->types[i].tris);
        	}

        	EG_free(discr->types);
        }

        discr->nTypes  = 0;
        discr->types   = NULL;

        if (discr->elems != NULL) { // Element connectivity

        	for (i = 0; i < discr->nElems; i++) {
        		if (discr->elems[i].gIndices != NULL) EG_free(discr->elems[i].gIndices);
        	}

        	EG_free(discr->elems);
        }

        discr->nElems  = 0;
        discr->elems   = NULL;

        if (discr->ptrm != NULL) EG_free(discr->ptrm); // Extra information to store into the discr void pointer
        discr->ptrm    = NULL;


        discr->nVerts = 0;    // Not currently used
        discr->verts  = NULL; // Not currently used
        discr->celem = NULL; // Not currently used

        discr->nDtris = 0; // Not currently used
        discr->dtris  = NULL; // Not currently used
    }

    return CAPS_SUCCESS;
}

int aimDiscr(char *tname, capsDiscr *discr) {

	int i, body, face; // Indexing

	int status; // Function return status

	int iIndex; // Instance index

	int numBody;

	// EGADS objects
	ego tess, *bodies = NULL, *faces = NULL, tempBody;

	const char   *string, *capsGroup; // capsGroups strings

	// EGADS function returns
	int plen, tlen, elen;
	const int    *ptype, *pindex, *tris, *nei;
	const double *xyz, *uv;

	// Body Tessellation
	int numFace = 0;
	int numFaceFound = 0;
	int numPoint = 0, numTri = 0, numGlobalPoint = 0;
	int *bodyFaceMap = NULL; // size=[2*numFaceFound], [2*numFaceFound + 0] = body, [2*numFaceFoun + 1] = face

	int *globalID = NULL, *localStitchedID = NULL, gID = 0;

	int nodeOffSet = 0;

	int *storage= NULL; // Extra information to store into the discr void pointer

	int numCAPSGroup = 0, attrIndex = 0, foundAttr = (int) false;
	int *capsGroupList = NULL;
	int dataTransferBodyIndex=-99;

	// Data transfer
	int  nrow, ncol, rank;
	void *dataTransfer = NULL;
	enum capsvType vtype;
	char *units;

	// Volume Mesh obtained from meshing AIM
	genUnstrMeshStruct volumeMesh;

	iIndex = discr->instance;
	#ifdef DEBUG
		printf(" fun3dAIM/aimDiscr: tname = %s, instance = %d!\n", tname, iIndex);
	#endif

	if ((iIndex < 0) || (iIndex >= numInstance)) return CAPS_BADINDEX;


	if (tname == NULL) return CAPS_NOTFOUND;

	if (fun3dInstance[iIndex].dataTransferCheck == (int) false) {
	    printf("The volume is not suitable for data transfer - possibly the volume mesher "
	            "added unaccounted for points\n");
	    return CAPS_BADVALUE;
	}

	// Currently this ONLY works if the capsTranfer lives on single body!
	status = aim_getBodies(discr->aInfo, &numBody, &bodies);
	if (status != CAPS_SUCCESS) {
		printf(" fun3dAIM/aimDiscr: %d aim_getBodies = %d!\n", iIndex, status);
		return status;
	}

	status = aimFreeDiscr(discr);
	if (status != CAPS_SUCCESS) return status;

	numFaceFound = 0;
	numPoint = numTri = 0;
	// Find any faces with our boundary marker and get how many points and triangles there are
	for (body = 0; body < numBody; body++) {

		status = EG_getBodyTopos(bodies[body], NULL, FACE, &numFace, &faces);
		if (status != EGADS_SUCCESS) {
			printf(" fun3dAIM: getBodyTopos (Face) = %d for Body %d!\n", status, body);
			return status;
		}

		for (face = 0; face < numFace; face++) {

			// Retrieve the string following a capsTransfer tag
			status = retrieve_CAPSTransferAttr(faces[face], &string);
			if (status != CAPS_SUCCESS) continue;

			if (strcmp(string, tname) != 0) continue;

			#ifdef DEBUG
				printf(" fun3dAIM/aimDiscr: Body %d/Face %d matches %s!\n", body, face+1, tname);
			#endif


			status = retrieve_CAPSGroupAttr(faces[face], &capsGroup);
			if (status != CAPS_SUCCESS) {
			    printf("capsTransfer found on face %d, but no capGroup found!!!\n", face);
			    continue;
			} else {

			    status = get_mapAttrToIndexIndex(&fun3dInstance[iIndex].attrMap, capsGroup, &attrIndex);
			    if (status != CAPS_SUCCESS) {
			        printf("capsGroup %s NOT found in attrMap\n",capsGroup);
			        continue;
			    } else {

			        // If first index create arrays and store index
			        if (numCAPSGroup == 0) {
			            numCAPSGroup += 1;
			            capsGroupList = (int *) EG_alloc(numCAPSGroup*sizeof(int));
			            if (capsGroupList == NULL) {
			                status =  EGADS_MALLOC;
			                goto cleanup;
			            }

			            capsGroupList[numCAPSGroup-1] = attrIndex;
			        } else { // If we already have an index(es) let make sure it is unique
			            foundAttr = (int) false;
			            for (i = 0; i < numCAPSGroup; i++) {
			                if (attrIndex == capsGroupList[i]) {
			                    foundAttr = (int) true;
			                    break;
			                }
			            }

			            if (foundAttr == (int) false) {
			                numCAPSGroup += 1;
			                capsGroupList = (int *) EG_reall(capsGroupList, numCAPSGroup*sizeof(int));
			                if (capsGroupList == NULL) {
			                    status =  EGADS_MALLOC;
			                    goto cleanup;
			                }

			                capsGroupList[numCAPSGroup-1] = attrIndex;
			            }
			        }
			    }
			}

			// Get face tessellation
			status = EG_getTessFace(bodies[body+numBody], face+1, &plen, &xyz, &uv, &ptype, &pindex, &tlen, &tris, &nei);
			if (status != EGADS_SUCCESS) {
				printf(" fun3dAIM: EG_getTessFace %d = %d for Body %d!\n", face+1, status, body+1);
				continue;
			}

			numFaceFound += 1;
			dataTransferBodyIndex = body;
			if (numFaceFound == 1) {
				bodyFaceMap = (int *) EG_alloc(2*numFaceFound*sizeof(int));
			} else {
				bodyFaceMap = (int *) EG_reall(bodyFaceMap, 2*numFaceFound*sizeof(int));
			}

			if (bodyFaceMap == NULL) {
				status = EGADS_MALLOC;
				goto cleanup;
			}

			// Get number of points and triangles
			bodyFaceMap[2*(numFaceFound-1) + 0] = body+1;
			bodyFaceMap[2*(numFaceFound-1) + 1] = face+1;

			// Sum number of points and triangles
			numPoint  += plen;
			numTri  += tlen;
		}

		if (faces != NULL) EG_free(faces);
		faces = NULL;


	    if (dataTransferBodyIndex >=0) break; // Force that only one body can be used
	}

	if (numFaceFound == 0) {
		printf(" fun3dAIM/aimDiscr: No Faces match %s!\n", tname);

		status = CAPS_NOTFOUND;
		goto cleanup;
	}

	if ( numPoint == 0 || numTri == 0) {
		#ifdef DEBUG
	    	printf(" fun3dAIM/aimDiscr: ntris = %d, npts = %d!\n", numTri, numPoint);
		#endif
	    status = CAPS_SOURCEERR;
	    goto cleanup;
	}

	status = aim_getData(discr->aInfo, "Volume_Mesh", &vtype, &rank, &nrow, &ncol, &dataTransfer, &units);
	if (status != CAPS_SUCCESS){
	    if (status == CAPS_DIRTY) printf("Parent AIMS are dirty\n");
	    else printf("Linking statusus = %d\n",status);

	    printf("Didn't find a link to a volume mesh (Volume_Mesh) from parent - data transfer isn't possible.\n");

	    goto cleanup;

	} else {

	    volumeMesh = *(genUnstrMeshStruct *) dataTransfer;
	}

	#ifdef DEBUG
		printf(" fun3dAIM/aimDiscr: Instance %d, Body Index for data transfer = %d\n", iIndex, dataTransferBodyIndex);
	#endif


	// Specify our single element type
	status = EGADS_MALLOC;
	discr->nTypes = 1;

	discr->types  = (capsEleType *) EG_alloc(sizeof(capsEleType));
	if (discr->types == NULL) goto cleanup;
	discr->types[0].nref  = 3;
	discr->types[0].ndata = 0;            /* data at geom reference positions */
	discr->types[0].ntri  = 1;
	discr->types[0].nmat  = 0;            /* match points at geom ref positions */
	discr->types[0].tris  = NULL;
	discr->types[0].gst   = NULL;
	discr->types[0].dst   = NULL;
	discr->types[0].matst = NULL;

	discr->types[0].tris   = (int *) EG_alloc(3*sizeof(int));
	if (discr->types[0].tris == NULL) goto cleanup;
	discr->types[0].tris[0] = 1;
	discr->types[0].tris[1] = 2;
	discr->types[0].tris[2] = 3;

	discr->types[0].gst   = (double *) EG_alloc(6*sizeof(double));
	if (discr->types[0].gst == NULL) goto cleanup;
	discr->types[0].gst[0] = 0.0;
	discr->types[0].gst[1] = 0.0;
	discr->types[0].gst[2] = 1.0;
	discr->types[0].gst[3] = 0.0;
	discr->types[0].gst[4] = 0.0;
	discr->types[0].gst[5] = 1.0;

	// Get the tessellation and make up a simple linear continuous triangle discretization */

	discr->nElems = numTri;

	discr->elems = (capsElement *) EG_alloc(discr->nElems*sizeof(capsElement));
	if (discr->elems == NULL) {
		status = EGADS_MALLOC;
		goto cleanup;
	}

	discr->mapping = (int *) EG_alloc(2*numPoint*sizeof(int)); // Will be resized
	if (discr->mapping == NULL) goto cleanup;

	globalID = (int *) EG_alloc(numPoint*sizeof(int));
	if (globalID == NULL) {
		status = EGADS_MALLOC;
		goto cleanup;
	}

	numPoint = 0;
	numTri = 0;
	for (face = 0; face < numFaceFound; face++){

		tess = bodies[bodyFaceMap[2*face + 0]-1 + numBody];

		if (localStitchedID == NULL) {
			status = EG_statusTessBody(tess, &tempBody, &i, &numGlobalPoint);

			localStitchedID = (int *) EG_alloc(numGlobalPoint*sizeof(int));
			if (localStitchedID == NULL) {
				status = EGADS_MALLOC;
				goto cleanup;
			}

			for (i = 0; i < numGlobalPoint; i++) localStitchedID[i] = 0;
		}

		// Get face tessellation
		status = EG_getTessFace(tess, bodyFaceMap[2*face + 1], &plen, &xyz, &uv, &ptype, &pindex, &tlen, &tris, &nei);
		if (status != EGADS_SUCCESS) {
			printf(" fun3dAIM: EG_getTessFace %d = %d for Body %d!\n", bodyFaceMap[2*face + 1], status, bodyFaceMap[2*face + 0]);
			continue;
		}

		for (i = 0; i < plen; i++ ) {

			status = EG_localToGlobal(tess, bodyFaceMap[2*face+1], i+1, &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			if (localStitchedID[gID -1] != 0) continue;

			discr->mapping[2*numPoint  ] = bodyFaceMap[2*face + 0];
			discr->mapping[2*numPoint+1] = gID;

			localStitchedID[gID -1] = numPoint+1;

			globalID[numPoint] = gID;

			numPoint += 1;

		}

		// Get triangle connectivity in global sense
		for (i = 0; i < tlen; i++) {

			discr->elems[numTri].bIndex      = bodyFaceMap[2*face + 0];
			discr->elems[numTri].tIndex      = 1;
			discr->elems[numTri].eIndex      = bodyFaceMap[2*face + 1];

			discr->elems[numTri].gIndices    = (int *) EG_alloc(6*sizeof(int));
			if (discr->elems[numTri].gIndices == NULL) {
				status = EGADS_MALLOC;
				goto cleanup;
			}

			discr->elems[numTri].dIndices    = NULL;
			discr->elems[numTri].eTris.tq[0] = i+1;

			status = EG_localToGlobal(tess, bodyFaceMap[2*face + 1], tris[3*i + 0], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			discr->elems[numTri].gIndices[0] = localStitchedID[gID-1];
			discr->elems[numTri].gIndices[1] = tris[3*i + 0];

			status = EG_localToGlobal(tess, bodyFaceMap[2*face + 1], tris[3*i + 1], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			discr->elems[numTri].gIndices[2] = localStitchedID[gID-1];
			discr->elems[numTri].gIndices[3] = tris[3*i + 1];

			status = EG_localToGlobal(tess, bodyFaceMap[2*face + 1], tris[3*i + 2], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			discr->elems[numTri].gIndices[4] = localStitchedID[gID-1];
			discr->elems[numTri].gIndices[5] = tris[3*i + 2];

			numTri += 1;
		}
	}

	discr->nPoints = numPoint;

	#ifdef DEBUG
		printf(" fun3dAIM/aimDiscr: ntris = %d, npts = %d!\n", discr->nElems, discr->nPoints);
	#endif

	// Resize mapping to switched together number of points
	discr->mapping = (int *) EG_reall(discr->mapping, 2*numPoint*sizeof(int));

	// Local to global node connectivity + numCAPSGroup + sizeof(capGrouplist)
	storage  = (int *) EG_alloc((numPoint + 1 + numCAPSGroup) *sizeof(int));
	if (storage == NULL) goto cleanup;
	discr->ptrm = storage;

	// Store the local-to-Global (volume) id
	nodeOffSet = 0;
	for (i = 0; i < dataTransferBodyIndex; i++) {
	    nodeOffSet += volumeMesh.surfaceMesh.surfNodesLump[i];
	}

	#ifdef DEBUG
		printf(" fun3dAIM/aimDiscr: Instance = %d, nodeOffSet = %d\n", iIndex, nodeOffSet);
	#endif

	for (i = 0; i < numPoint; i++) {

		storage[i] = volumeMesh.surfaceMesh.nodeLocal2Global[(globalID[i]-1) + nodeOffSet];

	    //#ifdef DEBUG
		//	printf(" fun3dAIM/aimDiscr: Instance = %d, Global Node ID %d\n", iIndex, storage[i]);
		//#endif
	}

	// Save way the attrMap capsGroup list
	storage[numPoint] = numCAPSGroup;
	for (i = 0; i < numCAPSGroup; i++) {
	    storage[numPoint+1+i] = capsGroupList[i];
	}

	#ifdef DEBUG
        printf(" fun3dAIM/aimDiscr: Instance = %d, Finished!!\n", iIndex);
    #endif

	status = CAPS_SUCCESS;

	goto cleanup;

	cleanup:
		if (faces != NULL) EG_free(faces);

		if (globalID  != NULL) EG_free(globalID);
		if (localStitchedID != NULL) EG_free(localStitchedID);

		if (capsGroupList != NULL) EG_free(capsGroupList);
		if (bodyFaceMap != NULL) EG_free(bodyFaceMap);

		if (status != CAPS_SUCCESS) {
			aimFreeDiscr(discr);
		}

		return status;
}

int aimLocateElement(capsDiscr *discr, double *params, double *param, int *eIndex,
                 	 double *bary)
{
	int    i, in[3], stat, ismall;
	double we[3], w, smallw = -1.e300;

	if (discr == NULL) return CAPS_NULLOBJ;

	for (ismall = i = 0; i < discr->nElems; i++) {
		in[0] = discr->elems[i].gIndices[0] - 1;
		in[1] = discr->elems[i].gIndices[2] - 1;
		in[2] = discr->elems[i].gIndices[4] - 1;
		stat  = EG_inTriExact(&params[2*in[0]], &params[2*in[1]], &params[2*in[2]], param, we);

		if (stat == EGADS_SUCCESS) {
			*eIndex = i+1;
			bary[0] = we[1];
			bary[1] = we[2];
			return CAPS_SUCCESS;
		}

		w = we[0];
		if (we[1] < w) w = we[1];
		if (we[2] < w) w = we[2];
		if (w > smallw) {
			ismall = i+1;
			smallw = w;
		}
	}

	/* must extrapolate! */
	if (ismall == 0) return CAPS_NOTFOUND;
	in[0] = discr->elems[ismall-1].gIndices[0] - 1;
	in[1] = discr->elems[ismall-1].gIndices[2] - 1;
	in[2] = discr->elems[ismall-1].gIndices[4] - 1;
	EG_inTriExact(&params[2*in[0]], &params[2*in[1]], &params[2*in[2]], param, we);

	*eIndex = ismall;
	bary[0] = we[1];
	bary[1] = we[2];

	/*
  	  printf(" aimLocateElement: extropolate to %d (%lf %lf %lf)  %lf\n", ismall, we[0], we[1], we[2], smallw);
	*/
	return CAPS_SUCCESS;
}

int aimTransfer(capsDiscr *discr, char *dataName, int numPoint, int dataRank, double *dataVal, /*@unused@*/ char **units)
{
    /*! \page dataTransferFUN3D FUN3D Data Transfer
     * Data transfer blah, blah, blah....
     */

	int status, status2; // Function return status
	int i, j, dataPoint; // Indexing

    // Current directory path
	char   currentPath[PATH_MAX];

	// Aero-Load data variables
    int numVariable;
    int numDataPoint;
    char **variableName = NULL;
    double **dataMatrix = NULL;
    double dataScaleFactor = 1.0;
    double dataScaleOffset = 0.0;
    char *dataUnits;

    // Indexing in data variables
    int globalIDIndex = -99;
    int variableIndex = -99;

    // Variables used in global node mapping
    int *nodeMap, *storage;
    int globalNodeID;
    int found = (int) false;

    // Filename stuff
    int *capsGroupList;
    char *filename = NULL; //"pyCAPS_FUN3D_Tetgen_ddfdrive_bndry1.dat";

	#ifdef DEBUG
		printf(" fun3dAIM/aimTransfer name = %s  instance = %d  npts = %d/%d!\n", dataName, discr->instance, numPoint, dataRank);
	#endif


	if (strcasecmp(dataName, "Pressure") != 0 &&
		strcasecmp(dataName, "P")        != 0 &&
		strcasecmp(dataName, "Cp")       != 0 &&
		strcasecmp(dataName, "CoeffientofPressure") != 0) {

		printf("Unrecognized data transfer variable - %s\n", dataName);
		return CAPS_NOTFOUND;
	}

	//Get the appropriate parts of the tessellation to data
	storage = (int *) discr->ptrm;
	nodeMap = &storage[0]; // Global indexing on the body

	capsGroupList = &storage[discr->nPoints]; // List of boundary ID (attrMap) in the transfer

	// Zero out data
	for (i = 0; i < numPoint; i++) {
		for (j = 0; j < dataRank; j++ ) {
			dataVal[dataRank*i+j] = 0;
		}
	}

	for (i = 0; i < capsGroupList[0]; i++) {

	    (void) getcwd(currentPath, PATH_MAX);
	    if (chdir(fun3dInstance[discr->instance].analysisPath) != 0) {
	        #ifdef DEBUG
	            printf(" fun3dAIM/aimTransfer Cannot chdir to %s!\n", fun3dInstance[discr->instance].analysisPath);
	        #endif

	        return CAPS_DIRERR;
	    }


	    filename = (char *) EG_alloc((strlen(fun3dInstance[discr->instance].projectName) +
	                                  strlen("_ddfdrive_bndry.dat") + 5)*sizeof(char));

	    sprintf(filename,"%s%s%d%s",fun3dInstance[discr->instance].projectName,
	                                "_ddfdrive_bndry",
	                                capsGroupList[i+1],
	                                ".dat");

	    status = fun3d_readAeroLoad(filename,
	                                &numVariable,
	                                &variableName,
	                                &numDataPoint,
	                                &dataMatrix);

	    if (filename != NULL) EG_free(filename);
	    filename = NULL;

	    // Restore the path we came in with
	    chdir(currentPath);
	    if (status != CAPS_SUCCESS) return status;

	    printf("Number of variables %d\n", numVariable);
	    // Output some of the first row of the data
	    //for (i = 0; i < numVariable; i++) printf("Variable %d - %.6f\n", i, dataMatrix[i][0]);

	    // Loop through the variable list to find which one is the global node ID variable
	    for (i = 0; i < numVariable; i++) {
	        if (strcasecmp("id", variableName[i]) ==0 ) {
	            globalIDIndex = i;
	            break;
	        }
	    }

	    if (globalIDIndex == -99) {
	        printf("Global node number variable not found in data file\n");
	        status = CAPS_NOTFOUND;
	        goto bail;
	    }

	    // Loop through the variable list to see if we can find the transfer data name
	    for (i = 0; i < numVariable; i++) {

	        if (strcasecmp(dataName, "Pressure") == 0 ||
	        	strcasecmp(dataName, "P")        == 0 ||
	            strcasecmp(dataName, "Cp")       == 0 ||
				strcasecmp(dataName, "CoeffientofPressure") == 0) {

	            if (strcasecmp("cp", variableName[i]) == 0) {
	                variableIndex = i;
	                break;
	            }

	            if (dataRank != 1) {
	            	printf("Data transfer rank should be 1 not %d\n", dataRank);
	            	status = CAPS_BADRANK;
	            	goto bail;
	            }

	            dataScaleFactor = fun3dInstance[discr->instance].pressureScaleFactor->vals.real;
	            dataScaleOffset = fun3dInstance[discr->instance].pressureScaleOffset->vals.real;

	            dataUnits = fun3dInstance[discr->instance].pressureScaleFactor->units;
	        }
	    }

	    if (variableIndex == -99) {
	        printf("Variable %s not found in data file\n", dataName);
	        status = CAPS_NOTFOUND;
	        goto bail;
	    }

	    for (i = 0; i < numPoint; i++) {

	    	globalNodeID = nodeMap[i];

	    	found = (int) false;
	    	for (dataPoint = 0; dataPoint < numDataPoint; dataPoint++) {
	    		if ((int) dataMatrix[globalIDIndex][dataPoint] ==  globalNodeID) {
	    			found = (int) true;
	    			break;
	    		}
	    	}

	    	if (found == (int) true) {
				for (j = 0; j < dataRank; j++) {

					// Add something for units - aim_covert()

					dataVal[dataRank*i+j] = dataMatrix[variableIndex][dataPoint]*dataScaleFactor + dataScaleOffset;
					//printf("DataValue = %f\n",dataVal[dataRank*i+j]);
					//dataVal[dataRank*i+j] = 99;

				}
	    	}
	    }


	    // Free data matrix
	    if (dataMatrix != NULL) {
	        for (i = 0; i < numVariable; i++) {
	            if (dataMatrix[i] != NULL) EG_free(dataMatrix[i]);
	        }

	        EG_free(dataMatrix);
	    }

	    // Free variable list
	    status = string_freeArray(numVariable, &variableName);
	    if (status != CAPS_SUCCESS) return status;

	}

	return CAPS_SUCCESS;

	bail:
        // Free data matrix
        if (dataMatrix != NULL) {
            for (i = 0; i < numVariable; i++) {
                if (dataMatrix[i] != NULL) EG_free(dataMatrix[i]);
            }

            EG_free(dataMatrix);
        }

        // Free variable list
        status2 = string_freeArray(numVariable, &variableName);
        if (status2 != CAPS_SUCCESS) return status2;

        return status;
}

int aimInterpolation(capsDiscr *discr, /*@unused@*/ char *name, int eIndex,
                 double *bary, int rank, double *data, double *result)
{
	int    in[3], i;
	double we[3];
	/*
	#ifdef DEBUG
  	  printf(" fun3dAIM/aimInterpolation  %s  instance = %d!\n",
         name, discr->instance);
	#endif
	 */
	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" fun3dAIM/Interpolation: eIndex = %d [1-%d]!\n",eIndex, discr->nElems);
	}

	we[0] = 1.0 - bary[0] - bary[1];
	we[1] = bary[0];
	we[2] = bary[1];
	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[2] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[4] - 1;
	for (i = 0; i < rank; i++){
		result[i] = data[rank*in[0]+i]*we[0] + data[rank*in[1]+i]*we[1] + data[rank*in[2]+i]*we[2];
	}

	return CAPS_SUCCESS;
}


int aimInterpolateBar(capsDiscr *discr, /*@unused@*/ char *name, int eIndex,
					  double *bary, int rank, double *r_bar, double *d_bar)
{
	int    in[3], i;
	double we[3];

	/*
	#ifdef DEBUG
  	  printf(" fun3dAIM/aimInterpolateBar  %s  instance = %d!\n", name, discr->instance);
	#endif
	 */

	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" fun3dAIM/InterpolateBar: eIndex = %d [1-%d]!\n", eIndex, discr->nElems);
	}

	we[0] = 1.0 - bary[0] - bary[1];
	we[1] = bary[0];
	we[2] = bary[1];
	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[2] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[4] - 1;

	for (i = 0; i < rank; i++) {
		/*  result[i] = data[rank*in[0]+i]*we[0] + data[rank*in[1]+i]*we[1] +
                data[rank*in[2]+i]*we[2];  */
		d_bar[rank*in[0]+i] += we[0]*r_bar[i];
		d_bar[rank*in[1]+i] += we[1]*r_bar[i];
		d_bar[rank*in[2]+i] += we[2]*r_bar[i];
	}

	return CAPS_SUCCESS;
}


int aimIntegration(capsDiscr *discr, /*@unused@*/ char *name, int eIndex, int rank,
				   /*@null@*/ double *data, double *result)
{
	int    i, in[3], stat, ptype, pindex, nBody;
	double x1[3], x2[3], x3[3], xyz1[3], xyz2[3], xyz3[3], area;
	ego    *bodies;

	/*
	#ifdef DEBUG
  	  printf(" fun3dAIM/aimIntegration  %s  instance = %d!\n", name, discr->instance);
	#endif
	 */
	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" fun3dAIM/aimIntegration: eIndex = %d [1-%d]!\n", eIndex, discr->nElems);
	}

	stat = aim_getBodies(discr->aInfo, &nBody, &bodies);
	if (stat != CAPS_SUCCESS) {
		printf(" fun3dAIM/aimIntegration: %d aim_getBodies = %d!\n", discr->instance, stat);
		return stat;
	}

	/* element indices */

	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[2] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[4] - 1;

	stat = EG_getGlobal(bodies[discr->mapping[2*in[0]]+nBody-1],
						discr->mapping[2*in[0]+1], &ptype, &pindex, xyz1);
	if (stat != CAPS_SUCCESS) {
		printf(" fun3dAIM/aimIntegration: %d EG_getGlobal %d = %d!\n", discr->instance, in[0], stat);
		return stat;
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[1]]+nBody-1],
                        discr->mapping[2*in[1]+1], &ptype, &pindex, xyz2);
	if (stat != CAPS_SUCCESS) {
		printf(" fun3dAIM/aimIntegration: %d EG_getGlobal %d = %d!\n", discr->instance, in[1], stat);
    	return stat;
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[2]]+nBody-1],
                        discr->mapping[2*in[2]+1], &ptype, &pindex, xyz3);
	if (stat != CAPS_SUCCESS) {
		printf(" fun3dAIM/aimIntegration: %d EG_getGlobal %d = %d!\n", discr->instance, in[2], stat);
		return stat;
	}

	x1[0] = xyz2[0] - xyz1[0];
	x2[0] = xyz3[0] - xyz1[0];
	x1[1] = xyz2[1] - xyz1[1];
	x2[1] = xyz3[1] - xyz1[1];
	x1[2] = xyz2[2] - xyz1[2];
	x2[2] = xyz3[2] - xyz1[2];

	//CROSS(x3, x1, x2);
	cross_DoubleVal(x1, x2, x3);

	//area  = sqrt(DOT(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */
	area  = sqrt(dot_DoubleVal(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */

	if (data == NULL) {
		*result = 3.0*area;
		return CAPS_SUCCESS;
	}

	for (i = 0; i < rank; i++) {
		result[i] = (data[rank*in[0]+i] + data[rank*in[1]+i] + data[rank*in[2]+i])*area;
	}

	return CAPS_SUCCESS;
}


int aimIntegrateBar(capsDiscr *discr, /*@unused@*/ char *name, int eIndex, int rank,
                    double *r_bar, double *d_bar)
{
	int    i, in[3], stat, ptype, pindex, nBody;
	double x1[3], x2[3], x3[3], xyz1[3], xyz2[3], xyz3[3], area;
	ego    *bodies;

	/*
	#ifdef DEBUG
  	  printf(" fun3dAIM/aimIntegrateBar  %s  instance = %d!\n", name, discr->instance);
	#endif
	 */

	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" fun3dAIM/aimIntegrateBar: eIndex = %d [1-%d]!\n", eIndex, discr->nElems);
	}

	stat = aim_getBodies(discr->aInfo, &nBody, &bodies);
	if (stat != CAPS_SUCCESS) {
		printf(" fun3dAIM/aimIntegrateBar: %d aim_getBodies = %d!\n", discr->instance, stat);
		return stat;
	}

	/* element indices */

	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[2] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[4] - 1;
	stat = EG_getGlobal(bodies[discr->mapping[2*in[0]]+nBody-1],
                       discr->mapping[2*in[0]+1], &ptype, &pindex, xyz1);
	if (stat != CAPS_SUCCESS) {
		printf(" fun3dAIM/aimIntegrateBar: %d EG_getGlobal %d = %d!\n", discr->instance, in[0], stat);
		return stat;
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[1]]+nBody-1],
                       discr->mapping[2*in[1]+1], &ptype, &pindex, xyz2);

	if (stat != CAPS_SUCCESS) {
		printf(" fun3dAIM/aimIntegrateBar: %d EG_getGlobal %d = %d!\n", discr->instance, in[1], stat);
		return stat;
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[2]]+nBody-1],
                      discr->mapping[2*in[2]+1], &ptype, &pindex, xyz3);

	if (stat != CAPS_SUCCESS) {
		printf(" fun3dAIM/aimIntegrateBar: %d EG_getGlobal %d = %d!\n", discr->instance, in[2], stat);
		return stat;
	}

	x1[0] = xyz2[0] - xyz1[0];
	x2[0] = xyz3[0] - xyz1[0];
	x1[1] = xyz2[1] - xyz1[1];
	x2[1] = xyz3[1] - xyz1[1];
	x1[2] = xyz2[2] - xyz1[2];
	x2[2] = xyz3[2] - xyz1[2];

	//CROSS(x3, x1, x2);
	cross_DoubleVal(x1, x2, x3);

	//area  = sqrt(DOT(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */
	area  = sqrt(dot_DoubleVal(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */

	for (i = 0; i < rank; i++) {
	/*  result[i] = (data[rank*in[0]+i] + data[rank*in[1]+i] +
				 data[rank*in[2]+i])*area;  */
		d_bar[rank*in[0]+i] += area*r_bar[i];
		d_bar[rank*in[1]+i] += area*r_bar[i];
		d_bar[rank*in[2]+i] += area*r_bar[i];
	}

	return CAPS_SUCCESS;
}
