/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             mystran AIM
 *
 *     Written by Dr. Ryan Durscher AFRL/RQVC
 *
 */


/*! \mainpage Introduction
 *
 * \section overviewMystran MYSTRAN AIM Overview
 * A module in the Computational Aircraft Prototype Syntheses (CAPS) has been developed to interact (primarily
 * through input files) with the finite element structural solver Mystran.
 *
 * Current issues include:
 *  - A thorough bug testing needs to be undertaken.
 *  .
 *
 * \section temporay BLAH
 *
 * Related pages:
 * - \ref feaMaterial
 * - \ref feaProperty
 * - \ref feaConstraint
 * - \ref feaLoad
 * - \ref feaAnalysis
 * - \ref geomFidelityMystran
 * - \ref aimInputsMystran
 * - \ref aimOutputsMystran
 * - \ref sharableDataMystran
 */
#include <string.h>
#include <math.h>
#include "capsTypes.h"
#include "aimUtil.h"

#include "meshUtils.h"
#include "miscUtils.h"
#include "feaUtils.h"
#include "nastranUtils.h"
#include "mystranUtils.h"

#ifdef WIN32
#define getcwd     _getcwd
#define snprintf   _snprintf
#define strcasecmp stricmp
#define PATH_MAX   _MAX_PATH
#else
#include <unistd.h>
#include <limits.h>
#endif

#define NUMINPUT   11
#define NUMOUTPUT  4

//#define DEBUG

typedef struct {
    // Project name
    char *projectName;

    // Analysis file path/directory
    char *analysisPath;

    //genFEAMeshStruct *feaMesh; // FEA mesh structure
    feaProblemStruct feaProblem;

    // Attribute to index map
    mapAttrToIndexStruct attrMap;

    /*
    // Check to make sure data transfer is ok
    int dataTransferCheck;
    */

    // Flag to see if we need to get an FEA mesh
    int haveFEAMesh;

    // Pointer to caps input value for tessellation parameters
    capsValue *tessParam;

    // Pointer to caps input value for Edge_Point- Min
    capsValue *edgePointMin;

    // Pointer to caps input value for Edge_Point- Max
    capsValue *edgePointMax;

    // Pointer to caps input value for No_Quad_Faces
    capsValue *noQuadFace;

} aimStorage;

static aimStorage *mystranInstance = NULL;
static int         numInstance  = 0;

extern char *EG_strdup(/*@null@*/ const char *str);
extern int   EG_inTriExact(double *t1, double *t2, double *t3, double *p,  double *w );

static int createMesh(int iIndex, void *aimInfo) {

	int status; // Function return status

    int i, body; // Indexing

	// Bodies
	int   numBody; // Number of Bodies
	ego  *bodies;

	// genFEAMeshStruct *feaMesh; // FEA meshes

	// Meshing related variables
	double  tessParam[3];
	int edgePointMin = 2;
	int edgePointMax = 50;
	int noQuadFace = 1;

	// Coordinate system
	mapAttrToIndexStruct coordSystemMap;

	// AIM Data transfer
	int  nrow, ncol, rank;
	void *dataTransfer;
	enum capsvType vtype;
	char *units;

	// Inherited fea/volume mesh related variables
	int numFEAMesh = 0;
	genUnstrMeshStruct volumeMesh;
	int feaMeshInherited = (int) false;

	// Get AIM bodies
	status = aim_getBodies(aimInfo, &numBody, &bodies);

	#ifdef DEBUG
	   	printf(" mystranAIM/createMesh instance = %d  nbody = %d!\n", iIndex, numBody);
	#endif

	if ((numBody <= 0) || (bodies == NULL)) {
		#ifdef DEBUG
		printf(" mystranAIM/createMesh No Bodies!\n");
	#endif

		return CAPS_SOURCEERR;
	}

	// Print number of body warning
	if (numBody != 1) {

		printf("mystranAIM only supports single body sources!!!!\n");
		return CAPS_SOURCEERR;
	}

	// Get CoordSystem attribute to index mapping
	status = initiate_mapAttrToIndexStruct(&coordSystemMap);
	if (status != CAPS_SUCCESS) return status;

	status = create_CoordSystemAttrToIndexMap(numBody,
									   bodies,
									   3, //>2 - search the body, faces, edges, and all the nodes
									   &coordSystemMap);
	if (status != CAPS_SUCCESS) goto cleanup;

	status = fea_getCoordSystem(numBody,
						 bodies,
						 coordSystemMap,
						 &mystranInstance[iIndex].feaProblem.numCoordSystem,
						 &mystranInstance[iIndex].feaProblem.feaCoordSystem);
	if (status != CAPS_SUCCESS) goto cleanup;

	// Get attribute to index mapping
	status = aim_getData(aimInfo, "Attribute_Map", &vtype, &rank, &nrow, &ncol, &dataTransfer, &units);
	if (status == CAPS_SUCCESS) {

		printf("Found link for attrMap (Attribute_Map) from parent\n");

		status = copy_mapAttrToIndexStruct( (mapAttrToIndexStruct *) dataTransfer, &mystranInstance[iIndex].attrMap);
		if (status != CAPS_SUCCESS) goto cleanup;

	} else {

		if (status == CAPS_DIRTY) printf("Parent AIMS are dirty\n");

		printf("Didn't find a link to attrMap from parent - getting it ourselves\n");

		// Get capsGroup name and index mapping to make sure all faces have a capsGroup value
		status = create_CAPSGroupAttrToIndexMap(numBody,
				                                bodies,
	                                            3, //>2 - search the body, faces, edges, and all the nodes
	                                            &mystranInstance[iIndex].attrMap);
		if (status != CAPS_SUCCESS) goto cleanup;
	}

	// See if a FEA mesh is available from parent
	status = aim_getData(aimInfo, "FEA_MESH", &vtype, &rank, &nrow, &ncol, &dataTransfer, &units);
	if (status == CAPS_SUCCESS) {

		printf("Found link for a FEA mesh (FEA_Mesh) from parent\n");

		numFEAMesh = 1;
		if      (nrow == 1 && ncol != 1) numFEAMesh = ncol;
		else if (nrow != 1 && ncol == 1) numFEAMesh = nrow;
		else {

			printf("Can not except 2D matrix of fea meshes\n");
			status = CAPS_BADVALUE;
			goto cleanup;
		}

		if (numFEAMesh != numBody) {
			printf("Number of inherited fea meshes does not match the number of bodies\n");
			return CAPS_SOURCEERR;
		}

		status = lump_genFEAMesh(numFEAMesh,
								 (genFEAMeshStruct *) dataTransfer,
								 &mystranInstance[iIndex].feaProblem.feaMesh);
		if (status != CAPS_SUCCESS) goto cleanup;

		feaMeshInherited = (int) true;

	} else if (status == CAPS_DIRTY) {

		printf("Link for a FEA mesh (FEA_Mesh) from parent found, but Parent AIMS are dirty\n");
		goto cleanup;

	} else { // Check to see if a general unstructured volume mesh is available

		printf("Didn't find a FEA mesh (FEA_Mesh) from parent - checking for general unstructured volume mesh\n");

		status = aim_getData(aimInfo, "Volume_MESH", &vtype, &rank, &nrow, &ncol, &dataTransfer, &units);
		if (status == CAPS_DIRTY) {

			printf("Link for a general unstructured volume mesh (Volume_Mesh) from parent found, but Parent AIMS are dirty\n");
			goto cleanup;

		} else if (status == CAPS_SUCCESS) {

			printf("Found link for a general unstructured volume mesh (Volume_Mesh) from parent\n");

			numFEAMesh = 1;
			if (nrow != 1 && ncol != 1) {

				printf("Can not except multiple volume meshes\n");
				status = CAPS_BADVALUE;
				goto cleanup;
			}

			if (numFEAMesh != numBody) {
				printf("Number of inherited volume mesh does not match the number of bodies - assuming volume mesh is already lumped\n");
			}

			status = initiate_genUnstrMeshStruct(&volumeMesh);
			if (status != CAPS_SUCCESS) goto cleanup;

			volumeMesh = *(genUnstrMeshStruct *) dataTransfer;

			// Transfer just the volume mesh
			mystranInstance[iIndex].feaProblem.feaMesh.numNodes = volumeMesh.numNodes;
			mystranInstance[iIndex].feaProblem.feaMesh.numTetra = volumeMesh.numTetra;
			mystranInstance[iIndex].feaProblem.feaMesh.numHexa  = volumeMesh.numHex;
			mystranInstance[iIndex].feaProblem.feaMesh.numElement3DPID = volumeMesh.numVolID;

			printf("Number of nodes - %d\n", mystranInstance[iIndex].feaProblem.feaMesh.numNodes);
			printf("Number of tetrahedral - %d\n", mystranInstance[iIndex].feaProblem.feaMesh.numTetra);
			printf("Number of hexahedral - %d\n", mystranInstance[iIndex].feaProblem.feaMesh.numHexa);
			printf("Number of 3D PIDs - %d\n", mystranInstance[iIndex].feaProblem.feaMesh.numElement3DPID);

			// Populate feaMesh from volume mesh - getting only solid elements
			if (volumeMesh.numTetra != 0 || volumeMesh.numHex != 0) {

				// Get grid points
				if (volumeMesh.numNodes != 0) {
					mystranInstance[iIndex].feaProblem.feaMesh.xyz = (double *) EG_alloc(3*volumeMesh.numNodes*sizeof(double));
					if (mystranInstance[iIndex].feaProblem.feaMesh.xyz == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					mystranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex = (int *) EG_alloc(volumeMesh.numNodes*sizeof(int));
					if (mystranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					mystranInstance[iIndex].feaProblem.feaMesh.xyzCoordID = (int *) EG_alloc(volumeMesh.numNodes*sizeof(int));
					if (mystranInstance[iIndex].feaProblem.feaMesh.xyzCoordID == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					memcpy(mystranInstance[iIndex].feaProblem.feaMesh.xyz,
						   volumeMesh.xyz,
						   3*volumeMesh.numNodes*sizeof(double));

					// Set default grid points to CAPS_MAGIC number
					for (i = 0; i < volumeMesh.numNodes; i++) {
						mystranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[i] = CAPSMAGIC;
						mystranInstance[iIndex].feaProblem.feaMesh.xyzCoordID[i] = 0;
					}

					// Search surface meshes to get grid attrIndex on boundaries
					for (i = 0; i < volumeMesh.surfaceMesh.numTriFace; i++) { // TriFace first
						mystranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[volumeMesh.surfaceMesh.triFaceList[3*i+0]-1] = volumeMesh.surfaceMesh.triFaceMarkList[i];
						mystranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[volumeMesh.surfaceMesh.triFaceList[3*i+1]-1] = volumeMesh.surfaceMesh.triFaceMarkList[i];
						mystranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[volumeMesh.surfaceMesh.triFaceList[3*i+2]-1] = volumeMesh.surfaceMesh.triFaceMarkList[i];
					}

					for (i = 0; i < volumeMesh.surfaceMesh.numQuadFace; i++) { // QuadFace first
						mystranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[volumeMesh.surfaceMesh.quadFaceList[4*i+0]-1] = volumeMesh.surfaceMesh.quadFaceMarkList[i];
						mystranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[volumeMesh.surfaceMesh.quadFaceList[4*i+1]-1] = volumeMesh.surfaceMesh.quadFaceMarkList[i];
						mystranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[volumeMesh.surfaceMesh.quadFaceList[4*i+2]-1] = volumeMesh.surfaceMesh.quadFaceMarkList[i];
						mystranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[volumeMesh.surfaceMesh.quadFaceList[4*i+3]-1] = volumeMesh.surfaceMesh.quadFaceMarkList[i];
					}
				}

				// Get global list volume IDs
				if (volumeMesh.numVolID != 0) {

					mystranInstance[iIndex].feaProblem.feaMesh.element3DPID = (int *) EG_alloc(volumeMesh.numVolID*sizeof(int));
					if (mystranInstance[iIndex].feaProblem.feaMesh.element3DPID == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					memcpy(mystranInstance[iIndex].feaProblem.feaMesh.element3DPID,
						   volumeMesh.volIDList,
						   volumeMesh.numVolID*sizeof(int));
				}

				// Get tetra elements
				if (volumeMesh.numTetra != 0) {

					mystranInstance[iIndex].feaProblem.feaMesh.ctetra    = (int *) EG_alloc(4*volumeMesh.numTetra*sizeof(int));
					if (mystranInstance[iIndex].feaProblem.feaMesh.ctetra == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					mystranInstance[iIndex].feaProblem.feaMesh.ctetraPID = (int *) EG_alloc(  volumeMesh.numTetra*sizeof(int));
					if (mystranInstance[iIndex].feaProblem.feaMesh.ctetraPID == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					mystranInstance[iIndex].feaProblem.feaMesh.ctetraElementID = (int *) EG_alloc(  volumeMesh.numTetra*sizeof(int));
					if (mystranInstance[iIndex].feaProblem.feaMesh.ctetraElementID == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					memcpy(mystranInstance[iIndex].feaProblem.feaMesh.ctetra,
						   volumeMesh.tetraList,
						   4*volumeMesh.numTetra*sizeof(int));

					memcpy(mystranInstance[iIndex].feaProblem.feaMesh.ctetraPID,
						   volumeMesh.tetraVolIDList,
						   volumeMesh.numTetra*sizeof(int));

					for (i = 0; i < mystranInstance[iIndex].feaProblem.feaMesh.numTetra; i++) {
						mystranInstance[iIndex].feaProblem.feaMesh.ctetraElementID[i] = i+1;
					}
				}

				// Get hexahedral elements
				if (volumeMesh.numHex != 0) {

					mystranInstance[iIndex].feaProblem.feaMesh.chexa    = (int *) EG_alloc(8*volumeMesh.numHex*sizeof(int));
					if (mystranInstance[iIndex].feaProblem.feaMesh.chexa == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					mystranInstance[iIndex].feaProblem.feaMesh.chexaPID = (int *) EG_alloc(  volumeMesh.numHex*sizeof(int));
					if (mystranInstance[iIndex].feaProblem.feaMesh.chexaPID == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					mystranInstance[iIndex].feaProblem.feaMesh.chexaElementID = (int *) EG_alloc(  volumeMesh.numHex*sizeof(int));
					if (mystranInstance[iIndex].feaProblem.feaMesh.chexaElementID == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					memcpy(mystranInstance[iIndex].feaProblem.feaMesh.chexa,
						   volumeMesh.hexList,
						   8*volumeMesh.numHex*sizeof(int));

					memcpy(mystranInstance[iIndex].feaProblem.feaMesh.chexaPID,
						   volumeMesh.hexVolIDList,
						   volumeMesh.numHex*sizeof(int));

					for (i = 0; i < mystranInstance[iIndex].feaProblem.feaMesh.numHexa; i++) {
						mystranInstance[iIndex].feaProblem.feaMesh.chexaElementID[i] = i+1 + mystranInstance[iIndex].feaProblem.feaMesh.numTetra;
					}
				}
			}

			feaMeshInherited = (int) true;
		} else {
			printf("Didn't find a general unstructured volume mesh (Volume_Mesh) from parent\n");

			feaMeshInherited = (int) false;
		}
	}


	// If we didn't inherit a FEA mesh we need to get one ourselves
	if (feaMeshInherited == (int) false) {

		tessParam[0] = mystranInstance[iIndex].tessParam->vals.reals[0]; // Gets multiplied by bounding box size
		tessParam[1] = mystranInstance[iIndex].tessParam->vals.reals[1]; // Gets multiplied by bounding box size
		tessParam[2] = mystranInstance[iIndex].tessParam->vals.reals[2];

		edgePointMin = mystranInstance[iIndex].edgePointMin->vals.integer;
		edgePointMax = mystranInstance[iIndex].edgePointMax->vals.integer;
		noQuadFace   = mystranInstance[iIndex].noQuadFace->vals.integer;

		if (edgePointMin < 2) {
			printf("The minimum number of allowable edge points is 2 not %d\n", edgePointMin);
			edgePointMin = 2;
		}

		if (edgePointMax < edgePointMin) {
			printf("The maximum number of edge points must be greater than the current minimum (%d)\n", edgePointMin);
			edgePointMax = edgePointMin+1;
		}

		// Allocate fea meshes
		//feaMesh = (genFEAMeshStruct *) EG_alloc(numBody*sizeof(genFEAMeshStruct));
		//if (feaMesh == NULL) return EGADS_MALLOC;

		// Get FEA Problem from EGADs body
		status = fea_bodyToBEM(bodies[0], // (in)  EGADS Body
	                           tessParam,          // (in)  Tessellation parameters
							   edgePointMin,        // (in)  minimum points along any Edge
							   edgePointMax,        // (in)  maximum points along any Edge
							   noQuadFace,
							   &mystranInstance[iIndex].attrMap,
							   &coordSystemMap,
							   &mystranInstance[iIndex].feaProblem.feaMesh);

		if (status != CAPS_SUCCESS) goto cleanup;

		printf("Number nodes = %d\n",mystranInstance[iIndex].feaProblem.feaMesh.numNodes);
		printf("Number Rods = %d\n" ,mystranInstance[iIndex].feaProblem.feaMesh.numRod);
		printf("Number Tria3 = %d\n",mystranInstance[iIndex].feaProblem.feaMesh.numTria3);
		printf("Number Quad4 = %d\n",mystranInstance[iIndex].feaProblem.feaMesh.numQuad4);
		printf("Number of 2D element PIDs = %d\n", mystranInstance[iIndex].feaProblem.feaMesh.numElement2DPID);

		status = aim_setTess(aimInfo, mystranInstance[iIndex].feaProblem.feaMesh.bodyTessMapping.egadsTess);
		if (status != CAPS_SUCCESS) goto cleanup;
	}

	status = CAPS_SUCCESS;

	goto cleanup;

	cleanup:

		(void ) destroy_mapAttrToIndexStruct(&coordSystemMap);

		return status;
}

/* ********************** Exposed AIM Functions ***************************** */

int aimInitialize(int ngIn, capsValue *gIn,
                  int *nFid, int **fids, int *nIn, int *nOut,
                  int *nFields, char ***fnames, int **ranks)
{
    int status; // Return flag
    int *ints;

    aimStorage *tmp;

    #ifdef DEBUG
        printf("\n mystranAIM/aimInitialize   ngIn = %d!\n", ngIn);
    #endif

    *nFid    = 6;
	ints     = (int *) EG_alloc(((int) *nFid)*sizeof(int));
	if (ints == NULL) return EGADS_MALLOC;

    /*! \page geomFidelityMystran Geometry Fidelity
     * The attribute capsFidelity may be set to either ALL or STRUCTURE for the Mystran AIM. The geometric fidelity for the
     * AIM requires that bodies be:
     *  - WIREBODY for purely 1D simulations
     *  - FACEBODY or SHEETBODY (non-manifold) for 2D simulations
     *  - SOLIDBODY or SHEETBODY (manifold) for 3D simulations
     */
	ints[0]  = ALL | ALL;
	ints[1]  = ALL | STRUCTURE;
	ints[2]  = SOLIDBODY | STRUCTURE;
	ints[3]  = SHEETBODY | STRUCTURE;
	ints[4]  = FACEBODY | STRUCTURE;
	ints[5]  = WIREBODY | STRUCTURE;
	*fids    = ints;

    /* specify the number of analysis input and out "parameters" */
    *nIn     = NUMINPUT;
    *nOut    = NUMOUTPUT;

    /* specify the field variables this analysis can generate */
    *nFields = 0;
    *ranks   = NULL;
    *fnames  = NULL;

    // Allocate mystranInstance
    if (mystranInstance == NULL) {

    	mystranInstance = (aimStorage *) EG_alloc(sizeof(aimStorage));

    	if (mystranInstance == NULL) {
    		EG_free(*fids);
    		*fids   = NULL;
    		return EGADS_MALLOC;
    	}

    } else {
    	tmp = (aimStorage *) EG_reall(mystranInstance, (numInstance+1)*sizeof(aimStorage));
    	if (tmp == NULL) {
    		EG_free(*fids);
    		*fids   = NULL;
    		return EGADS_MALLOC;
    	}

    	mystranInstance = tmp;
    }

    // Set initial values for mystranInstance
    mystranInstance[numInstance].projectName = NULL;

    // Analysis file path/directory
    mystranInstance[numInstance].analysisPath = NULL;

    /*
    // Check to make sure data transfer is ok
    mystranInstance[numInstance].dataTransferCheck = (int) true;
    */

    // Flag to see if we need to get an FEA mesh
    mystranInstance[numInstance].haveFEAMesh = (int) false;


    // Container for attribute to index map
    status = initiate_mapAttrToIndexStruct(&mystranInstance[numInstance].attrMap);
    if (status != CAPS_SUCCESS) return status;

    status = initiate_feaProblemStruct(&mystranInstance[numInstance].feaProblem);
    if (status != CAPS_SUCCESS) return status;

    // Pointer to caps input value for tessellation parameters
    mystranInstance[numInstance].tessParam = NULL;

    // Pointer to caps input value for Edge_Point- Min
    mystranInstance[numInstance].edgePointMin = NULL;

    // Pointer to caps input value for Edge_Point- Max
    mystranInstance[numInstance].edgePointMax = NULL;

    // Pointer to caps input value for No_Quad_Faces
    mystranInstance[numInstance].noQuadFace = NULL;



    // Increment number of instances
    numInstance += 1;

    return (numInstance -1);
}

int aimInputs(int iIndex, void *aimInfo, int index, char **ainame,
              capsValue *defval)
{
	/*! \page aimInputsMystran AIM Inputs
	 * The following list outlines the Mystran inputs along with their default value available
	 * through the AIM interface. Unless noted these values will be not be linked to
	 * any parent AIMs with variables of the same name.
	 */

	int status;

	#ifdef DEBUG
		printf(" mystranAIM/aimInputs instance = %d  index = %d!\n", iIndex, index);
	#endif

	*ainame = NULL;
  
	// Mystran Inputs
	if (index == 1) {
	    *ainame              = EG_strdup("Proj_Name");
		defval->type         = String;
		defval->nullVal      = NotNull;
		defval->vals.string  = EG_strdup("mystran_CAPS");
		defval->lfixed       = Change;

		/*! \page aimInputsMystran
		 * - <B> Proj_Name = "mystran_CAPS"</B> <br>
		 * This corresponds to the project name used for file naming.
		 */

	} else if (index == 2) {
        *ainame               = EG_strdup("Tess_Params");
        defval->type          = Double;
        defval->dim           = 1;
        defval->length        = 3;
        defval->nrow          = 3;
        defval->ncol          = 1;
        defval->units         = NULL;
        defval->lfixed        = Fixed;
        defval->vals.reals    = (double *) EG_alloc(defval->length*sizeof(double));
        if (defval->vals.reals != NULL) {
        	defval->vals.reals[0] = 0.025;
        	defval->vals.reals[1] = 0.001;
        	defval->vals.reals[2] = 15.00;
        } else return EGADS_MALLOC;

        mystranInstance[iIndex].tessParam = defval;

        /*! \page aimInputsMystran
		 * - <B> Tess_Params = [0.025, 0.001, 15.0]</B> <br>
		 * Body tessellation parameters used when creating a boundary element model.
		 * Tess_Params[0] and Tess_Params[1] get scaled by the bounding
		 * box of the body. (From the EGADS manual) A set of 3 parameters that drive the EDGE discretization
		 * and the FACE triangulation. The first is the maximum length of an EDGE segment or triangle side
		 * (in physical space). A zero is flag that allows for any length. The second is a curvature-based
		 * value that looks locally at the deviation between the centroid of the discrete object and the
		 * underlying geometry. Any deviation larger than the input value will cause the tessellation to
		 * be enhanced in those regions. The third is the maximum interior dihedral angle (in degrees)
		 * between triangle facets (or Edge segment tangents for a WIREBODY tessellation), note that a
		 * zero ignores this phase
		 */

	} else if (index == 3) {
        *ainame               = EG_strdup("Edge_Point_Min");
        defval->type          = Integer;
        defval->vals.integer  = 4;
        defval->lfixed        = Fixed;
        defval->nrow          = 1;
        defval->ncol          = 1;

        mystranInstance[iIndex].edgePointMin = defval;

        /*! \page aimInputsMystran
         * - <B> Edge_Point_Min = 4</B> <br>
         * Minimum number of points along an edge to use when creating a boundary element model.
         */

	} else if (index == 4) {
		*ainame               = EG_strdup("Edge_Point_Max");
		defval->type          = Integer;
		defval->vals.integer  = 10;
		defval->length        = 1;
		defval->lfixed        = Fixed;
		defval->nrow          = 1;
		defval->ncol          = 1;

		mystranInstance[iIndex].edgePointMax = defval;

		/*! \page aimInputsMystran
		 * - <B> Edge_Point_Max = 10</B> <br>
		 * Maximum number of points along an edge to use when creating a boundary element model.
		 */

	} else if (index == 5) {
		*ainame               = EG_strdup("No_Quad_Faces");
		defval->type          = Boolean;
		defval->vals.integer  = (int) false;

		mystranInstance[iIndex].noQuadFace = defval;

		/*! \page aimInputsMystran
		 * - <B> Edge_Point_Max = 10</B> <br>
		 * Maximum number of points along an edge to use when creating a boundary element model.
		 */

	} else if (index == 6) {
        *ainame              = EG_strdup("Property");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;
        defval->dim          = 1;

        /*! \page aimInputsMystran
		 * - <B> Property = NULL</B> <br>
		 * Property tuple used to input property information for the model, see \ref feaProperty for additional details.
		 */
	} else if (index == 7) {
	    *ainame              = EG_strdup("Material");
	    defval->type         = Tuple;
	    defval->nullVal      = IsNull;
	    //defval->units        = NULL;
	    defval->lfixed       = Change;
	    defval->vals.tuple   = NULL;
	    defval->dim          = 1;

	    /*! \page aimInputsMystran
		 * - <B> Material = NULL</B> <br>
		 * Material tuple used to input material information for the model, see \ref feaMaterial for additional details.
		 */
	} else if (index == 8) {
        *ainame              = EG_strdup("Constraint");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;
        defval->dim          = 1;

        /*! \page aimInputsMystran
		 * - <B> Constraint = NULL</B> <br>
		 * Constraint tuple used to input constraint information for the model, see \ref feaConstraint for additional details.
		 */
	} else if (index == 9) {
        *ainame              = EG_strdup("Load");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;
        defval->dim          = 1;

        /*! \page aimInputsMystran
		 * - <B> Load = NULL</B> <br>
		 * Load tuple used to input load information for the model, see \ref feaLoad for additional details.
		 */
    } else if (index == 10) {
        *ainame              = EG_strdup("Analysis");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;
        defval->dim          = 1;

        /*! \page aimInputsMystran
		 * - <B> Analysis = NULL</B> <br>
		 * Analysis tuple used to input analysis/case information for the model, see \ref feaAnalysis for additional details.
		 */
    } else if (index == 11) {
        *ainame              = EG_strdup("Analysis_Type");
        defval->type         = String;
        defval->nullVal      = NotNull;
        defval->vals.string  = EG_strdup("Modal");
        defval->lfixed       = Change;

        /*! \page aimInputsMystran
		 * - <B> Analysis_Type = "Modal"</B> <br>
		 * Type of analysis to generate files for, options include "Modal", "Static", and "Craig-Bampton".
		 */
    }

	// Link variable(s) to parent(s) if available
	/*if ((index != 1) && (*ainame != NULL) && index < 6 ) {
		status = aim_link(aimInfo, *ainame, ANALYSISIN, defval);
  	 	printf("Status = %d: Var Index = %d, Type = %d, link = %lX\n",
                        status, index, defval->type, defval->link);
	}*/

	return CAPS_SUCCESS;
}

int aimData(int iIndex, const char *name, enum capsvType *vtype, int *rank, int *nrow,
        int *ncol, void **data, char **units)
{

    /*! \page sharableDataMystran AIM Shareable Data
     * Currently the Mystran AIM does not have any shareable data types or values. It will try, however, to inherit a "FEA_MESH"
     * or "Volume_Mesh" from any parent AIMs. Note that the inheritance of the mesh is not required.
     */

    #ifdef DEBUG
        printf(" mystranAIM/aimData instance = %d  name = %s!\n", iIndex, name);
    #endif
   return CAPS_NOTFOUND;
}


int aimPreAnalysis(int iIndex, void *aimInfo, char *analysisPath,
                   capsValue *aimInputs, capsErrs **errs)
{

    int i, j, k; // Indexing

    int status; // Status return

    int found; // Boolean operator

    int *tempIntegerArray = NULL; // Temporary array to store a list of integers

    // CAPS function returns
	/*
	int 		  atype, alen;
	const int    *ints;
	const char   *string;
	const double *reals;
	*/

	// Analyis information
	char *analysisType = NULL;

	// File IO
	char *filename = NULL; // Output file name
    FILE *fp = NULL; // Output file pointer

    // Discrete data transfer variables
    capsDiscr *dataTransferDiscreteObj;
    char **transferName = NULL;
    int numTransferName, transferNameIndex;
    enum capsdMethod dataTransferMethod;
    int numDataTransferPoint;
    int numDataTransferElement = 0;
    int dataTransferRank;
    double *dataTransferData;

    // Variables used in global node mapping
    int *tessNodeMap = NULL, *storage = NULL;
    int globalNodeID;
    int nodeIndex[4], transferIndex[4], elementID, elementIndex;

    // NULL out errors
    *errs = NULL;
  
    // Store away the analysis path/directory
    mystranInstance[iIndex].analysisPath = analysisPath;

    // Get project name
    mystranInstance[iIndex].projectName = aimInputs[aim_getIndex(aimInfo, "Proj_Name", ANALYSISIN)-1].vals.string;

    // Get FEA mesh if we don't already have one
    if (mystranInstance[iIndex].haveFEAMesh == false) {
    	status = createMesh(iIndex, aimInfo);
    	if (status != CAPS_SUCCESS) goto cleanup;

    	mystranInstance[iIndex].haveFEAMesh = (int) true;
    }

    // Note: Setting order is important here.
    // 1. Materials should be set before properties.
    // 2. Coordinate system should be set before mesh and loads
	// 3. Mesh should be set before loads and constraints
	// 4. Constraints and loads should be set before analysis

    // Set material properties
    if (aimInputs[aim_getIndex(aimInfo, "Material", ANALYSISIN)-1].nullVal == NotNull) {
        status = fea_getMaterial(aimInputs[aim_getIndex(aimInfo, "Material", ANALYSISIN)-1].length,
                                 aimInputs[aim_getIndex(aimInfo, "Material", ANALYSISIN)-1].vals.tuple,
                                 &mystranInstance[iIndex].feaProblem.numMaterial,
                                 &mystranInstance[iIndex].feaProblem.feaMaterial);
        if (status != CAPS_SUCCESS) goto cleanup;
    } else printf("\nLoad tuple is NULL - No materials set\n");

    // Set property properties
    if (aimInputs[aim_getIndex(aimInfo, "Property", ANALYSISIN)-1].nullVal == NotNull) {
        status = fea_getProperty(aimInputs[aim_getIndex(aimInfo, "Property", ANALYSISIN)-1].length,
                                 aimInputs[aim_getIndex(aimInfo, "Property", ANALYSISIN)-1].vals.tuple,
                                 &mystranInstance[iIndex].attrMap,
                                 &mystranInstance[iIndex].feaProblem);
        if (status != CAPS_SUCCESS) goto cleanup;
    } else printf("\nProperty tuple is NULL - No properties set\n");

    // Set constraint properties
    if (aimInputs[aim_getIndex(aimInfo, "Constraint", ANALYSISIN)-1].nullVal == NotNull) {
        status = fea_getConstraint(aimInputs[aim_getIndex(aimInfo, "Constraint", ANALYSISIN)-1].length,
                                   aimInputs[aim_getIndex(aimInfo, "Constraint", ANALYSISIN)-1].vals.tuple,
                                   &mystranInstance[iIndex].attrMap,
                                   &mystranInstance[iIndex].feaProblem);
        if (status != CAPS_SUCCESS) goto cleanup;
    } else printf("\nConstraint tuple is NULL - No constraints applied\n");

    // Set load properties
    if (aimInputs[aim_getIndex(aimInfo, "Load", ANALYSISIN)-1].nullVal == NotNull) {
        status = fea_getLoad(aimInputs[aim_getIndex(aimInfo, "Load", ANALYSISIN)-1].length,
                             aimInputs[aim_getIndex(aimInfo, "Load", ANALYSISIN)-1].vals.tuple,
                             &mystranInstance[iIndex].attrMap,
                             &mystranInstance[iIndex].feaProblem);
        if (status != CAPS_SUCCESS) goto cleanup;

        // Loop through loads to see if any of them are supposed to be from an external source
        for (i = 0; i < mystranInstance[iIndex].feaProblem.numLoad; i++) {

        	if (mystranInstance[iIndex].feaProblem.feaLoad[i].loadType == PressureExternal) {

            	mystranInstance[iIndex].feaProblem.feaLoad[i].numElementID = 0;

                //See if we have data transfer information
                status = aim_getBounds(aimInfo, &numTransferName, &transferName);
                if (status != CAPS_SUCCESS) goto cleanup;

				numDataTransferElement = 0;
				elementIndex = 0;
				for (transferNameIndex = 0; transferNameIndex < numTransferName; transferNameIndex++) {

					status = aim_getDiscr(aimInfo, transferName[transferNameIndex], &dataTransferDiscreteObj);
					if (status != CAPS_SUCCESS) continue;

					status = aim_getDataSet(dataTransferDiscreteObj,
											"Pressure",
											&dataTransferMethod,
											&numDataTransferPoint,
											&dataTransferRank,
											&dataTransferData);

					if (status == CAPS_SUCCESS) { // If we do have data ready, how many elements there are?

						numDataTransferElement += dataTransferDiscreteObj->nElems;

					} else {

						continue; // If no elements in this object skip to next transfer name
					}

					if (dataTransferRank != 1) {
						printf("Pressure transfer data found however rank is %d not 1!!!!\n", dataTransferRank);

						status = CAPS_BADRANK;
						goto cleanup;
					}

					// If the first time we found elements allocate arrays
					if (mystranInstance[iIndex].feaProblem.feaLoad[i].numElementID == 0) {

						mystranInstance[iIndex].feaProblem.feaLoad[i].elementIDSet = (int *) EG_alloc(numDataTransferElement *sizeof(int));
						mystranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce = (double *) EG_alloc(4*numDataTransferElement *sizeof(double));

					} else { // Else reallocate arrays

						mystranInstance[iIndex].feaProblem.feaLoad[i].elementIDSet = (int *) EG_reall(mystranInstance[iIndex].feaProblem.feaLoad[i].elementIDSet,
																									  numDataTransferElement *sizeof(int));
						mystranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce = (double *) EG_reall(mystranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce,
																														4*numDataTransferElement *sizeof(double));
					}

					// Check to see if allocation failed
					if (mystranInstance[iIndex].feaProblem.feaLoad[i].elementIDSet == NULL ||
						mystranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce == NULL) {

						status = EGADS_MALLOC;
						goto cleanup;
					}

					// Increment the number of element ID
					mystranInstance[iIndex].feaProblem.feaLoad[i].numElementID = numDataTransferElement;

					//Get extra node information stored in the discrObj
					storage = (int *) dataTransferDiscreteObj->ptrm;
					tessNodeMap = &storage[0]; // Local indexing on the body

					//Now lets loop through our ctria3 mesh and get the node indexes for each element
					for (j = 0; j < mystranInstance[iIndex].feaProblem.feaMesh.numTria3; j++) {

						elementID = mystranInstance[iIndex].feaProblem.feaMesh.ctria3ElementID[j];

						// elementID is 1 bias
						nodeIndex[0] = mystranInstance[iIndex].feaProblem.feaMesh.ctria3[3*(elementID-1) +0];
						nodeIndex[1] = mystranInstance[iIndex].feaProblem.feaMesh.ctria3[3*(elementID-1) +1];
						nodeIndex[2] = mystranInstance[iIndex].feaProblem.feaMesh.ctria3[3*(elementID-1) +2];

						transferIndex[0] = -1;
						transferIndex[1] = -1;
						transferIndex[2] = -1;
						transferIndex[3] = -1;

						// Loop through the nodeMap of the data set getting nodeIDs trying
						//   to match the nodes in the element
						for (k = 0; k < numDataTransferPoint; k++) {

							globalNodeID = tessNodeMap[k];

							if (nodeIndex[0] == globalNodeID) transferIndex[0] = k;
							if (nodeIndex[1] == globalNodeID) transferIndex[1] = k;
							if (nodeIndex[2] == globalNodeID) transferIndex[2] = k;

							// If the nodes completely match the nodes on the element - break
							if (transferIndex[0] >= 0 && transferIndex[1] >= 0 && transferIndex[2] >= 0) {
								break;
							}
						}

						// If all the nodeIndexes match the transferIndex the element is in the data set
						//  so transfer the pressure forces
						if (transferIndex[0] >= 0 && transferIndex[1] >= 0 && transferIndex[2] >= 0) {

							mystranInstance[iIndex].feaProblem.feaLoad[i].elementIDSet[elementIndex] = elementID;

							mystranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce[4*elementIndex+0] = dataTransferData[transferIndex[0]];
							mystranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce[4*elementIndex+1] = dataTransferData[transferIndex[1]];
							mystranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce[4*elementIndex+2] = dataTransferData[transferIndex[2]];
							mystranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce[4*elementIndex+3] = 0.0;

							elementIndex += 1;
						}
					}

					if (elementIndex != numDataTransferElement) {
						printf("\nElement transfer mismatch: number of elements found = %d, number"
								" of elements in transfer data set %d\n", elementIndex, numDataTransferElement);
						if (transferName != NULL) EG_free(transferName);
						//if (status != CAPS_SUCCESS) printf("string_freeArray status = %d\n", status);
						status = CAPS_MISMATCH;
						if (status != CAPS_SUCCESS) goto cleanup;
					}
				} // End data transfer name loop


				if (transferName != NULL) EG_free(transferName);
				transferName = NULL;

        	} // End PressureExternal if
        } // End load for loop
    } else printf("\nLoad tuple is NULL - No loads applied\n");

    // Set analysis settings
    if (aimInputs[aim_getIndex(aimInfo, "Analysis", ANALYSISIN)-1].nullVal == NotNull) {
        status = fea_getAnalysis(aimInputs[aim_getIndex(aimInfo, "Analysis", ANALYSISIN)-1].length,
                                 aimInputs[aim_getIndex(aimInfo, "Analysis", ANALYSISIN)-1].vals.tuple,
                                 &mystranInstance[iIndex].feaProblem);
        if (status != CAPS_SUCCESS) goto cleanup; // It ok to not have an analysis tuple
    } else printf("\nAnalysis tuple is NULL\n");

    // Write Nastran Mesh
    filename = (char *) EG_alloc((PATH_MAX +1)*sizeof(char));
    if (filename == NULL) {
    	status = EGADS_MALLOC;
    	goto cleanup;
    }
    strcpy(filename, analysisPath);
    strcat(filename, "/");
    strcat(filename, mystranInstance[iIndex].projectName);

    status = write_NASTRAN(filename,
    					   1,
						   &mystranInstance[iIndex].feaProblem.feaMesh,
						   mystranInstance[iIndex].feaProblem.feaFileFormat.gridFileType);
    if (status != CAPS_SUCCESS) goto cleanup;

    if (filename != NULL) EG_free(filename);

    // Write mystran input file
    filename = (char *) EG_alloc((PATH_MAX +1)*sizeof(char));
    if (filename == NULL) {
    	status = EGADS_MALLOC;
    	goto cleanup;
    }

    strcpy(filename, analysisPath);
    strcat(filename, "/");
    strcat(filename, mystranInstance[iIndex].projectName);
    strcat(filename, ".dat");


    printf("\nWriting MYSTRAN instruction file....\n");
    fp = fopen(filename, "w");
    if (fp == NULL) {
    	printf("Unable to open file: %s\n", filename);
    	status = CAPS_IOERR;
    	goto cleanup;
    }
    if (filename != NULL) EG_free(filename);
    filename = NULL;

    //////////////// Executive control ////////////////
    fprintf(fp, "ID CAPS generated Problem FOR MYSTRAN\n");

    // Analysis type
    analysisType = aimInputs[aim_getIndex(aimInfo, "Analysis_Type", ANALYSISIN)-1].vals.string;

    if     (strcasecmp(analysisType, "Modal")         == 0) fprintf(fp, "SOL 3\n");
    else if(strcasecmp(analysisType, "Static")        == 0) fprintf(fp, "SOL 1\n");
    else if(strcasecmp(analysisType, "Craig-Bampton") == 0) fprintf(fp, "SOL 31\n");
    else {
        printf("Unrecognized \"Analysis_Type\", %s, defaulting to \"Modal\" analysis\n", analysisType);
        analysisType = "Modal";
        fprintf(fp, "SOL 3\n");
    }

    if(strcasecmp(analysisType, "Modal") == 0) fprintf(fp, "OUTPUT4 EIGEN_VAL, EIGEN_VEC, GEN_MASS, , // -1/21 $\n"); // Binary output of eigenvalues and vectors

    fprintf(fp, "CEND\n\n");

    //////////////// Case control ////////////////

    // Write output request information
    fprintf(fp, "DISP = ALL\n"); // Output all displacements

    if(strcasecmp(analysisType, "Static") == 0) fprintf(fp, "STRE = ALL\n"); // Output all stress

    if(strcasecmp(analysisType, "Static") == 0) fprintf(fp, "STRA = ALL\n"); // Output all strain

    //fprintf(fp, "ELDATA(6,PRINT) = ALL\n");

    // Write constraint information
    if (mystranInstance[iIndex].feaProblem.numConstraint != 0) {
    	fprintf(fp, "SPC = %d\n", mystranInstance[iIndex].feaProblem.numConstraint+1);
    } else {
    	printf("Warning: No constraints specified for job!!!!\n");
    }

    // Modal analysis - only
    // If modal - we are only going to use the first analysis structure we come across that has its type as modal
    if (strcasecmp(analysisType, "Modal") == 0) {

    	// Look through analysis structures for a modal one
    	found = (int) false;
        for (i = 0; i < mystranInstance[iIndex].feaProblem.numAnalysis; i++) {
            if (mystranInstance[iIndex].feaProblem.feaAnalysis[i].analysisType == Modal) {
            	found = (int) true;
            	break;
            }
        }

        // Write out analysis ID if a modal analysis structure was found
        if (found == (int) true)  {
            fprintf(fp, "METHOD = %d\n", mystranInstance[iIndex].feaProblem.feaAnalysis[i].analysisID);
        } else {
            printf("Warning: No eigenvalue analysis information specified in \"Analysis\" tuple, though "
            		"AIM input \"Analysis_Type\" is set to \"Modal\"!!!!\n");
            status = CAPS_NOTFOUND;
            goto cleanup;
        }

    }
    // Static analysis - only
    if (strcasecmp(analysisType, "Static") == 0) {

    	// If we have multiple analysis structures
    	if (mystranInstance[iIndex].feaProblem.numAnalysis != 0) {

			// Write subcase information if multiple analysis tuples were provide
			for (i = 0; i < mystranInstance[iIndex].feaProblem.numAnalysis; i++) {

				if (mystranInstance[iIndex].feaProblem.feaAnalysis[i].analysisType == Static) {
					fprintf(fp, "SUBCASE %d\n", i);
					fprintf(fp, "\tLABEL %s\n", mystranInstance[iIndex].feaProblem.feaAnalysis[i].name);

					// Write loads for sub-case
					if (mystranInstance[iIndex].feaProblem.numLoad != 0) {
						fprintf(fp, "\tLOAD = %d\n", mystranInstance[iIndex].feaProblem.numLoad+i+1);
					}

					// Issue some warnings regarding loads if necessary
					if (mystranInstance[iIndex].feaProblem.feaAnalysis[i].numLoad == 0 && mystranInstance[iIndex].feaProblem.numLoad !=0) {
						printf("Warning: No loads specified for static case %s, assuming "
								"all loads are applied!!!!\n", mystranInstance[iIndex].feaProblem.feaAnalysis[i].name);
					} else if (mystranInstance[iIndex].feaProblem.numLoad == 0) {
						printf("Warning: No loads specified for static case %s!!!!\n", mystranInstance[iIndex].feaProblem.feaAnalysis[i].name);
					}
				}
			}

    	} else { // // If no sub-cases

    		if (mystranInstance[iIndex].feaProblem.numLoad != 0) {
    			fprintf(fp, "LOAD = %d\n", mystranInstance[iIndex].feaProblem.numLoad+1);
    		} else {
    			printf("Warning: No loads specified for static a job!!!!\n");
    		}
    	}
    }

    //////////////// Bulk data ////////////////
    fprintf(fp, "\nBEGIN BULK\n");

    // Turn off auto SPC
    //fprintf(fp, "%-8s %7s %7s\n", "PARAM", "AUTOSPC", "N");

    //fprintf(fp, "%-8s %7s %7d\n", "PARAM", "PRTOU4", 1);
    //fprintf(fp, "%-8s %7s %7d\n", "PARAM", "PRTDLR", 1);

    //fprintf(fp, "%-8s %7s %7d\n", "PARAM", "PRTDOF", 1);

    //fprintf(fp, "%-8s %7d %7d\n", "DEBUG", 183, 1);
    // Turn on FEMAP
    //fprintf(fp, "%-8s %7s %7d\n", "PARAM", "POST", -1);

    // Analysis Cards - Eigenvalue included, as well as combined load
    if (mystranInstance[iIndex].feaProblem.numAnalysis != 0) {

    	for (i = 0; i < mystranInstance[iIndex].feaProblem.numAnalysis; i++) {
			status = nastran_writeAnalysisCard(fp, &mystranInstance[iIndex].feaProblem.feaAnalysis[i], &mystranInstance[iIndex].feaProblem.feaFileFormat);
			if (status != CAPS_SUCCESS) goto cleanup;


			if (mystranInstance[iIndex].feaProblem.feaAnalysis[i].numLoad != 0) {
				status = nastran_writeLoadADDCard(fp,
												  mystranInstance[iIndex].feaProblem.numLoad+i+1,
												  mystranInstance[iIndex].feaProblem.feaAnalysis[i].numLoad,
												  mystranInstance[iIndex].feaProblem.feaAnalysis[i].loadSetID,
												  mystranInstance[iIndex].feaProblem.feaLoad,
												  &mystranInstance[iIndex].feaProblem.feaFileFormat);
				if (status != CAPS_SUCCESS) goto cleanup;

			} else { // If no loads for an individual analysis are specified assume that all loads should be applied

				if (mystranInstance[iIndex].feaProblem.numLoad != 0) {

					// Create a temporary list of load IDs
					tempIntegerArray = (int *) EG_alloc(mystranInstance[iIndex].feaProblem.numLoad*sizeof(int));
					if (tempIntegerArray == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					for (j = 0; j < mystranInstance[iIndex].feaProblem.numLoad; j++) {
						tempIntegerArray[j] = mystranInstance[iIndex].feaProblem.feaLoad[j].loadID;
					}

					// Write combined load card
					status = nastran_writeLoadADDCard(fp,
													  mystranInstance[iIndex].feaProblem.numLoad+i+1,
													  mystranInstance[iIndex].feaProblem.numLoad,
													  tempIntegerArray,
													  mystranInstance[iIndex].feaProblem.feaLoad,
													  &mystranInstance[iIndex].feaProblem.feaFileFormat);
					if (status != CAPS_SUCCESS) goto cleanup;

					// Free temporary load ID list
					if (tempIntegerArray != NULL) EG_free(tempIntegerArray);
					tempIntegerArray = NULL;
				}

			}
		}

    } else { // If there aren't any analysis structures just write a single combined load card

    	// Combined loads
    	if (mystranInstance[iIndex].feaProblem.numLoad != 0) {

			// Create a temporary list of load IDs
			tempIntegerArray = (int *) EG_alloc(mystranInstance[iIndex].feaProblem.numLoad*sizeof(int));
			if (tempIntegerArray == NULL) {
				status = EGADS_MALLOC;
				goto cleanup;
			}

			for (i = 0; i < mystranInstance[iIndex].feaProblem.numLoad; i++) {
				tempIntegerArray[i] = mystranInstance[iIndex].feaProblem.feaLoad[i].loadID;
			}

			// Write combined load card
			status = nastran_writeLoadADDCard(fp,
											  mystranInstance[iIndex].feaProblem.numLoad+1,
											  mystranInstance[iIndex].feaProblem.numLoad,
											  tempIntegerArray,
											  mystranInstance[iIndex].feaProblem.feaLoad,
											  &mystranInstance[iIndex].feaProblem.feaFileFormat);

			if (status != CAPS_SUCCESS) goto cleanup;

			// Free temporary load ID list
			if (tempIntegerArray != NULL) EG_free(tempIntegerArray);
			tempIntegerArray = NULL;
    	}
    }

    // Combined constraints
    if (mystranInstance[iIndex].feaProblem.numConstraint != 0) {

    	// Create a temporary list of constraint IDs
    	tempIntegerArray = (int *) EG_alloc(mystranInstance[iIndex].feaProblem.numConstraint*sizeof(int));
    	if (tempIntegerArray == NULL) {
    		status = EGADS_MALLOC;
    		goto cleanup;
    	}

    	for (i = 0; i < mystranInstance[iIndex].feaProblem.numConstraint; i++) {
    		tempIntegerArray[i] = mystranInstance[iIndex].feaProblem.feaConstraint[i].constraintID;
    	}

    	// Write combined constraint card
    	status = nastran_writeConstraintADDCard(fp,
    											mystranInstance[iIndex].feaProblem.numConstraint+1,
												mystranInstance[iIndex].feaProblem.numConstraint,
												tempIntegerArray,
												&mystranInstance[iIndex].feaProblem.feaFileFormat);
    	if (status != CAPS_SUCCESS) goto cleanup;

    	// Free temporary constraint ID list
    	if (tempIntegerArray != NULL) EG_free(tempIntegerArray);
    	tempIntegerArray = NULL;
    }

    // Loads
    for (i = 0; i < mystranInstance[iIndex].feaProblem.numLoad; i++) {
    	status = nastran_writeLoadCard(fp, &mystranInstance[iIndex].feaProblem.feaLoad[i], &mystranInstance[iIndex].feaProblem.feaFileFormat);
    	if (status != CAPS_SUCCESS) goto cleanup;
    }

    // Constraints
    for (i = 0; i < mystranInstance[iIndex].feaProblem.numConstraint; i++) {

    	status = nastran_writeConstraintCard(fp, &mystranInstance[iIndex].feaProblem.feaConstraint[i], &mystranInstance[iIndex].feaProblem.feaFileFormat);
    	if (status != CAPS_SUCCESS) goto cleanup;
    }

  	// Materials
  	for (i = 0; i < mystranInstance[iIndex].feaProblem.numMaterial; i++) {
  	    status = nastran_writeMaterialCard(fp, &mystranInstance[iIndex].feaProblem.feaMaterial[i], &mystranInstance[iIndex].feaProblem.feaFileFormat);
  	    if (status != CAPS_SUCCESS) goto cleanup;
  	}

  	// Properties
  	for (i = 0; i < mystranInstance[iIndex].feaProblem.numProperty; i++) {
  	    status = nastran_writePropertyCard(fp, &mystranInstance[iIndex].feaProblem.feaProperty[i], &mystranInstance[iIndex].feaProblem.feaFileFormat);
  	    if (status != CAPS_SUCCESS) goto cleanup;
  	}

  	// Coordinate systems
  	for (i = 0; i < mystranInstance[iIndex].feaProblem.numCoordSystem; i++) {
  		status = nastran_writeCoordinateSystemCard(fp, &mystranInstance[iIndex].feaProblem.feaCoordSystem[i], &mystranInstance[iIndex].feaProblem.feaFileFormat);
  		if (status != CAPS_SUCCESS) goto cleanup;
  	}

  	// Include mesh file
  	fprintf(fp,"INCLUDE \'%s.bdf\'\n", mystranInstance[iIndex].projectName);

  	// End bulk data
    fprintf(fp,"ENDDATA\n");

    status = CAPS_SUCCESS;

    goto cleanup;

    cleanup:
		if (fp != NULL) fclose(fp);

		if (tempIntegerArray != NULL) EG_free(tempIntegerArray);

		if (filename != NULL) EG_free(filename);

		if (transferName != NULL) EG_free(transferName);
		transferName = NULL;


		return status;
}

// Set Mystran output variables
int aimOutputs(int iIndex, int index, char **aoname, capsValue *form)
{
	/*! \page aimOutputsMystran AIM Outputs
	 * The following list outlines the Mystran outputs available through the AIM interface.
	 */

	#ifdef DEBUG
        printf(" mystranAIM/aimOutputs instance = %d  index = %d!\n", iIndex, index);
    #endif

    /*! \page aimOutputsMystran AIM Outputs
     * - <B>EigenValue</B> = List of Eigen-Values (\f$ \lambda\f$) after a modal solve.
     * - <B>EigenRadian</B> = List of Eigen-Values in terms of radians (\f$ \omega = \sqrt{\lambda}\f$ ) after a modal solve.
     * - <B>EigenFrequency</B> = List of Eigen-Values in terms of frequencies (\f$ f = \frac{\omega}{2\pi}\f$) after a modal solve.
     * - <B>EigenGeneralMass</B> = List of generalized masses for the Eigen-Values.
     * .
     */

    if (index == 1) {
    	*aoname = EG_strdup("EigenValue");
    	form->type          = Double;
    	form->units         = NULL;
    	form->lfixed        = Change;
    	form->sfixed        = Change;
    } else if (index == 2) {
    	*aoname = EG_strdup("EigenRadian");
    	form->type          = Double;
    	form->units         = NULL;
    	form->lfixed        = Change;
    	form->sfixed        = Change;
    } else if (index == 3) {
        	*aoname = EG_strdup("EigenFrequency");
        	form->type          = Double;
        	form->units         = NULL;
        	form->lfixed        = Change;
        	form->sfixed        = Change;
    } else if (index == 4) {
    	*aoname = EG_strdup("EigenGeneralMass");
    	form->type          = Double;
    	form->units         = NULL;
    	form->lfixed        = Change;
    	form->sfixed        = Change;
    }
    /*else if (index == 3) {
    	*aoname = EG_strdup("EigenVector");
    	form->type          = Double;
    	form->units         = NULL;
    	form->lfixed        = Change;
    	form->sfixed        = Change;
    }*/

    return CAPS_SUCCESS;
}

// Calculate Mystran output
int aimCalcOutput(int iIndex, void *aimInfo, char *analysisPath,
              	  int index, capsValue *val, capsErrs **errors)
{
	int status; // Function return status

	int i; // Indexing

	char currentPath[PATH_MAX]; // Current directory path

	char *filename = NULL; // File to open
	char extOP1[] = ".OP1";
	char extF06[] = ".F06";
	FILE *fp; // File pointer

	const double pi = 4.0 * atan(1.0);

	(void) getcwd(currentPath, PATH_MAX); // Get our current path

	// Set path to analysis directory
	if (chdir(analysisPath) != 0) {
	#ifdef DEBUG
		printf(" mystranAIM/aimCalcOutput Cannot chdir to %s!\n", analysisPath);
	#endif

		return CAPS_DIRERR;
	}

	if (index == 1 || index == 2 || index == 3) {
		// Open mystran *.OP1 file - OUTPUT4
		filename = (char *) EG_alloc((strlen(mystranInstance[iIndex].projectName) + strlen(extOP1) +1)*sizeof(char));
		if (filename == NULL) return EGADS_MALLOC;

		sprintf(filename, "%s%s", mystranInstance[iIndex].projectName, extOP1);

		fp = fopen(filename, "rb");

		EG_free(filename); // Free filename allocation

		if (fp == NULL) {
			#ifdef DEBUG
				printf(" mystranAIM/aimCalcOutput Cannot open Output file!\n");
			#endif

			chdir(currentPath);

			return CAPS_IOERR;
		}

		status = mystran_readOutput4Data(fp, "EIGEN_VA", val);
		if (status == CAPS_SUCCESS) {
			if (index == 2) {
				for (i = 0; i < val->length; i++) {
					val->vals.reals[i] = sqrt(val->vals.reals[i]);
				}

				//val->units = EG_strdup("rads/s");
			}

			if (index == 3) {
				for (i = 0; i < val->length; i++) {
					val->vals.reals[i] = sqrt(val->vals.reals[i])/(2*pi);
				}

				//val->units = EG_strdup("Hz");
			}
		}

	} else if (index == 4) {
		// Open mystran *.OP1 file - OUTPUT4
		filename = (char *) EG_alloc((strlen(mystranInstance[iIndex].projectName) + strlen(extOP1) +1)*sizeof(char));
		if (filename == NULL) return EGADS_MALLOC;

		sprintf(filename, "%s%s", mystranInstance[iIndex].projectName, extOP1);

		fp = fopen(filename, "rb");

		EG_free(filename); // Free filename allocation

		if (fp == NULL) {
			#ifdef DEBUG
				printf(" mystranAIM/aimCalcOutput Cannot open Output file!\n");
			#endif

			chdir(currentPath);

			return CAPS_IOERR;
		}

		status = mystran_readOutput4Data(fp, "GEN_MASS", val);


	} else if (index == 5) {

        // Open mystran *.F06 file
        filename = (char *) EG_alloc((strlen(mystranInstance[iIndex].projectName) + strlen(extF06) +1)*sizeof(char));
        if (filename == NULL) return EGADS_MALLOC;

        sprintf(filename, "%s%s", mystranInstance[iIndex].projectName, extF06);

        fp = fopen(filename, "r");

        EG_free(filename); // Free filename allocation
        if (fp == NULL) {
            #ifdef DEBUG
                printf(" mystranAIM/aimCalcOutput Cannot open Output file!\n");
            #endif
            chdir(currentPath);

            return CAPS_IOERR;
        }

        /*
        double **dataMatrix = NULL;
        int numEigenVector = 0;
        int numGridPoint = 0;
        status = mystran_readF06Eigen(fp, &numEigenVector, &numGridPoint, &dataMatrix);
         */

	}


	// Restore the path we came in with
	chdir(currentPath);
	fclose(fp);

	return status;
}

int aimFreeDiscr(capsDiscr *discr)
{
	int iIndex, i; // Indexing

	int status; // Returning status

	#ifdef DEBUG
    	if (discr == NULL) {
    		printf(" mystranAIM/aimFreeDiscr w/ NULL!\n");
    	} else {
    		printf(" mystranAIM/aimFreeDiscr instance = %d!\n", discr->instance);
    	}
	#endif

    if (discr == NULL) {

        if (numInstance != 0) {

        	// Clean up mystranInstance data
        	for ( iIndex = 0; iIndex < numInstance; iIndex++) {

        		printf(" Cleaning up mystranInstance - %d\n", iIndex);


        		// Attribute to index map
        		status = destroy_mapAttrToIndexStruct(&mystranInstance[iIndex].attrMap);
        		if (status != CAPS_SUCCESS) return status;

        		// Destroy FEA problem structure
        		status = destroy_feaProblemStruct(&mystranInstance[iIndex].feaProblem);
        		if (status != CAPS_SUCCESS) return status;

        		// NULL projetName
        		mystranInstance[iIndex].projectName = NULL;

        		// NULL analysisPath
        		mystranInstance[iIndex].analysisPath = NULL;

        		// NULL pointer to caps input value for tessellation parameters
        		mystranInstance[iIndex].tessParam = NULL;

        		// NULL pointer to caps input value for Edge_Point- Min
        		mystranInstance[iIndex].edgePointMin = NULL;

        		// NULL pointer to caps input value for Edge_Point- Max
        		mystranInstance[iIndex].edgePointMax = NULL;

        		// NULL pointer to caps input value for No_Quad_Faces
        		mystranInstance[iIndex].noQuadFace = NULL;
        	}

          }
          numInstance = 0;

    } else {

    	// Free up this capsDiscr

		discr->nPoints = 0; // Points

		if (discr->mapping != NULL) EG_free(discr->mapping);
		discr->mapping = NULL;

		if (discr->types != NULL) { // Element types
			for (i = 0; i < discr->nTypes; i++) {
				if (discr->types[i].gst  != NULL) EG_free(discr->types[i].gst);
				if (discr->types[i].tris != NULL) EG_free(discr->types[i].tris);
			}

			EG_free(discr->types);
		}

		discr->nTypes  = 0;
		discr->types   = NULL;

		if (discr->elems != NULL) { // Element connectivity

			for (i = 0; i < discr->nElems; i++) {
				if (discr->elems[i].gIndices != NULL) EG_free(discr->elems[i].gIndices);
			}

			EG_free(discr->elems);
		}

		discr->nElems  = 0;
		discr->elems   = NULL;

		if (discr->ptrm != NULL) EG_free(discr->ptrm); // Extra information to store into the discr void pointer
		discr->ptrm    = NULL;


		discr->nVerts = 0;    // Not currently used
		discr->verts  = NULL; // Not currently used
		discr->celem = NULL; // Not currently used

		discr->nDtris = 0; // Not currently used
		discr->dtris  = NULL; // Not currently used

    }

    return CAPS_SUCCESS;
}

int aimDiscr(char *tname, capsDiscr *discr) {

	int i, body, face; // Indexing

	int status; // Function return status

	int iIndex; // Instance index

	int numBody;

	// EGADS objects
	ego tess, *bodies = NULL, *faces = NULL, tempBody;

	const char   *string, *capsGroup; // capsGroups strings

	// EGADS function returns
	int plen, tlen, elen;
	const int    *ptype, *pindex, *tris, *nei;
	const double *xyz, *uv;

	// Body Tessellation
	int numFace = 0;
	int numFaceFound = 0;
	int numPoint = 0, numTri = 0, numGlobalPoint = 0;
	int *bodyFaceMap = NULL; // size=[2*numFaceFound], [2*numFaceFound + 0] = body, [2*numFaceFoun + 1] = face

	int *globalID = NULL, *localStitchedID = NULL, gID = 0;

	int *storage= NULL; // Extra information to store into the discr void pointer

	int numCAPSGroup = 0, attrIndex = 0, foundAttr = (int) false;
	int *capsGroupList = NULL;
	int dataTransferBodyIndex=-99;

	iIndex = discr->instance;
	#ifdef DEBUG
		printf(" mystranAIM/aimDiscr: tname = %s, instance = %d!\n", tname, iIndex);
	#endif

	if ((iIndex < 0) || (iIndex >= numInstance)) return CAPS_BADINDEX;


	if (tname == NULL) return CAPS_NOTFOUND;

	/*if (mystranInstance[iIndex].dataTransferCheck == (int) false) {
	    printf("The volume is not suitable for data transfer - possibly the volume mesher "
	            "added unaccounted for points\n");
	    return CAPS_BADVALUE;
	}*/

	// Currently this ONLY works if the capsTranfer lives on single body!
	status = aim_getBodies(discr->aInfo, &numBody, &bodies);
	if (status != CAPS_SUCCESS) {
		printf(" mystranAIM/aimDiscr: %d aim_getBodies = %d!\n", iIndex, status);
		return status;
	}

	status = aimFreeDiscr(discr);
	if (status != CAPS_SUCCESS) return status;

	numFaceFound = 0;
	numPoint = numTri = 0;
	// Find any faces with our boundary marker and get how many points and triangles there are
	for (body = 0; body < numBody; body++) {

		status = EG_getBodyTopos(bodies[body], NULL, FACE, &numFace, &faces);
		if (status != EGADS_SUCCESS) {
			printf(" mystranAIM: getBodyTopos (Face) = %d for Body %d!\n", status, body);
			return status;
		}

		for (face = 0; face < numFace; face++) {

			// Retrieve the string following a capsTransfer tag
			status = retrieve_CAPSTransferAttr(faces[face], &string);
			if (status != CAPS_SUCCESS) continue;

			if (strcmp(string, tname) != 0) continue;

			#ifdef DEBUG
				printf(" mystranAIM/aimDiscr: Body %d/Face %d matches %s!\n", body, face+1, tname);
			#endif


			status = retrieve_CAPSGroupAttr(faces[face], &capsGroup);
			if (status != CAPS_SUCCESS) {
			    printf("capsTransfer found on face %d, but no capGroup found!!!\n", face);
			    continue;
			} else {

			    status = get_mapAttrToIndexIndex(&mystranInstance[iIndex].attrMap, capsGroup, &attrIndex);
			    if (status != CAPS_SUCCESS) {
			        printf("capsGroup %s NOT found in attrMap\n",capsGroup);
			        continue;
			    } else {

			        // If first index create arrays and store index
			        if (numCAPSGroup == 0) {
			            numCAPSGroup += 1;
			            capsGroupList = (int *) EG_alloc(numCAPSGroup*sizeof(int));
			            if (capsGroupList == NULL) {
			                status =  EGADS_MALLOC;
			                goto cleanup;
			            }

			            capsGroupList[numCAPSGroup-1] = attrIndex;
			        } else { // If we already have an index(es) let make sure it is unique
			            foundAttr = (int) false;
			            for (i = 0; i < numCAPSGroup; i++) {
			                if (attrIndex == capsGroupList[i]) {
			                    foundAttr = (int) true;
			                    break;
			                }
			            }

			            if (foundAttr == (int) false) {
			                numCAPSGroup += 1;
			                capsGroupList = (int *) EG_reall(capsGroupList, numCAPSGroup*sizeof(int));
			                if (capsGroupList == NULL) {
			                    status =  EGADS_MALLOC;
			                    goto cleanup;
			                }

			                capsGroupList[numCAPSGroup-1] = attrIndex;
			            }
			        }
			    }
			}

			// Get face tessellation
			status = EG_getTessFace(bodies[body+numBody], face+1, &plen, &xyz, &uv, &ptype, &pindex, &tlen, &tris, &nei);
			if (status != EGADS_SUCCESS) {
				printf(" mystranAIM: EG_getTessFace %d = %d for Body %d!\n", face+1, status, body+1);
				continue;
			}

			numFaceFound += 1;
			dataTransferBodyIndex = body;
			if (numFaceFound == 1) {
				bodyFaceMap = (int *) EG_alloc(2*numFaceFound*sizeof(int));
			} else {
				bodyFaceMap = (int *) EG_reall(bodyFaceMap, 2*numFaceFound*sizeof(int));
			}

			if (bodyFaceMap == NULL) {
				status = EGADS_MALLOC;
				goto cleanup;
			}

			// Get number of points and triangles
			bodyFaceMap[2*(numFaceFound-1) + 0] = body+1;
			bodyFaceMap[2*(numFaceFound-1) + 1] = face+1;

			// Sum number of points and triangles
			numPoint  += plen;
			numTri  += tlen;
		}

		if (faces != NULL) EG_free(faces);
		faces = NULL;


	    if (dataTransferBodyIndex >=0) break; // Force that only one body can be used
	}

	if (numFaceFound == 0) {
		printf(" mystranAIM/aimDiscr: No Faces match %s!\n", tname);

		status = CAPS_NOTFOUND;
		goto cleanup;
	}

	if ( numPoint == 0 || numTri == 0) {
		#ifdef DEBUG
	    	printf(" mystranAIM/aimDiscr: ntris = %d, npts = %d!\n", numTri, numPoint);
		#endif
	    status = CAPS_SOURCEERR;
	    goto cleanup;
	}

	#ifdef DEBUG
		printf(" mystranAIM/aimDiscr: Instance %d, Body Index for data transfer = %d\n", iIndex, dataTransferBodyIndex);
	#endif


	// Specify our single element type
	status = EGADS_MALLOC;
	discr->nTypes = 1;

	discr->types  = (capsEleType *) EG_alloc(sizeof(capsEleType));
	if (discr->types == NULL) goto cleanup;
	discr->types[0].nref  = 3;
	discr->types[0].ndata = 0;            /* data at geom reference positions */
	discr->types[0].ntri  = 1;
	discr->types[0].nmat  = 0;            /* match points at geom ref positions */
	discr->types[0].tris  = NULL;
	discr->types[0].gst   = NULL;
	discr->types[0].dst   = NULL;
	discr->types[0].matst = NULL;

	discr->types[0].tris   = (int *) EG_alloc(3*sizeof(int));
	if (discr->types[0].tris == NULL) goto cleanup;
	discr->types[0].tris[0] = 1;
	discr->types[0].tris[1] = 2;
	discr->types[0].tris[2] = 3;

	discr->types[0].gst   = (double *) EG_alloc(6*sizeof(double));
	if (discr->types[0].gst == NULL) goto cleanup;
	discr->types[0].gst[0] = 0.0;
	discr->types[0].gst[1] = 0.0;
	discr->types[0].gst[2] = 1.0;
	discr->types[0].gst[3] = 0.0;
	discr->types[0].gst[4] = 0.0;
	discr->types[0].gst[5] = 1.0;

	// Get the tessellation and make up a simple linear continuous triangle discretization */

	discr->nElems = numTri;

	discr->elems = (capsElement *) EG_alloc(discr->nElems*sizeof(capsElement));
	if (discr->elems == NULL) {
		status = EGADS_MALLOC;
		goto cleanup;
	}

	discr->mapping = (int *) EG_alloc(2*numPoint*sizeof(int)); // Will be resized
	if (discr->mapping == NULL) goto cleanup;

	globalID = (int *) EG_alloc(numPoint*sizeof(int));
	if (globalID == NULL) {
		status = EGADS_MALLOC;
		goto cleanup;
	}

	numPoint = 0;
	numTri = 0;
	for (face = 0; face < numFaceFound; face++){

		tess = bodies[bodyFaceMap[2*face + 0]-1 + numBody];

		if (localStitchedID == NULL) {
			status = EG_statusTessBody(tess, &tempBody, &i, &numGlobalPoint);

			localStitchedID = (int *) EG_alloc(numGlobalPoint*sizeof(int));
			if (localStitchedID == NULL) {
				status = EGADS_MALLOC;
				goto cleanup;
			}

			for (i = 0; i < numGlobalPoint; i++) localStitchedID[i] = 0;
		}

		// Get face tessellation
		status = EG_getTessFace(tess, bodyFaceMap[2*face + 1], &plen, &xyz, &uv, &ptype, &pindex, &tlen, &tris, &nei);
		if (status != EGADS_SUCCESS) {
			printf(" mystranAIM: EG_getTessFace %d = %d for Body %d!\n", bodyFaceMap[2*face + 1], status, bodyFaceMap[2*face + 0]);
			continue;
		}

		for (i = 0; i < plen; i++ ) {

			status = EG_localToGlobal(tess, bodyFaceMap[2*face+1], i+1, &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			if (localStitchedID[gID -1] != 0) continue;

			discr->mapping[2*numPoint  ] = bodyFaceMap[2*face + 0];
			discr->mapping[2*numPoint+1] = gID;

			localStitchedID[gID -1] = numPoint+1;

			globalID[numPoint] = gID;

			numPoint += 1;

		}

		// Get triangle connectivity in global sense
		for (i = 0; i < tlen; i++) {

			discr->elems[numTri].bIndex      = bodyFaceMap[2*face + 0];
			discr->elems[numTri].tIndex      = 1;
			discr->elems[numTri].eIndex      = bodyFaceMap[2*face + 1];

			discr->elems[numTri].gIndices    = (int *) EG_alloc(6*sizeof(int));
			if (discr->elems[numTri].gIndices == NULL) {
				status = EGADS_MALLOC;
				goto cleanup;
			}

			discr->elems[numTri].dIndices    = NULL;
			discr->elems[numTri].eTris.tq[0] = i+1;

			status = EG_localToGlobal(tess, bodyFaceMap[2*face + 1], tris[3*i + 0], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			discr->elems[numTri].gIndices[0] = localStitchedID[gID-1];
			discr->elems[numTri].gIndices[1] = tris[3*i + 0];

			status = EG_localToGlobal(tess, bodyFaceMap[2*face + 1], tris[3*i + 1], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			discr->elems[numTri].gIndices[2] = localStitchedID[gID-1];
			discr->elems[numTri].gIndices[3] = tris[3*i + 1];

			status = EG_localToGlobal(tess, bodyFaceMap[2*face + 1], tris[3*i + 2], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			discr->elems[numTri].gIndices[4] = localStitchedID[gID-1];
			discr->elems[numTri].gIndices[5] = tris[3*i + 2];

			numTri += 1;
		}
	}

	discr->nPoints = numPoint;

	#ifdef DEBUG
		printf(" mystranAIM/aimDiscr: ntris = %d, npts = %d!\n", discr->nElems, discr->nPoints);
	#endif

	// Resize mapping to switched together number of points
	discr->mapping = (int *) EG_reall(discr->mapping, 2*numPoint*sizeof(int));

	// Local to global node connectivity + numCAPSGroup + sizeof(capGrouplist)
	storage  = (int *) EG_alloc((numPoint + 1 + numCAPSGroup) *sizeof(int));
	if (storage == NULL) goto cleanup;
	discr->ptrm = storage;

	// Store the global node id
	for (i = 0; i < numPoint; i++) {

		storage[i] = globalID[i];

	    //#ifdef DEBUG
		//	printf(" mystranAIM/aimDiscr: Instance = %d, Global Node ID %d\n", iIndex, storage[i]);
		//#endif
	}

	// Save way the attrMap capsGroup list
	storage[numPoint] = numCAPSGroup;
	for (i = 0; i < numCAPSGroup; i++) {
	    storage[numPoint+1+i] = capsGroupList[i];
	}

	#ifdef DEBUG
        printf(" mystranAIM/aimDiscr: Instance = %d, Finished!!\n", iIndex);
    #endif

	status = CAPS_SUCCESS;

	goto cleanup;

	cleanup:
		if (faces != NULL) EG_free(faces);

		if (globalID  != NULL) EG_free(globalID);
		if (localStitchedID != NULL) EG_free(localStitchedID);

		if (capsGroupList != NULL) EG_free(capsGroupList);
		if (bodyFaceMap != NULL) EG_free(bodyFaceMap);

		if (status != CAPS_SUCCESS) {
			aimFreeDiscr(discr);
		}

		return status;
}

int aimLocateElement(capsDiscr *discr, double *params, double *param, int *eIndex,
                 	 double *bary)
{
	int    i, in[3], stat, ismall;
	double we[3], w, smallw = -1.e300;

	if (discr == NULL) return CAPS_NULLOBJ;

	for (ismall = i = 0; i < discr->nElems; i++) {
		in[0] = discr->elems[i].gIndices[0] - 1;
		in[1] = discr->elems[i].gIndices[2] - 1;
		in[2] = discr->elems[i].gIndices[4] - 1;
		stat  = EG_inTriExact(&params[2*in[0]], &params[2*in[1]], &params[2*in[2]], param, we);

		if (stat == EGADS_SUCCESS) {
			*eIndex = i+1;
			bary[0] = we[1];
			bary[1] = we[2];
			return CAPS_SUCCESS;
		}

		w = we[0];
		if (we[1] < w) w = we[1];
		if (we[2] < w) w = we[2];
		if (w > smallw) {
			ismall = i+1;
			smallw = w;
		}
	}

	/* must extrapolate! */
	if (ismall == 0) return CAPS_NOTFOUND;
	in[0] = discr->elems[ismall-1].gIndices[0] - 1;
	in[1] = discr->elems[ismall-1].gIndices[2] - 1;
	in[2] = discr->elems[ismall-1].gIndices[4] - 1;
	EG_inTriExact(&params[2*in[0]], &params[2*in[1]], &params[2*in[2]], param, we);

	*eIndex = ismall;
	bary[0] = we[1];
	bary[1] = we[2];

	/*
  	  printf(" aimLocateElement: extropolate to %d (%lf %lf %lf)  %lf\n", ismall, we[0], we[1], we[2], smallw);
	*/
	return CAPS_SUCCESS;
}

int aimTransfer(capsDiscr *discr, char *dataName, int numPoint, int dataRank, double *dataVal, char **units)
{
    int status; // Function return status
    int i, j, dataPoint; // Indexing


    // Current directory path
    char currentPath[PATH_MAX];

    char *extF06 = ".F06";

    // FO6 data variables
    int numGridPoint = 0;
    int numEigenVector = 0;

    double **dataMatrix = NULL;

    // Specific EigenVector to use
    int eigenVectorIndex = 0;

    // Variables used in global node mapping
    int *nodeMap, *storage;
    int globalNodeID;


    // Filename stuff
    char *filename = NULL;
    FILE *fp; // File pointer

    #ifdef DEBUG
        printf(" mystranAIM/aimTransfer name = %s  instance = %d  npts = %d/%d!\n", dataName, discr->instance, numPoint, dataRank);
    #endif

    //Get the appropriate parts of the tessellation to data
    storage = (int *) discr->ptrm;
    nodeMap = &storage[0]; // Global indexing on the body

    //capsGroupList = &storage[discr->nPoints]; // List of boundary ID (attrMap) in the transfer

    if (strcasecmp(dataName, "Displacement") != 0 &&
    	strncmp(dataName, "EigenVector", 11)  != 0) {

    	printf("Unrecognized data transfer variable - %s\n", dataName);
    	return CAPS_NOTFOUND;
    }

    (void) getcwd(currentPath, PATH_MAX);
    if (chdir(mystranInstance[discr->instance].analysisPath) != 0) {
        #ifdef DEBUG
            printf(" mystranAIM/aimTransfer Cannot chdir to %s!\n", mystranInstance[discr->instance].analysisPath);
        #endif

        return CAPS_DIRERR;
    }

    filename = (char *) EG_alloc((strlen(mystranInstance[discr->instance].projectName) +
                                  strlen(extF06) + 1)*sizeof(char));

    sprintf(filename,"%s%s",mystranInstance[discr->instance].projectName,extF06);

    // Open file
    fp = fopen(filename, "r");
    if (fp == NULL) {
        printf("Unable to open file: %s\n", filename);
        chdir(currentPath);

        if (filename != NULL) EG_free(filename);
        return CAPS_IOERR;
    }

    if (filename != NULL) EG_free(filename);
    filename = NULL;

    if (strcasecmp(dataName, "Displacement") == 0) {

        if (dataRank != 3) {

        	printf("Invalid rank for dataName \"%s\" - excepted a rank of 3!!!\n", dataName);
            status = CAPS_BADRANK;

        } else {

            status = mystran_readF06Displacement(fp,
                                                 1,
                                                 &numGridPoint,
                                                 &dataMatrix);
            fclose(fp);
        }

    } else if (strncmp(dataName, "EigenVector", 11) == 0) {

        // Which EigenVector do we want ?
    	for (i = 0; i < strlen(dataName); i++) {
    		if (dataName[i] == '_' ) break;
    	}

    	if (i == strlen(dataName)) {
    		eigenVectorIndex = 1;
    	} else {

    		status = sscanf(dataName, "EigenVector_%d", &eigenVectorIndex);
    		if (status != 1) {
    			printf("Unable to determine which EigenVector to use - Defaulting the first EigenVector!!!\n");
    			eigenVectorIndex = 1;
    		}
    	}

    	if (dataRank != 3) {

    		printf("Invalid rank for dataName \"%s\" - excepted a rank of 3!!!\n", dataName);
    		status = CAPS_BADRANK;

    	} else {

    		status = mystran_readF06EigenVector(fp,
												&numEigenVector,
												&numGridPoint,
												&dataMatrix);
    	}

        fclose(fp);

    } else {

        status = CAPS_NOTFOUND;
    }

    // Restore the path we came in
    chdir(currentPath);
    if (status != CAPS_SUCCESS) return status;


    for (i = 0; i < numPoint; i++) {

    	globalNodeID = nodeMap[i];

    	if (strcasecmp(dataName, "Displacement") == 0) {

    		for (dataPoint = 0; dataPoint < numGridPoint; dataPoint++) {
    			if ((int) dataMatrix[dataPoint][0] ==  globalNodeID) break;
    		}

    		dataVal[dataRank*i+0] = dataMatrix[dataPoint][2]; // T1
    		dataVal[dataRank*i+1] = dataMatrix[dataPoint][3]; // T2
    		dataVal[dataRank*i+2] = dataMatrix[dataPoint][4]; // T3

		} else if (strncmp(dataName, "EigenVector", 11) == 0) {


			for (dataPoint = 0; dataPoint < numGridPoint; dataPoint++) {
				if ((int) dataMatrix[eigenVectorIndex][8*dataPoint + 0] ==  globalNodeID) break;
			}

			dataVal[dataRank*i+0] = dataMatrix[eigenVectorIndex][8*dataPoint + 2]; // T1
			dataVal[dataRank*i+1] = dataMatrix[eigenVectorIndex][8*dataPoint + 3]; // T2
			dataVal[dataRank*i+2] = dataMatrix[eigenVectorIndex][8*dataPoint + 4]; // T3
			//dataVal[dataRank*i+3] = dataMatrix[eigenVectorIndex][8*dataPoint + 5]; // R1 - Don't use rotations
			//dataVal[dataRank*i+4] = dataMatrix[eigenVectorIndex][8*dataPoint + 6]; // R2
			//dataVal[dataRank*i+5] = dataMatrix[eigenVectorIndex][8*dataPoint + 7]; // R3

		}
    }

    // Free data matrix
    if (dataMatrix != NULL) {

        if      (strcasecmp(dataName, "Displacement") == 0) j = numGridPoint;
        else if (strcasecmp(dataName, "EigenVector")  == 0) j = numEigenVector;

        for (i = 0; i < j; i++) {
            if (dataMatrix[i] != NULL) EG_free(dataMatrix[i]);
        }
        EG_free(dataMatrix);
    }

    return CAPS_SUCCESS;
}
int aimInterpolation(capsDiscr *discr, /*@unused@*/ char *name, int eIndex,
                 double *bary, int rank, double *data, double *result)
{
	int    in[3], i;
	double we[3];
	/*
	#ifdef DEBUG
  	  printf(" mystranAIM/aimInterpolation  %s  instance = %d!\n",
         name, discr->instance);
	#endif
	 */
	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" mystranAIM/Interpolation: eIndex = %d [1-%d]!\n",eIndex, discr->nElems);
	}

	we[0] = 1.0 - bary[0] - bary[1];
	we[1] = bary[0];
	we[2] = bary[1];
	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[2] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[4] - 1;
	for (i = 0; i < rank; i++){
		result[i] = data[rank*in[0]+i]*we[0] + data[rank*in[1]+i]*we[1] + data[rank*in[2]+i]*we[2];
	}

	return CAPS_SUCCESS;
}


int aimInterpolateBar(capsDiscr *discr, /*@unused@*/ char *name, int eIndex,
					  double *bary, int rank, double *r_bar, double *d_bar)
{
	int    in[3], i;
	double we[3];

	/*
	#ifdef DEBUG
  	  printf(" mystranAIM/aimInterpolateBar  %s  instance = %d!\n", name, discr->instance);
	#endif
	 */

	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" mystranAIM/InterpolateBar: eIndex = %d [1-%d]!\n", eIndex, discr->nElems);
	}

	we[0] = 1.0 - bary[0] - bary[1];
	we[1] = bary[0];
	we[2] = bary[1];
	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[2] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[4] - 1;

	for (i = 0; i < rank; i++) {
		/*  result[i] = data[rank*in[0]+i]*we[0] + data[rank*in[1]+i]*we[1] +
                data[rank*in[2]+i]*we[2];  */
		d_bar[rank*in[0]+i] += we[0]*r_bar[i];
		d_bar[rank*in[1]+i] += we[1]*r_bar[i];
		d_bar[rank*in[2]+i] += we[2]*r_bar[i];
	}

	return CAPS_SUCCESS;
}


int aimIntegration(capsDiscr *discr, /*@unused@*/ char *name, int eIndex, int rank,
				   /*@null@*/ double *data, double *result)
{
	int    i, in[3], stat, ptype, pindex, nBody;
	double x1[3], x2[3], x3[3], xyz1[3], xyz2[3], xyz3[3], area;
	ego    *bodies;

	/*
	#ifdef DEBUG
  	  printf(" mystranAIM/aimIntegration  %s  instance = %d!\n", name, discr->instance);
	#endif
	 */
	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" mystranAIM/aimIntegration: eIndex = %d [1-%d]!\n", eIndex, discr->nElems);
	}

	stat = aim_getBodies(discr->aInfo, &nBody, &bodies);
	if (stat != CAPS_SUCCESS) {
		printf(" mystranAIM/aimIntegration: %d aim_getBodies = %d!\n", discr->instance, stat);
		return stat;
	}

	/* element indices */

	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[2] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[4] - 1;

	stat = EG_getGlobal(bodies[discr->mapping[2*in[0]]+nBody-1],
						discr->mapping[2*in[0]+1], &ptype, &pindex, xyz1);
	if (stat != CAPS_SUCCESS) {
		printf(" mystranAIM/aimIntegration: %d EG_getGlobal %d = %d!\n", discr->instance, in[0], stat);
		return stat;
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[1]]+nBody-1],
                        discr->mapping[2*in[1]+1], &ptype, &pindex, xyz2);
	if (stat != CAPS_SUCCESS) {
		printf(" mystranAIM/aimIntegration: %d EG_getGlobal %d = %d!\n", discr->instance, in[1], stat);
    	return stat;
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[2]]+nBody-1],
                        discr->mapping[2*in[2]+1], &ptype, &pindex, xyz3);
	if (stat != CAPS_SUCCESS) {
		printf(" mystranAIM/aimIntegration: %d EG_getGlobal %d = %d!\n", discr->instance, in[2], stat);
		return stat;
	}

	x1[0] = xyz2[0] - xyz1[0];
	x2[0] = xyz3[0] - xyz1[0];
	x1[1] = xyz2[1] - xyz1[1];
	x2[1] = xyz3[1] - xyz1[1];
	x1[2] = xyz2[2] - xyz1[2];
	x2[2] = xyz3[2] - xyz1[2];

	//CROSS(x3, x1, x2);
	cross_DoubleVal(x1, x2, x3);

	//area  = sqrt(DOT(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */
	area  = sqrt(dot_DoubleVal(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */

	if (data == NULL) {
		*result = 3.0*area;
		return CAPS_SUCCESS;
	}

	for (i = 0; i < rank; i++) {
		result[i] = (data[rank*in[0]+i] + data[rank*in[1]+i] + data[rank*in[2]+i])*area;
	}

	return CAPS_SUCCESS;
}


int aimIntegrateBar(capsDiscr *discr, /*@unused@*/ char *name, int eIndex, int rank,
                    double *r_bar, double *d_bar)
{
	int    i, in[3], stat, ptype, pindex, nBody;
	double x1[3], x2[3], x3[3], xyz1[3], xyz2[3], xyz3[3], area;
	ego    *bodies;

	/*
	#ifdef DEBUG
  	  printf(" mystranAIM/aimIntegrateBar  %s  instance = %d!\n", name, discr->instance);
	#endif
	 */

	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" mystranAIM/aimIntegrateBar: eIndex = %d [1-%d]!\n", eIndex, discr->nElems);
	}

	stat = aim_getBodies(discr->aInfo, &nBody, &bodies);
	if (stat != CAPS_SUCCESS) {
		printf(" mystranAIM/aimIntegrateBar: %d aim_getBodies = %d!\n", discr->instance, stat);
		return stat;
	}

	/* element indices */

	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[2] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[4] - 1;
	stat = EG_getGlobal(bodies[discr->mapping[2*in[0]]+nBody-1],
                       discr->mapping[2*in[0]+1], &ptype, &pindex, xyz1);
	if (stat != CAPS_SUCCESS) {
		printf(" mystranAIM/aimIntegrateBar: %d EG_getGlobal %d = %d!\n", discr->instance, in[0], stat);
		return stat;
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[1]]+nBody-1],
                       discr->mapping[2*in[1]+1], &ptype, &pindex, xyz2);

	if (stat != CAPS_SUCCESS) {
		printf(" mystranAIM/aimIntegrateBar: %d EG_getGlobal %d = %d!\n", discr->instance, in[1], stat);
		return stat;
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[2]]+nBody-1],
                      discr->mapping[2*in[2]+1], &ptype, &pindex, xyz3);

	if (stat != CAPS_SUCCESS) {
		printf(" mystranAIM/aimIntegrateBar: %d EG_getGlobal %d = %d!\n", discr->instance, in[2], stat);
		return stat;
	}

	x1[0] = xyz2[0] - xyz1[0];
	x2[0] = xyz3[0] - xyz1[0];
	x1[1] = xyz2[1] - xyz1[1];
	x2[1] = xyz3[1] - xyz1[1];
	x1[2] = xyz2[2] - xyz1[2];
	x2[2] = xyz3[2] - xyz1[2];

	//CROSS(x3, x1, x2);
	cross_DoubleVal(x1, x2, x3);

	//area  = sqrt(DOT(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */
	area  = sqrt(dot_DoubleVal(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */

	for (i = 0; i < rank; i++) {
	/*  result[i] = (data[rank*in[0]+i] + data[rank*in[1]+i] +
				 data[rank*in[2]+i])*area;  */
		d_bar[rank*in[0]+i] += area*r_bar[i];
		d_bar[rank*in[1]+i] += area*r_bar[i];
		d_bar[rank*in[2]+i] += area*r_bar[i];
	}

	return CAPS_SUCCESS;
}
