/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             FUN3D AIM
 *
 *     Written by Dr. Ryan Durscher and Dr. Ed Alyanak AFRL/RQVC
 *
 */

/*! \mainpage Introduction
 *
 * \section overviewNastran Nastran AIM Overview
 * A module in the Computational Aircraft Prototype Syntheses (CAPS) has been developed to interact (primarily
 * through input files) with the finite element structural solver Nastran.
 *
 * Current issues include:
 *  - A thorough bug testing needs to be undertaken.
 *  .
 *
 * \section temporay BLAH
 *
 * Related pages:
 * - \ref feaMaterial
 * - \ref feaProperty
 * - \ref feaConstraint
 * - \ref feaLoad
 * - \ref feaAnalysis
 * - \ref feaDesignVariable
 * - \ref feaDesignConstraint
 * - \ref geomFidelityNastran
 * - \ref aimInputsNastran
 * - \ref aimOutputsNastran
 * - \ref sharableDataNastran
 */

#include <string.h>
#include <math.h>
#include "capsTypes.h"
#include "aimUtil.h"

#include "meshUtils.h"
#include "miscUtils.h"
#include "feaUtils.h"
#include "nastranUtils.h"

#ifdef WIN32
#define getcwd     _getcwd
#define snprintf   _snprintf
#define strcasecmp stricmp
#define PATH_MAX   _MAX_PATH
#else
#include <unistd.h>
#include <limits.h>
#endif

#define NUMINPUT   17
#define NUMOUTPUT  1

#define Fidelity1D 100 // Base fidelity value for 1D simulations
#define Fidelity2D 200 // Base fidelity value for 2D simulations
#define Fidelity3D 300 // Base fidelity value for 3D simulations
#define NUMFID    6 // Number of fidelities contained in the aim

#define MXCHAR  255

//#define DEBUG

typedef struct {

	// Project name
	char *projectName; // Project name

    // Analysis file path/directory
    char *analysisPath;

    feaProblemStruct feaProblem;

    // Attribute to index map
    mapAttrToIndexStruct attrMap;

    // Flag to see if we need to get an FEA mesh
    int haveFEAMesh;

    // Pointer to caps input value for tessellation parameters
    capsValue *tessParam;

    // Pointer to caps input value for Edge_Point- Min
    capsValue *edgePointMin;

    // Pointer to caps input value for Edge_Point- Max
    capsValue *edgePointMax;

    // Pointer to caps input value for No_Quad_Faces
    capsValue *noQuadFace;

} aimStorage;

static aimStorage *nastranInstance = NULL;
static int         numInstance  = 0;

extern char *EG_strdup(const char *str);
extern int   EG_inTriExact(double *t1, double *t2, double *t3, double *p,  double *w );

static int createMesh(int iIndex, void *aimInfo) {

	int status; // Function return status

    int i, body; // Indexing

	// Bodies
	int   numBody; // Number of Bodies
	ego  *bodies;

	// genFEAMeshStruct *feaMesh; // FEA meshes

	// Meshing related variables
	double  tessParam[3];
	int edgePointMin = 2;
	int edgePointMax = 50;
	int noQuadFace = 1;

	// Coordinate system
	mapAttrToIndexStruct coordSystemMap;

	// AIM Data transfer
	int  nrow, ncol, rank;
	void *dataTransfer;
	enum capsvType vtype;
	char *units;

	// Inherited fea/volume mesh related variables
	int numFEAMesh = 0;
	genUnstrMeshStruct volumeMesh;
	int feaMeshInherited = (int) false;

	// Get AIM bodies
	status = aim_getBodies(aimInfo, &numBody, &bodies);

	#ifdef DEBUG
	   	printf(" nastranAIM/createMesh instance = %d  nbody = %d!\n", iIndex, numBody);
	#endif

	if ((numBody <= 0) || (bodies == NULL)) {
		#ifdef DEBUG
		printf(" nastranAIM/createMesh No Bodies!\n");
	#endif

		return CAPS_SOURCEERR;
	}

	// Print number of body warning
	if (numBody != 1) {

		printf("nastranAIM only supports single body sources!!!!\n");
		return CAPS_SOURCEERR;
	}

	// Get CoordSystem attribute to index mapping
	status = initiate_mapAttrToIndexStruct(&coordSystemMap);
	if (status != CAPS_SUCCESS) return status;

	status = create_CoordSystemAttrToIndexMap(numBody,
									   bodies,
									   3, //>2 - search the body, faces, edges, and all the nodes
									   &coordSystemMap);
	if (status != CAPS_SUCCESS) goto cleanup;

	status = fea_getCoordSystem(numBody,
						 	 	bodies,
								coordSystemMap,
								&nastranInstance[iIndex].feaProblem.numCoordSystem,
								&nastranInstance[iIndex].feaProblem.feaCoordSystem);
	if (status != CAPS_SUCCESS) goto cleanup;

	// Get attribute to index mapping
	status = aim_getData(aimInfo, "Attribute_Map", &vtype, &rank, &nrow, &ncol, &dataTransfer, &units);
	if (status == CAPS_SUCCESS) {

		printf("Found link for attrMap (Attribute_Map) from parent\n");

		status = copy_mapAttrToIndexStruct( (mapAttrToIndexStruct *) dataTransfer, &nastranInstance[iIndex].attrMap);
		if (status != CAPS_SUCCESS) goto cleanup;

	} else {

		if (status == CAPS_DIRTY) printf("Parent AIMS are dirty\n");

		printf("Didn't find a link to attrMap from parent - getting it ourselves\n");

		// Get capsGroup name and index mapping to make sure all faces have a capsGroup value
		status = create_CAPSGroupAttrToIndexMap(numBody,
				                                bodies,
	                                            3, //>2 - search the body, faces, edges, and all the nodes
	                                            &nastranInstance[iIndex].attrMap);
		if (status != CAPS_SUCCESS) goto cleanup;
	}

	// See if a FEA mesh is available from parent
	status = aim_getData(aimInfo, "FEA_MESH", &vtype, &rank, &nrow, &ncol, &dataTransfer, &units);
	if (status == CAPS_SUCCESS) {

		printf("Found link for a FEA mesh (FEA_Mesh) from parent\n");

		numFEAMesh = 1;
		if      (nrow == 1 && ncol != 1) numFEAMesh = ncol;
		else if (nrow != 1 && ncol == 1) numFEAMesh = nrow;
		else {

			printf("Can not except 2D matrix of fea meshes\n");
			status = CAPS_BADVALUE;
			goto cleanup;
		}

		if (numFEAMesh != numBody) {
			printf("Number of inherited fea meshes does not match the number of bodies\n");
			return CAPS_SOURCEERR;
		}

		status = lump_genFEAMesh(numFEAMesh,
								 (genFEAMeshStruct *) dataTransfer,
								 &nastranInstance[iIndex].feaProblem.feaMesh);
		if (status != CAPS_SUCCESS) goto cleanup;

		feaMeshInherited = (int) true;

	} else if (status == CAPS_DIRTY) {

		printf("Link for a FEA mesh (FEA_Mesh) from parent found, but Parent AIMS are dirty\n");
		goto cleanup;

	} else { // Check to see if a general unstructured volume mesh is available

		printf("Didn't find a FEA mesh (FEA_Mesh) from parent - checking for general unstructured volume mesh\n");

		status = aim_getData(aimInfo, "Volume_MESH", &vtype, &rank, &nrow, &ncol, &dataTransfer, &units);
		if (status == CAPS_DIRTY) {

			printf("Link for a general unstructured volume mesh (Volume_Mesh) from parent found, but Parent AIMS are dirty\n");
			goto cleanup;

		} else if (status == CAPS_SUCCESS) {

			printf("Found link for a general unstructured volume mesh (Volume_Mesh) from parent\n");

			numFEAMesh = 1;
			if (nrow != 1 && ncol != 1) {

				printf("Can not except multiple volume meshes\n");
				status = CAPS_BADVALUE;
				goto cleanup;
			}

			if (numFEAMesh != numBody) {
				printf("Number of inherited volume mesh does not match the number of bodies - assuming volume mesh is already lumped\n");
			}

			status = initiate_genUnstrMeshStruct(&volumeMesh);
			if (status != CAPS_SUCCESS) goto cleanup;

			volumeMesh = *(genUnstrMeshStruct *) dataTransfer;

			// Transfer just the volume mesh
			nastranInstance[iIndex].feaProblem.feaMesh.numNodes = volumeMesh.numNodes;
			nastranInstance[iIndex].feaProblem.feaMesh.numTetra = volumeMesh.numTetra;
			nastranInstance[iIndex].feaProblem.feaMesh.numHexa  = volumeMesh.numHex;
			nastranInstance[iIndex].feaProblem.feaMesh.numElement3DPID = volumeMesh.numVolID;

			printf("Number of nodes - %d\n", nastranInstance[iIndex].feaProblem.feaMesh.numNodes);
			printf("Number of tetrahedral - %d\n", nastranInstance[iIndex].feaProblem.feaMesh.numTetra);
			printf("Number of hexahedral - %d\n", nastranInstance[iIndex].feaProblem.feaMesh.numHexa);
			printf("Number of 3D PIDs - %d\n", nastranInstance[iIndex].feaProblem.feaMesh.numElement3DPID);

			// Populate feaMesh from volume mesh - getting only solid elements
			if (volumeMesh.numTetra != 0 || volumeMesh.numHex != 0) {

				// Get grid points
				if (volumeMesh.numNodes != 0) {
					nastranInstance[iIndex].feaProblem.feaMesh.xyz = (double *) EG_alloc(3*volumeMesh.numNodes*sizeof(double));
					if (nastranInstance[iIndex].feaProblem.feaMesh.xyz == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					nastranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex = (int *) EG_alloc(volumeMesh.numNodes*sizeof(int));
					if (nastranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					nastranInstance[iIndex].feaProblem.feaMesh.xyzCoordID = (int *) EG_alloc(volumeMesh.numNodes*sizeof(int));
					if (nastranInstance[iIndex].feaProblem.feaMesh.xyzCoordID == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					memcpy(nastranInstance[iIndex].feaProblem.feaMesh.xyz,
						   volumeMesh.xyz,
						   3*volumeMesh.numNodes*sizeof(double));

					// Set default grid points to CAPS_MAGIC number
					for (i = 0; i < volumeMesh.numNodes; i++) {
						nastranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[i] = CAPSMAGIC;
						nastranInstance[iIndex].feaProblem.feaMesh.xyzCoordID[i] = 0;
					}

					// Search surface meshes to get grid attrIndex on boundaries
					for (i = 0; i < volumeMesh.surfaceMesh.numTriFace; i++) { // TriFace first
						nastranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[volumeMesh.surfaceMesh.triFaceList[3*i+0]-1] = volumeMesh.surfaceMesh.triFaceMarkList[i];
						nastranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[volumeMesh.surfaceMesh.triFaceList[3*i+1]-1] = volumeMesh.surfaceMesh.triFaceMarkList[i];
						nastranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[volumeMesh.surfaceMesh.triFaceList[3*i+2]-1] = volumeMesh.surfaceMesh.triFaceMarkList[i];
					}

					for (i = 0; i < volumeMesh.surfaceMesh.numQuadFace; i++) { // QuadFace first
						nastranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[volumeMesh.surfaceMesh.quadFaceList[4*i+0]-1] = volumeMesh.surfaceMesh.quadFaceMarkList[i];
						nastranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[volumeMesh.surfaceMesh.quadFaceList[4*i+1]-1] = volumeMesh.surfaceMesh.quadFaceMarkList[i];
						nastranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[volumeMesh.surfaceMesh.quadFaceList[4*i+2]-1] = volumeMesh.surfaceMesh.quadFaceMarkList[i];
						nastranInstance[iIndex].feaProblem.feaMesh.xyzAttrIndex[volumeMesh.surfaceMesh.quadFaceList[4*i+3]-1] = volumeMesh.surfaceMesh.quadFaceMarkList[i];
					}
				}

				// Get global list volume IDs
				if (volumeMesh.numVolID != 0) {

					nastranInstance[iIndex].feaProblem.feaMesh.element3DPID = (int *) EG_alloc(volumeMesh.numVolID*sizeof(int));
					if (nastranInstance[iIndex].feaProblem.feaMesh.element3DPID == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					memcpy(nastranInstance[iIndex].feaProblem.feaMesh.element3DPID,
						   volumeMesh.volIDList,
						   volumeMesh.numVolID*sizeof(int));
				}

				// Get tetra elements
				if (volumeMesh.numTetra != 0) {

					nastranInstance[iIndex].feaProblem.feaMesh.ctetra    = (int *) EG_alloc(4*volumeMesh.numTetra*sizeof(int));
					if (nastranInstance[iIndex].feaProblem.feaMesh.ctetra == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					nastranInstance[iIndex].feaProblem.feaMesh.ctetraPID = (int *) EG_alloc(  volumeMesh.numTetra*sizeof(int));
					if (nastranInstance[iIndex].feaProblem.feaMesh.ctetraPID == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					nastranInstance[iIndex].feaProblem.feaMesh.ctetraElementID = (int *) EG_alloc(  volumeMesh.numTetra*sizeof(int));
					if (nastranInstance[iIndex].feaProblem.feaMesh.ctetraElementID == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					memcpy(nastranInstance[iIndex].feaProblem.feaMesh.ctetra,
						   volumeMesh.tetraList,
						   4*volumeMesh.numTetra*sizeof(int));

					memcpy(nastranInstance[iIndex].feaProblem.feaMesh.ctetraPID,
						   volumeMesh.tetraVolIDList,
						   volumeMesh.numTetra*sizeof(int));

					for (i = 0; i < nastranInstance[iIndex].feaProblem.feaMesh.numTetra; i++) {
						nastranInstance[iIndex].feaProblem.feaMesh.ctetraElementID[i] = i+1;
					}
				}

				// Get hexahedral elements
				if (volumeMesh.numHex != 0) {

					nastranInstance[iIndex].feaProblem.feaMesh.chexa    = (int *) EG_alloc(8*volumeMesh.numHex*sizeof(int));
					if (nastranInstance[iIndex].feaProblem.feaMesh.chexa == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					nastranInstance[iIndex].feaProblem.feaMesh.chexaPID = (int *) EG_alloc(  volumeMesh.numHex*sizeof(int));
					if (nastranInstance[iIndex].feaProblem.feaMesh.chexaPID == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					nastranInstance[iIndex].feaProblem.feaMesh.chexaElementID = (int *) EG_alloc(  volumeMesh.numHex*sizeof(int));
					if (nastranInstance[iIndex].feaProblem.feaMesh.chexaElementID == NULL) {
						status = EGADS_MALLOC;
						goto cleanup;
					}

					memcpy(nastranInstance[iIndex].feaProblem.feaMesh.chexa,
						   volumeMesh.hexList,
						   8*volumeMesh.numHex*sizeof(int));

					memcpy(nastranInstance[iIndex].feaProblem.feaMesh.chexaPID,
						   volumeMesh.hexVolIDList,
						   volumeMesh.numHex*sizeof(int));

					for (i = 0; i < nastranInstance[iIndex].feaProblem.feaMesh.numHexa; i++) {
						nastranInstance[iIndex].feaProblem.feaMesh.chexaElementID[i] = i+1 + nastranInstance[iIndex].feaProblem.feaMesh.numTetra;
					}
				}
			}

			feaMeshInherited = (int) true;
		} else {
			printf("Didn't find a general unstructured volume mesh (Volume_Mesh) from parent\n");

			feaMeshInherited = (int) false;
		}
	}


	// If we didn't inherit a FEA mesh we need to get one ourselves
	if (feaMeshInherited == (int) false) {
		// Tessellate body(ies)
		// Tess_Params -> -1 because of aim_getIndex 1 bias
		/*
		tessParams[0] = aimInputs[aim_getIndex(aimInfo, "Tess_Params", ANALYSISIN)-1].vals.reals[0]; // Gets multiplied by bounding box size
		tessParams[1] = aimInputs[aim_getIndex(aimInfo, "Tess_Params", ANALYSISIN)-1].vals.reals[1]; // Gets multiplied by bounding box size
		tessParams[2] = aimInputs[aim_getIndex(aimInfo, "Tess_Params", ANALYSISIN)-1].vals.reals[2];

		edgePointMin = aimInputs[aim_getIndex(aimInfo, "Edge_Point_Min", ANALYSISIN)-1].vals.integer;
		edgePointMax = aimInputs[aim_getIndex(aimInfo, "Edge_Point_Max", ANALYSISIN)-1].vals.integer;
		noQuadFaces  = aimInputs[aim_getIndex(aimInfo, "No_Quad_Faces", ANALYSISIN)-1].vals.integer;
		 */

		tessParam[0] = nastranInstance[iIndex].tessParam->vals.reals[0]; // Gets multiplied by bounding box size
		tessParam[1] = nastranInstance[iIndex].tessParam->vals.reals[1]; // Gets multiplied by bounding box size
		tessParam[2] = nastranInstance[iIndex].tessParam->vals.reals[2];

		edgePointMin = nastranInstance[iIndex].edgePointMin->vals.integer;
		edgePointMax = nastranInstance[iIndex].edgePointMax->vals.integer;
		noQuadFace   = nastranInstance[iIndex].noQuadFace->vals.integer;

		if (edgePointMin < 2) {
			printf("The minimum number of allowable edge points is 2 not %d\n", edgePointMin);
			edgePointMin = 2;
		}

		if (edgePointMax < edgePointMin) {
			printf("The maximum number of edge points must be greater than the current minimum (%d)\n", edgePointMin);
			edgePointMax = edgePointMin+1;
		}

		// Allocate fea meshes
		//feaMesh = (genFEAMeshStruct *) EG_alloc(numBody*sizeof(genFEAMeshStruct));
		//if (feaMesh == NULL) return EGADS_MALLOC;

		// Get FEA Problem from EGADs body
		status = fea_bodyToBEM(bodies[0], // (in)  EGADS Body
	                           tessParam,          // (in)  Tessellation parameters
							   edgePointMin,        // (in)  minimum points along any Edge
							   edgePointMax,        // (in)  maximum points along any Edge
							   noQuadFace,
							   &nastranInstance[iIndex].attrMap,
							   &coordSystemMap,
							   &nastranInstance[iIndex].feaProblem.feaMesh);

		if (status != CAPS_SUCCESS) goto cleanup;

		printf("Number nodes = %d\n",nastranInstance[iIndex].feaProblem.feaMesh.numNodes);
		printf("Number Rods = %d\n" ,nastranInstance[iIndex].feaProblem.feaMesh.numRod);
		printf("Number Tria3 = %d\n",nastranInstance[iIndex].feaProblem.feaMesh.numTria3);
		printf("Number Quad4 = %d\n",nastranInstance[iIndex].feaProblem.feaMesh.numQuad4);
		printf("Number of 2D element PIDs = %d\n", nastranInstance[iIndex].feaProblem.feaMesh.numElement2DPID);

		status = aim_setTess(aimInfo, nastranInstance[iIndex].feaProblem.feaMesh.bodyTessMapping.egadsTess);
		if (status != CAPS_SUCCESS) goto cleanup;
	}

	status = CAPS_SUCCESS;

	goto cleanup;

	cleanup:

		(void ) destroy_mapAttrToIndexStruct(&coordSystemMap);

		return status;
}


/* ********************** Exposed AIM Functions ***************************** */

int aimInitialize(int ngIn, capsValue *gIn,
                  int *nFid, int **fids, int *nIn, int *nOut,
                  int *nFields, char ***fnames, int **ranks)
{
    int status; // Return flag
    int *ints;

    aimStorage *tmp;

    #ifdef DEBUG
        printf("\n nastranAIM/aimInitialize   ngIn = %d!\n", ngIn);
    #endif

	*nFid    = 6;
	ints     = (int *) EG_alloc(((int) *nFid)*sizeof(int));
	if (ints == NULL) return EGADS_MALLOC;

	/*! \page geomFidelityNastran Geometry Fidelity
	 * The attribute capsFidelity may be set to either ALL or STRUCTURE for the Nastran AIM. The geometric fidelity for the
	 * AIM requires that bodies be:
	 *  - WIREBODY for purely 1D simulations
	 *  - FACEBODY or SHEETBODY (non-manifold) for 2D simulations
	 *  - SOLIDBODY or SHEETBODY (manifold) for 3D simulations
	 */
	ints[0]  = ALL | ALL;
	ints[1]  = ALL | STRUCTURE;
	ints[2]  = SOLIDBODY | STRUCTURE;
	ints[3]  = SHEETBODY | STRUCTURE;
	ints[4]  = FACEBODY | STRUCTURE;
	ints[5]  = WIREBODY | STRUCTURE;
	*fids    = ints;

    /* specify the number of analysis input and out "parameters" */
    *nIn     = NUMINPUT;
    *nOut    = NUMOUTPUT;

    /* specify the field variables this analysis can generate */
    *nFields = 0;
    *ranks   = NULL;
    *fnames  = NULL;

    // Allocate nastranInstance
    if (nastranInstance == NULL) {

    	nastranInstance = (aimStorage *) EG_alloc(sizeof(aimStorage));

    	if (nastranInstance == NULL) {
    		EG_free(*fids);
    		*fids   = NULL;
    		return EGADS_MALLOC;
    	}

    } else {
    	tmp = (aimStorage *) EG_reall(nastranInstance, (numInstance+1)*sizeof(aimStorage));
    	if (tmp == NULL) {
    		EG_free(*fids);
    		*fids   = NULL;
    		return EGADS_MALLOC;
    	}

    	nastranInstance = tmp;
    }

    // Set initial values for nastranInstance
    nastranInstance[numInstance].projectName = NULL;

    // Analysis file path/directory
    nastranInstance[numInstance].analysisPath = NULL;

    /*
    // Check to make sure data transfer is ok
    nastranInstance[numInstance].dataTransferCheck = (int) true;
    */

    // Flag to see if we need to get an FEA mesh
    nastranInstance[numInstance].haveFEAMesh = (int) false;

    // Container for attribute to index map
    status = initiate_mapAttrToIndexStruct(&nastranInstance[numInstance].attrMap);
    if (status != CAPS_SUCCESS) return status;

    status = initiate_feaProblemStruct(&nastranInstance[numInstance].feaProblem);
    if (status != CAPS_SUCCESS) return status;

    // Pointer to caps input value for tessellation parameters
    nastranInstance[numInstance].tessParam = NULL;

    // Pointer to caps input value for Edge_Point- Min
    nastranInstance[numInstance].edgePointMin = NULL;

    // Pointer to caps input value for Edge_Point- Max
    nastranInstance[numInstance].edgePointMax = NULL;

    // Pointer to caps input value for No_Quad_Faces
    nastranInstance[numInstance].noQuadFace = NULL;

    // Increment number of instances
    numInstance += 1;

    return (numInstance -1);
}

int aimInputs(int iIndex, void *aimInfo, int index, char **ainame, capsValue *defval)
{

	/*! \page aimInputsNastran AIM Inputs
	 * The following list outlines the Nastran inputs along with their default value available
	 * through the AIM interface. Unless noted these values will be not be linked to
	 * any parent AIMs with variables of the same name.
	 */

	int status;

	#ifdef DEBUG
		printf(" nastranAIM/aimInputs instance = %d  index = %d!\n", inst, index);
	#endif

	*ainame = NULL;
  
	// Nastran Inputs
	if (index == 1) {
	    *ainame              = EG_strdup("Proj_Name");
		defval->type         = String;
		defval->nullVal      = NotNull;
		defval->vals.string  = EG_strdup("nastran_CAPS");
		defval->lfixed       = Change;

		/*! \page aimInputsNastran
		 * - <B> Proj_Name = "nastran_CAPS"</B> <br>
		 * This corresponds to the project name used for file naming.
		 */

	} else if (index == 2) {
        *ainame               = EG_strdup("Tess_Params");
        defval->type          = Double;
        defval->dim           = 1;
        defval->length        = 3;
        defval->nrow          = 3;
        defval->ncol          = 1;
        defval->units         = NULL;
        defval->lfixed        = Fixed;
        defval->vals.reals    = (double *) EG_alloc(defval->length*sizeof(double));
        if (defval->vals.reals != NULL) {
        	defval->vals.reals[0] = 0.025;
        	defval->vals.reals[1] = 0.001;
        	defval->vals.reals[2] = 15.00;
        } else return EGADS_MALLOC;

        nastranInstance[iIndex].tessParam = defval;

        /*! \page aimInputsNastran
		 * - <B> Tess_Params = [0.025, 0.001, 15.0]</B> <br>
		 * Body tessellation parameters used when creating a boundary element model.
		 * Tess_Params[0] and Tess_Params[1] get scaled by the bounding
		 * box of the body. (From the EGADS manual) A set of 3 parameters that drive the EDGE discretization
		 * and the FACE triangulation. The first is the maximum length of an EDGE segment or triangle side
		 * (in physical space). A zero is flag that allows for any length. The second is a curvature-based
		 * value that looks locally at the deviation between the centroid of the discrete object and the
		 * underlying geometry. Any deviation larger than the input value will cause the tessellation to
		 * be enhanced in those regions. The third is the maximum interior dihedral angle (in degrees)
		 * between triangle facets (or Edge segment tangents for a WIREBODY tessellation), note that a
		 * zero ignores this phase
		 */

	} else if (index == 3) {
        *ainame               = EG_strdup("Edge_Point_Min");
        defval->type          = Integer;
        defval->vals.integer  = 2;
        defval->lfixed        = Fixed;
        defval->nrow          = 1;
        defval->ncol          = 1;

        nastranInstance[iIndex].edgePointMin = defval;

        /*! \page aimInputsNastran
		 * - <B> Edge_Point_Min = 2</B> <br>
		 * Minimum number of points along an edge to use when creating a boundary element model.
		 */

	} else if (index == 4) {
		*ainame               = EG_strdup("Edge_Point_Max");
		defval->type          = Integer;
		defval->vals.integer  = 10;
		defval->length        = 1;
		defval->lfixed        = Fixed;
		defval->nrow          = 1;
		defval->ncol          = 1;

		nastranInstance[iIndex].edgePointMax = defval;

		/*! \page aimInputsNastran
		 * - <B> Edge_Point_Max = 10</B> <br>
		 * Maximum number of points along an edge to use when creating a boundary element model.
		 */

	} else if (index == 5) {
		*ainame               = EG_strdup("No_Quad_Faces");
		defval->type          = Boolean;
		defval->vals.integer  = (int) false;

		nastranInstance[iIndex].noQuadFace = defval;

		/*! \page aimInputsNastran
		 * - <B> Edge_Point_Max = 10</B> <br>
		 * Maximum number of points along an edge to use when creating a boundary element model.
		 */

	} else if (index == 6) {
        *ainame              = EG_strdup("Property");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;
        defval->dim          = 1;

        /*! \page aimInputsNastran
		 * - <B> Property = NULL</B> <br>
		 * Property tuple used to input property information for the model, see \ref feaProperty for additional details.
		 */
	} else if (index == 7) {
	    *ainame              = EG_strdup("Material");
	    defval->type         = Tuple;
	    defval->nullVal      = IsNull;
	    //defval->units        = NULL;
	    defval->lfixed       = Change;
	    defval->vals.tuple   = NULL;
	    defval->dim          = 1;

	    /*! \page aimInputsNastran
		 * - <B> Material = NULL</B> <br>
		 * Material tuple used to input material information for the model, see \ref feaMaterial for additional details.
		 */
	} else if (index == 8) {
        *ainame              = EG_strdup("Constraint");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;
        defval->dim          = 1;

        /*! \page aimInputsNastran
		 * - <B> Constraint = NULL</B> <br>
		 * Constraint tuple used to input constraint information for the model, see \ref feaConstraint for additional details.
		 */
	} else if (index == 9) {
        *ainame              = EG_strdup("Load");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;
        defval->dim          = 1;

        /*! \page aimInputsNastran
		 * - <B> Load = NULL</B> <br>
		 * Load tuple used to input load information for the model, see \ref feaLoad for additional details.
		 */
	} else if (index == 10) {
        *ainame              = EG_strdup("Analysis");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;
        defval->dim          = 1;

        /*! \page aimInputsNastran
		 * - <B> Analysis = NULL</B> <br>
		 * Analysis tuple used to input analysis/case information for the model, see \ref feaAnalysis for additional details.
		 */
    } else if (index == 11) {
        *ainame              = EG_strdup("Analysis_Type");
        defval->type         = String;
        defval->nullVal      = NotNull;
        defval->vals.string  = EG_strdup("Modal"); // Modal, Static, StaticOpt
        defval->lfixed       = Change;

        /*! \page aimInputsNastran
		 * - <B> Analysis_Type = "Modal"</B> <br>
		 * Type of analysis to generate files for, options include "Modal", "Static", and "StaticOpt".
		 */
    } else if (index == 12) {
	    *ainame              = EG_strdup("File_Format");
        defval->type         = String;
        defval->vals.string  = EG_strdup("Small"); // Small, Large, Free
        defval->lfixed       = Change;

        /*! \page aimInputsNastran
         * - <B> File_Format = "Small"</B> <br>
         * Formatting type for the bulk file. Options: "Small", "Large", "Free".
         */

    } else if (index == 13) {
	    *ainame              = EG_strdup("Mesh_File_Format");
        defval->type         = String;
        defval->vals.string  = EG_strdup("Free"); // Small, Large, Free
        defval->lfixed       = Change;

        /*! \page aimInputsNastran
		 * - <B> Mesh_File_Format = "Small"</B> <br>
		 * Formatting type for the mesh file. Options: "Small", "Large", "Free".
		 */

    } else if (index == 14) {
        *ainame              = EG_strdup("Design_Variable");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;
        defval->dim          = 1;

        /*! \page aimInputsNastran
		 * - <B> Design_Variable = NULL</B> <br>
		 * The design variable tuple used to input design variable information for the model optimization, see \ref feaDesignVariable for additional details.
		 */

    } else if (index == 15) {
    	*ainame              = EG_strdup("Design_Constraint");
    	defval->type         = Tuple;
    	defval->nullVal      = IsNull;
    	//defval->units        = NULL;
    	defval->lfixed       = Change;
    	defval->vals.tuple   = NULL;
    	defval->dim          = 1;

    	/*! \page aimInputsNastran
    	 * - <B> Design_Constraint = NULL</B> <br>
		 * The design constraint tuple used to input design constraint information for the model optimization, see \ref feaDesignConstraint for additional details.
		 */

    } else if (index == 16) {
    	*ainame              = EG_strdup("ObjectiveMinMax");
    	defval->type         = String;
    	defval->nullVal      = NotNull;
    	defval->vals.string  = EG_strdup("Max"); // Max, Min
    	defval->lfixed       = Change;

    	/*! \page aimInputsNastran
		 * - <B> ObjectiveMinMax = "Max"</B> <br>
		 * Maximize or minimize the design objective during an optimization. Option: "Max" or "Min".
		 */

    } else if (index == 17) {
    	*ainame              = EG_strdup("ObjectiveResponseType");
    	defval->type         = String;
    	defval->nullVal      = NotNull;
    	defval->vals.string  = EG_strdup("Weight"); // Weight
    	defval->lfixed       = Change;

    	/*! \page aimInputsNastran
		 * - <B> ObjectiveResponseType = "Weight"</B> <br>
		 * Object response type (see Nastran manual).
		 */
    }

	// Link variable(s) to parent(s) if available
	if ((index != 1) && (*ainame != NULL) && index < 6 ) {
		status = aim_link(aimInfo, *ainame, ANALYSISIN, defval);

		/*
  	 	printf("Status = %d: Var Index = %d, Type = %d, link = %lX\n",
                        status, index, defval->type, defval->link);
      	*/
	}

	return CAPS_SUCCESS;
}

int aimData(int iIndex, const char *name, enum capsvType *vtype, int *rank, int *nrow,
        int *ncol, void **data, char **units)
{

    /*! \page sharableDataNastran AIM Shareable Data
     * Currently the Nastran AIM does not have any shareable data types or values. It will try, however, to inherit a "FEA_MESH"
     * or "Volume_Mesh" from any parent AIMs. Note that the inheritance of the mesh is not required.
     */

    #ifdef DEBUG
        printf(" nastranAIM/aimData instance = %d  name = %s!\n", iIndex, name);
    #endif
   return CAPS_NOTFOUND;
}

int aimPreAnalysis(int iIndex, void *aimInfo, char *analysisPath,
                   capsValue *aimInputs, capsErrs **errs)
{

    int i, j, k; // Indexing

    int status; // Status return

    int found; // Boolean operator

    int *tempIntegerArray = NULL; // Temporary array to store a list of integers

    // CAPS function returns
    /*
	int 		  atype, alen;
	const int    *ints;
	const char   *string;
	const double *reals;
     */

	// Analysis information
	char *analysisType;

	// Optimization Information
	char *objectiveMinMax, *objectiveResp;

	// File format information
	char *tempString, *delimiter;

	// File IO
	char *filename; // Output file name
    FILE *fp; // Output file pointer


    // Discrete data transfer variables
    capsDiscr *dataTransferDiscreteObj;
    char **transferName = NULL;
    int numTransferName, transferNameIndex;
    enum capsdMethod dataTransferMethod;
    int numDataTransferPoint;
    int numDataTransferElement = 0;
    int dataTransferRank;
    double *dataTransferData;

    // Variables used in global node mapping
    int *tessNodeMap, *storage;
    int globalNodeID;
    int nodeIndex[4], transferIndex[4], elementID, elementIndex;

    // NULL out errors
    *errs = NULL;
  
    // Store away the analysis path/directory
    nastranInstance[iIndex].analysisPath = analysisPath;

    // Get project name
    nastranInstance[iIndex].projectName = aimInputs[aim_getIndex(aimInfo, "Proj_Name", ANALYSISIN)-1].vals.string;

    // Get FEA mesh if we don't already have one
    if (nastranInstance[iIndex].haveFEAMesh == false) {
    	status = createMesh(iIndex, aimInfo);
    	if (status != CAPS_SUCCESS) return status;
    	nastranInstance[iIndex].haveFEAMesh = (int) true;
    }

    // Note: Setting order is important here.
    // 1. Materials should be set before properties.
    // 2. Coordinate system should be set before mesh and loads
	// 3. Mesh should be set before loads and constraints
	// 4. Constraints and loads should be set before analysis
    // 5. Optimization should be set after properties, but before analysis

    // Set material properties
    if (aimInputs[aim_getIndex(aimInfo, "Material", ANALYSISIN)-1].nullVal == NotNull) {
        status = fea_getMaterial(aimInputs[aim_getIndex(aimInfo, "Material", ANALYSISIN)-1].length,
                                 aimInputs[aim_getIndex(aimInfo, "Material", ANALYSISIN)-1].vals.tuple,
                                 &nastranInstance[iIndex].feaProblem.numMaterial,
                                 &nastranInstance[iIndex].feaProblem.feaMaterial);
        if (status != CAPS_SUCCESS) return status;
    } else printf("Load tuple is NULL - No materials set\n");

    // Set property properties
    if (aimInputs[aim_getIndex(aimInfo, "Property", ANALYSISIN)-1].nullVal == NotNull) {
        status = fea_getProperty(aimInputs[aim_getIndex(aimInfo, "Property", ANALYSISIN)-1].length,
                                 aimInputs[aim_getIndex(aimInfo, "Property", ANALYSISIN)-1].vals.tuple,
                                 &nastranInstance[iIndex].attrMap,
                                 &nastranInstance[iIndex].feaProblem);
        if (status != CAPS_SUCCESS) return status;
    } else printf("Property tuple is NULL - No properties set\n");

    // Set constraint properties
    if (aimInputs[aim_getIndex(aimInfo, "Constraint", ANALYSISIN)-1].nullVal == NotNull) {
        status = fea_getConstraint(aimInputs[aim_getIndex(aimInfo, "Constraint", ANALYSISIN)-1].length,
                                   aimInputs[aim_getIndex(aimInfo, "Constraint", ANALYSISIN)-1].vals.tuple,
                                   &nastranInstance[iIndex].attrMap,
                                   &nastranInstance[iIndex].feaProblem);
        if (status != CAPS_SUCCESS) return status;
    } else printf("Constraint tuple is NULL - No constraints applied\n");

    // Set load properties
    if (aimInputs[aim_getIndex(aimInfo, "Load", ANALYSISIN)-1].nullVal == NotNull) {
        status = fea_getLoad(aimInputs[aim_getIndex(aimInfo, "Load", ANALYSISIN)-1].length,
                             aimInputs[aim_getIndex(aimInfo, "Load", ANALYSISIN)-1].vals.tuple,
                             &nastranInstance[iIndex].attrMap,
                             &nastranInstance[iIndex].feaProblem);
        if (status != CAPS_SUCCESS) return status;

        // Loop through loads to see if any of them are supposed to be from an external source
		for (i = 0; i < nastranInstance[iIndex].feaProblem.numLoad; i++) {

			if (nastranInstance[iIndex].feaProblem.feaLoad[i].loadType == PressureExternal) {

				nastranInstance[iIndex].feaProblem.feaLoad[i].numElementID = 0;

				//See if we have data transfer information
				status = aim_getBounds(aimInfo, &numTransferName, &transferName);
				if (status != CAPS_SUCCESS) return status;

				numDataTransferElement = 0;
				elementIndex = 0;
				for (transferNameIndex = 0; transferNameIndex < numTransferName; transferNameIndex++) {

					status = aim_getDiscr(aimInfo, transferName[transferNameIndex], &dataTransferDiscreteObj);
					if (status != CAPS_SUCCESS) continue;

					status = aim_getDataSet(dataTransferDiscreteObj,
							                "Pressure",
											&dataTransferMethod,
											&numDataTransferPoint,
											&dataTransferRank,
											&dataTransferData);

					if (status == CAPS_SUCCESS) { // If we do have data ready, how many elements there are?

						numDataTransferElement += dataTransferDiscreteObj->nElems;

					} else {

						continue; // If no elements in this object skip to next transfer name
					}

					if (dataTransferRank != 1) {
						printf("Pressure transfer data found however rank is %d not 1!!!!\n", dataTransferRank);
						if (transferName != NULL) EG_free(transferName);
						return CAPS_BADRANK;
					}

					// If the first time we found elements allocate arrays
					if (nastranInstance[iIndex].feaProblem.feaLoad[i].numElementID == 0) {

						nastranInstance[iIndex].feaProblem.feaLoad[i].elementIDSet = (int *) EG_alloc(numDataTransferElement *sizeof(int));
						nastranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce = (double *) EG_alloc(4*numDataTransferElement *sizeof(double));

					} else { // Else reallocate arrays

						nastranInstance[iIndex].feaProblem.feaLoad[i].elementIDSet = (int *) EG_reall(nastranInstance[iIndex].feaProblem.feaLoad[i].elementIDSet,
																									  numDataTransferElement *sizeof(int));
						nastranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce = (double *) EG_reall(nastranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce,
																														4*numDataTransferElement *sizeof(double));
					}

					// Check to see if allocation failed
					if (nastranInstance[iIndex].feaProblem.feaLoad[i].elementIDSet == NULL ||
						nastranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce == NULL) {

						if (transferName != NULL) EG_free(transferName);
						return EGADS_MALLOC;
					}

					// Increment the number of element ID
					nastranInstance[iIndex].feaProblem.feaLoad[i].numElementID = numDataTransferElement;

					//Get extra node information stored in the discrObj
					storage = (int *) dataTransferDiscreteObj->ptrm;
					tessNodeMap = &storage[3*dataTransferDiscreteObj->nElems]; // Local indexing on the body

					//Now lets loop through our ctria3 mesh and get the node indexes for each element
					for (j = 0; j < nastranInstance[iIndex].feaProblem.feaMesh.numTria3; j++) {

						elementID = nastranInstance[iIndex].feaProblem.feaMesh.ctria3ElementID[j];

						// elementID is 1 bias
						nodeIndex[0] = nastranInstance[iIndex].feaProblem.feaMesh.ctria3[3*(elementID-1) +0];
						nodeIndex[1] = nastranInstance[iIndex].feaProblem.feaMesh.ctria3[3*(elementID-1) +1];
						nodeIndex[2] = nastranInstance[iIndex].feaProblem.feaMesh.ctria3[3*(elementID-1) +2];

						transferIndex[0] = -1;
						transferIndex[1] = -1;
						transferIndex[2] = -1;
						transferIndex[3] = -1;

						// Loop through the nodeMap of the data set getting nodeIDs trying
						//   to match the nodes in the element
						for (k = 0; k < numDataTransferPoint; k++) {

							globalNodeID = tessNodeMap[k];

							if (nodeIndex[0] == globalNodeID) transferIndex[0] = k;
							if (nodeIndex[1] == globalNodeID) transferIndex[1] = k;
							if (nodeIndex[2] == globalNodeID) transferIndex[2] = k;

							// If the nodes completely match the nodes on the element - break
							if (transferIndex[0] >= 0 && transferIndex[1] >= 0 && transferIndex[2] >= 0) {
								break;
							}
						}

						// If all the nodeIndexes match the transferIndex the element is in the data set
						//  so transfer the pressure forces
						if (transferIndex[0] >= 0 && transferIndex[1] >= 0 && transferIndex[2] >= 0) {

							nastranInstance[iIndex].feaProblem.feaLoad[i].elementIDSet[elementIndex] = elementID;

							nastranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce[4*elementIndex+0] = dataTransferData[transferIndex[0]];
							nastranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce[4*elementIndex+1] = dataTransferData[transferIndex[1]];
							nastranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce[4*elementIndex+2] = dataTransferData[transferIndex[2]];
							nastranInstance[iIndex].feaProblem.feaLoad[i].pressureMultiDistributeForce[4*elementIndex+3] = 0.0;

							elementIndex += 1;
						}
					}

					if (elementIndex != numDataTransferElement) {
						printf("\nElement transfer mismatch: number of elements found = %d, number"
								" of elements in transfer data set %d\n", elementIndex, numDataTransferElement);
						if (transferName != NULL) EG_free(transferName);
						return CAPS_MISMATCH;
					}
				} // End data transfer name loop

				if (transferName != NULL) EG_free(transferName);

			} // End PressureExternal if
		} // End load for loop
    } else printf("Load tuple is NULL - No loads applied\n");

    // Set design variables
    if (aimInputs[aim_getIndex(aimInfo, "Design_Variable", ANALYSISIN)-1].nullVal == NotNull) {
        status = fea_getDesignVariable(aimInputs[aim_getIndex(aimInfo, "Design_Variable", ANALYSISIN)-1].length,
                                       aimInputs[aim_getIndex(aimInfo, "Design_Variable", ANALYSISIN)-1].vals.tuple,
                                       &nastranInstance[iIndex].feaProblem);
        if (status != CAPS_SUCCESS) return status;
    } else printf("Design_Variable tuple is NULL - No design variables applied\n");

    // Set design constraints
    if (aimInputs[aim_getIndex(aimInfo, "Design_Constraint", ANALYSISIN)-1].nullVal == NotNull) {
        status = fea_getDesignConstraint(aimInputs[aim_getIndex(aimInfo, "Design_Constraint", ANALYSISIN)-1].length,
                                         aimInputs[aim_getIndex(aimInfo, "Design_Constraint", ANALYSISIN)-1].vals.tuple,
                                         &nastranInstance[iIndex].feaProblem);
        if (status != CAPS_SUCCESS) return status;
    } else printf("Design_Constraint tuple is NULL - No design constraints applied\n");

    // Set analysis settings
    if (aimInputs[aim_getIndex(aimInfo, "Analysis", ANALYSISIN)-1].nullVal == NotNull) {
        status = fea_getAnalysis(aimInputs[aim_getIndex(aimInfo, "Analysis", ANALYSISIN)-1].length,
                                 aimInputs[aim_getIndex(aimInfo, "Analysis", ANALYSISIN)-1].vals.tuple,
                                 &nastranInstance[iIndex].feaProblem);
        if (status != CAPS_SUCCESS) return status; // It ok to not have an analysis tuple
    } else printf("Analysis tuple is NULL\n");

    // Set file format type
    if        (strcasecmp(aimInputs[aim_getIndex(aimInfo, "File_Format", ANALYSISIN)-1].vals.string, "Small") == 0) {
    	nastranInstance[iIndex].feaProblem.feaFileFormat.fileType = SmallField;
    } else if (strcasecmp(aimInputs[aim_getIndex(aimInfo, "File_Format", ANALYSISIN)-1].vals.string, "Large") == 0) {
    	nastranInstance[iIndex].feaProblem.feaFileFormat.fileType = LargeField;
    } else if (strcasecmp(aimInputs[aim_getIndex(aimInfo, "File_Format", ANALYSISIN)-1].vals.string, "Free") == 0)  {
    	nastranInstance[iIndex].feaProblem.feaFileFormat.fileType = FreeField;
    } else {
    	printf("Unrecognized \"File_Format\", valid choices are [Small, Large, or Free]. Reverting to default\n");
    }

    // Set grid file format type
    if        (strcasecmp(aimInputs[aim_getIndex(aimInfo, "Mesh_File_Format", ANALYSISIN)-1].vals.string, "Small") == 0) {
    	nastranInstance[iIndex].feaProblem.feaFileFormat.gridFileType = SmallField;
    } else if (strcasecmp(aimInputs[aim_getIndex(aimInfo, "Mesh_File_Format", ANALYSISIN)-1].vals.string, "Large") == 0) {
    	nastranInstance[iIndex].feaProblem.feaFileFormat.gridFileType = LargeField;
    } else if (strcasecmp(aimInputs[aim_getIndex(aimInfo, "Mesh_File_Format", ANALYSISIN)-1].vals.string, "Free") == 0)  {
    	nastranInstance[iIndex].feaProblem.feaFileFormat.gridFileType = FreeField;
    } else {
    	printf("Unrecognized \"Mesh_File_Format\", valid choices are [Small, Large, or Free]. Reverting to default\n");
    }

    // Write Nastran Mesh
    filename = EG_alloc(MXCHAR +1);
    if (filename == NULL) return EGADS_MALLOC;
    strcpy(filename, analysisPath);
    strcat(filename, "/");
    strcat(filename, nastranInstance[iIndex].projectName);

    status = write_NASTRAN(filename,
    					   1,
						   &nastranInstance[iIndex].feaProblem.feaMesh,
						   nastranInstance[iIndex].feaProblem.feaFileFormat.gridFileType);
    EG_free(filename);
    if (status != CAPS_SUCCESS) return status;


    // Write nastran input file
    filename = EG_alloc(MXCHAR +1);
    if (filename == NULL) return EGADS_MALLOC;
    strcpy(filename, analysisPath);
    strcat(filename, "/");
    strcat(filename, nastranInstance[iIndex].projectName);
    strcat(filename, ".dat");


    printf("\nWriting nastran instruction file....\n");
    fp = fopen(filename, "w");
    EG_free(filename);
    if (fp == NULL) {
    	printf("Unable to open file: %s\n", filename);
    	return CAPS_IOERR;
    }

    // define file format delimiter type
	if (nastranInstance[iIndex].feaProblem.feaFileFormat.fileType == FreeField) {
		delimiter = ",";
	} else {
		delimiter = " ";
	}

    //////////////// Executive control ////////////////
    fprintf(fp, "ID CAPS generated Problem FOR nastran\n");

    // Analysis type
    analysisType = aimInputs[aim_getIndex(aimInfo, "Analysis_Type", ANALYSISIN)-1].vals.string;

    if     (strcasecmp(analysisType, "Modal")         == 0) fprintf(fp, "SOL 3\n");
    else if(strcasecmp(analysisType, "Static")        == 0) fprintf(fp, "SOL 1\n");
    else if(strcasecmp(analysisType, "Craig-Bampton") == 0) fprintf(fp, "SOL 31\n");
    else if(strcasecmp(analysisType, "StaticOpt")     == 0) fprintf(fp, "SOL 200\n");
    else {
        printf("Unrecognized \"Analysis_Type\", %s, defaulting to \"Modal\" analysis\n", analysisType);
        analysisType = "Modal";
        fprintf(fp, "SOL 3\n");
    }

    fprintf(fp, "CEND\n\n");
    fprintf(fp, "LINE=10000\n");


    //////////////// Case control ////////////////

    // Write output request information
    fprintf(fp, "DISP = ALL\n"); // Output all displacements

    fprintf(fp, "STRE = ALL\n"); // Output all stress

    fprintf(fp, "STRA = ALL\n"); // Output all strain

    fprintf(fp, "PARAM, POST, -1\n"); // Output OP2 file

    // Write constraint information
    if (nastranInstance[iIndex].feaProblem.numConstraint != 0) {
    	fprintf(fp, "SPC = %d\n", nastranInstance[iIndex].feaProblem.numConstraint+1);
    } else {
    	printf("Warning: No constraints specified for job!!!!\n");
    }

    // Design objective information, SOL 200 only
    if (strcasecmp(analysisType, "StaticOpt") == 0) {

    	objectiveMinMax = aimInputs[aim_getIndex(aimInfo, "ObjectiveMinMax", ANALYSISIN)-1].vals.string;
		if     (strcasecmp(objectiveMinMax, "Min") == 0) fprintf(fp, "DESOBJ(MIN) = 1\n");
		else if(strcasecmp(objectiveMinMax, "Max") == 0) fprintf(fp, "DESOBJ(MAX) = 1\n");
		else {
			printf("Unrecognized \"ObjectiveMinMax\", %s, defaulting to \"Min\"\n", objectiveMinMax);
			objectiveMinMax = "Min";
			fprintf(fp, "DESOBJ(MIN) = 1\n");
		}

    }

    // Modal analysis - only
    // If modal - we are only going to use the first analysis structure we come across that has its type as modal
    if (strcasecmp(analysisType, "Modal") == 0) {

    	// Look through analysis structures for a modal one
    	found = (int) false;
        for (i = 0; i < nastranInstance[iIndex].feaProblem.numAnalysis; i++) {
            if (nastranInstance[iIndex].feaProblem.feaAnalysis[i].analysisType == Modal) {
            	found = (int) true;
            	break;
            }
        }

        // Write out analysis ID if a modal analysis structure was found
        if (found == (int) true)  {
            fprintf(fp, "METHOD = %d\n", nastranInstance[iIndex].feaProblem.feaAnalysis[i].analysisID);
        } else {
            printf("Warning: No eigenvalue analysis information specified in \"Analysis\" tuple, through "
            		"AIM input \"Analysis_Type\" is set to \"Modal\"!!!!\n");
            return CAPS_NOTFOUND;
        }

    }
    // Static analysis and/or Static w/ optimization- only
    if (strcasecmp(analysisType, "Static") == 0 ||
        strcasecmp(analysisType, "StaticOpt") == 0) {

    	// If we have multiple analysis structures
    	if (nastranInstance[iIndex].feaProblem.numAnalysis != 0) {

			// Write subcase information if multiple analysis tuples were provide
			for (i = 0; i < nastranInstance[iIndex].feaProblem.numAnalysis; i++) {

				if (nastranInstance[iIndex].feaProblem.feaAnalysis[i].analysisType == Static ||
				    nastranInstance[iIndex].feaProblem.feaAnalysis[i].analysisType == StaticOpt) {

				    fprintf(fp, "SUBCASE %d\n", i);
				    fprintf(fp, "\tLABEL %s\n", nastranInstance[iIndex].feaProblem.feaAnalysis[i].name);

				    if (nastranInstance[iIndex].feaProblem.feaAnalysis[i].analysisType == StaticOpt) {
				    	fprintf(fp,"\tANALYSIS = STATICS\n");
				    }

					// Write loads for sub-case
					if (nastranInstance[iIndex].feaProblem.numLoad != 0) {
						fprintf(fp, "\tLOAD = %d\n", nastranInstance[iIndex].feaProblem.numLoad+i+1);
					}

					// Issue some warnings regarding loads if necessary
					if (nastranInstance[iIndex].feaProblem.feaAnalysis[i].numLoad == 0 && nastranInstance[iIndex].feaProblem.numLoad !=0) {
						printf("Warning: No loads specified for static case %s, assuming "
								"all loads are applied!!!!\n", nastranInstance[iIndex].feaProblem.feaAnalysis[i].name);
					} else if (nastranInstance[iIndex].feaProblem.numLoad == 0) {
						printf("Warning: No loads specified for static case %s!!!!\n", nastranInstance[iIndex].feaProblem.feaAnalysis[i].name);
					}
				}

				if (nastranInstance[iIndex].feaProblem.feaAnalysis[i].analysisType == StaticOpt) {
				    // Write objective function
				    //fprintf(fp, "\tDESOBJ(%s) = %d\n", nastranInstance[iIndex].feaProblem.feaAnalysis[i].objectiveMinMax,
				    //                                   nastranInstance[iIndex].feaProblem.feaAnalysis[i].analysisID);
				    // Write optimization constraints
					if (nastranInstance[iIndex].feaProblem.numDesignConstraint != 0) {
					    fprintf(fp, "\tDESSUB = %d\n", nastranInstance[iIndex].feaProblem.numDesignConstraint+i+1);
					}
				}
			}

    	} else { // // If no sub-cases

    	    // Write load card
    		if (nastranInstance[iIndex].feaProblem.numLoad != 0) {
    			fprintf(fp, "LOAD = %d\n", nastranInstance[iIndex].feaProblem.numLoad+1);
    		} else {
    			printf("Warning: No loads specified for static a job!!!!\n");
    		}

    		// What about an objective function if no analysis tuple? Do we need to add a new capsValue?

    		// Write design constraints
    		if (nastranInstance[iIndex].feaProblem.numDesignConstraint != 0) {
    		    fprintf(fp, "\tDESSUB = %d\n", nastranInstance[iIndex].feaProblem.numDesignConstraint+1);
    		}
    	}
    }

    //////////////// Bulk data ////////////////
    fprintf(fp, "\nBEGIN BULK\n");
    fprintf(fp, "$---1---|---2---|---3---|---4---|---5---|---6---|---7---|---8---|---9---|---10--|\n");

    // Turn off auto SPC
    //fprintf(fp, "%-8s %7s %7s\n", "PARAM", "AUTOSPC", "N");

    // Optimization Objective Response Response, SOL 200 only
    if (strcasecmp(analysisType, "StaticOpt") == 0) {

		objectiveResp = aimInputs[aim_getIndex(aimInfo, "ObjectiveResponseType", ANALYSISIN)-1].vals.string;
		if     (strcasecmp(objectiveResp, "Weight") == 0) objectiveResp = "WEIGHT";
		else {
			printf("Unrecognized \"ObjectiveResponseType\", %s, defaulting to \"Weight\"\n", objectiveResp);
			objectiveResp = "WEIGHT";
		}

		fprintf(fp,"%-8s", "DRESP1");

		tempString = convert_integerToString(1, 7, 1);
		fprintf(fp, "%s%s", delimiter, tempString);
		EG_free(tempString);

		fprintf(fp, "%s%7s", delimiter, objectiveResp);
		fprintf(fp, "%s%7s", delimiter, objectiveResp);


		fprintf(fp, "\n");


	}

    // Analysis Cards - Eigenvalue and design objective included, as well as combined load and combined design constraints
    printf("\tWriting analysis cards\n");
    if (nastranInstance[iIndex].feaProblem.numAnalysis != 0) {

    	for (i = 0; i < nastranInstance[iIndex].feaProblem.numAnalysis; i++) {
			status = nastran_writeAnalysisCard(fp,
											   &nastranInstance[iIndex].feaProblem.feaAnalysis[i],
											   &nastranInstance[iIndex].feaProblem.feaFileFormat);
			if (status != CAPS_SUCCESS) return status;


			if (nastranInstance[iIndex].feaProblem.feaAnalysis[i].numLoad != 0) {
				status = nastran_writeLoadADDCard(fp,
												  nastranInstance[iIndex].feaProblem.numLoad+i+1,
												  nastranInstance[iIndex].feaProblem.feaAnalysis[i].numLoad,
												  nastranInstance[iIndex].feaProblem.feaAnalysis[i].loadSetID,
												  nastranInstance[iIndex].feaProblem.feaLoad,
												  &nastranInstance[iIndex].feaProblem.feaFileFormat);
				if (status != CAPS_SUCCESS) return status;

			} else { // If no loads for an individual analysis are specified assume that all loads should be applied

				if (nastranInstance[iIndex].feaProblem.numLoad != 0) {

					// Create a temporary list of load IDs
					tempIntegerArray = (int *) EG_alloc(nastranInstance[iIndex].feaProblem.numLoad*sizeof(int));
					if (tempIntegerArray == NULL) return EGADS_MALLOC;

					for (j = 0; j < nastranInstance[iIndex].feaProblem.numLoad; j++) {
						tempIntegerArray[j] = nastranInstance[iIndex].feaProblem.feaLoad[j].loadID;
					}

					// Write combined load card
					status = nastran_writeLoadADDCard(fp,
													  nastranInstance[iIndex].feaProblem.numLoad+i+1,
													  nastranInstance[iIndex].feaProblem.numLoad,
													  tempIntegerArray,
													  nastranInstance[iIndex].feaProblem.feaLoad,
													  &nastranInstance[iIndex].feaProblem.feaFileFormat);

					if (status != CAPS_SUCCESS) return status;

					// Free temporary load ID list
					if (tempIntegerArray != NULL) EG_free(tempIntegerArray);
					tempIntegerArray = NULL;
				}

			}

            if (nastranInstance[iIndex].feaProblem.feaAnalysis[i].numDesignConstraint != 0) {
            	status = nastran_writeDesignConstraintADDCard(fp,
                                                              nastranInstance[iIndex].feaProblem.numDesignConstraint+i+1,
                                                              nastranInstance[iIndex].feaProblem.feaAnalysis[i].numDesignConstraint,
                                                              nastranInstance[iIndex].feaProblem.feaAnalysis[i].designConstraintSetID,
                                                              &nastranInstance[iIndex].feaProblem.feaFileFormat);
                if (status != CAPS_SUCCESS) return status;

            } else { // If no design constraints for an individual analysis are specified assume that all design constraints should be applied

                if (nastranInstance[iIndex].feaProblem.numDesignConstraint != 0) {

                    // Create a temporary list of design constraint IDs
                    tempIntegerArray = (int *) EG_alloc(nastranInstance[iIndex].feaProblem.numDesignConstraint*sizeof(int));
                    if (tempIntegerArray == NULL) return EGADS_MALLOC;

                    for (j = 0; j < nastranInstance[iIndex].feaProblem.numDesignConstraint; j++) {
                        tempIntegerArray[j] = nastranInstance[iIndex].feaProblem.feaDesignConstraint[j].designConstraintID;
                    }

                    // Write combined design constraint card
                    status = nastran_writeDesignConstraintADDCard(fp,
                                                                  nastranInstance[iIndex].feaProblem.numDesignConstraint+i+1,
                                                                  nastranInstance[iIndex].feaProblem.feaAnalysis[i].numDesignConstraint,
                                                                  tempIntegerArray,
                                                                  &nastranInstance[iIndex].feaProblem.feaFileFormat);
                    if (status != CAPS_SUCCESS) return status;

                    // Free temporary design constraint ID list
                    if (tempIntegerArray != NULL) EG_free(tempIntegerArray);
                    tempIntegerArray = NULL;
                }

            }
		}

    } else { // If there aren't any analysis structures just write a single combined load and design constraint card

    	// Combined loads
    	if (nastranInstance[iIndex].feaProblem.numLoad != 0) {

			// Create a temporary list of load IDs
			tempIntegerArray = (int *) EG_alloc(nastranInstance[iIndex].feaProblem.numLoad*sizeof(int));
			if (tempIntegerArray == NULL) return EGADS_MALLOC;

			for (i = 0; i < nastranInstance[iIndex].feaProblem.numLoad; i++) {
				tempIntegerArray[i] = nastranInstance[iIndex].feaProblem.feaLoad[i].loadID;
			}

			// Write combined load card
			status = nastran_writeLoadADDCard(fp,
											  nastranInstance[iIndex].feaProblem.numLoad+1,
											  nastranInstance[iIndex].feaProblem.numLoad,
											  tempIntegerArray,
											  nastranInstance[iIndex].feaProblem.feaLoad,
											  &nastranInstance[iIndex].feaProblem.feaFileFormat);

			if (status != CAPS_SUCCESS) return status;

			// Free temporary load ID list
			if (tempIntegerArray != NULL) EG_free(tempIntegerArray);
			tempIntegerArray = NULL;
    	}

        // Combined design constraints
    	printf("\tWriting design constraint cards\n");
        if (nastranInstance[iIndex].feaProblem.numDesignConstraint != 0) {

            // Create a temporary list of design constraint IDs
            tempIntegerArray = (int *) EG_alloc(nastranInstance[iIndex].feaProblem.numDesignConstraint*sizeof(int));
            if (tempIntegerArray == NULL) return EGADS_MALLOC;

            for (i = 0; i < nastranInstance[iIndex].feaProblem.numDesignConstraint; i++) {
                tempIntegerArray[i] = nastranInstance[iIndex].feaProblem.feaDesignConstraint[i].designConstraintID;
            }

            // Write combined design constraint card
            status = nastran_writeDesignConstraintADDCard(fp,
                                                          nastranInstance[iIndex].feaProblem.numDesignConstraint+1,
                                                          nastranInstance[iIndex].feaProblem.numDesignConstraint,
                                                          tempIntegerArray,
                                                          &nastranInstance[iIndex].feaProblem.feaFileFormat);

            if (status != CAPS_SUCCESS) return status;

            // Free temporary design constraint ID list
            if (tempIntegerArray != NULL) EG_free(tempIntegerArray);
            tempIntegerArray = NULL;
        }
    }

    // Combined constraints
    printf("\tWriting combined constraint cards\n");
    if (nastranInstance[iIndex].feaProblem.numConstraint != 0) {

    	// Create a temporary list of constraint IDs
    	tempIntegerArray = (int *) EG_alloc(nastranInstance[iIndex].feaProblem.numConstraint*sizeof(int));
    	if (tempIntegerArray == NULL) return EGADS_MALLOC;

    	for (i = 0; i < nastranInstance[iIndex].feaProblem.numConstraint; i++) {
    		tempIntegerArray[i] = nastranInstance[iIndex].feaProblem.feaConstraint[i].constraintID;
    	}

    	// Write combined constraint card
    	status = nastran_writeConstraintADDCard(fp,
    											nastranInstance[iIndex].feaProblem.numConstraint+1,
												nastranInstance[iIndex].feaProblem.numConstraint,
												tempIntegerArray,
												&nastranInstance[iIndex].feaProblem.feaFileFormat);
    	if (status != CAPS_SUCCESS) return status;

    	// Free temporary constraint ID list
    	if (tempIntegerArray != NULL) EG_free(tempIntegerArray);
    	tempIntegerArray = NULL;
    }

    // Loads
    printf("\tWriting load cards\n");
    for (i = 0; i < nastranInstance[iIndex].feaProblem.numLoad; i++) {
    	status = nastran_writeLoadCard(fp,
    	                               &nastranInstance[iIndex].feaProblem.feaLoad[i],
    	                               &nastranInstance[iIndex].feaProblem.feaFileFormat);
    	if (status != CAPS_SUCCESS) return status;
    }

    // Constraints
    printf("\tWriting constraint cards\n");
    for (i = 0; i < nastranInstance[iIndex].feaProblem.numConstraint; i++) {
    	status = nastran_writeConstraintCard(fp,
    	                                     &nastranInstance[iIndex].feaProblem.feaConstraint[i],
    	                                     &nastranInstance[iIndex].feaProblem.feaFileFormat);
    	if (status != CAPS_SUCCESS) return status;
    }

  	// Materials
    printf("\tWriting material cards\n");
  	for (i = 0; i < nastranInstance[iIndex].feaProblem.numMaterial; i++) {
  	    status = nastran_writeMaterialCard(fp,
  	                                       &nastranInstance[iIndex].feaProblem.feaMaterial[i],
  	                                       &nastranInstance[iIndex].feaProblem.feaFileFormat);
  	    if (status != CAPS_SUCCESS) return status;
  	}

  	// Properties
  	printf("\tWriting property cards\n");
  	for (i = 0; i < nastranInstance[iIndex].feaProblem.numProperty; i++) {
  	    status = nastran_writePropertyCard(fp,
  	                                       &nastranInstance[iIndex].feaProblem.feaProperty[i],
  	                                       &nastranInstance[iIndex].feaProblem.feaFileFormat);
  	    if (status != CAPS_SUCCESS) return status;
  	}

  	// Coordinate systems
  	printf("\tWriting coordinate system cards\n");
	for (i = 0; i < nastranInstance[iIndex].feaProblem.numCoordSystem; i++) {
		status = nastran_writeCoordinateSystemCard(fp, &nastranInstance[iIndex].feaProblem.feaCoordSystem[i], &nastranInstance[iIndex].feaProblem.feaFileFormat);
		if (status != CAPS_SUCCESS) return status;
	}

  	// Optimization - design variables
  	printf("\tWriting design variables and analysis - design variable relation cards\n");
  	for( i = 0; i < nastranInstance[iIndex].feaProblem.numDesignVariable; i++) {
  	  status = nastran_writeDesignVariableCard(fp,
  	                                           &nastranInstance[iIndex].feaProblem.feaDesignVariable[i],
  	                                           &nastranInstance[iIndex].feaProblem.feaFileFormat);
  	  if (status != CAPS_SUCCESS) return status;
  	}

  	// Optimization - design constraints
  	printf("\tWriting design constraints and responses cards\n");
  	for( i = 0; i < nastranInstance[iIndex].feaProblem.numDesignConstraint; i++) {
  	    status = nastran_writeDesignConstraintCard(fp,
  	                                               &nastranInstance[iIndex].feaProblem.feaDesignConstraint[i],
  	                                               &nastranInstance[iIndex].feaProblem.feaFileFormat);
  	    if (status != CAPS_SUCCESS) return status;
  	}

  	// Include mesh file
  	fprintf(fp,"\nINCLUDE \'%s.bdf\'\n\n", nastranInstance[iIndex].projectName);

  	// End bulk data
    fprintf(fp,"ENDDATA\n");

    fclose(fp);
/*
////////////////////////////////////////
    printf("\n\n\nTESTING F06 READER\n\n");

    // FO6 data variables
    int numGridPoint = 0;
    int numEigenVector = 0;
    double **dataMatrix = NULL;
    filename = (char *) EG_alloc((strlen(nastranInstance[iIndex].analysisPath) + strlen(nastranInstance[iIndex].projectName) +
                                      strlen(".f06") + 2)*sizeof(char));

    sprintf(filename,"%s/%s%s", nastranInstance[iIndex].analysisPath,
    						    nastranInstance[iIndex].projectName, ".f06");

    // Open file
    fp = fopen(filename, "r");
    if (filename != NULL) EG_free(filename);
    filename = NULL;

    if (fp == NULL) {
    	printf("Unable to open file: %s\n", filename);

    	return CAPS_IOERR;
    }

    status = nastran_readF06Displacement(fp,
                                         -9,
                                         &numGridPoint,
                                         &dataMatrix);

    for ( i = 0 ; i < numGridPoint ; i++) printf("dataMatrix  = %f %f %f %f %f %f %f %f\n",
    																	 dataMatrix[i][0],
    																	 dataMatrix[i][1],
																		 dataMatrix[i][2],
																		 dataMatrix[i][3],
																		 dataMatrix[i][4],
																		 dataMatrix[i][5],
																		 dataMatrix[i][6],
																		 dataMatrix[i][7]);
    fclose(fp);
///////////////////////////////////
*/

    return CAPS_SUCCESS;
}

// Set Nastran output variables
int aimOutputs(int iIndex, int index, char **aoname, capsValue *form)
{

	/*! \page aimOutputsNastran
	 * None currently.
	 */

	#ifdef DEBUG
        printf(" nastranAIM/aimOutputs instance = %d  index = %d!\n", iIndex, index);
    #endif

    return CAPS_SUCCESS;
}

// Calculate Nastran output
int aimCalcOutput(int iIndex, void *aimInfo, char *analysisPath,
              	  int index, capsValue *val, capsErrs **errors)
{

	return CAPS_SUCCESS;
}


int aimFreeDiscr(capsDiscr *discr)
{
	int iIndex, i; // Indexing

	int status; // Returning status

	#ifdef DEBUG
    	if (discr == NULL) {
    		printf(" nastranAIM/aimFreeDiscr w/ NULL!\n");
    	} else {
    		printf(" nastranAIM/aimFreeDiscr instance = %d!\n", discr->instance);
    	}
	#endif

    if (discr == NULL) {

        if (numInstance != 0) {

        	// Clean up nastranInstance data
        	for ( iIndex = 0; iIndex < numInstance; iIndex++) {

        		printf(" Cleaning up nastranInstance - %d\n", iIndex);


        		// Attribute to index map
        		status = destroy_mapAttrToIndexStruct(&nastranInstance[iIndex].attrMap);
        		if (status != CAPS_SUCCESS) return status;

        		// Destroy FEA problem structure
        		status = destroy_feaProblemStruct(&nastranInstance[iIndex].feaProblem);
        		if (status != CAPS_SUCCESS) return status;

        		// NULL projetName
        		nastranInstance[iIndex].projectName = NULL;

        		// NULL analysisPath
        		nastranInstance[iIndex].analysisPath = NULL;

        		// NULL pointer to caps input value for tessellation parameters
        		nastranInstance[iIndex].tessParam = NULL;

        		// NULL pointer to caps input value for Edge_Point- Min
        		nastranInstance[iIndex].edgePointMin = NULL;

        		// NULL pointer to caps input value for Edge_Point- Max
        		nastranInstance[iIndex].edgePointMax = NULL;

        		// NULL pointer to caps input value for No_Quad_Faces
        		nastranInstance[iIndex].noQuadFace = NULL;
        	}

          }
          numInstance = 0;

    } else {

    	// Free up this capsDiscr

		discr->nPoints = 0; // Points

		if (discr->mapping != NULL) EG_free(discr->mapping);
		discr->mapping = NULL;

		if (discr->types != NULL) { // Element types
			for (i = 0; i < discr->nTypes; i++) {
				if (discr->types[i].gst  != NULL) EG_free(discr->types[i].gst);
				if (discr->types[i].tris != NULL) EG_free(discr->types[i].tris);
			}

			EG_free(discr->types);
		}

		discr->nTypes  = 0;
		discr->types   = NULL;

		if (discr->elems != NULL) { // Element connectivity

			for (i = 0; i < discr->nElems; i++) {
				if (discr->elems[i].gIndices != NULL) EG_free(discr->elems[i].gIndices);
			}

			EG_free(discr->elems);
		}

		discr->nElems  = 0;
		discr->elems   = NULL;

		if (discr->ptrm != NULL) EG_free(discr->ptrm); // Extra information to store into the discr void pointer
		discr->ptrm    = NULL;


		discr->nVerts = 0;    // Not currently used
		discr->verts  = NULL; // Not currently used
		discr->celem = NULL; // Not currently used

		discr->nDtris = 0; // Not currently used
		discr->dtris  = NULL; // Not currently used

    }

    return CAPS_SUCCESS;
}

int aimDiscr(char *tname, capsDiscr *discr) {

	int i, body, face; // Indexing

	int status; // Function return status

	int iIndex; // Instance index

	int numBody;

	// EGADS objects
	ego tess, *bodies = NULL, *faces = NULL, tempBody;

	const char   *string, *capsGroup; // capsGroups strings

	// EGADS function returns
	int plen, tlen, elen;
	const int    *ptype, *pindex, *tris, *nei;
	const double *xyz, *uv;

	// Body Tessellation
	int numFace = 0;
	int numFaceFound = 0;
	int numPoint = 0, numTri = 0, numGlobalPoint = 0;
	int *bodyFaceMap = NULL; // size=[2*numFaceFound], [2*numFaceFound + 0] = body, [2*numFaceFoun + 1] = face

	int *globalID = NULL, *localStitchedID = NULL, gID = 0;

	int *storage= NULL; // Extra information to store into the discr void pointer

	int numCAPSGroup = 0, attrIndex = 0, foundAttr = (int) false;
	int *capsGroupList = NULL;
	int dataTransferBodyIndex=-99;

	iIndex = discr->instance;
	#ifdef DEBUG
		printf(" nastranAIM/aimDiscr: tname = %s, instance = %d!\n", tname, iIndex);
	#endif

	if ((iIndex < 0) || (iIndex >= numInstance)) return CAPS_BADINDEX;


	if (tname == NULL) return CAPS_NOTFOUND;

	/*if (nastranInstance[iIndex].dataTransferCheck == (int) false) {
	    printf("The volume is not suitable for data transfer - possibly the volume mesher "
	            "added unaccounted for points\n");
	    return CAPS_BADVALUE;
	}*/

	// Currently this ONLY works if the capsTranfer lives on single body!
	status = aim_getBodies(discr->aInfo, &numBody, &bodies);
	if (status != CAPS_SUCCESS) {
		printf(" nastranAIM/aimDiscr: %d aim_getBodies = %d!\n", iIndex, status);
		return status;
	}

	status = aimFreeDiscr(discr);
	if (status != CAPS_SUCCESS) return status;

	numFaceFound = 0;
	numPoint = numTri = 0;
	// Find any faces with our boundary marker and get how many points and triangles there are
	for (body = 0; body < numBody; body++) {

		status = EG_getBodyTopos(bodies[body], NULL, FACE, &numFace, &faces);
		if (status != EGADS_SUCCESS) {
			printf(" nastranAIM: getBodyTopos (Face) = %d for Body %d!\n", status, body);
			return status;
		}

		for (face = 0; face < numFace; face++) {

			// Retrieve the string following a capsTransfer tag
			status = retrieve_CAPSTransferAttr(faces[face], &string);
			if (status != CAPS_SUCCESS) continue;

			if (strcmp(string, tname) != 0) continue;

			#ifdef DEBUG
				printf(" nastranAIM/aimDiscr: Body %d/Face %d matches %s!\n", body, face+1, tname);
			#endif


			status = retrieve_CAPSGroupAttr(faces[face], &capsGroup);
			if (status != CAPS_SUCCESS) {
			    printf("capsTransfer found on face %d, but no capGroup found!!!\n", face);
			    continue;
			} else {

			    status = get_mapAttrToIndexIndex(&nastranInstance[iIndex].attrMap, capsGroup, &attrIndex);
			    if (status != CAPS_SUCCESS) {
			        printf("capsGroup %s NOT found in attrMap\n",capsGroup);
			        continue;
			    } else {

			        // If first index create arrays and store index
			        if (numCAPSGroup == 0) {
			            numCAPSGroup += 1;
			            capsGroupList = (int *) EG_alloc(numCAPSGroup*sizeof(int));
			            if (capsGroupList == NULL) {
			                status =  EGADS_MALLOC;
			                goto cleanup;
			            }

			            capsGroupList[numCAPSGroup-1] = attrIndex;
			        } else { // If we already have an index(es) let make sure it is unique
			            foundAttr = (int) false;
			            for (i = 0; i < numCAPSGroup; i++) {
			                if (attrIndex == capsGroupList[i]) {
			                    foundAttr = (int) true;
			                    break;
			                }
			            }

			            if (foundAttr == (int) false) {
			                numCAPSGroup += 1;
			                capsGroupList = (int *) EG_reall(capsGroupList, numCAPSGroup*sizeof(int));
			                if (capsGroupList == NULL) {
			                    status =  EGADS_MALLOC;
			                    goto cleanup;
			                }

			                capsGroupList[numCAPSGroup-1] = attrIndex;
			            }
			        }
			    }
			}

			// Get face tessellation
			status = EG_getTessFace(bodies[body+numBody], face+1, &plen, &xyz, &uv, &ptype, &pindex, &tlen, &tris, &nei);
			if (status != EGADS_SUCCESS) {
				printf(" nastranAIM: EG_getTessFace %d = %d for Body %d!\n", face+1, status, body+1);
				continue;
			}

			numFaceFound += 1;
			dataTransferBodyIndex = body;
			if (numFaceFound == 1) {
				bodyFaceMap = (int *) EG_alloc(2*numFaceFound*sizeof(int));
			} else {
				bodyFaceMap = (int *) EG_reall(bodyFaceMap, 2*numFaceFound*sizeof(int));
			}

			if (bodyFaceMap == NULL) {
				status = EGADS_MALLOC;
				goto cleanup;
			}

			// Get number of points and triangles
			bodyFaceMap[2*(numFaceFound-1) + 0] = body+1;
			bodyFaceMap[2*(numFaceFound-1) + 1] = face+1;

			// Sum number of points and triangles
			numPoint  += plen;
			numTri  += tlen;
		}

		if (faces != NULL) EG_free(faces);
		faces = NULL;


	    if (dataTransferBodyIndex >=0) break; // Force that only one body can be used
	}

	if (numFaceFound == 0) {
		printf(" nastranAIM/aimDiscr: No Faces match %s!\n", tname);

		status = CAPS_NOTFOUND;
		goto cleanup;
	}

	if ( numPoint == 0 || numTri == 0) {
		#ifdef DEBUG
	    	printf(" nastranAIM/aimDiscr: ntris = %d, npts = %d!\n", numTri, numPoint);
		#endif
	    status = CAPS_SOURCEERR;
	    goto cleanup;
	}

	#ifdef DEBUG
		printf(" nastranAIM/aimDiscr: Instance %d, Body Index for data transfer = %d\n", iIndex, dataTransferBodyIndex);
	#endif


	// Specify our single element type
	status = EGADS_MALLOC;
	discr->nTypes = 1;

	discr->types  = (capsEleType *) EG_alloc(sizeof(capsEleType));
	if (discr->types == NULL) goto cleanup;
	discr->types[0].nref  = 3;
	discr->types[0].ndata = 0;            /* data at geom reference positions */
	discr->types[0].ntri  = 1;
	discr->types[0].nmat  = 0;            /* match points at geom ref positions */
	discr->types[0].tris  = NULL;
	discr->types[0].gst   = NULL;
	discr->types[0].dst   = NULL;
	discr->types[0].matst = NULL;

	discr->types[0].tris   = (int *) EG_alloc(3*sizeof(int));
	if (discr->types[0].tris == NULL) goto cleanup;
	discr->types[0].tris[0] = 1;
	discr->types[0].tris[1] = 2;
	discr->types[0].tris[2] = 3;

	discr->types[0].gst   = (double *) EG_alloc(6*sizeof(double));
	if (discr->types[0].gst == NULL) goto cleanup;
	discr->types[0].gst[0] = 0.0;
	discr->types[0].gst[1] = 0.0;
	discr->types[0].gst[2] = 1.0;
	discr->types[0].gst[3] = 0.0;
	discr->types[0].gst[4] = 0.0;
	discr->types[0].gst[5] = 1.0;

	// Get the tessellation and make up a simple linear continuous triangle discretization */

	discr->nElems = numTri;

	discr->elems = (capsElement *) EG_alloc(discr->nElems*sizeof(capsElement));
	if (discr->elems == NULL) {
		status = EGADS_MALLOC;
		goto cleanup;
	}

	discr->mapping = (int *) EG_alloc(2*numPoint*sizeof(int)); // Will be resized
	if (discr->mapping == NULL) goto cleanup;

	globalID = (int *) EG_alloc(numPoint*sizeof(int));
	if (globalID == NULL) {
		status = EGADS_MALLOC;
		goto cleanup;
	}

	numPoint = 0;
	numTri = 0;
	for (face = 0; face < numFaceFound; face++){

		tess = bodies[bodyFaceMap[2*face + 0]-1 + numBody];

		if (localStitchedID == NULL) {
			status = EG_statusTessBody(tess, &tempBody, &i, &numGlobalPoint);

			localStitchedID = (int *) EG_alloc(numGlobalPoint*sizeof(int));
			if (localStitchedID == NULL) {
				status = EGADS_MALLOC;
				goto cleanup;
			}

			for (i = 0; i < numGlobalPoint; i++) localStitchedID[i] = 0;
		}

		// Get face tessellation
		status = EG_getTessFace(tess, bodyFaceMap[2*face + 1], &plen, &xyz, &uv, &ptype, &pindex, &tlen, &tris, &nei);
		if (status != EGADS_SUCCESS) {
			printf(" nastranAIM: EG_getTessFace %d = %d for Body %d!\n", bodyFaceMap[2*face + 1], status, bodyFaceMap[2*face + 0]);
			continue;
		}

		for (i = 0; i < plen; i++ ) {

			status = EG_localToGlobal(tess, bodyFaceMap[2*face+1], i+1, &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			if (localStitchedID[gID -1] != 0) continue;

			discr->mapping[2*numPoint  ] = bodyFaceMap[2*face + 0];
			discr->mapping[2*numPoint+1] = gID;

			localStitchedID[gID -1] = numPoint+1;

			globalID[numPoint] = gID;

			numPoint += 1;

		}

		// Get triangle connectivity in global sense
		for (i = 0; i < tlen; i++) {

			discr->elems[numTri].bIndex      = bodyFaceMap[2*face + 0];
			discr->elems[numTri].tIndex      = 1;
			discr->elems[numTri].eIndex      = bodyFaceMap[2*face + 1];

			discr->elems[numTri].gIndices    = (int *) EG_alloc(6*sizeof(int));
			if (discr->elems[numTri].gIndices == NULL) {
				status = EGADS_MALLOC;
				goto cleanup;
			}

			discr->elems[numTri].dIndices    = NULL;
			discr->elems[numTri].eTris.tq[0] = i+1;

			status = EG_localToGlobal(tess, bodyFaceMap[2*face + 1], tris[3*i + 0], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			discr->elems[numTri].gIndices[0] = localStitchedID[gID-1];
			discr->elems[numTri].gIndices[1] = tris[3*i + 0];

			status = EG_localToGlobal(tess, bodyFaceMap[2*face + 1], tris[3*i + 1], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			discr->elems[numTri].gIndices[2] = localStitchedID[gID-1];
			discr->elems[numTri].gIndices[3] = tris[3*i + 1];

			status = EG_localToGlobal(tess, bodyFaceMap[2*face + 1], tris[3*i + 2], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			discr->elems[numTri].gIndices[4] = localStitchedID[gID-1];
			discr->elems[numTri].gIndices[5] = tris[3*i + 2];

			numTri += 1;
		}
	}

	discr->nPoints = numPoint;

	#ifdef DEBUG
		printf(" nastranAIM/aimDiscr: ntris = %d, npts = %d!\n", discr->nElems, discr->nPoints);
	#endif

	// Resize mapping to switched together number of points
	discr->mapping = (int *) EG_reall(discr->mapping, 2*numPoint*sizeof(int));

	// Local to global node connectivity + numCAPSGroup + sizeof(capGrouplist)
	storage  = (int *) EG_alloc((numPoint + 1 + numCAPSGroup) *sizeof(int));
	if (storage == NULL) goto cleanup;
	discr->ptrm = storage;

	// Store the global node id
	for (i = 0; i < numPoint; i++) {

		storage[i] = globalID[i];

	    //#ifdef DEBUG
		//	printf(" nastranAIM/aimDiscr: Instance = %d, Global Node ID %d\n", iIndex, storage[i]);
		//#endif
	}

	// Save way the attrMap capsGroup list
	storage[numPoint] = numCAPSGroup;
	for (i = 0; i < numCAPSGroup; i++) {
	    storage[numPoint+1+i] = capsGroupList[i];
	}

	#ifdef DEBUG
        printf(" nastranAIM/aimDiscr: Instance = %d, Finished!!\n", iIndex);
    #endif

	status = CAPS_SUCCESS;

	goto cleanup;

	cleanup:
		if (faces != NULL) EG_free(faces);

		if (globalID  != NULL) EG_free(globalID);
		if (localStitchedID != NULL) EG_free(localStitchedID);

		if (capsGroupList != NULL) EG_free(capsGroupList);
		if (bodyFaceMap != NULL) EG_free(bodyFaceMap);

		if (status != CAPS_SUCCESS) {
			aimFreeDiscr(discr);
		}

		return status;
}

int aimLocateElement(capsDiscr *discr, double *params, double *param, int *eIndex,
                 	 double *bary)
{
	int    i, in[3], stat, ismall;
	double we[3], w, smallw = -1.e300;

	if (discr == NULL) return CAPS_NULLOBJ;

	for (ismall = i = 0; i < discr->nElems; i++) {
		in[0] = discr->elems[i].gIndices[0] - 1;
		in[1] = discr->elems[i].gIndices[2] - 1;
		in[2] = discr->elems[i].gIndices[4] - 1;
		stat  = EG_inTriExact(&params[2*in[0]], &params[2*in[1]], &params[2*in[2]], param, we);

		if (stat == EGADS_SUCCESS) {
			*eIndex = i+1;
			bary[0] = we[1];
			bary[1] = we[2];
			return CAPS_SUCCESS;
		}

		w = we[0];
		if (we[1] < w) w = we[1];
		if (we[2] < w) w = we[2];
		if (w > smallw) {
			ismall = i+1;
			smallw = w;
		}
	}

	/* must extrapolate! */
	if (ismall == 0) return CAPS_NOTFOUND;
	in[0] = discr->elems[ismall-1].gIndices[0] - 1;
	in[1] = discr->elems[ismall-1].gIndices[2] - 1;
	in[2] = discr->elems[ismall-1].gIndices[4] - 1;
	EG_inTriExact(&params[2*in[0]], &params[2*in[1]], &params[2*in[2]], param, we);

	*eIndex = ismall;
	bary[0] = we[1];
	bary[1] = we[2];

	/*
  	  printf(" aimLocateElement: extropolate to %d (%lf %lf %lf)  %lf\n", ismall, we[0], we[1], we[2], smallw);
	*/
	return CAPS_SUCCESS;
}

int aimTransfer(capsDiscr *discr, char *dataName, int numPoint, int dataRank, double *dataVal, char **units)
{
    int status; // Function return status
    int i, j, dataPoint; // Indexing


    // Current directory path
    char currentPath[PATH_MAX];

    char *extF06 = ".F06";

    // FO6 data variables
    int numGridPoint = 0;
    int numEigenVector = 0;
    double **dataMatrix = NULL;

    // Variables used in global node mapping
    int *nodeMap, *storage;
    int globalNodeID;

    // Filename stuff
    char *filename = NULL;
    FILE *fp; // File pointer

    #ifdef DEBUG
        printf(" nastranAIM/aimTransfer name = %s  instance = %d  npts = %d/%d!\n", dataName, discr->instance, numPoint, dataRank);
    #endif

    //Get the appropriate parts of the tessellation to data
    storage = (int *) discr->ptrm;
    nodeMap = &storage[0]; // Global indexing on the body

    //capsGroupList = &storage[discr->nPoints]; // List of boundary ID (attrMap) in the transfer

    if (strcasecmp(dataName, "Displacement") != 0 &&
    	strcasecmp(dataName, "EigenVector")  != 0) {

    	printf("Unrecognized data transfer variable - %s\n", dataName);
    	return CAPS_NOTFOUND;
    }

    (void) getcwd(currentPath, PATH_MAX);
    if (chdir(nastranInstance[discr->instance].analysisPath) != 0) {
        #ifdef DEBUG
            printf(" nastranAIM/aimTransfer Cannot chdir to %s!\n", nastranInstance[discr->instance].analysisPath);
        #endif

        return CAPS_DIRERR;
    }

    filename = (char *) EG_alloc((strlen(nastranInstance[discr->instance].projectName) +
                                  strlen(extF06) + 1)*sizeof(char));

    sprintf(filename,"%s%s",nastranInstance[discr->instance].projectName,extF06);

    // Open file
    fp = fopen(filename, "r");
    if (filename != NULL) EG_free(filename);
    filename = NULL;

    if (fp == NULL) {
        printf("Unable to open file: %s\n", filename);
        chdir(currentPath);
        return CAPS_IOERR;
    }

    if (strcasecmp(dataName, "Displacement") == 0) {

        if (dataRank != 3) {
            printf("Invalid rank for dataName \"Displacement\"!!!\n");
            status = CAPS_BADRANK;

        } else {

            status = nastran_readF06Displacement(fp,
                                                 1,
                                                 &numGridPoint,
                                                 &dataMatrix);
            fclose(fp);
        }

    } else if (strcasecmp(dataName, "EigenVector") == 0) {

        status = nastran_readF06EigenVector(fp,
                                            &numEigenVector,
                                            &numGridPoint,
                                            &dataMatrix);
        fclose(fp);
    } else {

        status = CAPS_NOTFOUND;
    }

    // Restore the path we came in
    chdir(currentPath);
    if (status != CAPS_SUCCESS) return status;


    for (i = 0; i < numPoint; i++) {

        globalNodeID = nodeMap[i];

        if (strcasecmp(dataName, "Displacement") == 0) {

            for (dataPoint = 0; dataPoint < numGridPoint; dataPoint++) {
                if ((int) dataMatrix[dataPoint][0] ==  globalNodeID) break;
            }

            dataVal[dataRank*i+0] = dataMatrix[dataPoint][2];// rvec[rank*map[i]+j];
            dataVal[dataRank*i+1] = dataMatrix[dataPoint][3];
            dataVal[dataRank*i+2] = dataMatrix[dataPoint][4];
        }
    }

    // Free data matrix
    if (dataMatrix != NULL) {

        if      (strcasecmp(dataName, "Displacement") == 0) j = numGridPoint;
        else if (strcasecmp(dataName, "EigenVector")  == 0) j = numEigenVector;


        for (i = 0; i < j; i++) {
            if (dataMatrix[i] != NULL) EG_free(dataMatrix[i]);
        }
        EG_free(dataMatrix);
    }

    return CAPS_SUCCESS;
}
int aimInterpolation(capsDiscr *discr, /*@unused@*/ char *name, int eIndex,
                 double *bary, int rank, double *data, double *result)
{
	int    in[3], i;
	double we[3];
	/*
	#ifdef DEBUG
  	  printf(" nastranAIM/aimInterpolation  %s  instance = %d!\n",
         name, discr->instance);
	#endif
	 */
	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" nastranAIM/Interpolation: eIndex = %d [1-%d]!\n",eIndex, discr->nElems);
	}

	we[0] = 1.0 - bary[0] - bary[1];
	we[1] = bary[0];
	we[2] = bary[1];
	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[2] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[4] - 1;
	for (i = 0; i < rank; i++){
		result[i] = data[rank*in[0]+i]*we[0] + data[rank*in[1]+i]*we[1] + data[rank*in[2]+i]*we[2];
	}

	return CAPS_SUCCESS;
}


int aimInterpolateBar(capsDiscr *discr, /*@unused@*/ char *name, int eIndex,
					  double *bary, int rank, double *r_bar, double *d_bar)
{
	int    in[3], i;
	double we[3];

	/*
	#ifdef DEBUG
  	  printf(" nastranAIM/aimInterpolateBar  %s  instance = %d!\n", name, discr->instance);
	#endif
	 */

	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" nastranAIM/InterpolateBar: eIndex = %d [1-%d]!\n", eIndex, discr->nElems);
	}

	we[0] = 1.0 - bary[0] - bary[1];
	we[1] = bary[0];
	we[2] = bary[1];
	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[2] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[4] - 1;

	for (i = 0; i < rank; i++) {
		/*  result[i] = data[rank*in[0]+i]*we[0] + data[rank*in[1]+i]*we[1] +
                data[rank*in[2]+i]*we[2];  */
		d_bar[rank*in[0]+i] += we[0]*r_bar[i];
		d_bar[rank*in[1]+i] += we[1]*r_bar[i];
		d_bar[rank*in[2]+i] += we[2]*r_bar[i];
	}

	return CAPS_SUCCESS;
}


int aimIntegration(capsDiscr *discr, /*@unused@*/ char *name, int eIndex, int rank,
				   /*@null@*/ double *data, double *result)
{
	int    i, in[3], stat, ptype, pindex, nBody;
	double x1[3], x2[3], x3[3], xyz1[3], xyz2[3], xyz3[3], area;
	ego    *bodies;

	/*
	#ifdef DEBUG
  	  printf(" nastranAIM/aimIntegration  %s  instance = %d!\n", name, discr->instance);
	#endif
	 */
	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" nastranAIM/aimIntegration: eIndex = %d [1-%d]!\n", eIndex, discr->nElems);
	}

	stat = aim_getBodies(discr->aInfo, &nBody, &bodies);
	if (stat != CAPS_SUCCESS) {
		printf(" nastranAIM/aimIntegration: %d aim_getBodies = %d!\n", discr->instance, stat);
		return stat;
	}

	/* element indices */

	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[2] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[4] - 1;

	stat = EG_getGlobal(bodies[discr->mapping[2*in[0]]+nBody-1],
						discr->mapping[2*in[0]+1], &ptype, &pindex, xyz1);
	if (stat != CAPS_SUCCESS) {
		printf(" nastranAIM/aimIntegration: %d EG_getGlobal %d = %d!\n", discr->instance, in[0], stat);
		return stat;
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[1]]+nBody-1],
                        discr->mapping[2*in[1]+1], &ptype, &pindex, xyz2);
	if (stat != CAPS_SUCCESS) {
		printf(" nastranAIM/aimIntegration: %d EG_getGlobal %d = %d!\n", discr->instance, in[1], stat);
    	return stat;
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[2]]+nBody-1],
                        discr->mapping[2*in[2]+1], &ptype, &pindex, xyz3);
	if (stat != CAPS_SUCCESS) {
		printf(" nastranAIM/aimIntegration: %d EG_getGlobal %d = %d!\n", discr->instance, in[2], stat);
		return stat;
	}

	x1[0] = xyz2[0] - xyz1[0];
	x2[0] = xyz3[0] - xyz1[0];
	x1[1] = xyz2[1] - xyz1[1];
	x2[1] = xyz3[1] - xyz1[1];
	x1[2] = xyz2[2] - xyz1[2];
	x2[2] = xyz3[2] - xyz1[2];

	//CROSS(x3, x1, x2);
	cross_DoubleVal(x1, x2, x3);

	//area  = sqrt(DOT(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */
	area  = sqrt(dot_DoubleVal(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */

	if (data == NULL) {
		*result = 3.0*area;
		return CAPS_SUCCESS;
	}

	for (i = 0; i < rank; i++) {
		result[i] = (data[rank*in[0]+i] + data[rank*in[1]+i] + data[rank*in[2]+i])*area;
	}

	return CAPS_SUCCESS;
}


int aimIntegrateBar(capsDiscr *discr, /*@unused@*/ char *name, int eIndex, int rank,
                    double *r_bar, double *d_bar)
{
	int    i, in[3], stat, ptype, pindex, nBody;
	double x1[3], x2[3], x3[3], xyz1[3], xyz2[3], xyz3[3], area;
	ego    *bodies;

	/*
	#ifdef DEBUG
  	  printf(" nastranAIM/aimIntegrateBar  %s  instance = %d!\n", name, discr->instance);
	#endif
	 */

	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" nastranAIM/aimIntegrateBar: eIndex = %d [1-%d]!\n", eIndex, discr->nElems);
	}

	stat = aim_getBodies(discr->aInfo, &nBody, &bodies);
	if (stat != CAPS_SUCCESS) {
		printf(" nastranAIM/aimIntegrateBar: %d aim_getBodies = %d!\n", discr->instance, stat);
		return stat;
	}

	/* element indices */

	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[2] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[4] - 1;
	stat = EG_getGlobal(bodies[discr->mapping[2*in[0]]+nBody-1],
                       discr->mapping[2*in[0]+1], &ptype, &pindex, xyz1);
	if (stat != CAPS_SUCCESS) {
		printf(" nastranAIM/aimIntegrateBar: %d EG_getGlobal %d = %d!\n", discr->instance, in[0], stat);
		return stat;
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[1]]+nBody-1],
                       discr->mapping[2*in[1]+1], &ptype, &pindex, xyz2);

	if (stat != CAPS_SUCCESS) {
		printf(" nastranAIM/aimIntegrateBar: %d EG_getGlobal %d = %d!\n", discr->instance, in[1], stat);
		return stat;
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[2]]+nBody-1],
                      discr->mapping[2*in[2]+1], &ptype, &pindex, xyz3);

	if (stat != CAPS_SUCCESS) {
		printf(" nastranAIM/aimIntegrateBar: %d EG_getGlobal %d = %d!\n", discr->instance, in[2], stat);
		return stat;
	}

	x1[0] = xyz2[0] - xyz1[0];
	x2[0] = xyz3[0] - xyz1[0];
	x1[1] = xyz2[1] - xyz1[1];
	x2[1] = xyz3[1] - xyz1[1];
	x1[2] = xyz2[2] - xyz1[2];
	x2[2] = xyz3[2] - xyz1[2];

	//CROSS(x3, x1, x2);
	cross_DoubleVal(x1, x2, x3);

	//area  = sqrt(DOT(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */
	area  = sqrt(dot_DoubleVal(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */

	for (i = 0; i < rank; i++) {
	/*  result[i] = (data[rank*in[0]+i] + data[rank*in[1]+i] +
				 data[rank*in[2]+i])*area;  */
		d_bar[rank*in[0]+i] += area*r_bar[i];
		d_bar[rank*in[1]+i] += area*r_bar[i];
		d_bar[rank*in[2]+i] += area*r_bar[i];
	}

	return CAPS_SUCCESS;
}
