/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             SU2 AIM Based on FUN3D AIM
 *
 *     Modified from code written by Dr. Ryan Durscher AFRL/RQVC
 *
 */


/*!\mainpage Introduction
 *
 * \section overviewSU2 SU2 AIM Overview
 * This module can be used to interface SU2 with geometry in CAPS system. The 
 * SU2 AIM has both c-version and a python-version for the interface. For 
 * SU2 capabilities and related documentation, please refer to 
 * http://su2.stanford.edu/. SU2 expects a volume mesh file and a corresponding
 * configuration file to perform the analysis. 
 *
 * \subsection meshSU2 Automatic generation of SU2 Mesh file
 * The volume mesh file from SU2 AIM is written in native SU2 
 * format ("filename.su2"). The description of the native SU2 mesh can be 
 * found SU2 website. For the automatic generation of mesh file, SU2 AIM 
 * depends on Mesh AIMs, for example, TetGen or AFLR4/3 AIM. 
 *
 * \subsection configSU2 Automatic generation of SU2 Configuration file
 * The configuration file ("filename.cfg") from SU2 AIM is automatically 
 * created by using the flow features and boundary conditions that were set in 
 * the driver program as a user input. For the rest of the configuration 
 * variables, defualt set of values are provided for a general execution.
 * If desired, a user has freedom to manually (a) change these variables based 
 * on  personal preference, or (b) override the configuration file with unique 
 * configuration variables.
 
 
 * Related pages:
 * - \ref cfdBoundaryConditions
 * - \ref geomFidelityFUN3D
 * - \ref geomFidelitySU2
 * - \ref aimInputsSU2
 * - \ref aimOutputsSU2
 * - \ref sharableDataSU2
 */


#include <string.h>
#include <ctype.h>
#include <math.h>
#include "capsTypes.h"
#include "aimUtil.h"

#include "meshUtils.h"
#include "cfdUtils.h"
#include "miscUtils.h"
#include "fun3dUtils.h"


#ifdef HAVE_PYTHON
#include "Python.h" // Bring in Python API
#endif


#ifdef WIN32
#define getcwd     _getcwd
#define snprintf   _snprintf
#define strcasecmp stricmp
#define PATH_MAX   _MAX_PATH
#else
#include <unistd.h>
#include <limits.h>
#endif

#define NUMINPUT   38
#define NUMOUTPUT  24

#define Fidelity2D 200 // Base fidelity value for 2D simulations
#define Fidelity3D 300 // Base fiedlity value for 3D simulations
#define NUMFID    4    // Number of fidelities contained in the aim

#define MXCHAR  255

//#define DEBUG

typedef struct {

    // SU2 project name
    char *projectName;

    // Attribute to index map
    mapAttrToIndexStruct attrMap;

	// Flag to see if we need to get an FEA mesh
   int haveVolumeMesh;

    // Check to make sure data transfer is ok
    int dataTransferCheck;

    // Analysis file path/directory
    char *analysisPath;

    // Pointer to caps input value for scaling pressure during data transfer
    capsValue *pressureScaleFactor;

    // Pointer to caps input value for offset pressure during data transfer
    capsValue *pressureScaleOffset;

} aimStorage;

static aimStorage *su2Instance = NULL;
static int         numInstance  = 0;


extern /*@null@*/ /*@only@*/
	char       *EG_strdup(/*@null@*/ const char *str);
extern  int         EG_inTriExact(double *t1, double *t2, double *t3,
                                  double *p,  double *w );

// Convert a string to upper case, SU2 requirement
void forceUpperCase ( char *sPtr ) 
{
    while ( *sPtr != '\0' ) {   
        *sPtr = toupper ( ( unsigned char ) *sPtr );
        ++sPtr;
    }   
}


#include "su2DataExchange.c"

/* ********************** Exposed AIM Functions ***************************** */

int aimInitialize(/*@unused@*/ int ngIn, /*@unused@*/ /*@null@*/ capsValue *gIn,
                  int *nFid, int **fids, int *nIn, int *nOut,
                  int *nFields, char ***fnames, int **ranks)
{
    int *ints;

    int status; // Function return

    aimStorage *tmp;

    #ifdef DEBUG
        printf("\n su2AIM/aimInitialize   ngIn = %d!\n", ngIn);
    #endif

    *nFid    = 5;
    ints     = (int *) EG_alloc(((int) *nFid)*sizeof(int));
    if (ints == NULL) return EGADS_MALLOC;

    /*! \page geomFidelitySU2 Geometry Fidelity
	* The geometric fidelity for the SU2 AIM requires that the body(ies) be either a face body(ies), solid
	* body(ies) or non- and manifold sheet body(ies). For 2D simulations the attribute capsFidelity should
	* be set to CFD with the geometric fidelity being either a planar SHEETBODY or FACEBODY, while for 3D
    * simulations be set to CFD but the both SOLIDBODY(ies) or SHEETBODY(ies) are acceptable.
	*/
    ints[0]  = ALL | ALL;
    ints[1]  = ALL | CFD;
    ints[2]  = SOLIDBODY | CFD;
    ints[3]  = SHEETBODY | CFD;
    ints[4]  = FACEBODY | CFD;

    *fids    = ints;

    /* specify the number of analysis input and out "parameters" */
    *nIn     = NUMINPUT;
    *nOut    = NUMOUTPUT;

    /* specify the field variables this analysis can generate */
    *nFields = 0;
    *ranks   = NULL;
    *fnames  = NULL;

    // Allocate su2Instance
    if (su2Instance == NULL) {

    	su2Instance = (aimStorage *) EG_alloc(sizeof(aimStorage));

    	if (su2Instance == NULL) {
    		EG_free(*fids);
    		*fids   = NULL;
    		return EGADS_MALLOC;
    	}

    } else {
    	tmp = (aimStorage *) EG_reall(su2Instance, (numInstance+1)*sizeof(aimStorage));
    	if (tmp == NULL) {
    		EG_free(*fids);
    		*fids   = NULL;
    		return EGADS_MALLOC;
    	}

    	su2Instance = tmp;
    }

    // Set initial values for su2Instance
    su2Instance[numInstance].projectName = NULL;

    // Analysis file path/directory
    su2Instance[numInstance].analysisPath = NULL;

    su2Instance[numInstance].haveVolumeMesh = (int) false;

    // Container for attribute to index map
    status = initiate_mapAttrToIndexStruct(&su2Instance[numInstance].attrMap);
    if (status != CAPS_SUCCESS) return status;

    // Check to make sure data transfer is ok
    su2Instance[numInstance].dataTransferCheck = (int) true;

    // Pointer to caps input value for scaling pressure during data transfer
    su2Instance[numInstance].pressureScaleFactor = NULL;

    // Pointer to caps input value for off setting pressure during data transfer
    su2Instance[numInstance].pressureScaleOffset = NULL;

    // Increment number of instances
    numInstance += 1;

    return (numInstance -1);
}

int aimInputs(int iIndex, void *aimInfo, int index, char **ainame, capsValue *defval)
{
	int status;

	#ifdef DEBUG
        printf(" su2AIM/aimInputs instance = %d  index = %d!\n", iIndex, index);
	#endif

    *ainame = NULL;
  
	// SU2 Inputs
    /*! \page aimInputsSU2 User Inputs For SU2 Configuration File 
     * For the description of the configuration variables, associated values, 
     * and available options refer to the template configuration file that is 
     * distributed with SU2. 
     * Note: The configuration file is dependent on the version of SU2 used. 
     * This configuration file that will be auto generated is compatible with 
     * SU2 4.1.1.
     *
     * Following variables are automatically set based on the input from 
     * driver program:
     *
     * Keyword                 |  Value
     * :--:                    | :--:
     * MESH_FILENAME           | [Proj_Name].su2
     * MESH_FORMAT             | ASCII
     *
     * Following is a list of configuration variables that can be set by the 
     * driver program that will be used in creating SU2 configuration file.
     *
     * Keyword                 |  Default
     * :--:                    | :--:
     * Boundary_Condition      | Refer \ref cfdBoundaryConditions
     * Overwrite_CFG           | false
     *   
     * Free stream conditions:
     * Keyword                 |  Default
     * :--:                    | :--:
     * Mach                    | 0.8
     * Alpha                   | 0.0
     * Beta                    | 0.0
     *
     * Numerical Scheme:
     * Keyword                 |  Default
     * :--:                    | :--:
     * Eqn                     | COMPRESSIBLE
     * Num_Iter                | 999
     * CFL_NUMBER              | 10.0
     * MGLEVEL                 | 0
     * RESIDUAL_REDUCTION      | 6
     *
     * Dimensionalization:
     * Keyword                 |  Default
     * :--:                    | :--:
     * SYSTEM_MEASUREMENTS     | SI
     * FREESTREAM_PRESSURE     | 101325.0 N/m^2
     * FREESTREAM_TEMPERATURE  | 288.15K
     * FREESTREAM_DENSITY      | 1.2886 Kg/m^3
     * FREESTREAM_VELOCITY     | 1.0 m/s
     * FREESTREAM_VISCOSITY    | 1.853E-5 N s/m^2
     * REF_DIMENSIONALIZATION  | DIMENSIONAL
     * REF_ORIGIN_MOMENT_X     | 0.25 m
     * REF_ORIGIN_MOMENT_Y     | 0.00 m
     * REF_ORIGIN_MOMENT_Z     | 0.00 m
     * REF_LENGTH_MOMENT       | 1.00 m
     * REF_AREA                | 0, compute automagically
     *
     * Output
     * Keyword                 |  Default
     * :--:                    | :--:
     *  OUTPUT_FORMAT          | PARAVIEW
     *  
     *  
     */
	if (index == 1) {
	    *ainame              = EG_strdup("Proj_Name");
		defval->type         = String;
		defval->nullVal      = NotNull;
		defval->vals.string  = EG_strdup("su2_CAPS");
		defval->lfixed       = Change;

	} else if (index == 2) {
		*ainame              = EG_strdup("Mach"); // Mach number
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 3) {
		*ainame              = EG_strdup("Re"); // Reynolds number
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 4) {
		*ainame              = EG_strdup("Viscous"); // Viscous term
		defval->type         = String;
		defval->vals.string  = NULL;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 5) {
		*ainame              = EG_strdup("Eqn"); // Equation type
		defval->type         = String;
		defval->vals.string  = NULL;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
	} else if (index == 6) {
		*ainame              = EG_strdup("Alpha");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 7) {
		*ainame              = EG_strdup("Beta");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 8) {
        *ainame              = EG_strdup("Overwrite_CFG");
        defval->type         = Boolean;
        defval->vals.integer = false;
        defval->nullVal      = NotNull;
	} else if (index == 9) {
	    *ainame              = EG_strdup("Mesh_Format");
        defval->type         = String;
        defval->vals.string  = EG_strdup("SU2");
        defval->lfixed       = Change;
	} else if (index == 10) {
        *ainame              = EG_strdup("Mesh_ASCII_Flag");
        defval->type         = Boolean;
        defval->vals.integer = true;
	} else if (index == 11) {
        *ainame              = EG_strdup("Num_Iter");
        defval->type         = Integer;
        defval->nullVal      = IsNull;
        defval->units        = NULL;
        defval->lfixed       = Change;
        defval->dim          = 0;
	} else if (index == 12) {
		*ainame              = EG_strdup("CFL_NUMBER");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
    } else if (index == 13) {  //not used in SU2 CFG
        *ainame               = EG_strdup("CFL_Schedule");
        defval->type          = Double;
        defval->dim           = 1;
        defval->length        = 2;
        defval->nrow          = 2;
        defval->ncol          = 1;
        defval->units         = NULL;
        defval->vals.reals    = (double *) EG_alloc(defval->length*sizeof(double));
        defval->nullVal       = IsNull;
        defval->lfixed        = Fixed;
    } else if (index == 14) { //not used in SU2 CFG
        *ainame               = EG_strdup("CFL_Schedule_Iter");
        defval->type          = Integer;
        defval->dim           = 1;
        defval->length        = 2;
        defval->nrow          = 2;
        defval->ncol          = 1;
        defval->units         = NULL;
        defval->vals.integers = (int *) EG_alloc(defval->length*sizeof(int));
        defval->nullVal       = IsNull;
        defval->lfixed        = Fixed;
    } else if (index == 15) { //not used in SU2 CFG
	    *ainame              = EG_strdup("Restart_Read");
		defval->type         = String;
		defval->vals.string  = NULL;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
    } else if (index == 16) {
        *ainame              = EG_strdup("Boundary_Condition");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->dim          = 1;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;
    } else if (index == 17) { //not used in SU2 CFG
        *ainame              = EG_strdup("Use_Python_NML");
        defval->type         = Boolean;
        defval->vals.integer = (int) false;
	} else if (index == 18) {
		*ainame              = EG_strdup("MGLEVEL");
		defval->type         = Integer;
		defval->vals.integer = 2;
		defval->units        = NULL;
		defval->dim          = 0;
	} else if (index == 19) {
		*ainame              = EG_strdup("RESIDUAL_REDUCTION");
		defval->type         = Integer;
		defval->vals.integer = 6;
		defval->units        = NULL;
		defval->dim          = 0;
	} else if (index == 20) {
	    *ainame              = EG_strdup("SYSTEM_MEASUREMENTS");
        defval->type         = String;
		defval->nullVal      = NotNull;
        defval->vals.string  = EG_strdup("SI");
        defval->lfixed       = Change;
	} else if (index == 21) {
	    *ainame              = EG_strdup("REF_DIMENSIONALIZATION");
        defval->type         = String;
		defval->nullVal      = NotNull;
        defval->vals.string  = EG_strdup("DIMENSIONAL");
        defval->lfixed       = Change;
	} else if (index == 22) {
		*ainame              = EG_strdup("FREESTREAM_PRESSURE");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 23) {
		*ainame              = EG_strdup("FREESTREAM_TEMPERATURE");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 24) {
		*ainame              = EG_strdup("FREESTREAM_DENSITY");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 25) {
		*ainame              = EG_strdup("FREESTREAM_VELOCITY");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 26) {
		*ainame              = EG_strdup("FREESTREAM_VISCOSITY");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 27) {
		*ainame              = EG_strdup("REF_ORIGIN_MOMENT_X");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 28) {
		*ainame              = EG_strdup("REF_ORIGIN_MOMENT_Y");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 29) {
		*ainame              = EG_strdup("REF_ORIGIN_MOMENT_Z");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 30) {
		*ainame              = EG_strdup("REF_LENGTH_MOMENT");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 31) {
		*ainame              = EG_strdup("REF_AREA");
		defval->type         = Double;
		defval->nullVal      = IsNull;
		defval->units        = NULL;
		defval->lfixed       = Change;
		defval->dim          = 0;
	} else if (index == 32) { //not used in SU2 CFG
        *ainame              = EG_strdup("Pressure_Scale_Factor");
        defval->type         = Double;
        defval->vals.real    = 1.0;
        defval->units      	 = NULL;

        su2Instance[iIndex].pressureScaleFactor = defval;
    } else if (index == 33) {
    	*ainame              = EG_strdup("Pressure_Scale_Offset");
    	defval->type         = Double;
    	defval->vals.real    = 0.0;
    	defval->units      	 = NULL;

    	su2Instance[iIndex].pressureScaleOffset = defval;


    } else if (index == 34) { //not used in SU2 CFG
    	*ainame              = EG_strdup("NonInertial_Rotation_Rate");
    	defval->type          = Double;
    	defval->dim           = 1;
    	defval->length        = 3;
    	defval->nrow          = 3;
    	defval->ncol          = 1;
    	defval->units         = NULL;
    	defval->vals.reals    = (double *) EG_alloc(defval->length*sizeof(double));
    	if (defval->vals.reals == NULL) {
    		return EGADS_MALLOC;
    	} else {
    		defval->vals.reals[0] = 0.0;
    		defval->vals.reals[1] = 0.0;
    		defval->vals.reals[2] = 0.0;
    	}
    	defval->nullVal       = IsNull;
    	defval->lfixed        = Fixed;
    } else if (index == 35) { //not used in SU2 CFG
    	*ainame              = EG_strdup("NonInertial_Rotation_Center");
    	defval->type          = Double;
    	defval->dim           = 1;
    	defval->length        = 3;
    	defval->nrow          = 3;
    	defval->ncol          = 1;
    	defval->units         = NULL;
    	defval->vals.reals    = (double *) EG_alloc(defval->length*sizeof(double));
    	if (defval->vals.reals == NULL) {
    		return EGADS_MALLOC;
    	} else {
    		defval->vals.reals[0] = 0.0;
    		defval->vals.reals[1] = 0.0;
    		defval->vals.reals[2] = 0.0;
    	}
    	defval->nullVal       = IsNull;
    	defval->lfixed        = Fixed;
	} else if (index == 36) {
	    *ainame              = EG_strdup("OUTPUT_FORMAT");
        defval->type         = String;
		defval->nullVal      = NotNull;
        defval->vals.string  = EG_strdup("PARAVIEW");
    } else if (index == 37) {
            *ainame              = EG_strdup("PHYSICAL_PROBLEM");
            defval->type         = String;
            defval->nullVal      = NotNull;
            defval->vals.string  = EG_strdup("EULER");
            defval->lfixed       = Change;
            defval->lfixed       = Change;

    }  else if (index == 38) {
    	*ainame              = EG_strdup("Two_Dimensional");
    	defval->type         = Boolean;
    	defval->type         = Boolean;
    	defval->vals.integer = (int) false;

    	/*! \page aimInputsSU2
    	 * - <B>Two_Dimensional = False</B> <br>
    	 * Run SU2 in 2D mode.
    	 */
    }



	// Link variable(s) to parent(s) if available
	if ((index != 1) && (*ainame != NULL) && (index !=15)) {
            status = aim_link(aimInfo, *ainame, ANALYSISIN, defval);
/*	    printf("Status = %d: Var Index = %d, Type = %d, link = %lX\n",
                   status, index, defval->type, defval->link);  */
	}

	return CAPS_SUCCESS;
}

int aimPreAnalysis(int iIndex, void *aimInfo, char *analysisPath,
                   capsValue *aimInputs, capsErrs **errs)
{

    // Function return flag
    int status;

    // Python linking
    int pythonLinked = (int) false;
    int usePython = (int) false;

    // Indexing
    int i, j, k, surf, body;

    // Data transfer
    int  nrow, ncol, rank;
    void *dataTransfer = NULL;
    enum capsvType vtype;
    char *units = NULL;

    // EGADS return values
    int          atype, alen;
    const int    *ints;
    const char   *string;
    const double *reals;

    // Output filename
    char *filename = NULL;

    // File pointer if we are going to internally write the namelist
    // FILE *fnml;

    // Current CAPS fidelity
    int currentFidelity = 0;

    // AIM input bodies
    int  numBody;
    ego *bodies = NULL;
    int  bodySubType = 0, bodyOClass, bodyNumChild, *bodySense;
    ego  bodyRef, *bodyChild;
	double bodyData[4];
  
    // For SU2 boundary tagging
    int counter;

    // Boundary/surface properties
    cfdBCsStruct bcProps;
  
    // Boundary conditions container - for writing .mapbc file
    bndCondStruct bndConds;
  
    // Volume Mesh obtained from meshing AIM
    genUnstrMeshStruct volumeMesh;

    int xMeshConstant = (int) true, yMeshConstant = (int) true, zMeshConstant= (int) true; // 2D mesh checks
    double tempCoord;

    int numNodeCheck, numTriFaceCheck, numQuadFaceCheck; // Consistency checkers for volume-surface meshes
  

    // Discrete data transfer variables
    capsDiscr *dataTransferDiscreteObj;
    char **transferName = NULL;
    int numTransferName;
    enum capsdMethod dataTransferMethod;
    int numDataTransferPoint;
    int numDataTransferElement = 0;
    int dataTransferRank;
    double *dataTransferData;

    // Variables used in global node mapping
    int *nodeMap, *storage;
    int globalNodeID;
    int *capsGroupList;

    // Data transfer Out variables
    const char *dataOutName[] = {"x","y","z", "id", "dx", "dy", "dz"};
    double **dataOutMatrix = NULL;
    int *dataOutFormat = NULL;
    int *dataConnectMatrix = NULL;
    int numOutVariable = 7;
    int numOutDataPoint = 0;
    int numOutDataConnect = 0;
    int foundDisplacement = (int) false;

    // NULL out errors
    *errs = NULL;
  
    // Store away the analysis path/directory
    su2Instance[iIndex].analysisPath = analysisPath;

    // Get AIM bodies
    status = aim_getBodies(aimInfo, &numBody, &bodies);
  
    #ifdef DEBUG
        printf(" su2AIM/aimPreAnalysis instance = %d  numBody = %d!\n", iIndex, numBody);
    #endif
  
    if ((numBody <= 0) || (bodies == NULL)) {
        #ifdef DEBUG
            printf(" su2AIM/aimPreAnalysis No Bodies!\n");
        #endif
        return CAPS_SOURCEERR;
    }

    // Check to see if python was linked
    #ifdef HAVE_PYTHON
        pythonLinked = (int) true;
    #else
        pythonLinked = (int) false;
    #endif

    // Should we use python even if it was linked?
    usePython = aimInputs[aim_getIndex(aimInfo, "Use_Python_NML", ANALYSISIN)-1].vals.integer;
    if (usePython == (int) true && pythonLinked == (int) false) {

        printf("Use of Python library requested but not linked!\n");
        usePython = (int) false;

    } else if (usePython == (int) false && pythonLinked == (int) true) {

        printf("Python library was linked, but will not be used!\n");
    }

    // Get project name
    su2Instance[iIndex].projectName = aimInputs[aim_getIndex(aimInfo, "Proj_Name", ANALYSISIN)-1].vals.string;

    // Get fidelity
    for (body = 0; body < numBody; body++) {
        status = EG_attributeRet(bodies[body], "capsFidelity", &atype, &alen, &ints, &reals, &string);
        if (status   != EGADS_SUCCESS) continue; // Need to add an error code

        if (body == 0){
            currentFidelity = (int) reals[0];

        } else {

        	if (currentFidelity != (int) reals[0]) {
        	    printf("All bodies must have the same capsFidelity value - Body 1 = %d, Body %d = %d!\n", currentFidelity,
                																								  body+1,
																												  (int) reals[0]);
        	    return CAPS_BADVALUE;
        	}
        }
    }

    // Get attribute to index mapping
    status = aim_getData(aimInfo, "Attribute_Map", &vtype, &rank, &nrow, &ncol, &dataTransfer, &units);
    if (status == CAPS_SUCCESS) {

        printf("Found link for attrMap (Attribute_Map) from parent\n");

        status = copy_mapAttrToIndexStruct( (mapAttrToIndexStruct *) dataTransfer, &su2Instance[iIndex].attrMap);
        if (status != CAPS_SUCCESS) return status;
    } else {

        if (status == CAPS_DIRTY) printf("Parent AIMS are dirty\n");
        else printf("Linking status during attrMap (Attribute_Map) = %d\n",status);

        printf("Didn't find a link to attrMap from parent - getting it ourselves\n");

        // Get capsGroup name and index mapping to make sure all faces have a capsGroup value
        status = create_CAPSGroupAttrToIndexMap(numBody,
                                                bodies,
                                                1, // Only search down to the face level of the EGADS body
                                                &su2Instance[iIndex].attrMap);
        if (status != CAPS_SUCCESS) return status;
    }

    // Get boundary conditions
    status = initiate_cfdBCsStruct(&bcProps);
    if (status != CAPS_SUCCESS) return status;

    // Get boundary conditions - if the boundary condition has been set
    if (aimInputs[aim_getIndex(aimInfo, "Boundary_Condition", ANALYSISIN)-1].nullVal ==  NotNull) {

        status = cfd_getBoundaryCondition( aimInputs[aim_getIndex(aimInfo, "Boundary_Condition", ANALYSISIN)-1].length,
                                           aimInputs[aim_getIndex(aimInfo, "Boundary_Condition", ANALYSISIN)-1].vals.tuple,
                                           &su2Instance[iIndex].attrMap,
                                           &bcProps);
        if (status != CAPS_SUCCESS) goto cleanup;

    } else {
        printf("Warning: No boundary conditions provided !!!!\n");
    }

    // Are we running in two-mode
    if (aimInputs[aim_getIndex(aimInfo, "Two_Dimensional", ANALYSISIN)-1].vals.integer == (int) true ) {

    	if (numBody > 1) {
    		printf("Only 1 body may be provided when running in two mode!!\n");
    		status = CAPS_BADVALUE;
    		goto cleanup;
    	}

    	for (body = 0; body < numBody; body++) {
			// What type of BODY do we have?
			status = EG_getTopology(bodies[body], &bodyRef, &bodyOClass, &bodySubType, bodyData, &bodyNumChild, &bodyChild, &bodySense);
			if (status != EGADS_SUCCESS) goto cleanup;

			if (bodySubType != FACEBODY && bodySubType != SHEETBODY) {
				printf("Body type must be either FACEBODY (%d) or a SHEETBODY (%d) when running in two mode!\n", FACEBODY, SHEETBODY);
				status = CAPS_BADTYPE;
				goto cleanup;
			}
    	}


    	// Initiate volume mesh since we will be creating it instead of inheriting it
    	status = initiate_genUnstrMeshStruct(&volumeMesh);
    	if (status != CAPS_SUCCESS) return status;

    	// Get Surface mesh
    	status = aim_getData(aimInfo, "Surface_Mesh", &vtype, &rank, &nrow, &ncol, &dataTransfer, &units);
    	if (status != CAPS_SUCCESS){

    		if (status == CAPS_DIRTY) printf("Parent AIMS are dirty\n");
    		else printf("Linking status = %d\n",status);

    		printf("Didn't find a link to a surface mesh (Surface_Mesh) from parent - a mesh will NOT be created.\n");

    	} else {

    		printf("Found link for surface mesh (Surface_Mesh) from parent\n");

    		volumeMesh.numNodes = (*(genUnstrSurfMeshStruct *) dataTransfer).numNodes;
    		volumeMesh.surfaceMesh.numTriFace = (*(genUnstrSurfMeshStruct *) dataTransfer).numTriFace;
    		volumeMesh.surfaceMesh.numQuadFace = (*(genUnstrSurfMeshStruct *) dataTransfer).numQuadFace;
    		volumeMesh.surfaceMesh.numBoundaryEdge = (*(genUnstrSurfMeshStruct *) dataTransfer).numBoundaryEdge;

    		volumeMesh.xyz = (*(genUnstrSurfMeshStruct *) dataTransfer).xyz;
    		volumeMesh.surfaceMesh.triFaceList = (*(genUnstrSurfMeshStruct *) dataTransfer).localTriFaceList;
    		volumeMesh.surfaceMesh.quadFaceList = (*(genUnstrSurfMeshStruct *) dataTransfer).localQuadFaceList;
    		volumeMesh.surfaceMesh.triFaceMarkList = (*(genUnstrSurfMeshStruct *) dataTransfer).triFaceMarkList;
    		volumeMesh.surfaceMesh.quadFaceMarkList = (*(genUnstrSurfMeshStruct *) dataTransfer).quadFaceMarkList;

    		volumeMesh.surfaceMesh.localBoundaryEdgeList = (*(genUnstrSurfMeshStruct *) dataTransfer).localBoundaryEdgeList;
    		volumeMesh.surfaceMesh.boundaryEdgeMarkList = (*(genUnstrSurfMeshStruct *) dataTransfer).boundaryEdgeMarkList;


    		// Constant x?
    		for (i = 0; i < volumeMesh.numNodes; i++) {
    			if ((volumeMesh.xyz[3*i + 0] - volumeMesh.xyz[3*0 + 0]) > 1E-7) {
    				xMeshConstant = (int) false;
    				break;
    			}
    		}

    		// Constant y?
    		for (i = 0; i < volumeMesh.numNodes; i++) {
    			if ((volumeMesh.xyz[3*i + 1] - volumeMesh.xyz[3*0 + 1] ) > 1E-7) {
    				yMeshConstant = (int) false;
    				break;
    			}
    		}

    		// Constant z?
    		for (i = 0; i < volumeMesh.numNodes; i++) {
    			if ((volumeMesh.xyz[3*i + 2]-volumeMesh.xyz[3*0 + 2]) > 1E-7) {
    				zMeshConstant = (int) false;
    				break;
    			}
    		}

    		if (zMeshConstant != (int) true) {
    			printf("SU2 expects 2D meshes be in the x-y plane... attempting to rotate mesh!\n");

    			if (xMeshConstant == (int) true && yMeshConstant == (int) false) {
    				printf("Swapping z and x coordinates!\n");
    	    		for (i = 0; i < volumeMesh.numNodes; i++) {
    	    			tempCoord = volumeMesh.xyz[3*i + 0];
    	    			volumeMesh.xyz[3*i + 0] = volumeMesh.xyz[3*i + 2];
    	    			volumeMesh.xyz[3*i + 2] = tempCoord;
    	    		}

    			} else if(xMeshConstant == (int) false && yMeshConstant == (int) true) {

    				printf("Swapping z and y coordinates!\n");
    				for (i = 0; i < volumeMesh.numNodes; i++) {
    					tempCoord = volumeMesh.xyz[3*i + 1];
    					volumeMesh.xyz[3*i + 1] = volumeMesh.xyz[3*i + 2];
    					volumeMesh.xyz[3*i + 2] = tempCoord;
    				}

    			} else {
    				printf("Unable to rotate mesh!\n");
    				status = CAPS_NOTFOUND;
    			}

    		}

    	}
    	// Can't currently do data transfer in 2D-mode
    	su2Instance[iIndex].dataTransferCheck = (int) false;
    } else {
		// Get Volume mesh
		status = aim_getData(aimInfo, "Volume_Mesh", &vtype, &rank, &nrow, &ncol, &dataTransfer, &units);
		if (status != CAPS_SUCCESS){

			if (status == CAPS_DIRTY) printf("Parent AIMS are dirty\n");
			else printf("Linking status = %d\n",status);

			printf("Didn't find a link to a volume mesh (Volume_Mesh) from parent - a volume mesh will NOT be created.\n");

			// No volume currently means we can't do datatransfer
			su2Instance[iIndex].dataTransferCheck = (int) false;
		} else {

			printf("Found link for volume mesh (Volume_Mesh) from parent\n");
			 volumeMesh = *(genUnstrMeshStruct *) dataTransfer;
		}
    }

    if (status == CAPS_SUCCESS) {

        status = initiate_bndCondStruct(&bndConds);
        if (status != CAPS_SUCCESS) return status;

        status = populate_bndCondStruct_from_bcPropsStruct(&bcProps, &bndConds);
        if (status != CAPS_SUCCESS) return status;

        // Replace dummy values in bcVal with FUN3D specific values (more to add here)
        printf("Writing boundary flags\n");
        for (i = 0; i < bcProps.numBCID ; i++) {
            printf(" - bcProps.surfaceProps[%d].surfaceType = %u\n", i, bcProps.surfaceProps[i].surfaceType); 

            // {UnknownBoundary, Inviscid, Viscous, Farfield, Freestream,
            //  BackPressure, Symmetry, SubsonicInflow, SubsonicOutflow}

            if      (bcProps.surfaceProps[i].surfaceType == Inviscid) bndConds.bcVal[i] = 3000;
            else if (bcProps.surfaceProps[i].surfaceType == Viscous)  bndConds.bcVal[i] = 4000;
            else if (bcProps.surfaceProps[i].surfaceType == Farfield) bndConds.bcVal[i] = 5000;
            else if (bcProps.surfaceProps[i].surfaceType == Extrapolate) bndConds.bcVal[i] = 5026;
            else if (bcProps.surfaceProps[i].surfaceType == Freestream)  bndConds.bcVal[i] = 5050;
            else if (bcProps.surfaceProps[i].surfaceType == BackPressure)    bndConds.bcVal[i] = 5051;
            else if (bcProps.surfaceProps[i].surfaceType == SubsonicInflow)  bndConds.bcVal[i] = 7011;
            else if (bcProps.surfaceProps[i].surfaceType == SubsonicOutflow) bndConds.bcVal[i] = 7012;
            else if (bcProps.surfaceProps[i].surfaceType == MassflowIn)      bndConds.bcVal[i] = 7036;
            else if (bcProps.surfaceProps[i].surfaceType == MassflowOut)     bndConds.bcVal[i] = 7031;
            else if (bcProps.surfaceProps[i].surfaceType == FixedInflow)     bndConds.bcVal[i] = 7100;
            else if (bcProps.surfaceProps[i].surfaceType == FixedOutflow)    bndConds.bcVal[i] = 7105;
            else if (bcProps.surfaceProps[i].surfaceType == Symmetry) {

                if      (bcProps.surfaceProps[i].symmetryPlane == 1) bndConds.bcVal[i] = 6021;
                else if (bcProps.surfaceProps[i].symmetryPlane == 2) bndConds.bcVal[i] = 6022;
                else if (bcProps.surfaceProps[i].symmetryPlane == 3) bndConds.bcVal[i] = 6023;
            }
        }
        printf("Done boundary flags\n");

        // Write SU2 Mesh
        filename = EG_alloc(MXCHAR +1);
        if (filename == NULL) return EGADS_MALLOC;
        strcpy(filename, analysisPath);
        strcat(filename, "/");
        strcat(filename, su2Instance[iIndex].projectName);

        status = write_SU2( filename, 
                            aimInputs[aim_getIndex(aimInfo, "Mesh_ASCII_Flag", ANALYSISIN)-1].vals.integer, 
                            &volumeMesh, bndConds.numBND, bndConds.bndID);


        if (filename != NULL) EG_free(filename);
        filename = NULL;

        status = destroy_bndCondStruct(&bndConds);
        if (status != CAPS_SUCCESS) return status;

        // Lets check the volume mesh

        // Do we have an individual surface mesh for each body
        if (volumeMesh.surfaceMesh.numSurface != numBody) {
        	printf("Number of inherited surface meshes in the volume mesh does not match the number "
        			"of bodies - data transfer will NOT be possible.\n\n");
        	su2Instance[iIndex].dataTransferCheck = (int) false;
        }

        // Check to make sure the volume mesher didn't add any unaccounted for points/faces
        numNodeCheck = numTriFaceCheck = numQuadFaceCheck = 0;
        for (i = 0; i < volumeMesh.surfaceMesh.numSurface; i++) {
        	numNodeCheck += volumeMesh.surfaceMesh.surfNodesLump[i];
        	numTriFaceCheck += volumeMesh.surfaceMesh.surfTriFaceLump[i];
        	numQuadFaceCheck+= volumeMesh.surfaceMesh.surfQuadFaceLump[i];
        }

        if (numNodeCheck != volumeMesh.surfaceMesh.numNodes) {

        	su2Instance[iIndex].dataTransferCheck = (int) false;
        	printf("Volume mesher added nodes - data transfer will NOT be possible.\n");

        } else if (numTriFaceCheck != volumeMesh.surfaceMesh.numTriFace) {

        	su2Instance[iIndex].dataTransferCheck = (int) false;
        	printf("Volume mesher added trifaces - data transfer will NOT be possible.\n");

        } else if (numQuadFaceCheck != volumeMesh.surfaceMesh.numQuadFace) {

        	su2Instance[iIndex].dataTransferCheck = (int) false;
        	printf("Volume mesher added quadfaces - data transfer will NOT be possible.\n");

        } else { // Data transfer appears to be ok
        	su2Instance[iIndex].dataTransferCheck = (int) true;
        }

        // To this point is doesn't appear that the volume mesh has done anything bad to our surface mesh(es)
        if (su2Instance[iIndex].dataTransferCheck == (int) true) {

        	// Lets store away our tessellation now
        	for (i = 0; i < volumeMesh.surfaceMesh.numSurface; i++) {

        		if (su2Instance[iIndex].haveVolumeMesh == (int) false) {
                	#ifdef DEBUG
        	        	printf(" su2AIM/aimPreAnalysis instance = %d Setting body tessellation number %d!\n", iIndex, i);
                	#endif

        	        status = aim_setTess(aimInfo, volumeMesh.surfaceMesh.bodyTessMappingLump[i]->egadsTess);
        	        if (status != CAPS_SUCCESS) {
        	        	printf(" aim_setTess return = %d\n", status);
        	        	return status;
        	        }

        	        su2Instance[iIndex].haveVolumeMesh = (int) true;
        		}
        	}
        }
    }

    // Clean up the volume mesh that was created for 2D mode
    if (aimInputs[aim_getIndex(aimInfo, "Two_Dimensional", ANALYSISIN)-1].vals.integer == (int) true ) {
    	volumeMesh.numNodes = 0;
    	volumeMesh.surfaceMesh.numTriFace = 0;
    	volumeMesh.surfaceMesh.numQuadFace = 0;
    	volumeMesh.surfaceMesh.numBoundaryEdge = 0;

    	volumeMesh.xyz = NULL;
    	volumeMesh.surfaceMesh.triFaceList = NULL;
    	volumeMesh.surfaceMesh.quadFaceList = NULL;
    	volumeMesh.surfaceMesh.triFaceMarkList = NULL;
    	volumeMesh.surfaceMesh.quadFaceMarkList = NULL;

    	volumeMesh.surfaceMesh.localBoundaryEdgeList = NULL;
    	volumeMesh.surfaceMesh.boundaryEdgeMarkList = NULL;
    }

    if (usePython == (int) true && pythonLinked == (int) true) {
        ///////////////////////////////////////////////////////////////////////////////////
        // Place holder for writing SU2 CFG file using python. Currently NOT implemented //
        ///////////////////////////////////////////////////////////////////////////////////
    } else {

        if (aimInputs[aim_getIndex(aimInfo, "Overwrite_CFG",ANALYSISIN)-1].vals.integer == (int) false) {

            printf("Since Python was not linked and/or being used, the \"Overwrite_CFG\" input needs to be set to \"true\" to give");
            printf(" permission to create a new SU2 cfg. SU2 CFG will NOT be updated!!\n");

        } else {

            FILE *fcfg;
          
            printf("Warning: The su2 cfg file will be overwritten!\n");

            filename = EG_alloc(MXCHAR +1);
            if (filename == NULL) return EGADS_MALLOC;
            sprintf(filename,"%s/%s.cfg", analysisPath, aimInputs[aim_getIndex(aimInfo, "Proj_Name",  ANALYSISIN)-1].vals.string);

            fcfg = fopen(filename,"w");
            EG_free(filename);
            if (fcfg == NULL) return CAPS_IOERR; // ADD error code

            fprintf(fcfg,"%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n");
            fprintf(fcfg,"%%                                                                              %%\n");
            fprintf(fcfg,"%% SU2 configuration file                                                       %%\n");
            fprintf(fcfg,"%% Created by SU2AIM for Project: \"%s\"\n", aimInputs[aim_getIndex(aimInfo, "Proj_Name",  ANALYSISIN)-1].vals.string);
            fprintf(fcfg,"%% File Version 4.1.1/2 \"Cardinal\"                                              %%\n");
            fprintf(fcfg,"%%                                                                              %%\n");
            fprintf(fcfg,"%% Please report bugs/comments/suggestions to NBhagat1@UDayton.edu              %%\n");
            fprintf(fcfg,"%%                                                                              %%\n");
            fprintf(fcfg,"%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ------------- DIRECT, ADJOINT, AND LINEARIZED PROBLEM DEFINITION ------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Physical governing equations (EULER, NAVIER_STOKES,\n");
            fprintf(fcfg,"%%                               WAVE_EQUATION, HEAT_EQUATION, FEM_ELASTICITY,\n");
            fprintf(fcfg,"%%                               POISSON_EQUATION)\n");
            if(aimInputs[aim_getIndex(aimInfo, "PHYSICAL_PROBLEM",  ANALYSISIN)-1].nullVal == NotNull) {
                forceUpperCase(aimInputs[aim_getIndex(aimInfo, "PHYSICAL_PROBLEM",  ANALYSISIN)-1].vals.string);
                fprintf(fcfg,"PHYSICAL_PROBLEM= %s\n", aimInputs[aim_getIndex(aimInfo, "PHYSICAL_PROBLEM",  ANALYSISIN)-1].vals.string);

            } else if (aimInputs[aim_getIndex(aimInfo, "Viscous",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"PHYSICAL_PROBLEM= NAVIER_STOKES\n");
            } else {
                fprintf(fcfg,"PHYSICAL_PROBLEM= EULER\n");
            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Specify turbulence model (NONE, SA, SA_NEG, SST)\n");
            fprintf(fcfg,"KIND_TURB_MODEL= NONE\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Mathematical problem (DIRECT, CONTINUOUS_ADJOINT)\n");
            fprintf(fcfg,"MATH_PROBLEM= DIRECT\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Restart solution (NO, YES)\n");
            fprintf(fcfg,"RESTART_SOL= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Regime type (COMPRESSIBLE, INCOMPRESSIBLE)\n");
            if (aimInputs[aim_getIndex(aimInfo, "Eqn",  ANALYSISIN)-1].nullVal == NotNull) {
                forceUpperCase( aimInputs[aim_getIndex(aimInfo, "Eqn",  ANALYSISIN)-1].vals.string );
                fprintf(fcfg,"REGIME_TYPE= %s\n", aimInputs[aim_getIndex(aimInfo, "Eqn",  ANALYSISIN)-1].vals.string);
            } else {
                fprintf(fcfg,"REGIME_TYPE= COMPRESSIBLE\n");
            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% System of measurements (SI, US)\n");
            fprintf(fcfg,"%% International system of units (SI): ( meters, kilograms, Kelvins,\n");
            fprintf(fcfg,"%%                                       Newtons = kg m/s^2, Pascals = N/m^2, \n");
            fprintf(fcfg,"%%                                       Density = kg/m^3, Speed = m/s,\n");
            fprintf(fcfg,"%%                                       Equiv. Area = m^2 )\n");
            fprintf(fcfg,"%% United States customary units (US): ( inches, slug, Rankines, lbf = slug ft/s^2, \n");
            fprintf(fcfg,"%%                                       psf = lbf/ft^2, Density = slug/ft^3, \n");
            fprintf(fcfg,"%%                                       Speed = ft/s, Equiv. Area = ft^2 )\n");
            forceUpperCase(aimInputs[aim_getIndex(aimInfo, "SYSTEM_MEASUREMENTS",  ANALYSISIN)-1].vals.string); 
            fprintf(fcfg,"SYSTEM_MEASUREMENTS= %s\n", aimInputs[aim_getIndex(aimInfo, "SYSTEM_MEASUREMENTS",  ANALYSISIN)-1].vals.string); 
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% -------------------- COMPRESSIBLE FREE-STREAM DEFINITION --------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Mach number (non-dimensional, based on the free-stream values)\n");
            if (aimInputs[aim_getIndex(aimInfo, "Mach",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"MACH_NUMBER= %f\n", aimInputs[aim_getIndex(aimInfo, "Mach",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"MACH_NUMBER= 0.8\n");
            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Angle of attack (degrees, only for compressible flows)\n");
            if (aimInputs[aim_getIndex(aimInfo, "Alpha",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"AoA= %f\n", aimInputs[aim_getIndex(aimInfo, "Alpha",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"AoA= 1.25\n");
            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Activate fixed lift mode (specify a CL instead of AoA, NO/YES)\n");
            fprintf(fcfg,"FIXED_CL_MODE= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Target coefficient of lift for fixed lift mode (0.0 by default)\n");
            fprintf(fcfg,"TARGET_CL= 0.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Iterations to re-evaluate the angle of attack (100 by default)\n");
            fprintf(fcfg,"ITER_FIXED_CL= 100\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Damping factor for fixed CL mode (0.1 by default)\n");
            //fprintf(fcfg,"DAMP_FIXED_CL= 0.2\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Side-slip angle (degrees, only for compressible flows)\n");
            if (aimInputs[aim_getIndex(aimInfo, "Beta",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"SIDESLIP_ANGLE= %f\n", aimInputs[aim_getIndex(aimInfo, "Beta",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"SIDESLIP_ANGLE= 0.0\n");
            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Init option to choose between Reynolds (default) or thermodynamics quantities\n");
            fprintf(fcfg,"%% for initializing the solution (REYNOLDS, TD_CONDITIONS)\n");
            fprintf(fcfg,"INIT_OPTION= REYNOLDS\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Free-stream option to choose between density and temperature (default) for\n");
            fprintf(fcfg,"%% initializing the solution (TEMPERATURE_FS, DENSITY_FS)\n");
            fprintf(fcfg,"FREESTREAM_OPTION= TEMPERATURE_FS\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Free-stream pressure (101325.0 N/m^2, 2116.216 psf by default)\n");
            if (aimInputs[aim_getIndex(aimInfo, "FREESTREAM_PRESSURE",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"FREESTREAM_PRESSURE= %f\n", aimInputs[aim_getIndex(aimInfo, "FREESTREAM_PRESSURE",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"FREESTREAM_PRESSURE= 101325.0\n");
            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Free-stream temperature (288.15 K, 518.67 R by default)\n");
            if (aimInputs[aim_getIndex(aimInfo, "FREESTREAM_TEMPERATURE",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"FREESTREAM_TEMPERATURE= %f\n", aimInputs[aim_getIndex(aimInfo, "FREESTREAM_TEMPERATURE",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"FREESTREAM_TEMPERATURE= 288.15\n");
            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Reynolds number (non-dimensional, based on the free-stream values)\n");
            if (aimInputs[aim_getIndex(aimInfo, "Re",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"REYNOLDS_NUMBER= %e\n", aimInputs[aim_getIndex(aimInfo, "Re",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"REYNOLDS_NUMBER= 6.5E6\n");
            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Reynolds length (1 m, 1 inch by default)\n");
            fprintf(fcfg,"REYNOLDS_LENGTH= 1.0\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% -------------------- INCOMPRESSIBLE FREE-STREAM DEFINITION ------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Free-stream density (1.2886 Kg/m^3, 0.0025 slug/ft^3 by default)\n");
            if (aimInputs[aim_getIndex(aimInfo, "FREESTREAM_DENSITY",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"FREESTREAM_DENSITY= %f\n", aimInputs[aim_getIndex(aimInfo, "FREESTREAM_DENSITY",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"FREESTREAM_DENSITY= 1.2886\n");
            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Free-stream velocity (1.0 m/s, 1.0 ft/s by default)\n");
            if (aimInputs[aim_getIndex(aimInfo, "FREESTREAM_VELOCITY",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"FREESTREAM_VELOCITY= (%f, 0.0, 0.0) \n", aimInputs[aim_getIndex(aimInfo, "FREESTREAM_VELOCITY",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"FREESTREAM_VELOCITY= (1.0, 0.0, 0.0)\n");
            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Free-stream viscosity (1.853E-5 N s/m^2, 3.87E-7 lbf s/ft^2 by default)\n");
            if (aimInputs[aim_getIndex(aimInfo, "FREESTREAM_VISCOSITY",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"FREESTREAM_VISCOSITY= %e\n", aimInputs[aim_getIndex(aimInfo, "FREESTREAM_VISCOSITY",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"FREESTREAM_VISCOSITY= 1.853E-5\n");
            }
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ---------------------- REFERENCE VALUE DEFINITION ---------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Reference origin for moment computation (m or in)\n");
            if (aimInputs[aim_getIndex(aimInfo, "REF_ORIGIN_MOMENT_X",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"REF_ORIGIN_MOMENT_X= %f\n", aimInputs[aim_getIndex(aimInfo, "REF_ORIGIN_MOMENT_X",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"REF_ORIGIN_MOMENT_X= 0.25\n");
            }
            if (aimInputs[aim_getIndex(aimInfo, "REF_ORIGIN_MOMENT_Y",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"REF_ORIGIN_MOMENT_Y= %f\n", aimInputs[aim_getIndex(aimInfo, "REF_ORIGIN_MOMENT_Y",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"REF_ORIGIN_MOMENT_Y= 0.00\n");
            }
            if (aimInputs[aim_getIndex(aimInfo, "REF_ORIGIN_MOMENT_Z",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"REF_ORIGIN_MOMENT_Z= %f\n", aimInputs[aim_getIndex(aimInfo, "REF_ORIGIN_MOMENT_Z",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"REF_ORIGIN_MOMENT_Z= 0.00\n");
            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Reference length for pitching, rolling, and yawing non-dimensional\n");
            fprintf(fcfg,"%% moment (m or in)\n");
            if (aimInputs[aim_getIndex(aimInfo, "REF_LENGTH_MOMENT",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"REF_LENGTH_MOMENT= %f\n", aimInputs[aim_getIndex(aimInfo, "REF_LENGTH_MOMENT",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"REF_LENGTH_MOMENT= 1.00\n");
            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Reference area for force coefficients (0 implies automatic\n");
            fprintf(fcfg,"%% calculation) (m^2 or in^2)\n");
            if (aimInputs[aim_getIndex(aimInfo, "REF_AREA",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"REF_AREA= %f\n", aimInputs[aim_getIndex(aimInfo, "REF_AREA",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"REF_AREA= 1.00\n");
            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Flow non-dimensionalization (DIMENSIONAL, FREESTREAM_PRESS_EQ_ONE,\n");
            fprintf(fcfg,"%%                              FREESTREAM_VEL_EQ_MACH, FREESTREAM_VEL_EQ_ONE)\n");
            forceUpperCase(aimInputs[aim_getIndex(aimInfo, "REF_DIMENSIONALIZATION",  ANALYSISIN)-1].vals.string); 
            fprintf(fcfg,"REF_DIMENSIONALIZATION= %s\n", aimInputs[aim_getIndex(aimInfo, "REF_DIMENSIONALIZATION",  ANALYSISIN)-1].vals.string); 
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ---- IDEAL GAS, POLYTROPIC, VAN DER WAALS AND PENG ROBINSON CONSTANTS -------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Different gas model (STANDARD_AIR, IDEAL_GAS, VW_GAS, PR_GAS)\n");
            fprintf(fcfg,"FLUID_MODEL= STANDARD_AIR\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Ratio of specific heats (1.4 default and the value is hardcoded\n");
            fprintf(fcfg,"%%                          for the model STANDARD_AIR)\n");
            fprintf(fcfg,"GAMMA_VALUE= 1.4\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Specific gas constant (287.058 J/kg*K default and this value is hardcoded \n");
            fprintf(fcfg,"%%                        for the model STANDARD_AIR)\n");
            fprintf(fcfg,"GAS_CONSTANT= 287.058\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Critical Temperature (131.00 K by default)\n");
            fprintf(fcfg,"CRITICAL_TEMPERATURE= 131.00\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Critical Pressure (3588550.0 N/m^2 by default)\n");
            fprintf(fcfg,"CRITICAL_PRESSURE= 3588550.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Critical Density (263.0 Kg/m3 by default)\n");
            fprintf(fcfg,"CRITICAL_DENSITY= 263.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Acentri factor (0.035 (air))\n");
            fprintf(fcfg,"ACENTRIC_FACTOR= 0.035\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% --------------------------- VISCOSITY MODEL ---------------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Viscosity model (SUTHERLAND, CONSTANT_VISCOSITY).\n");
            fprintf(fcfg,"VISCOSITY_MODEL= SUTHERLAND\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Molecular Viscosity that would be constant (1.716E-5 by default)\n");
            fprintf(fcfg,"MU_CONSTANT= 1.716E-5\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Sutherland Viscosity Ref (1.716E-5 default value for AIR SI)\n");
            fprintf(fcfg,"MU_REF= 1.716E-5\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Sutherland Temperature Ref (273.15 K default value for AIR SI)\n");
            fprintf(fcfg,"MU_T_REF= 273.15\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Sutherland constant (110.4 default value for AIR SI)\n");
            fprintf(fcfg,"SUTHERLAND_CONSTANT= 110.4\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% --------------------------- THERMAL CONDUCTIVITY MODEL ----------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Conductivity model (CONSTANT_CONDUCTIVITY, CONSTANT_PRANDTL).\n");
            fprintf(fcfg,"CONDUCTIVITY_MODEL= CONSTANT_PRANDTL\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Molecular Thermal Conductivity that would be constant (0.0257 by default)\n");
            fprintf(fcfg,"KT_CONSTANT= 0.0257\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ------------------------- UNSTEADY SIMULATION -------------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Unsteady simulation (NO, TIME_STEPPING, DUAL_TIME_STEPPING-1ST_ORDER, \n");
            fprintf(fcfg,"%%                      DUAL_TIME_STEPPING-2ND_ORDER, TIME_SPECTRAL)\n");
            fprintf(fcfg,"UNSTEADY_SIMULATION= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Time Step for dual time stepping simulations (s) -- Only used when UNST_CFL_NUMBER = 0.0\n");
            fprintf(fcfg,"UNST_TIMESTEP= 0.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Total Physical Time for dual time stepping simulations (s)\n");
            fprintf(fcfg,"UNST_TIME= 50.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Unsteady Courant-Friedrichs-Lewy number of the finest grid\n");
            fprintf(fcfg,"UNST_CFL_NUMBER= 0.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Number of internal iterations (dual time method)\n");
            fprintf(fcfg,"UNST_INT_ITER= 200\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Integer number of periodic time instances for Time Spectral\n");
            fprintf(fcfg,"TIME_INSTANCES= 1\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Iteration number to begin unsteady restarts\n");
            fprintf(fcfg,"UNST_RESTART_ITER= 0\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ----------------------- DYNAMIC MESH DEFINITION -----------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Dynamic mesh simulation (NO, YES)\n");
            fprintf(fcfg,"GRID_MOVEMENT= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Type of dynamic mesh (NONE, RIGID_MOTION, DEFORMING, ROTATING_FRAME,\n");
            fprintf(fcfg,"%%                       MOVING_WALL, STEADY_TRANSLATION, FLUID_STRUCTURE,\n");
            fprintf(fcfg,"%%                       AEROELASTIC, ELASTICITY, EXTERNAL,\n");
            fprintf(fcfg,"%%                       AEROELASTIC_RIGID_MOTION, GUST)\n");
            fprintf(fcfg,"GRID_MOVEMENT_KIND= DEFORMING\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Motion mach number (non-dimensional). Used for initializing a viscous flow\n");
            fprintf(fcfg,"%% with the Reynolds number and for computing force coeffs. with dynamic meshes.\n");
            fprintf(fcfg,"MACH_MOTION= 0.8\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Moving wall boundary marker(s) (NONE = no marker, ignored for RIGID_MOTION)\n");
            fprintf(fcfg,"MARKER_MOVING= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Coordinates of the motion origin\n");
            fprintf(fcfg,"MOTION_ORIGIN_X= 0.25\n");
            fprintf(fcfg,"MOTION_ORIGIN_Y= 0.0\n");
            fprintf(fcfg,"MOTION_ORIGIN_Z= 0.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Angular velocity vector (rad/s) about the motion origin\n");
            fprintf(fcfg,"ROTATION_RATE_X = 0.0\n");
            fprintf(fcfg,"ROTATION_RATE_Y = 0.0\n");
            fprintf(fcfg,"ROTATION_RATE_Z = 0.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Pitching angular freq. (rad/s) about the motion origin\n");
            fprintf(fcfg,"PITCHING_OMEGA_X= 0.0 \n");
            fprintf(fcfg,"PITCHING_OMEGA_Y= 0.0\n");
            fprintf(fcfg,"PITCHING_OMEGA_Z= 106.69842\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Pitching amplitude (degrees) about the motion origin\n");
            fprintf(fcfg,"PITCHING_AMPL_X= 0.0\n");
            fprintf(fcfg,"PITCHING_AMPL_Y= 0.0\n");
            fprintf(fcfg,"PITCHING_AMPL_Z= 1.01\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Pitching phase offset (degrees) about the motion origin\n");
            fprintf(fcfg,"PITCHING_PHASE_X= 0.0\n");
            fprintf(fcfg,"PITCHING_PHASE_Y= 0.0\n");
            fprintf(fcfg,"PITCHING_PHASE_Z= 0.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Translational velocity (m/s) in the x, y, & z directions\n");
            fprintf(fcfg,"TRANSLATION_RATE_X = 0.0\n");
            fprintf(fcfg,"TRANSLATION_RATE_Y = 0.0\n");
            fprintf(fcfg,"TRANSLATION_RATE_Z = 0.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Plunging angular freq. (rad/s) in x, y, & z directions\n");
            fprintf(fcfg,"PLUNGING_OMEGA_X= 0.0\n");
            fprintf(fcfg,"PLUNGING_OMEGA_Y= 0.0\n");
            fprintf(fcfg,"PLUNGING_OMEGA_Z= 0.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Plunging amplitude (m) in x, y, & z directions\n");
            fprintf(fcfg,"PLUNGING_AMPL_X= 0.0\n");
            fprintf(fcfg,"PLUNGING_AMPL_Y= 0.0\n");
            fprintf(fcfg,"PLUNGING_AMPL_Z= 0.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Move Motion Origin for marker moving (1 or 0)\n");
            fprintf(fcfg,"MOVE_MOTION_ORIGIN = 0\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% -------------- AEROELASTIC SIMULATION (Typical Section Model) ---------------%%\n");
            fprintf(fcfg,"%% Activated by GRID_MOVEMENT_KIND option\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% The flutter speed index (modifies the freestream condition in the solver)\n");
            fprintf(fcfg,"FLUTTER_SPEED_INDEX = 0.6\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Natural frequency of the spring in the plunging direction (rad/s)\n");
            fprintf(fcfg,"PLUNGE_NATURAL_FREQUENCY = 100\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Natural frequency of the spring in the pitching direction (rad/s)\n");
            fprintf(fcfg,"PITCH_NATURAL_FREQUENCY = 100\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% The airfoil mass ratio\n");
            fprintf(fcfg,"AIRFOIL_MASS_RATIO = 60\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Distance in semichords by which the center of gravity lies behind\n");
            fprintf(fcfg,"%% the elastic axis\n");
            fprintf(fcfg,"CG_LOCATION = 1.8\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% The radius of gyration squared (expressed in semichords)\n");
            fprintf(fcfg,"%% of the typical section about the elastic axis\n");
            fprintf(fcfg,"RADIUS_GYRATION_SQUARED = 3.48\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Solve the aeroelastic equations every given number of internal iterations\n");
            fprintf(fcfg,"AEROELASTIC_ITER = 3\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% --------------------------- GUST SIMULATION ---------------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Apply a wind gust (NO, YES)\n");
            fprintf(fcfg,"WIND_GUST = NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Type of gust (NONE, TOP_HAT, SINE, ONE_M_COSINE, VORTEX, EOG)\n");
            fprintf(fcfg,"GUST_TYPE = NONE\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Direction of the gust (X_DIR or Y_DIR)\n");
            fprintf(fcfg,"GUST_DIR = Y_DIR\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Gust wavelenght (meters)\n");
            fprintf(fcfg,"GUST_WAVELENGTH= 10.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Number of gust periods\n");
            fprintf(fcfg,"GUST_PERIODS= 1.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Gust amplitude (m/s)\n");
            fprintf(fcfg,"GUST_AMPL= 10.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Time at which to begin the gust (sec)\n");
            fprintf(fcfg,"GUST_BEGIN_TIME= 0.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Location at which the gust begins (meters) */\n");
            fprintf(fcfg,"GUST_BEGIN_LOC= 0.0\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ------------------------ SUPERSONIC SIMULATION ------------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Evaluate equivalent area on the Near-Field (NO, YES)\n");
            fprintf(fcfg,"EQUIV_AREA= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Integration limits of the equivalent area ( xmin, xmax, Dist_NearField )\n");
            fprintf(fcfg,"EA_INT_LIMIT= ( 1.6, 2.9, 1.0 )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Equivalent area scale factor ( EA should be ~ force based objective functions )\n");
            fprintf(fcfg,"EA_SCALE_FACTOR= 1.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Fix an azimuthal line due to misalignments of the near-field\n");
            fprintf(fcfg,"FIX_AZIMUTHAL_LINE= 90.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Drag weight in sonic boom Objective Function (from 0.0 to 1.0)\n");
            fprintf(fcfg,"DRAG_IN_SONICBOOM= 0.0\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% -------------------------- ENGINE SIMULATION --------------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Damping factor for the engine inflow.\n");
            fprintf(fcfg,"DAMP_ENGINE_INFLOW= 0.95\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Damping factor for the engine exhaust.\n");
            fprintf(fcfg,"DAMP_ENGINE_EXHAUST= 0.95\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Damping factor for the engine bleed.\n");
            fprintf(fcfg,"DAMP_ENGINE_BLEED= 0.95\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Engine nu factor (SA model).\n");
            fprintf(fcfg,"ENGINE_NU_FACTOR= 30.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Initialization with a subsonic flow around the engine.\n");
            fprintf(fcfg,"SUBSONIC_ENGINE= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Coordinates of the box that defines the subsonic region (Xmin, Ymin, Zmin,\n");
            fprintf(fcfg,"%%                                                          Xmax, Ymax, Zmax)\n");
            fprintf(fcfg,"SUBSONIC_ENGINE_BOX= ( -0.5, -0.49, 0.0, 2.5, 0.49, 0.0 )\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% --------------------- INVERSE DESIGN SIMULATION -----------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Evaluate an inverse design problem using Cp (NO, YES)\n");
            fprintf(fcfg,"INV_DESIGN_CP= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Evaluate an inverse design problem using heat flux (NO, YES)\n");
            fprintf(fcfg,"INV_DESIGN_HEATFLUX= NO\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% -------------------- BOUNDARY CONDITION DEFINITION --------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Euler wall boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"MARKER_EULER= (" );

            counter = 0; // Euler boundary
            for (i = 0; i < bcProps.numBCID; i++) {

            	if (bcProps.surfaceProps[i].surfaceType == Inviscid) {

            		if (counter > 0) fprintf(fcfg, ",");
            		fprintf(fcfg," %d", bcProps.surfaceProps[i].bcID);

            		counter += 1;
            	}
            }

            if(counter == 0) fprintf(fcfg," NONE");
            fprintf(fcfg," )\n");

            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Navier-Stokes (no-slip), constant heat flux wall  marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"%% Format: ( marker name, constant heat flux (J/m^2), ... )\n");
            fprintf(fcfg,"MARKER_HEATFLUX= (");

            counter = 0; // Viscous boundary w/ heat flux
            for (i = 0; i < bcProps.numBCID; i++) {
            	if (bcProps.surfaceProps[i].surfaceType == Viscous &&
            	    bcProps.surfaceProps[i].wallTemperatureFlag == (int) true &&
					bcProps.surfaceProps[i].wallTemperature < 0) {

            		if (counter > 0) fprintf(fcfg, ",");
            		fprintf(fcfg," %d, %f", bcProps.surfaceProps[i].bcID, bcProps.surfaceProps[i].wallHeatFlux);

            		counter += 1;
            	}
            }

            if(counter == 0) fprintf(fcfg," NONE");
			fprintf(fcfg," )\n");

            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Navier-Stokes (no-slip), isothermal wall marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"%% Format: ( marker name, constant wall temperature (K), ... )\n");
            fprintf(fcfg,"MARKER_ISOTHERMAL= (");

            counter = 0; // Viscous boundary w/ isothermal wall
            for (i = 0; i < bcProps.numBCID; i++) {
            	if (bcProps.surfaceProps[i].surfaceType == Viscous &&
            		bcProps.surfaceProps[i].wallTemperatureFlag == (int) true &&
					bcProps.surfaceProps[i].wallTemperature >= 0) {

            		if (counter > 0) fprintf(fcfg, ",");
            		fprintf(fcfg," %d, %f", bcProps.surfaceProps[i].bcID, bcProps.surfaceProps[i].wallTemperature);

            		counter += 1;
            	}
            }

            if(counter == 0) fprintf(fcfg," NONE");
			fprintf(fcfg," )\n");

            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Far-field boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"MARKER_FAR= (" );

            counter = 0; // Farfield boundary
            for (i = 0; i < bcProps.numBCID; i++) {
            	if (bcProps.surfaceProps[i].surfaceType == Farfield) {

            		if (counter > 0) fprintf(fcfg, ",");
            		fprintf(fcfg," %d", bcProps.surfaceProps[i].bcID);

            		counter += 1;
            	}
            }

            if(counter == 0) fprintf(fcfg," NONE");
            fprintf(fcfg," )\n");

            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Symmetry boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"MARKER_SYM= (" );

            counter = 0; // Symmetry boundary
            for (i = 0; i < bcProps.numBCID; i++) {
            	if (bcProps.surfaceProps[i].surfaceType == Symmetry) {

            		if (counter > 0) fprintf(fcfg, ",");
            		fprintf(fcfg," %d", bcProps.surfaceProps[i].bcID);

            		counter += 1;
				}
            }

            if(counter == 0) fprintf(fcfg," NONE");
            fprintf(fcfg," )\n");

            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Near-Field boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"MARKER_NEARFIELD= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Zone interface boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"MARKER_INTERFACE= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Actuator disk boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"%% Format: ( inlet face marker, outlet face marker,\n");
            fprintf(fcfg,"%% rotation_angle_x-axis, rotation_angle_y-axis, rotation_angle_z-axis,\n");
            fprintf(fcfg,"%% root radius, tip radius, pressure jump, temperature jump, rev/min,\n");
            fprintf(fcfg,"%% uniform(0)/linear(1) distribution, ... )\n");
            fprintf(fcfg,"MARKER_ACTDISK= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Inlet boundary type (TOTAL_CONDITIONS, MASS_FLOW)\n");
            fprintf(fcfg,"INLET_TYPE= TOTAL_CONDITIONS\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Inlet boundary marker(s) with the following formats (NONE = no marker) \n");
            fprintf(fcfg,"%% Total Conditions: (inlet marker, total temp, total pressure, flow_direction_x, \n");
            fprintf(fcfg,"%%           flow_direction_y, flow_direction_z, ... ) where flow_direction is\n");
            fprintf(fcfg,"%%           a unit vector.\n");
            fprintf(fcfg,"%% Mass Flow: (inlet marker, density, velocity magnitude, flow_direction_x, \n");
            fprintf(fcfg,"%%           flow_direction_y, flow_direction_z, ... ) where flow_direction is\n");
            fprintf(fcfg,"%%           a unit vector.\n");
            fprintf(fcfg,"%% Incompressible: (inlet marker, NULL, velocity magnitude, flow_direction_x,\n");
            fprintf(fcfg,"%%           flow_direction_y, flow_direction_z, ... ) where flow_direction is\n");
            fprintf(fcfg,"%%           a unit vector.\n");
            fprintf(fcfg,"MARKER_INLET= ( ");

            counter = 0; // Subsonic Inflow
            for (i = 0; i < bcProps.numBCID ; i++) {
                if (bcProps.surfaceProps[i].surfaceType == SubsonicInflow) {

                	if (counter > 0) fprintf(fcfg, ",");
                	fprintf(fcfg," %d, %f, %f, %f, %f, %f", bcProps.surfaceProps[i].bcID,
                                                            bcProps.surfaceProps[i].totalTemperature,
                                                            bcProps.surfaceProps[i].totalPressure,
                                                            bcProps.surfaceProps[i].uVelocity,
                                                            bcProps.surfaceProps[i].vVelocity,
                                                            bcProps.surfaceProps[i].wVelocity);
                	counter += 1;
                }
            }

            if(counter == 0) fprintf(fcfg," NONE");
            fprintf(fcfg," )\n");

            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Supersonic inlet boundary marker(s) (NONE = no marker) \n");
            fprintf(fcfg,"%% Format: (inlet marker, temperature, static pressure, velocity_x, \n");
            fprintf(fcfg,"%%           velocity_y, velocity_z, ... ), i.e. primitive variables specified.\n");
            fprintf(fcfg,"MARKER_SUPERSONIC_INLET= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Outlet boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"%% Format: ( outlet marker, back pressure (static), ... )\n");
            fprintf(fcfg,"MARKER_OUTLET= ( ");

            counter = 0; // Outlet boundary
            for (i = 0; i < bcProps.numBCID ; i++) {
            	if (bcProps.surfaceProps[i].surfaceType == BackPressure ||
                    bcProps.surfaceProps[i].surfaceType == SubsonicOutflow) {    

            		if (counter > 0) fprintf(fcfg, ",");
            		fprintf(fcfg," %d, %f", bcProps.surfaceProps[i].bcID,
            								bcProps.surfaceProps[i].staticPressure);

            		counter += 1;
                }
            }

            if(counter == 0) fprintf(fcfg," NONE");
            fprintf(fcfg," )\n");

            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Supersonic outlet boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"MARKER_SUPERSONIC_OUTLET= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Periodic boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"%% Format: ( periodic marker, donor marker, rotation_center_x, rotation_center_y, \n");
            fprintf(fcfg,"%% rotation_center_z, rotation_angle_x-axis, rotation_angle_y-axis, \n");
            fprintf(fcfg,"%% rotation_angle_z-axis, translation_x, translation_y, translation_z, ... )\n");
            fprintf(fcfg,"MARKER_PERIODIC= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Engine inflow boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"%% Format: (engine inflow marker, fan face Mach, ... )\n");
            fprintf(fcfg,"MARKER_ENGINE_INFLOW= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Engine bleed boundary marker(s) with the following formats (NONE = no marker)\n");
            fprintf(fcfg,"%% Format: (engine bleed marker, mass flow rate, total temp, ... )\n");
            fprintf(fcfg,"MARKER_ENGINE_BLEED= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Engine exhaust boundary marker(s) with the following formats (NONE = no marker) \n");
            fprintf(fcfg,"%% Format: (engine exhaust marker, total nozzle temp, total nozzle pressure, ... )\n");
            fprintf(fcfg,"MARKER_ENGINE_EXHAUST= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Displacement boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"%% Format: ( displacement marker, displacement value normal to the surface, ... )\n");
            fprintf(fcfg,"MARKER_NORMAL_DISPL= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Load boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"%% Format: ( load marker, force value normal to the surface, ... )\n");
            fprintf(fcfg,"MARKER_NORMAL_LOAD= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Pressure boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"%% Format: ( pressure marker )\n");
            fprintf(fcfg,"MARKER_PRESSURE= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Neumann bounday marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"MARKER_NEUMANN= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Dirichlet boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"MARKER_DIRICHLET= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Riemann boundary marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"%% Format: (marker, data kind flag, list of data)\n");
            fprintf(fcfg,"MARKER_RIEMANN= ( NONE )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Non Reflecting boundary conditions marker(s) (NONE = no marker)\n");
            fprintf(fcfg,"%% Format: (marker, data kind flag, list of data)\n");
            fprintf(fcfg,"MARKER_NRBC= ( NONE )\n");
            fprintf(fcfg,"%% ------------------------ SURFACES IDENTIFICATION ----------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Marker(s) of the surface in the surface flow solution file\n");
            fprintf(fcfg,"MARKER_PLOTTING= (" );

            counter = 0; // Surface marker
            for (i = 0; i < bcProps.numBCID ; i++) {
                if (bcProps.surfaceProps[i].surfaceType == Inviscid ||
                	bcProps.surfaceProps[i].surfaceType == Viscous) {

                	if (counter > 0) fprintf(fcfg, ",");
                	fprintf(fcfg," %d", bcProps.surfaceProps[i].bcID);

                	counter += 1;
                }
            }
            if(counter == 0) fprintf(fcfg," NONE");
            fprintf(fcfg," )\n");

            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Marker(s) of the surface where the non-dimensional coefficients are evaluated.\n");
            fprintf(fcfg,"MARKER_MONITORING= (" );

            counter = 0;
            for (i = 0; i < bcProps.numBCID ; i++) {
                if (bcProps.surfaceProps[i].surfaceType == Inviscid ||
                   	bcProps.surfaceProps[i].surfaceType == Viscous) {

                	if (counter > 0) fprintf(fcfg, ",");
                	fprintf(fcfg," %d", bcProps.surfaceProps[i].bcID);

                	counter += 1;
                }
            }

            if(counter == 0) fprintf(fcfg," NONE");
            fprintf(fcfg," )\n");

            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Marker(s) of the surface where obj. func. (design problem) will be evaluated\n");
            fprintf(fcfg,"MARKER_DESIGNING = ( NONE )\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ------------- COMMON PARAMETERS DEFINING THE NUMERICAL METHOD ---------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Numerical method for spatial gradients (GREEN_GAUSS, WEIGHTED_LEAST_SQUARES)\n");
            fprintf(fcfg,"NUM_METHOD_GRAD= GREEN_GAUSS\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% CFL number (stating value for the adaptive CFL number)\n");
            if (aimInputs[aim_getIndex(aimInfo, "CFL_NUMBER",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"CFL_NUMBER= %f\n", aimInputs[aim_getIndex(aimInfo, "CFL_NUMBER",  ANALYSISIN)-1].vals.real);
            } else {
                fprintf(fcfg,"CFL_NUMBER= 10.0\n");
            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Adaptive CFL number (NO, YES)\n");
            fprintf(fcfg,"CFL_ADAPT= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Parameters of the adaptive CFL number (factor down, factor up, CFL min value,\n");
            fprintf(fcfg,"%%                                        CFL max value )\n");
            fprintf(fcfg,"CFL_ADAPT_PARAM= ( 1.5, 0.5, 1.25, 50.0 )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Maximum Delta Time in local time stepping simulations\n");
            fprintf(fcfg,"MAX_DELTA_TIME= 1E6\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Runge-Kutta alpha coefficients\n");
            fprintf(fcfg,"RK_ALPHA_COEFF= ( 0.66667, 0.66667, 1.000000 )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Objective function in optimization problem (DRAG, LIFT, SIDEFORCE, MOMENT_X,\n");
            fprintf(fcfg,"%%                                             MOMENT_Y, MOMENT_Z, EFFICIENCY,\n");
            fprintf(fcfg,"%%                                             EQUIVALENT_AREA, NEARFIELD_PRESSURE,\n");
            fprintf(fcfg,"%%                                             FORCE_X, FORCE_Y, FORCE_Z, THRUST,\n");
            fprintf(fcfg,"%%                                             TORQUE, FREE_SURFACE, TOTAL_HEATFLUX,\n");
            fprintf(fcfg,"%%                                             MAXIMUM_HEATFLUX, INVERSE_DESIGN_PRESSURE,\n");
            fprintf(fcfg,"%%                                             INVERSE_DESIGN_HEATFLUX, AVG_TOTAL_PRESSURE, \n");
            fprintf(fcfg,"%%                                             MASS_FLOW_RATE)\n");
            fprintf(fcfg,"OBJECTIVE_FUNCTION= DRAG\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ----------------------- SLOPE LIMITER DEFINITION ----------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Reference element length for computing the slope and sharp edges \n");
            fprintf(fcfg,"%%                              limiters (0.1 m, 5.0 in by default)\n");
            fprintf(fcfg,"REF_ELEM_LENGTH= 0.1\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Coefficient for the limiter\n");
            fprintf(fcfg,"LIMITER_COEFF= 0.3\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Freeze the value of the limiter after a number of iterations\n");
            fprintf(fcfg,"LIMITER_ITER= 999999\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Coefficient for the sharp edges limiter\n");
            fprintf(fcfg,"SHARP_EDGES_COEFF= 3.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Reference coefficient (sensitivity) for detecting sharp edges.\n");
            fprintf(fcfg,"REF_SHARP_EDGES= 3.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Remove sharp edges from the sensitivity evaluation (NO, YES)\n");
            fprintf(fcfg,"SENS_REMOVE_SHARP= NO\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ------------------------ LINEAR SOLVER DEFINITION ---------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Linear solver or smoother for implicit formulations (BCGSTAB, FGMRES, SMOOTHER_JACOBI, \n");
            fprintf(fcfg,"%%                                                      SMOOTHER_ILU0, SMOOTHER_LUSGS, \n");
            fprintf(fcfg,"%%                                                      SMOOTHER_LINELET)\n");
            fprintf(fcfg,"LINEAR_SOLVER= FGMRES\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Preconditioner of the Krylov linear solver (ILU0, LU_SGS, LINELET, JACOBI)\n");
            fprintf(fcfg,"LINEAR_SOLVER_PREC= LU_SGS\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Minimum error of the linear solver for implicit formulations\n");
            fprintf(fcfg,"LINEAR_SOLVER_ERROR= 1E-4\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Max number of iterations of the linear solver for the implicit formulation\n");
            fprintf(fcfg,"LINEAR_SOLVER_ITER= 5\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% -------------------------- MULTIGRID PARAMETERS -----------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Multi-grid levels (0 = no multi-grid)\n");
            fprintf(fcfg,"MGLEVEL= %d\n", aimInputs[aim_getIndex(aimInfo, "MGLEVEL",  ANALYSISIN)-1].vals.integer);
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Multi-grid cycle (V_CYCLE, W_CYCLE, FULLMG_CYCLE)\n");
            fprintf(fcfg,"MGCYCLE= V_CYCLE\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Multi-grid pre-smoothing level\n");
            fprintf(fcfg,"MG_PRE_SMOOTH= ( 1, 2, 3, 3 )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Multi-grid post-smoothing level\n");
            fprintf(fcfg,"MG_POST_SMOOTH= ( 0, 0, 0, 0 )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Jacobi implicit smoothing of the correction\n");
            fprintf(fcfg,"MG_CORRECTION_SMOOTH= ( 0, 0, 0, 0 )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Damping factor for the residual restriction\n");
            fprintf(fcfg,"MG_DAMP_RESTRICTION= 0.75\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Damping factor for the correction prolongation\n");
            fprintf(fcfg,"MG_DAMP_PROLONGATION= 0.75\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% -------------------- FLOW NUMERICAL METHOD DEFINITION -----------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Convective numerical method (JST, LAX-FRIEDRICH, CUSP, ROE, AUSM, HLLC,\n");
            fprintf(fcfg,"%%                              TURKEL_PREC, MSW)\n");
            fprintf(fcfg,"CONV_NUM_METHOD_FLOW= JST\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Spatial numerical order integration (1ST_ORDER, 2ND_ORDER, 2ND_ORDER_LIMITER)\n");
            fprintf(fcfg,"SPATIAL_ORDER_FLOW= 2ND_ORDER_LIMITER\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Slope limiter (VENKATAKRISHNAN, BARTH_JESPERSEN)\n");
            fprintf(fcfg,"SLOPE_LIMITER_FLOW= VENKATAKRISHNAN\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Entropy fix coefficient (0.0 implies no entropy fixing, 1.0 implies scalar\n");
            fprintf(fcfg,"%%                          artificial dissipation)\n");
            fprintf(fcfg,"ENTROPY_FIX_COEFF= 0.001\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% 1st, 2nd and 4th order artificial dissipation coefficients\n");
            fprintf(fcfg,"AD_COEFF_FLOW= ( 0.15, 0.5, 0.02 )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Viscous limiter (NO, YES)\n");
            fprintf(fcfg,"VISCOUS_LIMITER_FLOW= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Time discretization (RUNGE-KUTTA_EXPLICIT, EULER_IMPLICIT, EULER_EXPLICIT)\n");
            fprintf(fcfg,"TIME_DISCRE_FLOW= EULER_IMPLICIT\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Relaxation coefficient\n");
            fprintf(fcfg,"RELAXATION_FACTOR_FLOW= 1.0\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% -------------------- TURBULENT NUMERICAL METHOD DEFINITION ------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Convective numerical method (SCALAR_UPWIND)\n");
            fprintf(fcfg,"CONV_NUM_METHOD_TURB= SCALAR_UPWIND\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Spatial numerical order integration (1ST_ORDER, 2ND_ORDER, 2ND_ORDER_LIMITER)\n");
            fprintf(fcfg,"SPATIAL_ORDER_TURB= 1ST_ORDER\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Slope limiter (VENKATAKRISHNAN)\n");
            fprintf(fcfg,"SLOPE_LIMITER_TURB= VENKATAKRISHNAN\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Viscous limiter (NO, YES)\n");
            fprintf(fcfg,"VISCOUS_LIMITER_TURB= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Time discretization (EULER_IMPLICIT)\n");
            fprintf(fcfg,"TIME_DISCRE_TURB= EULER_IMPLICIT\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Reduction factor of the CFL coefficient in the turbulence problem\n");
            fprintf(fcfg,"CFL_REDUCTION_TURB= 1.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Relaxation coefficient\n");
            fprintf(fcfg,"RELAXATION_FACTOR_TURB= 1.0\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% --------------------- HEAT NUMERICAL METHOD DEFINITION ----------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Value of the thermal diffusivity\n");
            fprintf(fcfg,"THERMAL_DIFFUSIVITY= 1.0\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ---------------- ADJOINT-FLOW NUMERICAL METHOD DEFINITION -------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Convective numerical method (JST, LAX-FRIEDRICH, ROE)\n");
            fprintf(fcfg,"CONV_NUM_METHOD_ADJFLOW= JST\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Spatial numerical order integration (1ST_ORDER, 2ND_ORDER, 2ND_ORDER_LIMITER)\n");
            fprintf(fcfg,"SPATIAL_ORDER_ADJFLOW= 2ND_ORDER\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Slope limiter (VENKATAKRISHNAN, SHARP_EDGES, WALL_DISTANCE)\n");
            fprintf(fcfg,"SLOPE_LIMITER_ADJFLOW= VENKATAKRISHNAN\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% 1st, 2nd, and 4th order artificial dissipation coefficients\n");
            fprintf(fcfg,"AD_COEFF_ADJFLOW= ( 0.15, 0.5, 0.02 )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Time discretization (RUNGE-KUTTA_EXPLICIT, EULER_IMPLICIT)\n");
            fprintf(fcfg,"TIME_DISCRE_ADJFLOW= EULER_IMPLICIT\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Relaxation coefficient\n");
            fprintf(fcfg,"RELAXATION_FACTOR_ADJFLOW= 1.0\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Reduction factor of the CFL coefficient in the adjoint problem\n");
            fprintf(fcfg,"CFL_REDUCTION_ADJFLOW= 0.8\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Limit value for the adjoint variable\n");
            fprintf(fcfg,"LIMIT_ADJFLOW= 1E6\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Multigrid adjoint problem (NO, YES)\n");
            fprintf(fcfg,"MG_ADJFLOW= YES\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ---------------- ADJOINT-TURBULENT NUMERICAL METHOD DEFINITION --------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Convective numerical method (SCALAR_UPWIND)\n");
            fprintf(fcfg,"CONV_NUM_METHOD_ADJTURB= SCALAR_UPWIND\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Spatial numerical order integration (1ST_ORDER, 2ND_ORDER, 2ND_ORDER_LIMITER)\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"SPATIAL_ORDER_ADJTURB= 1ST_ORDER\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Slope limiter (VENKATAKRISHNAN)\n");
            fprintf(fcfg,"SLOPE_LIMITER_ADJTURB= VENKATAKRISHNAN\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Time discretization (EULER_IMPLICIT)\n");
            fprintf(fcfg,"TIME_DISCRE_ADJTURB= EULER_IMPLICIT\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Reduction factor of the CFL coefficient in the adjoint turbulent problem\n");
            fprintf(fcfg,"CFL_REDUCTION_ADJTURB= 0.01\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ----------------------- GEOMETRY EVALUATION PARAMETERS ----------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Geometrical evaluation mode (FUNCTION, GRADIENT)\n");
            fprintf(fcfg,"GEO_MODE= FUNCTION\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Marker(s) of the surface where geometrical based func. will be evaluated\n");
            fprintf(fcfg,"GEO_MARKER= ( airfoil )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Number of airfoil sections\n");
            fprintf(fcfg,"GEO_NUMBER_SECTIONS= 5\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Orientation of airfoil sections (X_AXIS, Y_AXIS, Z_AXIS)\n");
            fprintf(fcfg,"GEO_ORIENTATION_SECTIONS= Y_AXIS\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Location (coordinate) of the airfoil sections (MinValue, MaxValue)\n");
            fprintf(fcfg,"GEO_LOCATION_SECTIONS= (1.5, 3.5)\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Plot loads and Cp distributions on each airfoil section\n");
            fprintf(fcfg,"GEO_PLOT_SECTIONS= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Number of section cuts to make when calculating internal volume\n");
            fprintf(fcfg,"GEO_VOLUME_SECTIONS= 101\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ------------------------- GRID ADAPTATION STRATEGY --------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Kind of grid adaptation (NONE, PERIODIC, FULL, FULL_FLOW, GRAD_FLOW,\n");
            fprintf(fcfg,"%%                          FULL_ADJOINT, GRAD_ADJOINT, GRAD_FLOW_ADJ, ROBUST,\n");
            fprintf(fcfg,"%%                          FULL_LINEAR, COMPUTABLE, COMPUTABLE_ROBUST,\n");
            fprintf(fcfg,"%%                          REMAINING, WAKE, SMOOTHING, SUPERSONIC_SHOCK)\n");
            fprintf(fcfg,"KIND_ADAPT= FULL_FLOW\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Percentage of new elements (%% of the original number of elements)\n");
            fprintf(fcfg,"NEW_ELEMS= 5\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Scale factor for the dual volume\n");
            fprintf(fcfg,"DUALVOL_POWER= 0.5\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Adapt the boundary elements (NO, YES)\n");
            fprintf(fcfg,"ADAPT_BOUNDARY= YES\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ----------------------- DESIGN VARIABLE PARAMETERS --------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Kind of deformation (TRANSLATION, ROTATION, SCALE,\n");
            fprintf(fcfg,"%%                      FFD_SETTING,\n");
            fprintf(fcfg,"%%                      FFD_CONTROL_POINT, FFD_CAMBER, FFD_THICKNESS\n");
            fprintf(fcfg,"%%                      FFD_DIHEDRAL_ANGLE, FFD_TWIST_ANGLE, FFD_ROTATION,\n");
            fprintf(fcfg,"%%                      FFD_CONTROL_POINT_2D, FFD_CAMBER_2D, FFD_THICKNESS_2D,\n");
            fprintf(fcfg,"%%                      HICKS_HENNE, PARABOLIC, NACA_4DIGITS, AIRFOIL)\n");
            fprintf(fcfg,"DV_KIND= FFD_SETTING \n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Marker of the surface in which we are going apply the shape deformation\n");
            fprintf(fcfg,"DV_MARKER= ( airfoil )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Parameters of the shape deformation\n");
            fprintf(fcfg,"%% - TRANSLATION ( x_Disp, y_Disp, z_Disp ), as a unit vector\n");
            fprintf(fcfg,"%% - ROTATION ( x_Orig, y_Orig, z_Orig, x_End, y_End, z_End )\n");
            fprintf(fcfg,"%% - SCALE ( 1.0 )\n");
            fprintf(fcfg,"%% - FFD_SETTING ( 1.0 )\n");
            fprintf(fcfg,"%% - FFD_CONTROL_POINT ( FFD_BoxTag, i_Ind, j_Ind, k_Ind, x_Disp, y_Disp, z_Disp )\n");
            fprintf(fcfg,"%% - FFD_CAMBER ( FFD_BoxTag, i_Ind, j_Ind )\n");
            fprintf(fcfg,"%% - FFD_THICKNESS ( FFD_BoxTag, i_Ind, j_Ind )\n");
            fprintf(fcfg,"%% - FFD_DIHEDRAL_ANGLE ( FFD_BoxTag, x_Orig, y_Orig, z_Orig, x_End, y_End, z_End )\n");
            fprintf(fcfg,"%% - FFD_TWIST_ANGLE ( FFD_BoxTag, x_Orig, y_Orig, z_Orig, x_End, y_End, z_End )\n");
            fprintf(fcfg,"%% - FFD_ROTATION ( FFD_BoxTag, x_Orig, y_Orig, z_Orig, x_End, y_End, z_End )\n");
            fprintf(fcfg,"%% - FFD_CONTROL_POINT_2D ( FFD_BoxTag, i_Ind, j_Ind, x_Disp, y_Disp )\n");
            fprintf(fcfg,"%% - FFD_CAMBER_2D ( FFD_BoxTag, i_Ind )\n");
            fprintf(fcfg,"%% - FFD_THICKNESS_2D ( FFD_BoxTag, i_Ind )\n");
            fprintf(fcfg,"%% - HICKS_HENNE ( Lower Surface (0)/Upper Surface (1)/Only one Surface (2), x_Loc )\n");
            fprintf(fcfg,"%% - PARABOLIC ( Center, Thickness )\n");
            fprintf(fcfg,"%% - NACA_4DIGITS ( 1st digit, 2nd digit, 3rd and 4th digit )\n");
            fprintf(fcfg,"%% - AIRFOIL ( 1.0 )\n");
            fprintf(fcfg,"DV_PARAM= ( 1, 0.5 )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Value of the shape deformation\n");
            fprintf(fcfg,"DV_VALUE= 0.01\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ------------------------ GRID DEFORMATION PARAMETERS ------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Linear solver or smoother for implicit formulations (FGMRES, RESTARTED_FGMRES, BCGSTAB)\n");
            fprintf(fcfg,"DEFORM_LINEAR_SOLVER= FGMRES\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Number of smoothing iterations for mesh deformation\n");
            fprintf(fcfg,"DEFORM_LINEAR_ITER= 500\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Number of nonlinear deformation iterations (surface deformation increments)\n");
            fprintf(fcfg,"DEFORM_NONLINEAR_ITER= 2\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Print the residuals during mesh deformation to the console (YES, NO)\n");
            fprintf(fcfg,"DEFORM_CONSOLE_OUTPUT= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Factor to multiply smallest cell volume for deform tolerance (0.001 default)\n");
            fprintf(fcfg,"DEFORM_TOL_FACTOR = 0.001\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Type of element stiffness imposed for FEA mesh deformation (INVERSE_VOLUME, \n");
            fprintf(fcfg,"%%                                          WALL_DISTANCE, CONSTANT_STIFFNESS)\n");
            fprintf(fcfg,"DEFORM_STIFFNESS_TYPE= INVERSE_VOLUME\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Visualize the deformation (NO, YES)\n");
            fprintf(fcfg,"VISUALIZE_DEFORMATION= NO\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% -------------------- FREE-FORM DEFORMATION PARAMETERS -----------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Tolerance of the Free-Form Deformation point inversion\n");
            fprintf(fcfg,"FFD_TOLERANCE= 1E-10\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Maximum number of iterations in the Free-Form Deformation point inversion\n");
            fprintf(fcfg,"FFD_ITERATIONS= 500\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% FFD box definition: 3D case (FFD_BoxTag, X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3, X4, Y4, Z4,\n");
            fprintf(fcfg,"%%                              X5, Y5, Z5, X6, Y6, Z6, X7, Y7, Z7, X8, Y8, Z8)\n");
            fprintf(fcfg,"%%                     2D case (FFD_BoxTag, X1, Y1, 0.0, X2, Y2, 0.0, X3, Y3, 0.0, X4, Y4, 0.0,\n");
            fprintf(fcfg,"%%                              0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)\n");
            fprintf(fcfg,"FFD_DEFINITION= (MAIN_BOX, 0.5, 0.25, -0.25, 1.5, 0.25, -0.25, 1.5, 0.75, -0.25, 0.5, 0.75, -0.25, 0.5, 0.25, 0.25, 1.5, 0.25, 0.25, 1.5, 0.75, 0.25, 0.5, 0.75, 0.25)\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% FFD box degree: 3D case (x_degree, y_degree, z_degree)\n");
            fprintf(fcfg,"%%                 2D case (x_degree, y_degree, 0)\n");
            fprintf(fcfg,"FFD_DEGREE= (10, 10, 1)\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Surface continuity at the intersection with the FFD (1ST_DERIVATIVE, 2ND_DERIVATIVE)\n");
            fprintf(fcfg,"FFD_CONTINUITY= 2ND_DERIVATIVE\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% --------------------------- CONVERGENCE PARAMETERS --------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Number of total iterations\n");
            if (aimInputs[aim_getIndex(aimInfo, "Num_Iter",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"EXT_ITER= %d\n", aimInputs[aim_getIndex(aimInfo, "Num_Iter",  ANALYSISIN)-1].vals.integer);
            } else {
                fprintf(fcfg,"EXT_ITER= 9999\n");

            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Convergence criteria (CAUCHY, RESIDUAL)\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"CONV_CRITERIA= RESIDUAL\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Residual reduction (order of magnitude with respect to the initial value)\n");
            if (aimInputs[aim_getIndex(aimInfo, "RESIDUAL_REDUCTION",  ANALYSISIN)-1].nullVal == NotNull) {
                fprintf(fcfg,"RESIDUAL_REDUCTION= %d\n", aimInputs[aim_getIndex(aimInfo, "RESIDUAL_REDUCTION",  ANALYSISIN)-1].vals.integer);
            } else {
                fprintf(fcfg,"RESIDUAL_REDUCTION= 5\n");

            }
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Min value of the residual (log10 of the residual)\n");
            fprintf(fcfg,"RESIDUAL_MINVAL= -8\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Start convergence criteria at iteration number\n");
            fprintf(fcfg,"STARTCONV_ITER= 10\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Number of elements to apply the criteria\n");
            fprintf(fcfg,"CAUCHY_ELEMS= 100\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Epsilon to control the series convergence\n");
            fprintf(fcfg,"CAUCHY_EPS= 1E-10\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Direct function to apply the convergence criteria (LIFT, DRAG, NEARFIELD_PRESS)\n");
            fprintf(fcfg,"CAUCHY_FUNC_FLOW= DRAG\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Adjoint function to apply the convergence criteria (SENS_GEOMETRY, SENS_MACH)\n");
            fprintf(fcfg,"CAUCHY_FUNC_ADJFLOW= SENS_GEOMETRY\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% ------------------------- INPUT/OUTPUT INFORMATION --------------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Mesh input file\n");
            fprintf(fcfg,"MESH_FILENAME= %s.su2\n", aimInputs[aim_getIndex(aimInfo, "Proj_Name",  ANALYSISIN)-1].vals.string); 
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Mesh input file format (SU2, CGNS)\n");
            fprintf(fcfg,"MESH_FORMAT= SU2\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Mesh output file\n");
            fprintf(fcfg,"MESH_OUT_FILENAME= mesh_out.su2\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Restart flow input file\n");
            fprintf(fcfg,"SOLUTION_FLOW_FILENAME= solution_flow.dat\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Restart adjoint input file\n");
            fprintf(fcfg,"SOLUTION_ADJ_FILENAME= solution_adj.dat\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Output file format (TECPLOT, TECPLOT_BINARY, PARAVIEW,\n");
            fprintf(fcfg,"%%                     FIELDVIEW, FIELDVIEW_BINARY)\n");
            fprintf(fcfg,"OUTPUT_FORMAT= %s\n", aimInputs[aim_getIndex(aimInfo, "OUTPUT_FORMAT",  ANALYSISIN)-1].vals.string);
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Output file convergence history (w/o extension) \n");
            fprintf(fcfg,"CONV_FILENAME= history_%s\n", aimInputs[aim_getIndex(aimInfo, "Proj_Name",  ANALYSISIN)-1].vals.string); 
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Output file with the forces breakdown\n");
            fprintf(fcfg,"BREAKDOWN_FILENAME= forces_breakdown_%s.dat\n", aimInputs[aim_getIndex(aimInfo, "Proj_Name",  ANALYSISIN)-1].vals.string); 
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Output file restart flow\n");
            fprintf(fcfg,"RESTART_FLOW_FILENAME= restart_flow_%s.dat\n", aimInputs[aim_getIndex(aimInfo, "Proj_Name",  ANALYSISIN)-1].vals.string); 
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Output file restart adjoint\n");
            fprintf(fcfg,"RESTART_ADJ_FILENAME= restart_adj.dat\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Output file flow (w/o extension) variables\n");
            fprintf(fcfg,"VOLUME_FLOW_FILENAME= flow_%s\n", aimInputs[aim_getIndex(aimInfo, "Proj_Name",  ANALYSISIN)-1].vals.string); 
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Output file adjoint (w/o extension) variables\n");
            fprintf(fcfg,"VOLUME_ADJ_FILENAME= adjoint\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Output Objective function\n");
            fprintf(fcfg,"VALUE_OBJFUNC_FILENAME= of_eval.dat\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Output objective function gradient (using continuous adjoint)\n");
            fprintf(fcfg,"GRAD_OBJFUNC_FILENAME= of_grad.dat\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Output file surface flow coefficient (w/o extension)\n");
            fprintf(fcfg,"SURFACE_FLOW_FILENAME= surface_flow_%s\n", aimInputs[aim_getIndex(aimInfo, "Proj_Name",  ANALYSISIN)-1].vals.string); 
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Output file surface adjoint coefficient (w/o extension)\n");
            fprintf(fcfg,"SURFACE_ADJ_FILENAME= surface_adjoint\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Writing solution file frequency\n");
            fprintf(fcfg,"WRT_SOL_FREQ= 1000\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Writing solution file frequency for physical time steps (dual time)\n");
            fprintf(fcfg,"WRT_SOL_FREQ_DUALTIME= 1\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Writing convergence history frequency\n");
            fprintf(fcfg,"WRT_CON_FREQ= 1\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Writing convergence history frequency (dual time, only written to screen)\n");
            fprintf(fcfg,"WRT_CON_FREQ_DUALTIME= 10\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Output residual values in the solution files\n");
            fprintf(fcfg,"WRT_RESIDUALS= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Output limiters values in the solution files\n");
            fprintf(fcfg,"WRT_LIMITERS= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Output the sharp edges detector\n");
            fprintf(fcfg,"WRT_SHARPEDGES= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Minimize the required output memory\n");
            fprintf(fcfg,"LOW_MEMORY_OUTPUT= NO\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Verbosity of console output: NONE removes minor MPI overhead (NONE, HIGH)\n");
            fprintf(fcfg,"CONSOLE_OUTPUT_VERBOSITY= HIGH\n");
            fprintf(fcfg,"\n");
            fprintf(fcfg,"%% --------------------- OPTIMAL SHAPE DESIGN DEFINITION -----------------------%%\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Available flow based objective functions or constraint functions\n");
            fprintf(fcfg,"%%    DRAG, LIFT, SIDEFORCE, EFFICIENCY,\n");
            fprintf(fcfg,"%%    FORCE_X, FORCE_Y, FORCE_Z,\n");
            fprintf(fcfg,"%%    MOMENT_X, MOMENT_Y, MOMENT_Z,\n");
            fprintf(fcfg,"%%    THRUST, TORQUE, FIGURE_OF_MERIT,\n");
            fprintf(fcfg,"%%    EQUIVALENT_AREA, NEARFIELD_PRESSURE, \n");
            fprintf(fcfg,"%%    TOTAL_HEATFLUX, MAXIMUM_HEATFLUX,\n");
            fprintf(fcfg,"%%    INVERSE_DESIGN_PRESSURE, INVERSE_DESIGN_HEATFLUX,\n");
            fprintf(fcfg,"%%    FREE_SURFACE, AVG_TOTAL_PRESSURE, MASS_FLOW_RATE\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Available geometrical based objective functions or constraint functions\n");
            fprintf(fcfg,"%%    MAX_THICKNESS, 1/4_THICKNESS, 1/2_THICKNESS, 3/4_THICKNESS, AREA, AOA, CHORD, \n");
            fprintf(fcfg,"%%    MAX_THICKNESS_SEC1, MAX_THICKNESS_SEC2, MAX_THICKNESS_SEC3, MAX_THICKNESS_SEC4, MAX_THICKNESS_SEC5, \n");
            fprintf(fcfg,"%%    1/4_THICKNESS_SEC1, 1/4_THICKNESS_SEC2, 1/4_THICKNESS_SEC3, 1/4_THICKNESS_SEC4, 1/4_THICKNESS_SEC5, \n");
            fprintf(fcfg,"%%    1/2_THICKNESS_SEC1, 1/2_THICKNESS_SEC2, 1/2_THICKNESS_SEC3, 1/2_THICKNESS_SEC4, 1/2_THICKNESS_SEC5, \n");
            fprintf(fcfg,"%%    3/4_THICKNESS_SEC1, 3/4_THICKNESS_SEC2, 3/4_THICKNESS_SEC3, 3/4_THICKNESS_SEC4, 3/4_THICKNESS_SEC5, \n");
            fprintf(fcfg,"%%    AREA_SEC1, AREA_SEC2, AREA_SEC3, AREA_SEC4, AREA_SEC5, \n");
            fprintf(fcfg,"%%    AOA_SEC1, AOA_SEC2, AOA_SEC3, AOA_SEC4, AOA_SEC5, \n");
            fprintf(fcfg,"%%    CHORD_SEC1, CHORD_SEC2, CHORD_SEC3, CHORD_SEC4, CHORD_SEC5\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Available design variables\n");
            fprintf(fcfg,"%%    HICKS_HENNE 	(  1, Scale | Mark. List | Lower(0)/Upper(1) side, x_Loc )\n");
            fprintf(fcfg,"%%    COSINE_BUMP	(  2, Scale | Mark. List | Lower(0)/Upper(1) side, x_Loc, x_Size )\n");
            fprintf(fcfg,"%%    SPHERICAL		(  3, Scale | Mark. List | ControlPoint_Index, Theta_Disp, R_Disp )\n");
            fprintf(fcfg,"%%    NACA_4DIGITS	(  4, Scale | Mark. List | 1st digit, 2nd digit, 3rd and 4th digit )\n");
            fprintf(fcfg,"%%    DISPLACEMENT	(  5, Scale | Mark. List | x_Disp, y_Disp, z_Disp )\n");
            fprintf(fcfg,"%%    ROTATION		(  6, Scale | Mark. List | x_Axis, y_Axis, z_Axis, x_Turn, y_Turn, z_Turn )\n");
            fprintf(fcfg,"%%    FFD_CONTROL_POINT	(  7, Scale | Mark. List | FFD_BoxTag, i_Ind, j_Ind, k_Ind, x_Mov, y_Mov, z_Mov )\n");
            fprintf(fcfg,"%%    FFD_DIHEDRAL_ANGLE	(  8, Scale | Mark. List | FFD_BoxTag, x_Orig, y_Orig, z_Orig, x_End, y_End, z_End )\n");
            fprintf(fcfg,"%%    FFD_TWIST_ANGLE 	(  9, Scale | Mark. List | FFD_BoxTag, x_Orig, y_Orig, z_Orig, x_End, y_End, z_End )\n");
            fprintf(fcfg,"%%    FFD_ROTATION 	( 10, Scale | Mark. List | FFD_BoxTag, x_Orig, y_Orig, z_Orig, x_End, y_End, z_End )\n");
            fprintf(fcfg,"%%    FFD_CAMBER 	( 11, Scale | Mark. List | FFD_BoxTag, i_Ind, j_Ind )\n");
            fprintf(fcfg,"%%    FFD_THICKNESS 	( 12, Scale | Mark. List | FFD_BoxTag, i_Ind, j_Ind )\n");
            fprintf(fcfg,"%%    FOURIER 		( 14, Scale | Mark. List | Lower(0)/Upper(1) side, index, cos(0)/sin(1) )\n");
            fprintf(fcfg,"%%    FFD_CONTROL_POINT_2D (  15, Scale | Mark. List | FFD_BoxTag, i_Ind, j_Ind, x_Mov, y_Mov )\n");
            fprintf(fcfg,"%%    FFD_CAMBER_2D 	( 16, Scale | Mark. List | FFD_BoxTag, i_Ind )\n");
            fprintf(fcfg,"%%    FFD_THICKNESS_2D 	( 17, Scale | Mark. List | FFD_BoxTag, i_Ind )\n");
            fprintf(fcfg,"%%    FFD_CONTROL_SURFACE 	( 18, Scale | Mark. List | FFD_BoxTag, x_Orig, y_Orig, z_Orig, x_End, y_End, z_End )\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Optimization objective function with scaling factor\n");
            fprintf(fcfg,"%% ex= Objective * Scale\n");
            fprintf(fcfg,"OPT_OBJECTIVE= DRAG * 0.001\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Optimization constraint functions with scaling factors, separated by semicolons\n");
            fprintf(fcfg,"%% ex= (Objective = Value ) * Scale, use '>','<','='\n");
            fprintf(fcfg,"OPT_CONSTRAINT= ( LIFT > 0.328188 ) * 0.001; ( MOMENT_Z > 0.034068 ) * 0.001; ( MAX_THICKNESS > 0.11 ) * 0.001\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Maximum number of iterations\n");
            fprintf(fcfg,"OPT_ITERATIONS= 100\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Requested accuracy\n");
            fprintf(fcfg,"OPT_ACCURACY= 1E-6\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Upper bound for each design variable\n");
            fprintf(fcfg,"OPT_BOUND_UPPER= 0.1\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Lower bound for each design variable\n");
            fprintf(fcfg,"OPT_BOUND_LOWER= -0.1\n");
            fprintf(fcfg,"%%\n");
            fprintf(fcfg,"%% Optimization design variables, separated by semicolons\n");
            fprintf(fcfg,"DEFINITION_DV= ( 1, 1.0 | airfoil | 0, 0.05 ); ( 1, 1.0 | airfoil | 0, 0.10 ); ( 1, 1.0 | airfoil | 0, 0.15 ); ( 1, 1.0 | airfoil | 0, 0.20 ); ( 1, 1.0 | airfoil | 0, 0.25 ); ( 1, 1.0 | airfoil | 0, 0.30 ); ( 1, 1.0 | airfoil | 0, 0.35 ); ( 1, 1.0 | airfoil | 0, 0.40 ); ( 1, 1.0 | airfoil | 0, 0.45 ); ( 1, 1.0 | airfoil | 0, 0.50 ); ( 1, 1.0 | airfoil | 0, 0.55 ); ( 1, 1.0 | airfoil | 0, 0.60 ); ( 1, 1.0 | airfoil | 0, 0.65 ); ( 1, 1.0 | airfoil | 0, 0.70 ); ( 1, 1.0 | airfoil | 0, 0.75 ); ( 1, 1.0 | airfoil | 0, 0.80 ); ( 1, 1.0 | airfoil | 0, 0.85 ); ( 1, 1.0 | airfoil | 0, 0.90 ); ( 1, 1.0 | airfoil | 0, 0.95 ); ( 1, 1.0 | airfoil | 1, 0.05 ); ( 1, 1.0 | airfoil | 1, 0.10 ); ( 1, 1.0 | airfoil | 1, 0.15 ); ( 1, 1.0 | airfoil | 1, 0.20 ); ( 1, 1.0 | airfoil | 1, 0.25 ); ( 1, 1.0 | airfoil | 1, 0.30 ); ( 1, 1.0 | airfoil | 1, 0.35 ); ( 1, 1.0 | airfoil | 1, 0.40 ); ( 1, 1.0 | airfoil | 1, 0.45 ); ( 1, 1.0 | airfoil | 1, 0.50 ); ( 1, 1.0 | airfoil | 1, 0.55 ); ( 1, 1.0 | airfoil | 1, 0.60 ); ( 1, 1.0 | airfoil | 1, 0.65 ); ( 1, 1.0 | airfoil | 1, 0.70 ); ( 1, 1.0 | airfoil | 1, 0.75 ); ( 1, 1.0 | airfoil | 1, 0.80 ); ( 1, 1.0 | airfoil | 1, 0.85 ); ( 1, 1.0 | airfoil | 1, 0.90 ); ( 1, 1.0 | airfoil | 1, 0.95 )\n");

            // Close file
            fclose(fcfg);
        }
    }

    status = destroy_cfdBCsStruct(&bcProps);
    if (status != CAPS_SUCCESS) return status;

    // If data transfer is ok ....
    if (su2Instance[iIndex].dataTransferCheck == (int) true) {

    	//See if we have data transfer information
    	status = aim_getBounds(aimInfo, &numTransferName, &transferName);
    	if (status == CAPS_SUCCESS) {
    		foundDisplacement = (int) false;
    		for (i = 0; i < numTransferName; i++) {

    			status = aim_getDiscr(aimInfo, transferName[i], &dataTransferDiscreteObj);
    			if (status != CAPS_SUCCESS) continue;

    			status = aim_getDataSet(dataTransferDiscreteObj,
    									"Displacement",
										&dataTransferMethod,
										&numDataTransferPoint,
										&dataTransferRank,
										&dataTransferData);

    			if (status == CAPS_SUCCESS) { // If we do have data ready how many elements there are

    				foundDisplacement = (int) true;

    			} else {

    				continue; // If no elements in this object skip to next transfer name
    			}

    			if (dataTransferRank != 3) {
    				printf("Displacement transfer data found however rank is %d not 3!!!!\n", dataTransferRank);
    				if (transferName != NULL) EG_free(transferName);
    				return CAPS_BADRANK;
    			}

    			// Ok we found at least 1 discplacement transfer variable - let break so we can allocate some arrays
    			if (foundDisplacement == (int) true) break;

    		} // End dataTransferDiscreteObj loop

    		// Ok looks like we have displacements to get
    		if (foundDisplacement == (int) true) {

    			// Right now we are just going to output a body containing all surface nodes
    			numOutDataPoint    = volumeMesh.surfaceMesh.numNodes;
    			numOutDataConnect  = volumeMesh.surfaceMesh.numTriFace + volumeMesh.surfaceMesh.numQuadFace;

    			// Allocate data arrays that are going to be output
				dataOutMatrix = (double **) EG_alloc(numOutVariable*sizeof(double));
				dataOutFormat = (int *) EG_alloc(numOutVariable*sizeof(int));
				dataConnectMatrix = (int *) EG_alloc(4*numOutDataConnect*sizeof(int));

				if (dataOutMatrix == NULL || dataOutFormat == NULL || dataConnectMatrix == NULL) {
					if (transferName != NULL) EG_free(transferName);
					if (dataOutMatrix != NULL) EG_free(dataOutMatrix);
					if (dataOutFormat != NULL) EG_free(dataOutFormat);
					if (dataConnectMatrix != NULL) EG_free(dataConnectMatrix);
					return EGADS_MALLOC;
				}


				for (i = 0; i < numOutVariable; i++) {

					dataOutMatrix[i] = (double *) EG_alloc(numOutDataPoint*sizeof(double));

					if (dataOutMatrix[i] == NULL) { // If allocation failed ....
						for (j = 0; j < numOutVariable; j++) {

							if (dataOutMatrix[j] != NULL ) EG_free(dataOutMatrix[j]);
						}

						if (dataOutMatrix != NULL) EG_free(dataOutMatrix);
						if (transferName != NULL) EG_free(transferName);

						return EGADS_MALLOC;
					}
				}

				// Set data out formatting
				for (i = 0; i < numOutVariable; i++) {
					if (strcasecmp(dataOutName[i], "id") == 0) {
						dataOutFormat[i] = (int) Integer;
					} else {
						dataOutFormat[i] = (int) Double;
					}
				}

				// Fill data out matrix with current surface mesh and global id
				for (i = 0; i < numOutDataPoint; i++ ) {

					// Coordinates
					dataOutMatrix[0][i] = volumeMesh.surfaceMesh.xyz[3*i + 0];
					dataOutMatrix[1][i] = volumeMesh.surfaceMesh.xyz[3*i + 1];
					dataOutMatrix[2][i] = volumeMesh.surfaceMesh.xyz[3*i + 2];

					// Global ID
					dataOutMatrix[3][i] = (double) volumeMesh.surfaceMesh.nodeLocal2Global[i];

					// Delta displacements
					dataOutMatrix[4][i] = 0;
					dataOutMatrix[5][i] = 0;
					dataOutMatrix[6][i] = 0;
				}

				// Fill out connectivity - Tri
				for (i = 0; i < volumeMesh.surfaceMesh.numTriFace; i++) {
					dataConnectMatrix[4*i+ 0] = volumeMesh.surfaceMesh.localTriFaceList[3*i + 0];
					dataConnectMatrix[4*i+ 1] = volumeMesh.surfaceMesh.localTriFaceList[3*i + 1];
					dataConnectMatrix[4*i+ 2] = volumeMesh.surfaceMesh.localTriFaceList[3*i + 2];
					dataConnectMatrix[4*i+ 3] = volumeMesh.surfaceMesh.localTriFaceList[3*i + 2];
				}

				for (i = 0; i < volumeMesh.surfaceMesh.numQuadFace; i++) {
					j = i + volumeMesh.surfaceMesh.numTriFace;
					dataConnectMatrix[4*j+ 0] = volumeMesh.surfaceMesh.localQuadFaceList[4*i + 0];
					dataConnectMatrix[4*j+ 1] = volumeMesh.surfaceMesh.localQuadFaceList[4*i + 1];
					dataConnectMatrix[4*j+ 2] = volumeMesh.surfaceMesh.localQuadFaceList[4*i + 2];
					dataConnectMatrix[4*j+ 3] = volumeMesh.surfaceMesh.localQuadFaceList[4*i + 3];
				}
				// Re-loop through transfers
				for (i = 0; i < numTransferName; i++) {

					status = aim_getDiscr(aimInfo, transferName[i], &dataTransferDiscreteObj);
					if (status != CAPS_SUCCESS) continue;

					status = aim_getDataSet(dataTransferDiscreteObj,
											"Displacement",
											&dataTransferMethod,
											&numDataTransferPoint,
											&dataTransferRank,
											&dataTransferData);

					//if (status != CAPS_SUCCESS) continue; // If no elements in this object skip to next transfer name

					//Get extra node information stored in the discrObj
					storage = (int *) dataTransferDiscreteObj->ptrm;
					nodeMap = &storage[3*dataTransferDiscreteObj->nElems + dataTransferDiscreteObj->nPoints]; // Global indexing on the body
					capsGroupList = &storage[3*dataTransferDiscreteObj->nElems + dataTransferDiscreteObj->nPoints + dataTransferDiscreteObj->nPoints]; // List of boundary ID (attrMap) in the transfer

					for (j = 0; j < numDataTransferPoint; j++) {

						globalNodeID = nodeMap[j];
						for (k = 0; k < numOutDataPoint; k++) {

							// If the global node IDs match store the displacement values in the dataOutMatrix
							if (globalNodeID  == (int) dataOutMatrix[3][k]) {

								// A rank of 3 should have already been checked
								// Delta displacements
								dataOutMatrix[4][k] = dataTransferData[3*j+0];
								dataOutMatrix[5][k] = dataTransferData[3*j+1];
								dataOutMatrix[6][k] = dataTransferData[3*j+2];
								break;
							}
						}
					}
				} // End dataTransferDiscreteObj loop

				// Update surface coordinates based on displacements
				for (i = 0; i < numOutDataPoint; i++ ) {

					// Coordinates + displacements at nodes
					dataOutMatrix[0][i] = dataOutMatrix[0][i] + dataOutMatrix[4][i]; // x
					dataOutMatrix[1][i] = dataOutMatrix[1][i] + dataOutMatrix[5][i]; // y
					dataOutMatrix[2][i] = dataOutMatrix[2][i] + dataOutMatrix[6][i]; // z

				}

				if (filename != NULL) EG_free(filename);

				filename = EG_alloc(MXCHAR +1);
				if (filename == NULL) {
					if (transferName != NULL) EG_free(transferName);
					return EGADS_MALLOC;
				}

				strcpy(filename, analysisPath);
				strcat(filename, "/");
				strcat(filename, su2Instance[iIndex].projectName);
				strcat(filename,"_body1.dat");

				// Write out discplacement in tecplot file
				status = fun3d_writeTecplotFEPOINT(filename,
				                                   "SU2 blah blah",
				    							   numOutVariable,
												   (char **) dataOutName,
												   numOutDataPoint,
												   dataOutMatrix,
												   dataOutFormat,
												   numOutDataConnect, // numConnectivity
												   dataConnectMatrix, // connectivity matrix
												   NULL); // Solution time


				// Clean-up
				for (i = 0; i < numOutVariable; i++) {
					if (dataOutMatrix[i] != NULL)  EG_free(dataOutMatrix[i]);
				}
				if (dataOutMatrix != NULL) EG_free(dataOutMatrix);
				if (dataOutFormat != NULL) EG_free(dataOutFormat);
				if (dataConnectMatrix != NULL) EG_free(dataConnectMatrix);

				if (filename != NULL) EG_free(filename);

    		} // End if found displacements

    	} // End bounds loop

    	if (transferName != NULL) EG_free(transferName);

    } // End if data transfer ok

    return CAPS_SUCCESS;

    goto cleanup;

    cleanup:

		if (transferName != NULL) EG_free(transferName);
		if (filename != NULL) EG_free(filename);

		(void) destroy_cfdBCsStruct(&bcProps);
		// (void) destroy_modalAeroelasticStruct(&modalAeroelastic);

		(void) destroy_bndCondStruct(&bndConds);

		// Clean up the volume mesh that was created for 2D mode
		if (aimInputs[aim_getIndex(aimInfo, "Two_Dimensional", ANALYSISIN)-1].vals.integer == (int) true ) {

			// Destroy volume mesh since we created here instead of inheriting it
			(void) destroy_genUnstrMeshStruct(&volumeMesh);
		}

		return status;

}


int aimOutputs(int iIndex, int index, char **aoname, capsValue *form)
{
	// SU2 Outputs
    /*! \page aimOutputsSU2 Post Analysis AIM Outputs 
     * After successful completion, SU2 writes results in various files. 
     * The data from these files can be directly viewed, visulaized, and or
     * used for further postprocessing.
     * 
     * One of the files is ("forces_breakdown.dat") which summarizes 
     * convergence including flow properties, numerical paramters, and 
     * resulting force and moment values. As an AIM output, this file is 
     * parsed for force and moment coefficients, and printed as closing 
     * remarks.
     */


	int numOutVars = 9; // Grouped

	#ifdef DEBUG
        printf(" su2AIM/aimOutputs instance = %d  index = %d!\n", iIndex, index);
    #endif

    // Total Forces - Pressure + Viscous
    if 		(index == 1) *aoname = EG_strdup("CLtot");
    else if (index == 2) *aoname = EG_strdup("CDtot");
    else if (index == 3) *aoname = EG_strdup("CSFtot");
    else if (index == 4) *aoname = EG_strdup("CMXtot");
    else if (index == 5) *aoname = EG_strdup("CMYtot");
    else if (index == 6) *aoname = EG_strdup("CMZtot");
    else if (index == 7) *aoname = EG_strdup("CXtot");
    else if (index == 8) *aoname = EG_strdup("CYtot");
    else if (index == 9) *aoname = EG_strdup("CZtot");

    // Pressure Forces
    else if (index == 1 + numOutVars) *aoname = EG_strdup("CLtot_p");
    else if (index == 2 + numOutVars) *aoname = EG_strdup("CDtot_p");
    else if (index == 3 + numOutVars) *aoname = EG_strdup("CSFtot_p");
    else if (index == 4 + numOutVars) *aoname = EG_strdup("CMXtot_p");
    else if (index == 5 + numOutVars) *aoname = EG_strdup("CMYtot_p");
    else if (index == 6 + numOutVars) *aoname = EG_strdup("CMZtot_p");
    else if (index == 7 + numOutVars) *aoname = EG_strdup("CXtot_p");
    else if (index == 8 + numOutVars) *aoname = EG_strdup("CYtot_p");
    else if (index == 9 + numOutVars) *aoname = EG_strdup("CZtot_p");

    // Viscous Forces
    else if (index == 1 + 2*numOutVars) *aoname = EG_strdup("CLtot_v");
    else if (index == 2 + 2*numOutVars) *aoname = EG_strdup("CDtot_v");
    else if (index == 3 + 2*numOutVars) *aoname = EG_strdup("CSFtot_v");
    else if (index == 4 + 2*numOutVars) *aoname = EG_strdup("CMXtot_v");
    else if (index == 5 + 2*numOutVars) *aoname = EG_strdup("CMYtot_v");
    else if (index == 6 + 2*numOutVars) *aoname = EG_strdup("CMZtot_v");
    else if (index == 7 + 2*numOutVars) *aoname = EG_strdup("CXtot_v");
    else if (index == 8 + 2*numOutVars) *aoname = EG_strdup("CYtot_v");
    else if (index == 9 + 2*numOutVars) *aoname = EG_strdup("CZtot_v");

    //if (index <= 3*numOutVars) {
    form->type   = Double;
    form->dim    = 1;
    form->length = 1;
    form->nrow   = 1;
    form->ncol   = 1;
    form->units  = NULL;
    //}

    return CAPS_SUCCESS;
}

// Calculate SU2 output
int aimCalcOutput(int iIndex, void *aimInfo, char *analysisPath,
              	  int index, capsValue *val, capsErrs **errors)
{
	size_t linecap = 0;

	char currentPath[PATH_MAX]; // Current directory path
	char *strKeyword = NULL, *bndSectionKeyword = NULL, *bndSubSectionKeyword = NULL; // Keyword strings

	char *strValue;    // Holder string for value of keyword
	char *line = NULL; // Temporary line holder

	char *filename = NULL; // File to open
	char fileExtension[] = ".dat";
	char fileSuffix[] = "forces_breakdown_";

	int bndSectionFound = (int) false;

	FILE *fp; // File pointer

	int numOutVars = 9; // Grouped

	#ifdef DEBUG
		printf(" su2AIM/aimCalcOutput instance = %d  index = %d!\n", iIndex, index);
	#endif

	*errors        = NULL;
	val->vals.real = 0.0; // Set default value

	//printf("index = %d\n",index);

	// Set the "search" string(s)
	if (index <= numOutVars) {
		bndSectionKeyword  = (char *) "Forces";
	}

	if      (index == 1 || index == 1 + numOutVars || index == 1 + 2*numOutVars) strKeyword = (char *) "CL:";
	else if (index == 2 || index == 2 + numOutVars || index == 2 + 2*numOutVars) strKeyword = (char *) "CD:";
	else if (index == 3 || index == 3 + numOutVars || index == 3 + 2*numOutVars) strKeyword = (char *) "CSF:";
	else if (index == 4 || index == 4 + numOutVars || index == 4 + 2*numOutVars) strKeyword = (char *) "CMx:";
	else if (index == 5 || index == 5 + numOutVars || index == 5 + 2*numOutVars) strKeyword = (char *) "CMy:";
	else if (index == 6 || index == 6 + numOutVars || index == 6 + 2*numOutVars) strKeyword = (char *) "CMz:";
	else if (index == 7 || index == 7 + numOutVars || index == 7 + 2*numOutVars) strKeyword = (char *) "CFx:";
	else if (index == 8 || index == 8 + numOutVars || index == 8 + 2*numOutVars) strKeyword = (char *) "CFy:";
	else if (index == 9 || index == 9 + numOutVars || index == 9 + 2*numOutVars) strKeyword = (char *) "CFz:";
	else {
		printf("Unrecognized output variable index - %d\n", index);
		return CAPS_BADINDEX;
	}

	(void) getcwd(currentPath, PATH_MAX); // Get our current path

	// Set path to analysis directory
	if (chdir(analysisPath) != 0) {
		#ifdef DEBUG
			printf(" su2AIM/aimCalcOutput Cannot chdir to %s!\n", analysisPath);
		#endif

		return CAPS_DIRERR;
	}

	// Open SU2 force file
	filename = (char *) EG_alloc((strlen(su2Instance[iIndex].projectName) + strlen(fileSuffix) + strlen(fileExtension) +1)*sizeof(char));
	if (filename == NULL) return EGADS_MALLOC;

	sprintf(filename, "%s%s%s", fileSuffix, su2Instance[iIndex].projectName, fileExtension);

	fp = fopen(filename, "r");
	EG_free(filename); // Free filename allocation


	if (fp == NULL) {
		#ifdef DEBUG
			printf(" su2AIM/aimCalcOutput Cannot open Output file!\n");
		#endif

		chdir(currentPath);

		return CAPS_IOERR;
	} else {
		#ifdef DEBUG
	        printf(" Found force file: %s%s%s", fileSuffix, su2Instance[iIndex].projectName, fileExtension);
        #endif
    }

	// Scan the file for the string
	strValue = NULL;
    while( !feof(fp) ) {
        char str[80];

        fscanf(fp, "%s", str);
        if( strcmp(str,strKeyword) == 0) {

	 		strValue = strstr(str, strKeyword);

            fscanf(fp, "%s", str);
		    
            #ifdef DEBUG
                printf("%s\n", str);
            #endif

            sscanf(str, "%lf", &val->vals.real);

            #ifdef DEBUG
                printf("  found %s = %f\n", strKeyword, val->vals.real);
            #endif

            break;
        }
    }

	// Restore the path we came in with
	chdir(currentPath);
	fclose(fp);

	#ifdef DEBUG
        printf(" Done\n Closing force file.\n");
    #endif
    
	if (strValue == NULL) {
		#ifdef DEBUG
			printf(" su2AIM/aimCalcOutput Cannot find %s in Output file!\n", strKeyword);
		#endif

		return CAPS_NOTFOUND;
	}

	return CAPS_SUCCESS;
}

/************************************************************************/
// CAPS transferring functions

int aimFreeDiscr(capsDiscr *discr)
{
    int iIndex, i;

    int status; // Function return

    #ifdef DEBUG
        if (discr == NULL) {
            printf(" su2AIM/aimFreeDiscr w/ NULL!\n");
        } else {
            printf(" su2AIM/aimFreeDiscr instance = %d!\n", discr->instance);
        }
    #endif

    if (discr == NULL) {

        // Clean up su2Instance data
        for ( iIndex = 0; iIndex < numInstance; iIndex++) {

            printf(" Cleaning up su2Instance - %d\n", iIndex);

            // Attribute to index map
            status = destroy_mapAttrToIndexStruct(&su2Instance[iIndex].attrMap);
            if (status != CAPS_SUCCESS) return status;

            // SU2 project name
            su2Instance[iIndex].projectName = NULL;

            // Analysis file path/directory
            su2Instance[iIndex].analysisPath = NULL;
        }

        if (su2Instance != NULL) EG_free(su2Instance);
        su2Instance = NULL;
        numInstance = 0;

    } else {

    	// Free up this capsDiscr

		discr->nPoints = 0; // Points

		if (discr->mapping != NULL) EG_free(discr->mapping);
		discr->mapping = NULL;

		if (discr->types != NULL) { // Element types
			for (i = 0; i < discr->nTypes; i++) {
				if (discr->types[i].gst  != NULL) EG_free(discr->types[i].gst);
				if (discr->types[i].tris != NULL) EG_free(discr->types[i].tris);
			}

			EG_free(discr->types);
		}

		discr->nTypes  = 0;
		discr->types   = NULL;

		if (discr->elems != NULL) { // Element connectivity

			for (i = 0; i < discr->nElems; i++) {
				if (discr->elems[i].gIndices != NULL) EG_free(discr->elems[i].gIndices);
			}

			EG_free(discr->elems);
		}

		discr->nElems  = 0;
		discr->elems   = NULL;

		if (discr->ptrm != NULL) EG_free(discr->ptrm); // Extra information to store into the discr void pointer
		discr->ptrm    = NULL;


		discr->nVerts = 0;    // Not currently used
		discr->verts  = NULL; // Not currently used
		discr->celem = NULL; // Not currently used

		discr->nDtris = 0; // Not currently used
		discr->dtris  = NULL; // Not currently used
	}

	return CAPS_SUCCESS;
}

int aimDiscr(char *tname, capsDiscr *discr) {

	int i, body, face; // Indexing

	int status; // Function return status

	int iIndex; // Instance index

	int numBody;

	// EGADS objects
	ego tess, *bodies = NULL, *faces = NULL;

	const char   *string, *capsGroup; // capsGroups strings

	// EGADS function returns
	int plen, tlen, elen;
	const int    *ptype, *pindex, *tris, *nei;
	const double *xyz, *uv;

	// Body Tessellation
	int numFace = 0;
	int numFaceFound = 0;
	int numPoint = 0, numTri = 0;
	int *bodyFaceMap = NULL; // size=[2*numFaceFound], [2*numFaceFound + 0] = body, [2*numFaceFoun + 1] = face

	int *globalID = NULL, *localStitchedID = NULL, gID = 0;

	int nodeOffSet = 0;

	int *storage= NULL; // Extra information to store into the discr void pointer

	int numCAPSGroup = 0, attrIndex = 0, foundAttr = (int) false;
	int *capsGroupList = NULL;
	int dataTransferBodyIndex=-99;

	// Data transfer
	int  nrow, ncol, rank;
	void *dataTransfer = NULL;
	enum capsvType vtype;
	char *units;

	// Volume Mesh obtained from meshing AIM
	genUnstrMeshStruct volumeMesh;

	iIndex = discr->instance;
	#ifdef DEBUG
		printf(" su2AIM/aimDiscr: tname = %s, instance = %d!\n", tname, inst);
	#endif

	if ((iIndex < 0) || (iIndex >= numInstance)) return CAPS_BADINDEX;


	if (tname == NULL) return CAPS_NOTFOUND;

	if (su2Instance[iIndex].dataTransferCheck == (int) false) {
		printf("The volume is not suitable for data transfer - possibly the volume mesher "
				"added unaccounted for points\n");
		return CAPS_BADVALUE;
	}

	// Currently this works with 1 body!
	status = aim_getBodies(discr->aInfo, &numBody, &bodies);
	if (status != CAPS_SUCCESS) {
		printf(" su2AIM/aimDiscr: %d aim_getBodies = %d!\n", iIndex, status);
		return status;
	}

	status = aimFreeDiscr(discr);
	if (status != CAPS_SUCCESS) return status;

	numFaceFound = 0;
	numPoint = numTri = 0;
	// Find any faces with our boundary marker and get how many points and triangles there are
	for (body = 0; body < numBody; body++) {

		status = EG_getBodyTopos(bodies[body], NULL, FACE, &numFace, &faces);
		if (status != EGADS_SUCCESS) {
			printf(" su2AIM: getBodyTopos (Face) = %d for Body %d!\n", status, body);
			return status;
		}

		for (face = 0; face < numFace; face++) {

			// Retrieve the string following a capsTransfer tag
			status = retrieve_CAPSTransferAttr(faces[face], &string);
			if (status != CAPS_SUCCESS) continue;

			if (strcmp(string, tname) != 0) continue;

			#ifdef DEBUG
				printf(" su2AIM/aimDiscr: Body %d/Face %d matches %s!\n", body, face+1, tname);
			#endif


			status = retrieve_CAPSGroupAttr(faces[face], &capsGroup);
			if (status != CAPS_SUCCESS) {
				printf("capsTransfer found on face %d, but no capGroup found!!!\n", face);
				continue;
			} else {

				status = get_mapAttrToIndexIndex(&su2Instance[iIndex].attrMap, capsGroup, &attrIndex);
				if (status != CAPS_SUCCESS) {
					printf("capsGroup %s NOT found in attrMap\n",capsGroup);
					continue;
				} else {

					// If first index create arrays and store index
					if (numCAPSGroup == 0) {
						numCAPSGroup += 1;
						capsGroupList = (int *) EG_alloc(numCAPSGroup*sizeof(int));
						if (capsGroupList == NULL) {
							status =  EGADS_MALLOC;
							goto bail;
						}

						capsGroupList[numCAPSGroup-1] = attrIndex;
					} else { // If we already have an index(es) let make sure it is unique
						foundAttr = (int) false;
						for (i = 0; i < numCAPSGroup; i++) {
							if (attrIndex == capsGroupList[i]) {
								foundAttr = (int) true;
								break;
							}
						}

						if (foundAttr == (int) false) {
							numCAPSGroup += 1;
							capsGroupList = (int *) EG_reall(capsGroupList, numCAPSGroup*sizeof(int));
							if (capsGroupList == NULL) {
								status =  EGADS_MALLOC;
								goto bail;
							}

							capsGroupList[numCAPSGroup-1] = attrIndex;
						}
					}
				}
			}

			// Get face tessellation
			status = EG_getTessFace(bodies[body+numBody], face+1, &plen, &xyz, &uv, &ptype, &pindex, &tlen, &tris, &nei);
			if (status != EGADS_SUCCESS) {
				printf(" su2AIM: EG_getTessFace %d = %d for Body %d!\n", face+1, status, body+1);
				continue;
			}

			numFaceFound += 1;
			dataTransferBodyIndex = body;
			if (numFaceFound == 1) {
				bodyFaceMap = (int *) EG_alloc(2*numFaceFound*sizeof(int));
			} else {
				bodyFaceMap = (int *) EG_reall(bodyFaceMap, 2*numFaceFound*sizeof(int));
			}

			if (bodyFaceMap == NULL) {
				status = EGADS_MALLOC;
				goto bail;
			}

			// Get number of points and triangles
			bodyFaceMap[2*(numFaceFound-1) + 0] = body+1;
			bodyFaceMap[2*(numFaceFound-1) + 1] = face+1;

			// Sum number of points and triangles
			numPoint  += plen;
			numTri  += tlen;

		}

		if (faces != NULL) EG_free(faces);
		faces = NULL;


		if (dataTransferBodyIndex >=0) break; // Force that only one body can be used
	}

	if (numFaceFound == 0) {
		printf(" su2AIM/aimDiscr: No Faces match %s!\n", tname);

		status = CAPS_SUCCESS;
		goto bail;
	}

	if ((numPoint == 0) || (numTri == 0)) {
		#ifdef DEBUG
			printf(" su2AIM/aimDiscr: ntris = %d, npts = %d!\n", numTri, numPoint);
		#endif
		status = CAPS_SOURCEERR;
		goto bail;
	}

	status = aim_getData(discr->aInfo, "Volume_Mesh", &vtype, &rank, &nrow, &ncol, &dataTransfer, &units);
	if (status != CAPS_SUCCESS){
		if (status == CAPS_DIRTY) printf("Parent AIMS are dirty\n");
		else printf("Linking statusus = %d\n",status);

		printf("Didn't find a link to a volume mesh (Volume_Mesh) from parent - data transfer isn't possible.\n");

		goto bail;

	} else {

		volumeMesh = *(genUnstrMeshStruct *) dataTransfer;
	}

	#ifdef DEBUG
		printf(" su2AIM/aimDiscr: Instance %d, Body Index for data transfer = %d\n", inst, dataTransferBodyIndex);
	#endif


	// Specify our single element type
	status = EGADS_MALLOC;
	discr->nTypes = 1;

	discr->types  = (capsEleType *) EG_alloc(sizeof(capsEleType));
	if (discr->types == NULL) goto bail;
	discr->types[0].nref  = 3;
	discr->types[0].ndata = 0;            /* data at geom reference positions */
	discr->types[0].ntri  = 1;
	discr->types[0].nmat  = 0;            /* match points at geom ref positions */
	discr->types[0].tris  = NULL;
	discr->types[0].gst   = NULL;
	discr->types[0].dst   = NULL;
	discr->types[0].matst = NULL;

	discr->types[0].tris   = (int *) EG_alloc(3*sizeof(int));
	if (discr->types[0].tris == NULL) goto bail;
	discr->types[0].tris[0] = 1;
	discr->types[0].tris[1] = 2;
	discr->types[0].tris[2] = 3;

	discr->types[0].gst   = (double *) EG_alloc(6*sizeof(double));
	if (discr->types[0].gst == NULL) goto bail;
	discr->types[0].gst[0] = 0.0;
	discr->types[0].gst[1] = 0.0;
	discr->types[0].gst[2] = 1.0;
	discr->types[0].gst[3] = 0.0;
	discr->types[0].gst[4] = 0.0;
	discr->types[0].gst[5] = 1.0;

	// Get the tessellation and make up a simple linear continuous triangle discretization */

	discr->nElems = numTri;

	discr->elems = (capsElement *) EG_alloc(discr->nElems*sizeof(capsElement));
	if (discr->elems == NULL) goto bail;

	discr->mapping = (int *) EG_alloc(2*numPoint*sizeof(int)); // Will be resized
	if (discr->mapping == NULL) goto bail;

	globalID = (int *) EG_alloc(numPoint*sizeof(int));
	if (globalID == NULL) goto bail;

	localStitchedID = (int *) EG_alloc(numPoint*sizeof(int));
	if (localStitchedID == NULL) goto bail;

	for (i = 0; i < numPoint; i++) localStitchedID[i] = 0;

	numPoint = 0;
	numTri = 0;
	for (face = 0; face < numFaceFound; face++){

		tess = bodies[bodyFaceMap[2*face + 0]-1 + numBody];

		// Get face tessellation
		status = EG_getTessFace(tess, bodyFaceMap[2*face + 1], &plen, &xyz, &uv, &ptype, &pindex, &tlen, &tris, &nei);
		if (status != EGADS_SUCCESS) {
			printf(" su2AIM: EG_getTessFace %d = %d for Body %d!\n", bodyFaceMap[2*face + 1], status, bodyFaceMap[2*face + 0]);
			continue;
		}

		for (i = 0; i < plen; i++ ) {

			status = EG_localToGlobal(tess, bodyFaceMap[2*face+1], i+1, &gID);
			if (status != EGADS_SUCCESS) goto bail;


			if (localStitchedID[gID -1] != 0) continue;

			discr->mapping[2*numPoint  ] = bodyFaceMap[2*face + 0];
			discr->mapping[2*numPoint+1] = gID;

			localStitchedID[gID -1] = numPoint;

			globalID[numPoint] = gID;

			numPoint += 1;
		}

		// Get triangle connectivity in global sense
		for (i = 0; i < tlen; i++) {

			discr->elems[numTri].bIndex      = bodyFaceMap[2*face + 0];
			discr->elems[numTri].tIndex      = 1;
			discr->elems[numTri].eIndex      = bodyFaceMap[2*face + 1];

			discr->elems[numTri].gIndices    = (int *) EG_alloc(6*sizeof(int));
			if (discr->elems[numTri].gIndices == NULL) {
				status = EGADS_MALLOC;
				goto bail;
			}

			discr->elems[numTri].dIndices    = NULL;
			discr->elems[numTri].eTris.tq[0] = i+1;

			status = EG_localToGlobal(tess, bodyFaceMap[2*face + 1], tris[3*i + 0], &gID);
			if (status != EGADS_SUCCESS) goto bail;

			discr->elems[numTri].gIndices[0] = localStitchedID[gID-1];
			discr->elems[numTri].gIndices[1] = tris[3*i + 0];

			status = EG_localToGlobal(tess, bodyFaceMap[2*face + 1], tris[3*i + 1], &gID);
			if (status != EGADS_SUCCESS) goto bail;

			discr->elems[numTri].gIndices[2] = localStitchedID[gID-1];
			discr->elems[numTri].gIndices[3] = tris[3*i + 1];

			status = EG_localToGlobal(tess, bodyFaceMap[2*face + 1], tris[3*i + 2], &gID);
			if (status != EGADS_SUCCESS) goto bail;

			discr->elems[numTri].gIndices[4] = localStitchedID[gID-1];
			discr->elems[numTri].gIndices[5] = tris[3*i + 2];

			numTri += 1;
		}
	}

	discr->nPoints = numPoint;

	#ifdef DEBUG
		printf(" su2AIM/aimDiscr: ntris = %d, npts = %d!\n", discr->nElems, discr->nPoints);
	#endif

	// Resize mapping to switched together number of points
	discr->mapping = (int *) EG_reall(discr->mapping, 2*numPoint*sizeof(int));

	// Local to global node connectivity + numCAPSGroup + sizeof(capGrouplist)
	storage  = (int *) EG_alloc((numPoint + 1 + numCAPSGroup) *sizeof(int));
	if (discr->ptrm == NULL) goto bail;
	discr->ptrm = storage;

	// Store the local-to-Global (volume) id
	nodeOffSet = 0;
	for (i = 0; i < dataTransferBodyIndex; i++) {
		nodeOffSet += volumeMesh.surfaceMesh.surfNodesLump[i];
	}

	#ifdef DEBUG
		printf(" su2AIM/aimDiscr: Instance = %d, nodeOffSet = %d\n", iIndex, nodeOffSet);
	#endif

	for (i = 0; i < numPoint; i++) {

		storage[i] = volumeMesh.surfaceMesh.nodeLocal2Global[(globalID[i]-1) + nodeOffSet];

		//#ifdef DEBUG
		//	printf(" su2AIM/aimDiscr: Instance = %d, Global Node ID %d\n", iIndex, storage[i]);
		//#endif
	}

	// Save way the attrMap capsGroup list
	storage[numPoint] = numCAPSGroup;
	for (i = 0; i < numCAPSGroup; i++) {
		storage[numPoint+1+i] = capsGroupList[i];
	}
	if (capsGroupList != NULL) EG_free(capsGroupList);


	// Free up our stuff
	if (globalID  != NULL) EG_free(globalID);

	if (localStitchedID != NULL) EG_free(localStitchedID);

	if (bodyFaceMap != NULL) EG_free(bodyFaceMap);

	#ifdef DEBUG
		printf(" su2AIM/aimDiscr: Instance = %d, Finished!!\n", inst);
	#endif

	return CAPS_SUCCESS;

bail:
	if (faces != NULL) EG_free(faces);

	if (globalID  != NULL) EG_free(globalID);
	if (localStitchedID != NULL) EG_free(localStitchedID);

	if (capsGroupList != NULL) EG_free(capsGroupList);
	if (bodyFaceMap != NULL) EG_free(bodyFaceMap);

	aimFreeDiscr(discr);

	return status;
}

int aimLocateElement(capsDiscr *discr, double *params, double *param, int *eIndex,
					 double *bary)
{
	int    i, in[3], stat, ismall;
	double we[3], w, smallw = -1.e300;

	if (discr == NULL) return CAPS_NULLOBJ;

	for (ismall = i = 0; i < discr->nElems; i++) {
		in[0] = discr->elems[i].gIndices[0] - 1;
		in[1] = discr->elems[i].gIndices[2] - 1;
		in[2] = discr->elems[i].gIndices[4] - 1;
		stat  = EG_inTriExact(&params[2*in[0]], &params[2*in[1]], &params[2*in[2]], param, we);

		if (stat == EGADS_SUCCESS) {
			*eIndex = i+1;
			bary[0] = we[1];
			bary[1] = we[2];
			return CAPS_SUCCESS;
		}

		w = we[0];
		if (we[1] < w) w = we[1];
		if (we[2] < w) w = we[2];
		if (w > smallw) {
			ismall = i+1;
			smallw = w;
		}
	}

	/* must extrapolate! */
	if (ismall == 0) return CAPS_NOTFOUND;
	in[0] = discr->elems[ismall-1].gIndices[0] - 1;
	in[1] = discr->elems[ismall-1].gIndices[2] - 1;
	in[2] = discr->elems[ismall-1].gIndices[4] - 1;
	EG_inTriExact(&params[2*in[0]], &params[2*in[1]], &params[2*in[2]], param, we);

	*eIndex = ismall;
	bary[0] = we[1];
	bary[1] = we[2];

	/*
	  printf(" aimLocateElement: extropolate to %d (%lf %lf %lf)  %lf\n", ismall, we[0], we[1], we[2], smallw);
	*/
	return CAPS_SUCCESS;
}

int aimTransfer(capsDiscr *discr, char *dataName, int numPoint, int dataRank, double *dataVal, /*@unused@*/ char **units)
{
	/*! \page dataTransferFUN3D Data Transfer
	 * Data transfer blah, blah, blah....
	 */

	int status, status2; // Function return status
	int i, j, dataPoint; // Indexing

	// Current directory path
	char   currentPath[PATH_MAX];

	// Aero-Load data variables
	int numVariable;
	int numDataPoint;
	char **variableName = NULL;
	double **dataMatrix = NULL;
	double dataScaleFactor = 1.0;
	double dataScaleOffset = 0.0;
	char *dataUnits;

	// Indexing in data variables
	int globalIDIndex = -99;
	int variableIndex = -99;

	// Variables used in global node mapping
	int *nodeMap, *storage;
	int globalNodeID;

	// Filename stuff
	int *capsGroupList;
	char *filename = NULL; //"pyCAPS_FUN3D_Tetgen_ddfdrive_bndry1.dat";

	#ifdef DEBUG
		printf(" su2AIM/aimTransfer name = %s  instance = %d  npts = %d/%d!\n", dataName, discr->instance, numPoint, dataRank);
	#endif


	if (strcasecmp(dataName, "Pressure") != 0 &&
		strcasecmp(dataName, "P")        != 0 &&
		strcasecmp(dataName, "Cp")       != 0 &&
		strcasecmp(dataName, "CoeffientofPressure") != 0) {

		printf("Unrecognized data transfer variable - %s\n", dataName);
		return CAPS_NOTFOUND;
	}

	//Get the appropriate parts of the tessellation to data
	storage = (int *) discr->ptrm;
	//  tessNodeMap = &storage[3*discr->nElems ]; // Local indexing on the body
	nodeMap = &storage[0]; // Global indexing on the body

	capsGroupList = &storage[discr->nPoints]; // List of boundary ID (attrMap) in the transfer

	//printf("\n\n...............HACKING FUN3D AIM TRANSFER........\n");
	dataVal[0] = 99;
	for (i = 0; i < capsGroupList[0]; i++) {

		(void) getcwd(currentPath, PATH_MAX);
		if (chdir(su2Instance[discr->instance].analysisPath) != 0) {
			#ifdef DEBUG
				printf(" su2AIM/aimTransfer Cannot chdir to %s!\n", su2Instance[discr->instance].analysisPath);
			#endif

			return CAPS_DIRERR;
		}


		filename = (char *) EG_alloc((strlen(su2Instance[discr->instance].projectName) +
									  strlen("_ddfdrive_bndry.dat") + 5)*sizeof(char));

		sprintf(filename,"%s%s%d%s",su2Instance[discr->instance].projectName,
									"_ddfdrive_bndry",
									capsGroupList[i+1],
									".dat");

		status = su2_readAeroLoad(filename,
									&numVariable,
									&variableName,
									&numDataPoint,
									&dataMatrix);

		if (filename != NULL) EG_free(filename);
		filename = NULL;

		// Restore the path we came in with
		chdir(currentPath);
		if (status != CAPS_SUCCESS) return status;

		printf("Number of variables %d\n", numVariable);
		// Output some of the first row of the data
		//for (i = 0; i < numVariable; i++) printf("Variable %d - %.6f\n", i, dataMatrix[i][0]);

		// Loop through the variable list to find which one is the global node ID variable
		for (i = 0; i < numVariable; i++) {
			if (strcasecmp("id", variableName[i]) ==0 ) {
				globalIDIndex = i;
				break;
			}
		}

		if (globalIDIndex == -99) {
			printf("Global node number variable not found in data file\n");
			status = CAPS_NOTFOUND;
			goto bail;
		}

		// Loop through the variable list to see if we can find the transfer data name
		for (i = 0; i < numVariable; i++) {

			if (strcasecmp(dataName, "Pressure") == 0 ||
				strcasecmp(dataName, "P")        == 0 ||
				strcasecmp(dataName, "Cp")       == 0 ||
				strcasecmp(dataName, "CoeffientofPressure") == 0) {

				if (strcasecmp("cp", variableName[i]) == 0) {
					variableIndex = i;
					break;
				}

				if (dataRank != 1) {
					printf("Data transfer rank should be 1 not %d\n", dataRank);
					status = CAPS_BADRANK;
					goto bail;
				}

				dataScaleFactor = su2Instance[discr->instance].pressureScaleFactor->vals.real;
				dataScaleOffset = su2Instance[discr->instance].pressureScaleOffset->vals.real;

				dataUnits = su2Instance[discr->instance].pressureScaleFactor->units;
			}
		}

		if (variableIndex == -99) {
			printf("Variable %s not found in data file\n", dataName);
			status = CAPS_NOTFOUND;
			goto bail;
		}

		for (i = 0; i < numPoint; i++) {

			globalNodeID = nodeMap[i];

			for (dataPoint = 0; dataPoint < numDataPoint; dataPoint++) {
				if ((int) dataMatrix[globalIDIndex][dataPoint] ==  globalNodeID) break;
			}

			for (j = 0; j < dataRank; j++) {

				// Add something for units - aim_covert()

				dataVal[dataRank*i+j] = dataMatrix[variableIndex][dataPoint]*dataScaleFactor + dataScaleOffset;
				//printf("DataValue = %f\n",dataVal[dataRank*i+j]);
				//dataVal[dataRank*i+j] = 99;

			}
		}


		// Free data matrix
		if (dataMatrix != NULL) {
			for (i = 0; i < numVariable; i++) {
				if (dataMatrix[i] != NULL) EG_free(dataMatrix[i]);
			}

			EG_free(dataMatrix);
		}

		// Free variable list
		status = string_freeArray(numVariable, &variableName);
		if (status != CAPS_SUCCESS) return status;

	}

	return CAPS_SUCCESS;

	bail:
		// Free data matrix
		if (dataMatrix != NULL) {
			for (i = 0; i < numVariable; i++) {
				if (dataMatrix[i] != NULL) EG_free(dataMatrix[i]);
			}

			EG_free(dataMatrix);
		}

		// Free variable list
		status2 = string_freeArray(numVariable, &variableName);
		if (status2 != CAPS_SUCCESS) return status2;

		return status;
}

int aimInterpolation(capsDiscr *discr, /*@unused@*/ char *name, int eIndex,
					 double *bary, int rank, double *data, double *result)
{
	int    in[3], i;
	double we[3];
	/*
	#ifdef DEBUG
		printf(" su2AIM/aimInterpolation  %s  instance = %d!\n", name, discr->instance);
	#endif
	*/

	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" su2AIM/Interpolation: eIndex = %d [1-%d]!\n", eIndex, discr->nElems);
	}

	we[0] = 1.0 - bary[0] - bary[1];
	we[1] = bary[0];
	we[2] = bary[1];
	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[1] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[2] - 1;

	for (i = 0; i < rank; i++) {
		result[i] = data[rank*in[0]+i]*we[0] + data[rank*in[1]+i]*we[1] + data[rank*in[2]+i]*we[2];
	}

	return CAPS_SUCCESS;
}

int aimInterpolateBar(capsDiscr *discr, /*@unused@*/ char *name, int eIndex,
					  double *bary, int rank, double *r_bar, double *d_bar)
{
	int    in[3], i;
	double we[3];

	/*
	#ifdef DEBUG
		printf(" su2AIM/aimInterpolateBar  %s  instance = %d!\n", name, discr->instance);
	#endif
	*/

	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" su2AIM/InterpolateBar: eIndex = %d [1-%d]!\n", eIndex, discr->nElems);
	}

	we[0] = 1.0 - bary[0] - bary[1];
	we[1] = bary[0];
	we[2] = bary[1];

	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[1] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[2] - 1;

	for (i = 0; i < rank; i++) {
		// result[i] = data[rank*in[0]+i]*we[0] + data[rank*in[1]+i]*we[1] + data[rank*in[2]+i]*we[2];
		d_bar[rank*in[0]+i] += we[0]*r_bar[i];
		d_bar[rank*in[1]+i] += we[1]*r_bar[i];
		d_bar[rank*in[2]+i] += we[2]*r_bar[i];
	}

	return CAPS_SUCCESS;
}

int aimIntegration(capsDiscr *discr, /*@unused@*/ char *name, int eIndex, int rank,
				   /*@null@*/ double *data, double *result)
{
	int    i, in[3], stat, ptype, pindex, nBody;
	double x1[3], x2[3], x3[3], xyz1[3], xyz2[3], xyz3[3], area;
	ego    *bodies;

	/*
	#ifdef DEBUG
	  printf(" su2AIM/aimIntegration  %s  instance = %d!\n", name, discr->instance);
	#endif
	*/

	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" su2AIM/aimIntegration: eIndex = %d [1-%d]!\n", eIndex, discr->nElems);
	}

	stat = aim_getBodies(discr->aInfo, &nBody, &bodies);
	if (stat != CAPS_SUCCESS) {
		printf(" su2AIM/aimIntegration: %d aim_getBodies = %d!\n", discr->instance, stat);
		return stat;
	}

	/* element indices */

	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[2] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[4] - 1;

	stat = EG_getGlobal(bodies[discr->mapping[2*in[0]]+nBody-1], discr->mapping[2*in[0]+1], &ptype, &pindex, xyz1);

	if (stat != CAPS_SUCCESS) {
		printf(" su2AIM/aimIntegration: %d EG_getGlobal %d = %d!\n", discr->instance, in[0], stat);
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[1]]+nBody-1], discr->mapping[2*in[1]+1], &ptype, &pindex, xyz2);

	if (stat != CAPS_SUCCESS) {
		printf(" su2AIM/aimIntegration: %d EG_getGlobal %d = %d!\n", discr->instance, in[1], stat);
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[2]]+nBody-1], discr->mapping[2*in[2]+1], &ptype, &pindex, xyz3);

	if (stat != CAPS_SUCCESS) {
		printf(" su2AIM/aimIntegration: %d EG_getGlobal %d = %d!\n", discr->instance, in[2], stat);
	}

	x1[0] = xyz2[0] - xyz1[0];
	x2[0] = xyz3[0] - xyz1[0];
	x1[1] = xyz2[1] - xyz1[1];
	x2[1] = xyz3[1] - xyz1[1];
	x1[2] = xyz2[2] - xyz1[2];
	x2[2] = xyz3[2] - xyz1[2];
	//CROSS(x3, x1, x2);
	cross_DoubleVal(x1,x2,x3);
	//area  = sqrt(DOT(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */
	area  = sqrt(dot_DoubleVal(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */

	if (data == NULL) {
		*result = 3.0*area;
		return CAPS_SUCCESS;
	}

	for (i = 0; i < rank; i++) {
		result[i] = (data[rank*in[0]+i] + data[rank*in[1]+i] + data[rank*in[2]+i])*area;
	}

	return CAPS_SUCCESS;
}

int aimIntegrateBar(capsDiscr *discr, /*@unused@*/ char *name, int eIndex, int rank,
					double *r_bar, double *d_bar)
{
	int    i, in[3], stat, ptype, pindex, nBody;
	double x1[3], x2[3], x3[3], xyz1[3], xyz2[3], xyz3[3], area;
	ego    *bodies;

	/*
	#ifdef DEBUG
	  printf(" su2AIM/aimIntegrateBar  %s  instance = %d!\n",name, discr->instance);
	#endif
	*/

	if ((eIndex <= 0) || (eIndex > discr->nElems)) {
		printf(" su2AIM/aimIntegrateBar: eIndex = %d [1-%d]!\n", eIndex, discr->nElems);
	}

	stat = aim_getBodies(discr->aInfo, &nBody, &bodies);
	if (stat != CAPS_SUCCESS) {
		printf(" su2AIM/aimIntegrateBar: %d aim_getBodies = %d!\n", discr->instance, stat);
		return stat;
	}

	/* element indices */

	in[0] = discr->elems[eIndex-1].gIndices[0] - 1;
	in[1] = discr->elems[eIndex-1].gIndices[2] - 1;
	in[2] = discr->elems[eIndex-1].gIndices[4] - 1;

	stat = EG_getGlobal(bodies[discr->mapping[2*in[0]]+nBody-1], discr->mapping[2*in[0]+1], &ptype, &pindex, xyz1);
	if (stat != CAPS_SUCCESS) {
		printf(" su2AIM/aimIntegrateBar: %d EG_getGlobal %d = %d!\n", discr->instance, in[0], stat);
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[1]]+nBody-1], discr->mapping[2*in[1]+1], &ptype, &pindex, xyz2);
	if (stat != CAPS_SUCCESS) {
		printf(" su2AIM/aimIntegrateBar: %d EG_getGlobal %d = %d!\n", discr->instance, in[1], stat);
	}

	stat = EG_getGlobal(bodies[discr->mapping[2*in[2]]+nBody-1], discr->mapping[2*in[2]+1], &ptype, &pindex, xyz3);

	if (stat != CAPS_SUCCESS) {
		printf(" su2AIM/aimIntegrateBar: %d EG_getGlobal %d = %d!\n", discr->instance, in[2], stat);
	}

	x1[0] = xyz2[0] - xyz1[0];
	x2[0] = xyz3[0] - xyz1[0];
	x1[1] = xyz2[1] - xyz1[1];
	x2[1] = xyz3[1] - xyz1[1];
	x1[2] = xyz2[2] - xyz1[2];
	x2[2] = xyz3[2] - xyz1[2];
	//CROSS(x3, x1, x2);
	cross_DoubleVal(x1,x2,x3);
	//area  = sqrt(DOT(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */
	area  = sqrt(dot_DoubleVal(x3, x3))/6.0;      /* 1/2 for area and then 1/3 for sum */

	for (i = 0; i < rank; i++) {
		//result[i] = (data[rank*in[0]+i] + data[rank*in[1]+i] + data[rank*in[2]+i])*area;
		d_bar[rank*in[0]+i] += area*r_bar[i];
		d_bar[rank*in[1]+i] += area*r_bar[i];
		d_bar[rank*in[2]+i] += area*r_bar[i];
	}

	return CAPS_SUCCESS;
}
