// Tetgen interface functions - Written by Dr. Ryan Durscher AFRL/RQVC

#include "tetgen.h"
#include "meshTypes.h"
#include "egads.h"
#include "capsTypes.h"

static int tetgen_to_GenUnstrMesh(tetgenio *mesh, genUnstrMeshStruct *genUnstrMesh) {

    int i; // Indexing

    int defaultVolID = 1; // Default volume ID

	// Numbers:
	genUnstrMesh->numNodes = mesh->numberofpoints;
	genUnstrMesh->surfaceMesh.numQuadFace = 0;
	genUnstrMesh->surfaceMesh.numTriFace = mesh->numberoftrifaces;

	// Copy nodes
	if (genUnstrMesh->xyz != NULL) EG_free(genUnstrMesh->xyz);
	genUnstrMesh->xyz = (double *) EG_alloc(genUnstrMesh->numNodes*3 * sizeof(double));
	if (genUnstrMesh->xyz == NULL) return EGADS_MALLOC;

	memcpy(genUnstrMesh->xyz, mesh->pointlist, genUnstrMesh->numNodes*3 * sizeof(double));

/*
	if (mesh->pointmarkerlist != NULL) {
	    for (i = 0; i < genUnstrMesh->numNodes; i++) printf("Point %d - attr - %d\n", i+1, mesh->pointmarkerlist[i]);
	} else printf("Point markerlist is NULL");
*/

	// Set quadface number
	if (genUnstrMesh->surfaceMesh.quadFaceList != NULL) EG_free(genUnstrMesh->surfaceMesh.quadFaceList);
	genUnstrMesh->surfaceMesh.quadFaceList      = NULL;

	if (genUnstrMesh->surfaceMesh.quadFaceMarkList != NULL) EG_free(genUnstrMesh->surfaceMesh.quadFaceMarkList);
	genUnstrMesh->surfaceMesh.quadFaceMarkList  = NULL;

	// Copy surface tri connectivity and marker
	if (genUnstrMesh->surfaceMesh.triFaceList != NULL) EG_free(genUnstrMesh->surfaceMesh.triFaceList);

	genUnstrMesh->surfaceMesh.triFaceList = (int *) EG_alloc(genUnstrMesh->surfaceMesh.numTriFace*3 * sizeof(int));
	if (genUnstrMesh->surfaceMesh.triFaceList == NULL) return EGADS_MALLOC;

	memcpy(genUnstrMesh->surfaceMesh.triFaceList, mesh->trifacelist, genUnstrMesh->surfaceMesh.numTriFace*3*sizeof(int));

	if (genUnstrMesh->surfaceMesh.triFaceMarkList != NULL) EG_free(genUnstrMesh->surfaceMesh.triFaceMarkList);

	genUnstrMesh->surfaceMesh.triFaceMarkList = (int *) EG_alloc(genUnstrMesh->surfaceMesh.numTriFace*sizeof(int));
	if (genUnstrMesh->surfaceMesh.triFaceMarkList == NULL) return EGADS_MALLOC;

	memcpy(genUnstrMesh->surfaceMesh.triFaceMarkList, mesh->trifacemarkerlist, genUnstrMesh->surfaceMesh.numTriFace*sizeof(int));

	// Copy tetrahedral
	genUnstrMesh->numTetra = mesh->numberoftetrahedra;

	if (genUnstrMesh->tetraList != NULL) EG_free(genUnstrMesh->tetraList);

	genUnstrMesh->tetraList = (int *) EG_alloc(genUnstrMesh->numTetra*4 * sizeof(int));
	if (genUnstrMesh->tetraList == NULL) return EGADS_MALLOC;

	memcpy(genUnstrMesh->tetraList,mesh->tetrahedronlist, genUnstrMesh->numTetra*4*sizeof(int));

	// Copy tetrahedral volume id in case mesh will be used for FEA
	genUnstrMesh->tetraVolIDList = (int *) EG_alloc(genUnstrMesh->numTetra * sizeof(int));
	if (genUnstrMesh->tetraVolIDList == NULL) return EGADS_MALLOC;

	if (mesh->numberoftetrahedronattributes == 0) {
	    genUnstrMesh->numVolID = 1;
	    genUnstrMesh->volIDList =  (int *) EG_alloc(genUnstrMesh->numVolID * sizeof(int));
	    if (genUnstrMesh->volIDList == NULL) return EGADS_MALLOC;

	    genUnstrMesh->volIDList[0] = defaultVolID;
	}

	for (i = 0; i < genUnstrMesh->numTetra; i++) {
	    if (mesh->numberoftetrahedronattributes == 0) genUnstrMesh->tetraVolIDList[i] = defaultVolID;
	    else {
	        printf("Populating tetraVolID array from tetrahedronattributelist - This is currently untested - SEG Fault may ensue\n");
	        genUnstrMesh->tetraVolIDList[i] = (int) mesh->tetrahedronattributelist[mesh->numberoftetrahedronattributes*i + (mesh->numberoftetrahedronattributes-1)];
	    }
	}

	return 0;
}

//#ifdef __cplusplus
extern "C" {
//#endif

int tetgen_Volume_Mesh(int numHoles,
                       double *holes,
                       meshInputStruct meshInput,
                       genUnstrSurfMeshStruct *surfaceMesh,
					   genUnstrMeshStruct *genUnstrMesh)
{
    // Initialize variables
	int status = 0; //Function status return
    int i, j; // Indexing
    char *inputString = NULL; // Input string to tetgen

    // TetGen variables
	tetgenio in, out;
	tetgenio::facet *f;
	tetgenio::polygon *p;

	printf("\nGenerating volume mesh using TetGen.....\n");

	// First index - either 0 or 1
	in.firstnumber = 1;

	// Copy (point) surfaceMesh inputs into output volume mesh
	genUnstrMesh->surfaceMesh.numNodes = surfaceMesh->numNodes;
	genUnstrMesh->surfaceMesh.xyz = surfaceMesh->xyz;

	genUnstrMesh->surfaceMesh.numTriFace = surfaceMesh->numTriFace;
	genUnstrMesh->surfaceMesh.numQuadFace = surfaceMesh->numQuadFace;

	genUnstrMesh->surfaceMesh.localTriFaceList = surfaceMesh->localTriFaceList;
	genUnstrMesh->surfaceMesh.triFaceMarkList = surfaceMesh->triFaceMarkList;

	genUnstrMesh->surfaceMesh.localQuadFaceList = surfaceMesh->localQuadFaceList;
	genUnstrMesh->surfaceMesh.quadFaceMarkList  = surfaceMesh->quadFaceMarkList;

	// Check inputs
	if (surfaceMesh->numNodes == 0) return CAPS_BADVALUE;
	if (surfaceMesh->numTriFace == 0) return CAPS_BADVALUE;

	// Set the number of surface nodes
	in.numberofpoints = surfaceMesh->numNodes;

	// Create surface point array
	in.pointlist = new REAL[in.numberofpoints *3];

	// Transfer input coordinates to tetgen array
	for(j = 0; j < surfaceMesh->numNodes; j++) {
		in.pointlist[3*j  ] = surfaceMesh->xyz[3*j  ];
		in.pointlist[3*j+1] = surfaceMesh->xyz[3*j+1];
		in.pointlist[3*j+2] = surfaceMesh->xyz[3*j+2];
	}

	// Set the number of surface tri
	in.numberoffacets = surfaceMesh->numTriFace;

	// Create surface tri arrays
	in.facetlist = new tetgenio::facet[in.numberoffacets];

	// Create surface marker/BC arrays
	in.facetmarkerlist = new int[in.numberoffacets];

	// Transfer input surfaceMesh->localTriFaceList array to tetgen array
	for(j = 0; j < surfaceMesh->numTriFace; j++) {
		f =&in.facetlist[j];
		f->numberofpolygons = 1;
		f->polygonlist = new tetgenio::polygon[f->numberofpolygons];
		f->numberofholes = 0;
		f->holelist = NULL;
	    p = &f->polygonlist[0];
		p->numberofvertices = 3;
		p->vertexlist = new int[p->numberofvertices];
		p->vertexlist[0] = surfaceMesh->localTriFaceList[3*j  ];
		p->vertexlist[1] = surfaceMesh->localTriFaceList[3*j+1];
		p->vertexlist[2] = surfaceMesh->localTriFaceList[3*j+2];
	}

	// Transfer input BC array to tetgen array
	for(j = 0; j< surfaceMesh->numTriFace;j++) {
		in.facetmarkerlist[j] = surfaceMesh->triFaceMarkList[j];
	}

	// Add hole information
	if ((holes != NULL) || (numHoles != 0)) {
	    //printf("\n WE HAVE HOLES\n");

	    // Initialize variables
	    in.numberofholes = numHoles;
	    in.holelist = new REAL[in.numberofholes *3];

	    // Transfer hole points to tetgen array
	    for (j = 0; j < in.numberofholes; j++) {
	        in.holelist[3*j  ] = holes[3*j  ];
	        in.holelist[3*j+1] = holes[3*j+1];
	        in.holelist[3*j+2] = holes[3*j+2];
	    }
	}

    // If no input string is provided create a simple one based on exposed parameters
	if (meshInput.tetgenInput.meshInputString == NULL) {

	    // Create Tetgen input string
	    char temp[120] = "p"; // Tetrahedralize a piecewise linear complex flag
	    char q[80];

	    if (meshInput.preserveSurfMesh !=0) strcat(temp, "Y"); // Preserve surface mesh flag

	    if ((meshInput.tetgenInput.meshQuality_rad_edge != 0) ||
                (meshInput.tetgenInput.meshQuality_angle != 0)) {

	        strcat(temp, "q"); // Mesh quality flag

	        if (meshInput.tetgenInput.meshQuality_rad_edge >= 0) {

	            sprintf(q, "%.3f",meshInput.tetgenInput.meshQuality_rad_edge);
	            strcat(temp,q);

	        } else if (meshInput.tetgenInput.meshQuality_rad_edge < 0)  {

	            printf("Not setting meshQuality radius-edge ratio. Value needs to be positive\n");
	        }

	        if (meshInput.tetgenInput.meshQuality_angle >= 0) {

	            sprintf(q, "/%.3f",meshInput.tetgenInput.meshQuality_angle);
	            strcat(temp,q);

	        } else if (meshInput.tetgenInput.meshQuality_angle < 0) {

	            printf("Not setting meshQuality dihedral angle. Value needs to be positive\n");
	        }

	    }

	    if ((meshInput.quiet != 0) && (meshInput.tetgenInput.verbose == 0)) strcat(temp, "Q"); // Quiet: No terminal output except errors
	    if (meshInput.tetgenInput.verbose != 0) strcat(temp, "V"); // Verbose: Detailed information, more terminal output.

	    if (meshInput.tetgenInput.meshTolerance > 0) {
	        sprintf(q, "T%.2e",meshInput.tetgenInput.meshTolerance);
	        strcat(temp,q);
	    }

	    // Transfer temp character array to input
	    inputString = temp;

	} else {

	    inputString = meshInput.tetgenInput.meshInputString;
	    // Dont think this is needed - maybe p is the default
	    // Check to make sure a 'p' is included in the input string provided by a user
	    /*if (strchr(input,'p') == NULL) {
	        printf("Warning: No \'p\' was found in the Tetgen input string. Input into Tetgen is currently based on PLC's\n");
	    }*/
	}

	printf("\nTetgen input string = %s""\n", inputString);


	// Create volume mesh
	try {
		tetrahedralize(inputString, &in, &out);
	} catch (...){
		printf("Tetgen failed to generate a volume mesh......!!!\n");
		return -335;
	}

    // Save data
    //in.save_nodes((char *)"TETGEN_Test");
    //in.save_poly((char *) "TETGEN_Test");
	//out.save_faces((char *) "TETGEN_Test");

	// Transfer tetgen mesh structure to genUnstrMesh format
	status = tetgen_to_GenUnstrMesh(&out, genUnstrMesh);
	if (status != 0) return status;

	// Populate surface mesh
    if (meshInput.preserveSurfMesh == (int) false &&
        meshInput.tetgenInput.meshInputString == NULL &&
        meshInput.tetgenInput.ignoreSurfaceExtract == (int) false) {

        printf("Extracting surface mesh from volume\n");

        int nodeCount = 0;
        int node;
        bool found = false;

        int *unique_Nodelist;

        unique_Nodelist = new int [3*out.numberoftrifaces]; // At most 3*numTriFaces

        for (i = 0; i < 3*out.numberoftrifaces; i++) {

            node = out.trifacelist[i];

            found = false;
            for (j = 0; j < nodeCount; j++){

                if (unique_Nodelist[j] == node) {
                    found = true;
                    break;
                }
            }

            if (found == false) {
                unique_Nodelist[nodeCount] = node;
                nodeCount += 1;
            }
        }

        // Populate local node list
        //printf("Num. nodes = %d\n",nodeCount);
        genUnstrMesh->surfaceMesh.numNodes = nodeCount;

        if (genUnstrMesh->surfaceMesh.nodeLocal2Global != NULL) EG_free(genUnstrMesh->surfaceMesh.nodeLocal2Global);

        genUnstrMesh->surfaceMesh.nodeLocal2Global = (int *) EG_alloc(genUnstrMesh->surfaceMesh.numNodes * sizeof(int));
        if (genUnstrMesh->surfaceMesh.nodeLocal2Global == NULL) return EGADS_MALLOC;

        if (genUnstrMesh->surfaceMesh.xyz != NULL) EG_free(genUnstrMesh->surfaceMesh.xyz);

        genUnstrMesh->surfaceMesh.xyz  = (double *) EG_alloc(3*genUnstrMesh->surfaceMesh.numNodes *sizeof(double));
        if (genUnstrMesh->surfaceMesh.xyz == NULL) return EGADS_MALLOC;

        for (i = 0; i < genUnstrMesh->surfaceMesh.numNodes; i++) {
        	genUnstrMesh->surfaceMesh.xyz[3*i + 0] = out.pointlist[(unique_Nodelist[i]-1) + 0]; // -1 -> 1 bias of node numbering
        	genUnstrMesh->surfaceMesh.xyz[3*i + 1] = out.pointlist[(unique_Nodelist[i]-1) + 1];
        	genUnstrMesh->surfaceMesh.xyz[3*i + 2] = out.pointlist[(unique_Nodelist[i]-1) + 2];

        	genUnstrMesh->surfaceMesh.nodeLocal2Global[i] = unique_Nodelist[i];
        }

        // Populate local node connectivity
        //genUnstrMesh->surfaceMesh.numTriFace = out.numberoftrifaces;
        if (genUnstrMesh->surfaceMesh.localTriFaceList != NULL) EG_free(genUnstrMesh->surfaceMesh.localTriFaceList);

        genUnstrMesh->surfaceMesh.localTriFaceList = (int *) EG_alloc(3*genUnstrMesh->surfaceMesh.numTriFace * sizeof(int));
        if (genUnstrMesh->surfaceMesh.localTriFaceList == NULL) return EGADS_MALLOC;

        for (i = 0; i < 3*genUnstrMesh->surfaceMesh.numTriFace; i++) {

            for (j = 0; j < genUnstrMesh->surfaceMesh.numNodes; j++) {

                if (out.trifacelist[i] == genUnstrMesh->surfaceMesh.nodeLocal2Global[j]) {
                	genUnstrMesh->surfaceMesh.localTriFaceList[i] = j+1; // Indexing starts at 1 -> 1 bias of node numbering
                    break;
                }
            }
        }

        // Clean up unique node list
        delete [] unique_Nodelist;
        unique_Nodelist = NULL;

    } else if (meshInput.preserveSurfMesh == (int) true &&
               meshInput.tetgenInput.meshInputString == NULL) {
        printf("Surface mesh as been preserved getting global node connectivity and local-2-global node mapping\n");

        if (genUnstrMesh->surfaceMesh.nodeLocal2Global != NULL) EG_free(genUnstrMesh->surfaceMesh.nodeLocal2Global);

        genUnstrMesh->surfaceMesh.nodeLocal2Global = (int *) EG_alloc(genUnstrMesh->surfaceMesh.numNodes * sizeof(int));
        if (genUnstrMesh->surfaceMesh.nodeLocal2Global == NULL) return EGADS_MALLOC;

        //Get local to global node numbering
        for (i = 0; i < genUnstrMesh->surfaceMesh.numNodes; i++) {
        	for(j = 0; j < out.numberofpoints; j++){
        		if ( genUnstrMesh->surfaceMesh.xyz[3*i + 0] == out.pointlist[3*j + 0] &&
        			 genUnstrMesh->surfaceMesh.xyz[3*i + 1] == out.pointlist[3*j + 1] &&
					 genUnstrMesh->surfaceMesh.xyz[3*i + 2] == out.pointlist[3*j + 2]) {
        			break;
        		}
        	}

        	genUnstrMesh->surfaceMesh.nodeLocal2Global[i] = j+in.firstnumber; // Indexing starts at 1 -> 1 bias of node numbering
        }
    }

    // Release TetGen memory - get released when function goes out of scope
    //in.deinitialize();
    //out.deinitialize();

    printf("Done meshing using TetGen!\n");

    return 0;
}

//#ifdef __cplusplus
}
//#endif


