// Structures for general meshing - Written by Dr. Ryan Durscher AFRL/RQVC

#ifndef MESHTYPES_H
#define MESHTYPES_H

#include "egads.h"
// AFLR3 Mesh properties
    //  STD_UG3_GBC    : standard surface
    //  -STD_UG3_GBC   : standard BL generating surface
    //  BL_INT_UG3_GBC : surface that intersects BL region
    //  TRANSP_SRC_UG3_GBC  : embedded/transparent surface that is converted to source nodes by AFLR3
    //  TRANSP_BL_INT_UG3_GBC : embedded/transparent surface that intersects BL region
    //  TRANSP_UG3_GBC       : embedded/transparent surface
    //  -TRANSP_UG3_GBC      : embedded/transparent BL generating surface
    //  TRANSP_INTRNL_UG3_GBC : embedded/transparent surface that is converted to internal faces by AFLR3
    //  -TRANSP_INTRNL_UG3_GBC : embedded/transparent BL generating surface
    //                              that is converted to internal faces by AFLR3
    //  FIXED_BL_INT_UG3_GBC : fixed surface with a BL region that intersects volume BL region

//typedef enum {Standard, BoundarLayer, IntersectBoundaryLayer, Transparent,

typedef enum {UnknownDistribution, EvenDistribution, TanhDistribution } edgeDistributionEnum;

typedef struct {
	// EGADS body tessellation storage
	ego egadsTess; // EGADS body tessellation
	int numTessFace; // Number of faces in the tessellation
	int *tessFaceQuadMap; // List to keep track of whether or not the tessObj has quads that have been split into tris
					   // size = [numTessFace]. In general if the quads have been split they should be added to the end
					   // of the tri list in the face tessellation
} bodyTessMappingStruct;

// General container for unstructured surface meshes
typedef struct {
	int numTriFace;  // Number of triangular faces
	int numQuadFace; // Number of quadrilateral faces
	int numNodes;    // Number of nodes

	double *xyz; // (Local) x1 = xyz[0], y1 = xyz[1], z1 = xyz[2], x2 = xyz[3],.... size  = [3*numNodes]
	int *nodeLocal2Global; // Local to global node mapping - size = [numNodes]

	int *triFaceList;  // Surface triface node (connectivity) index w.r.t to volume mesh - Three corners for face1 are
	                   // at indices [0], [1] and [2], followed by the remaining faces. Three ints per face. size = [3*numTriFace]

	int *quadFaceList; // Surface quadface node index w.r.t to volume mesh - Same as above size = [4*numQuadFace]

	int *localTriFaceList;  // Local node connectivity list for triface. size = [3*numTriFace]
	int *localQuadFaceList; // Local node connectivity list for quadface. size = [4*numQuadFace]

	int *triFaceMarkList;   // BC flag index for triface. size = [numTriFace]
	int *quadFaceMarkList;  // BC flag index for quadface. size = [numQuadFace]

	int *triFaceBLList;   // Boundary layer flag for triface - for volume mesher if applicable. size = [numTriFace]
	int *quadFaceBLList;  // Boundary layer flag for quadface. size = [numQuadFace]

	double *nodeBLSpacingList;   // Boundary layer initial normal spacing at each node - for volume mesher if applicable. size = [numNodes]
	double *nodeBLThicknessList;  // Boundary layer desired thickness at each node. size = [numNodes]

	int numBoundaryEdge; // Number of edges for a face - Used for 2D meshes - otherwise 0
	int *localBoundaryEdgeList; // Local node connectivity for an edge. size = [2*numBoundaryEdge]
	int *boundaryEdgeMarkList; // BC flag index for edge. size = [2*numBoundaryEdge]

	// If multiple surfaces have been lumped in a single array these arrays may be used to un-lump them
	int numSurface; // Number of surfaces lumped together
	int *surfTriFaceLump; // Number of TriFace for an individual surface. size = [numSurface]
						  // surfTriFaceLump[0] = number of trifaces that were in surface 0,
					      // surfTriFaceLump[1] = number of trifaces that were in surface 1, etc.
						  // The order of this array should correspond to the same order as the surfaces were lumped together
						  // into a single array
	int *surfQuadFaceLump;// Number of QuadFace for an individual surface. size = [numSurface]
	int *surfNodesLump;   // Number of Nodes for an individual surface. size = [numSurface]

	bodyTessMappingStruct **bodyTessMappingLump; // size = [numSurface]

	bodyTessMappingStruct bodyTessMapping;

} genUnstrSurfMeshStruct;

// General container for unstructured volume meshes
typedef struct {
    int numNodes; // Number of nodes
    int numTetra; // Number of tetrahedrals
    int numPyr;   // Number of pyramids
    int numPrz;	  // Number of prisms
    int numHex;   // Number of hexahedrals


    double *xyz;    // x1 = xyz[0], y1 = xyz[1], z1 = xyz[2], x2 = xyz[3],.... size  = [3*numPoints]
    int *tetraList; // Connectivity list for tetrahedrals. Corner of the first tetrahedral
    			    // are tetra[0] = [0], [1], [2], [3]; tetra[1] = [4], [5], [6], [7], etc.... . size = [4*numTetra]
    int *pyrList;   // Connectivity list for pyramids. Similar to tetra corner information.. see above ^ . size = [5*numPyr]
    int *przList;   // Connectivity list for prisms. Similar to tetra corner information.. see above ^ . size = [6*numPrz]
    int *hexList;   // Connectivity list for hexahedral. Similar to tetra corner information.. see above ^ . size = [8*numHex]

    int numVolID; // Global number of volume ID's
    int *volIDList;  // Global list of volume ID's. size = [numVolID];
    int *tetraVolIDList; // Volume ID associated with each tetrahedral element. size = [numTetra]
    int *pyrVolIDList;   // Volume ID associated with each pyramid element. size = [numPyr]
    int *przVolIDList;   // Volume ID associated with each prism  element. size = [numPrz]
    int *hexVolIDList;   // Volume ID associated with each hexahedral element. size = [numHex]

    genUnstrSurfMeshStruct surfaceMesh; // Container for surface mesh in the volume

} genUnstrMeshStruct;

// General container for FEA meshes
typedef struct {

    // Grid points
    int numNodes;  // Number of nodes in grid
    double *xyz;   // x1 = xyz[0], y1 = xyz[1], z1 = xyz[2], x2 = xyz[3],.... size  = [3*numPoints]
    int    *xyzCoordID; // Grid point coordinate identification. size = [numNodes]
    int    *xyzAttrIndex; // Attribute index for each grid point. size = [numNodes]

    // 1-D

    // Combined 1D PID tracker
    int numElement1DPID;
    int *element1DPID; // size = [numElement1DPID]

    int numRod;// Number of rod elements
    int *crod; // Connectivity list for crod elements. Points of the first rod
               // are crod[0] = [0], [1]; crod[1] = [2], [3], etc.... . size = [2*numRod]
    int *crodPID; // Property identification for crod elements. size = [numRod]
    int *crodElementID; // Element identification for crod element. size[numRod]
    int *crodCoordID; // Coordinate identification for crod element. size[numRod]

    /*
    int numBar;
    int *cbar;

    int numBeam;
    int *cbeam;
    */

    // 2-D

    // Combined 2D PID tracker
    int numElement2DPID;
    int *element2DPID; // size = [numElement2DPID]

    int numTria3;  // Number of tria3 elements
    int *ctria3;   // Connectivity list for ctria3 elements. Corner of the first tria
    			   // are ctria3[0] = [0], [1], [2]; ctria3[1] = [3], [4], [5], etc.... . size = [3*numTria3]
    int *ctria3PID;// Property identification for ctria3 elements. size = [numTria3]
    int *ctria3ElementID; // Element identification for ctria3 element. size[numTria3]
    int *ctria3CoordID; // Coordinate identification for ctria3 element. size[numTria3]

    //int *ctria3T;  // What the hell was it supposed to be?

    int numQuad4;   // Number of cquad4 elements
    int *cquad4;    // Connectivity list for cquad4 elements. Corner of the first quad
    			    // are cquad4[0] = [0], [1], [2], [3]; cquad4[1] = [4], [5], [6], [7], etc.... . size = [4*numQuad4]
    int *cquad4PID; // Property identification for cquad4 elements. size = [numQuad4]
    int *cquad4ElementID; // Element identification for cquad4 element. size[numQuad4]
    int *cquad4CoordID; // Coordinate identification for cquad4 element. size[numQuad4]

    //int *cquad4T;   // What the hell was this supposed to be?

/*
    int numShear;
    int *cshear;
    int *cshearPID;
    int *cshearT;

    int numQuad8;
    int *cquad8;
    int *cquad8PID;
    int *cquad8T;

    int numTria6;
    int *ctria6;
    int *ctria6PID;
    int *ctria6T;
*/
    // 3-D

    // Combined 3D PID tracker
    int numElement3DPID;
    int *element3DPID; // size = [numElement3DPID]

    int numHexa;   // Number of hexahedral elements
    int *chexa;    // Connectivity list for chexa .  size = [8*numHexa]
    int *chexaPID; // Property identification for chexa elements. size = [numHexa]
    int *chexaElementID; // Element identification for chexa element. size[numHexa]

/*
    int numPenta;
    int *cpenta;
    int *cpentaPID;
*/
    int numTetra; // Number of tetrahedral elements
    int *ctetra;  // Connectivity list of ctetra. size = [4*numTetra]
    int *ctetraPID; // Property identification of ctetra elements. size = [numTetra]
    int *ctetraElementID; // Element identification for ctetra element. size[numTetra]
/*

    int numHexaHO; // High Order
    int *chexaHO;
    int *chexaHOPID;

    int numPentaHO;
    int *cpentaHO;
    int *cpentaHOPID;

    int numTetraHO;
    int *ctetraHO;
    int *ctetraHOPID;
*/

    // This will probably be removed in the future
    int numMesh; // Number of surfaces lumped together

    bodyTessMappingStruct **bodyTessMappingLump; // size = [numMesh]

    bodyTessMappingStruct bodyTessMapping;


} genFEAMeshStruct;

// Container for boundary conditions
typedef struct {
    int numBND;
    int *bndID;
    int *bcVal;
} bndCondStruct;

// Container for Tetgen specific inputs
typedef struct {
    double meshQuality_rad_edge;// Tetgen: maximum radius-edge ratio
    double meshQuality_angle;   // Tetgen: minimum dihedral angle
    char  *meshInputString;     // Tetgen: Input string (optional) if NULL use default values
    int   verbose;              // 0 = False, anything else True - Verbose output from mesh generator
    int   ignoreSurfaceExtract; // 0 = False, anything else True - Don't extract the new surface mesh if
                                //      Steiner points are added.
    double meshTolerance; // Tetgen : mesh tolerance
} tetgenInputStruct;


// Container for AFLR3 specific inputs
typedef struct {
    char  *meshInputString;    // AFLR3: Input string (optional) if NULL use default values
} aflr3InputStruct;

// Container for AFLR4 specific inputs
typedef struct {
    char  *meshInputString;    // AFLR4: Input string (optional) if NULL use default values
} aflr4InputStruct;


// Container for meshing inputs
typedef struct {
    double paramTess[3]; /*a set of 3 parameters that drive the EDGE discretization and the
                           FACE triangulation. The first is the maximum length of an EDGE
                           segment or triangle side (in physical space). A zero is flag that allows
                           for any length. The second is a curvature-based value that looks
                           locally at the deviation between the centroid of the discrete object and
                           the underlying geometry. Any deviation larger than the input value will
                           cause the tessellation to be enhanced in those regions. The third is
                           the maximum interior dihedral angle (in degrees) between triangle
                           facets (or Edge segment tangents for a WIREBODY tessellation),
                           note that a zero ignores this phase*/

    int preserveSurfMesh; // 0 = False , anything else True - Use the body tessellation as the surface mesh

    int quiet;            // 0 = False , anything else True - No output from mesh generator
    char *outputFormat;   // Mesh output formats - AFLR3, TECPLOT, VTK, SU2
    char *outputFileName; // Filename prefix for mesh
    char *outputDirectory;// Directory to write mesh to
    int outputASCIIFlag;  // 0 = Binary output, anything else for ASCII

    bndCondStruct bndConds; // Structure of boundary conditions
    tetgenInputStruct tetgenInput; // Structure of Tetgen specific inputs
    aflr3InputStruct aflr3Input; // Structure of AFLR3 specific inputs
    aflr4InputStruct aflr4Input; // Structure of AFLR4 specific inputs

} meshInputStruct;

typedef struct {
	char *name; // Attribute name
	int attrIndex;  // Attribute index

	int numEdgePoints; // Number of points along an edge

	edgeDistributionEnum edgeDistribution; // Distribution function along an edge

	double initialNodeSpacing[2]; // Initial node spacing along an edge

	int useTessParams; // Trigger to use the specified face tessellation parameters
	double tessParams[3]; // Tessellation parameters on a face.

	double boundaryLayerThickness; // Boundary layer thickness on a face (3D) or edge (2D)
	double boundaryLayerSpacing; // Boundary layer spacing on a face (3D) or edge (2D)

} meshSizingStruct;

#endif
