// Mesh related utility functions - Written by Dr. Ryan Durscher AFRL/RQVC

#include "egads.h"
#include <math.h>
#include <string.h>
#include <stdio.h>

#ifdef WIN32
#define strcasecmp  stricmp
#endif

#include "meshTypes.h"  // Bring in mesh structurs
#include "capsTypes.h"  // Bring in CAPS types
#include "cfdTypes.h"   // Bring in cfd structures
#include "feaTypes.h"   // Bring in fea structures


#include "cfdUtils.h"
#include "meshUtils.h"
#include "miscUtils.h"

int mesh_bodyTessellation(ego tess, mapAttrToIndexStruct *attrMap, int *numNodes, double *xyzCoord[],
         	 	 	 	  int *numTriFace, int *triFaceConn[], int *triFaceCompID[],
						  int *numBndEdge, int *bndEdgeConn[], int *bndEdgeCompID[],
						  int *tessFaceQuadMap,
						  int *numQuadFace, int *quadFaceConn[], int *quadFaceCompID[])
{

	/*
		 * 	calculates and returns a complete Body tessellation
		 *
		 * 	In : tess	 - ego of a body tessellation
		 *       attrMap - pointer to an attribute to index mapping for component
		 *		 tessFaceQuadMap - List to keeping track of whether or not the tessObj has quads that
		 *		 				have been split into tris. size = [numFace on Tess]. Order of index is assumed
		 *		 				to correspond to the order of faces.
		 *		 				In general if the quads have been split they were added to the end
		 *		 				of the tri-list in the face tessellation. CAN be NULL.
		 *
		 *
		 *  Out :
		 *       numNodes   - number of vertices (xyz coordinates)
		 *       xyzCoord   - coordinates, 3*numPoints in length -- freeable
		 * 		 numTriFace - number of triangles
		 * 		 triFaceConn- triangle connectivity, 3*numTriFace in length -- freeable
		 * 		 triFaceCompID - triangle boundary marking index found in attrMap, numTriFace in length --freeable
		 * 		 numBndEdge - number of boundary segments if body only has 1 face - 2D meshing
		 * 		 bndEdgeConn- edge segment connectivity, 2*numBndEdge in length -- freeable
		 * 		 bndEdgeCompID - edge segment boundary marking index found in attrMap, numBndEdge in length --freeable
		 *
		 *		 numQuadFace - number of quadrilaterals - tessFaceQuadMap be none NULL for this information to be extracted
		 *		 quadFaceConn - quadrilateral connectivity, 4*numQuadFace in length -- freable
		 *		 quadFaceCompID - quadrilateral boundary marking index found in attrMap, numQuadFace in length --freeable
		 */

	int status; // Function return

	int i, j, face, edge, offSetIndex; // Indexing

	// EGADS function return variables
	int           plen = 0, tlen = 0;
	const int    *tris = NULL, *triNeighbor = NULL, *ptype = NULL, *pindex = NULL;
	const double *points = NULL, *uv = NULL;

	int 	*triConn = NULL, *quadConn = NULL;
	double  *xyzs = NULL;

	// Totals
	int numFace = 0, numEdge = 0, numPoints = 0, numTri =0, numQuad = 0, numEdgeSeg = 0;

	// EGO objects
	ego *faces = NULL, *edges = NULL;
	egTessel *btess = NULL;

	int *globalID = NULL, gID = 0, foundID = (int) false; // Global ID tracker

	int  cID = 0, *compID = NULL, *quadCompID = NULL; // Component marker

	const char *groupName = NULL;

	*numNodes  = 0;
	*numTriFace = 0;
	*xyzCoord  = NULL;
	*triFaceConn = NULL;
	*triFaceCompID = NULL;

	*numBndEdge = 0;
	*bndEdgeConn = NULL;
	*bndEdgeCompID = NULL;

	*numQuadFace = 0;
	*quadFaceConn = NULL;
	*quadFaceCompID = NULL;

	if (tess == NULL)                 return EGADS_NULLOBJ;
	if (tess->magicnumber != MAGIC)   return EGADS_NOTOBJ;
	if (tess->oclass != TESSELLATION) return EGADS_NOTTESS;
	if (tess->blind == NULL)          return EGADS_NOTFOUND;


	btess   = (egTessel *) tess->blind;

	// Get faces and edges so we can check for attributes on them
	status = EG_getBodyTopos(btess->src, NULL, FACE, &numFace, &faces);
	if (status != EGADS_SUCCESS) {
		printf(" Error: EG_getBodyTopos = %d!\n", status);
		goto cleanup;
	}

	if (numFace == 1) {
		status = EG_getBodyTopos(btess->src, NULL, EDGE, &numEdge, &edges);
		if (status != EGADS_SUCCESS) {
			printf(" Error: EG_getBodyTopos = %d!\n", status);
			goto cleanup;
		}
	}

	numPoints = numTri = numEdgeSeg = 0;
	for (face = 1; face <= numFace; face++) {
		status = EG_getTessFace(tess, face, &plen, &points, &uv, &ptype, &pindex,
								&tlen, &tris, &triNeighbor);

		if (status != EGADS_SUCCESS) {
			printf(" Face %d: EG_getTessFace status = %d (bodyTessellation)!\n", face, status);
			goto cleanup;
		}

		numPoints += plen;
		numTri += tlen;

		if (numFace == 1) {

			for (edge = 1; edge <= numEdge; edge++) {

				status = EG_getTessEdge(tess, edge, &plen, &points, &uv);
				if (status != EGADS_SUCCESS) goto cleanup;

				numEdgeSeg += plen-1;

			}
		}
	}

	// Allocate memory associated with the nodes

	xyzs = (double *) EG_alloc(3*numPoints*sizeof(double));
	if (xyzs == NULL) {
		printf(" Error: Can not allocate XYZs (bodyTessellation)!\n");
		status = EGADS_MALLOC;
		goto cleanup;
	}

	globalID = (int *) EG_alloc(numPoints*sizeof(int));
	if (globalID == NULL) {
		printf(" Error: Can not allocate node component ID (bodyTessellation)!\n");
		status = EGADS_MALLOC;
		goto cleanup;
	}

	// Allocate memory associated with triangles

	triConn = (int *) EG_alloc(3*numTri*sizeof(int));
	if (triConn == NULL) {
		printf(" Error: Can not allocate triangles (bodyTessellation)!\n");
		status = EGADS_MALLOC;
		goto cleanup;
	}

	compID = (int *) EG_alloc(numTri*sizeof(int));
	if (compID == NULL) {
		printf(" Error: Can not allocate components (bodyTessellation)!\n");
		status = EGADS_MALLOC;
		goto cleanup;
	}

	// Do we have split quads?
	if (tessFaceQuadMap != NULL) {

		// How many?
		numQuad = 0;
		for (face = 0; face < numFace; face++) numQuad += tessFaceQuadMap[face];

		// Allocate quad arrays
		if (numQuad != 0) {
			quadConn = (int *) EG_alloc(4*numQuad*sizeof(int));
			if (quadConn == NULL) {
				printf(" Error: Can not allocate quadrilaterals (bodyTessellation)!\n");
				status = EGADS_MALLOC;
				goto cleanup;
			}

			quadCompID = (int *) EG_alloc(numQuad*sizeof(int));
			if (quadCompID == NULL) {
				printf(" Error: Can not allocate quad components (bodyTessellation)!\n");
				status = EGADS_MALLOC;
				goto cleanup;
			}

			for (i = 0; i < numQuad; i++) quadCompID[i] = 1;
		}
	}

	//Set default value for compID
	for (i = 0; i < numTri; i++) compID[i] = 1;


	// Loop through faces and build global xyz and connectivity
	numPoints = 0;
	numTri = 0;
	numQuad = 0;
	for (face = 1; face <= numFace; face++) {

		// Look for component/boundary ID for attribute mapper based  on capsGroup

		status = retrieve_CAPSGroupAttr(faces[face-1], &groupName);
		if (status != CAPS_SUCCESS) {
			printf("Error: No capsGroup attribute found on Face %d, unable to assign a boundary index value\n",face-1);
			goto cleanup;
		}

		status = get_mapAttrToIndexIndex(attrMap, groupName, &cID);
		if (status != CAPS_SUCCESS) {
			printf("Error: Unable to retrieve boundary index from capsGroup %s\n", groupName);
			goto cleanup;
		}

		status = EG_getTessFace(tess, face, &plen, &points, &uv, &ptype, &pindex,
								&tlen, &tris, &triNeighbor);
		if (status != EGADS_SUCCESS) continue;

		// Get xyz in global senses
		for (i = 0; i < plen; i++) {

			status = EG_localToGlobal(tess, face, i+1, &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			foundID = (int) false;

			// For not interior points see if you have already saved away this point
			if (ptype[i] != -1) {

				// Look for
				for (j = 0; j < numPoints; j++) {

					if ( gID ==  globalID[j]) {
						foundID = (int) true;
						break;
					}
				}
			}

			if (foundID == (int) true) continue;

			xyzs[3*numPoints  ]  = points[3*i  ];
			xyzs[3*numPoints+1]  = points[3*i+1];
			xyzs[3*numPoints+2]  = points[3*i+2];

			globalID[numPoints] = gID;

			//nCompID[npts] = cID;

			numPoints += 1;
		}

		// Do we possibly have quads?
		if (tessFaceQuadMap != NULL) {
			tlen = tlen - 2*tessFaceQuadMap[face-1]; // Get the number of "true" triangles

			// Get connectivity in global sense
			offSetIndex = 3*tlen;
			for (i = 0; i < tessFaceQuadMap[face-1]; i++){
				status = EG_localToGlobal(tess, face, tris[3*i + offSetIndex + 0], &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				quadConn[4*numQuad + 0] = gID;

				status = EG_localToGlobal(tess, face, tris[3*i + offSetIndex + 1], &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				quadConn[4*numQuad + 1] = gID;


				status = EG_localToGlobal(tess, face, tris[3*i + offSetIndex + 2], &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				quadConn[4*numQuad + 3] = gID;

				offSetIndex += 3;

				status = EG_localToGlobal(tess, face, tris[3*i + offSetIndex + 1], &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				quadConn[4*numQuad + 2] = gID;

				quadCompID[numQuad] = cID;

				numQuad +=1;
			}
		}

		// Get triangle connectivity in global sense
		for (i = 0; i < tlen; i++) {

			status = EG_localToGlobal(tess, face, tris[3*i + 0], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			triConn[3*numTri + 0] = gID;

			status = EG_localToGlobal(tess, face, tris[3*i + 1], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			triConn[3*numTri + 1] = gID;

			status = EG_localToGlobal(tess, face, tris[3*i + 2], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			triConn[3*numTri + 2] = gID;

			compID[numTri] = cID;

			numTri += 1;
		}

	}

	// Resize the xyz array
	xyzs = (double *) EG_reall(xyzs, 3*numPoints*sizeof(double));
	if (xyzs == NULL) {
		status =  EGADS_MALLOC;
		goto cleanup;
	}

	// Resize triConn and compId (triFace) if we ended up have some quad faces
	if (numQuad != 0) {

		triConn = (int *) EG_reall(triConn, 3*numTri*sizeof(int));
		if (triConn == NULL) {
			status = EGADS_MALLOC;
			goto cleanup;
		}

		compID = (int *) EG_reall(compID, numTri*sizeof(int));
		if (compID == NULL) {
			status = EGADS_MALLOC;
			goto cleanup;
		}
	}

	// Get boundary edge information if we only have 1 face
	if ( numFace == 1 && numEdge != 0) {

		*numBndEdge = numEdgeSeg;

		*bndEdgeConn = (int *) EG_alloc(2*numEdgeSeg*sizeof(int));
		if (bndEdgeConn == NULL) {
			printf(" Error: Can not allocate components (bodyTessellation)!\n");
			status = EGADS_MALLOC;
			goto cleanup;
		}

		*bndEdgeCompID = (int *) EG_alloc(numEdgeSeg*sizeof(int));
		if (bndEdgeCompID == NULL) {
			printf(" Error: Can not allocate components (bodyTessellation)!\n");
			status = EGADS_MALLOC;
			goto cleanup;
		}

		numEdgeSeg = 0;
		// Fill up boundary edge list
		for (edge = 1; edge <= numEdge; edge++) {

			status = retrieve_CAPSGroupAttr(edges[edge-1], &groupName);
			if (status != CAPS_SUCCESS) continue;

			status = get_mapAttrToIndexIndex(attrMap, groupName, &cID);
			if (status != CAPS_SUCCESS) continue;

			status = EG_getTessEdge(tess, edge, &plen, &points, &uv);
			if (status != EGADS_SUCCESS) goto cleanup;

			for (i = 0; i < plen -1; i++) {

				status = EG_localToGlobal(tess, -edge, i+1, &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				(*bndEdgeConn)[2*numEdgeSeg + 0] = gID;

				status = EG_localToGlobal(tess, -edge, i+2, &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				(*bndEdgeConn)[2*numEdgeSeg + 1] = gID;

				(*bndEdgeCompID)[numEdgeSeg] = cID;

				numEdgeSeg += 1;
			}
		}
	}


	*numNodes     = numPoints;
	*xyzCoord      = xyzs;
	*numTriFace    = numTri;
	*triFaceConn   = triConn;
	*triFaceCompID = compID;

	*numQuadFace = numQuad;
	*quadFaceConn = quadConn;
	*quadFaceCompID = quadCompID;

	status = CAPS_SUCCESS;

	goto cleanup;

	cleanup:
		if (status != CAPS_SUCCESS) {
			if (xyzs != NULL) EG_free(xyzs);
			if (triConn != NULL) EG_free(triConn);
			if (compID != NULL) EG_free(compID);

			if (quadConn != NULL) EG_free(quadConn);
			if (quadCompID != NULL) EG_free(quadCompID);

			*numBndEdge = 0;

			if (*bndEdgeConn != NULL) EG_free(*bndEdgeConn);
			*bndEdgeConn = NULL;

			if (*bndEdgeCompID != NULL) EG_free(*bndEdgeCompID);
			*bndEdgeCompID = NULL;
		}

		if (globalID != NULL) EG_free(globalID);

		if (faces != NULL) EG_free(faces);
		if (edges != NULL) EG_free(edges);

		return status;
}



// Create a surface mesh in genUnstrSurfMeshStruct format using the EGADS body tessellation
int mesh_genUnstrSurfaceMeshEGADS(ego body, double tessParams[3], mapAttrToIndexStruct *attrMap, genUnstrSurfMeshStruct *surfMesh) {

	int status; // Function return integer

	// Meshing related variables
	double  box[6], size;
	double  params[3];

	// Get Tessellation
	printf("\tTessellating body\n");

	// Get bounding box for scaling
	status = EG_getBoundingBox(body, box);
	if (status != EGADS_SUCCESS) {
		printf(" EG_getBoundingBox = %d\n\n", status);
		return status;
	}

	size = sqrt((box[0]-box[3])*(box[0]-box[3]) +
			    (box[1]-box[4])*(box[1]-box[4]) +
				(box[2]-box[5])*(box[2]-box[5]));

	params[0] = tessParams[0]*size;
	params[1] = tessParams[1]*size;
	params[2] = tessParams[2];

	status = EG_makeTessBody(body, params, &surfMesh->bodyTessMapping.egadsTess);
	if (status != EGADS_SUCCESS) {
		printf(" EG_makeTessBody = %d\n", status);
		return status;
	}

	status = mesh_bodyTessellation(surfMesh->bodyTessMapping.egadsTess,
								   attrMap,
								   &surfMesh->numNodes,
								   &surfMesh->xyz,
								   &surfMesh->numTriFace,
								   &surfMesh->localTriFaceList,
								   &surfMesh->triFaceMarkList,
								   &surfMesh->numBoundaryEdge,
								   &surfMesh->localBoundaryEdgeList,
								   &surfMesh->boundaryEdgeMarkList,
								   NULL,
								   &surfMesh->numQuadFace,
								   &surfMesh->localQuadFaceList,
								   &surfMesh->quadFaceMarkList);

	if (status != CAPS_SUCCESS) return status;

	return CAPS_SUCCESS;
}

// Lump multiple surface meshes in genUnstrSurfMeshStruct format into a single surface mesh in genUnstrSurfMeshStruct format
int lump_genUnstrSurfaceMesh(int numSurfMesh, genUnstrSurfMeshStruct surfMesh[], genUnstrSurfMeshStruct *surfMeshComb) {

	// Default values get set to 0!

    int i, surf; // Indexing

    int foundTriFaceList         = (int) false; // Boolean testers
    int foundQuadFaceList        = (int) false;

    int foundNodeLocal2Global 	 = (int) false;
    int foundLocalTriFaceList    = (int) false;
    int foundLocalQuadFaceList   = (int) false;
    int foundTriFaceMarkList     = (int) false;
    int foundQuadFaceMarkList    = (int) false;
    int foundTriFaceBLList       = (int) false;
    int foundQuadFaceBLList      = (int) false;
    int foundNodeBLSpacingList   = (int) false;
    int foundNodeBLThicknessList = (int) false;

    int numTriFace = 0;
    int numQuadFace = 0;
    int numNodes = 0;

    int offSet = 0, offSetConn = 0; // Indexing offsets

    printf("Combining multiple surface meshes into a single surface mesh\n");

    // Surface un-lumping arrays
    surfMeshComb->numSurface = numSurfMesh;
    surfMeshComb->surfNodesLump    = (int *) EG_alloc( surfMeshComb->numSurface * sizeof(int));
    surfMeshComb->surfTriFaceLump  = (int *) EG_alloc( surfMeshComb->numSurface * sizeof(int));
    surfMeshComb->surfQuadFaceLump = (int *) EG_alloc( surfMeshComb->numSurface * sizeof(int));

    surfMeshComb->bodyTessMappingLump = (bodyTessMappingStruct **) EG_alloc(surfMeshComb->numSurface * sizeof(bodyTessMappingStruct *));

    if ((surfMeshComb->surfNodesLump == NULL)    ||
    	(surfMeshComb->surfTriFaceLump == NULL)  ||
        (surfMeshComb->surfQuadFaceLump == NULL) ||
		(surfMeshComb->bodyTessMappingLump == NULL)) {
        return EGADS_MALLOC;
    }

	// Loop through surfaces
    for (surf = 0; surf < numSurfMesh; surf++) {

        // Increment and set totals
    	surfMeshComb->numNodes += surfMesh[surf].numNodes;
    	surfMeshComb->numTriFace += surfMesh[surf].numTriFace;
    	surfMeshComb->numQuadFace += surfMesh[surf].numQuadFace;

    	// Array used to un-lump surfaces
    	surfMeshComb->surfNodesLump[surf]    = surfMesh[surf].numNodes;
    	surfMeshComb->surfTriFaceLump[surf]  = surfMesh[surf].numTriFace;
    	surfMeshComb->surfQuadFaceLump[surf] = surfMesh[surf].numQuadFace;
    }

    // Allocate array
    if (surfMeshComb->numNodes > 0) {
    	surfMeshComb->xyz = (double *) EG_alloc( 3*surfMeshComb->numNodes * sizeof(double));
    	if (surfMeshComb->xyz == NULL) return EGADS_MALLOC;
    } else {
    	surfMeshComb->xyz = NULL;
    }



    // Loop through surfaces to test if other arrays are not NULL
    foundTriFaceList = (int) false;
    foundQuadFaceList = (int) false;

    foundNodeLocal2Global = (int) false;

    foundLocalTriFaceList  = (int) false;
    foundLocalQuadFaceList = (int) false;

    foundTriFaceMarkList   = (int) false;
    foundQuadFaceMarkList  = (int) false;

    foundTriFaceBLList = (int) false;
    foundQuadFaceBLList = (int) false;

    foundNodeBLSpacingList = (int) false;
    foundNodeBLThicknessList = (int) false;

    for (surf = 0; surf < numSurfMesh; surf++) {

    	if (surfMesh[surf].triFaceList  != NULL) foundTriFaceList  = (int) true;
    	if (surfMesh[surf].quadFaceList != NULL) foundQuadFaceList = (int) true;

    	if (surfMesh[surf].localTriFaceList  != NULL) foundLocalTriFaceList  = (int) true;
    	if (surfMesh[surf].localQuadFaceList != NULL) foundLocalQuadFaceList = (int) true;

    	if (surfMesh[surf].triFaceMarkList  != NULL)  foundTriFaceMarkList  = (int) true;
    	if (surfMesh[surf].quadFaceMarkList != NULL)  foundQuadFaceMarkList = (int) true;

    	if (surfMesh[surf].triFaceBLList  != NULL)   foundTriFaceBLList  = (int) true;
    	if (surfMesh[surf].quadFaceBLList != NULL)   foundQuadFaceBLList = (int) true;
    }

    if (foundTriFaceList == (int) true && surfMeshComb->numTriFace > 0) {
    	surfMeshComb->triFaceList = (int *) EG_alloc( 3*surfMeshComb->numTriFace * sizeof(int));
    	if (surfMeshComb->triFaceList == NULL) return EGADS_MALLOC;
    } else {
    	surfMeshComb->triFaceList = NULL;
    }

    if (foundQuadFaceList == (int) true && surfMeshComb->numQuadFace > 0) {
    	surfMeshComb->quadFaceList = (int *) EG_alloc( 4*surfMeshComb->numQuadFace * sizeof(int));
    	if (surfMeshComb->quadFaceList == NULL) return EGADS_MALLOC;
    } else {
    	surfMeshComb->quadFaceList = NULL;
    }

	for (surf = 0; surf < numSurfMesh; surf++) {
		if (surfMesh[surf].nodeLocal2Global == NULL && foundNodeLocal2Global == (int) true) {
			printf("nodeLocal2Global must be non-NULL for all surfaces in order to lump nodeLocal2Global arrays, surface - %d!\n", surf);
			foundNodeLocal2Global   = (int) false;
			break;
		} else if (surfMesh[surf].nodeLocal2Global != NULL) foundNodeLocal2Global   = (int) true;
	}


    for (surf = 0; surf < numSurfMesh; surf++) {
    	if (surfMesh[surf].nodeBLSpacingList == NULL && foundNodeBLSpacingList == (int) true) {
    		printf("nodeBLSpacingList must be non-NULL for all surfaces in order to lump nodeBLSpacingList arrays, surface - %d!\n", surf);
    		foundNodeBLSpacingList   = (int) false;
    		break;

    	} else if (surfMesh[surf].nodeBLSpacingList != NULL) foundNodeBLSpacingList   = (int) true;
    }

    for (surf = 0; surf < numSurfMesh; surf++) {
    	if (surfMesh[surf].nodeBLThicknessList == NULL && foundNodeBLThicknessList == (int) true) {
    		printf("nodeBLThicknessList must be non-NULL for all surfaces in order to lump nodeBLThicknessList arrays, surface - %d!\n", surf);
    		foundNodeBLThicknessList = (int) false;
    		break;
    	} else if (surfMesh[surf].nodeBLThicknessList != NULL ) foundNodeBLThicknessList = (int) true;
    }

    if (foundLocalTriFaceList == (int) true && surfMeshComb->numTriFace > 0) {
    	surfMeshComb->localTriFaceList = (int *) EG_alloc(3*surfMeshComb->numTriFace  * sizeof(int));
    	if (surfMeshComb->localTriFaceList == NULL) return EGADS_MALLOC;
    } else {
    	surfMeshComb->localTriFaceList = NULL;
    }

    if (foundLocalQuadFaceList == (int) true && surfMeshComb->numQuadFace > 0) {
    	surfMeshComb->localQuadFaceList = (int *) EG_alloc(4*surfMeshComb->numQuadFace * sizeof(int));
    	if (surfMeshComb->localQuadFaceList == NULL) return EGADS_MALLOC;
    } else {
    	surfMeshComb->localQuadFaceList = NULL;
    }

    if (foundTriFaceMarkList == (int) true && surfMeshComb->numTriFace > 0) {
    	surfMeshComb->triFaceMarkList = (int *) EG_alloc( surfMeshComb->numTriFace  * sizeof(int));
    	if (surfMeshComb->triFaceMarkList == NULL) return EGADS_MALLOC;
    } else {
    	surfMeshComb->triFaceMarkList = NULL;
    }

    if (foundQuadFaceMarkList == (int) true && surfMeshComb->numQuadFace > 0) {
    	surfMeshComb->quadFaceMarkList = (int *) EG_alloc(surfMeshComb->numQuadFace  * sizeof(int));
    	if (surfMeshComb->quadFaceMarkList == NULL) return EGADS_MALLOC;
    } else {
    	surfMeshComb->quadFaceMarkList = NULL;
    }

    if (foundTriFaceBLList == (int) true && surfMeshComb->numTriFace > 0) {
    	surfMeshComb->triFaceBLList = (int *) EG_alloc( surfMeshComb->numTriFace * sizeof(int));
    	if (surfMeshComb->triFaceBLList == NULL) return EGADS_MALLOC;
    } else {
    	surfMeshComb->triFaceBLList = NULL;
    }

    if (foundQuadFaceBLList == (int) true && surfMeshComb->numQuadFace > 0) {
    	surfMeshComb->quadFaceBLList = (int *) EG_alloc( surfMeshComb->numQuadFace * sizeof(int));
    	if (surfMeshComb->quadFaceBLList == NULL) return EGADS_MALLOC;
    } else {
    	surfMeshComb->quadFaceBLList = NULL;
    }

    if (foundNodeLocal2Global == (int) true && surfMeshComb->numNodes > 0) {
    	surfMeshComb->nodeLocal2Global = (int *) EG_alloc(surfMeshComb->numNodes * sizeof(int));
    	if (surfMeshComb->nodeLocal2Global == NULL) return EGADS_MALLOC;
    } else {
    	surfMeshComb->nodeLocal2Global = NULL;
    }

    if (foundNodeBLSpacingList == (int) true && surfMeshComb->numNodes > 0) {
    	surfMeshComb->nodeBLSpacingList = (double *) EG_alloc(surfMeshComb->numNodes * sizeof(double));
    	if (surfMeshComb->nodeBLSpacingList == NULL) return EGADS_MALLOC;
    } else {
    	surfMeshComb->nodeBLSpacingList = NULL;
    }

    if (foundNodeBLThicknessList == (int) true && surfMeshComb->numNodes > 0) {
    	surfMeshComb->nodeBLThicknessList = (double *) EG_alloc(surfMeshComb->numNodes * sizeof(double));
    	if (surfMeshComb->nodeBLThicknessList == NULL) return EGADS_MALLOC;
    } else {
    	surfMeshComb->nodeBLThicknessList = NULL;
    }

    for (surf = 0; surf < numSurfMesh; surf++) {
        // Increment totals
     	numNodes += surfMesh[surf].numNodes;
     	numTriFace += surfMesh[surf].numTriFace;
     	numQuadFace += surfMesh[surf].numQuadFace;

    	// Fill arrays - if a surface mesh array is NULL, free and NULL the combined array
    	if (surfMesh[surf].xyz != NULL && surfMeshComb->xyz != NULL) {

    		offSet = 3*(numNodes-surfMesh[surf].numNodes);

    		memcpy(surfMeshComb->xyz + offSet,
    			   surfMesh[surf].xyz,
				   3*surfMesh[surf].numNodes*sizeof(double));
    	} else {
    		printf("NULL array of xyz coordinates found for surface %d!!\n",surf);
    		EG_free(surfMeshComb->xyz);
    		surfMeshComb->xyz = NULL;
    	}

     	if (surfMesh[surf].nodeLocal2Global != NULL && surfMeshComb->nodeLocal2Global != NULL) {

    		offSet = (numNodes-surfMesh[surf].numNodes);

    		/// I dont think this is right if you are using this you shouldn't have multiple surfaces
    		// If coming in this function nodeLocal2Global is most likely NULL
    		memcpy(surfMeshComb->nodeLocal2Global + offSet,
    			   surfMesh[surf].nodeLocal2Global,
				   surfMesh[surf].numNodes*sizeof(int));

    	} /*else {
    		EG_free(surfMeshComb->nodeLocal2Global);
    		surfMeshComb->nodeLocal2Global = NULL;
    	}*/


     	// Set connectivity offset
     	offSetConn = (numNodes-surfMesh[surf].numNodes);

     	if (surfMesh[surf].triFaceList != NULL && surfMeshComb->triFaceList != NULL) {

     		offSet = 3*(numTriFace-surfMesh[surf].numTriFace);


     		for (i = 0; i < surfMesh[surf].numTriFace; i++) {
     			surfMeshComb->triFaceList[3*i + 0 + offSet] = surfMesh[surf].triFaceList[3*i + 0] + offSetConn;
     			surfMeshComb->triFaceList[3*i + 1 + offSet] = surfMesh[surf].triFaceList[3*i + 1] + offSetConn;
     			surfMeshComb->triFaceList[3*i + 2 + offSet] = surfMesh[surf].triFaceList[3*i + 2] + offSetConn;
     		}

     	}/* else {
    		EG_free(surfMeshComb->triFaceList);
    		surfMeshComb->triFaceList = NULL;
    	}*/

     	if (surfMesh[surf].quadFaceList != NULL && surfMeshComb->quadFaceList != NULL) {

     		offSet = 4*(numQuadFace-surfMesh[surf].numQuadFace);

     		for (i = 0; i < surfMesh[surf].numQuadFace; i++) {
         			surfMeshComb->quadFaceList[4*i + 0 + offSet] = surfMesh[surf].quadFaceList[4*i + 0] + offSetConn;
         			surfMeshComb->quadFaceList[4*i + 1 + offSet] = surfMesh[surf].quadFaceList[4*i + 1] + offSetConn;
         			surfMeshComb->quadFaceList[4*i + 2 + offSet] = surfMesh[surf].quadFaceList[4*i + 2] + offSetConn;
         			surfMeshComb->quadFaceList[4*i + 3 + offSet] = surfMesh[surf].quadFaceList[4*i + 3] + offSetConn;
     		}

     	}/* else {
    		EG_free(surfMeshComb->quadFaceList);
    		surfMeshComb->quadFaceList = NULL;
    	}*/

     	if (surfMesh[surf].localTriFaceList != NULL && surfMeshComb->localTriFaceList != NULL) {

     		offSet = 3*(numTriFace-surfMesh[surf].numTriFace);

      		for (i = 0; i < surfMesh[surf].numTriFace; i++) {
         			surfMeshComb->localTriFaceList[3*i + 0 + offSet] = surfMesh[surf].localTriFaceList[3*i + 0] + offSetConn;
         			surfMeshComb->localTriFaceList[3*i + 1 + offSet] = surfMesh[surf].localTriFaceList[3*i + 1] + offSetConn;
         			surfMeshComb->localTriFaceList[3*i + 2 + offSet] = surfMesh[surf].localTriFaceList[3*i + 2] + offSetConn;
         	}

     	}/* else {
    		EG_free(surfMeshComb->localTriFaceList);
    		surfMeshComb->localTriFaceList = NULL;
    	}*/

     	if (surfMesh[surf].localQuadFaceList != NULL && surfMeshComb->localQuadFaceList != NULL) {

     		offSet = 4*(numQuadFace-surfMesh[surf].numQuadFace);

      		for (i = 0; i < surfMesh[surf].numQuadFace; i++) {
      			surfMeshComb->localQuadFaceList[4*i + 0 + offSet] = surfMesh[surf].localQuadFaceList[4*i + 0] + offSetConn;
      			surfMeshComb->localQuadFaceList[4*i + 1 + offSet] = surfMesh[surf].localQuadFaceList[4*i + 1] + offSetConn;
      			surfMeshComb->localQuadFaceList[4*i + 2 + offSet] = surfMesh[surf].localQuadFaceList[4*i + 2] + offSetConn;
      			surfMeshComb->localQuadFaceList[4*i + 3 + offSet] = surfMesh[surf].localQuadFaceList[4*i + 3] + offSetConn;
      		}
     	}/* else {
    		EG_free(surfMeshComb->localQuadFaceList);
    		surfMeshComb->localQuadFaceList = NULL;
    	}*/

     	if (surfMesh[surf].triFaceMarkList != NULL && surfMeshComb->triFaceMarkList != NULL) {

     		offSet = (numTriFace-surfMesh[surf].numTriFace);

     		memcpy(surfMeshComb->triFaceMarkList + offSet,
     			   surfMesh[surf].triFaceMarkList,
				   surfMesh[surf].numTriFace*sizeof(int));
     	}/* else {
    		EG_free(surfMeshComb->triFaceMarkList);
    		surfMeshComb->triFaceMarkList = NULL;
    	}*/

     	if (surfMesh[surf].quadFaceMarkList != NULL && surfMeshComb->quadFaceMarkList != NULL) {

     		offSet = (numQuadFace-surfMesh[surf].numQuadFace);

     		memcpy(surfMeshComb->quadFaceMarkList + offSet,
     			   surfMesh[surf].quadFaceMarkList,
				   surfMesh[surf].numQuadFace*sizeof(int));
     	}/* else {
    		EG_free(surfMeshComb->quadFaceMarkList);
    		surfMeshComb->quadFaceMarkList = NULL;
    	}*/

     	if (surfMesh[surf].triFaceBLList != NULL && surfMeshComb->triFaceBLList != NULL) {

     		offSet = (numTriFace-surfMesh[surf].numTriFace);

     		memcpy(surfMeshComb->triFaceBLList + offSet,
     			   surfMesh[surf].triFaceBLList,
				   surfMesh[surf].numTriFace*sizeof(int));
     	}/* else {
    		EG_free(surfMeshComb->triFaceBLList);
    		surfMeshComb->triFaceBLList = NULL;
    	}*/

      	if (surfMesh[surf].quadFaceBLList != NULL && surfMeshComb->quadFaceBLList != NULL) {

     		offSet = (numQuadFace-surfMesh[surf].numQuadFace);

     		memcpy(surfMeshComb->quadFaceBLList + offSet,
     			   surfMesh[surf].quadFaceBLList,
				   surfMesh[surf].numQuadFace*sizeof(int));
     	}/* else {
    		EG_free(surfMeshComb->quadFaceBLList);
    		surfMeshComb->quadFaceBLList = NULL;
    	}*/

    	if (surfMesh[surf].nodeBLSpacingList != NULL && surfMeshComb->nodeBLSpacingList != NULL) {

    		offSet = (numNodes-surfMesh[surf].numNodes);

    		memcpy(surfMeshComb->nodeBLSpacingList + offSet,
    			   surfMesh[surf].nodeBLSpacingList,
				   surfMesh[surf].numNodes*sizeof(double));
    	}/* else {
    		EG_free(surfMeshComb->nodeBLSpacingList);
    		surfMeshComb->nodeBLSpacingList = NULL;
    	}*/

    	if (surfMesh[surf].nodeBLThicknessList != NULL && surfMeshComb->nodeBLThicknessList != NULL) {

    		offSet = (numNodes-surfMesh[surf].numNodes);

    		memcpy(surfMeshComb->nodeBLThicknessList + offSet,
    			   surfMesh[surf].nodeBLThicknessList,
				   surfMesh[surf].numNodes*sizeof(double));
    	}/* else {
    		EG_free(surfMeshComb->nodeBLThicknessList);
    		surfMeshComb->nodeBLThicknessList = NULL;
    	}*/

    	// Point the lumped bodyTessMapping pointer that the surface bodyTessMapping
    	surfMeshComb->bodyTessMappingLump[surf] = &surfMesh[surf].bodyTessMapping;
    }

    return CAPS_SUCCESS;
}

// Lump multiple FEA meshes in genFEAMeshStructStruct format into a single mesh
// IMPORTANT: PID numbers are already considered to be with respect to the cumulative bodies
int lump_genFEAMesh(int numMesh, genFEAMeshStruct feaMesh[], genFEAMeshStruct *feaMeshComb) {

    int i, mesh; // Indexing

    int numElement1DPID = 0;
    int numElement2DPID = 0;
    int numElement3DPID = 0;

    int numNodes = 0;
    int numRod = 0;
    int numTria3 = 0;
    int numQuad4 = 0;
    int numHexa = 0;
    int numTetra = 0;

    int offSet = 0, offSetNode = 0, offSetConn = 0; // Indexing offsets

    // Mesh un-lumping arrays

    feaMeshComb->numMesh = numMesh;

    /*
    surfMeshComb->surfNodesLump    = (int *) EG_alloc( surfMeshComb->numSurface * sizeof(int));
    surfMeshComb->surfTriFaceLump  = (int *) EG_alloc( surfMeshComb->numSurface * sizeof(int));
    surfMeshComb->surfQuadFaceLump = (int *) EG_alloc( surfMeshComb->numSurface * sizeof(int));
    if ((surfMeshComb->surfNodesLump == NULL) || (surfMeshComb->surfTriFaceLump == NULL) ||
            (surfMeshComb->surfQuadFaceLump == NULL)) {
        return EGADS_MALLOC;
    }*/


    // If just one mesh we are just copying over values
    if (numMesh > 1) printf("Combining feaMeshes into a single Mesh\n");

    // Loop through feaMeshes
    for (mesh = 0; mesh < numMesh; mesh++) {

        // Increment and set totals
        feaMeshComb->numElement1DPID += feaMesh[mesh].numElement1DPID;
        feaMeshComb->numElement2DPID += feaMesh[mesh].numElement2DPID;
        feaMeshComb->numElement3DPID += feaMesh[mesh].numElement3DPID;

        feaMeshComb->numNodes += feaMesh[mesh].numNodes;
        feaMeshComb->numRod   += feaMesh[mesh].numRod;
        feaMeshComb->numTria3 += feaMesh[mesh].numTria3;
        feaMeshComb->numQuad4 += feaMesh[mesh].numQuad4;
        feaMeshComb->numHexa  += feaMesh[mesh].numHexa;
        feaMeshComb->numTetra += feaMesh[mesh].numTetra;


        // Array used to un-lump surfaces
        //surfMeshComb->surfNodesLump[surf] = surfMesh[surf].numNodes;
        //surfMeshComb->surfTriFaceLump[surf]  = surfMesh[surf].numTriFace;
        //surfMeshComb->surfQuadFaceLump[surf] = surfMesh[surf].numQuadFace;

    }

    // Allocate arrays
    if (feaMeshComb->numElement1DPID > 0) {
        feaMeshComb->element1DPID = (int *) EG_alloc( feaMeshComb->numElement1DPID * sizeof(int));
    } else {
        feaMeshComb->element1DPID = NULL;
    }

    if (feaMeshComb->numElement2DPID > 0) {
        feaMeshComb->element2DPID = (int *) EG_alloc( feaMeshComb->numElement2DPID * sizeof(int));
    } else {
        feaMeshComb->element2DPID = NULL;
    }

    if (feaMeshComb->numElement3DPID > 0) {
        feaMeshComb->element3DPID = (int *) EG_alloc( feaMeshComb->numElement3DPID * sizeof(int));
    } else {
        feaMeshComb->element3DPID = NULL;
    }


    if (feaMeshComb->numNodes > 0) {
        feaMeshComb->xyz          = (double *) EG_alloc( 3*feaMeshComb->numNodes * sizeof(double));
        feaMeshComb->xyzCoordID   = (int *)    EG_alloc(   feaMeshComb->numNodes * sizeof(int));
        feaMeshComb->xyzAttrIndex = (int *)    EG_alloc(   feaMeshComb->numNodes * sizeof(int));
    } else {
        feaMeshComb->xyz          = NULL;
        feaMeshComb->xyzCoordID   = NULL;
        feaMeshComb->xyzAttrIndex = NULL;
    }

    if (feaMeshComb->numRod > 0) {
        feaMeshComb->crod    = (int *) EG_alloc( 2*feaMeshComb->numRod * sizeof(int));
        feaMeshComb->crodPID = (int *) EG_alloc(   feaMeshComb->numRod * sizeof(int));
    } else {
        feaMeshComb->crod    = NULL;
        feaMeshComb->crodPID = NULL;
    }

    if (feaMeshComb->numTria3 > 0) {
        feaMeshComb->ctria3    = (int *) EG_alloc( 3*feaMeshComb->numTria3 * sizeof(int));
        feaMeshComb->ctria3PID = (int *) EG_alloc(   feaMeshComb->numTria3 * sizeof(int));
    } else {
        feaMeshComb->ctria3    = NULL;
        feaMeshComb->ctria3PID = NULL;
    }

    if (feaMeshComb->numQuad4 > 0) {
        feaMeshComb->cquad4    = (int *) EG_alloc( 4*feaMeshComb->numQuad4 * sizeof(int));
        feaMeshComb->cquad4PID = (int *) EG_alloc(   feaMeshComb->numQuad4 * sizeof(int));
    } else {
        feaMeshComb->cquad4    = NULL;
        feaMeshComb->cquad4PID = NULL;
    }

    if (feaMeshComb->numHexa > 0) {
        feaMeshComb->chexa    = (int *) EG_alloc( 8*feaMeshComb->numHexa * sizeof(int));
        feaMeshComb->chexaPID = (int *) EG_alloc(   feaMeshComb->numHexa * sizeof(int));
    } else {
        feaMeshComb->chexa    = NULL;
        feaMeshComb->chexaPID = NULL;
    }

    if (feaMeshComb->numTetra > 0) {
        feaMeshComb->ctetra    = (int *) EG_alloc( 4*feaMeshComb->numTetra * sizeof(int));
        feaMeshComb->ctetraPID = (int *) EG_alloc(   feaMeshComb->numTetra * sizeof(int));
    } else {
        feaMeshComb->ctetra    = NULL;
        feaMeshComb->ctetraPID = NULL;
    }

    // Loop through mesh and fill arrays
    for (mesh = 0; mesh < numMesh; mesh++) {

        // Increment totals
        numElement1DPID += feaMesh[mesh].numElement1DPID;
        numElement2DPID += feaMesh[mesh].numElement2DPID;
        numElement3DPID += feaMesh[mesh].numElement3DPID;

        numNodes += feaMesh[mesh].numNodes;
        numRod   += feaMesh[mesh].numRod;
        numTria3 += feaMesh[mesh].numTria3;
        numQuad4 += feaMesh[mesh].numQuad4;
        numHexa  += feaMesh[mesh].numHexa;
        numTetra += feaMesh[mesh].numTetra;

        if (feaMesh[mesh].xyz != NULL && feaMeshComb->xyz != NULL) {

            offSet = 3*(numNodes-feaMesh[mesh].numNodes);

            for (i = 0; i < feaMesh[mesh].numNodes; i++) {

                feaMeshComb->xyz[3*i+0 + offSet] = feaMesh[mesh].xyz[3*i+0];
                feaMeshComb->xyz[3*i+1 + offSet] = feaMesh[mesh].xyz[3*i+1];
                feaMeshComb->xyz[3*i+2 + offSet] = feaMesh[mesh].xyz[3*i+2];
            }

        }

        if (feaMesh[mesh].xyzCoordID != NULL && feaMeshComb->xyzCoordID != NULL) {

            offSet = (numNodes-feaMesh[mesh].numNodes);

            // May need to offset the CoordID in the future - similar to elementPID
            for (i = 0; i < feaMesh[mesh].numNodes; i++){

                feaMeshComb->xyzCoordID[i + offSet]   = feaMesh[mesh].xyzCoordID[i];
                feaMeshComb->xyzAttrIndex[i + offSet] = feaMesh[mesh].xyzAttrIndex[i];
            }
        }

        if (feaMesh[mesh].element1DPID != NULL && feaMeshComb->element1DPID != NULL) {

            offSet           = (numElement1DPID - feaMesh[mesh].numElement1DPID);

            for (i = 0; i < feaMesh[mesh].numElement1DPID; i++){

                feaMeshComb->element1DPID[i + offSet] = feaMesh[mesh].element1DPID[i];
            }
        }

        if (feaMesh[mesh].element2DPID != NULL && feaMeshComb->element2DPID != NULL) {

            offSet           = (numElement2DPID - feaMesh[mesh].numElement2DPID);

            for (i = 0; i < feaMesh[mesh].numElement2DPID; i++){

                feaMeshComb->element2DPID[i + offSet] = feaMesh[mesh].element2DPID[i];
            }
        }

        if (feaMesh[mesh].element3DPID != NULL && feaMeshComb->element3DPID != NULL) {

            offSet           = (numElement3DPID - feaMesh[mesh].numElement3DPID);

            for (i = 0; i < feaMesh[mesh].numElement3DPID; i++){

                feaMeshComb->element3DPID[i + offSet] = feaMesh[mesh].element3DPID[i];
            }
        }

        // Set connectivity offset
        offSetConn = (numNodes-feaMesh[mesh].numNodes);

        if (feaMesh[mesh].crod != NULL && feaMeshComb->crod != NULL) {

            offSet = 2*(numRod-feaMesh[mesh].numRod);

            for (i = 0; i < feaMesh[mesh].numRod; i++) {
                feaMeshComb->crod[2*i + 0 + offSet] = feaMesh[mesh].crod[2*i + 0] + offSetConn;
                feaMeshComb->crod[2*i + 1 + offSet] = feaMesh[mesh].crod[2*i + 1] + offSetConn;

                feaMeshComb->crodPID[i    + offSet/2] = feaMesh[mesh].crodPID[i];
            }
        }

        if (feaMesh[mesh].ctria3 != NULL && feaMeshComb->ctria3 != NULL) {

            offSet = 3*(numTria3-feaMesh[mesh].numTria3);

            for (i = 0; i < feaMesh[mesh].numTria3; i++) {
                feaMeshComb->ctria3[3*i + 0 + offSet] = feaMesh[mesh].ctria3[3*i + 0] + offSetConn;
                feaMeshComb->ctria3[3*i + 1 + offSet] = feaMesh[mesh].ctria3[3*i + 1] + offSetConn;
                feaMeshComb->ctria3[3*i + 2 + offSet] = feaMesh[mesh].ctria3[3*i + 2] + offSetConn;

                feaMeshComb->ctria3PID[i    + offSet/3] = feaMesh[mesh].ctria3PID[i];
            }
        }

        if (feaMesh[mesh].cquad4 != NULL && feaMeshComb->cquad4 != NULL) {

            offSet = 4*(numQuad4-feaMesh[mesh].numQuad4);

            for (i = 0; i < feaMesh[mesh].numQuad4; i++) {
                feaMeshComb->cquad4[4*i + 0 + offSet] = feaMesh[mesh].cquad4[4*i + 0] + offSetConn;
                feaMeshComb->cquad4[4*i + 1 + offSet] = feaMesh[mesh].cquad4[4*i + 1] + offSetConn;
                feaMeshComb->cquad4[4*i + 2 + offSet] = feaMesh[mesh].cquad4[4*i + 2] + offSetConn;
                feaMeshComb->cquad4[4*i + 3 + offSet] = feaMesh[mesh].cquad4[4*i + 3] + offSetConn;

                feaMeshComb->cquad4PID[i    + offSet/4] = feaMesh[mesh].cquad4PID[i];
            }
        }

        if (feaMesh[mesh].chexa != NULL && feaMeshComb->chexa != NULL) {

            offSet = 8*(numHexa-feaMesh[mesh].numHexa);

            for (i = 0; i < feaMesh[mesh].numHexa; i++) {
                feaMeshComb->chexa[8*i + 0 + offSet] = feaMesh[mesh].chexa[8*i + 0] + offSetConn;
                feaMeshComb->chexa[8*i + 1 + offSet] = feaMesh[mesh].chexa[8*i + 1] + offSetConn;
                feaMeshComb->chexa[8*i + 2 + offSet] = feaMesh[mesh].chexa[8*i + 2] + offSetConn;
                feaMeshComb->chexa[8*i + 3 + offSet] = feaMesh[mesh].chexa[8*i + 3] + offSetConn;
                feaMeshComb->chexa[8*i + 4 + offSet] = feaMesh[mesh].chexa[8*i + 4] + offSetConn;
                feaMeshComb->chexa[8*i + 5 + offSet] = feaMesh[mesh].chexa[8*i + 5] + offSetConn;
                feaMeshComb->chexa[8*i + 6 + offSet] = feaMesh[mesh].chexa[8*i + 6] + offSetConn;
                feaMeshComb->chexa[8*i + 7 + offSet] = feaMesh[mesh].chexa[8*i + 7] + offSetConn;

                feaMeshComb->chexaPID[i    + offSet/8] = feaMesh[mesh].chexaPID[i];
            }
        }

        if (feaMesh[mesh].ctetra != NULL && feaMeshComb->ctetra != NULL) {

            offSet = 4*(numTetra-feaMesh[mesh].numTetra);

            for (i = 0; i < feaMesh[mesh].numTetra; i++) {
                feaMeshComb->ctetra[4*i + 0 + offSet] = feaMesh[mesh].ctetra[4*i + 0] + offSetConn;
                feaMeshComb->ctetra[4*i + 1 + offSet] = feaMesh[mesh].ctetra[4*i + 1] + offSetConn;
                feaMeshComb->ctetra[4*i + 2 + offSet] = feaMesh[mesh].ctetra[4*i + 2] + offSetConn;
                feaMeshComb->ctetra[4*i + 3 + offSet] = feaMesh[mesh].ctetra[4*i + 3] + offSetConn;

                feaMeshComb->ctetraPID[i    + offSet/4] = feaMesh[mesh].ctetraPID[i];
            }
        }


        // Point the lumped bodyTessMapping pointer that the surface bodyTessMapping
        feaMeshComb->bodyTessMappingLump[mesh] = &feaMesh[mesh].bodyTessMapping;

    }


    return CAPS_SUCCESS;
}


// Populate bndCondStruct boundary condition information - Boundary condition values get filled with 99
int populate_bndCondStruct_from_bcPropsStruct(cfdBCsStruct *bcProps,
                                              bndCondStruct *bndConds)
{

    // *bcProps [IN]
    // *bndConds [IN/OUT]


    int i;

    bndConds->numBND = bcProps->numBCID;

    // Transfers bcIDS into bndConds
    if (bndConds->numBND > 0){
        bndConds->bndID = (int *) EG_alloc(bndConds->numBND*sizeof(int));
        bndConds->bcVal = (int *) EG_alloc(bndConds->numBND*sizeof(int));
        if ((bndConds->bndID == NULL) || (bndConds->bcVal == NULL)) {
            return EGADS_MALLOC;
        }
    }

    for (i = 0; i < bndConds->numBND ; i++) {
        bndConds->bndID[i] = bcProps->surfaceProps[i].bcID;
    }

    // Fill in rest of bcVal  with dummy values
    for (i = 0; i < bndConds->numBND ; i++) bndConds->bcVal[i] = 99;

    return CAPS_SUCCESS;
}

// Populate bndCondStruct boundary condition information from attribute map - Boundary condition values get filled with 99
int populate_bndCondStruct_from_mapAttrToIndexStruct(mapAttrToIndexStruct *attrMap,
                                                     bndCondStruct *bndConds)
{
    // *bndConds [OUT]
    // *attrMap [IN]

    int i;

    bndConds->numBND = attrMap->numAttribute;

    // Transfers bcIDS into bndConds
    if (bndConds->numBND > 0){
        bndConds->bndID = (int *) EG_alloc(bndConds->numBND*sizeof(int));
        bndConds->bcVal = (int *) EG_alloc(bndConds->numBND*sizeof(int));
        if ((bndConds->bndID == NULL) || (bndConds->bcVal == NULL)) {
            return EGADS_MALLOC;
        }
    }

    for (i = 0; i < bndConds->numBND; i++) {
        bndConds->bndID[i] = attrMap->attributeIndex[i];
    }

    // Fill in rest of bcVal  with dummy values
    for (i = 0; i < bndConds->numBND; i++) bndConds->bcVal[i] = 99;

    return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a FEA mesh in the genFEAMeshStruct structure format
int initiate_genFEAMeshStruct(genFEAMeshStruct *feaMesh) {

    int status; // Function status return

    // ElementPID tracking information

    feaMesh->numElement1DPID = 0;
    feaMesh->element1DPID    = NULL;

    feaMesh->numElement2DPID = 0;
    feaMesh->element2DPID    = NULL;

    feaMesh->numElement3DPID = 0;
    feaMesh->element3DPID    = NULL;


	// Grid points
	feaMesh-> numNodes = 0;  // Number of nodes in grid
	feaMesh->xyz = NULL;   // x1 = xyz[0], y1 = xyz[1], z1 = xyz[2], x2 = xyz[3],.... size  = [3*numPoints]
	feaMesh->xyzCoordID = NULL; // Grid point coordinate identification. size = [numNodes]
	feaMesh->xyzAttrIndex = NULL;

	// 1-D
	feaMesh->numRod = 0;
	feaMesh->crod = NULL;
	feaMesh->crodPID = NULL;
	feaMesh->crodElementID = NULL;
	feaMesh->crodCoordID = NULL; // Coordinate identification for crod element. size[numRod]

	// 2-D
	feaMesh->numTria3 = 0;  // Number of tria3 elements
	feaMesh->ctria3 = NULL;   // Connectivity list for ctria3 elements.
	feaMesh->ctria3PID = NULL;// Property identification for ctria3 elements. size = [numTria3]
    feaMesh->ctria3ElementID = NULL;
    feaMesh->ctria3CoordID = NULL; // Coordinate identification for ctria3 element. size[numTria3]

	feaMesh->numQuad4 = 0;   // Number of cquad4 elements
	feaMesh->cquad4 = NULL;    // Connectivity list for cquad4 elements.
	feaMesh->cquad4PID = NULL; // Property identification for cquad4 elements. size = [numQuad4]
    feaMesh->cquad4ElementID = NULL;
    feaMesh->cquad4CoordID = NULL; // Coordinate identification for cquad4 element. size[numQuad4]

	// 3-D
	feaMesh->numHexa = 0;   // Number of hexahedral elements
	feaMesh->chexa = NULL;    // Connectivity list for chexa .  size = [8*numHexa]
	feaMesh->chexaPID = NULL; // Property identification for chexa elements. size = [numHexa]
    feaMesh->chexaElementID = NULL;

	feaMesh->numTetra = 0; // Number of tetrahedral elements
	feaMesh->ctetra = NULL;  // Connectivity list of ctetra. size = [4*numTetra]
	feaMesh->ctetraPID = NULL; // Property identification of ctetra elements. size = [numTetra]
    feaMesh->ctetraElementID = NULL;

	feaMesh->numMesh = 0; // Number of surfaces lumped together
	feaMesh->bodyTessMappingLump = NULL; // size = [numMesh]
	status = initiate_bodyTessMappingStruct(&feaMesh->bodyTessMapping);
	if (status != CAPS_SUCCESS) return status;

	return CAPS_SUCCESS;

}

// Destroy (0 out all values, free all arrays and NULL all pointers) a FEA mesh in the genFEAMeshStruct structure format
int destroy_genFEAMeshStruct(genFEAMeshStruct *feaMesh) {

    int status; // Function status return

    // ElementPID tracking information
    feaMesh->numElement1DPID = 0;

    if (feaMesh->element1DPID != NULL) EG_free(feaMesh->element1DPID);
    feaMesh->element1DPID = NULL;

    feaMesh->numElement2DPID = 0;

    if (feaMesh->element2DPID != NULL) EG_free(feaMesh->element2DPID);
    feaMesh->element2DPID = NULL;

    feaMesh->numElement3DPID = 0;

    if (feaMesh->element3DPID != NULL) EG_free(feaMesh->element3DPID);
    feaMesh->element3DPID = NULL;

    // CoordID tracking information
    if (feaMesh->crodCoordID != NULL) EG_free(feaMesh->crodCoordID);
    feaMesh->crodCoordID = NULL;

    if (feaMesh->ctria3CoordID != NULL) EG_free(feaMesh->ctria3CoordID);
    feaMesh->ctria3CoordID = NULL;

    if (feaMesh->cquad4CoordID != NULL) EG_free(feaMesh->cquad4CoordID);
    feaMesh->cquad4CoordID = NULL;

	// Grid points
	feaMesh-> numNodes = 0;  // Number of nodes in grid

	if (feaMesh->xyz != NULL) EG_free(feaMesh->xyz);
	feaMesh->xyz = NULL;

	if (feaMesh->xyzCoordID != NULL) EG_free(feaMesh->xyzCoordID);
	feaMesh->xyzCoordID = NULL;

	if (feaMesh->xyzAttrIndex != NULL) EG_free(feaMesh->xyzAttrIndex);
	feaMesh->xyzAttrIndex = NULL;

	// 1-D
	feaMesh->numRod = 0;

	if ( feaMesh->crod != NULL) EG_free(feaMesh->crod);
	feaMesh->crod = NULL;

	if ( feaMesh->crodPID != NULL) EG_free(feaMesh->crodPID);
	feaMesh->crodPID = NULL;

	if ( feaMesh->crodElementID != NULL) EG_free(feaMesh->crodElementID);
	feaMesh->crodElementID = NULL;


	// 2-D
	feaMesh->numTria3 = 0;

	if ( feaMesh->ctria3 != NULL) EG_free(feaMesh->ctria3);
	feaMesh->ctria3 = NULL;

	if ( feaMesh->ctria3PID != NULL) EG_free(feaMesh->ctria3PID);
	feaMesh->ctria3PID = NULL;

    if ( feaMesh->ctria3ElementID != NULL) EG_free(feaMesh->ctria3ElementID);
    feaMesh->ctria3ElementID = NULL;

	feaMesh->numQuad4 = 0;

	if ( feaMesh->cquad4 != NULL) EG_free(feaMesh->cquad4);
	feaMesh->cquad4 = NULL;

	if ( feaMesh->cquad4PID != NULL) EG_free(feaMesh->cquad4PID);
	feaMesh->cquad4PID = NULL;

    if ( feaMesh->cquad4ElementID != NULL) EG_free(feaMesh->cquad4ElementID);
    feaMesh->cquad4ElementID = NULL;

	// 3-D
	feaMesh->numHexa = 0;

	if ( feaMesh->chexa != NULL) EG_free(feaMesh->chexa);
	feaMesh->chexa = NULL;

	if ( feaMesh->chexaPID != NULL) EG_free(feaMesh->chexaPID);
	feaMesh->chexaPID = NULL;

    if ( feaMesh->chexaElementID != NULL) EG_free(feaMesh->chexaElementID);
    feaMesh->chexaElementID = NULL;

	feaMesh->numTetra = 0;

	if ( feaMesh->ctetra != NULL) EG_free(feaMesh->ctetra);
	feaMesh->ctetra = NULL;

	if ( feaMesh->ctetraPID != NULL) EG_free(feaMesh->ctetraPID);
	feaMesh->ctetraPID = NULL;

	if ( feaMesh->ctetraElementID != NULL) EG_free(feaMesh->ctetraElementID);
    feaMesh->ctetraElementID = NULL;

	feaMesh->numMesh = 0; // Number of surfaces lumped together
	if (feaMesh->bodyTessMappingLump != NULL) EG_free(feaMesh->bodyTessMappingLump);
	feaMesh->bodyTessMappingLump = NULL;

	status = destroy_bodyTessMappingStruct(&feaMesh->bodyTessMapping);
	if (status != CAPS_SUCCESS) return status;

	return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a bodyTessMapping in the bodyTessMappingStruct structure format
int initiate_bodyTessMappingStruct (bodyTessMappingStruct *bodyTessMapping) {
	// EGADS body tessellation storage
	//bodyTessMapping->egadsTess = NULL;

	bodyTessMapping->numTessFace = 0; // Number of faces in the tessellation

	bodyTessMapping->tessFaceQuadMap = NULL; // List to keep track of whether or not the tessObj has quads that have been split into tris
					   	   	   	   	   	   	 // size = [numTessFace]. In general if the quads have been split they should be added to the end
					   	   	   	   	   	   	 // of the tri list in the face tessellation

	return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a surface mesh in the genUnstrSurfMeshStruct structure format
int initiate_genUnstrSurfMeshStruct(genUnstrSurfMeshStruct *surfaceMesh) {

	int status; // Function status return

    // Set all numbers to zero
    surfaceMesh->numTriFace  = 0; // Number of triangular faces
    surfaceMesh->numQuadFace = 0; // Number of quadrilateral faces
    surfaceMesh->numNodes    = 0; // Number of nodes

    surfaceMesh->numSurface = 0; // Number of surfaces lumped together

    // NULL pointers
    surfaceMesh->xyz = NULL;

    surfaceMesh->nodeLocal2Global = NULL;

    surfaceMesh->triFaceList = NULL;

    surfaceMesh->quadFaceList = NULL;

    surfaceMesh->localTriFaceList = NULL;

    surfaceMesh->localQuadFaceList = NULL;

    surfaceMesh->triFaceMarkList = NULL;

    surfaceMesh->quadFaceMarkList = NULL;

    surfaceMesh->triFaceBLList = NULL;

    surfaceMesh->quadFaceBLList = NULL;

    surfaceMesh->nodeBLSpacingList = NULL;

    surfaceMesh->nodeBLThicknessList = NULL;

    surfaceMesh->surfTriFaceLump  = NULL; // Number of TriFace for an individual surface. size = [numSurface]
    surfaceMesh->surfQuadFaceLump = NULL; // Number of QuadFace for an individual surface. size = [numSurface]
    surfaceMesh->surfNodesLump = NULL; // Number of Nodes for an individual surface. size = [numSurface]

    surfaceMesh->bodyTessMappingLump = NULL;
    status = initiate_bodyTessMappingStruct(&surfaceMesh->bodyTessMapping);
    if (status != CAPS_SUCCESS) return status;

    surfaceMesh->numBoundaryEdge = 0; // Number of edges for a face - Used for 2D meshes - otherwise 0
    surfaceMesh->localBoundaryEdgeList = NULL; // Local node connectivity for an edge. size = [numBoundaryEdge]
    surfaceMesh->boundaryEdgeMarkList = NULL; // BC flag index for edge. size = [numBoundaryEdge]

    return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a volume Mesh in the genUnstrMeshStruct structure format
int initiate_genUnstrMeshStruct(genUnstrMeshStruct *mesh) {

	int status;

    // Set all numbers to zero
    mesh->numNodes = 0; // Number of nodes
    mesh->numTetra = 0; // Number of tetrahedrals
    mesh->numPyr   = 0; // Number of pyramids
    mesh->numPrz   = 0; // Number of prisms
    mesh->numHex   = 0; // Number of hexahedrals

    // NULL pointers
    mesh->xyz = NULL;

    mesh->tetraList = NULL;

    mesh->pyrList = NULL;

    mesh->przList = NULL;

    mesh->hexList = NULL;

    mesh->numVolID = 0; // Global number of volume ID's
    mesh->volIDList = NULL;  // Global list of volume ID's. size = [numVolID];

    mesh->tetraVolIDList = NULL; // Volume ID associated with each tetrahedral element. size = [numTetra]
    mesh->pyrVolIDList   = NULL;   // Volume ID associated with each pyramid element. size = [numPyr]
    mesh->przVolIDList   = NULL;   // Volume ID associated with each prism  element. size = [numPrz]
    mesh->hexVolIDList   = NULL;   // Volume ID associated with each hexahedral element. size = [numHex]

    status = initiate_genUnstrSurfMeshStruct(&mesh->surfaceMesh);

    return status;
}

// Destroy (0 out all values and NULL all pointers) a bodyTessMapping in the bodyTessMappingStruct structure format
int destroy_bodyTessMappingStruct (bodyTessMappingStruct *bodyTessMapping) {

	int status; // Function return

	// EGADS body tessellation storage
	/*if (bodyTessMapping->egadsTess != NULL) {
		status = EG_deleteObject(bodyTessMapping->egadsTess);
		if (status != EGADS_SUCCESS) printf("Status delete tess object = %d", status);
	}*/
	//bodyTessMapping->egadsTess= NULL;

	bodyTessMapping->numTessFace = 0; // Number of faces in the tessellation

	if (bodyTessMapping->tessFaceQuadMap != NULL) EG_free(bodyTessMapping->tessFaceQuadMap);
	bodyTessMapping->tessFaceQuadMap = NULL; // List to keep track of whether or not the tessObj has quads that have been split into tris
					   	   	   	   	   	     // size = [numTessFace]. In general if the quads have been split they should be added to the end
					   	   	   	   	   	   	 // of the tri list in the face tessellation

	return CAPS_SUCCESS;
}

// Destroy (0 out all values, free all arrays and NULL all pointers) a surface mesh in the genUnstrSurfMeshStruct structure format
int destroy_genUnstrSurfMeshStruct(genUnstrSurfMeshStruct *surfaceMesh) {

	int status; // Function return status

	int i; // Indexing

    // Set all numbers to zero
    surfaceMesh->numTriFace  = 0; // Number of triangular faces
    surfaceMesh->numQuadFace = 0; // Number of quadrilateral faces
    surfaceMesh->numNodes    = 0; // Number of nodes

    surfaceMesh->numSurface = 0; // Number of surfaces lumped together

    if (surfaceMesh->xyz != NULL) EG_free(surfaceMesh->xyz);
    surfaceMesh->xyz = NULL;

    if (surfaceMesh->nodeLocal2Global != NULL) EG_free(surfaceMesh->nodeLocal2Global);
    surfaceMesh->nodeLocal2Global = NULL;

    if (surfaceMesh->triFaceList != NULL) EG_free(surfaceMesh->triFaceList);
    surfaceMesh->triFaceList = NULL;

    if (surfaceMesh->quadFaceList != NULL) EG_free(surfaceMesh->quadFaceList);
    surfaceMesh->quadFaceList = NULL;

    if (surfaceMesh->localTriFaceList != NULL) EG_free(surfaceMesh->localTriFaceList);
    surfaceMesh->localTriFaceList = NULL;

    if (surfaceMesh->localQuadFaceList != NULL) EG_free(surfaceMesh->localQuadFaceList);
    surfaceMesh->localQuadFaceList = NULL;

    if (surfaceMesh->triFaceMarkList != NULL) EG_free(surfaceMesh->triFaceMarkList);
    surfaceMesh->triFaceMarkList = NULL;

    if (surfaceMesh->quadFaceMarkList != NULL) EG_free(surfaceMesh->quadFaceMarkList);
    surfaceMesh->quadFaceMarkList = NULL;

    if (surfaceMesh->triFaceBLList != NULL) EG_free(surfaceMesh->triFaceBLList);
    surfaceMesh->triFaceBLList = NULL;

    if (surfaceMesh->quadFaceBLList != NULL) EG_free(surfaceMesh->quadFaceBLList);
    surfaceMesh->quadFaceBLList = NULL;

    if (surfaceMesh->nodeBLSpacingList != NULL) EG_free(surfaceMesh->nodeBLSpacingList);
    surfaceMesh->nodeBLSpacingList = NULL;

    if (surfaceMesh->nodeBLThicknessList != NULL) EG_free(surfaceMesh->nodeBLThicknessList);
    surfaceMesh->nodeBLThicknessList = NULL;

    if (surfaceMesh->surfTriFaceLump != NULL) EG_free(surfaceMesh->surfTriFaceLump);
    surfaceMesh->surfTriFaceLump = NULL;

    if (surfaceMesh->surfQuadFaceLump != NULL) EG_free(surfaceMesh->surfQuadFaceLump);
    surfaceMesh->surfQuadFaceLump = NULL;

    if (surfaceMesh->surfNodesLump != NULL) EG_free(surfaceMesh->surfNodesLump);
    surfaceMesh->surfNodesLump = NULL;

    if (surfaceMesh->bodyTessMappingLump != NULL) EG_free(surfaceMesh->bodyTessMappingLump);
    surfaceMesh->bodyTessMappingLump = NULL;

    status = destroy_bodyTessMappingStruct(&surfaceMesh->bodyTessMapping);
    if (status != CAPS_SUCCESS) printf("Status during destroy_bodyTessMappStruct = %d\n", status);

    surfaceMesh->numBoundaryEdge = 0; // Number of edges for a face - Used for 2D meshes - otherwise 0
    if (surfaceMesh->localBoundaryEdgeList != NULL) EG_free(surfaceMesh->localBoundaryEdgeList);
    surfaceMesh->localBoundaryEdgeList = NULL; // Local node connectivity for an edge. size = [numBoundaryEdge]

    if (surfaceMesh->boundaryEdgeMarkList != NULL) EG_free(surfaceMesh->boundaryEdgeMarkList);
    surfaceMesh->boundaryEdgeMarkList = NULL; // BC flag index for edge. size = [numBoundaryEdge]

    return CAPS_SUCCESS;
}

// Destroy (0 out all values, free all arrays and NULL all pointers) a volume  in the genUnstrMeshStruct structure format
int destroy_genUnstrMeshStruct(genUnstrMeshStruct *mesh) {

	int status;

    // Set all numbers to zero
    mesh->numNodes = 0; // Number of nodes
    mesh->numTetra = 0; // Number of tetrahedrals
    mesh->numPyr   = 0; // Number of pyramids
    mesh->numPrz   = 0; // Number of prisms
    mesh->numHex   = 0; // Number of hexahedrals

    // Free and NULL pointers
    if (mesh->xyz != NULL) EG_free(mesh->xyz);
    mesh->xyz = NULL;

    if (mesh->tetraList != NULL) EG_free(mesh->tetraList);
    mesh->tetraList = NULL;

    if (mesh->pyrList != NULL) EG_free(mesh->pyrList);
    mesh->pyrList = NULL;

    if (mesh->przList != NULL) EG_free(mesh->przList);
    mesh->przList = NULL;

    if (mesh->hexList != NULL) EG_free(mesh->hexList);
    mesh->hexList = NULL;

    // Volume ID's
    mesh->numVolID = 0; // Global number of volume ID's

    if (mesh->volIDList != NULL) EG_free(mesh->volIDList);
    mesh->volIDList = NULL;  // Global list of volume ID's. size = [numVolID];

    if (mesh->tetraVolIDList != NULL) EG_free(mesh->tetraVolIDList);
    mesh->tetraVolIDList= NULL; // Volume ID associated with each tetrahedral element. size = [numTetra]

    if (mesh->pyrVolIDList != NULL) EG_free(mesh->pyrVolIDList);
    mesh->pyrVolIDList= NULL;   // Volume ID associated with each pyramid element. size = [numPyr]

    if (mesh->przVolIDList != NULL) EG_free(mesh->przVolIDList);
    mesh->przVolIDList= NULL;   // Volume ID associated with each prism  element. size = [numPrz]

    if (mesh->hexVolIDList != NULL) EG_free(mesh->hexVolIDList);
    mesh->hexVolIDList= NULL;   // Volume ID associated with each hexahedral element. size = [numHex]

    // Destroy surface mesh
    status = destroy_genUnstrSurfMeshStruct(&mesh->surfaceMesh);

    return status;
}

// Initiate (0 out all values and NULL all pointers) a bndCond in the bndCondStruct structure format
int initiate_bndCondStruct(bndCondStruct *bndCond) {

    bndCond->numBND = 0;
    bndCond->bndID = NULL;
    bndCond->bcVal = NULL;

    return CAPS_SUCCESS;
}

// Destroy (0 out all values, free all arrays, and NULL all pointers) a bndCond in the bndCondStruct structure format
int destroy_bndCondStruct(bndCondStruct *bndCond) {

    bndCond->numBND = 0;

    if (bndCond->bndID != NULL) EG_free(bndCond->bndID);
    bndCond->bndID = NULL;

    if (bndCond->bcVal != NULL) EG_free(bndCond->bcVal);
    bndCond->bcVal = NULL;

    return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a input in the tetgenInputStruct structure format
static int initiate_tetgenInputStruct(tetgenInputStruct *input) {

	input->meshQuality_rad_edge = 0.0;// Tetgen: maximum radius-edge ratio
	input->meshQuality_angle = 0.0;   // Tetgen: minimum dihedral angle
	input->meshInputString = NULL;     // Tetgen: Input string (optional) if NULL use default values
	input->verbose = (int) false;              // 0 = False, anything else True - Verbose output from mesh generator
	input->ignoreSurfaceExtract = (int) false; // 0 = False, anything else True - Don't extract the new surface mesh if
	                                //      Steiner points are added.
	input->meshTolerance = 0.0; // Tetgen : mesh tolerance

	return CAPS_SUCCESS;
}


// Destroy (0 out all values and NULL all pointers) a input in the tetgenInputStruct structure format
static int destroy_tetgenInputStruct(tetgenInputStruct *input) {

	input->meshQuality_rad_edge = 0.0;// Tetgen: maximum radius-edge ratio
	input->meshQuality_angle = 0.0;   // Tetgen: minimum dihedral angle
	if (input->meshInputString != NULL) EG_free(input->meshInputString);

	input->meshInputString = NULL;     // Tetgen: Input string (optional) if NULL use default values
	input->verbose = (int) false;              // 0 = False, anything else True - Verbose output from mesh generator
	input->ignoreSurfaceExtract = (int) false; // 0 = False, anything else True - Don't extract the new surface mesh if
	                                //      Steiner points are added.
	input->meshTolerance = 0.0; // Tetgen : mesh tolerance

	return CAPS_SUCCESS;
}


// Initiate (0 out all values and NULL all pointers) a input in the aflr3InputStruct structure format
static int initiate_aflr3InputStruct(aflr3InputStruct *input) {

	input->meshInputString = NULL;     // Tetgen: Input string (optional) if NULL use default values

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) a input in the aflr3InputStruct structure format
static int destroy_aflr3InputStruct(aflr3InputStruct *input) {

	if (input->meshInputString != NULL) EG_free(input->meshInputString);
	input->meshInputString = NULL;     // Tetgen: Input string (optional) if NULL use default values

	return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a input in the aflr4InputStruct structure format
static int initiate_aflr4InputStruct(aflr4InputStruct *input) {

	input->meshInputString = NULL;     // Tetgen: Input string (optional) if NULL use default values

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) a input in the aflr4InputStruct structure format
static int destroy_aflr4InputStruct(aflr4InputStruct *input) {

	if (input->meshInputString != NULL) EG_free(input->meshInputString);
	input->meshInputString = NULL;     // Tetgen: Input string (optional) if NULL use default values

	return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a meshInput in the meshInputStruct structure format
int initiate_meshInputStruct(meshInputStruct *meshInput) {

	int status; // Function return
	meshInput->paramTess[0] = 0;
	meshInput->paramTess[1] = 0;
	meshInput->paramTess[2] = 0;

	meshInput->preserveSurfMesh = (int) false; // 0 = False , anything else True - Use the body tessellation as the surface mesh

	meshInput->quiet = (int) false; // 0 = False , anything else True - No output from mesh generator
	meshInput->outputFormat = NULL;   // Mesh output formats - AFLR3, TECPLOT, VTK, SU2
	meshInput->outputFileName = NULL; // Filename prefix for mesh
	meshInput->outputDirectory = NULL;// Directory to write mesh to
	meshInput->outputASCIIFlag = (int) true;  // 0 = Binary output, anything else for ASCII

	status = initiate_bndCondStruct(&meshInput->bndConds);
	if (status != CAPS_SUCCESS) return status;

	status = initiate_tetgenInputStruct(&meshInput->tetgenInput);
	if (status != CAPS_SUCCESS) return status;

	status = initiate_aflr3InputStruct(&meshInput->aflr3Input);
	if (status != CAPS_SUCCESS) return status;

	status = initiate_aflr4InputStruct(&meshInput->aflr4Input);
	if (status != CAPS_SUCCESS) return status;

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) a meshInput in the meshInputStruct structure format
int destroy_meshInputStruct(meshInputStruct *meshInput) {

	int status; // Function return
	meshInput->paramTess[0] = 0;
	meshInput->paramTess[1] = 0;
	meshInput->paramTess[2] = 0;

	meshInput->preserveSurfMesh = (int) false; // 0 = False , anything else True - Use the body tessellation as the surface mesh

	meshInput->quiet = (int) false; // 0 = False , anything else True - No output from mesh generator
	if (meshInput->outputFormat != NULL) EG_free(meshInput->outputFormat);
	meshInput->outputFormat = NULL;   // Mesh output formats - AFLR3, TECPLOT, VTK, SU2

	if (meshInput->outputFileName != NULL) EG_free(meshInput->outputFileName);
	meshInput->outputFileName = NULL; // Filename prefix for mesh

	if (meshInput->outputDirectory != NULL) EG_free(meshInput->outputDirectory);
	meshInput->outputDirectory = NULL;// Directory to write mesh to

	meshInput->outputASCIIFlag = (int) true;  // 0 = Binary output, anything else for ASCII

	status = destroy_bndCondStruct(&meshInput->bndConds);
	if (status != CAPS_SUCCESS) return status;

	status = destroy_tetgenInputStruct(&meshInput->tetgenInput);
	if (status != CAPS_SUCCESS) return status;

	status = destroy_aflr3InputStruct(&meshInput->aflr3Input);
	if (status != CAPS_SUCCESS) return status;

	status = destroy_aflr4InputStruct(&meshInput->aflr4Input);
	if (status != CAPS_SUCCESS) return status;

	return CAPS_SUCCESS;
}

// Plot array of surface meshes
int plot_SurfaceMesh(meshInputStruct meshInput,
                     int numSurfMesh,
                     genUnstrSurfMeshStruct surfaceMesh[])
{

    int status; // Function return status

    int i; // Indexing

    char *filename = NULL; // Mesh file name

    genUnstrMeshStruct mesh; // Temporary volume mesh to plot mesh

    // Variables to aid in STL output
    int numTri = 0, *triConn = NULL, offSetIndex = 0;

    printf("Plotting surface mesh\n");

    // Initiate volume mesh
    status = initiate_genUnstrMeshStruct(&mesh);
    if (status != CAPS_SUCCESS) return status;

    // Lump surface meshes together
    status = lump_genUnstrSurfaceMesh(numSurfMesh, surfaceMesh, &mesh.surfaceMesh);
    if (status != CAPS_SUCCESS) return status;

    // Set number of nodes
    mesh.numNodes = mesh.surfaceMesh.numNodes;

    // Point xyz array at surfaceMesh xyz - avoids copying mesh
    mesh.xyz = mesh.surfaceMesh.xyz;

    // Point Tri and Quad Face list to local node numbers -avoids copying mesh
    if (mesh.surfaceMesh.triFaceList == NULL)  mesh.surfaceMesh.triFaceList = mesh.surfaceMesh.localTriFaceList;
    if (mesh.surfaceMesh.quadFaceList == NULL) mesh.surfaceMesh.quadFaceList = mesh.surfaceMesh.localQuadFaceList;

    filename= EG_alloc(strlen(meshInput.outputDirectory) + strlen(meshInput.outputFileName) + 2);
    if (filename == NULL) {

    	status = EGADS_MALLOC;
    	goto cleanup;
    }

    strcpy(filename, meshInput.outputDirectory);
    strcat(filename, "/");
    strcat(filename, meshInput.outputFileName);

    // Write mesh in grids formats
    if (meshInput.outputFormat == NULL) {
        printf("\tNo mesh outputFormat specified - defaulting to AFLR3\n");
        meshInput.outputFormat = (char *) "AFLR3";
    }

    if (strcasecmp(meshInput.outputFormat,"AFLR3") == 0) {

        status = write_AFLR3(filename, meshInput.outputASCIIFlag, &mesh);
        if (status != CAPS_SUCCESS) {
            printf("\tplot_SurfaceMesh: write_AFLR3 = %d\n", status);
        }

    } else if (strcasecmp(meshInput.outputFormat,"VTK") == 0) {

        status = write_VTK(filename, meshInput.outputASCIIFlag, &mesh);
        if (status != CAPS_SUCCESS) {
            printf("\tplot_SurfaceMesh: write_VTK = %d\n", status);
        }

        //} else if (strcasecmp(meshInput.outputFormat, "FUN2D") == 0) {

    } else if (strcasecmp(meshInput.outputFormat,"Tecplot") == 0) {

        status = write_Tecplot_Tet(filename, meshInput.outputASCIIFlag, &mesh);
        if (status != CAPS_SUCCESS) {
            printf("\tplot_SurfaceMesh: write_Tecplot = %d\n", status);
        }

        //} else if (strcasecmp(meshInput.outputFormat, "FUN2D") == 0) {

    } else if (strcasecmp(meshInput.outputFormat,"STL") == 0) {

    	numTri = mesh.surfaceMesh.numTriFace + 2*mesh.surfaceMesh.numQuadFace;

    	if (numTri > 0) triConn = (int *) EG_alloc(3*numTri*sizeof(int));

    	if (triConn == NULL) {
    		status = EGADS_MALLOC;
    		goto cleanup;
    	}

    	// Copy triangle connectivity
    	for (i = 0; i < mesh.surfaceMesh.numTriFace; i++) {

    		triConn[3*i + 0] = mesh.surfaceMesh.triFaceList[3*i + 0];
    		triConn[3*i + 1] = mesh.surfaceMesh.triFaceList[3*i + 1];
    		triConn[3*i + 2] = mesh.surfaceMesh.triFaceList[3*i + 2];
    	}

    	// Copy and split quad connectivity
		offSetIndex = 3*mesh.surfaceMesh.numTriFace;
		for (i = 0; i < mesh.surfaceMesh.numQuadFace; i++) {

			triConn[3*i + 0 + offSetIndex] = mesh.surfaceMesh.quadFaceList[4*i+0];
			triConn[3*i + 1 + offSetIndex] = mesh.surfaceMesh.quadFaceList[4*i+1];
			triConn[3*i + 2 + offSetIndex] = mesh.surfaceMesh.quadFaceList[4*i+3];

			offSetIndex += 3;
			triConn[3*i + 0 + offSetIndex] = mesh.surfaceMesh.quadFaceList[4*i+1];
			triConn[3*i + 1 + offSetIndex] = mesh.surfaceMesh.quadFaceList[4*i+2];
			triConn[3*i + 2 + offSetIndex] = mesh.surfaceMesh.quadFaceList[4*i+3];
		}

		status = write_STL(filename,
    	          	  	   numTri,
						   mesh.surfaceMesh.xyz,
						   triConn);

		if (status != CAPS_SUCCESS) {
			printf("\tplot_SurfaceMesh: write_STL = %d\n", status);
		}

    } else {

        printf("\tUnrecognized mesh format in plot_SurfaceMesh \"%s\"\n", meshInput.outputFormat);
    }

    goto cleanup;

    cleanup:

		if (status != CAPS_SUCCESS) printf("\tplot_SurfaceMesh status = %d\n", status);

		if (triConn != NULL) EG_free(triConn);

    	if (filename != NULL) EG_free(filename);

    	// NULL out mesh xyz pointer to avoid a double free event during mesh destruction
    	mesh.xyz = NULL;
    	mesh.surfaceMesh.triFaceList  = NULL;
    	mesh.surfaceMesh.quadFaceList = NULL;

		(void) destroy_genUnstrMeshStruct(&mesh);

		return status;
}

/*

static int calculate_CellVolume(cfdVolumeCellTypeEnum cellType, int elemenIndex, genUnstrMeshStruct volumeMesh, double *cellVolume ) {

	int i, j; // Indexing

	double term0, term1, term2;

	double x[6], y[6], z[6];
	double x1, x2, x3, y1, y2, y3, z1, z2, z3, xavg, yavg, zavg, areax, areay, areaz;

	enum {UnknownCFDCellSurface, Tri, Quad} cfdSurfaceType;

	term0 = 0.0;
	term1 = 0.0;
	term2 = 0.0;

	if (cellType == Prism) {

		for (i = 0; i < 6; i++) {
			x[i] = volumeMesh.xyz[3*volumeMesh->przList[6*elemenIndex + i] + 0];
			y[i] = volumeMesh.xyz[3*volumeMesh->przList[6*elemenIndex + i] + 1];
			z[i] = volumeMesh.xyz[3*volumeMesh->przList[6*elemenIndex + i] + 2];
		}

		// Loop through both Tri-faces
		for (i = 0; i < 1; i++){

			x1 = x[i];
		    x2 = x[i+1];
			x3 = x[i+2];

			y1 = y[i];
			y2 = y[i+1];
			y3 = y[i+2];

			z1 = z[i];
			z2 = z[i+1];
			z3 = z[i+2];


			xavg = (x1 + x2 + x3)/3.0;
			yavg = (y1 + y2 + y3)/3.0;
			zavg = (z1 + z2 + z3)/3.0;

			areax = 0.5*( (y2-y1)*(z3-z1) - (z2-z1)*(y3-y1) );
			areay = 0.5*( (z2-z1)*(x3-x1) - (x2-x1)*(z3-z1) );
			areaz = 0.5*( (x2-x1)*(y3-y1) - (y2-y1)*(x3-x1) );

			term0 = term0 + xavg*areax + yavg*areay + zavg*areaz
		}

		term0 = term0/3.0;

		// Loop through both Tri-faces
		for (i = 0; i < 2; i++)	{

			x1 = x(n1);
			x2 = x(n2);
			x3 = x(n3);
			x4 = x(n4);

			y1 = y(n1)
	    			        		  y2 = y(n2)
	    			          y3 = y(n	3)
	    			          y4 = y(n4)

	    			          z1 = z(n1)
	    			          z2 = z(n2)
	    			          z3 = z(n3)
	    			          z4 = z(n4)

	    			!         quad faces are split into 2 triangles - this splitting can
	    			!         be done in 2 ways (one for each diagonal). Start by
	    			!         picking one diagonal, then calculate volumes. Later,
	    			!         pick the other diagonal and repeat. Finally can average
	    			!         the two results to eliminate bias.

	    			!         break face up into triangles 1-2-3 and 1-3-4

	    			!         triangle 1-2-3

	    			          xavg = (x1 + x2 + x3)/3._dp
	    			          yavg = (y1 + y2 + y3)/3._dp
	    			          zavg = (z1 + z2 + z3)/3._dp

	    			          areax = 0.5_dp*( (y2-y1)*(z3-z1) - (z2-z1)*(y3-y1) )
	    			          areay = 0.5_dp*( (z2-z1)*(x3-x1) - (x2-x1)*(z3-z1) )
	    			          areaz = 0.5_dp*( (x2-x1)*(y3-y1) - (y2-y1)*(x3-x1) )

	    			          term1 = term1 + xavg*areax + yavg*areay + zavg*areaz

	    			          xn(iface) = xn(iface) + areax
	    			          yn(iface) = yn(iface) + areay
	    			          zn(iface) = zn(iface) + areaz

	    			!         triangle 1-3-4

	    			          xavg = (x1 + x3 + x4)/3._dp
	    			          yavg = (y1 + y3 + y4)/3._dp
	    			          zavg = (z1 + z3 + z4)/3._dp

	    			          areax = 0.5_dp*( (y3-y1)*(z4-z1) - (z3-z1)*(y4-y1) )
	    			          areay = 0.5_dp*( (z3-z1)*(x4-x1) - (x3-x1)*(z4-z1) )
	    			          areaz = 0.5_dp*( (x3-x1)*(y4-y1) - (y3-y1)*(x4-x1) )

	    			          term1 = term1 + xavg*areax + yavg*areay + zavg*areaz

	    			          xn(iface) = xn(iface) + areax
	    			          yn(iface) = yn(iface) + areay
	    			          zn(iface) = zn(iface) + areaz

	    			!         now repeat calculation, but this time use alternate face
	    			!         diagonal when splitting faces into triangles

	    			!         break face up into triangles 1-2-4 and 2-3-4

	    			!         triangle 1-2-4

	    			          xavg = (x1 + x2 + x4)/3._dp
	    			          yavg = (y1 + y2 + y4)/3._dp
	    			          zavg = (z1 + z2 + z4)/3._dp

	    			          areax = 0.5_dp*( (y2-y1)*(z4-z1) - (z2-z1)*(y4-y1) )
	    			          areay = 0.5_dp*( (z2-z1)*(x4-x1) - (x2-x1)*(z4-z1) )
	    			          areaz = 0.5_dp*( (x2-x1)*(y4-y1) - (y2-y1)*(x4-x1) )

	    			          term2 = term2 + xavg*areax + yavg*areay + zavg*areaz

	    			          xn(iface) = xn(iface) + areax
	    			          yn(iface) = yn(iface) + areay
	    			          zn(iface) = zn(iface) + areaz

	    			!         triangle 2-3-4

	    			          xavg = (x2 + x3 + x4)/3._dp
	    			          yavg = (y2 + y3 + y4)/3._dp
	    			          zavg = (z2 + z3 + z4)/3._dp

	    			          areax = 0.5_dp*( (y3-y2)*(z4-z2) - (z3-z2)*(y4-y2) )
	    			          areay = 0.5_dp*( (z3-z2)*(x4-x2) - (x3-x2)*(z4-z2) )
	    			          areaz = 0.5_dp*( (x3-x2)*(y4-y2) - (y3-y2)*(x4-x2) )

	    			          term2 = term2 + xavg*areax + yavg*areay + zavg*areaz

	    			          xn(iface) = xn(iface) + areax
	    			          yn(iface) = yn(iface) + areay
	    			          zn(iface) = zn(iface) + areaz

	    			!         average face normals from the two different quad splittings

	    			          xn(iface) = xn(iface)/2._dp
	    			          yn(iface) = yn(iface)/2._dp
	    			          zn(iface) = zn(iface)/2._dp

	    			!         unit normal

	    			          ra = sqrt( xn(iface)*xn(iface) + yn(iface)*yn(iface) +               &
	    			                     zn(iface)*zn(iface) )
	    			          xn(iface) = xn(iface)/ra
	    			          yn(iface) = yn(iface)/ra
	    			          zn(iface) = zn(iface)/ra

	    			          if (present(cell_area)) then
	    			            cell_area(n) = cell_area(n) + ra
	    			          end if

	    			        end if

	    			      end do quad_faces_loop

	    			      term1 = term1/3._dp
	    			      term2 = term2/3._dp

	    			!     add the volume contribution from the tria faces and average the
	    			!     resultant volume contributions from the two different quad splittings


	}

	*cellVolume = term0 + (term1 + term2)/2.0

}
*/

// Extrude a surface mesh a single unit the length of extrusionLength - return a
// volume mesh, cell volume and left-handness is not checked
int extrude_SurfaceMesh(double extrusionLength, genUnstrSurfMeshStruct surfaceMesh, genUnstrMeshStruct *volumeMesh)
{

    int status; // Status return

    int i, j, offset; // Indexing

    int normalFixed = (int) false;

    double p1[3], p2[3], p3[3], normalVector[3], normalVectorInitial[3]; // Surface normal

    printf("\nCreating a 3D volume mesh from a 2D surface\n");

    // Initiate volume mesh
    status = destroy_genUnstrMeshStruct(volumeMesh);
    if (status != CAPS_SUCCESS) return status;

    // Check arrays
    if (surfaceMesh.numBoundaryEdge == 0 ||
    	surfaceMesh.localBoundaryEdgeList == NULL ||
		surfaceMesh.boundaryEdgeMarkList == NULL) {
    	printf("\tNo edge boundaries saved - cannot use extrude_SurfaceMesh function!\n");
    	status = CAPS_BADVALUE;
    	goto bail;
    }

    // Set number of nodes
    volumeMesh->numNodes = 2*surfaceMesh.numNodes;

    volumeMesh->xyz = (double *) EG_alloc(3*volumeMesh->numNodes *sizeof(double));
    if (volumeMesh->xyz == NULL) {
    	status = EGADS_MALLOC;
    	goto bail;
    }
    // Set number of volume elements
    volumeMesh->numPrz = surfaceMesh.numTriFace;
    volumeMesh->numHex = surfaceMesh.numQuadFace;

    printf("\tNumber of Nodes =  %d\n", volumeMesh->numNodes);
    printf("\tNumber of Prisms =  %d\n", volumeMesh->numPrz);
    printf("\tNumber of Hexahedral =  %d\n", volumeMesh->numHex);

    if (volumeMesh->numPrz != 0) {
    	volumeMesh->przList = (int *) EG_alloc(6*volumeMesh->numPrz *sizeof(int));
    	if (volumeMesh->przList == NULL) {
    		status = EGADS_MALLOC;
    		goto bail;
    	}
    }

    if (volumeMesh->numHex != 0) {
    	volumeMesh->hexList = (int *) EG_alloc(8*volumeMesh->numHex *sizeof(int));
    	if (volumeMesh->hexList == NULL) {
    		status = EGADS_MALLOC;
    		goto bail;
    	}
    }

    // Determine surface normal sure all points are on the same plane
    if (surfaceMesh.numTriFace != 0) {

    	i = 0;
    	//for (i = 0; i < 1surfaceMesh.numTriFace; i++) {
    	p1[0] = surfaceMesh.xyz[3*(surfaceMesh.localTriFaceList[3*i +0]-1) + 0];
    	p1[1] = surfaceMesh.xyz[3*(surfaceMesh.localTriFaceList[3*i +0]-1) + 1];
		p1[2] = surfaceMesh.xyz[3*(surfaceMesh.localTriFaceList[3*i +0]-1) + 2];

		p2[0] = surfaceMesh.xyz[3*(surfaceMesh.localTriFaceList[3*i +1]-1) + 0];
		p2[1] = surfaceMesh.xyz[3*(surfaceMesh.localTriFaceList[3*i +1]-1) + 1];
		p2[2] = surfaceMesh.xyz[3*(surfaceMesh.localTriFaceList[3*i +1]-1) + 2];

		p3[0] = surfaceMesh.xyz[3*(surfaceMesh.localTriFaceList[3*i +2]-1) + 0];
		p3[1] = surfaceMesh.xyz[3*(surfaceMesh.localTriFaceList[3*i +2]-1) + 1];
		p3[2] = surfaceMesh.xyz[3*(surfaceMesh.localTriFaceList[3*i +2]-1) + 2];
		get_Surface_Norm(p1, p2, p3, normalVector);

		if (normalFixed == (int) false) {

			normalVectorInitial[0] = fabs(normalVector[0]);
			normalVectorInitial[1] = fabs(normalVector[1]);
			normalVectorInitial[2] = fabs(normalVector[2]);

			normalFixed = (int) true;

			//printf("Surface normal vector INITIAL = %f %f %f\n", normalVectorInitial[0], normalVectorInitial[1], normalVectorInitial[2]);
		}

		/*
		if (normalVectorInitial[0] != fabs(normalVector[0]) &&
			normalVectorInitial[1] != fabs(normalVector[1]) &&
			normalVectorInitial[2] != fabs(normalVector[2])) {

			printf("Warning points are not all on a single plane!!!\n");
		}
		*/

		//printf("Surface normal vector = %f %f %f\n", normalVector[0], normalVector[1], normalVector[2]);
    	//}
	} else if (surfaceMesh.numQuadFace != 0 ) {
		i = 0;
		//for (i = 0; i < surfaceMesh.numQuadFace; i++) {
		p1[0] = surfaceMesh.xyz[3*(surfaceMesh.localQuadFaceList[4*i + 0]-1) + 0];
		p1[1] = surfaceMesh.xyz[3*(surfaceMesh.localQuadFaceList[4*i + 0]-1) + 1];
		p1[2] = surfaceMesh.xyz[3*(surfaceMesh.localQuadFaceList[4*i + 0]-1) + 2];

		p2[0] = surfaceMesh.xyz[3*(surfaceMesh.localQuadFaceList[4*i + 1]-1) + 0];
		p2[1] = surfaceMesh.xyz[3*(surfaceMesh.localQuadFaceList[4*i + 1]-1) + 1];
		p2[2] = surfaceMesh.xyz[3*(surfaceMesh.localQuadFaceList[4*i + 1]-1) + 2];

		p3[0] = surfaceMesh.xyz[3*(surfaceMesh.localQuadFaceList[4*i + 2]-1) + 0];
		p3[1] = surfaceMesh.xyz[3*(surfaceMesh.localQuadFaceList[4*i + 2]-1) + 1];
		p3[2] = surfaceMesh.xyz[3*(surfaceMesh.localQuadFaceList[4*i + 2]-1) + 2];

		get_Surface_Norm(p1, p2, p3, normalVector);

		if (normalFixed == (int) false) {

			normalVectorInitial[0] = fabs(normalVector[0]);
			normalVectorInitial[1] = fabs(normalVector[1]);
			normalVectorInitial[2] = fabs(normalVector[2]);

			normalFixed = (int) true;
		}


		/*
		if (normalVectorInitial[0] != fabs(normalVector[0]) &&
			normalVectorInitial[1] != fabs(normalVector[1]) &&
			normalVectorInitial[2] != fabs(normalVector[2])) {

			printf("Warning points are not all on a single plane!!!\n");
		}
		*/

		//printf("Surface normal vector = %f %f %f\n", normalVector[0], normalVector[1], normalVector[2]);
		//}

	} else {
		status = CAPS_BADVALUE;
		goto bail;
	}

    // Copy initial surface mesh
    for (i = 0; i < surfaceMesh.numNodes; i++) {

    	volumeMesh->xyz[3*i + 0] = surfaceMesh.xyz[3*i + 0];
    	volumeMesh->xyz[3*i + 1] = surfaceMesh.xyz[3*i + 1];
    	volumeMesh->xyz[3*i + 2] = surfaceMesh.xyz[3*i + 2];
    }

    // Get extrusion plane
    for (i = 0; i < surfaceMesh.numNodes; i++) {

    	volumeMesh->xyz[3*(i+surfaceMesh.numNodes) + 0] = surfaceMesh.xyz[3*i + 0] + normalVectorInitial[0]*extrusionLength;
    	volumeMesh->xyz[3*(i+surfaceMesh.numNodes) + 1] = surfaceMesh.xyz[3*i + 1] + normalVectorInitial[1]*extrusionLength;
    	volumeMesh->xyz[3*(i+surfaceMesh.numNodes) + 2] = surfaceMesh.xyz[3*i + 2] + normalVectorInitial[2]*extrusionLength;
    }

    // Create prisms
    for (i = 0; i < volumeMesh->numPrz; i++) {

    	volumeMesh->przList[6*i + 0] = surfaceMesh.localTriFaceList[3*i + 0];
    	volumeMesh->przList[6*i + 1] = surfaceMesh.localTriFaceList[3*i + 1];
    	volumeMesh->przList[6*i + 2] = surfaceMesh.localTriFaceList[3*i + 2];

    	volumeMesh->przList[6*i + 3] = surfaceMesh.localTriFaceList[3*i + 0] + surfaceMesh.numNodes;
    	volumeMesh->przList[6*i + 4] = surfaceMesh.localTriFaceList[3*i + 1] + surfaceMesh.numNodes;
    	volumeMesh->przList[6*i + 5] = surfaceMesh.localTriFaceList[3*i + 2] + surfaceMesh.numNodes;
    }

    // Create hexs
    for (i = 0; i < volumeMesh->numHex; i++) {

    	volumeMesh->hexList[8*i + 0] = surfaceMesh.localQuadFaceList[4*i + 0];
    	volumeMesh->hexList[8*i + 1] = surfaceMesh.localQuadFaceList[4*i + 1];
    	volumeMesh->hexList[8*i + 2] = surfaceMesh.localQuadFaceList[4*i + 2];
    	volumeMesh->hexList[8*i + 3] = surfaceMesh.localQuadFaceList[4*i + 3];

    	volumeMesh->hexList[8*i + 4] = surfaceMesh.localQuadFaceList[4*i + 0] + surfaceMesh.numNodes;
    	volumeMesh->hexList[8*i + 5] = surfaceMesh.localQuadFaceList[4*i + 1] + surfaceMesh.numNodes;
    	volumeMesh->hexList[8*i + 6] = surfaceMesh.localQuadFaceList[4*i + 2] + surfaceMesh.numNodes;
    	volumeMesh->hexList[8*i + 7] = surfaceMesh.localQuadFaceList[4*i + 3] + surfaceMesh.numNodes;
    }

    if (surfaceMesh.numBoundaryEdge == 0) {
    	printf("\tNo boundary edges!!!\n");
    	status = CAPS_BADVALUE;
    	goto bail;
    }

    printf("\tNumber of Edges = %d\n", surfaceMesh.numBoundaryEdge);

    volumeMesh->surfaceMesh.numQuadFace = surfaceMesh.numBoundaryEdge + 2*surfaceMesh.numQuadFace;
    volumeMesh->surfaceMesh.numTriFace = 2*surfaceMesh.numTriFace;

    volumeMesh->surfaceMesh.quadFaceList = (int *) EG_alloc(4*volumeMesh->surfaceMesh.numQuadFace *sizeof(int));
    if (volumeMesh->surfaceMesh.quadFaceList == NULL) {
    	status = EGADS_MALLOC;
    	goto bail;
    }

    volumeMesh->surfaceMesh.quadFaceMarkList = (int *) EG_alloc(volumeMesh->surfaceMesh.numQuadFace *sizeof(int));
    if (volumeMesh->surfaceMesh.quadFaceMarkList == NULL) {
    	status = EGADS_MALLOC;
    	goto bail;
    }

    volumeMesh->surfaceMesh.triFaceList = (int *) EG_alloc(3*volumeMesh->surfaceMesh.numTriFace *sizeof(int));
    if (volumeMesh->surfaceMesh.triFaceList == NULL) {
    	status = EGADS_MALLOC;
    	goto bail;
    }

    volumeMesh->surfaceMesh.triFaceMarkList = (int *) EG_alloc(volumeMesh->surfaceMesh.numTriFace *sizeof(int));
    if (volumeMesh->surfaceMesh.triFaceMarkList == NULL) {
    	status = EGADS_MALLOC;
    	goto bail;
    }

    for (i = 0; i < surfaceMesh.numBoundaryEdge; i++) {

    	volumeMesh->surfaceMesh.quadFaceList[4*i + 0] = surfaceMesh.localBoundaryEdgeList[2*i + 0];
    	volumeMesh->surfaceMesh.quadFaceList[4*i + 1] = surfaceMesh.localBoundaryEdgeList[2*i + 1];

    	volumeMesh->surfaceMesh.quadFaceList[4*i + 2] = surfaceMesh.localBoundaryEdgeList[2*i + 1] + surfaceMesh.numNodes;
    	volumeMesh->surfaceMesh.quadFaceList[4*i + 3] = surfaceMesh.localBoundaryEdgeList[2*i + 0] + surfaceMesh.numNodes;

    	volumeMesh->surfaceMesh.quadFaceMarkList[i] = surfaceMesh.boundaryEdgeMarkList[i];
    }


    offset = surfaceMesh.numBoundaryEdge;
    for (i = 0; i < surfaceMesh.numQuadFace; i++) {

    	volumeMesh->surfaceMesh.quadFaceList[4*(i+offset) + 0] = surfaceMesh.localQuadFaceList[4*i + 0];
    	volumeMesh->surfaceMesh.quadFaceList[4*(i+offset) + 1] = surfaceMesh.localQuadFaceList[4*i + 1];
    	volumeMesh->surfaceMesh.quadFaceList[4*(i+offset) + 2] = surfaceMesh.localQuadFaceList[4*i + 2];
    	volumeMesh->surfaceMesh.quadFaceList[4*(i+offset) + 3] = surfaceMesh.localQuadFaceList[4*i + 3];

    	volumeMesh->surfaceMesh.quadFaceMarkList[i+offset] = surfaceMesh.quadFaceMarkList[i];
    }

    offset = offset+surfaceMesh.numQuadFace;
    for (i = 0; i < surfaceMesh.numQuadFace; i++) {

    	volumeMesh->surfaceMesh.quadFaceList[4*(i+offset) + 3] = surfaceMesh.localQuadFaceList[4*i + 0] + surfaceMesh.numNodes;
    	volumeMesh->surfaceMesh.quadFaceList[4*(i+offset) + 2] = surfaceMesh.localQuadFaceList[4*i + 1] + surfaceMesh.numNodes;
    	volumeMesh->surfaceMesh.quadFaceList[4*(i+offset) + 1] = surfaceMesh.localQuadFaceList[4*i + 2] + surfaceMesh.numNodes;
    	volumeMesh->surfaceMesh.quadFaceList[4*(i+offset) + 0] = surfaceMesh.localQuadFaceList[4*i + 3] + surfaceMesh.numNodes;

    	volumeMesh->surfaceMesh.quadFaceMarkList[i+offset] = surfaceMesh.quadFaceMarkList[i];
    }


    offset = 0;
    for (i = 0; i < surfaceMesh.numTriFace; i++) {


    	volumeMesh->surfaceMesh.triFaceList[3*(i+offset) + 0] = surfaceMesh.localTriFaceList[3*i + 0];
    	volumeMesh->surfaceMesh.triFaceList[3*(i+offset) + 1] = surfaceMesh.localTriFaceList[3*i + 1];
    	volumeMesh->surfaceMesh.triFaceList[3*(i+offset) + 2] = surfaceMesh.localTriFaceList[3*i + 2];

    	volumeMesh->surfaceMesh.triFaceMarkList[i+offset] = surfaceMesh.triFaceMarkList[i];
    }

    offset = surfaceMesh.numTriFace;
    for (i = 0; i < surfaceMesh.numTriFace; i++) {


    	volumeMesh->surfaceMesh.triFaceList[3*(i+offset) + 2] = surfaceMesh.localTriFaceList[3*i + 0] + surfaceMesh.numNodes;
    	volumeMesh->surfaceMesh.triFaceList[3*(i+offset) + 1] = surfaceMesh.localTriFaceList[3*i + 1] + surfaceMesh.numNodes;
    	volumeMesh->surfaceMesh.triFaceList[3*(i+offset) + 0] = surfaceMesh.localTriFaceList[3*i + 2] + surfaceMesh.numNodes;

    	volumeMesh->surfaceMesh.triFaceMarkList[i+offset] = surfaceMesh.triFaceMarkList[i];
    }

    return CAPS_SUCCESS;

    bail:

		// Destroy volume mesh
		(void ) destroy_genUnstrMeshStruct(volumeMesh);

		return status;

}

// Write a mesh contained in the general unstructured mesh structure (genUnstrMesh) in AFLR3 format
int write_AFLR3(char *fname,
                 int asciiFlag, // 0 for binary, anything else for ascii
                 genUnstrMeshStruct *mesh)
{

    FILE *fp;
    int i; // Indexing variable

    int sint = sizeof(int); // Size of an integer
    int sdouble = sizeof(double); // Size of a double

    //int numBytes = 0; // Number bytes written (unformatted option) // Un-comment if writing an unformatted file
    int machineENDIANNESS = 99;
    char *filename = NULL;
    char *postFix = NULL;

    printf("\nWriting AFLR3 file ....\n");

    if (asciiFlag == 0) {

    	machineENDIANNESS = get_MachineENDIANNESS();

    	if (machineENDIANNESS == 0) {
    		postFix = ".lb8.ugrid";

    	} else if (machineENDIANNESS == 1) {
    	   postFix = ".b8.ugrid";

    	} else {
    		printf("Unable to determine the ENDIANNESS of the current machine for binary file output\n");
    		return  CAPS_IOERR;
    	}

        filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(postFix)) *sizeof(char));
        if (filename == NULL) return EGADS_MALLOC;

        sprintf(filename, "%s%s", fname, postFix);
    	fp = fopen(filename, "wb");

        if (fp == NULL) {
            printf("Unable to open file: %s\n", filename);
            if (filename != NULL) EG_free(filename);
            return CAPS_IOERR;
        }

        //numBytes = 7 * sint; // Un-comment if writing an unformatted file
        //fwrite(&numBytes, sint,1,fp); // Un-comment if writing an unformatted file

        fwrite(&mesh->numNodes,   sint ,1,fp);
        fwrite(&mesh->surfaceMesh.numTriFace, sint ,1,fp);
        fwrite(&mesh->surfaceMesh.numQuadFace,sint ,1,fp);
        fwrite(&mesh->numTetra,   sint ,1,fp);
        fwrite(&mesh->numPyr,     sint ,1,fp);
        fwrite(&mesh->numPrz,     sint ,1,fp);
        fwrite(&mesh->numHex,     sint ,1,fp);

        //fwrite(&numBytes, sint,1,fp); // Un-comment if writing an unformatted file

        /*
        numBytes = 3*mesh->surfaceMesh.numTriFace  * sint +
        		   4*mesh->surfaceMesh.numQuadFace * sint +
				     mesh->surfaceMesh.numTriFace  * sint +
					 mesh->surfaceMesh.numQuadFace * sint +
				   4*mesh->numTetra * sint +
				   5*mesh->numPyr   * sint +
				   6*mesh->numPrz   * sint +
				   8*mesh->numHex   * sint +
				   3*mesh->numNodes * sdouble;
 	 	 */
		 // Un-comment if writing an unformatted file

        //fwrite(&numBytes, sint,1,fp); // Un-comment if writing an unformatted file

        // Write nodal coordinates
        for (i = 0; i < mesh->numNodes; i++) {
        	fwrite(&mesh->xyz[3*i+0], sdouble, 1, fp);
        	fwrite(&mesh->xyz[3*i+1], sdouble, 1, fp);
        	fwrite(&mesh->xyz[3*i+2], sdouble, 1, fp);

        }
        // Write tri-faces
        for (i = 0; i < mesh->surfaceMesh.numTriFace; i++) {
            fwrite(&mesh->surfaceMesh.triFaceList[3*i+0], sint, 1, fp);
            fwrite(&mesh->surfaceMesh.triFaceList[3*i+1], sint, 1, fp);
            fwrite(&mesh->surfaceMesh.triFaceList[3*i+2], sint, 1, fp);
        }

        // Write quad-faces
        for (i = 0; i < mesh->surfaceMesh.numQuadFace; i++) {
        	fwrite(&mesh->surfaceMesh.quadFaceList[4*i+0], sint, 1, fp);
            fwrite(&mesh->surfaceMesh.quadFaceList[4*i+1], sint, 1, fp);
            fwrite(&mesh->surfaceMesh.quadFaceList[4*i+2], sint, 1, fp);
            fwrite(&mesh->surfaceMesh.quadFaceList[4*i+3], sint, 1, fp);
        }

        // Write tri-face boundaries
        for (i = 0; i < mesh->surfaceMesh.numTriFace; i++) {
            fwrite(&mesh->surfaceMesh.triFaceMarkList[i], sint, 1, fp);
        }

        // Write quad-face boundaries
        for (i = 0; i < mesh->surfaceMesh.numQuadFace; i++) {
            fwrite(&mesh->surfaceMesh.quadFaceMarkList[i], sint, 1, fp);
        }

        // Write tetrahedrals connectivity
        for (i = 0; i < mesh->numTetra; i++) {
            fwrite(&mesh->tetraList[4*i+0], sint, 1, fp);
            fwrite(&mesh->tetraList[4*i+1], sint, 1, fp);
            fwrite(&mesh->tetraList[4*i+2], sint, 1, fp);
            fwrite(&mesh->tetraList[4*i+3], sint, 1, fp);
        }

        // Write pyramid connectivity
        for (i = 0; i < mesh->numPyr; i++) {
        	fwrite(&mesh->pyrList[5*i+0], sint, 1, fp);
            fwrite(&mesh->pyrList[5*i+1], sint, 1, fp);
            fwrite(&mesh->pyrList[5*i+2], sint, 1, fp);
            fwrite(&mesh->pyrList[5*i+3], sint, 1, fp);
            fwrite(&mesh->pyrList[5*i+4], sint, 1, fp);
        }

        // Write prisms connectivity
        for (i = 0; i < mesh->numPrz; i++) {
        	fwrite(&mesh->przList[6*i+0], sint, 1, fp);
            fwrite(&mesh->przList[6*i+1], sint, 1, fp);
            fwrite(&mesh->przList[6*i+2], sint, 1, fp);
            fwrite(&mesh->przList[6*i+3], sint, 1, fp);
            fwrite(&mesh->przList[6*i+4], sint, 1, fp);
            fwrite(&mesh->przList[6*i+5], sint, 1, fp);
        }

        // Write hex connectivity
        for (i = 0; i < mesh->numHex; i++) {
            fwrite(&mesh->hexList[8*i+0], sint, 1, fp);
            fwrite(&mesh->hexList[8*i+1], sint, 1, fp);
            fwrite(&mesh->hexList[8*i+2], sint, 1, fp);
            fwrite(&mesh->hexList[8*i+3], sint, 1, fp);
            fwrite(&mesh->hexList[8*i+4], sint, 1, fp);
            fwrite(&mesh->hexList[8*i+5], sint, 1, fp);
            fwrite(&mesh->hexList[8*i+6], sint, 1, fp);
            fwrite(&mesh->hexList[8*i+7], sint, 1, fp);
        }

        //fwrite(&numBytes, sint,1,fp); // Un-comment if writing an unformatted file

    } else {

        postFix = ".ugrid";
        filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(postFix)) *sizeof(char));
        if (filename == NULL) return EGADS_MALLOC;

        sprintf(filename, "%s%s", fname, postFix);

        fp = fopen(filename, "w");
        if (fp == NULL) {
            printf("Unable to open file: %s\n", filename);
            if (filename != NULL) EG_free(filename);
            return CAPS_IOERR;
        }

        //nodes, tri-face, quad-face, numTetra, numPyr, numPrz, numHex
        fprintf(fp,"%d, %d, %d, %d, %d, %d, %d\n", mesh->numNodes,
                                                   mesh->surfaceMesh.numTriFace,
                                                   mesh->surfaceMesh.numQuadFace,
                                                   mesh->numTetra,
                                                   mesh->numPyr,
                                                   mesh->numPrz,
                                                   mesh->numHex);
        // Write nodal coordinates
        for (i = 0; i < mesh->numNodes; i++) {
            fprintf(fp,"%f %f %f\n", mesh->xyz[3*i  ],
                                     mesh->xyz[3*i+1],
                                     mesh->xyz[3*i+2]);
        }

        // Write tri-faces
        for (i = 0; i < mesh->surfaceMesh.numTriFace; i++) {
            fprintf(fp,"%d %d %d\n",mesh->surfaceMesh.triFaceList[3*i  ],
                                    mesh->surfaceMesh.triFaceList[3*i+1],
                                    mesh->surfaceMesh.triFaceList[3*i+2]);
        }

        // Write quad-faces
        for (i = 0; i < mesh->surfaceMesh.numQuadFace; i++) {
            fprintf(fp,"%d %d %d %d\n",mesh->surfaceMesh.quadFaceList[4*i  ],
                                       mesh->surfaceMesh.quadFaceList[4*i+1],
                                       mesh->surfaceMesh.quadFaceList[4*i+2],
                                       mesh->surfaceMesh.quadFaceList[4*i+3]);
        }

        // Write tri-face boundaries
        for (i = 0; i < mesh->surfaceMesh.numTriFace; i++) {
            fprintf(fp,"%d\n",mesh->surfaceMesh.triFaceMarkList[i]);
        }

        // Write quad-face boundaries
        for (i = 0; i < mesh->surfaceMesh.numQuadFace; i++) {
            fprintf(fp,"%d\n",mesh->surfaceMesh.quadFaceMarkList[i]);
        }

        // Write tetrahedrals connectivity
        for (i = 0; i < mesh->numTetra; i++) {
            fprintf(fp,"%d %d %d %d\n",mesh->tetraList[4*i  ],
                                       mesh->tetraList[4*i+1],
                                       mesh->tetraList[4*i+2],
                                       mesh->tetraList[4*i+3]);
        }

        // Write pyramid connectivity
        for (i = 0; i < mesh->numPyr; i++) {
            fprintf(fp,"%d %d %d %d %d\n",mesh->pyrList[5*i  ],
                                          mesh->pyrList[5*i+1],
                                          mesh->pyrList[5*i+2],
                                          mesh->pyrList[5*i+3],
                                          mesh->pyrList[5*i+4]);
        }

        // Write prisms connectivity
        for (i = 0; i < mesh->numPrz; i++) {
            fprintf(fp,"%d %d %d %d %d %d\n",mesh->przList[6*i  ],
                                             mesh->przList[6*i+1],
                                             mesh->przList[6*i+2],
                                             mesh->przList[6*i+3],
                                             mesh->przList[6*i+4],
                                             mesh->przList[6*i+5]);
        }

        // Write hex connectivity
        for (i = 0; i < mesh->numHex; i++) {
            fprintf(fp,"%d %d %d %d %d %d %d %d\n",mesh->hexList[8*i  ],
                                                   mesh->hexList[8*i+1],
                                                   mesh->hexList[8*i+2],
                                                   mesh->hexList[8*i+3],
                                                   mesh->hexList[8*i+4],
                                                   mesh->hexList[8*i+5],
                                                   mesh->hexList[8*i+6],
                                                   mesh->hexList[8*i+7]);
        }
    }

    fclose(fp);

    if (filename != NULL) EG_free(filename);

    printf("Finished writing AFLR3 file\n\n");

    return CAPS_SUCCESS;
}

// Write a mesh contained in the general unstructured mesh structure (genUnstrMesh) in SU2 format
int write_SU2(char *fname,
              int asciiFlag, // 0 for binary, anything else for ascii
              genUnstrMeshStruct *mesh,
              int numBnds,
              int *bndID//,
              //int *bndVals // Eventually probably want to put this as a list of strings for tag names
              )
{

    // Important: SU2 wants elements/index to start at 0 we assume everything coming in starts at 1

	int status; // Function return status

    FILE *fp = NULL;
    int  i, j, *numMarkerList = NULL;
    int  elemNum = 0;
    char *filename = NULL;
    char fileExt[] = ".su2";

    int twodMode = (int) false;

    numMarkerList = (int *) EG_alloc(numBnds*sizeof(int));  // Array to keep track of the number of
                                                            // surface faces with a particular boundary id.
    if (numMarkerList == NULL) {
    	status  = EGADS_MALLOC;
    	goto cleanup;
    }

    printf("\nWriting SU2 file ....\n");

    if (asciiFlag == 0) {
        printf("Binary output is not supported by SU2\n");
        printf("..... switching to ASCII!\n");
        asciiFlag = 1;
    }

    filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(fileExt)) *sizeof(char));
    if (filename == NULL) {
    	status  = EGADS_MALLOC;
    	goto cleanup;
    }

    sprintf(filename,"%s%s",fname, fileExt);

    fp = fopen(filename, "w");

    if (fp == NULL) {
        printf("Unable to open file: %s\n", filename);

        status  = CAPS_IOERR;
        goto cleanup;
    }


    // Will this be a 2D mesh ?
    if (mesh->numTetra == 0 &&
        mesh->numPyr  == 0 &&
		mesh->numPrz == 0 &&
		mesh->numHex == 0) {

    	// Must have boundary edges stored to write out a 2D mesh
    	if (mesh->surfaceMesh.numBoundaryEdge !=0) {

    		printf("The supplied volume mesh appears to be a 2D mesh!\n");
    		twodMode = (int) true;
    	}
    }

    if (asciiFlag == 0) {} // Mute point as SU2 only wants ascii - future extension though?
    else {

        // Dimensionality
    	if (twodMode == (int) true) {
    		fprintf(fp,"NDIME= %d\n",2);
    	} else {
    		fprintf(fp,"NDIME= %d\n",3);
    	}

    	if (twodMode == (int) true) {
    		// Number of elements
    		fprintf(fp,"NELEM= %d\n",mesh->surfaceMesh.numTriFace +
    								 mesh->surfaceMesh.numQuadFace);
    	} else {
    		// Number of elements
			fprintf(fp,"NELEM= %d\n",mesh->numTetra +
									 mesh->numPyr +
									 mesh->numPrz +
									 mesh->numHex);
    	}
        // SU2 wants elements/index to start at 0 - assume everything starts at 1

        elemNum = 0;

        // Write tetrahedral connectivity
        for (i = 0; i < mesh->numTetra; i++) {
            fprintf(fp,"10 %d %d %d %d %d\n",mesh->tetraList[4*i  ]-1,
                                             mesh->tetraList[4*i+1]-1,
                                             mesh->tetraList[4*i+2]-1,
                                             mesh->tetraList[4*i+3]-1,
                                             elemNum);
            elemNum+=1;
        }

        // Write hex connectivity
        for (i = 0; i < mesh->numHex; i++) {
            fprintf(fp,"12 %d %d %d %d %d %d %d %d %d\n",mesh->hexList[8*i  ]-1,
                                                         mesh->hexList[8*i+1]-1,
                                                         mesh->hexList[8*i+2]-1,
                                                         mesh->hexList[8*i+3]-1,
                                                         mesh->hexList[8*i+4]-1,
                                                         mesh->hexList[8*i+5]-1,
                                                         mesh->hexList[8*i+6]-1,
                                                         mesh->hexList[8*i+7]-1,
                                                         elemNum);
            elemNum+=1;
        }

        // Write prisms connectivity
        for (i = 0; i < mesh->numPrz; i++) {
            fprintf(fp,"13 %d %d %d %d %d %d %d\n",mesh->przList[6*i  ]-1,
                                                   mesh->przList[6*i+1]-1,
                                                   mesh->przList[6*i+2]-1,
                                                   mesh->przList[6*i+3]-1,
                                                   mesh->przList[6*i+4]-1,
                                                   mesh->przList[6*i+5]-1,
                                                   elemNum);
            elemNum+=1;
        }

        // Write pyramid connectivity
        for (i = 0; i < mesh->numPyr; i++) {
            fprintf(fp,"14 %d %d %d %d %d %d\n",mesh->pyrList[5*i  ]-1,
                                                mesh->pyrList[5*i+1]-1,
                                                mesh->pyrList[5*i+2]-1,
                                                mesh->pyrList[5*i+3]-1,
                                                mesh->pyrList[5*i+4]-1,
                                                elemNum);
            elemNum+=1;
        }

        if (twodMode == (int) true) {

        	for (i = 0; i < mesh->surfaceMesh.numTriFace; i++) {

        		fprintf(fp,"5 %d %d %d\n", mesh->surfaceMesh.triFaceList[3*i  ]-1, // Again -1 since assumed index started a 1
        								   mesh->surfaceMesh.triFaceList[3*i+1]-1,
										   mesh->surfaceMesh.triFaceList[3*i+2]-1);

        	}
        	for (i = 0; i < mesh->surfaceMesh.numQuadFace; i++) {
        		fprintf(fp,"9 %d %d %d %d\n",mesh->surfaceMesh.quadFaceList[4*i  ]-1,
        									 mesh->surfaceMesh.quadFaceList[4*i+1]-1,
											 mesh->surfaceMesh.quadFaceList[4*i+2]-1,
											 mesh->surfaceMesh.quadFaceList[4*i+3]-1);
        	}
        }

        // Number of points
        fprintf(fp,"NPOIN= %d\n",mesh->numNodes);

        // Write nodal coordinates
        for (i = 0; i < mesh->numNodes; i++) {
            fprintf(fp,"%f %f %f %d\n",mesh->xyz[3*i  ],
                                       mesh->xyz[3*i+1],
                                       mesh->xyz[3*i+2],
                                       i); // Connectivity starts at 0
        }

        // Number of boundary ID
        fprintf(fp,"NMARK= %d\n", numBnds);

        // We need the number of surface elements that have a particular boundary ID
        // First initialize numMarkerList components to zero
        for (i = 0; i < numBnds; i++) numMarkerList[i] = 0;

        // Next count elements with a particular id
        for (i = 0; i < numBnds; i++) {

        	if (twodMode == (int) true) {

        		for (j = 0; j < mesh->surfaceMesh.numBoundaryEdge; j++) {

        			if (mesh->surfaceMesh.boundaryEdgeMarkList[j] == bndID[i]) numMarkerList[i] += 1;
        		}

        	} else {

				for (j = 0; j < mesh->surfaceMesh.numTriFace; j++) {

					if (mesh->surfaceMesh.triFaceMarkList[j] == bndID[i]) numMarkerList[i] += 1;
				}

				for (j = 0; j < mesh->surfaceMesh.numQuadFace; j++) {

					if (mesh->surfaceMesh.quadFaceMarkList[j] == bndID[i]) numMarkerList[i] += 1;
				}
        	}

        }

        // Finally lets write the connectivity
        for (i = 0; i < numBnds; i++) {

        	if (numMarkerList[i] == 0) continue;

            fprintf(fp,"MARKER_TAG= %d\n", bndID[i]); // Probably eventually want to change this to a string tag
                                                      // see note at the beginning of function

            fprintf(fp,"MARKER_ELEMS= %d\n", numMarkerList[i]); //Number of elements with a particular ID

            if (twodMode == (int) true) {

            	for (j = 0; j < mesh->surfaceMesh.numBoundaryEdge; j++) {

            		if (mesh->surfaceMesh.boundaryEdgeMarkList[j] == bndID[i]) {
            			fprintf(fp, "3 %d %d\n", mesh->surfaceMesh.localBoundaryEdgeList[2*j+0]-1,
            									 mesh->surfaceMesh.localBoundaryEdgeList[2*j+1]-1);
            		}
            	}

            } else {
				for (j = 0; j < mesh->surfaceMesh.numTriFace; j++) {

					if (mesh->surfaceMesh.triFaceMarkList[j] == bndID[i]) {

						fprintf(fp,"5 %d %d %d\n", mesh->surfaceMesh.triFaceList[3*j  ]-1, // Again -1 since assumed index started a 1
												   mesh->surfaceMesh.triFaceList[3*j+1]-1,
												   mesh->surfaceMesh.triFaceList[3*j+2]-1);
					}
				}

				for (j = 0; j < mesh->surfaceMesh.numQuadFace; j++) {

					if (mesh->surfaceMesh.quadFaceMarkList[j] == bndID[i]) {

						fprintf(fp,"9 %d %d %d %d\n",mesh->surfaceMesh.quadFaceList[4*j  ]-1,
													 mesh->surfaceMesh.quadFaceList[4*j+1]-1,
													 mesh->surfaceMesh.quadFaceList[4*j+2]-1,
													 mesh->surfaceMesh.quadFaceList[4*j+3]-1);
					}
				}
			}
        }
    }

    printf("Finished writing SU2 file\n\n");

    status = CAPS_SUCCESS;

    goto cleanup;

    cleanup:
		if (filename != NULL) EG_free(filename);
		if (numMarkerList != NULL) EG_free(numMarkerList);

		if (fp != NULL) fclose(fp);
		return status;
}

// Write a mesh contained in the general unstructured mesh structure (genUnstrMesh) in VTK format
int write_VTK(char *fname,
              int asciiFlag, // 0 for binary, anything else for ascii
              genUnstrMeshStruct *mesh)
{

    FILE *fp;
    int i, j, len, numCells, temp;
    int sint = sizeof(int);
    int sdouble = sizeof(double);

    char *filename = NULL;
  
    // NEED change if indices already start at 0
    int m1 = -1; // VTK indices start at 0 !!!!

    printf("\nWriting VTK file ....\n");

    if (asciiFlag == 0) {
        printf("Binary output is not currently working correctly for VTK\n");
        printf("..... switching to ASCII!\n");
        asciiFlag = 1;
    }

    //TODO - add input checks

    filename = (char *) EG_alloc((strlen(fname) + 1 + 4) *sizeof(char));
    sprintf(filename,"%s.vtk",fname);

    fp = fopen(filename, "wb");
    if (fp == NULL) {
        printf("Unable to open file: %s\n", filename);
        EG_free(filename);
        return CAPS_IOERR;
    }

    fprintf(fp,"# vtk DataFile Version 2.0\n");
    fprintf(fp,"Unstructured Grid\n");

    if (asciiFlag == 0) {
        fprintf(fp,"BINARY\n");
    } else {
        fprintf(fp,"ASCII\n");
    }

    fprintf(fp,"DATASET UNSTRUCTURED_GRID\n");

    fprintf(fp,"POINTS %d double\n",mesh->numNodes);

    if (asciiFlag == 0) {
        // Write nodal coordinates
        len = sdouble*mesh->numNodes*3;
        fwrite(mesh->xyz, sdouble, len,fp);

    } else {
        // Write nodal coordinates
        for (i = 0; i < mesh->numNodes; i++) {
            fprintf(fp,"%.17g %.17g %.17g\n",mesh->xyz[3*i  ],
                                             mesh->xyz[3*i+1],
                                             mesh->xyz[3*i+2]);
        }

    }

    numCells = mesh->numTetra + mesh->numPyr + mesh->numPrz + mesh->numHex;
    len = numCells + mesh->numTetra*4 + mesh->numPyr*5 + mesh->numPrz*6 + mesh->numHex*8;
    fprintf(fp,"\nCELLS %d %d\n",numCells,len);

    if (asciiFlag == 0) {
        // Write tetrahedrals connectivity
        for (i = 0; i < mesh->numTetra; i++) {
            len = 4;
            fwrite(&len,sint,1,fp);
            for (j = 0; j < len ; j++) {
                temp = mesh->tetraList[len*i + j]+m1;
                fwrite(&temp, sint, 1,fp);
            }
        }

        // Write pyramid connectivity
        for (i = 0; i < mesh->numPyr; i++) {
            len = 5;
            fwrite(&len,sint,1,fp);
            for (j = 0; j < len ; j++) {
                temp = mesh->pyrList[len*i + j]+m1;
                fwrite(&temp, sint, 1,fp);
            }
        }

        // Write prisms connectivity
        for (i = 0; i < mesh->numPrz; i++) {
            len = 6;
            fwrite(&len,sint,1,fp);
            for (j = 0; j < len ; j++) {
                temp = mesh->przList[len*i + j]+m1;
                fwrite(&temp, sint, 1,fp);
            }
        }

        // Write hex connectivity
        for (i = 0; i < mesh->numHex; i++) {
            len = 8;
            fwrite(&len,sint,1,fp);
            for (j = 0; j < len ; j++) {
                temp = mesh->hexList[len*i + j]+m1;
                fwrite(&temp, sint, 1,fp);
            }
        }

    } else {
        // Write tetrahedrals connectivity
        for (i = 0; i < mesh->numTetra; i++) {
            fprintf(fp,"4 %d %d %d %d\n",mesh->tetraList[4*i  ]+m1,
                                         mesh->tetraList[4*i+1]+m1,
                                         mesh->tetraList[4*i+2]+m1,
                                         mesh->tetraList[4*i+3]+m1);
        }

        // Write pyramid connectivity
        for (i = 0; i < mesh->numPyr; i++) {
            fprintf(fp,"5 %d %d %d %d %d\n",mesh->pyrList[5*i  ]+m1,
                                            mesh->pyrList[5*i+1]+m1,
                                            mesh->pyrList[5*i+2]+m1,
                                            mesh->pyrList[5*i+3]+m1,
                                            mesh->pyrList[5*i+4]+m1);
        }

        // Write prisms connectivity
        for (i = 0; i < mesh->numPrz; i++) {
            fprintf(fp,"6 %d %d %d %d %d %d\n",mesh->przList[6*i  ]+m1,
                                               mesh->przList[6*i+1]+m1,
                                               mesh->przList[6*i+2]+m1,
                                               mesh->przList[6*i+3]+m1,
                                               mesh->przList[6*i+4]+m1,
                                               mesh->przList[6*i+5]+m1);
        }

        // Write hex connectivity
        for (i = 0; i < mesh->numHex; i++) {
            fprintf(fp,"8 %d %d %d %d %d %d %d %d\n",mesh->hexList[8*i  ]+m1,
                                                     mesh->hexList[8*i+1]+m1,
                                                     mesh->hexList[8*i+2]+m1,
                                                     mesh->hexList[8*i+3]+m1,
                                                     mesh->hexList[8*i+4]+m1,
                                                     mesh->hexList[8*i+5]+m1,
                                                     mesh->hexList[8*i+6]+m1,
                                                     mesh->hexList[8*i+7]+m1);
        }
    }

    fprintf(fp,"\nCELL_TYPES %d\n",numCells);

    if (asciiFlag == 0) {
        // Write tetrahedrals type
        for (i = 0; i < mesh->numTetra; i++) {
            len = 10;
            fwrite(&len,sint,1,fp);
        }

        // Write pyramid type
        for (i = 0; i < mesh->numPyr; i++) {
            len = 14;
            fwrite(&len,sint,1,fp);
        }

        // Write prisms type
        for (i = 0; i < mesh->numPrz; i++) {
            len = 13;
            fwrite(&len,sint,1,fp);
        }

        // Write hex type
        for (i = 0; i < mesh->numHex; i++) {
            len = 12;
            fwrite(&len,sint,1,fp);
        }

    } else {
        // Write tetrahedrals type
        for (i = 0; i < mesh->numTetra; i++) {
            fprintf(fp,"%d\n",10);
        }

        // Write pyramid type
        for (i = 0; i < mesh->numPyr; i++) {
            fprintf(fp,"%d\n",14);
        }

        // Write prisms type
        for (i = 0; i < mesh->numPrz; i++) {
            fprintf(fp,"%d\n",13);
        }

        // Write hex type
        for (i = 0; i < mesh->numHex; i++) {
            fprintf(fp,"%d\n",12);
        }
    }


    fclose(fp);

    EG_free(filename);
    printf("Finished writing VTK file\n\n");

    return CAPS_SUCCESS;

}

// Write a mesh contained in the general unstructured mesh structure (genUnstrMesh) in Tecplot format
int write_Tecplot_Tet(char *fname,
                      int asciiFlag,
                      genUnstrMeshStruct *mesh)
{

	int status; // Function status return
	int i, surf; // Indexing

	int numSurface; // Number of surface that have been lummped together
	int numTriFace = 0, numQuadFace = 0, numNodes = 0;
	int triFaceOffset = 0, quadFaceOffset = 0, nodeOffset = 0;

    FILE *fp = NULL, *fpSurface = NULL;
    int numSurfaceVars, cellCenterVars[2] ={-1, -1};
    char filename[512], filenameSurface[512];
    char surfaceVars[512];
    char zoneHeader[512];
    char cellCenterVarsString[512];

    printf("\nWriting TECPLOT file ....\n");

    if (asciiFlag == 0) {
    	printf("\tBinary output is not supported for Tecplot output\n");
    	printf("\t..... switching to ASCII!\n");
    }

    sprintf(filename,"%s.dat",fname);

    sprintf(filenameSurface,"%s_Surface.dat",fname);

    // Write out surface mesh individual if surfaceMesh.xyz is populated
    if (mesh->surfaceMesh.xyz != NULL) {

    	fpSurface = fopen(filenameSurface,"w");
    	if (fpSurface == NULL) {
    		printf("\tUnable to open file: %s\n", filename);
    		status = CAPS_IOERR;
    		goto cleanup;
    	}

    	sprintf(surfaceVars, "VARIABLES = \"x\", \"y\", \"z\"");
    	numSurfaceVars = 3;

    	if (mesh->surfaceMesh.nodeLocal2Global != NULL) {
    		strcat(surfaceVars,", \"local2global\"");
    		numSurfaceVars += 1;
    	}
    	if (mesh->surfaceMesh.triFaceMarkList  != NULL ||
    		mesh->surfaceMesh.quadFaceMarkList != NULL) {

    		strcat(surfaceVars,", \"BC-Marker\"");
    		numSurfaceVars += 1;
    		cellCenterVars[0] = numSurfaceVars;

    	}
    	if (mesh->surfaceMesh.triFaceBLList  != NULL ||
    		mesh->surfaceMesh.quadFaceBLList != NULL) {

    		strcat(surfaceVars,", \"BL-Marker\"");
    		numSurfaceVars += 1;
    		cellCenterVars[1] = numSurfaceVars;
    	}
    	if (mesh->surfaceMesh.nodeBLSpacingList   != NULL) {
    		strcat(surfaceVars,", \"BL-Spacing\"");
    		numSurfaceVars += 1;
    	}
    	if (mesh->surfaceMesh.nodeBLThicknessList != NULL) {
    		strcat(surfaceVars,", \"BL-Thickness\"");
    		numSurfaceVars += 1;
    	}

    	strcat(surfaceVars,"\n");

    	fprintf(fpSurface,"TITLE = \"Surface-Mesh(es)\"\n");
    	fprintf(fpSurface,"%s", surfaceVars);


    	if (mesh->surfaceMesh.numSurface > 0) {
    		numSurface = mesh->surfaceMesh.numSurface;
    	} else {
    		numSurface = 1;
    	}

    	for (surf = 0; surf < numSurface; surf++) {

    		printf("\tSurface %d\n", surf);
    		if (mesh->surfaceMesh.surfTriFaceLump != NULL) {
    			numTriFace = mesh->surfaceMesh.surfTriFaceLump[surf];
    		} else {
    			numTriFace = mesh->surfaceMesh.numTriFace;
    		}

    		if (mesh->surfaceMesh.surfQuadFaceLump != NULL) {
    			numQuadFace = mesh->surfaceMesh.surfQuadFaceLump[surf];
    		} else {
    			numQuadFace = mesh->surfaceMesh.numQuadFace;
    		}

    		if (mesh->surfaceMesh.surfNodesLump != NULL) {
    			numNodes = mesh->surfaceMesh.surfNodesLump[surf];
    		} else {
    			numNodes = mesh->surfaceMesh.numNodes;
    		}

			sprintf(zoneHeader, "ZONE N = %d, E = %d, DATAPACKING = BLOCK, ZONETYPE= FEQUADRILATERAL,",
														             numNodes, numTriFace + numQuadFace);

			if (cellCenterVars[0] > 0 || cellCenterVars[1] > 0 ) {

				strcat(zoneHeader," VARLOCATION= ([");

				if (cellCenterVars[0] > 0 && cellCenterVars[1] > 0) {
					sprintf(cellCenterVarsString,"%d, %d", cellCenterVars[0], cellCenterVars[1] );

				} else if (cellCenterVars[0] > 0 && cellCenterVars[1] <= 0) {

					sprintf(cellCenterVarsString,"%d",cellCenterVars[0]);

				} else if (cellCenterVars[0] <= 0 && cellCenterVars[1] > 0) {

					sprintf(cellCenterVarsString,"%d",cellCenterVars[1]);

				}
				strcat(zoneHeader,cellCenterVarsString);
				strcat(zoneHeader,"]=CELLCENTERED)");
			}

			strcat(zoneHeader,"\n");

			if (surf > 0) fprintf(fpSurface, "%s", "\n");

			fprintf(fpSurface,"%s",zoneHeader);

			// X
			for (i = 0; i < numNodes; i++) {
				fprintf(fpSurface,"%f ", mesh->surfaceMesh.xyz[3*i  + 3*nodeOffset]);
				if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numNodes) fprintf(fpSurface,"\n");
			}

			fprintf(fpSurface,"\n");

			// Y
			for (i = 0; i < numNodes; i++) {
				fprintf(fpSurface,"%f ",mesh->surfaceMesh.xyz[3*i+1 + 3*nodeOffset]);
				if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numNodes) fprintf(fpSurface,"\n");
			}

			fprintf(fpSurface,"\n");

			// Z
			for (i = 0; i < numNodes; i++) {
				fprintf(fpSurface,"%f ", mesh->surfaceMesh.xyz[3*i+2 + 3*nodeOffset]);
				if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numNodes) fprintf(fpSurface,"\n");
			}

			fprintf(fpSurface,"\n");

			// Local to 2 Global connectivity
			if (mesh->surfaceMesh.nodeLocal2Global != NULL) {
				for (i = 0; i < numNodes; i++) {
					fprintf(fpSurface,"%d ", mesh->surfaceMesh.nodeLocal2Global[i + nodeOffset]);
					if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numNodes) fprintf(fpSurface,"\n");

				}

				fprintf(fpSurface,"\n");
			}

			// TriFace BC Marker
			if (mesh->surfaceMesh.triFaceMarkList!= NULL) {
				for (i = 0; i < numTriFace; i++) {
					fprintf(fpSurface,"%d ", mesh->surfaceMesh.triFaceMarkList[i + triFaceOffset]);
					if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numTriFace) fprintf(fpSurface,"\n");

				}

				fprintf(fpSurface,"\n");
			}

			// QuadFace BC Marker
			if (mesh->surfaceMesh.quadFaceMarkList!= NULL) {
				for (i = 0; i < numQuadFace; i++) {
					fprintf(fpSurface,"%d ", mesh->surfaceMesh.quadFaceMarkList[i + quadFaceOffset]);
					if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numQuadFace) fprintf(fpSurface,"\n");

				}
				fprintf(fpSurface,"\n");
			}

			// TriFace BL Marker
			if (mesh->surfaceMesh.triFaceBLList!= NULL) {
				for (i = 0; i < numTriFace; i++) {
					fprintf(fpSurface,"%d ", mesh->surfaceMesh.triFaceBLList[i + triFaceOffset]);
					if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numTriFace) fprintf(fpSurface,"\n");

				}
				fprintf(fpSurface,"\n");
			}

			// QuadFace BL Marker
			if (mesh->surfaceMesh.quadFaceBLList!= NULL) {
				for (i = 0; i < numQuadFace; i++) {
					fprintf(fpSurface,"%d ", mesh->surfaceMesh.quadFaceBLList[i + quadFaceOffset]);
					if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numQuadFace) fprintf(fpSurface,"\n");

				}
				fprintf(fpSurface,"\n");
			}

			// BL initial spacing
			if (mesh->surfaceMesh.nodeBLSpacingList != NULL) {
				for (i = 0; i < numNodes; i++) {
					fprintf(fpSurface,"%f ", mesh->surfaceMesh.nodeBLSpacingList[i + nodeOffset]);
					if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numNodes) fprintf(fpSurface,"\n");

				}
				fprintf(fpSurface,"\n");
			}

			// BL thickness
			if (mesh->surfaceMesh.nodeBLThicknessList != NULL) {
				for (i = 0; i < numNodes; i++) {
					fprintf(fpSurface,"%f ", mesh->surfaceMesh.nodeBLThicknessList[i + nodeOffset]);
					if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numNodes) fprintf(fpSurface,"\n");

				}
				fprintf(fpSurface,"\n");
			}

			// TriFace connectivity
			for (i = 0; i <numTriFace; i++) {
				fprintf(fpSurface,"%d %d %d %d\n",
						mesh->surfaceMesh.localTriFaceList[3*i + 0 + 3*triFaceOffset] - nodeOffset,
						mesh->surfaceMesh.localTriFaceList[3*i + 1 + 3*triFaceOffset] - nodeOffset,
						mesh->surfaceMesh.localTriFaceList[3*i + 2 + 3*triFaceOffset] - nodeOffset,
						mesh->surfaceMesh.localTriFaceList[3*i + 2 + 3*triFaceOffset] - nodeOffset);

			}

			// QuadFace connectivity
			for (i = 0; i < numQuadFace; i++) {
				fprintf(fpSurface,"%d %d %d %d\n",
						mesh->surfaceMesh.localQuadFaceList[4*i + 0 + 4*quadFaceOffset] - nodeOffset,
						mesh->surfaceMesh.localQuadFaceList[4*i + 1 + 4*quadFaceOffset] - nodeOffset,
						mesh->surfaceMesh.localQuadFaceList[4*i + 2 + 4*quadFaceOffset] - nodeOffset,
						mesh->surfaceMesh.localQuadFaceList[4*i + 3 + 4*quadFaceOffset] - nodeOffset);

			}

			triFaceOffset += numTriFace;
			quadFaceOffset += numQuadFace;
			nodeOffset += numNodes;
    	} // End of surface lump loop;

    	fclose(fpSurface);
    	fpSurface = NULL;

    } // Done writing surface

    if (mesh->numTetra == 0) {
    	printf("\tWarning: No tetrahedrals present in volume mesh!\n");
    	status = CAPS_SUCCESS;
    	goto cleanup;

    }

    fp = fopen(filename, "w");
	if (fp == NULL) {
		printf("\tUnable to open file: %s\n", filename);
		status = CAPS_IOERR;
		goto cleanup;
	}

    fprintf(fp,"TITLE = \"Tetra-Mesh\"\n");
    fprintf(fp,"VARIABLES = \"x\", \"y\", \"z\"\n");
    fprintf(fp,"ZONE N = %d, E = %d, DATAPACKING = BLOCK, ZONETYPE= FETETRAHEDRON\n",
            mesh->numNodes, mesh->numTetra);

    // Write nodal coordinates

    // X
    for (i = 0; i < mesh->numNodes; i++) {
    	fprintf(fp,"%f ", mesh->xyz[3*i  ]);
    	if ((i+1) % 7 == 0 && (i+1) != mesh->numNodes) fprintf(fp,"\n");
    }

    fprintf(fp,"\n");

    // Y
    for (i = 0; i < mesh->numNodes; i++) {
    	fprintf(fp,"%f ",mesh->xyz[3*i+1]);
    	if ((i+1) % 7 == 0 && (i+1) != mesh->numNodes) fprintf(fp,"\n");
    }

    fprintf(fp,"\n");

    // Z
    for (i = 0; i < mesh->numNodes; i++) {
    	fprintf(fp,"%f ", mesh->xyz[3*i+2]);
    	if ((i+1) % 7 == 0 && (i+1) != mesh->numNodes) fprintf(fp,"\n");
    }

    fprintf(fp,"\n");

    // Write tetrahedrals connectivity
    for (i = 0; i < mesh->numTetra; i++) {
    	fprintf(fp,"%d %d %d %d\n",
                mesh->tetraList[4*i  ], mesh->tetraList[4*i+1],
                mesh->tetraList[4*i+2], mesh->tetraList[4*i+3]);
    }

    printf("\tFinished writing TECPLOT file\n\n");

    status = CAPS_SUCCESS;

    cleanup:

		if (fp != NULL) fclose(fp);
		if (fpSurface != NULL) fclose(fpSurface);

		return status;
}

int write_MAPBC(char *fname,
                 int numBnds,
                 int *bndID,
                 int *bndVals)
{


	int status; // Function return status

    int i, j; // Indexing

    FILE *fp = NULL;
    char *filename = NULL;
    char fileExt[] = ".mapbc";

    int bndIDMin = 0, bndIDMinIndex =0;
    int *wroteBnd = NULL; // Array to keep track if the boundary has been written

    printf("\nWriting MAPBC file ....\n");

    if (numBnds <= 0) {
    	printf("Warning: Number of boundaries for MAPBC = 0 !\n");
    	status = CAPS_BADVALUE;
    	goto cleanup;
    }

    wroteBnd = (int *) EG_alloc(numBnds*sizeof(int));
    if (wroteBnd == NULL) {
    	status = EGADS_MALLOC;
    	goto cleanup;
    }

    for (i = 0; i < numBnds; i++) wroteBnd[i] = (int) false;

    filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(fileExt)) *sizeof(char));
    if (filename == NULL) {
    	status = EGADS_MALLOC;
    	goto cleanup;
    }

    sprintf(filename,"%s%s",fname, fileExt);

    fp = fopen(filename, "w");

    if (fp == NULL) {
        printf("Unable to open file: %s\n", filename);
        status = CAPS_IOERR;
        goto cleanup;
    }

    // Write out the total number of IDs
    fprintf(fp,"%d\n",numBnds);

    // Make sure the boundary IDs are written out in order
    for (j = 0; j < numBnds; j++) {

    	// Determine lowest boundary ID that hasn't already been written
    	bndIDMin = 1E6;
    	for (i = 0; i < numBnds; i++) {
    		if (bndIDMin >= bndID[i] && wroteBnd[i] != (int) true) {
    			bndIDMin = bndID[i];
    			bndIDMinIndex = i;
    		}
    	}

    	// Write out boundary ID and value
    	fprintf(fp,"%d %d\n", bndID[bndIDMinIndex], bndVals[bndIDMinIndex]);

    	// Mark boundary has being written
    	wroteBnd[bndIDMinIndex] = (int) true;
    }

    printf("Finished writing MAPBC file\n\n");

    status = CAPS_SUCCESS;

    cleanup:
		if (filename != NULL) EG_free(filename);

		if (fp != NULL) fclose(fp);

		if (wroteBnd != NULL) EG_free(wroteBnd);

		return status;
}

static void get_TriangleArea(double p1[3],
					   	   	 double p2[3],
							 double p3[3],
							 double *area) {

	double a = 0.0, b = 0.0, c = 0.0, s = 0.0;

	a = sqrt( (p2[0] - p1[0])*(p2[0] - p1[0]) +
			  (p2[1] - p1[1])*(p2[1] - p1[1]) +
			  (p2[2] - p1[2])*(p2[2] - p1[2]));

	b = sqrt( (p3[0] - p1[0])*(p3[0] - p1[0]) +
			  (p3[1] - p1[1])*(p3[1] - p1[1]) +
			  (p3[2] - p1[2])*(p3[2] - p1[2]));

	c = sqrt( (p3[0] - p2[0])*(p3[0] - p2[0]) +
			  (p3[1] - p2[1])*(p3[1] - p2[1]) +
			  (p3[2] - p2[2])*(p3[2] - p2[2]));


	s = (a + b + c)/2.0;

	*area = sqrt(s*(s - a)*(s - b)*(s - c));
}

void get_Surface_Norm(double p1[3],
                      double p2[3],
                      double p3[3],
                      double norm[3])
{
    double a[3]={0.0,0.0,0.0};
    double b[3]={0.0,0.0,0.0};
    double mag = 0.0;

    // a x b

    // Create two vectors from points
    a[0]= p2[0]-p1[0];
    a[1]= p2[1]-p1[1];
    a[2]= p2[2]-p1[2];

    b[0]= p3[0]-p1[0];
    b[1]= p3[1]-p1[1];
    b[2]= p3[2]-p1[2];

    // Take the cross product
    norm[0] = a[1]*b[2]-a[2]*b[1];

    norm[1] = a[2]*b[0]-a[0]*b[2];

    norm[2] = a[0]*b[1]-a[1]*b[0];

    // Normalize vector
    mag = sqrt(norm[0]*norm[0] + norm[1]*norm[1] + norm[2]*norm[2]);

    norm[0] = norm[0]/mag;
    norm[1] = norm[1]/mag;
    norm[2] = norm[2]/mag;
}

int write_STL(char   *fname,
              int    numTris,
              double *verts,
              int    *conn)
{
	int i; // Indexing

	FILE *fp = NULL;

    double norm[3] = {0,0,0}, p1[3] = {0,0,0}, p2[3] = {0,0,0}, p3[3] = {0,0,0};
    char *filename = NULL;
    char fileExt[] = ".stl";

    printf("Writing STL file ....\n");

    filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(fileExt)) *sizeof(char));
    if (filename == NULL) return EGADS_MALLOC;

    sprintf(filename,"%s%s",fname, fileExt);

    fp = fopen(filename,"w");

    if (fp == NULL) {
        printf("Unable to open file: %s\n", filename);
        if (filename != NULL) EG_free(filename);
        return CAPS_IOERR;
    }

    if (filename != NULL) EG_free(filename);

    fprintf(fp,"solid STL_GEOM\n");

    for (i = 0; i < numTris; i++) {

        p1[0] = verts[3*(conn[3*i + 0] - 1) + 0];
        p1[1] = verts[3*(conn[3*i + 0] - 1) + 1];
        p1[2] = verts[3*(conn[3*i + 0] - 1) + 2];

        p2[0] = verts[3*(conn[3*i + 1] - 1) + 0];
        p2[1] = verts[3*(conn[3*i + 1] - 1) + 1];
        p2[2] = verts[3*(conn[3*i + 1] - 1) + 2];

        p3[0] = verts[3*(conn[3*i + 2] - 1) + 0];
        p3[1] = verts[3*(conn[3*i + 2] - 1) + 1];
        p3[2] = verts[3*(conn[3*i + 2] - 1) + 2];

        get_Surface_Norm(p1,
                         p2,
                         p3,
                         norm);

        fprintf(fp,"\tfacet normal %f %f %f\n",norm[0],norm[1],norm[2]);
        fprintf(fp,"\t\touter loop\n");

        fprintf(fp,"\t\t\tvertex %f %f %f\n",p1[0],p1[1],p1[2]);
        fprintf(fp,"\t\t\tvertex %f %f %f\n",p2[0],p2[1],p2[2]);
        fprintf(fp,"\t\t\tvertex %f %f %f\n",p3[0],p3[1],p3[2]);

        fprintf(fp,"\t\tendloop\n");
        fprintf(fp,"\tendfacet\n");
    }


    fprintf(fp,"endsolid");

    if (fp != NULL) fclose(fp);

    printf("Done Writing STL\n");

    return CAPS_SUCCESS;
}


// Write a NASTRAN mesh file
int write_NASTRAN(char *fname,
                  int asciiFlag, // 0 for binary, anything else for ascii
                  genFEAMeshStruct *nasMesh,
				  feaFileTypeEnum gridFileType)
{

    FILE *fp;
    int i, j, index, gridFields;
    char *filename, *delimiter, *tempString;
    char fileExt[] = ".bdf";

    printf("\nWriting NASTRAN grid and connectivity file (in large field format) ....\n");

    if (asciiFlag == 0) {
        printf("Binary output is not currently supported for working with NASTRAN\n");
        printf("..... switching to ASCII!\n");
        asciiFlag = 1;
    }

    filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(fileExt)) *sizeof(char));
    sprintf(filename,"%s%s", fname, fileExt);

    fp = fopen(filename, "w");
    if (fp == NULL) {
        printf("Unable to open file: %s\n", filename);
        if (filename != NULL) EG_free(filename);
        return CAPS_IOERR;
    }

    if (filename != NULL) EG_free(filename);

    if (gridFileType == LargeField) {
    	fprintf(fp,"$---1A--|-------2-------|-------3-------|-------4-------|-------5-------|-10A--|\n");
    	fprintf(fp,"$---1B--|-------6-------|-------7-------|-------8-------|-------9-------|-10B--|\n");
    } else {
    	fprintf(fp,"$---1---|---2---|---3---|---4---|---5---|---6---|---7---|---8---|---9---|---10--|\n");
    }

    if (gridFileType == FreeField) {
		delimiter = ",";
		gridFields = 8;
	} else {
		delimiter = " ";
		gridFields = 7;
	}


	// Write nodal coordinates
	if (gridFileType == LargeField) {

		for (i = 0; i < nasMesh->numNodes; i++) {
			index = i + 1;
			fprintf(fp,"%-8s %15d", "GRID*", index);

			// If the coord ID == 0 leave blank
			if (nasMesh->xyzCoordID != NULL) {

				if (nasMesh->xyzCoordID[i] != 0) fprintf(fp," %15d", nasMesh->xyzCoordID[i]);
				else                             fprintf(fp,"%16s", "");
			} else {
				fprintf(fp,"%16s", "");
			}

			fprintf(fp," %+14.8e %+14.8e%-8s\n", nasMesh->xyz[3*i  ],
												 nasMesh->xyz[3*i+1],
												 "*");
			fprintf(fp,"%-8s %+14.8e\n", "*",
										  nasMesh->xyz[3*i+2]);
		}

	} else {

		for (i = 0; i < nasMesh->numNodes; i++) {
			index = i + 1;
			fprintf(fp,"%-8s", "GRID");

			tempString = convert_integerToString(index  , 7, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			// If the coord ID == 0 leave blank
			if (nasMesh->xyzCoordID != NULL) {

				if (nasMesh->xyzCoordID[i] != 0) fprintf(fp,"%s%7d", delimiter, nasMesh->xyzCoordID[i]);
				else                             fprintf(fp,"%s%7s", delimiter, "");
			} else {
				fprintf(fp,"%s%7s", delimiter, "");
			}

			for (j = 0; j < 3; j++) {
				tempString = convert_doubleToString(nasMesh->xyz[3*i+j]  , gridFields, 1);
				fprintf(fp,"%s%s", delimiter, tempString);
				EG_free(tempString);
			}

			fprintf(fp,"\n");
		}
	}

	fprintf(fp,"$---1---|---2---|---3---|---4---|---5---|---6---|---7---|---8---|---9---|---10--|\n");
	index = 0; // Element ID

	// 1D elements
	for (i = 0; i < nasMesh->numRod; i++) {
		index += 1;

		fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d\n", "CROD",
											 delimiter, nasMesh->crodElementID[i],
											 delimiter, nasMesh->crodPID[i],
											 delimiter, nasMesh->crod[2*i  ],
											 delimiter, nasMesh->crod[2*i+1]);

	}

	fprintf(fp,"$---1---|---2---|---3---|---4---|---5---|---6---|---7---|---8---|---9---|---10--|\n");

	// 2D elements
	// Write tria3 connectivity
	for (i = 0; i < nasMesh->numTria3; i++) {
		index += 1;

		fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d%s%7d","CTRIA3",
												   delimiter, nasMesh->ctria3ElementID[i],
												   delimiter, nasMesh->ctria3PID[i],
												   delimiter, nasMesh->ctria3[3*i  ],
												   delimiter, nasMesh->ctria3[3*i+1],
												   delimiter, nasMesh->ctria3[3*i+2]);
		// Write coordinate id
		if (nasMesh->ctria3CoordID != NULL) {
			if (nasMesh->ctria3CoordID[i] != 0){
				fprintf(fp, "%s%7d",delimiter,nasMesh->ctria3CoordID[i]);
			}
		}
		fprintf(fp,"\n");
	}

	fprintf(fp,"$---1---|---2---|---3---|---4---|---5---|---6---|---7---|---8---|---9---|---10--|\n");
	// Write cquad4 connectivity
	for (i = 0; i < nasMesh->numQuad4; i++) {
		index += 1;

		fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d", "CQUAD4",
													     delimiter, nasMesh->cquad4ElementID[i],
														 delimiter, nasMesh->cquad4PID[i],
														 delimiter, nasMesh->cquad4[4*i  ],
														 delimiter, nasMesh->cquad4[4*i+1],
														 delimiter, nasMesh->cquad4[4*i+2],
														 delimiter, nasMesh->cquad4[4*i+3]);
		// Write coordinate id
		if (nasMesh->cquad4CoordID != NULL) {
			if (nasMesh->cquad4CoordID[i] != 0){
				fprintf(fp, "%s%7d",delimiter,nasMesh->cquad4CoordID[i]);
			}
		}
		fprintf(fp,"\n");
	}

	fprintf(fp,"$---1---|---2---|---3---|---4---|---5---|---6---|---7---|---8---|---9---|---10--|\n");

	// 3D elements
	// Write tetrahedrals connectivity
	for (i = 0; i < nasMesh->numTetra; i++) {
		index += 1;
		fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d\n", "CTETRA",
													 delimiter, nasMesh->ctetraElementID[i],
													 delimiter, nasMesh->ctetraPID[i],
													 delimiter, nasMesh->ctetra[4*i  ],
													 delimiter, nasMesh->ctetra[4*i+1],
													 delimiter, nasMesh->ctetra[4*i+2],
													 delimiter, nasMesh->ctetra[4*i+3]);
	}

	fprintf(fp,"$---1---|---2---|---3---|---4---|---5---|---6---|---7---|---8---|---9---|---10--|\n");
	// Write hex connectivity
	for (i = 0; i < nasMesh->numHexa; i++) {
		index += 1;

		fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d%s%-8s\n", "CHEXA",
																 delimiter, nasMesh->chexaElementID[i],
																 delimiter, nasMesh->chexaPID[i],
																 delimiter, nasMesh->chexa[8*i  ],
																 delimiter, nasMesh->chexa[8*i+1],
																 delimiter, nasMesh->chexa[8*i+2],
																 delimiter, nasMesh->chexa[8*i+3],
																 delimiter, nasMesh->chexa[8*i+4],
																 delimiter, nasMesh->chexa[8*i+5],
																 delimiter, "+CH");
		fprintf(fp,"%-8s%s%7d%s%7d\n", "+CH",
									delimiter, nasMesh->chexa[8*i+6],
									delimiter, nasMesh->chexa[8*i+7]);
	}

    fclose(fp);

    printf("Finished writing Nastran grid file\n\n");

    return CAPS_SUCCESS;
}

// Write a Abaqus mesh file
int write_Abaqus(char *fname,
                  genFEAMeshStruct *feaMesh,
                  mapAttrToIndexStruct *attrMap)
{

    FILE *fp;
    int i, pidIndex; // Indexing
    int index; // Node/Element tracker
    int status; // Function return status
    char *elemSet;
    char *filename;
    char fileExt[] = "_Mesh.inp";

    if (attrMap == NULL) {
        printf("Error in writing Abaqus mesh file: No attribute to index map provided\n");
        return CAPS_BADVALUE;
    }

    printf("\nWriting Abaqus grid and connectivity file ....\n");

    filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(fileExt)) *sizeof(char));
    sprintf(filename,"%s%s", fname, fileExt);

    fp = fopen(filename, "w");

    if (fp == NULL) {
        printf("Unable to open file: %s\n", filename);
        if (filename != NULL) EG_free(filename);
        return CAPS_IOERR;
    }

    if (filename != NULL) EG_free(filename);

    // Write nodal coordinates

    fprintf(fp, "*NODE\n");
    for (i = 0; i < feaMesh->numNodes; i++) {
        index = i + 1;
        fprintf(fp,"%d, %.6e, %.6e, %.6e\n", index,
                                             feaMesh->xyz[3*i  ],
                                             feaMesh->xyz[3*i+1],
                                             feaMesh->xyz[3*i+2]);
    }

    index = 0; // Element ID

    //////////   1D elements ///////////////////

    // Write rod connectivity
    if (feaMesh->numRod > 0) {

        for (pidIndex = 0; pidIndex < feaMesh->numElement1DPID; pidIndex++) {

            status =  get_mapAttrToIndexKeyword(attrMap, feaMesh->element1DPID[pidIndex], &elemSet);
            if (status != CAPS_SUCCESS) printf("Error finding element set name while writing Abaqus file\n");

            fprintf(fp,"*ELEMENT, TYPE=%s, ELSET=%s\n", "B21", elemSet);

            for (i = 0; i < feaMesh->numRod; i++) {

                if (feaMesh->element1DPID[pidIndex] == feaMesh->crodPID[i]) {

                    index += 1;
                    fprintf(fp,"%d, %d, %d\n",index,
                                              feaMesh->crod[2*i  ],
                                              feaMesh->crod[2*i+1]);
                }
            }
        }
    }

    //////////   2D elements ///////////////////

    // Write tria3 connectivity
    if (feaMesh->numTria3 > 0) {

        for (pidIndex = 0; pidIndex < feaMesh->numElement2DPID; pidIndex++) {

            status =  get_mapAttrToIndexKeyword(attrMap, feaMesh->element2DPID[pidIndex], &elemSet);
            if (status != CAPS_SUCCESS) printf("Error finding element set name while writing Abaqus file\n");

            fprintf(fp,"*ELEMENT, TYPE=%s, ELSET=%s\n", "S3", elemSet);

            for (i = 0; i < feaMesh->numTria3; i++) {

                if (feaMesh->element2DPID[pidIndex] == feaMesh->ctria3PID[i]) {

                    index += 1;
                    fprintf(fp,"%d, %d, %d, %d\n", index,
                                                   feaMesh->ctria3[3*i  ],
                                                   feaMesh->ctria3[3*i+1],
                                                   feaMesh->ctria3[3*i+2]);
                }
            }
        }
    }

    // Write cquad4 connectivity
    if (feaMesh->numQuad4 > 0) {

        for (pidIndex = 0; pidIndex < feaMesh->numElement2DPID; pidIndex++) {

            status =  get_mapAttrToIndexKeyword(attrMap, feaMesh->element2DPID[pidIndex], &elemSet);
            if (status != CAPS_SUCCESS) printf("Error finding element set name while writing Abaqus file\n");

            fprintf(fp,"*ELEMENT, TYPE=%s, ELSET=%s\n", "S4", elemSet);

            for (i = 0; i < feaMesh->numQuad4; i++) {

                if (feaMesh->element2DPID[pidIndex] == feaMesh->cquad4PID[i]) {

                    index += 1;
                    fprintf(fp,"%d, %d, %d, %d, %d\n", index,
                                                       feaMesh->cquad4[4*i  ],
                                                       feaMesh->cquad4[4*i+1],
                                                       feaMesh->cquad4[4*i+2],
                                                       feaMesh->cquad4[4*i+3]);
                }
            }
        }
    }

    //////////   3D elements ///////////////////

    // Write tetrahedrals connectivity
    if (feaMesh->numTetra > 0) {

        for (pidIndex = 0; pidIndex < feaMesh->numElement3DPID; pidIndex++) {

            status =  get_mapAttrToIndexKeyword(attrMap, feaMesh->element3DPID[pidIndex], &elemSet);
            if (status != CAPS_SUCCESS) printf("Error finding element set name while writing Abaqus file\n");

            fprintf(fp,"*ELEMENT, TYPE=%s, ELSET=%s\n", "C3D4", elemSet);

            for (i = 0; i < feaMesh->numTetra; i++) {

                if (feaMesh->element3DPID[pidIndex] == feaMesh->ctetraPID[i]) {

                    index += 1;
                    fprintf(fp,"%d, %d, %d, %d, %d\n", index,
                                                       feaMesh->ctetra[4*i  ],
                                                       feaMesh->ctetra[4*i+1],
                                                       feaMesh->ctetra[4*i+2],
                                                       feaMesh->ctetra[4*i+3]);
                }
            }
        }
    }

    // Write hex connectivity
    if (feaMesh->numHexa > 0) {

        for (pidIndex = 0; pidIndex < feaMesh->numElement3DPID; pidIndex++) {

            status =  get_mapAttrToIndexKeyword(attrMap, feaMesh->element3DPID[pidIndex], &elemSet);
            if (status != CAPS_SUCCESS) printf("Error finding element set name while writing Abaqus file\n");

            fprintf(fp,"*ELEMENT, TYPE=%s, ELSET=%s\n", "C3D8", elemSet);

            for (i = 0; i < feaMesh->numHexa; i++) {

                if (feaMesh->element3DPID[pidIndex] == feaMesh->chexaPID[i]) {

                    index += 1;
                    fprintf(fp,"%d, %d, %d, %d, %d, %d, %d, %d, %d\n", index,
                                                                       feaMesh->chexa[8*i  ],
                                                                       feaMesh->chexa[8*i+1],
                                                                       feaMesh->chexa[8*i+2],
                                                                       feaMesh->chexa[8*i+3],
                                                                       feaMesh->chexa[8*i+4],
                                                                       feaMesh->chexa[8*i+5],
                                                                       feaMesh->chexa[8*i+6],
                                                                       feaMesh->chexa[8*i+7]);
                }
            }
        }
    }

    // Close file
    fclose(fp);

    printf("Finished writing Abaqus grid file\n\n");

    return CAPS_SUCCESS;
}


// Initiate (0 out all values and NULL all pointers) a meshProp in the meshSizingStruct structure format
int initiate_meshSizingStruct (meshSizingStruct *meshProp) {

	meshProp->name = NULL; // Attribute name
	meshProp->attrIndex = 0;  // Attribute index

	meshProp->numEdgePoints = 0; // Number of points along and edge

	meshProp->edgeDistribution = UnknownDistribution; // Distribution function along an edge

	meshProp->initialNodeSpacing[0] = 0.0; // Initial node spacing along an edge
	meshProp->initialNodeSpacing[1] = 0.0;

	meshProp->useTessParams = (int) false; // Trigger to use the specified face tessellation parameters
	meshProp->tessParams[0] = 0.0; // Tessellation parameters on a face.
	meshProp->tessParams[1] = 0.0;
	meshProp->tessParams[2] = 0.0;

	meshProp->boundaryLayerThickness = 0.0; // Boundary layer thickness on a face (3D) or edge (2D)
	meshProp->boundaryLayerSpacing = 0.0; // Boundary layer spacing on a face (3D) or edge (2D)

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) a meshProp in the meshSizingStruct structure format
int destroy_meshSizingStruct (meshSizingStruct *meshProp) {

	if (meshProp->name != NULL) EG_free(meshProp->name);

	meshProp->name = NULL; // Attribute name
	meshProp->attrIndex = 0;  // Attribute index

	meshProp->numEdgePoints = 0; // Number of points along and edge

	meshProp->edgeDistribution = UnknownDistribution; // Distribution function along an edge

	meshProp->initialNodeSpacing[0] = 0.0; // Initial node spacing along an edge
    meshProp->initialNodeSpacing[1] = 0.0;

	meshProp->useTessParams = (int) false; // Trigger to use the specified face tessellation parameters
	meshProp->tessParams[0] = 0.0; // Tessellation parameters on a face.
	meshProp->tessParams[1] = 0.0;
	meshProp->tessParams[2] = 0.0;

	meshProp->boundaryLayerThickness = 0.0; // Boundary layer thickness on a face (3D) or edge (2D)
	meshProp->boundaryLayerSpacing = 0.0; // Boundary layer spacing on a face (3D) or edge (2D)

	return CAPS_SUCCESS;
}


// Fill meshProps in a meshBCStruct format with mesh boundary condition information from incoming Mesh Sizing Tuple
int mesh_getSizingProp(int numTuple,
                   	   capsTuple meshBCTuple[],
					   mapAttrToIndexStruct *attrMap,
					   int *numMeshProp,
					   meshSizingStruct *meshProps[]) {

    /*! \page meshSizingProp Mesh Sizing
     * Structure for the mesh sizing tuple  = ("CAPS Group Name", "Value").
     * "CAPS Group Name" defines the capsGroup on which the sizing information should be applied.
     *	The "Value" can either be a JSON String dictionary (see Section \ref jsonStringMeshSizing) or a single string keyword string
     *	(see Section \ref keyStringMeshSizing)
     */

    int status; //Function return

    int i; // Indexing

    int bcIndex;

    char *keyValue = NULL;
    char *keyWord = NULL;

    // Destroy our meshProps structures coming in if aren't 0 and NULL already
    for (i = 0; i < *numMeshProp; i++) {
    	status = destroy_meshSizingStruct(&(*meshProps)[i]);
    	if (status != CAPS_SUCCESS) return status;
    }

    if (*meshProps != NULL) EG_free(*meshProps);
    *meshProps = NULL;
    *numMeshProp = 0;

    printf("\nGetting mesh sizing parameters\n");

   *numMeshProp = numTuple;

    if (*numMeshProp > 0) {
        *meshProps = (meshSizingStruct *) EG_alloc(*numMeshProp * sizeof(meshSizingStruct));
        if (*meshProps == NULL) {
        	*numMeshProp = 0;
        	return EGADS_MALLOC;
        }

    } else {
        printf("\tNumber of mesh sizing values in input tuple is 0\n");
        return CAPS_NOTFOUND;
    }

    for (i = 0; i < *numMeshProp; i++) {
        status = initiate_meshSizingStruct(&(*meshProps)[i]);
        if (status != CAPS_SUCCESS) return status;
    }

    //printf("Number of tuple pairs = %d\n", numTuple);

    for (i = 0; i < *numMeshProp; i++) {

        printf("\tMesh sizing name - %s\n", meshBCTuple[i].name);

        status = get_mapAttrToIndexIndex(attrMap, (const char *) meshBCTuple[i].name, &(*meshProps)[i].attrIndex);
        if (status == CAPS_NOTFOUND) {
            printf("\tMesh Sizing name \"%s\" not found in attrMap\n", meshBCTuple[i].name);
            return status;
        }

        // Copy mesh sizing name
        if ((*meshProps)[i].name != NULL) EG_free((*meshProps)[i].name);

        (*meshProps)[i].name = (char *) EG_alloc((strlen(meshBCTuple[i].name) + 1)*sizeof(char));
        if ((*meshProps)[i].name == NULL) return EGADS_MALLOC;

        memcpy((*meshProps)[i].name,
        	   meshBCTuple[i].name,
			   strlen(meshBCTuple[i].name)*sizeof(char));

        (*meshProps)[i].name[strlen(meshBCTuple[i].name)] = '\0';

        // Do we have a json string?
        if (strncmp(meshBCTuple[i].value, "{", 1) == 0) {
            //printf("JSON String - %s\n",meshBCTuple[i].value);

            /*! \page meshSizingProp
             * \section jsonStringMeshSizing JSON String Dictionary
             *
             * If "Value" is a JSON string dictionary
			 *  (eg. "Value" = {"edgeDistribution": "Even", "numEdgePoints": 100})
			 *  the following keywords ( = default values) may be used:
             */

        	// Edge properties
        	/*! \page meshSizingProp
			 *
			 * <ul>
             * <li> <B>edgeDistribution = "Even"</B> </li> <br>
             *      Edge Distribution types. Options: Even (even distribution), Tanh (hyperbolic tangent
             *      distribution).
             * </ul>
			 */

        	(*meshProps)[i].edgeDistribution = EvenDistribution;
        	keyWord = "edgeDistribution";
        	status = search_jsonDictionary( meshBCTuple[i].value, keyWord, &keyValue);
        	if (status == CAPS_SUCCESS) {

        		//{Even}
        		if      (strcasecmp(keyValue, "\"Even\"") == 0) (*meshProps)[i].edgeDistribution = EvenDistribution;
        		else if (strcasecmp(keyValue, "\"Tanh\"") == 0) (*meshProps)[i].edgeDistribution = TanhDistribution;
        		else {

        			printf("\tUnrecognized \"%s\" specified (%s) for Mesh_Condition tuple %s, current options are "
        					"\" Even, ... \"\n", keyWord,
												 keyValue,
												 meshBCTuple[i].name);
        			if (keyValue != NULL) EG_free(keyValue);

        			return CAPS_NOTFOUND;
        		}

        		if (keyValue != NULL) {
        			EG_free(keyValue);
        			keyValue = NULL;
        		}
        	}

        	/*! \page meshSizingProp
        	 *
             * <ul>
             * <li>  <B>numEdgePoints = 0</B> </li> <br>
             *  Number of points along an edge.
             * </ul>
             */
        	keyWord = "numEdgePoints";
            status = search_jsonDictionary( meshBCTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

            	status = string_toInteger(keyValue, &(*meshProps)[i].numEdgePoints);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page meshSizingProp
             *
             *
             *  <ul>
             *   <li>  <B>initialNodeSpacing = [0.0, 0.0]</B> </li> <br>
             *   Initial (absolute) node spacing along an edge.
             *  </ul>
             *
             */
            keyWord = "initialNodeSpacing";
            status = search_jsonDictionary( meshBCTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDoubleArray(keyValue, 2, (*meshProps)[i].initialNodeSpacing);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            // Edge/Face properties
			/*! \page meshSizingProp
			 *
			 * \if (AFLR3)
			 *
			 * <ul>
			 * 	<li>  <B>boundaryLayerThickness = 0.0</B> </li> <br>
			 *   Desired boundary layer thickness on a face.
			 * </ul>
			 *
			 * \elseif (AFLR4 || TETGEN || DELAUNDO)
			 * \else
			 *
			 * <ul>
			 * <li>  <B>boundaryLayerThickness = 0.0</B> </li> <br>
			 *  Desired boundary layer thickness on an edge (2D meshing) or face (3D meshing).
			 * </ul>
			 * \endif
			 */
			keyWord = "boundaryLayerThickness";
			status = search_jsonDictionary( meshBCTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				status = string_toDouble(keyValue, &(*meshProps)[i].boundaryLayerThickness);
				if (keyValue != NULL) {
					EG_free(keyValue);
					keyValue = NULL;
				}
				if (status != CAPS_SUCCESS) return status;
			}

			/*! \page meshSizingProp
			 *
			 *\if (AFLR3)
			 *
			 * <ul>
			 * 	<li>  <B>boundaryLayerSpacing = 0.0</B> </li> <br>
			 *   Initial spacing for boundary layer mesh growth on as face.
			 * </ul>
			 *
			 *\elseif (AFLR4 || TETGEN || DELAUNDO)
			 * \else
			 *
			 * <ul>
			 * <li>  <B>boundaryLayerSpacing = 0.0</B> </li> <br>
			 *  Initial spacing for boundary layer mesh growth on an edge (2D meshing) or face (3D meshing).
			 * </ul>
			 * \endif
			 */
			keyWord = "boundaryLayerSpacing";
			status = search_jsonDictionary( meshBCTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				status = string_toDouble(keyValue, &(*meshProps)[i].boundaryLayerSpacing);
				if (keyValue != NULL) {
					EG_free(keyValue);
					keyValue = NULL;
				}
				if (status != CAPS_SUCCESS) return status;
			}


            // Face properties
            /*! \page meshSizingProp
			 *
			 * <ul>
			 * <li>  <B>tessParams = (no default) </B> </li> <br>
			 * Face tessellation parameters, example [0.1, 0.01, 20.0]. (From the EGADS manual) A set of 3 parameters that drive the EDGE discretization
			 * and the FACE triangulation. The first is the maximum length of an EDGE segment or triangle side
			 * (in physical space). A zero is flag that allows for any length. The second is a curvature-based
			 * value that looks locally at the deviation between the centroid of the discrete object and the
			 * underlying geometry. Any deviation larger than the input value will cause the tessellation to
			 * be enhanced in those regions. The third is the maximum interior dihedral angle (in degrees)
			 * between triangle facets (or Edge segment tangents for a WIREBODY tessellation), note that a
			 * zero ignores this phase
			 * </ul>
			 */
            keyWord = "tessParams";
			status = search_jsonDictionary( meshBCTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				string_toDoubleArray(keyValue, 3, (*meshProps)[i].tessParams);
				if (keyValue != NULL) {
					EG_free(keyValue);
					keyValue = NULL;
				}
				if (status != CAPS_SUCCESS) return status;

				(*meshProps)[i].useTessParams = (int) true;

			}

        } else {
            /*! \page meshSizingProp
             * \section keyStringMeshSizing Single Value String
             *
             * If "Value" is a single string the following options maybe used:
             * - (NONE Currently)
             * */

        }
    }

    printf("\tDone getting mesh sizing parameters\n");
    return CAPS_SUCCESS;
}

// Return the desired scale factor "delta" for a given spacing "ds" evaluated at point "epi" along
// interval "I" - Used when only a single end of an edge has a desired spacing - Vinokur 1980
static double eqn_stretchingFactorSingleSided(double delta, double inputVars[]) {

    double value;
    double epi = inputVars[0];
    double I   = inputVars[1];
    double ds  = inputVars[2];

    // Derivative of of 1 + tanh(delta(epi/I - 1))/tanh(delta)
    value = -1.0*(tanh(delta*(epi*I + 1))*(tanh(delta)*tanh(delta) - 1)) / (tanh(delta)*tanh(delta)) +
    		  ((tanh(delta*(epi*I + 1))*tanh(delta*(epi*I + 1)) - 1)*(epi*I + 1))/tanh(delta);

    return value - ds;
}

// Return the desired scale factor "delta" for a given "B" - Used when both ends of edge have a
// desired spacing - Vinokur 1980
static double eqn_stretchingFactorDoubleSided(double delta, double inputVars[]) {

    double value;
    double B  = inputVars[0];

    value = sinh(delta)/delta;

    return value - B;
}

static double root_BisectionMethod( double (*f)(double, double *),
                                    double lowerBnd,
                                    double upperBnd,
                                    double inputVars[]){

    // BiSection variables
    double epsilon = 1E-6;

    double midBnd = 0.0, midVal = 0.0, upperVal = 0.0;

    do {
        midBnd = (lowerBnd + upperBnd)/2;

        upperVal = (*f)(upperBnd, inputVars);
        midVal   = (*f)(midBnd, inputVars);

        if (midVal == 0.0) {
            upperBnd = midBnd;
            break;
        }

        if (upperVal * midVal >= 0) {
            upperBnd = midBnd;

        } else {
            lowerBnd = midBnd;
        }

    } while ((upperBnd - lowerBnd) > epsilon);

    return upperBnd;
}


// Modify the EGADS body tessellation based on given inputs
int mesh_modifyBodyTess(int numMeshProp,
						meshSizingStruct meshProp[],
						int minEdgePointGlobal,
						int maxEdgePointGlobal,
						double tessParamGlobal[],
						mapAttrToIndexStruct attrMap,
						int numBody,
						ego *bodies[]) {

	int status; // Function return

	int i, j, bodyIndex, faceIndex, edgeIndex, attrIndex; // Indexing

	// Body parameters
	double bodySize = 0, box[6];

	// Face parameters
	int numFace = 0;
	ego *faces = NULL;

	// Edge parameters
	int numEdge = 0; // Number of edges
	ego *edges = NULL; // EGADS edge object;


	int numEdgePoint = 0, periodicEdge = 0;

	edgeDistributionEnum edgeDistribution;
	double initialNodeSpacing[2] = {0.0, 0.0};

	double *rPos = NULL;

	double arcLength = 0.0, edgeRange[2], edgeTessParam = 0.0;

	double A, B, I, U, stretchingFactor = 0.0, epi, inputVars[3]; // Tanh Helpers



    // Mesh attribute parameters
	const char *groupName = NULL;


	// Set body tessellation parameters
	for (bodyIndex = 0; bodyIndex < numBody; bodyIndex++) {

		status = EG_getBodyTopos((*bodies)[bodyIndex], NULL, EDGE, &numEdge, &edges);
		if (status !=  EGADS_SUCCESS) goto bail;

		status = EG_getBodyTopos((*bodies)[bodyIndex], NULL, FACE, &numFace, &faces);
		if (status !=  EGADS_SUCCESS) goto bail;

		if (minEdgePointGlobal != 0 && maxEdgePointGlobal != 0) {

			// Get bounding box for scaling
			status = EG_getBoundingBox((*bodies)[bodyIndex], box);
			if (status != EGADS_SUCCESS) goto bail;

			bodySize = sqrt((box[0]-box[3])*(box[0]-box[3]) +
							(box[1]-box[4])*(box[1]-box[4]) +
							(box[2]-box[5])*(box[2]-box[5]));
		}

		// Loop over edges for each body
		for (edgeIndex = 0; edgeIndex < numEdge; edgeIndex++) {

			numEdgePoint = 0;
			edgeDistribution = EvenDistribution;

			status = EG_getRange(edges[edgeIndex], edgeRange, &periodicEdge);
			if (status != EGADS_SUCCESS) goto bail;

			status = EG_arcLength(edges[edgeIndex], edgeRange[0], edgeRange[1], &arcLength);
			if (status != EGADS_SUCCESS) goto bail;

			if (minEdgePointGlobal != 0 && maxEdgePointGlobal != 0) {

				edgeTessParam = tessParamGlobal[0]*bodySize;

				numEdgePoint = (int) min_DoubleVal(
										 max_DoubleVal(
												 max_DoubleVal( (double) minEdgePointGlobal, 2.0) ,
												 (double) (1+arcLength/edgeTessParam)),
										 (double) maxEdgePointGlobal);
			}

			status = retrieve_CAPSGroupAttr(edges[edgeIndex], &groupName);
			if (status == EGADS_SUCCESS) {
				status = get_mapAttrToIndexIndex(&attrMap, groupName, &attrIndex);
				if (status == CAPS_SUCCESS) {

					for (i = 0; i < numMeshProp; i++) {

						if (meshProp[i].attrIndex == attrIndex) {
							numEdgePoint = meshProp[i].numEdgePoints;
							edgeDistribution =	meshProp[i].edgeDistribution;

							//initialNodeSpacing[0] = meshProp[i].initialNodeSpacing[0]/arcLength;
							//initialNodeSpacing[1] = meshProp[i].initialNodeSpacing[1]/arcLength;

							initialNodeSpacing[0] = meshProp[i].initialNodeSpacing[0];
							initialNodeSpacing[1] = meshProp[i].initialNodeSpacing[1];

							break;
						}
					}
				}
			} else if (status != EGADS_SUCCESS && status != EGADS_NOTFOUND) {
				goto bail;
			}

			if (numEdgePoint >= 2) {

				if (numEdgePoint == 2) {

					i = 0;

					status = EG_attributeAdd(edges[edgeIndex],
											 ".rPos", ATTRINT, 1, &i, NULL, NULL);

					if (status != EGADS_SUCCESS) goto bail;

				} else {

					rPos  = (double *) EG_alloc(numEdgePoint*sizeof(double));
					if (rPos == NULL) {
						status = EGADS_MALLOC;
						goto bail;
					}

					if (edgeDistribution == EvenDistribution) {

						// Even distribution
						for (i = 0; i < numEdgePoint; i++) {
							rPos[i] = (double) (i+1) / (double)(numEdgePoint-1);
						}

					} else if (edgeDistribution == TanhDistribution) {

					    // Default to "even" in case something is wrong
					    for (i = 0; i < numEdgePoint; i++) {
					        rPos[i] = (double) (i+1) / (double)(numEdgePoint-1);
					    }

					    // Tanh distribution - both nodes specified
					    if (initialNodeSpacing[0] > 0.0 &&
					        initialNodeSpacing[1] > 0.0) { // Both ends are specified

					    	I  = (double) numEdgePoint;

					        A = sqrt(initialNodeSpacing[1]) / sqrt(initialNodeSpacing[0]);

					        B = 1/ (I*sqrt(initialNodeSpacing[0] * initialNodeSpacing[1]));

					        inputVars[0] = B;

					        stretchingFactor = root_BisectionMethod( &eqn_stretchingFactorDoubleSided, 0, 1000, inputVars);
					        //printf("StretchingFactor = %f\n", stretchingFactor);

					        for (i = 0; i < numEdgePoint; i++) {

					        	epi = (double) i+1;

					        	U = 0.5 * ( 1.0 + tanh(stretchingFactor*(epi/I-0.5))/tanh(stretchingFactor/2));

					            rPos[i] = U/ (A + (1-A)*U);
					        }

					    } else if (initialNodeSpacing[0] > 0.0 &&
					               initialNodeSpacing[1] <= 0.0) {

					    	I = (double)(numEdgePoint);

					    	inputVars[0] = 1.0; // epi = 1
					    	inputVars[1] = I;
					    	inputVars[2] = initialNodeSpacing[0];

					    	stretchingFactor = root_BisectionMethod( &eqn_stretchingFactorSingleSided, 0, 1000, inputVars);
					    	//printf("0 - stretchingFactor = %f\n", stretchingFactor);

					    	for (i = 0; i < numEdgePoint; i++) {

					    		epi = (double) i+1;

								rPos[i] = 1.0 + tanh(stretchingFactor*(epi/I-1)) /tanh(stretchingFactor);
							}

					    } else if (initialNodeSpacing[0] <= 0.0 &&
					               initialNodeSpacing[1] > 0.0) {

					    	I = (double)(numEdgePoint);

					    	inputVars[0] = 1.0; // epi = 1
					    	inputVars[1] = I;
					    	inputVars[2] = initialNodeSpacing[1];

					    	stretchingFactor = root_BisectionMethod( &eqn_stretchingFactorSingleSided, 0, 1000, inputVars);
					    	//printf("1 - stretchingFactor = %f\n", stretchingFactor);

					    	j = numEdgePoint-1;
							for (i = 0; i < numEdgePoint; i++) {

								epi = (double) i;

								rPos[j] = 1.0 - (1.0 + tanh(stretchingFactor*(epi/I-1)) /tanh(stretchingFactor));
								j -= 1;
							}
					    }

					    // Debug
					    //for (i = 0; i < numEdgePoint-1; i++) printf("EdgeIndex = %d, Rpos(%d) delta= %f\n", edgeIndex, i, rPos[i+1]-rPos[i]);

					} else {

						printf("Unknown distribution function\n");
						goto bail;
					}


					//printf("Number of points added to edge %d (body = %d) = %d\n", edgeIndex, bodyIndex, numEdgePoint);
					status = EG_attributeAdd(edges[edgeIndex],
											 ".rPos", ATTRREAL, numEdgePoint-2, NULL, rPos, NULL);

					if (status != EGADS_SUCCESS) goto bail;
				}
			}

			if (rPos != NULL) EG_free(rPos);
			rPos = NULL;
		}

		if (edges != NULL) EG_free(edges);
		edges = NULL;

		// Loop over faces for each body
		for (faceIndex = 0; faceIndex < numFace; faceIndex++) {

			status = retrieve_CAPSGroupAttr(faces[faceIndex], &groupName);
			if (status == EGADS_SUCCESS) {
				status = get_mapAttrToIndexIndex(&attrMap, groupName, &attrIndex);
				if (status == CAPS_SUCCESS) {

					for (i = 0; i < numMeshProp; i++) {

						if (meshProp[i].attrIndex == attrIndex) {

							if (meshProp[i].useTessParams == (int) true) {
								status = EG_attributeAdd(faces[faceIndex],
														 ".tParams",
														 ATTRREAL,
														 3,
														 NULL,
														 meshProp[i].tessParams, NULL);

							}

							break;
						}
					}
				}
			} else if (status != EGADS_SUCCESS && status != EGADS_NOTFOUND) {
				goto bail;
			}
		}

		if (faces != NULL) EG_free(faces);
		faces = NULL;
	}

	if (faces != NULL) EG_free(faces);
	if (edges != NULL) EG_free(edges);
	if (rPos != NULL) EG_free(rPos);

	return CAPS_SUCCESS;

	bail:
		if (faces != NULL) EG_free(faces);
		if (edges != NULL) EG_free(edges);
		if (rPos != NULL) EG_free(rPos);

		return status;
}
