#include <string.h>
#include "feaTypes.h"  // Bring in FEA structures
#include "capsTypes.h" // Bring in CAPS types

#include "miscUtils.h" //Bring in misc. utility functions
#include "nastranUtils.h" // Bring in nastran utility header

#ifdef WIN32
#define strcasecmp  stricmp
#endif

// Write Nastran coordinate system card from a feaCoordSystemStruct structure
int nastran_writeCoordinateSystemCard(FILE *fp, feaCoordSystemStruct *feaCoordSystem, feaFileFormatStruct *feaFileFormat) {

	int  fieldWidth;
	char *tempString, *delimiter;

	double pointA[3] = {0, 0, 0};
	double pointB[3] = {0, 0, 0};
	double pointC[3] = {0, 0, 0};

    if (feaFileFormat->fileType == FreeField) {
		delimiter = ",";
		fieldWidth = 8;
	} else {
		delimiter = " ";
		fieldWidth = 7;
	}

    pointA[0] = feaCoordSystem->origin[0];
    pointA[1] = feaCoordSystem->origin[1];
    pointA[2] = feaCoordSystem->origin[2];

    pointB[0] = feaCoordSystem->normal3[0] + pointA[0];
    pointB[1] = feaCoordSystem->normal3[1] + pointA[1];
    pointB[2] = feaCoordSystem->normal3[2] + pointA[2];

    pointC[0] = feaCoordSystem->normal1[0] + pointB[0];
    pointC[1] = feaCoordSystem->normal1[1] + pointB[1];
    pointC[2] = feaCoordSystem->normal1[2] + pointB[2];


    /*
    pointB[0] = feaCoordSystem->normal1[0] + pointA[0];
	pointB[1] = feaCoordSystem->normal1[1] + pointA[1];
	pointB[2] = feaCoordSystem->normal1[2] + pointA[2];

	pointC[0] = feaCoordSystem->normal2[0] + pointB[0];
	pointC[1] = feaCoordSystem->normal2[1] + pointB[1];
	pointC[2] = feaCoordSystem->normal2[2] + pointB[2];
	*/


	// Rectangular
	if (feaCoordSystem->coordSystemType == RectangularCoordSystem) fprintf(fp,"%-8s", "CORD2R");

	// Spherical
	if (feaCoordSystem->coordSystemType == SphericalCoordSystem)  fprintf(fp,"%-8s", "CORD2S");

	// Cylindrical
	if (feaCoordSystem->coordSystemType == CylindricalCoordSystem) fprintf(fp,"%-8s", "CORD2C");

	if (feaCoordSystem->coordSystemType == RectangularCoordSystem ||
		feaCoordSystem->coordSystemType == SphericalCoordSystem   ||
		feaCoordSystem->coordSystemType == CylindricalCoordSystem) {

		tempString = convert_integerToString( feaCoordSystem->coordSystemID, 7, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		tempString = convert_integerToString( feaCoordSystem->refCoordSystemID, fieldWidth, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString( pointA[0], fieldWidth, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString( pointA[1], fieldWidth, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString( pointA[2], fieldWidth, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString( pointB[0], fieldWidth, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString( pointB[1], fieldWidth, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString( pointB[2], fieldWidth, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		// Continuation line
		if (feaFileFormat->fileType == FreeField) {
			fprintf(fp, ",+ ");
		} else {
			fprintf(fp, "+%7s", "");
		}

		fprintf(fp, "\n");

		// Continuation line
		if (feaFileFormat->fileType == FreeField) {
			fprintf(fp, ",+ ");
		} else {
			fprintf(fp, "+%7s", "");
		}

		tempString = convert_doubleToString( pointC[0], fieldWidth, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString( pointC[1], fieldWidth, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString( pointC[2], fieldWidth, 1);
		fprintf(fp, "%s%s\n",delimiter, tempString);
		EG_free(tempString);

	} else {

		printf("Unrecognized coordinate system type !!\n");
		return CAPS_BADVALUE;
	}

	return CAPS_SUCCESS;
}

// Write combined Nastran constraint card from a set of constraint IDs.
// 	The combined constraint ID is set through the constraintID variable.
int nastran_writeConstraintADDCard(FILE *fp, int constraintID, int numSetID, int constraintSetID[], feaFileFormatStruct *feaFileFormat) {

	int i; // Indexing

	int sidIndex= 0, lineCount = 0; // Counters

	int status; // Return status;

	char *delimiter;

	if (feaFileFormat->fileType == FreeField) {
		delimiter = ",";
	} else {
		delimiter = " ";
	}

  	if (numSetID != 0) fprintf(fp,"%-8s%s%7d", "SPCADD",delimiter , constraintID);

  	lineCount = 1;
  	for (i = 0; i < numSetID; i++) {
  	    sidIndex += 1;
  	    if (sidIndex % (8*lineCount) == 0) {

  	    	if (feaFileFormat->fileType == FreeField) {
  	    		fprintf(fp, ",+C%-5d\n", lineCount-1); // Start of continuation
  	    		fprintf(fp, "+C%-5d,"  , lineCount-1);  // End of continuation
  	    	} else {
  	    		fprintf(fp, "+C%-6d\n", lineCount-1); // Start of continuation
  	    		fprintf(fp, "+C%-6d"  , lineCount-1);  // End of continuation
  	    	}

  			lineCount += 1;
  		}

  	    fprintf(fp,"%s%7d",delimiter, constraintSetID[i]);
  	}

  	if (numSetID != 0) fprintf(fp,"\n\n");

	return CAPS_SUCCESS;
}

// Write Nastran constraint card from a feaConstraint structure
int nastran_writeConstraintCard(FILE *fp, feaConstraintStruct *feaConstraint, feaFileFormatStruct *feaFileFormat) {

    int i; // Indexing;
    char *tempString, *delimiter;

    if (feaFileFormat->fileType == FreeField) {
		delimiter = ",";
	} else {
		delimiter = " ";
	}

    if (feaConstraint->constraintType == Displacement) {

        for (i = 0; i < feaConstraint->numGridID; i++) {
            fprintf(fp,"%-8s", "SPC");

            tempString = convert_integerToString(feaConstraint->constraintID, 7, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_integerToString(feaConstraint->gridIDSet[i], 7, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_integerToString(feaConstraint->dofConstraint, 7, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString( feaConstraint->gridDisplacement, 7, 1);
            fprintf(fp, "%s%s\n",delimiter, tempString);
            EG_free(tempString);
        }
	}

	if (feaConstraint->constraintType == ZeroDisplacement) {

	    for (i = 0; i < feaConstraint->numGridID; i++) {

            fprintf(fp,"%-8s", "SPC1");

            tempString = convert_integerToString(feaConstraint->constraintID , 7, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_integerToString(feaConstraint->dofConstraint, 7, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_integerToString(feaConstraint->gridIDSet[i], 7, 1);
            fprintf(fp, "%s%s\n",delimiter, tempString);
            EG_free(tempString);
        }
	}

	return CAPS_SUCCESS;
}

// Write a Nastran Material card from a feaMaterial structure
int nastran_writeMaterialCard(FILE *fp, feaMaterialStruct *feaMaterial, feaFileFormatStruct *feaFileFormat) {

	int  fieldWidth;
	char *tempString, *delimiter;

    if (feaFileFormat->fileType == FreeField) {
		delimiter = ",";
		fieldWidth = 8;
	} else {
		delimiter = " ";
		fieldWidth = 7;
	}

	// Isotropic
	if (feaMaterial->materialType == Isotropic) {

	    /*
        fprintf(fp,"%-8s %s %s %s %s %s\n", "MAT1",
                                            convert_integerToString(feaMaterial->materialID  , 7, 1),
                                            convert_doubleToString( feaMaterial->youngModulus, 7, 1),
                                            convert_doubleToString( feaMaterial->shearModulus, 7, 1),
                                            convert_doubleToString( feaMaterial->poissonRatio, 7, 1),
                                            convert_doubleToString( feaMaterial->density     , 7, 1));
        */

		fprintf(fp,"%-8s", "MAT1");

		tempString = convert_integerToString(feaMaterial->materialID, 7, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString( feaMaterial->youngModulus, fieldWidth, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString( feaMaterial->shearModulus, fieldWidth, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString( feaMaterial->poissonRatio, fieldWidth, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString( feaMaterial->density, fieldWidth, 1);
		fprintf(fp, "%s%s\n",delimiter, tempString);
		EG_free(tempString);
	}

	// Orthotropic
	if (feaMaterial->materialType == Orthotropic) {

	    fprintf(fp,"%-8s", "MAT8");

        tempString = convert_integerToString(feaMaterial->materialID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaMaterial->youngModulus, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaMaterial->youngModulusLateral, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaMaterial->poissonRatio, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaMaterial->shearModulus, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaMaterial->shearModulusTrans1Z, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaMaterial->shearModulusTrans2Z, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaMaterial->density, fieldWidth, 1);
        fprintf(fp, "%s%s\n",delimiter, tempString);
        EG_free(tempString);
    }

	return CAPS_SUCCESS;
}

// Write a Nastran Property card from a feaProperty structure
int nastran_writePropertyCard(FILE *fp, feaPropertyStruct *feaProperty, feaFileFormatStruct *feaFileFormat) {

	int i,j; // Indexing
	int  fieldWidth;
    char *tempString=NULL;
    char *delimiter;

    if (feaFileFormat->fileType == FreeField) {
   		delimiter = ",";
   		fieldWidth = 8;
   	} else {
   		delimiter = " ";
   		fieldWidth = 7;
   	}

	//          1D Elements        //

	// Rod
	if (feaProperty->propertyType ==  Rod) {

        fprintf(fp,"%-8s", "PROD");

        tempString = convert_integerToString(feaProperty->propertyID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_integerToString(feaProperty->materialID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->crossSecArea, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->torsionalConst, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->torsionalStressReCoeff, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->massPerLength, fieldWidth, 1);
        fprintf(fp, "%s%s\n",delimiter, tempString);
        EG_free(tempString);

	}

	// Bar
	if (feaProperty->propertyType ==  Bar) {

	    fprintf(fp,"%-8s", "PBAR");

		tempString = convert_integerToString(feaProperty->propertyID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_integerToString(feaProperty->materialID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->crossSecArea, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->zAxisInertia, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->yAxisInertia, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->torsionalConst, fieldWidth, 1);
        fprintf(fp, "%s%s\n",delimiter, tempString);
        EG_free(tempString);

	}

	//          2D Elements        //

	// Shell
	if (feaProperty->propertyType == Shell) {

	    fprintf(fp,"%-8s", "PSHELL");

		tempString = convert_integerToString(feaProperty->propertyID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_integerToString(feaProperty->materialID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString(feaProperty->membraneThickness, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        if (feaProperty->materialBendingID != 0) {
            tempString = convert_integerToString(feaProperty->materialBendingID, 7, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString(feaProperty->bendingInertiaRatio, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

        } else { // Print a blank
        	if (feaProperty->materialShearID != 0 | feaFileFormat->fileType == FreeField) {
        		if (feaProperty->massPerArea != 0) {
					fprintf(fp, ", , ");
				}
        	} else {
        		fprintf(fp, " %7s  %7s", "", "");
        	}
        }

        if (feaProperty->materialShearID != 0) {
            tempString = convert_integerToString(feaProperty->materialShearID, 7, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString(feaProperty->shearMembraneRatio, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);
        } else {// Print a blank
        	if (feaFileFormat->fileType == FreeField) {
        		if (feaProperty->massPerArea != 0) {
        			fprintf(fp, ", , ");
        		}
			} else {
				fprintf(fp, " %7s  %7s", "", "");
			}
        }

        if (feaProperty->massPerArea != 0){
			tempString = convert_doubleToString(feaProperty->massPerArea, fieldWidth, 1);
			fprintf(fp, "%s%s",delimiter, tempString);
			EG_free(tempString);
        }

        fprintf(fp, "\n");
	}

	// Composite
		if (feaProperty->propertyType == Composite) {
			fprintf(fp,"%-8s", "PCOMP");

			// PID
			tempString = convert_integerToString(feaProperty->propertyID, 7, 1);
			fprintf(fp, "%s%s",delimiter, tempString);
			if (tempString != NULL) EG_free(tempString);

			// BLANK FIELD Z0
			if (feaFileFormat->fileType == FreeField) {
				fprintf(fp, ", ");
			} else {
				fprintf(fp, " %7s", "");
			}

			// NSM
			if (feaProperty->massPerArea != 0){
				tempString = convert_doubleToString(feaProperty->massPerArea, fieldWidth, 1);
				fprintf(fp, "%s%s",delimiter, tempString);
				EG_free(tempString);
			} else {// Print a blank
				if (feaFileFormat->fileType == FreeField) {
					fprintf(fp, ", ");
				} else {
					fprintf(fp, " %7s", "");
				}
	        }

			// SHEAR BOND ALLOWABLE SB
			tempString = convert_doubleToString(feaProperty->compositeShearBondAllowable, fieldWidth, 1);
			fprintf(fp, "%s%s",delimiter, tempString);
			if (tempString != NULL) EG_free(tempString);

			// FAILURE THEORY
			fprintf(fp, "%s%7s",delimiter, feaProperty->compositeFailureTheory);

			// BLANK FIELD TREF
			if (feaFileFormat->fileType == FreeField) {
				fprintf(fp, ", ");
			} else {
				fprintf(fp, " %7s", "");
			}

			// BLANK FIELD DAMPING
			if (feaFileFormat->fileType == FreeField) {
				fprintf(fp, ", ");
			} else {
				fprintf(fp, " %7s", "");
			}

			// LAM
			if (feaProperty->compositeSymmetricLaminate == 1){
				fprintf(fp, "%s%7s",delimiter, "SYM");
			} else {// Print a blank
				if (feaFileFormat->fileType == FreeField) {
					fprintf(fp, ", ");
				} else {
					fprintf(fp, " %7s", "");
				}
			}

			// CONTINUATION LINE
			if (feaFileFormat->fileType == FreeField) {
				fprintf(fp, ",+ ");
			} else {
				fprintf(fp, "+%7s", "");
			}

			fprintf(fp, "\n");

			// LOOP OVER PLYS
			for (i = 0; i < feaProperty->numPly; i++) {

				if ( (double)i/2.0 == (double) (i/2) ) {
					// CONINUATION LINE
					if (feaFileFormat->fileType == FreeField) {
						fprintf(fp, "+ ");
					} else {
						fprintf(fp, "+%7s", "");
					}
				}

				// MID
				tempString = convert_integerToString(feaProperty->compositeMaterialID[i], 7, 1);
				fprintf(fp, "%s%s",delimiter, tempString);
				if (tempString != NULL) EG_free(tempString);

				// THICKNESS
				if (feaProperty->compositeThickness != NULL) {
					tempString = convert_doubleToString(feaProperty->compositeThickness[i], fieldWidth, 1);
					fprintf(fp, "%s%s",delimiter, tempString);
					if (tempString != NULL) EG_free(tempString);

				} else {
					if (feaFileFormat->fileType == FreeField) {
						fprintf(fp, ", ");
					} else {
						fprintf(fp, " %7s", "");
					}
				}

				// THETA
				if (feaProperty->compositeOrientation != NULL) {
					tempString = convert_doubleToString(feaProperty->compositeOrientation[i], fieldWidth, 1);
					fprintf(fp, "%s%s",delimiter, tempString);
					if (tempString != NULL) EG_free(tempString);
				} else {

					if (feaFileFormat->fileType == FreeField) {
						fprintf(fp, ", ");
					} else {
						fprintf(fp, " %7s", "");
					}
				}

				// BLANK STRESS / STRAIN OUTPUT
				if (i < feaProperty->numPly-1) {
					if (feaFileFormat->fileType == FreeField) {
						fprintf(fp, ", ");
					} else {
						fprintf(fp, " %7s", "");
					}
				}

				if ( (double)i/2.0 != (double) (i/2) ) {
					// CONINUATION LINE
					if (i < feaProperty->numPly-1) {
						if (feaFileFormat->fileType == FreeField) {
							fprintf(fp, ",+ ");
						} else {
							fprintf(fp, "+%7s", "");
						}
					}
					fprintf(fp, "\n");
				}

				if (i == feaProperty->numPly-1){
					fprintf(fp, "\n");
				}

			}

		}

    //          3D Elements        //

    // Solid
    if (feaProperty->propertyType == Solid) {

        fprintf(fp,"%-8s", "PSOLID");

        tempString = convert_integerToString(feaProperty->propertyID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_integerToString(feaProperty->materialID, 7, 1);
        fprintf(fp, "%s%s\n",delimiter, tempString);
        EG_free(tempString);
    }

	return CAPS_SUCCESS;
}

// Write a combined Nastran load card from a set of load IDs. Uses the feaLoad structure to get the local scale factor
// 	for the load. The overall load scale factor is set to 1. The combined load ID is set through the loadID variable.
int nastran_writeLoadADDCard(FILE *fp, int loadID, int numSetID, int loadSetID[], feaLoadStruct feaLoad[], feaFileFormatStruct *feaFileFormat) {

    int i, loadIndex; // Indexing

    int sidIndex= 0, lineCount = 0; // Counters

    int status; // Return status;

    double overallScale = 1.0;

    char *tempString, *delimiter;

    if (feaFileFormat->fileType == FreeField) {
		delimiter = ",";
	} else {
		delimiter = " ";
	}

    if (numSetID != 0) {
        fprintf(fp,"%-8s%s%7d", "LOAD", delimiter,
                               loadID);

        tempString = convert_doubleToString(overallScale, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);
    }

    lineCount = 1;
    for (i = 0; i < numSetID; i++) {
    	sidIndex += 2;
    	if (sidIndex % (8*lineCount) == 0) {

    		if (feaFileFormat->fileType == FreeField) {
				fprintf(fp, ",+L%-5d\n", lineCount-1); // Start of continuation
				fprintf(fp, "+L%-5d,"  , lineCount-1);  // End of continuation
			} else {
				fprintf(fp, "+L%-6d\n", lineCount-1);  // Start of continuation
				fprintf(fp, "+L%-6d"  , lineCount-1);  // End of continuation
			}

    		lineCount += 1;
    	}

    	// ID's are 1 bias
		tempString = convert_doubleToString(feaLoad[loadSetID[i]-1].loadScaleFactor, 7, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

    	fprintf(fp,"%s%7d",delimiter, loadSetID[i]);
    }

    if (numSetID != 0) fprintf(fp,"\n\n");

    return CAPS_SUCCESS;
}

// Write a Nastran Load card from a feaLoad structure
int nastran_writeLoadCard(FILE *fp, feaLoadStruct *feaLoad, feaFileFormatStruct *feaFileFormat)
{
    int i, j; // Indexing
    int fieldWidth;
    char *tempString=NULL;
    char *delimiter;

    if (feaFileFormat->fileType == FreeField) {
		delimiter = ",";
		fieldWidth = 8;
	} else {
		delimiter = " ";
		fieldWidth = 7;
	}

	// Concentrated force at a grid point
	if (feaLoad->loadType ==  GridForce) {

	    for (i = 0; i < feaLoad->numGridID; i++) {

            fprintf(fp,"%-8s", "FORCE");

            tempString = convert_integerToString(feaLoad->loadID, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_integerToString(feaLoad->gridIDSet[i], fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_integerToString(feaLoad->coordSystemID, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString(feaLoad->forceScaleFactor, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString(feaLoad->directionVector[0], fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString(feaLoad->directionVector[1], fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString(feaLoad->directionVector[2], fieldWidth, 1);
            fprintf(fp, "%s%s\n",delimiter, tempString);
            EG_free(tempString);
	    }
	}

	// Concentrated moment at a grid point
	if (feaLoad->loadType ==  GridMoment) {

	    for (i = 0; i < feaLoad->numGridID; i++) {
            fprintf(fp,"%-8s", "MOMENT");

            tempString = convert_integerToString(feaLoad->loadID, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_integerToString(feaLoad->gridIDSet[i], fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_integerToString(feaLoad->coordSystemID, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString(feaLoad->momentScaleFactor, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString(feaLoad->directionVector[0], fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString(feaLoad->directionVector[1], fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString(feaLoad->directionVector[2], fieldWidth, 1);
            fprintf(fp, "%s%s\n",delimiter, tempString);
            EG_free(tempString);
	    }
	}

	// Gravitational load
	if (feaLoad->loadType ==  Gravity) {

		fprintf(fp,"%-8s", "GRAV");

		tempString = convert_integerToString(feaLoad->loadID, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_integerToString(feaLoad->coordSystemID, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString(feaLoad->gravityAcceleration, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString(feaLoad->directionVector[0], fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString(feaLoad->directionVector[1], fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString(feaLoad->directionVector[2], fieldWidth, 1);
        fprintf(fp, "%s%s\n",delimiter, tempString);
        EG_free(tempString);
	}

	// Pressure load
	if (feaLoad->loadType ==  Pressure) {

	    for (i = 0; i < feaLoad->numElementID; i++) {
	        fprintf(fp,"%-8s", "PLOAD2");

            tempString = convert_integerToString(feaLoad->loadID, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString(feaLoad->pressureForce, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_integerToString(feaLoad->elementIDSet[i], fieldWidth, 1);
            fprintf(fp, "%s%s\n",delimiter, tempString);
            EG_free(tempString);
	    }
	}

	// Pressure load at element Nodes
    if (feaLoad->loadType ==  PressureDistribute) {

        for (i = 0; i < feaLoad->numElementID; i++) {
            fprintf(fp,"%-8s", "PLOAD4");

            tempString = convert_integerToString(feaLoad->loadID, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_integerToString(feaLoad->elementIDSet[i], fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            for (j = 0; j < 4; j++) {
                tempString = convert_doubleToString(feaLoad->pressureDistributeForce[j], fieldWidth, 1);
                fprintf(fp, "%s%s",delimiter, tempString);
                EG_free(tempString);
            }

            fprintf(fp,"\n");
        }
    }

    // Pressure load at element Nodes - different distribution per element
    if (feaLoad->loadType ==  PressureExternal) {

    	for (i = 0; i < feaLoad->numElementID; i++) {
    		fprintf(fp,"%-8s", "PLOAD4");

    		tempString = convert_integerToString(feaLoad->loadID, fieldWidth, 1);
    		fprintf(fp, "%s%s",delimiter, tempString);
    		EG_free(tempString);

    		tempString = convert_integerToString(feaLoad->elementIDSet[i], fieldWidth, 1);
    		fprintf(fp, "%s%s",delimiter, tempString);
    		EG_free(tempString);

    		for (j = 0; j < 4; j++) {
    			tempString = convert_doubleToString(feaLoad->pressureMultiDistributeForce[4*i + j], fieldWidth, 1);
    			fprintf(fp, "%s%s",delimiter, tempString);
    			EG_free(tempString);
    		}

    		fprintf(fp,"\n");
    	}
    }

	// Rotational velocity
	if (feaLoad->loadType ==  Rotational) {

	    for (i = 0; i < feaLoad->numGridID; i++) {
            fprintf(fp,"%-8s", "RFORCE");

            tempString = convert_integerToString(feaLoad->loadID, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_integerToString(feaLoad->gridIDSet[i], fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_integerToString(feaLoad->coordSystemID, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString( feaLoad->angularVelScaleFactor, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString( feaLoad->directionVector[0], fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString( feaLoad->directionVector[1], fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString( feaLoad->directionVector[2], fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            if (feaFileFormat->fileType == FreeField) {
            	fprintf(fp,", ,+RF\n");
            	fprintf(fp,"+RF     ");
            } else {
            	fprintf(fp," %7s%-7s\n", "", "+RF");
            	fprintf(fp,"%-8s", "+RF");
            }

            tempString = convert_doubleToString( feaLoad->angularAccScaleFactor, fieldWidth, 1);
            fprintf(fp, "%s%s\n",delimiter, tempString);
            EG_free(tempString);
	    }
	}
	// Thermal load at a grid point
    if (feaLoad->loadType ==  Thermal) {

        for (i = 0; i < feaLoad->numGridID; i++) {
            fprintf(fp,"%-8s", "TEMP");

            tempString = convert_integerToString(feaLoad->loadID, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_integerToString(feaLoad->gridIDSet[i], fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            //tempString = convert_doubleToString(feaLoad->temperature, 7, 1);
            //fprintf(fp, "%s%s\n",delimiter, tempString);
            //EG_free(tempString);
        }
    }

	return CAPS_SUCCESS;
}

// Write a Nastran Analysis card from a feaAnalysis structure
int nastran_writeAnalysisCard(FILE *fp, feaAnalysisStruct *feaAnalysis, feaFileFormatStruct *feaFileFormat) {

	int fieldWidth;

    char *tempString, *delimiter;

    if (feaFileFormat->fileType == FreeField) {
        delimiter = ",";
        fieldWidth = 8;
    } else {
        delimiter = " ";
        fieldWidth  = 7;
    }

    // Eigenvalue
    if (feaAnalysis->analysisType == Modal) {

        if (strcasecmp(feaAnalysis->extractionMethod, "Lanczos") == 0) {

			fprintf(fp,"%-8s", "EIGRL");

			tempString = convert_integerToString(feaAnalysis->analysisID  , fieldWidth, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			tempString = convert_doubleToString(feaAnalysis->frequencyRange[0], fieldWidth, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			tempString = convert_doubleToString(feaAnalysis->frequencyRange[1], fieldWidth, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			tempString = convert_integerToString(feaAnalysis->numDesiredEigenvalue, fieldWidth, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			// fprintf(fp, "%s%7s%s%7s%s%-7s", delimiter, "", delimiter, "", delimiter, "");

			if (feaAnalysis->eigenNormaliztion != NULL) {

				fprintf(fp, "%s%7s%s%7s%s%-7s", delimiter, "", delimiter, "", delimiter, "");

				fprintf(fp, "%s%7s\n", delimiter, feaAnalysis->eigenNormaliztion);
				//fprintf(fp, "%s%7s", delimiter, feaAnalysis->eigenNormaliztion);
			} else {
				fprintf(fp,"\n");
				//fprintf(fp,"%s%7s", delimiter, ""); // Print blank space
			}

			/*
			fprintf(fp, "%s%-7s\n", delimiter, "+E1");

			fprintf(fp, "%-8s", "+E1");

			tempString = convert_integerToString(feaAnalysis->lanczosMode, 7, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			if (feaAnalysis->lanczosType != NULL) {
				fprintf(fp, "%s%7s\n", delimiter, feaAnalysis->lanczosType);
			} else {
				fprintf(fp,"\n");
				//fprintf(fp, "%s%7s\n", delimiter, ""); // Print blank space
			}
			*/

        } else {

			fprintf(fp,"%-8s", "EIGR");

			tempString = convert_integerToString(feaAnalysis->analysisID, fieldWidth, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			if (feaAnalysis->extractionMethod != NULL) {
				fprintf(fp, "%s%7s", delimiter, feaAnalysis->extractionMethod);
			} else {
				fprintf(fp, "%s%7s", delimiter, ""); // Print blank space
			}

			tempString = convert_doubleToString(feaAnalysis->frequencyRange[0], fieldWidth, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			tempString = convert_doubleToString(feaAnalysis->frequencyRange[1], fieldWidth, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			tempString = convert_integerToString(feaAnalysis->numEstEigenvalue, fieldWidth, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			tempString = convert_integerToString(feaAnalysis->numDesiredEigenvalue, fieldWidth, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			if (feaFileFormat->fileType == FreeField) {
				fprintf(fp, "%s%7s%s%7s%s%-7s\n", delimiter, "", delimiter, "", delimiter, "+E1");
			} else {
				fprintf(fp, "%s%7s%s%7s%-7s\n", delimiter, "", delimiter, "", "+E1");
			}

			fprintf(fp, "%-8s", "+E1");

			if (feaAnalysis->eigenNormaliztion != NULL) {
				fprintf(fp, "%s%7s", delimiter, feaAnalysis->eigenNormaliztion);
			} else {
				fprintf(fp,"%s%7s", delimiter,  ""); // Print blank space
			}

			tempString = convert_integerToString(feaAnalysis->gridNormaliztion, fieldWidth, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			tempString = convert_integerToString(feaAnalysis->componentNormaliztion, fieldWidth, 1);
			fprintf(fp, "%s%s\n", delimiter, tempString);
			EG_free(tempString);

        }
    }

    /*
    if (feaAnalysis->analysisType == StaticOpt) {

        fprintf(fp,"%-8s", "DRESP1");

        tempString = convert_integerToString(feaAnalysis->analysisID, fieldWidth, 1);
        fprintf(fp, "%s%s", delimiter, tempString);
        EG_free(tempString);

        if (feaAnalysis->objectiveResponseType != NULL) {

            fprintf(fp, "%s%7s", delimiter, &feaAnalysis->objectiveResponseType[0]);
        } else {
            fprintf(fp, "%s%7s", delimiter, ""); // Print blank space
        }

        if (feaAnalysis->objectiveResponseType != NULL) {
            fprintf(fp, "%s%7s", delimiter, feaAnalysis->objectiveResponseType);
        } else {
            fprintf(fp, "%s%7s", delimiter,  ""); // Print blank space
        }

        fprintf(fp, "\n");

    }
    */

    return CAPS_SUCCESS;
}

// Write a combined Nastran design constraint card from a set of constraint IDs.
//  The combined design constraint ID is set through the constraint ID variable.
int nastran_writeDesignConstraintADDCard(FILE *fp, int constraintID, int numSetID, int designConstraintSetID[], feaFileFormatStruct *feaFileFormat) {

    int i, loadIndex; // Indexing

    int sidIndex= 0, lineCount = 0; // Counters

    int status; // Return status;

    double overallScale = 1.0;

    char *tempString, *delimiter;

    if (feaFileFormat->fileType == FreeField) {
        delimiter = ",";
    } else {
        delimiter = " ";
    }

    if (numSetID != 0) {
        fprintf(fp,"%-8s%s%7d", "DCONADD", delimiter, constraintID);
    }

    lineCount = 1;
    for (i = 0; i < numSetID; i++) {
        sidIndex += 2;
        if (sidIndex % (8*lineCount) == 0) {

            if (feaFileFormat->fileType == FreeField) {
                fprintf(fp, ",+L%-5d\n", lineCount-1); // Start of continuation
                fprintf(fp, "+L%-5d,"  , lineCount-1);  // End of continuation
            } else {
                fprintf(fp, "+L%-6d\n", lineCount-1);  // Start of continuation
                fprintf(fp, "+L%-6d"  , lineCount-1);  // End of continuation
            }

            lineCount += 1;
        }

        fprintf(fp,"%s%7d",delimiter, designConstraintSetID[i]);
    }

    if (numSetID != 0) fprintf(fp,"\n\n");

    return CAPS_SUCCESS;
}

// Write design constraint/optimization information from a feaDesignConstraint structure
int nastran_writeDesignConstraintCard(FILE *fp, feaDesignConstraintStruct *feaDesignConstraint, feaFileFormatStruct *feaFileFormat) {

    int  i, j; // Index

    int fieldWidth;

    int elementStressLocation[4];

    char *tempString=NULL;
    char *delimiter;

    if (feaFileFormat->fileType == FreeField) {
        delimiter = ",";
        fieldWidth = 8;
    } else {
        delimiter = " ";
        fieldWidth = 7;
    }

    for (i = 0; i < feaDesignConstraint->numPropertyID; i++) {

        if (feaDesignConstraint->propertySetType[i] == Rod) {
            // Nothing set yet
        } else if (feaDesignConstraint->propertySetType[i] == Bar) {
            // Nothing set yet
        } else if (feaDesignConstraint->propertySetType[i] == Shell) {

            elementStressLocation[0] = 7;
            elementStressLocation[1] = 9;
            elementStressLocation[2] = 16;
            elementStressLocation[3] = 17;

            for (j = 0; j < 4; j++) {

                // DCONSTR, DCID, RID, LALLOW, UALLOW
                fprintf(fp,"%-8s", "DCONSTR");

                tempString = convert_integerToString(feaDesignConstraint->designConstraintID, fieldWidth, 1);
                fprintf(fp,"%s%s",delimiter, tempString);
                EG_free(tempString);

                tempString = convert_integerToString(feaDesignConstraint->designConstraintID+10000+j*1000, fieldWidth, 1);
                fprintf(fp,"%s%s",delimiter, tempString);
                EG_free(tempString);

                tempString = convert_doubleToString(feaDesignConstraint->lowerBound, fieldWidth, 1);
                fprintf(fp,"%s%s",delimiter, tempString);
                EG_free(tempString);

                tempString = convert_doubleToString(feaDesignConstraint->upperBound, fieldWidth, 1);
                fprintf(fp,"%s%s\n",delimiter, tempString);
                EG_free(tempString);


                // DRESP1, ID, LABEL, RTYPE, PTYPE, REGION, ATTA, ATTB, ATT1
                //         ATT2, ...

                fprintf(fp,"%-8s", "DRESP1");

                tempString = convert_integerToString(feaDesignConstraint->designConstraintID+10000+j*1000, fieldWidth, 1);
                fprintf(fp,"%s%s",delimiter, tempString);
                EG_free(tempString);

                tempString = convert_integerToString(feaDesignConstraint->designConstraintID+10000+j*1000, fieldWidth-2, 0);
                //fprintf(fp, "%s%s%.*s", delimiter,  tempString,1,feaDesignConstraint->name);
                fprintf(fp, "%sR%s", delimiter,  tempString);
                EG_free(tempString);

                fprintf(fp,"%s%7s",delimiter,feaDesignConstraint->responseType);

                fprintf(fp,"%s%7s",delimiter,"PSHELL");

                fprintf(fp,"%s%7s", delimiter,"");

                //printf("Printing DRESP1 isn't very robust yet.....\n");
                tempString = convert_integerToString(elementStressLocation[j], fieldWidth, 1);
                fprintf(fp, "%s%s",delimiter, tempString);
                EG_free(tempString);

                fprintf(fp,"%s%7s", delimiter,"");

                tempString = convert_integerToString(feaDesignConstraint->propertySetID[i], fieldWidth, 1);
                fprintf(fp, "%s%s\n",delimiter, tempString);
                EG_free(tempString);
            }

        } else if (feaDesignConstraint->propertySetType[i] == Composite) {
            // Nothing set yet
        } else if (feaDesignConstraint->propertySetType[i] == Solid) {
            // Nothing set yet
        }
    }

    return CAPS_SUCCESS;
}

// Write design variable/optimization information from a feaDesignVariable structure
int nastran_writeDesignVariableCard(FILE *fp, feaDesignVariableStruct *feaDesignVariable, feaFileFormatStruct *feaFileFormat) {

    int  i, fieldWidth;

    int status; // Status function returnd

    char *tempString=NULL;
    char *delimiter;

    if (feaFileFormat->fileType == FreeField) {
        delimiter = ",";
        fieldWidth = 8;
    } else {
        delimiter = " ";
        fieldWidth = 7;
        fprintf(fp, "$---1---|---2---|---3---|---4---|---5---|---6---|---7---|---8---|---9---|---10--|\n");
    }

    // DESVAR, ID, LABEL, XINIT, XLB, XUB, DELXV
    fprintf(fp,"%-8s", "DESVAR");

    tempString = convert_integerToString(feaDesignVariable->designVariableID, fieldWidth, 1);
    fprintf(fp, "%s%s",delimiter, tempString);
    EG_free(tempString);

    fprintf(fp, "%s%7s", delimiter,feaDesignVariable->name);

    tempString = convert_doubleToString(feaDesignVariable->initialValue, fieldWidth, 1);
    fprintf(fp, "%s%s",delimiter, tempString);
    EG_free(tempString);

    /*
    tempString = convert_doubleToString(feaDesignVariable->lowerBound, fieldWidth, 1);
    fprintf(fp, "%s%s",delimiter, tempString);
    EG_free(tempString);

    tempString = convert_doubleToString(feaDesignVariable->upperBound, fieldWidth, 1);
    fprintf(fp, "%s%s",delimiter, tempString);
    EG_free(tempString);
    */

    if (feaDesignVariable->numDiscreteValue == 0) {

    	tempString = convert_doubleToString(feaDesignVariable->lowerBound, fieldWidth, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString(feaDesignVariable->upperBound, fieldWidth, 1);
		fprintf(fp, "%s%s",delimiter, tempString);
		EG_free(tempString);

    	tempString = convert_doubleToString(feaDesignVariable->maxDelta, fieldWidth, 1);
    	fprintf(fp, "%s%s\n",delimiter, tempString);
    	EG_free(tempString);

    } else {

    	if (feaFileFormat->fileType == FreeField) {
			fprintf(fp,", , ");
		} else {
			fprintf(fp," %7s %7s", "", "");
		}

    	// ID to DDVAL entry
    	tempString = convert_integerToString(feaDesignVariable->designVariableID, fieldWidth, 1);
    	fprintf(fp, "%s%s\n",delimiter, tempString);
    	EG_free(tempString);

    	// Write DDVAL card
    	status = nastran_writeDDVALCard(fp,
    									feaDesignVariable->designVariableID,
										feaDesignVariable->numDiscreteValue,
										feaDesignVariable->discreteValue,
										feaFileFormat);
		if (status != CAPS_SUCCESS) return status;

    }


    if (feaDesignVariable->designVariableType == MaterialDesignVar) {

    	for (i = 0; i < feaDesignVariable->numMaterialID; i++) {
    		fprintf(fp,"%-8s", "DVMREL1");

    		tempString = convert_integerToString(feaDesignVariable->designVariableID, fieldWidth, 1);
    		fprintf(fp, "%s%s", delimiter, tempString);
    		EG_free(tempString);

    		// UnknownMaterial, Isotropic, Anisothotropic, Orthotropic, Anisotropic
    		if (feaDesignVariable->materialSetType[i] == Isotropic) {
    			fprintf(fp,"%s%7s",delimiter, "MAT1");
    		} else if (feaDesignVariable->materialSetType[i] == Anisothotropic) {
    			fprintf(fp,"%s%7s",delimiter, "MAT2");
    		} else if (feaDesignVariable->materialSetType[i] == Orthotropic) {
    			fprintf(fp,"%s%7s",delimiter, "MAT8");
    		} else if (feaDesignVariable->materialSetType[i] == Anisotropic) {
    			fprintf(fp,"%s%7s",delimiter, "MAT9");
    		}

    		tempString = convert_integerToString(feaDesignVariable->materialSetID[i], fieldWidth, 1);
    		fprintf(fp, "%s%s",delimiter, tempString);
    		EG_free(tempString);

    		fprintf(fp, "%s%7s", delimiter, feaDesignVariable->fieldName);

    		if (feaFileFormat->fileType == FreeField) {
    			fprintf(fp,", , , , ");
    		} else {
    			fprintf(fp," %7s %7s %7s %7s", "", "", "", "");
    		}

    		tempString = convert_integerToString(feaDesignVariable->designVariableID, 5, 0);
    		if (feaFileFormat->fileType == FreeField) {
    			fprintf(fp, "%s+DV%s\n+DV%s",delimiter, tempString, tempString);
    		} else {
    			fprintf(fp, "+DV%s\n+DV%s", tempString, tempString);
    		}
    		EG_free(tempString);

    		tempString = convert_integerToString(feaDesignVariable->designVariableID, fieldWidth, 1);
    		fprintf(fp, "%s%s",delimiter, tempString);
    		EG_free(tempString);

            tempString = convert_doubleToString(1.0, fieldWidth, 1);
            fprintf(fp, "%s%s\n",delimiter, tempString);
            EG_free(tempString);

    	}
    }

    if (feaDesignVariable->designVariableType == PropertyDesignVar) {

        // DVPREL1, ID, TYPE, PID, FID, PMIN, PMAX, CO
        //          DVID1, COEF1, DVID2, COEF2 ...
        for (i = 0; i < feaDesignVariable->numPropertyID; i++) {

            fprintf(fp,"%-8s", "DVPREL1");

            tempString = convert_integerToString(feaDesignVariable->designVariableID, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            // UnknownProperty, Rod, Bar, Shear, Shell, Composite, Solid
            if (feaDesignVariable->propertySetType[i] == Rod) {
            	fprintf(fp,"%s%7s",delimiter, "PROD");
            } else if (feaDesignVariable->propertySetType[i] == Bar) {
            	fprintf(fp,"%s%7s",delimiter, "PBAR");
            } else if (feaDesignVariable->propertySetType[i] == Shell) {
            	fprintf(fp,"%s%7s",delimiter, "PSHELL");
            } else if (feaDesignVariable->propertySetType[i] == Composite) {
            	fprintf(fp,"%s%7s",delimiter, "PCOMP");
            } else if (feaDesignVariable->propertySetType[i] == Solid) {
            	fprintf(fp,"%s%7s",delimiter, "PSOLID");
            }

            tempString = convert_integerToString(feaDesignVariable->propertySetID[i], fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            fprintf(fp,"%s%7s",delimiter, feaDesignVariable->fieldName);

            if (feaFileFormat->fileType == FreeField) {
                fprintf(fp,", , , , ");
            } else {
                fprintf(fp," %7s %7s %7s %7s", "", "", "", "");
            }

            tempString = convert_integerToString(feaDesignVariable->designVariableID, 5, 0);
            if (feaFileFormat->fileType == FreeField) {
            	fprintf(fp, "%s+DV%s\n+DV%s",delimiter, tempString, tempString);
            } else {
            	fprintf(fp, "+DV%s\n+DV%s", tempString, tempString);
            }
            EG_free(tempString);

            tempString = convert_integerToString(feaDesignVariable->designVariableID, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString(1.0, fieldWidth, 1);
            fprintf(fp, "%s%s\n",delimiter, tempString);
            EG_free(tempString);

        }
    }

    return CAPS_SUCCESS;
}

// Write a Nastran DDVAL card from a set of ddvalSet. The id is set through the ddvalID variable.
int nastran_writeDDVALCard(FILE *fp, int ddvalID, int numDDVALSet, double ddvalSet[], feaFileFormatStruct *feaFileFormat) {

    int i; // Indexing
    int fieldWidth;
    int sidIndex= 0, lineCount = 0; // Counters
    char *tempString = NULL, *delimiter;

    if (feaFileFormat->fileType == FreeField) {
		delimiter = ",";
		fieldWidth = 8;
	} else {
		delimiter = " ";
		fieldWidth = 7;
	}

    if (numDDVALSet != 0) {
        fprintf(fp,"%-8s", "DDVAL");

        tempString = convert_integerToString(ddvalID, fieldWidth, 1);
        fprintf(fp, "%s%s", delimiter, tempString);
        EG_free(tempString);
    }

    lineCount = 1;
    for (i = 0; i < numDDVALSet; i++) {
    	sidIndex += 2;
    	if (sidIndex % (8*lineCount) == 0) {

    		if (feaFileFormat->fileType == FreeField) {
				fprintf(fp, ",+L%-5d\n", lineCount-1); // Start of continuation
				fprintf(fp, "+L%-5d,"  , lineCount-1);  // End of continuation
			} else {
				fprintf(fp, "+L%-6d\n", lineCount-1);  // Start of continuation
				fprintf(fp, "+L%-6d"  , lineCount-1);  // End of continuation
			}

    		lineCount += 1;
    	}

    	tempString = convert_doubleToString(ddvalSet[i], fieldWidth, 1);
    	fprintf(fp, "%s%s", delimiter, tempString);
    	EG_free(tempString);
    }

    if (numDDVALSet != 0) fprintf(fp,"\n");

    return CAPS_SUCCESS;
}

// Read data from a Nastran F06 file and load it into a dataMatrix[numEigenVector][numGridPoint*8]
// where variables are Grid Id, Coord Id, T1, T2, T3, R1, R2, R3
int nastran_readF06EigenVector(FILE *fp, int *numEigenVector, int *numGridPoint, double ***dataMatrix) {

    int status; // Function return

    int i, j, eigenValue; // Indexing

    size_t linecap = 0;

    char *line = NULL; // Temporary line holder

    char *numEigenLine = "                                     NUMBER OF ROOTS FOUND ................";
    char *outputEigenLine = "      EIGENVALUE =";
    char *beginEigenLine=NULL;
    char *endEigenLine = "1";
    char tempString[1];

    int numVariable = 8; // Grid Id, Coord Id, T1, T2, T3, R1, R2, R3
    int intLength;

    printf("Reading Nastran FO6 file - extracting Eigen-Vectors!\n");

    *numEigenVector = 0;
    *numGridPoint = 0;

    // Loop through file line by line until we have determined how many Eigen-Values and grid points we have
    while (*numGridPoint == 0) {

        // Get line from file
        status = getline(&line, &linecap, fp);
        if (status < 0) break;

        // See how many Eigen-Values we have
        if (strncmp(numEigenLine, line,strlen(numEigenLine)) == 0) {
            sscanf(&line[strlen(numEigenLine)], "%d", numEigenVector);

            // Build begin Eigen-Value string
            beginEigenLine = (char *) EG_alloc((strlen(outputEigenLine)+1)*sizeof(char));
            if (beginEigenLine == NULL) return EGADS_MALLOC;

            sprintf(beginEigenLine,"%s",outputEigenLine);
            beginEigenLine[strlen(outputEigenLine)+1] = '\0';
        }

        // Once we know how many Eigen-Values we have, we need to determine how many grid points exist
        if (*numEigenVector > 0) {

            // Look for start of Eigen-Vector 1
            if (strncmp(beginEigenLine, line, strlen(beginEigenLine)) == 0) {

                // Fast forward 3 lines
                for (i = 0; i < 3; i++) {
                    status = getline(&line, &linecap, fp);
                    if (status < 0) break;
                }

                // Loop through lines counting the number of grid points
                while (getline(&line, &linecap, fp)) {
                    *numGridPoint +=1;
                    if (strncmp(endEigenLine, line,strlen(endEigenLine)) == 0) {
                        *numGridPoint -= 1; // Get rid of last line
                        break;
                    }
                }
            }
        }
    }

    printf("Number of Eigen-Vectors = %d\n", *numEigenVector);
    printf("Number of Grid Points = %d for each Eigen-Vector\n", *numGridPoint);

    // Free begin Eigen-Value string
    if (beginEigenLine != NULL) EG_free(beginEigenLine);
    beginEigenLine = NULL;

    if (*numGridPoint == 0 || *numEigenVector == 0) {
        printf("Either data points  = 0 and/or how many Eigen-Values = 0\n");
        return CAPS_NOTFOUND;
    }
    // Rewind the file
    rewind(fp);

    // Allocate dataMatrix array
    if (*dataMatrix != NULL) EG_free(*dataMatrix);

    *dataMatrix = (double **) EG_alloc(*numEigenVector *sizeof(double *));
    if (*dataMatrix == NULL) return EGADS_MALLOC; // If allocation failed ....

    for (i = 0; i < *numEigenVector; i++) {

        (*dataMatrix)[i] = (double *) EG_alloc(*numGridPoint*numVariable*sizeof(double));

        if ((*dataMatrix)[i] == NULL) { // If allocation failed ....
            for (j = 0; j < *numEigenVector; j++) {

                if ((*dataMatrix)[j] != NULL ) EG_free((*dataMatrix)[j]);
            }

            if ((*dataMatrix) != NULL) EG_free((*dataMatrix));

            return EGADS_MALLOC;
        }
    }


    if      (*numEigenVector >= 1000) intLength = 5;
    else if (*numEigenVector >= 100) intLength = 4;
    else if (*numEigenVector >= 10) intLength = 3;
    else intLength = 2;

    intLength = 1;

    // Loop through the file again and pull out data
    eigenValue = 1;
    while (eigenValue <= *numEigenVector) {

        if (beginEigenLine == NULL) {
            // Build begin Eigen-Value string
            beginEigenLine = (char *) EG_alloc((strlen(outputEigenLine)+intLength)*sizeof(char));
            if (beginEigenLine == NULL) return EGADS_MALLOC;

            sprintf(beginEigenLine,"%s",outputEigenLine);
            beginEigenLine[strlen(outputEigenLine)+intLength] = '\0';
        }

        // Get line from file
        status = getline(&line, &linecap, fp);
        if (status < 0) break;

        // Look for start of Eigen-Vector
        if (strncmp(beginEigenLine, line, strlen(beginEigenLine)) == 0) {

            printf("Loading Eigen-Vector = %d\n", eigenValue);

            // Fast forward 3 lines
            for (i = 0; i < 3; i++) {
                status = getline(&line, &linecap, fp);
                if (status < 0) break;
            }

            // Loop through the file and fill up the data matrix
            for (j = 0; j < (*numGridPoint)*numVariable; j++) {
                // eigenValue is 1 bias
                fscanf(fp, "%lf", &(*dataMatrix)[eigenValue-1][j]);

                if (j == 0 || j % numVariable+1 == 0) {
                	fscanf(fp, "%lf", &(*dataMatrix)[eigenValue-1][j]);
                	fscanf(fp, "%s", tempString);
                	j = j + 1;
					(*dataMatrix)[eigenValue-1][j] = 0.0;
                } else fscanf(fp, "%lf", &(*dataMatrix)[i][j]);

            }

            if (beginEigenLine != NULL) EG_free(beginEigenLine);
            beginEigenLine = NULL;

            eigenValue += 1;
        }
    }

    if (beginEigenLine != NULL) EG_free(beginEigenLine);
    beginEigenLine = NULL;

    return CAPS_SUCCESS;
}

// Read data from a Nastran F06 file and load it into a dataMatrix[numGridPoint][8]
// where variables are Grid Id, Coord Id, T1, T2, T3, R1, R2, R3
int nastran_readF06Displacement(FILE *fp, int subcaseId, int *numGridPoint, double ***dataMatrix) {

    int status; // Function return

    int i, j; // Indexing

    size_t linecap = 0;

    char *line = NULL; // Temporary line holder

    char *outputSubcaseLine = "0                                                                                                            SUBCASE ";
    char *displacementLine ="                                             D I S P L A C E M E N T   V E C T O R";
    char *beginSubcaseLine=NULL;
    char *endSubcaseLine = "1";
    char tempString[1];

    int numVariable = 8; // Grid Id, coord Id T1, T2, T3, R1, R2, R3
    int intLength;
    int lineFastForward = 0;

    printf("Reading Nastran FO6 file - extracting Displacements!\n");

    *numGridPoint = 0;

    if      (subcaseId >= 1000) intLength = 4;
    else if (subcaseId >= 100) intLength = 3;
    else if (subcaseId >= 10) intLength = 2;
    else intLength = 1;

    if (subcaseId > 0) {
    	beginSubcaseLine = (char *) EG_alloc((strlen(outputSubcaseLine)+intLength+1)*sizeof(char));
    	if (beginSubcaseLine == NULL) return EGADS_MALLOC;

    	sprintf(beginSubcaseLine,"%s%d",outputSubcaseLine, subcaseId);

    	lineFastForward = 4;

    } else {

    	intLength = 0;
    	beginSubcaseLine = (char *) EG_alloc((strlen(displacementLine)+1)*sizeof(char));
    	if (beginSubcaseLine == NULL) return EGADS_MALLOC;
    	sprintf(beginSubcaseLine,"%s",displacementLine);

    	lineFastForward = 2;
    }

    beginSubcaseLine[strlen(outputSubcaseLine)+intLength] = '\0';

    // Loop through file line by line until we have determined how many grid points we have
    while (*numGridPoint == 0) {

        // Get line from file
        status = getline(&line, &linecap, fp);
        if (status < 0) break;

        // Look for start of subcaseId
        if (strncmp(beginSubcaseLine, line, strlen(beginSubcaseLine)) == 0) {

            // Fast forward lines
            for (i = 0; i < lineFastForward; i++) {
                status = getline(&line, &linecap, fp);
                if (status < 0) break;
            }

            // Loop through lines counting the number of grid points
            while (getline(&line, &linecap, fp)) {
                *numGridPoint +=1;
                if (strncmp(endSubcaseLine, line,strlen(endSubcaseLine)) == 0) {
                    *numGridPoint -= 1; // Get rid of last line
                    break;
                }
            }
        }
    }

    printf("Number of Grid Points = %d\n", *numGridPoint);

    if (*numGridPoint == 0) {
        printf("Either data points  = 0 and/or subcase wasn't found\n");

        if (beginSubcaseLine != NULL) EG_free(beginSubcaseLine);
        return CAPS_NOTFOUND;
    }

    // Rewind the file
    rewind(fp);

    // Allocate dataMatrix array
    if (*dataMatrix != NULL) EG_free(*dataMatrix);

    *dataMatrix = (double **) EG_alloc(*numGridPoint *sizeof(double *));
    if (*dataMatrix == NULL) return EGADS_MALLOC; // If allocation failed ....

    for (i = 0; i < *numGridPoint; i++) {

        (*dataMatrix)[i] = (double *) EG_alloc(numVariable*sizeof(double));

        if ((*dataMatrix)[i] == NULL) { // If allocation failed ....
            for (j = 0; j < *numGridPoint; j++) {

                if ((*dataMatrix)[j] != NULL ) EG_free((*dataMatrix)[j]);
            }

            if ((*dataMatrix) != NULL) EG_free((*dataMatrix));

            return EGADS_MALLOC;
        }
    }

    // Loop through the file again and pull out data
    while (getline(&line, &linecap, fp)) {

        // Look for start of Displacement
        if (strncmp(beginSubcaseLine, line, strlen(beginSubcaseLine)) == 0) {

            printf("Loading displacements for Subcase = %d\n", subcaseId);

            // Fast forward lines
            for (i = 0; i < lineFastForward; i++) {
                status = getline(&line, &linecap, fp);
                if (status < 0) break;
            }


            // Loop through the file and fill up the data matrix
            for (i = 0; i < (*numGridPoint); i++) {
                for (j = 0; j < numVariable; j++) {

                	if (j == 0 || j % numVariable+1 == 0) {
                		fscanf(fp, "%lf", &(*dataMatrix)[i][j]);
                		fscanf(fp, "%s", tempString);
                		j = j + 1;
                		(*dataMatrix)[i][j] = 0.0;
                	} else fscanf(fp, "%lf", &(*dataMatrix)[i][j]);
                }
            }

            break;
        }
    }

    if (beginSubcaseLine != NULL) EG_free(beginSubcaseLine);

    return CAPS_SUCCESS;
}
