// Vortex lattice analysis related utility functions - Written by Dr. Ryan Durscher AFRL/RQVC

#include <string.h>
#include <math.h>
#include "capsTypes.h"  // Bring in CAPS types
#include "vlmTypes.h"   // Bring in Vortex Lattice Method structures
#include "vlmUtils.h"
#include "miscUtils.h"  // Bring in miscellaneous utilities

#ifdef WIN32
#define strcasecmp  stricmp
#endif

// Fill vlmSurface in a vlmSurfaceStruct format with vortex lattice information
// from an incoming surfaceTuple
int get_vlmSurface(int numTuple,
                   capsTuple surfaceTuple[],
                   mapAttrToIndexStruct *attrMap,
                   int *numVLMSurface,
                   vlmSurfaceStruct *vlmSurface[]) {

    /*! \page vlmSurface Vortex Lattice Surface
     * Structure for the Vortex Lattice Surface tuple  = ("Name of Surface", "Value").
     * "Name of surface defines the name of the surface in which the data should be applied.
     *  The "Value" can either be a JSON String dictionary (see Section \ref jsonStringVLMSurface)
     *  or a single string keyword string (see Section \ref keyStringVLMSurface).
     */

    int status; //Function return

    int i, groupIndex, attrIndex; // Indexing

    char *keyValue = NULL; // Key values from tuple searches
    char *keyWord = NULL; // Key words to find in the tuples

    int numGroupName = 0;
    char **groupName = NULL;

    int stringLength = 0;

    // Clear out vlmSurface is it has already been allocated
    if (*vlmSurface != NULL) {
        for (i = 0; i < (*numVLMSurface); i++) {
            status = destroy_vlmSurface(&(*vlmSurface)[i]);
            if (status != CAPS_SUCCESS) printf("destroy_vlmSurface status = %d\n", status);
        }
    }

    printf("Getting vortex lattice surface data\n");

    if (numTuple <= 0){
        printf("\tNumber of VLM Surface tuples is %d\n", numTuple);
        return CAPS_NOTFOUND;
    }

    *numVLMSurface = numTuple;
    *vlmSurface = (vlmSurfaceStruct *) EG_alloc((*numVLMSurface) * sizeof(vlmSurfaceStruct));
    if (*vlmSurface == NULL) {
        *numVLMSurface = 0;
        return EGADS_MALLOC;
    }

    // Initiate vlmSurfaces
    for (i = 0; i < (*numVLMSurface); i++) {
        status = initiate_vlmSurface(&(*vlmSurface)[i]);
        if (status != CAPS_SUCCESS) goto cleanup;
    }

    for (i = 0; i < (*numVLMSurface); i++) {

        printf("\tVLM surface name - %s\n", surfaceTuple[i].name);

        // Copy surface name
        stringLength = strlen(surfaceTuple[i].name);
        (*vlmSurface)[i].name = (char *) EG_alloc((stringLength + 1)*sizeof(char));
        if ((*vlmSurface)[i].name == NULL) return EGADS_MALLOC;

        memcpy((*vlmSurface)[i].name,
                surfaceTuple[i].name,
                stringLength);
        (*vlmSurface)[i].name[stringLength] = '\0';

        // Do we have a json string?
        if (strncmp(surfaceTuple[i].value, "{", 1) == 0) {

            //printf("JSON String - %s\n",surfaceTuple[i].value);

            /*! \page vlmSurface
             * \section jsonStringVLMSurface JSON String Dictionary
             *
             * If "Value" is a JSON string dictionary the following keywords ( = default values) may be used:
             *
             * \if (AVL)
             * <ul>
             *  <li> <B>groupName = "(no default)"</B> </li> <br>
             *  Single or list of <em>capsGroup</em> names used to define the surface (e.g. "Name1" or ["Name1","Name2",...].
             *  If no groupName variable is provided an attempted will be made to use the tuple name instead;
             * </ul>
             * \else
             * <ul>
             *  <li> <B>groupName = "(no default)"</B> </li> <br>
             *  Single or list of <em>capsGroup</em> names used to define the surface (e.g. "Name1" or ["Name1","Name2",...].
             *  If no groupName variable is provided an attempted will be made to use the tuple name instead;
             * </ul>
             * \endif
             *
             */

            // Get surface variables
            keyWord = "groupName";
            status = search_jsonDictionary( surfaceTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

               status = string_toStringDynamicArray(keyValue, &numGroupName, &groupName);
               if (status != CAPS_SUCCESS) goto cleanup;

               if (keyValue != NULL) EG_free(keyValue);
               keyValue = NULL;

               // Determine many capsGroups go into making the surface
               for (groupIndex = 0; groupIndex < numGroupName; groupIndex++) {

                   status = get_mapAttrToIndexIndex(attrMap, (const char *) groupName[groupIndex], &attrIndex);

                   if (status == CAPS_NOTFOUND) {
                       printf("\tgroupName name %s not found in attribute map of capsGroups!!!!\n", groupName[groupIndex]);
                       continue;

                   } else if (status != CAPS_SUCCESS) goto cleanup;

                   (*vlmSurface)[i].numAttr += 1;
                   if ((*vlmSurface)[i].numAttr == 1) {
                       (*vlmSurface)[i].attrIndex = (int *) EG_alloc(1*sizeof(int));

                   } else{
                       (*vlmSurface)[i].attrIndex = (int *) EG_reall((*vlmSurface)[i].attrIndex,
                                                                     (*vlmSurface)[i].numAttr*sizeof(int));
                   }

                   if ((*vlmSurface)[i].attrIndex == NULL) {
                       status = EGADS_MALLOC;
                       goto cleanup;
                   }

                   (*vlmSurface)[i].attrIndex[(*vlmSurface)[i].numAttr-1] = attrIndex;
               }

               status = string_freeArray(numGroupName, &groupName);
               if (status != CAPS_SUCCESS) goto cleanup;
               groupName = NULL;
               numGroupName = 0;

            } else {
                printf("\tNo \"groupName\" variable provided or no matches found, going to use tuple name\n");
            }

            if ((*vlmSurface)[i].numAttr == 0) {

                status = get_mapAttrToIndexIndex(attrMap, (const char *) (*vlmSurface)[i].name, &attrIndex);
                if (status == CAPS_NOTFOUND) {
                    printf("\tTuple name %s not found in attribute map of capsGroups!!!!\n", (*vlmSurface)[i].name);
                    goto cleanup;
                }

                (*vlmSurface)[i].numAttr += 1;
                if ((*vlmSurface)[i].numAttr == 1) {
                    (*vlmSurface)[i].attrIndex = (int *) EG_alloc(1*sizeof(int));
                } else{
                    (*vlmSurface)[i].attrIndex = (int *) EG_reall((*vlmSurface)[i].attrIndex,
                                                                  (*vlmSurface)[i].numAttr*sizeof(int));
                }

                if ((*vlmSurface)[i].attrIndex == NULL) {
                    status = EGADS_MALLOC;
                    goto cleanup;
                }

                (*vlmSurface)[i].attrIndex[(*vlmSurface)[i].numAttr-1] = attrIndex;
            }

            if (keyValue != NULL) EG_free(keyValue);
            keyValue = NULL;

            /*! \page vlmSurface
             * \if (AVL)
             * <ul>
             * <li> <B>noKeyword = "(no default)"</B> </li> <br>
             *      "No" type. Options: NOWAKE, NOALBE, NOLOAD.
             * </ul>
             * \endif
             */

            keyWord = "noKeyword";
            status = search_jsonDictionary( surfaceTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                if      (strcasecmp(keyValue, "\"NOWAKE\"") == 0) (*vlmSurface)[i].nowake = (int) true;
                else if (strcasecmp(keyValue, "\"NOALBE\"") == 0) (*vlmSurface)[i].noalbe = (int) true;
                else if (strcasecmp(keyValue, "\"NOLOAD\"") == 0) (*vlmSurface)[i].noload = (int) true;
                else {

                    printf("\tUnrecognized \"%s\" specified (%s) for VLM Section tuple %s, current options are "
                            "\" NOWAKE, NOALBE, or  NOLOAD\"\n", keyWord, keyValue, surfaceTuple[i].name);
                }

                if (keyValue != NULL) EG_free(keyValue);
                keyValue = NULL;
            }

            /*! \page vlmSurface
             * <ul>
             * <li> <B>numChord = 0.0</B> </li> <br>
             *  The number of chordwise horseshoe vortices placed on the surface.
             * </ul>
             *
             */
            keyWord = "numChord";
            status = search_jsonDictionary( surfaceTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toInteger(keyValue, &(*vlmSurface)[i].Nchord);
                if (status != CAPS_SUCCESS) goto cleanup;

                if (keyValue != NULL) EG_free(keyValue);
                keyValue = NULL;
            }

            /*! \page vlmSurface
             * <ul>
             * <li> <B>spaceChord = 0.0</B> </li> <br>
             *  The chordwise vortex spacing parameter.
             * </ul>
              *
              */
             keyWord = "spaceChord";
             status = search_jsonDictionary( surfaceTuple[i].value, keyWord, &keyValue);
             if (status == CAPS_SUCCESS) {

                 status = string_toDouble(keyValue, &(*vlmSurface)[i].Cspace);
                 if (status != CAPS_SUCCESS) goto cleanup;

                 if (keyValue != NULL) EG_free(keyValue);
                 keyValue = NULL;
             }

             /*! \page vlmSurface
              * <ul>
              * <li> <B>numSpan = 0</B> </li> <br>
              *  The number of spanwise horseshoe vortices placed on the surface.
              * </ul>
              *
              */
             keyWord = "numSpan";
             status = search_jsonDictionary( surfaceTuple[i].value, keyWord, &keyValue);
             if (status == CAPS_SUCCESS) {

                 status = string_toInteger(keyValue, &(*vlmSurface)[i].Nspan);
                 if (status != CAPS_SUCCESS) goto cleanup;

                 if (keyValue != NULL) EG_free(keyValue);
                 keyValue = NULL;
              }

             /*! \page vlmSurface
              * <ul>
              * <li> <B>spaceSpan = 0.0</B> </li> <br>
              *  The spanwise vortex spacing parameter.
              * </ul>
              *
              */
             keyWord = "spaceSpan";
             status = search_jsonDictionary( surfaceTuple[i].value, keyWord, &keyValue);
             if (status == CAPS_SUCCESS) {

                 status = string_toDouble(keyValue, &(*vlmSurface)[i].Sspace);
                 if (status != CAPS_SUCCESS) goto cleanup;

                 if (keyValue != NULL) EG_free(keyValue);
                 keyValue = NULL;
             }

            /*! \page vlmSurface
              * <ul>
              * <li> <B>yMirror = False</B> </li> <br>
              *  Mirror surface about the y-direction.
              * </ul>
              *
              */
             keyWord = "yMirror";
             status = search_jsonDictionary( surfaceTuple[i].value, keyWord, &keyValue);
             if (status == CAPS_SUCCESS) {

                 status = string_toBoolean(keyValue, &(*vlmSurface)[i].iYdup);
                 if (status != CAPS_SUCCESS) goto cleanup;

                 if (keyValue != NULL) EG_free(keyValue);
                 keyValue = NULL;
             }
        } else {

            /*! \page vlmSurface
             * \section keyStringVLMSurface Single Value String
             *
             * If "Value" is a single string the following options maybe used:
             * - (NONE Currently)
             *
             */
            printf("\tNo current defaults for get_vlmSurface, tuple value must be a JSON string\n");
            status = CAPS_BADVALUE;
            goto cleanup;
        }
    }

    printf("\tDone getting vortex lattice surface data\n");

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("\tError in get_vlmSurface, status = %d", status);

        if (keyValue != NULL) EG_free(keyValue);

        if (numGroupName != 0 && groupName != NULL){
            (void) string_freeArray(numGroupName, &groupName);
        }
        return status;
}


// Fill vlmControl in a vlmControlStruct format with vortex lattice information
// from an incoming controlTuple
int get_vlmControl(int numTuple,
                   capsTuple controlTuple[],
                   int *numVLMControl,
				   vlmControlStruct *vlmControl[]) {

    /*! \page vlmControl Vortex Lattice Control Surface
     * Structure for the Vortex Lattice Control Surface tuple  = ("Name of Control Surface", "Value").
     * "Name of control surface defines the name of the control surface in which the data should be applied.
     *  The "Value" must be a JSON String dictionary (see Section \ref jsonStringVLMSection).
     */

    int status; //Function return

    int i; // Indexing

    char *keyValue = NULL; // Key values from tuple searches
    char *keyWord = NULL; // Key words to find in the tuples

    int stringLength = 0;

    // Clear out vlmSurface is it has already been allocated
    if (*vlmControl != NULL) {
        for (i = 0; i < (*numVLMControl); i++) {
            status = destroy_vlmControl(&(*vlmControl)[i]);
            if (status != CAPS_SUCCESS) printf("destroy_vlmControl status = %d\n", status);
        }
    }

    printf("Getting vortex lattice control surface data\n");

    if (numTuple <= 0){
        printf("\tNumber of VLM Surface tuples is %d\n", numTuple);
        return CAPS_NOTFOUND;
    }

    *numVLMControl = numTuple;
    *vlmControl = (vlmControlStruct *) EG_alloc((*numVLMControl) * sizeof(vlmControlStruct));
    if (*vlmControl == NULL) {
        *numVLMControl = 0;
        return EGADS_MALLOC;
    }

    // Initiate vlmSurfaces
    for (i = 0; i < (*numVLMControl); i++) {
        status = initiate_vlmControl(&(*vlmControl)[i]);
        if (status != CAPS_SUCCESS) goto cleanup;
    }

    for (i = 0; i < (*numVLMControl); i++) {

        printf("\tVLM control surface name - %s\n", controlTuple[i].name);

        // Copy surface name
        stringLength = strlen(controlTuple[i].name);
        (*vlmControl)[i].name = (char *) EG_alloc((stringLength + 1)*sizeof(char));
        if ((*vlmControl)[i].name == NULL) return EGADS_MALLOC;

        memcpy((*vlmControl)[i].name,
        		controlTuple[i].name,
                stringLength);
        (*vlmControl)[i].name[stringLength] = '\0';

        // Do we have a json string?
        if (strncmp(controlTuple[i].value, "{", 1) == 0) {

            //printf("JSON String - %s\n",surfaceTuple[i].value);

            /*! \page vlmControl
             * \section jsonStringVLMSection JSON String Dictionary
             *
             * If "Value" is a JSON string dictionary the following keywords ( = default values) may be used:
             *
             */

        	/*! \page vlmControl
             * \if (AVL)
             * <ul>
             * <li> <B>deflectionAngle = 0.0</B> </li> <br>
             *      Deflection angle of the control surface.
             * </ul>
             * \endif
             */
            keyWord = "deflectionAngle";
            status = search_jsonDictionary( controlTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

            	status = string_toDouble(keyValue, &(*vlmControl)[i].deflectionAngle);
            	if (status != CAPS_SUCCESS) goto cleanup;

            	if (keyValue != NULL) EG_free(keyValue);
            	keyValue = NULL;
            }

            /*
            ! \page vlmControl
			 * \if (AVL)
			 * <ul>
			 * <li> <B>percentChord = 0.0</B> </li> <br>
			 *      Percentage along the airfoil chord the control surface's hinge line resides.
			 * </ul>
			 * \endif

			keyWord = "percentChord";
			status = search_jsonDictionary( controlTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				status = string_toDouble(keyValue, &(*vlmControl)[i].percentChord);
				if (status != CAPS_SUCCESS) goto cleanup;

				if (keyValue != NULL) EG_free(keyValue);
				keyValue = NULL;
			}
             */

			/*! \page vlmControl
			 * \if (AVL)
			 * <ul>
			 * <li> <B>leOrTe = (no default) </B> </li> <br>
			 *      Is the control surface a leading ( = 0) or trailing (> 0) edge effector? Overrides
			 *      the assumed default value set by the geometry: If the percentage along
			 *      the airfoil chord is < 50% a leading edge flap is assumed, while >= 50% indicates a
			 *      trailing edge flap.
			 * </ul>
			 * \endif
			 */
			keyWord = "leOrTe";
			status = search_jsonDictionary( controlTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				(*vlmControl)[i].leOrTeOverride = (int) true;
				status = string_toInteger(keyValue, &(*vlmControl)[i].leOrTe);
				if (status != CAPS_SUCCESS) goto cleanup;

				if (keyValue != NULL) EG_free(keyValue);
				keyValue = NULL;
			}

			/*! \page vlmControl
			 * \if (AVL)
			 * <ul>
			 * <li> <B>controlGain = 1.0</B> </li> <br>
			 *      Control deflection gain, units:  degrees deflection / control variable
			 * </ul>
			 * \endif
			 */
			keyWord = "controlGain";
			status = search_jsonDictionary( controlTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				status = string_toDouble(keyValue, &(*vlmControl)[i].controlGain);
				if (status != CAPS_SUCCESS) goto cleanup;

				if (keyValue != NULL) EG_free(keyValue);
				keyValue = NULL;
			}

			/*! \page vlmControl
			 * \if (AVL)
			 * <ul>
			 * <li> <B>hingeLine = [0.0 0.0 0.0]</B> </li> <br>
			 *      Alternative vector giving hinge axis about which surface rotates
			 * </ul>
			 * \endif
			 */
			keyWord = "hingeLine";
			status = search_jsonDictionary( controlTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				status = string_toDoubleArray(keyValue, 3, (*vlmControl)[i].xyzHingeVec);
				if (status != CAPS_SUCCESS) goto cleanup;

				if (keyValue != NULL) EG_free(keyValue);
				keyValue = NULL;
			}

			/*! \page vlmControl
			 * \if (AVL)
			 * <ul>
			 * <li> <B>deflectionDup = 0 </B> </li> <br>
			 *      Sign of deflection for duplicated surface
			 * </ul>
			 * \endif
			 */
			keyWord = "deflectionDup";
			status = search_jsonDictionary( controlTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				status = string_toInteger(keyValue, &(*vlmControl)[i].deflectionDup);
				if (status != CAPS_SUCCESS) goto cleanup;

				if (keyValue != NULL) EG_free(keyValue);
				keyValue = NULL;
			}



        } else {

            /*! \page vlmControl
             * \section keyStringVLMControl Single Value String
             *
             * If "Value" is a single string the following options maybe used:
             * - (NONE Currently)
             *
             */
            printf("\tNo current defaults for get_vlmControl, tuple value must be a JSON string\n");
            status = CAPS_BADVALUE;
            goto cleanup;
        }
    }

    printf("\tDone getting vortex lattice control surface data\n");

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("\tError in get_vlmSurface, status = %d", status);

        if (keyValue != NULL) EG_free(keyValue);

        return status;
}


// Initiate (0 out all values and NULL all pointers) of a control in the vlmcontrol structure format
int initiate_vlmControl(vlmControlStruct *control) {

	control->name = NULL; // Control surface name

    control->deflectionAngle = 0.0; // Deflection angle of the control surface
    control->controlGain = 1.0; //Control deflection gain, units:  degrees deflection / control variable

    control->percentChord = 0.0; // Percentage along chord

    control->xyzHinge[0] = 0.0; // xyz location of the hinge
    control->xyzHinge[1] = 0.0;
    control->xyzHinge[2] = 0.0;

    control->xyzHingeVec[0] = 0.0; // Vector of hinge line at xyzHinge
    control->xyzHingeVec[1] = 0.0;
    control->xyzHingeVec[2] = 0.0;


    control->leOrTeOverride = (int) false; // Does the user want to override the geometry set value?
    control->leOrTe = 0; // Leading = 0 or trailing > 0 edge control surface

    control->deflectionDup = 0; // Sign of deflection for duplicated surface

    return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) of a control in the vlmcontrol structure format
int destroy_vlmControl(vlmControlStruct *control) {

	if (control->name != NULL) EG_free(control->name);
	control->name = NULL; // Control surface name

    control->deflectionAngle = 0.0; // Deflection angle of the control surface

    control->controlGain = 1.0; //Control deflection gain, units:  degrees deflection / control variable

    control->percentChord = 0.0; // Percentage along chord

    control->xyzHinge[0] = 0.0; // xyz location of the hinge
    control->xyzHinge[1] = 0.0;
    control->xyzHinge[2] = 0.0;

    control->xyzHingeVec[0] = 0.0; // Vector of hinge line at xyzHinge
    control->xyzHingeVec[1] = 0.0;
    control->xyzHingeVec[2] = 0.0;

    control->leOrTeOverride = (int) false; // Does the user want to override the geometry set value?
    control->leOrTe = 0; // Leading = 0 or trailing > 0 edge control surface

    control->deflectionDup = 0; // Sign of deflection for duplicated surface

    return CAPS_SUCCESS;
}


// Initiate (0 out all values and NULL all pointers) of a section in the vlmSection structure format
int initiate_vlmSection(vlmSectionStruct *section) {

    section->bodyIndex = 0;
    section->sectionIndex = 0;

    section->xyzLE[0] = 0.0;
    section->xyzLE[1] = 0.0;
    section->xyzLE[2] = 0.0;
    section->nodeIndexLE = 0;

    section->xyzTE[0] = 0.0;
    section->xyzTE[1] = 0.0;
    section->xyzTE[2] = 0.0;
    section->nodeIndexTE = 0;


    section->numControl = 0;
    section->vlmControl = NULL;

    return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) of a section in the vlmSection structure format
int destroy_vlmSection(vlmSectionStruct *section) {

    int status; // Function return status
    int i; // Indexing

    section->bodyIndex = 0;
    section->sectionIndex = 0;

    section->xyzLE[0] = 0.0;
    section->xyzLE[1] = 0.0;
    section->xyzLE[2] = 0.0;
    section->nodeIndexLE = 0;

    section->xyzTE[0] = 0.0;
    section->xyzTE[1] = 0.0;
    section->xyzTE[2] = 0.0;
    section->nodeIndexTE = 0;

    for (i = 0; i < section->numControl; i++) {
        status = destroy_vlmControl(&section->vlmControl[i]);
        if (status != CAPS_SUCCESS) printf("destroy_vlmControl %d\n", status);
    }

    if (section->vlmControl != NULL) EG_free(section->vlmControl);
    section->vlmControl = NULL;
    section->numControl = 0;

    return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) of a surface in the vlmSurface structure format
int initiate_vlmSurface(vlmSurfaceStruct  *surface) {

    // Surface information
    surface->name   = NULL;

    surface->numAttr = 0;    // Number of capsGroup/attributes used to define a given surface
    surface->attrIndex = NULL; // List of attribute map integers that correspond given capsGroups

    surface->Nchord = 0;
    surface->Cspace = 0.0;

    surface->Nspan = 0;
    surface->Sspace = 0.0;

    surface->nowake = (int) false;
    surface->noalbe = (int) false;;
    surface->noload = (int) false;;
    surface->compon = 0;
    surface->iYdup  = (int) false;

    // Section storage
    surface->numSection = 0;
    surface->vlmSection = NULL;

    return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) of a surface in the vlmSurface structure format
int destroy_vlmSurface(vlmSurfaceStruct  *surface) {

    int status; // Function return status
    int i; // Indexing

    // Surface information
    if (surface->name != NULL) EG_free(surface->name);
    surface->name = NULL;

    surface->numAttr = 0;    // Number of capsGroup/attributes used to define a given surface
    if (surface->attrIndex != NULL) EG_free(surface->attrIndex);
    surface->attrIndex = NULL; // List of attribute map integers that correspond given capsGroups


    surface->Nchord = 0;
    surface->Cspace = 0.0;

    surface->Nspan = 0;
    surface->Sspace = 0.0;

    surface->nowake = (int) false;
    surface->noalbe = (int) false;;
    surface->noload = (int) false;;
    surface->compon = 0;
    surface->iYdup  = (int) false;

    // Section storage
    for (i = 0; i < surface->numSection; i++ ) {
        status = destroy_vlmSection(&surface->vlmSection[i]);
        if (status != CAPS_SUCCESS) printf("destroy_vlmSection status = %d", status);
    }

    if (surface->vlmSection != NULL) EG_free(surface->vlmSection = NULL);
    surface->vlmSection = NULL;
    surface->numSection = 0;

    return CAPS_SUCCESS;
}

// Populate vlmSurface-section control surfaces from geometry attributes, modify control properties based on
// incoming vlmControl structures
int get_ControlSurface(ego bodies[],
					   int numControl,
					   vlmControlStruct vlmControl[],
					   vlmSurfaceStruct *vlmSurface) {

	int status = 0; // Function status return

	int section, control, attr, index; // Indexing

	int atype, alen; // EGADS return variables
	const int    *ints;
	const double *reals;
	const char *string;

	const char *attributeKey = "vlmControl", *aName = NULL;
	const char *attrName = NULL;

	int numAttr = 0; // Number of attributes

	int stringLength = 0; // Length of a string

	// Control related variables
	double chordLength, chordPercent, chordVec[3];

	for (section = 0; section < vlmSurface->numSection; section++) {

		vlmSurface->vlmSection[section].numControl = 0;

		status = EG_attributeNum(bodies[vlmSurface->vlmSection[section].bodyIndex], &numAttr);
		if (status != EGADS_SUCCESS)  return status;

		// Control attributes
		for (attr = 0; attr < numAttr; attr++) {

			status = EG_attributeGet(bodies[vlmSurface->vlmSection[section].bodyIndex],
									 attr+1,
									 &aName,
									 &atype,
									 &alen,
									 &ints,
									 &reals, &string);


			if (status != EGADS_SUCCESS) continue;
			if (atype  != ATTRREAL)      continue;

			if (strncmp(aName, attributeKey, strlen(attributeKey)) != 0) continue;

			if (alen == 0) {
				printf( "Warning: %s should be followed by a single value corresponding to the flap location "
						"as a function of the chord. 0 - 1 (fraction - %% / 100), 1-100 (%%)\n", attributeKey);
				continue;
			}

			if (reals[0] > 100) {
				printf( "Warning: %s value (%f) must be less than 100\n", attributeKey, reals[0]);
				continue;
			}

			//printf("Attribute name = %s\n", aName);

			if (aName[strlen(attributeKey)] == '\0')

				attrName = "Flap";

			else if	(aName[strlen(attributeKey)] == '_') {

				attrName = &aName[strlen(attributeKey)+1];

			} else {

				attrName = &aName[strlen(attributeKey)];
			}

			//printf("AttrName = %s\n", attrName);

			vlmSurface->vlmSection[section].numControl += 1;

			index = vlmSurface->vlmSection[section].numControl-1; // Make copy to shorten the following lines of code

			if (vlmSurface->vlmSection[section].numControl == 1) {
				vlmSurface->vlmSection[section].vlmControl = (vlmControlStruct *) EG_alloc(vlmSurface->vlmSection[section].numControl*sizeof(vlmControlStruct));

			} else {

				vlmSurface->vlmSection[section].vlmControl = (vlmControlStruct *) EG_reall(vlmSurface->vlmSection[section].vlmControl,
																						   vlmSurface->vlmSection[section].numControl*sizeof(vlmControlStruct));
			}

			if (vlmSurface->vlmSection[section].vlmControl == NULL) return EGADS_MALLOC;

			status = initiate_vlmControl(&vlmSurface->vlmSection[section].vlmControl[index]);
			if (status != CAPS_SUCCESS) return status;

			// Get name of control surface
			if (vlmSurface->vlmSection[section].vlmControl[index].name != NULL) EG_free(vlmSurface->vlmSection[section].vlmControl[index].name);

			stringLength = strlen(attrName);

			vlmSurface->vlmSection[section].vlmControl[index].name = (char *) EG_alloc((stringLength+1)*sizeof(char));
			if (vlmSurface->vlmSection[section].vlmControl[index].name == NULL) return EGADS_MALLOC;

			memcpy(vlmSurface->vlmSection[section].vlmControl[index].name,
					attrName,
					stringLength*sizeof(char));

			vlmSurface->vlmSection[section].vlmControl[index].name[stringLength] = '\0';

			// Loop through control surfaces from input Tuple and see if defaults have be augmented
			for (control = 0; control < numControl; control++) {

				if (strcasecmp(vlmControl[control].name, attrName) != 0) continue;

				status = copy_vlmControlStruct(&vlmControl[control],
						  	  	  	  	  	   &vlmSurface->vlmSection[section].vlmControl[index]);
				if (status != CAPS_SUCCESS) return status;
			}

			// Get percent of chord from attribute
			vlmSurface->vlmSection[section].vlmControl[index].percentChord = reals[0];

			// Was value given as a percentage or fraction
			if (vlmSurface->vlmSection[section].vlmControl[index].percentChord  >= 1.0) {
				vlmSurface->vlmSection[section].vlmControl[index].percentChord  = vlmSurface->vlmSection[section].vlmControl[index].percentChord / 100;
			}

			if (vlmSurface->vlmSection[section].vlmControl[index].leOrTeOverride == (int) false) {

				if (vlmSurface->vlmSection[section].vlmControl[index].percentChord  < 0.5) {
					vlmSurface->vlmSection[section].vlmControl[index].leOrTe = 0;
				} else {
					vlmSurface->vlmSection[section].vlmControl[index].leOrTe = 1;
				}
			}

			// Get xyz of hinge location
			chordVec[0] = vlmSurface->vlmSection[section].xyzTE[0] - vlmSurface->vlmSection[section].xyzLE[0];
			chordVec[1] = vlmSurface->vlmSection[section].xyzTE[1] - vlmSurface->vlmSection[section].xyzLE[1];
			chordVec[2] = vlmSurface->vlmSection[section].xyzTE[2] - vlmSurface->vlmSection[section].xyzLE[2];
			chordLength = sqrt(pow(chordVec[0], 2) +
							   pow(chordVec[1], 2) +
							   pow(chordVec[2], 2));
			chordPercent = vlmSurface->vlmSection[section].vlmControl[index].percentChord;

			vlmSurface->vlmSection[section].vlmControl[index].xyzHinge[0] = chordLength*chordVec[0] +
																			vlmSurface->vlmSection[section].xyzLE[0];

			vlmSurface->vlmSection[section].vlmControl[index].xyzHinge[1] = chordLength*chordVec[1] +
			    														    vlmSurface->vlmSection[section].xyzLE[1];

			vlmSurface->vlmSection[section].vlmControl[index].xyzHinge[2] = chordLength*chordVec[2] +
																			vlmSurface->vlmSection[section].xyzLE[2];
			/*
			printf("Hinge location = %f %f %f\n", vlmSurface->vlmSection[section].vlmControl[index].xyzHinge[0],
												  vlmSurface->vlmSection[section].vlmControl[index].xyzHinge[1],
												  vlmSurface->vlmSection[section].vlmControl[index].xyzHinge[2]);
			 */

			/*// Set the hinge vector to the plane normal - TOBE ADDED
			vlmSurface->vlmSection[section].vlmControl[index].xyzHingeVec[0] =
			vlmSurface->vlmSection[section].vlmControl[index].xyzHingeVec[1] =
			vlmSurface->vlmSection[section].vlmControl[index].xyzHingeVec[2] =
			*/
		}
	}

	return CAPS_SUCCESS;
}

/*
// Retrieve the string following a vlmControl tag
int retrieve_vlmControlAttr(ego geomEntity, const char **attrName, const char **string) {
	//int retrieve_vlmControlAttr(ego geomEntity, const char **string) {

    int status; // Status return

    int i; // Indexing

    int atype, alen; // EGADS return variables
    const int    *ints;
    const double *reals;

    const char *attributeKey = "vlmControl", *aName = NULL;

    int numAttr = 0;


    status = EG_attributeRet(geomEntity, attributeKey, &atype, &alen, &ints, &reals, string);
    if (status != EGADS_SUCCESS) return status;

    if (atype != ATTRSTRING) {
        printf ("Error: Attribute vlmControl should be followed by a string\n");
        return EGADS_NOTFOUND;
    }


    status = EG_attributeNum(geomEntity, &numAttr);
    if (status != EGADS_SUCCESS)  return status;

    for (i = 0; i < numAttr; i++) {
    	status = EG_attributeGet(geomEntity, i+1, &aName, &atype, &alen, &ints, &reals, string);
    	if (status != EGADS_SUCCESS) continue;
    	if (atype  != ATTRSTRING)    continue;

    	if (strncmp(aName, attributeKey, strlen(attributeKey)) != 0) continue;

    	*attrName = &aName[strlen(attributeKey)];

    	// Debug
    	//printf("AttrName = %s, length = %d\n", *attrName, strlen(*attrName));

    	return CAPS_SUCCESS;
    }

    //return CAPS_SUCCESS;

    return EGADS_NOTFOUND;
}
*/


// Make a copy of attribute map (controlIn)
int copy_vlmControlStruct(vlmControlStruct *controlIn, vlmControlStruct *controlOut) {

	int stringLength = 0;

	if (controlIn->name != NULL) {
		stringLength = strlen(controlIn->name);

		if (controlOut->name != NULL) EG_free(controlOut->name);

		controlOut->name = (char *) EG_alloc((stringLength+1)*sizeof(char));
		if (controlOut->name == NULL) return EGADS_MALLOC;

		memcpy(controlOut->name,
			   controlIn->name,
			   stringLength*sizeof(char));

		controlOut->name[stringLength] = '\0';
	}

	controlOut->deflectionAngle = controlIn->deflectionAngle; // Deflection angle of the control surface
	controlOut->controlGain = controlIn->controlGain; //Control deflection gain, units:  degrees deflection / control variable


	controlOut->percentChord = controlIn->percentChord; // Percentage along chord

	controlOut->xyzHinge[0] = controlIn->xyzHinge[0]; // xyz location of the hinge
	controlOut->xyzHinge[1] = controlIn->xyzHinge[1];
	controlOut->xyzHinge[2] = controlIn->xyzHinge[2];

	controlOut->xyzHingeVec[0] = controlIn->xyzHingeVec[0]; // xyz location of the hinge
	controlOut->xyzHingeVec[1] = controlIn->xyzHingeVec[1];
	controlOut->xyzHingeVec[2] = controlIn->xyzHingeVec[2];


	controlOut->leOrTeOverride = controlIn->leOrTeOverride; // Does the user want to override the geometry set value?
	controlOut->leOrTe = controlIn->leOrTe; // Leading = 0 or trailing > 0 edge control surface

	controlOut->deflectionDup = controlIn->deflectionDup; // Sign of deflection for duplicated surface

	return CAPS_SUCCESS;
}
