# Import pyCAPS class file
from pyCAPS import capsProblem

# Import os module   
try: 
    import os
except:
    print "Unable to import os module"
    raise SystemError


try: 
    import argparse
except: 
    print "Unable to import argparse module"
    raise SystemError

# Setup and read command line options. Please note that this isn't required for pyCAPS
parser = argparse.ArgumentParser(description = 'FUN3D and Tetgen Pytest Example',
                                 prog = 'fun3d_and_Tetgen_PyTest.py',
                                 formatter_class = argparse.ArgumentDefaultsHelpFormatter)

#Setup the available commandline options
parser.add_argument('-workDir', default = "./", nargs=1, type=str, help = 'Set working/run directory')
parser.add_argument('-noAnalysis', action='store_true', default = False, help = "Don't run analysis code")
args = parser.parse_args()


# Initialize pyCAPS object - directory where the pyCAPS.so resides
myProblem = capsProblem(libDir = os.environ['ESP_ROOT'] + "/lib")

# Create working directory variable 
workDir = str(args.workDir[0]) + "/FUN3DTetgenAnalysisTest"

# Load CSM file 
myProblem.loadCAPS("./csmData/cfdMultiBody.csm")

# Set fidelity - could also be set during .loadAIM
myProblem.capsFidelity = "CFD"

# Change a design parameter - area in the geometry 
myProblem.geometry.setGeometryVal("area", 50)

# Write out a new egads file for the geometry 
myProblem.geometry.saveGeometry("pyCAPS_EGADS", directory = workDir, extension="egads") # .egads extension added by default

# Load Tetgen aim 
myProblem.loadAIM(aim = "tetgenAIM", analysisDir = workDir)

# Set new EGADS body tessellation parameters
myProblem.analysis["tetgenAIM"].setAnalysisVal("Tess_Params", [.05, 0.01, 20.0])

# Preserve surface mesh while messhing
myProblem.analysis["tetgenAIM"].setAnalysisVal("Preserve_Surf_Mesh", True)

# Run AIM pre-analysis
myProblem.analysis["tetgenAIM"].aimPreAnalysis()

# NO analysis is needed - Tetgen was already ran dureing preAnalysis

# Run AIM post-analysis
myProblem.analysis["tetgenAIM"].aimPostAnalysis()

# Get analysis info. 
myProblem.analysis["tetgenAIM"].getAnalysisInfo()

# Load FUN3D aim - child of Tetgen AIM  
myProblem.loadAIM(aim = "fun3dAIM", 
                  altName = "fun3d", 
                  analysisDir = workDir, parents = ["tetgenAIM"])

# Set project name
myProblem.analysis["fun3d"].setAnalysisVal("Proj_Name", "fun3dTetgenTest")

myProblem.analysis["fun3d"].setAnalysisVal("Mesh_ASCII_Flag", False)
  
# Set AoA number 
myProblem.analysis["fun3d"].setAnalysisVal("Alpha", 1.0) 

# Set Mach number 
myProblem.analysis["fun3d"].setAnalysisVal("Mach", 0.5901)

# Set equation type
myProblem.analysis["fun3d"].setAnalysisVal("Eqn","compressible")

# Set Viscous term      
myProblem.analysis["fun3d"].setAnalysisVal("Viscous", "inviscid")

# Set number of iterations
myProblem.analysis["fun3d"].setAnalysisVal("Num_Iter",10)

# Set CFL number schedule
myProblem.analysis["fun3d"].setAnalysisVal("CFL_Schedule",[0.5, 3.0])

# Set read restart option
myProblem.analysis["fun3d"].setAnalysisVal("Restart_Read","off")

# Set CFL number iteration schedule
myProblem.analysis["fun3d"].setAnalysisVal("CFL_Schedule_Iter", [1, 40])
 
# Set overwrite fun3d.nml if not linking to Python library
myProblem.analysis["fun3d"].setAnalysisVal("Overwrite_NML", True)

# Set boundary conditions
inviscidBC1 = {"bcType" : "Inviscid", "wallTemperature" : 1}
inviscidBC2 = {"bcType" : "Inviscid", "wallTemperature" : 1.2}
myProblem.analysis["fun3d"].setAnalysisVal("Boundary_Condition", [("Wing1", inviscidBC1),
                                                                  ("Wing2", inviscidBC2),
                                                                  ("Farfield","farfield")])
 
# Run AIM pre-analysis
myProblem.analysis["fun3d"].aimPreAnalysis()

####### Run fun3d ####################
print "\n\nRunning FUN3D......"  
currentDirectory = os.getcwd() # Get our current working directory 

os.chdir(myProblem.analysis["fun3d"].analysisDir) # Move into test directory

if (args.noAnalysis == False): 
    os.system("nodet_mpi --animation_freq -1000 --write_aero_loads_to_file> Info.out"); # Run fun3d via system call
   
os.chdir(currentDirectory) # Move back to top directory 
#######################################

# Run AIM post-analysis
myProblem.analysis["fun3d"].aimPostAnalysis()

# Get force results
print "Total Force - Pressure + Viscous"
# Get Lift and Drag coefficients
print "Cl = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CLtot"), \
      "Cd = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CDtot")

# Get Cmx, Cmy, and Cmz coefficients 
print "Cmx = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CMXtot"), \
      "Cmy = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CMYtot"), \
      "Cmz = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CMZtot")
    
# Get Cx, Cy, Cz coefficients 
print "Cx = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CXtot"), \
      "Cy = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CYtot"), \
      "Cz = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CZtot")
    
print "Pressure Contribution"
# Get Lift and Drag coefficients 
print "Cl_p = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CLtot_p"), \
      "Cd_p = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CDtot_p")
    
# Get Cmx, Cmy, and Cmz coefficients 
print "Cmx_p = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CMXtot_p"), \
      "Cmy_p = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CMYtot_p"), \
      "Cmz_p = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CMZtot_p")
   
# Get Cx, Cy, and Cz, coefficients 
print "Cx_p = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CXtot_p"), \
      "Cy_p = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CYtot_p"), \
      "Cz_p = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CZtot_p")
    
print "Viscous Contribution"
# Get Lift and Drag coefficients 
print "Cl_v = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CLtot_v"), \
      "Cd_v = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CDtot_v")
  
# Get Cmx, Cmy, and Cmz coefficients 
print "Cmx_v = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CMXtot_v"), \
      "Cmy_v = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CMYtot_v"), \
      "Cmz_v = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CMZtot_v")
   
# Get Cx, Cy, and Cz, coefficients 
print "Cx_v = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CXtot_v"), \
      "Cy_v = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CYtot_v"), \
      "Cz_v = " , myProblem.analysis["fun3d"].getAnalysisOutVal("CZtot_v")

# Close CAPS 
myProblem.closeCAPS()