/*
    Simplified wrapper functions to CAPS - All inputs are typical c-types (int, double, etc.)
     so the user doesn't have explicit access to the CAPS objects. When compiled into a shared object
     having all of the inputs in typical c-types allows for easy integration into Python (using the ctypes module)

     Written by Dr. Ryan Durscher AFRL/RQVC
 */

#include "caps.h"
#include <string.h>

#ifdef WIN32
#define unlink   _unlink
#define getcwd   _getcwd
#define PATH_MAX _MAX_PATH
#else
#include <limits.h>
#include <unistd.h>
#endif

typedef struct{
	capsObj dobjParent;
	capsObj dobjChild;
} dataTransferDataSetStruct;


typedef struct{
  capsObj bobj; // Data Transfer (bound) object array

  int numDataSet;
  dataTransferDataSetStruct *dobj; // size[numDataSet]

} dataTransferStruct;

typedef struct {
    capsObj  pobj; // Problem object
    capsObj  *aobj; // Analysis object array

    dataTransferStruct *dataTransferobj;

    capsObj  obj, link_, parent, mobj; // CAPS objects

    int      numAobj; // Number of analysis objects
    int      numDobj; // Number of data transfer (bound) objects

    char     *apath, **fnames, **lines, *name;

    int      fidelity, nFields, *ranks, dirty, nErr, nLines;
    int      vlen, vdim, vnrow, vncol, naobjParents;

    enum     capsoType type;
    enum     capsvType vtype;
    enum     capssType subtype;

    enum     capsFixed vlfixed, vsfixed;
    enum     capsNull  vntype;

    capsErrs *errors;
    capsOwn  current;

} capsProblemStruct;

static capsProblemStruct *myProblem = NULL;
static int               numProblem = 0;


// Initialize a new analysis object.
static int initialize_Aobj(capsProblemStruct *pobj)
{
    pobj->numAobj += 1;

    if (pobj->numAobj == 1) {
        pobj->aobj = (capsObj *) EG_alloc(            pobj->numAobj*sizeof(capsObj));
    }
    else {
        pobj->aobj = (capsObj *) EG_reall(pobj->aobj, pobj->numAobj*sizeof(capsObj));
    }
  
    if (pobj->aobj == NULL) return EGADS_MALLOC;

    return 0;
}

// Initialize a new data transfer (bound) object.
static int initialize_dataTransferobj(capsProblemStruct *pobj)
{
    pobj->numDobj += 1;

    if (pobj->numDobj == 1) {
        pobj->dataTransferobj = (dataTransferStruct *) EG_alloc( pobj->numDobj*sizeof(dataTransferStruct));
    }
    else {
        pobj->dataTransferobj = (dataTransferStruct *) EG_reall(pobj->dataTransferobj, pobj->numDobj*sizeof(dataTransferStruct));
    }

    if (pobj->dataTransferobj == NULL) return EGADS_MALLOC;

    return 0;
}

static void report_Errors(capsProblemStruct *pobj)
{
    int i, j;
    int status;

    if (pobj->errors != NULL) {
        for (i = 0; i < pobj->nErr; i++) {

            status = caps_errorInfo(pobj->errors, i+1, &pobj->obj, &pobj->nLines, &pobj->lines);
            if (status != CAPS_SUCCESS) {
                printf(" CAPS Error: caps_errorInfo[%d] = %d!\n", i+1, status);
                continue;
            }

            for (j = 0; j < pobj->nLines; j++) printf(" %s\n", pobj->lines[j]);
        }

        caps_freeError(pobj->errors);
    }
}

// Close a pyCAPS
void closePYCAPS()
{
	if (myProblem != NULL) EG_free(myProblem);
    myProblem = NULL;
    numProblem = 0;
}

// Load a *.csm, *.caps, or a *.egads file
int loadCAPS(int *pIndex, char *file)
{
    char *pname = file;       // project name is filename
    int status;

    numProblem += 1;

    if (myProblem == NULL) {
      myProblem  = (capsProblemStruct *) EG_alloc(           numProblem*sizeof(capsProblemStruct));
    } else {
      myProblem  = (capsProblemStruct *) EG_reall(myProblem, numProblem*sizeof(capsProblemStruct));
    }
    if (myProblem == NULL) return EGADS_MALLOC;

    *pIndex = numProblem-1;

    myProblem[*pIndex].numAobj = 0;
    myProblem[*pIndex].numDobj = 0;
    myProblem[*pIndex].dataTransferobj = NULL;
    myProblem[*pIndex].aobj = NULL;

    status = caps_open(file, pname, &myProblem[*pIndex].pobj);

    if (status != CAPS_SUCCESS) printf(" caps_open = %d\n", status);

    return status;
}

// Close a CAPS problem
int closeCAPS(int pIndex)
{

    int status; // Function return status
    int i; //Indexing

    printf("Closing CAPS\n");

    status = caps_close(myProblem[pIndex].pobj);

    //if (status != CAPS_SUCCESS) printf(" caps_close = %d\n", status);

    if (myProblem[pIndex].aobj != NULL) EG_free(myProblem[pIndex].aobj); // Analysis object array

    if (myProblem[pIndex].dataTransferobj != NULL) {

    	for (i = 0; i < myProblem[pIndex].numDobj; i++) {

    		if (myProblem[pIndex].dataTransferobj[i].dobj != NULL) EG_free(myProblem[pIndex].dataTransferobj[i].dobj);
    	}

    	EG_free(myProblem[pIndex].dataTransferobj); // Data transfer array
    }

    return status;
}

// Save a CAPS problem
int saveCAPS(int pIndex, char *file)
{
    int status;

    status = caps_save(myProblem[pIndex].pobj, file);

    if (status != CAPS_SUCCESS) printf(" caps_save = %d\n",status);

    return status;
}

// Report Analysis info on dirty analyses
int dirtyCAPSAnalysis(int pIndex) {

	int status; // Function return
	int i, j; //Indexing

	int numDirty;

	capsObj *aobjDirty; // Freeable list of dirty analysis objects
	capsObj *aobjTemp;

	status = caps_dirtyAnalysis(myProblem[pIndex].pobj, &numDirty, &aobjDirty);
	if (status != CAPS_SUCCESS) return status;

	printf("\n\tNumber of dirty analyses = %d\n", numDirty);
	for (i = 0; i < numDirty; i++) {
		status = caps_analysisInfo(aobjDirty[i],
								   &myProblem[pIndex].apath,
								   &myProblem[pIndex].fidelity,
								   &myProblem[pIndex].naobjParents,
								   &aobjTemp,
								   &myProblem[pIndex].nFields,
								   &myProblem[pIndex].fnames,
								   &myProblem[pIndex].ranks,
								   &myProblem[pIndex].dirty);

		if (status != CAPS_SUCCESS) {
			printf(" caps_analysisInfo = %d\n", status);
			return status;
		}

		printf("\n Fidelity       = %d", myProblem[pIndex].fidelity);
		printf("\n APath          = %s", myProblem[pIndex].apath);
		printf("\n No. of Parents = %d", myProblem[pIndex].naobjParents);
		printf("\n Fields         =");

		for (j = 0; j < myProblem[pIndex].nFields; j++) {

			printf("  %s (%d)", myProblem[pIndex].fnames[j],
								myProblem[pIndex].ranks[j]);
		}

		printf("\n Dirty    =  %d\n", myProblem[pIndex].dirty);
	}

	return CAPS_SUCCESS;
}
// Load an AIM, set working directory and fidelity value
int loadAIM(int pIndex, char *aimName, char *directory, int fidel, int *index,
            int numParents, int *parentIndexes)
{
    int i;
    int status;

    capsObj *aobjTemp;

    aobjTemp = NULL;

    if (numParents != 0) {

        aobjTemp = (capsObj *) EG_alloc(numParents*sizeof(capsObj));

        if (aobjTemp == NULL) {

            numParents = 0;
            printf(" Warning: Malloc error on Parents!\n");

        } else {

            for (i = 0; i < numParents; i++)
                aobjTemp[i] = myProblem[pIndex].aobj[parentIndexes[i]];

        }
    }

    status = initialize_Aobj(&myProblem[pIndex]);

    if (status != CAPS_SUCCESS) {
        printf(" Initializing analysis objects = %d\n", status);
        if (aobjTemp != NULL)  EG_free(aobjTemp);
        //caps_close(myProblem[pIndex].pobj);
        return status;
    }

    *index = myProblem[pIndex].numAobj - 1;

    status = caps_load(myProblem[pIndex].pobj,
                       aimName,
                       directory,
                       fidel,
                       numParents,
                       aobjTemp,
                       &myProblem[pIndex].aobj[*index]);

    if (aobjTemp != NULL) EG_free(aobjTemp);

    if (status != CAPS_SUCCESS) {
        printf(" caps_load = %d\n", status);
        //caps_close(myProblem[pIndex].pobj);
        return status;
    }

    status = caps_info(myProblem[pIndex].pobj,
                      &myProblem[pIndex].name,
                      &myProblem[pIndex].type,
                      &myProblem[pIndex].subtype,
                      &myProblem[pIndex].link_,
                      &myProblem[pIndex].parent,
                      &myProblem[pIndex].current);

    if (status != CAPS_SUCCESS) {
        printf(" caps_info on Problem = %d\n", status);
        //caps_close(myProblem[pIndex].pobj);
    }

    return status;
}

// Get the ANALYSISIN type and length for a given variable
int getAnalysisInVarType(int pIndex,char *name, int *vartype,
                                                int *varlen,
                                                int *vardim,
                                                int *varnrow,
                                                int *varncol,
                                                int *varlfixed,
                                                int *varsfixed,
                                                int *varntype,
                                                int aIndex)
{
    int status;

    const char *valunits;

    status = caps_childByName(myProblem[pIndex].aobj[aIndex],
                              VALUE,
                              ANALYSISIN,
                              name,
                              &myProblem[pIndex].mobj);

    if (status != CAPS_SUCCESS) return status;


    status = caps_getValue(myProblem[pIndex].mobj,
                          &myProblem[pIndex].vtype,
                          &myProblem[pIndex].vlen,
                          NULL,
                          &valunits,
                          &myProblem[pIndex].nErr,
                          &myProblem[pIndex].errors);

    report_Errors(&myProblem[pIndex]);

    if (status != CAPS_SUCCESS) return status;

    status = caps_getValueShape(myProblem[pIndex].mobj,
                               &myProblem[pIndex].vdim,
                               &myProblem[pIndex].vlfixed,
                               &myProblem[pIndex].vsfixed,
                               &myProblem[pIndex].vntype,
                               &myProblem[pIndex].vnrow,
                               &myProblem[pIndex].vncol);

    *vartype   = myProblem[pIndex].vtype;
    *varlen    = myProblem[pIndex].vlen;
    *vardim    = myProblem[pIndex].vdim;
    *varnrow   = myProblem[pIndex].vnrow;
    *varncol   = myProblem[pIndex].vncol;
    *varlfixed = myProblem[pIndex].vlfixed;
    *varsfixed = myProblem[pIndex].vsfixed;
    *varntype  = myProblem[pIndex].vntype;

    return status;
}

// Set an ANALYSISIN variable - type = VOID - Other setAnalysisInVar_* are redundant with
//  proper casting in python
int setAnalysisInVar_VOID_P(int pIndex, char *name, void **val, int numRow, int numCol, int aIndex)
{

    int status;    status = caps_childByName(myProblem[pIndex].aobj[aIndex],
                              VALUE,
                              ANALYSISIN,
                              name,
                              &myProblem[pIndex].mobj);

    if (status != CAPS_SUCCESS) {
        printf(" caps_childByName = %d\n", status);
        return status;
    }

    if (numRow*numCol == 1) status = caps_setValue(myProblem[pIndex].mobj, numRow, numCol, &val[0]);
    else                    status = caps_setValue(myProblem[pIndex].mobj, numRow, numCol,  val);

    if (status == CAPS_LINKERR) {

        printf("CAPS_LINKERR detected while trying to set analysisIN variable \"%s\" - link will be broken in order to set value!!\n", name);

        // Break the link
        status = caps_makeLinkage(NULL, Copy, myProblem[pIndex].mobj);

        // If successful - try to reset the value
        if (status == CAPS_SUCCESS) {

            if (numRow*numCol == 1) status = caps_setValue(myProblem[pIndex].mobj, numRow, numCol, &val[0]);
            else                    status = caps_setValue(myProblem[pIndex].mobj, numRow, numCol,  val);

            if (status != CAPS_SUCCESS) {

                printf(" caps_setValue = %d\n", status);
                return status;
            }

        } else {

            printf(" caps_makeLinkage = %d\n", status);
            return status;
        }

    } else if (status != CAPS_SUCCESS) {

        printf(" caps_setValue = %d\n", status);
    }

    return status;
}

// Set an ANALYSISOUT variable - type = VOID
int getAnalysisOutVar_VOID_P(int pIndex, char *name, int *vartype,
                                                     int *varlen,
                                                     int *vardim,
                                                     int *varnrow,
                                                     int *varncol,
                                                     int *varlfixed,
                                                     int *varsfixed,
                                                     int *varntype,
                                                     void **val,
                                                     int aIndex)
{
    int status;

    const char *valunits;
    const void *value;

    status = caps_childByName(myProblem[pIndex].aobj[aIndex],
                              VALUE,
                              ANALYSISOUT,
                              name,
                              &myProblem[pIndex].mobj);

    if (status != CAPS_SUCCESS) {

        printf(" caps_childByName = %d\n", status);
        return status;
    }

    status = caps_getValue(myProblem[pIndex].mobj,
                          &myProblem[pIndex].vtype,
                          &myProblem[pIndex].vlen,
                          &value,
                          &valunits,
                          &myProblem[pIndex].nErr,
                          &myProblem[pIndex].errors);

    report_Errors(&myProblem[pIndex]);

    if (status != CAPS_SUCCESS) return status;

    status = caps_getValueShape(myProblem[pIndex].mobj,
                               &myProblem[pIndex].vdim,
                               &myProblem[pIndex].vlfixed,
                               &myProblem[pIndex].vsfixed,
                               &myProblem[pIndex].vntype,
                               &myProblem[pIndex].vnrow,
                               &myProblem[pIndex].vncol);

    *vartype   = myProblem[pIndex].vtype;
    *varlen    = myProblem[pIndex].vlen;
    *vardim    = myProblem[pIndex].vdim;
    *varnrow   = myProblem[pIndex].vnrow;
    *varncol   = myProblem[pIndex].vncol;
    *varlfixed = myProblem[pIndex].vlfixed;
    *varsfixed = myProblem[pIndex].vsfixed;
    *varntype  = myProblem[pIndex].vntype;

    *val = (void *) value;

    return status;
}

// Get analysis into
int getAnalysisInfo(int pIndex, int aIndex, int *cleanliness)
{
    int i;
    int status;

    capsObj *aobjTemp;

    status = caps_analysisInfo(myProblem[pIndex].aobj[aIndex],
                              &myProblem[pIndex].apath,
                              &myProblem[pIndex].fidelity,
                              &myProblem[pIndex].naobjParents,
                              &aobjTemp,
                              &myProblem[pIndex].nFields,
                              &myProblem[pIndex].fnames,
                              &myProblem[pIndex].ranks,
                              &myProblem[pIndex].dirty);

    if (status != CAPS_SUCCESS) {
        printf(" caps_analysisInfo = %d\n", status);
        //caps_close(myProblem[pIndex].pobj);
        return status;
    }

    printf("\n Fidelity       = %d", myProblem[pIndex].fidelity);
    printf("\n APath          = %s", myProblem[pIndex].apath);
    printf("\n No. of Parents = %d", myProblem[pIndex].naobjParents);
    printf("\n Fields         =");

    for (i = 0; i < myProblem[pIndex].nFields; i++) {

        printf("  %s (%d)", myProblem[pIndex].fnames[i],
                            myProblem[pIndex].ranks[i]);
    }

    printf("\n Dirty    =  %d\n", myProblem[pIndex].dirty);

    *cleanliness = myProblem[pIndex].dirty;

    return status;
}

// Get the GEOMETRYIN type and length for a given variable
int getGeometryInVarType(int pIndex, char *name, int *vartype,
                                                 int *varlen,
                                                 int *vardim,
                                                 int *varnrow,
                                                 int *varncol,
                                                 int *varlfixed,
                                                 int *varsfixed,
                                                 int *varntype)
{
    int status;

    const char *valunits;

    status = caps_childByName(myProblem[pIndex].pobj,
                              VALUE,
                              GEOMETRYIN,
                              name,
                              &myProblem[pIndex].mobj);

    if (status != CAPS_SUCCESS) {
        printf(" caps_childByName = %d\n", status);
        return status;
    }

    status = caps_getValue(myProblem[pIndex].mobj,
                          &myProblem[pIndex].vtype,
                          &myProblem[pIndex].vlen,
                          NULL,
                          &valunits,
                          &myProblem[pIndex].nErr,
                          &myProblem[pIndex].errors);

    report_Errors(&myProblem[pIndex]);

    if (status != CAPS_SUCCESS) return status;

    status = caps_getValueShape(myProblem[pIndex].mobj,
                               &myProblem[pIndex].vdim,
                               &myProblem[pIndex].vlfixed,
                               &myProblem[pIndex].vsfixed,
                               &myProblem[pIndex].vntype,
                               &myProblem[pIndex].vnrow,
                               &myProblem[pIndex].vncol);

    *vartype   = myProblem[pIndex].vtype;
    *varlen    = myProblem[pIndex].vlen;
    *vardim    = myProblem[pIndex].vdim;
    *varnrow   = myProblem[pIndex].vnrow;
    *varncol   = myProblem[pIndex].vncol;
    *varlfixed = myProblem[pIndex].vlfixed;
    *varsfixed = myProblem[pIndex].vsfixed;
    *varntype  = myProblem[pIndex].vntype;

    return status;
}

// Set an GEOMETRYIN variable - type = VOID
int setGeometryInVar_VOID_P(int pIndex, char *name, void **val, int numRow, int numCol)
{

    int status;

    status = caps_childByName(myProblem[pIndex].pobj,
                              VALUE,
                              GEOMETRYIN,
                              name,
                              &myProblem[pIndex].mobj);

    if (status != CAPS_SUCCESS) {
        printf(" caps_childByName = %d\n", status);
        return status;
    }

    if (numRow*numCol == 1)status = caps_setValue(myProblem[pIndex].mobj, numRow, numCol, &val[0]);
    else                   status = caps_setValue(myProblem[pIndex].mobj, numRow, numCol, val);

    if (status != CAPS_SUCCESS) printf(" caps_setValue = %d\n", status);

    return status;
}

// Save current geometry using EGADS
int saveGeometry(int pIndex, char *name, char*fileExt) {

	int status; // Function return status
	int i; // Indexing
	char *filename = NULL;

	int numBody=0;
	capsProblem *problem;

	ego body; // Temporary body container
	ego *bodies; // Model bodies
	ego model;
	char *units;

	if (name == NULL) return CAPS_BADVALUE;
	if (fileExt == NULL) return CAPS_BADVALUE;

	// Force regeneration of the geometry
	status = caps_preAnalysis(myProblem[pIndex].pobj,
	                         &myProblem[pIndex].nErr,
	                         &myProblem[pIndex].errors);
	if (status != CAPS_SUCCESS) printf(" caps_preAnalysis = %d\n", status);

	report_Errors(&myProblem[pIndex]);

	// How many bodies do we have
	status = caps_size(myProblem[pIndex].pobj, BODIES, NONE, &numBody);
	if (status != CAPS_SUCCESS) return status;

	if (numBody < 1) {
		printf("The number of bodies in the problem is less than 1!!!\n");
		return CAPS_BADVALUE;
	}
	// Allocate body array
	bodies = (ego *) EG_alloc(numBody*sizeof(ego));
	if (bodies == NULL) return EGADS_MALLOC;

	// Make copies of the bodies to had to the model
	for (i = 0; i < numBody; i++) {
		status = caps_bodyByIndex(myProblem[pIndex].pobj, i+1, &body, &units);
		if (status != CAPS_SUCCESS) goto bail;

		status = EG_copyObject(body, NULL, &bodies[i]);
		if (status != EGADS_SUCCESS) goto bail;
	}

	if (myProblem[pIndex].pobj->type  != PROBLEM) {
	    status = CAPS_BADTYPE;
	    goto bail;
	}

	if (myProblem[pIndex].pobj->blind == NULL){
	    status = CAPS_NULLBLIND;
	    goto bail;
	}

	problem = (capsProblem *) myProblem[pIndex].pobj->blind;

	// Make a model from the context and bodies
	status = EG_makeTopology(problem->context, NULL, MODEL, 0,
			                 NULL, numBody, bodies, NULL, &model);
	if (status != EGADS_SUCCESS) goto bail;

    // Allocate space for file name
    filename = (char *) EG_alloc((strlen(name)+strlen(fileExt) + 1)*sizeof(char));
    if (filename == NULL) {
        status = EGADS_MALLOC;
        goto bail;
    }

    sprintf(filename, "%s%s",name, fileExt);
    filename[strlen(name)+strlen(fileExt)] ='\0';

	// Save the model
	status = EG_saveModel(model, filename);
	if (status != EGADS_SUCCESS) goto bail;

	// Clean up - delete the model
	status = EG_deleteObject(model);
	if (status != EGADS_SUCCESS) goto bail;

	status = CAPS_SUCCESS;
	goto bail;

	bail:
	    // Free filename
        if (filename != NULL) EG_free(filename);

        // Clean up - delete copies bodies
        if (bodies != NULL) EG_free(bodies);

        return status;
}

// Run AIM pre-analysis
int aimPreAnalysis(int pIndex, int aIndex)
{
    int status;

    status = caps_preAnalysis(myProblem[pIndex].aobj[aIndex],
                             &myProblem[pIndex].nErr,
                             &myProblem[pIndex].errors);

    if (status != CAPS_SUCCESS) printf(" caps_preAnalysis = %d\n", status);

    report_Errors(&myProblem[pIndex]);

    return status;
}

// Run AIM post-analysis
int aimPostAnalysis(int pIndex, int aIndex)
{
    int status;

    status = caps_postAnalysis(myProblem[pIndex].aobj[aIndex],
                               myProblem[pIndex].current,
                              &myProblem[pIndex].nErr,
                              &myProblem[pIndex].errors);

    if (status != CAPS_SUCCESS) printf(" caps_postAnalysis = %d\n", status);

    report_Errors(&myProblem[pIndex]);

    return status;
}

// Create a data transfer set
int createDataTransfer(int pIndex,
					   int numDataSet,
					   int numAIndexUnique,
					   int *aIndexUnique, // size[numAIndexUnique]
                       int *aIndexParent, // size[numDataSet]
                       int *aIndexChild,  // size[numDataSet]
                       char *capsTransfer,
                       char **varName, // size[numDataSet]
                       int *varRank, // size[numDataSet]
					   int *dataTransferMethod, // size[numDataSet]
                       int *dIndex) {

	int status; // Function return status

	int i, j; // Indexing

    // Setting up vsets
    capsObj *vset = NULL;     // size[numAIndexUnique]
    enum capsdMethod dataTransferMethodType;

    printf("Building data bound - %s\n", capsTransfer);
    status = initialize_dataTransferobj(&myProblem[pIndex]);

    if (status != CAPS_SUCCESS) {
        printf(" Initializing data transfer objects = %d\n", status);
        return status;
    }

    *dIndex = myProblem[pIndex].numDobj - 1;

    status = caps_makeBound(myProblem[pIndex].pobj,
                            2, //int dim,
                            capsTransfer,
                            &myProblem[pIndex].dataTransferobj[*dIndex].bobj);
    if (status != CAPS_SUCCESS) {
        printf("  Making bounds = %d\n",status);
        return status;
    }


    myProblem[pIndex].dataTransferobj[*dIndex].numDataSet = numDataSet;

    myProblem[pIndex].dataTransferobj[*dIndex].dobj = (dataTransferDataSetStruct *) EG_alloc(myProblem[pIndex].dataTransferobj[*dIndex].numDataSet *sizeof(dataTransferDataSetStruct));
    if (myProblem[pIndex].dataTransferobj[*dIndex].dobj == NULL) return EGADS_MALLOC;

    vset = (capsObj *) EG_alloc(numAIndexUnique*sizeof(capsObj));
    if (vset == NULL) return EGADS_MALLOC;

    for (i = 0; i < numAIndexUnique; i++) {

    	status = caps_makeVertexSet(myProblem[pIndex].dataTransferobj[*dIndex].bobj,
    								myProblem[pIndex].aobj[aIndexUnique[i]],
									NULL,
									&vset[i]);
    	if (status != CAPS_SUCCESS) {
    		printf("Make vertexSet - %d\n", status);
    		if (vset != NULL) EG_free(vset);
    		return status;
    	}
    }

    for (i = 0; i < numDataSet; i++) {

    	for (j = 0; j < numAIndexUnique; j++) {
    		if (aIndexParent[i] == aIndexUnique[j]) break;
    	}

    	//printf("Parent index = %d, varname = %s\n", aIndexUnique[j],varName[i]);
    	status = caps_makeDataSet(vset[j],
    							  varName[i],
								  Analysis,
								  varRank[i],
								  &myProblem[pIndex].dataTransferobj[*dIndex].dobj[i].dobjParent);
    	if (status != CAPS_SUCCESS) {
    		printf("Make dataSet parent - %d\n", status);
    		if (vset != NULL) EG_free(vset);
    		return status;
    	}

    	for (j = 0; j < numAIndexUnique; j++) {
    		if (aIndexChild[i] == aIndexUnique[j]) break;
    	}

    	//printf("Child index = %d, varname = %s\n", aIndexUnique[j], varName[i]);

    	if (dataTransferMethod[i] == 0) dataTransferMethodType = Interpolate;
    	else                            dataTransferMethodType = Conserve;

    	status = caps_makeDataSet(vset[j],
    							  varName[i],
								  dataTransferMethodType, //Conserve,
								  varRank[i],
								  &myProblem[pIndex].dataTransferobj[*dIndex].dobj[i].dobjChild);
    	if (status != CAPS_SUCCESS) {
    		printf("Make dataSet child - %d\n", status);
    		if (vset != NULL) EG_free(vset);
    		return status;
    	}
    }

    status = caps_completeBound(myProblem[pIndex].dataTransferobj[*dIndex].bobj);
    if (status != CAPS_SUCCESS) {
    	printf("Complete bound - %d\n", status);
    	if (vset != NULL) EG_free(vset);
    	return status;
    }

    if (vset != NULL) EG_free(vset);

    return CAPS_SUCCESS;
}

int getDataTransfer(int pIndex, int dIndex, char *varName, int varIndex) {
	int status; // Function return status

    //int i; // Indexing
    //int numDataSet;
    int npts;
    int rank;
    double *data;
    char *units;

    //capsObj *dsets;

    /*
 	status  = caps_getDataSets(myProblem[pIndex].dataTransferobj[dIndex].bobj, varName, &numDataSet, &dsets);
    printf("Status %d\n",status);
    printf("Varname %s\n",varName);
    printf("Number of datasets numDataSet = %d\n",numDataSet);
    */

    //printf(" # xyz dsets = %d, stat = %d\n", n, stat);
    //stat  = caps_getData(dsets[0], &len, &rank, &data, &units);

    /*
    for (i = 0; i < numDataSet; i++) {
    	 status = caps_getData(dsets[i], &npts, &rank, &data, &units);
    }
    */


    status = caps_getData(myProblem[pIndex].dataTransferobj[dIndex].dobj[varIndex].dobjParent, &npts, &rank, &data, &units);
    if (status != CAPS_SUCCESS) return status;

    printf("Source: %s\n", varName);
    printf("Number of points = %d\n",npts);
    printf("Rank = %d\n",rank);
    /*
    for (i = 0; i < npts; i++) {
         printf("i = %d, dataPoint = %f\n",i, data[i]);
         if (i > 10) break;
     }
    */
    status = caps_getData(myProblem[pIndex].dataTransferobj[dIndex].dobj[varIndex].dobjChild, &npts, &rank, &data, &units);
    if (status != CAPS_SUCCESS) return status;


    printf("Destination: %s\n", varName);
    printf("Number of points = %d\n",npts);
    printf("Rank = %d\n",rank);
    /*
    for (i = 0; i < npts; i++) {
        printf("i = %d, dataPoint = %f\n",i, data[i]);
        if (i > 10) break;
    }
    */

    return CAPS_SUCCESS;
}
