"""
Python interface to the pyCAPS.c wrapper:
The primary proposes of this file are to provide a miniature python framework to use CAPS and 
    to simplify the inputs to pyCAPS.c for the user (i.e. the user doesn't need to worry about 
    using c_types type conversions such as c_char_p, c_int, etc).
"""

try:
    from ctypes import c_bool, c_int, c_double, c_char_p, c_void_p
    from ctypes import cdll, CDLL
    from ctypes import cast, POINTER, byref, Structure
except:
    print "Unable to import ctypes\n"
    raise SystemError

try: 
    from json import dumps as jsonDumps
except:
    print "Unable to import json\n"
    raise SystemError

try:
    import os
except:
    print "Unable to import os\n"
    raise SystemError

try: 
    import time
except: 
    print "Unable to import time\n"
    raise SystemError


class _capsTuple(Structure):
    """
    Defines the equalvalent capsTuple structure
    """
    _fields_ = [("name" , c_char_p),
                ("value", c_char_p)]

class capsProblem(object):
    """

     Defines a CAPS problem object

    """
    def __init__(self, **kwargs):

        self.libCAPS = None
        if "libDir" in kwargs.keys():
            self.libDir = kwargs["libDir"]
            self.fileSO = self.libDir + "/" + "pyCAPS.so"

            self.loadSO()

        else:
            self.fileSO = None
            print "Path to pyCAPS.so needs to be provided via kwarg - libDir"
            raise SystemExit

        if "capsFile" in kwargs.keys():
            self.capsFile = kwargs["capsFile"]
            self.loadCAPS()

        else:
            self.capsFile = None

        # Problem
        self.problemIndex = None

        # Geometry
        self.geometry = None

        # Analysis
        self.currentAIM = None
        self.aimGlobalCount = 0

        self.analysisDir = None
        self.analysis = {}

        # Data Transfer
        self.dataTransfer = {}
        
        # Other
        self.capsFidelity = None

        self.status = 0

        self.errorCode = {   0 : 'CAPS_SUCCESS',
                          -301 : 'CAPS_BADRANK',
                          -302 : 'CAPS_BADDSETNAME',
                          -303 : 'CAPS_NOTFOUND',
                          -304 : 'CAPS_BADINDEX',
                          -305 : 'CAPS_NOTCHANGED',
                          -306 : 'CAPS_BADTYPE',
                          -307 : 'CAPS_NULLVALUE',
                          -308 : 'CAPS_NULLNAME',
                          -309 : 'CAPS_NULLOBJ',
                          -310 : 'CAPS_BADOBJECT',
                          -311 : 'CAPS_BADVALUE',
                          -312 : 'CAPS_PARAMBNDERR',
                          -313 : 'CAPS_NOTCONNECT',
                          -314 : 'CAPS_NOTPARMTRIC',
                          -315 : 'CAPS_READONLYERR',
                          -316 : 'CAPS_FIXEDLEN',
                          -317 : 'CAPS_BADNAME',
                          -318 : 'CAPS_BADMETHOD',
                          -319 : 'CAPS_CIRCULARLINK',
                          -320 : 'CAPS_UNITERR',
                          -321 : 'CAPS_NULLBLIND',
                          -322 : 'CAPS_SHAPEERR',
                          -323 : 'CAPS_LINKERR',
                          -324 : 'CAPS_MISMATCH',
                          -325 : 'CAPS_NOTPROBLEM',
                          -326 : 'CAPS_RANGEERR',
                          -327 : 'CAPS_DIRTY',
                          -328 : 'CAPS_HIERARCHERR',
                          -329 : 'CAPS_STATEERR',
                          -330 : 'CAPS_SOURCEERR',
                          -331 : 'CAPS_EXISTS',
                          -332 : 'CAPS_IOERR',
                          -333 : 'CAPS_DIRERR',
                          -334 : 'CAPS_NOTIMPLEMENT',
                          -335 : 'CAPS_EXECERR',
                          -336 : 'CAPS_CLEAN',
                          -337 : 'CAPS_BADFIDELITY', 
                          
                          -30 : 'EGADS_EXISTS',
                          -29 : 'EGADS_ATTRERR',
                          -28 : 'EGADS_TOPOCNT',
                          -27 : 'EGADS_OCSEGFLT',
                          -26 : 'EGADS_BADSCALE',
                          -25 : 'EGADS_NOTORTHO',
                          -24 : 'EGADS_DEGEN',
                          -23 : 'EGADS_CONSTERR',
                          -22 : 'EGADS_TOPOERR',
                          -21 : 'EGADS_GEOMERR',
                          -20 : 'EGADS_NOTBODY',
                          -19 : 'EGADS_WRITERR',
                          -18 : 'EGADS_NOTMODEL',
                          -17 : 'EGADS_NOLOAD',
                          -16 : 'EGADS_RANGERR',
                          -15 : 'EGADS_NOTGEOM',
                          -14 : 'EGADS_NOTTESS',
                          -13 : 'EGADS_EMPTY',
                          -12 : 'EGADS_NOTTOPO',
                          -11 : 'EGADS_REFERCE',
                          -10 : 'EGADS_NOTXFORM',
                           -9 : 'EGADS_NOTCNTX',
                           -8 : 'EGADS_MIXCNTX',
                           -7 : 'EGADS_NODATA',
                           -6 : 'EGADS_NONAME',
                           -5 : 'EGADS_INDEXERR',
                           -4 : 'EGADS_MALLOC',
                           -3 : 'EGADS_NOTOBJ',
                           -2 : 'EGADS_NULLOBJ',
                           -1 : 'EGADS_NOTFOUND'}
        
        self.analysisFidelity = {'ALL' : 0, 
                                 'WAKE' : 32,
                                 'STRUCTURE' : 64,
                                 'LINEARAERO' : 128,
                                 'FULLPOTENTIAL' : 256,
                                 'CFD' : 512}
        
        self.geometricFidelity = {'ALL'  : 0, 
                                  'NODE' : 20,
                                  'WIREBODY' : 6,
                                  'FACEBODY' : 7,
                                  'SHEETBODY': 8,
                                  'SOLIDBODY': 9}

    def __del__(self):
        #print "pyCAPS closing pyCAPS.so"
        if self.fileSO is not None:
            # Probably need to change this for a Windows
            cdll.LoadLibrary(self.fileSO).dlclose(self.libCAPS._handle)
           # self.libCAPS.dlclose(self.libCAPS._handle)

    def loadSO(self):
        """
        Load pyCAPS shared object
        """

        if self.fileSO is not None:
            # Load pyCAPS shared object
            self.libCAPS = CDLL(self.fileSO)
            #self.libCAPS = cdll.LoadLibrary(self.fileSO)

    def loadCAPS(self, fileName=None):
        """
        Loads an *.csm, *.caps, or *.egads file
        """

        if fileName is not None:
            self.capsFile = fileName

        if self.checkCAPSFileStaus():
            index = c_int()
            self.status = self.libCAPS.loadCAPS(byref(index), c_char_p(self.capsFile))

            self.problemIndex = index.value

            self.checkStatus(msg = "\twhile loading file - " + self.capsFile, useTimer=False)
            self.geometry = _capsGeomety(self)

    def closeCAPS(self):
        """
        Close a CAPS problem
        """

        #print "Closing caps problem"
        if self.checkCAPSFileStaus():
            self.status = self.libCAPS.closeCAPS(c_int(self.problemIndex))
            self.checkStatus(msg = "\twhile closing CAPS", useTimer=False, closeCAPS=False)

    def saveCAPS(self, fname=None):
        """
        Save a CAPS problem
        """

        if fname is None:
            fname = "CAPS_File.caps"
            print "Using default file name - " + str(fname)

        self.status = self.libCAPS.saveCAPS(c_int(self.problemIndex), c_char_p(fname))
        self.checkStatus(msg = "\twhile saving CAPS")

    def dirtyCAPSAnalysis(self):
        """
        Report what analyses are dirty in the problem 
        """
        self.status = self.libCAPS.dirtyCAPSAnalysis(c_int(self.problemIndex))
        self.checkStatus(msg="\twhile check which analysis are dirty")
        
    def loadAIM(self, **kwargs):
        """
        Load an AIM - valid keyword arguements:
            aim - Name of the requested aim
            analysisDir - Directory for aim analysis
            capsFidelity - Fidelity in which to invoke the aim
            parents - List of parent aim names to initilize the aim with
        """
        
        geometricFidelity = 0
        analysisFidelity = 0
        resetFidelity = False
        
        if not self.checkCAPSFileStaus():
            return

        if "aim" in kwargs.keys():
            self.currentAIM = kwargs["aim"]

        if "capsFidelity" in kwargs.keys():
            if kwargs["capsFidelity"].upper() in self.analysisFidelity:
                analysisFidelity = self.analysisFidelity[kwargs["capsFidelity"].upper()]
                resetFidelity = True
            else:
                print "Not a valid capsFidelity!!!"
                self.status = -337
                self.checkStatus(msg = "\twhile loading aim - " + self.currentAIM)
        
        if "geometricFidelity" in kwargs.keys():
            if kwargs["geometricFidelity"].upper() in self.geometricFidelity:
                geometricFidelity = self.analysisFidelity[kwargs["geometricFidelity"].upper()]
                resetFidelity = True
            else:
                print "Not a valid geometric fidelity!!!"
                self.status = -337
                self.checkStatus(msg = "\twhile loading aim - " + self.currentAIM)

        if resetFidelity == True:
            self.capsFidelity = geometricFidelity + analysisFidelity
        else:
            if not isinstance(self.capsFidelity, (int, long)):
                if self.capsFidelity.upper() in self.analysisFidelity:
                    self.capsFidelity = self.analysisFidelity[self.capsFidelity.upper()]
                else:
                    self.status = -337
                    self.checkStatus(msg = "\twhile loading aim - " + self.currentAIM)
        
        if "analysisDir" in kwargs.keys():
            self.analysisDir = kwargs["analysisDir"]

        if "altName" in kwargs.keys():
            altName = kwargs["altName"]
        else:
            altName = self.currentAIM

        if "parents" in kwargs.keys():
            # Get parent index from anaylsis directionary
            parentIndexes = []

            if self.getAIMParents(kwargs["parents"], parentIndexes): # parentIndexes returned

                numParents = c_int(len(parentIndexes))
                parentList = (c_int*len(parentIndexes))(* parentIndexes)

            else:
                return

        else:
            numParents = c_int(0)
            parentList = (c_int*1)() # If no parents pass arbitary parent list

        aimIndex = c_int() # Returned aim index for referencing

        # Load the aim
        if self.checkAIMStatus() and not self.checkAnalysisDict(altName):

            self.status = self.libCAPS.loadAIM(c_int(self.problemIndex),
                                               c_char_p(self.currentAIM),
                                               c_char_p(self.analysisDir),
                                               c_int(self.capsFidelity),
                                               byref(aimIndex),
                                               numParents,
                                               parentList)

            self.checkStatus(msg = "\twhile loading aim - " + self.currentAIM)

            self.aimGlobalCount = self.aimGlobalCount + 1

            self.analysis[altName] = _capsAnalysis(altName,
                                                   self.currentAIM,
                                                   self.analysisDir,
                                                   aimIndex.value,
                                                   self)

            #print "A-object index = ", index.value
    
    def createDataTransfer(self, **kwargs):
        
        if self.aimGlobalCount < 2:
            print "At least two AIMs need to be loaded before a data transfer can be setup"
            if self.statusContinueTimer():
                self.closeCAPS()
                raise SystemExit
 
        numDataSet = 1
        if "variableName" in kwargs.keys():
            varName = kwargs["variableName"]
            
            if isinstance(varName, list):
                numDataSet = len(varName)
            else:
                numDataSet = 1
                varName = [varName] # Convert to list
            
            varNameList = varName
            varName = (c_char_p * numDataSet)(*varName) # Convert to c_char_p array
            
            #varName = (ctypes.c_char_p * numDataSet)()
            #arr[:] = temp
        else:
            print "No variable name (variableName) provided"
            if self.statusContinueTimer():
                self.closeCAPS()
                raise SystemExit
        
        if "variableRank" in kwargs.keys():
            varRank = kwargs["variableRank"]
            
            if isinstance(varRank, list):
                numTest = len(varRank)
            else:
                numTest = 1
                varRank = [varRank] # Convert to list
            
            varRank = (c_int* numTest)(*varRank) # Convert to c_int array
        else:
            print "No variable rank (variableRank) provided - assuming variable is a scalar"
            numTest = numDataSet
            varRank = [1]*numTest # Convert to list

            varRank = (c_int* numTest)(*varRank) # Convert to c_int array
        
        if (numTest != numDataSet):
            print "Dimensional mismatch between inputs!!!!"
            return
        
        aimIndex = [] # Collect all the aim Index so we can see howmany unique indexed there are 
        aimSrcIndex = []
        if "aimSrc" in kwargs.keys():
            if isinstance(kwargs["aimSrc"], list):
                pass
            else:
                kwargs["aimSrc"] = [kwargs["aimSrc"]]
                
            for i in kwargs["aimSrc"]:
                if i not in self.analysis.keys():
                    print "aimSrc = " + i + " not found in analysis dictionary!!!"
                    if self.statusContinueTimer():
                        self.closeCAPS()
                        raise SystemExit
                else:
                    aimSrcIndex.append(self.analysis[i].aIndex)
                
        
        else:
            for i in self.analysis.keys():
                if self.aimGlobalCount-1 == self.analysis[i].aIndex:
                    break
                
            print "No source AIM  (aimSrc) provided - assuming it is the last" + \
                  " loaded AIM - " + str(self.analysis[i].aimName)
            
            for i in xrange(numDataSet):
                aimSrcIndex.append(self.aimGlobalCount-1)           
        
        if len(aimSrcIndex) != numDataSet:
            print "Dimensional mismatch between inputs!!!!"
            return
        else:
            aimIndex = aimIndex + aimSrcIndex
            aimSrcIndex = (c_int* numDataSet)(*aimSrcIndex) # Convert to c_int array
      
        aimDestIndex = []
        if "aimDest" in kwargs.keys():
            
            if isinstance(kwargs["aimDest"], list):
                pass
            else:
                kwargs["aimDest"] = [kwargs["aimDest"]]
            
            for i in kwargs["aimDest"]:
                if i not in self.analysis.keys():
                    print "aimDest = " + kwargs["aimDest"] + " not found in analysis dictionary!!!"
                    if self.statusContinueTimer():
                        self.closeCAPS()
                        raise SystemExit
                else:
                    aimDestIndex.append(self.analysis[i].aIndex)
        else:
            for i in self.analysis.keys():
                if self.aimGlobalCount-2 == self.analysis[i].aIndex:
                    break
            
            print "No destination AIM (aimDest) provided - assuming it is the second to last" + \
                  " loaded AIM - " + str(self.analysis[i].aimName)
            
            for i in xrange(numDataSet):
                aimDestIndex.append(self.aimGlobalCount-2)
        
        if len(aimDestIndex) != numDataSet:
            print "Dimensional mismatch between inputs!!!!"
            return
        else:
            aimIndex = aimIndex + aimDestIndex
            aimDestIndex = (c_int* numDataSet)(*aimDestIndex) # Convert to c_int array
        
        dataTransferMethod = []
        if "transferMethod" in kwargs.keys():
            
            if isinstance(kwargs["transferMethod"], list):
                pass
            else:
                kwargs["transferMethod"] = [kwargs["transferMethod"]]
            
            for i in kwargs["transferMethod"]:
                if "conserve" == i.lower():
                    dataTransferMethod.append(1)
                elif "interpolate" == i.lower():
                    dataTransferMethod.append(0)
                else:
                    print "Unreconized data transfer method - defaulting to \"Interpolate\""
                    dataTransferMethod.append(0)
        else:
            print "No data transfer method (transferMethod) provided - defaulting to \"Interpolate\""
            dataTransferMethod =  [0]*numDataSet
        
        if len(dataTransferMethod) != numDataSet:
            print "Dimensional mismatch between inputs for transferMethod!!!!"
            return
        else:
            dataTransferMethod = (c_int* numDataSet)(*dataTransferMethod) # Convert to c_int array
        
        if "capsTransfer" in kwargs.keys():
            capsTransfer = kwargs["capsTransfer"]
        else:
            print "No transfer name (capsTransfer) name provided - the data variable name will be used" 
            return
            
        dataIndex = c_int() # Returned bound/data index for reference
     
        self.status = self.libCAPS.createDataTransfer(c_int(self.problemIndex),
                                                      c_int(numDataSet),
                                                      c_int(len(set(aimIndex))),
                                                      (c_int*len(set(aimIndex)))(*set(aimIndex)),
                                                      aimSrcIndex,
                                                      aimDestIndex,
                                                      c_char_p(capsTransfer),
                                                      varName,
                                                      varRank,
                                                      dataTransferMethod,
                                                      byref(dataIndex))
        self.checkStatus(msg = "\twhile creating data transfer - " + capsTransfer)
       
        self.dataTransfer[capsTransfer] = _capsDataTransfer(capsTransfer,
                                                            varNameList,
                                                            dataIndex.value,
                                                            self)
        
    def checkStatus(self, msg=None, useTimer=True, closeCAPS=True):
        """
        Check status flag returned by pyCAPS.c
        """
        errorFound = False
        if self.status not in self.errorCode:
            print "CAPS error detected: error code " + str(self.status) + " = " + " Undefined error"
            errorFound = True
        
        elif self.errorCode[self.status] != 'CAPS_SUCCESS':
            print "CAPS error detected: error code " + str(self.status) + " = " + \
                                                       self.errorCode[self.status] + \
                                                       " received"
            errorFound = True
                                                       
        if errorFound is True:
            if msg is not None:
                print msg
            
            try:
                import sys
            except:
                    print "Unable to import sys\n"
                    raise SystemError

            sys.tracebacklimit=0

            if useTimer is True:
                if self.statusContinueTimer():
                    if closeCAPS is True:
                        self.closeCAPS()
                        raise Exception("CAPS error detected!!!")
            else:
                if closeCAPS is True:
                    self.closeCAPS()
                raise Exception("CAPS error detected!!!")

    def statusContinueTimer(self):
        print ".....press Ctrl-C if you wish to proceed (you have 5 seconds)."
        try:
            time.sleep(5)
            return True    
        except KeyboardInterrupt:
            return False
        
        return True
    def checkCAPSFileStaus(self):
        """
        Check to see if a CAPS file has been specified
        """

        if self.capsFile is None:
            print "Error: A *.csm, *.caps, or *.egads file has not been loaded "
            return False
        else:
            return True

    def getAIMParents(self, parentList, parentIndexes):
        """
        Check the requested parent list against the analysis dictionary
        """

        for i in parentList:
            if i in self.analysis.keys():
                parentIndexes.append(self.analysis[i].aIndex)
            else:
                print "Requested parent", i, "has not been initialized/loaded"
                return False

        return True

    def checkAnalysisDict(self, name):
        """
        Check the analysis dictionary for redundant names - need unique names
        """

        if name in self.analysis.keys():
            print "Analysis object with that name has already been initilized -" + \
                  "Please use a different (alternative) name"
            return True
        else:
            return False

    def checkAIMStatus(self):
        """
        Check to see if an AIM has been fully specified
        """

        if self.capsFidelity is None or self.analysisDir is None or self.currentAIM is None:
            print "Error: Unable to load AIM, either the capsFidelity, analysisDir " + \
                  "(analysis directory) and/or aim variables have not been set"
            return False
        else:
            return True

class _capsGeomety(object):
    """
    Functions to interact with caps Geometry - intiataed within capsProblem (not a standalone class)
    """

    def __init__(self, problemParent):
        self.capsProblem = problemParent

    def saveGeometry(self, filename, directory = None, extension=None):
        
        if filename is None:
            print "No filename provided"
            return 
        
        if directory is None: 
            directory = os.getcwd()
        else:
            # Check to see if directory exists
            if not os.path.isdir(directory):
                print "Directory does not currently exist while saving geometry file - it will be made automatically"
                os.makedirs(directory)
        
        filename = directory + "/" + filename
         
        if extension is None: 
            extension = ".egads"
        elif "." not in extension:
            extension = "." + extension
        
        # Check to see if file exists
        try:
            os.remove(filename+extension)
            print "Geometry file already exists - file " + filename+extension + " will be deleted"
        except OSError:
            pass    
            
        self.capsProblem.status =self.capsProblem.libCAPS.saveGeometry(c_int(self.capsProblem.problemIndex), 
                                                                       c_char_p(filename),
                                                                       c_char_p(extension))
        
        self.capsProblem.checkStatus(msg = "\twhile saving geometry")
        
    def setGeometryVal(self, varname=None, value=None):
        """
        Sets an GEOMETRYIN variable
        """

        if varname is None or value is None:
            print "No variable name and/or value provided"
            return

        var_type = c_int()
        var_len  = c_int()
        #var_units= c_char_p()

        var_dim  = c_int()
        var_nrow = c_int()
        var_ncol = c_int()
        var_lfixed = c_int()
        var_sfixed = c_int()
        var_ntype  = c_int()

        self.capsProblem.status = self.capsProblem.libCAPS.getGeometryInVarType(c_int(self.capsProblem.problemIndex),
                                                                                c_char_p(varname),
                                                                                byref(var_type),
                                                                                byref(var_len),
                                                                                byref(var_dim),
                                                                                byref(var_nrow),
                                                                                byref(var_ncol),
                                                                                byref(var_lfixed),
                                                                                byref(var_sfixed),
                                                                                byref(var_ntype))
        self.capsProblem.checkStatus(msg = "\twhile getting geometry variable - " + str(varname))

        # Parse variable types and cast python value(s) accordingly
        if var_type.value == 0: # Boolean
            if var_len.value != 1:
                print "Boolean type returned, but length doesn't equal 1"
                return
            valueLength = 1
            value = (c_int * valueLength)(*[int(value)]) # Convert to c_int array
            
        elif var_type.value == 1: # Integer

            if isinstance(value, list):
                valueLength = len(value)
            else:
                valueLength = 1
                value = [value] # Convert to list

            value = (c_int * valueLength)(*value) # Convert to c_int array

        elif var_type.value == 2: # Double

            if isinstance(value, list):
                valueLength = len(value)

            else:
                valueLength = 1
                value = [value] # Convert to list

            value = (c_double * valueLength)(*value) # Convert to c_double array

        elif var_type.value == 3: # String
            valueLength = len(value)
            value = c_char_p(value)
            
            #var_len.value is the length of the string

        else:
            print "Unsupported variable type at this time"

        if valueLength != var_nrow.value*var_ncol.value and valueLength != var_len.value:
            print "The length of the input value (" + str(valueLength) + ") does not match the length of the CAPS " + \
                  "variable with name " + varname + "(" + str(var_len.value) + ") - geometry variable lengths are FIXED!"
            self.capsProblem.status = -324
            self.capsProblem.checkStatus(msg = "\twhile setting geometry variable - " + str(varname)) 
            
        # Set value
        self.capsProblem.status = self.capsProblem.libCAPS.setGeometryInVar_VOID_P(c_int(self.capsProblem.problemIndex),
                                                                                   c_char_p(varname),
                                                                                   cast(value, c_void_p),
                                                                                   c_int(valueLength), # numRow
                                                                                   c_int(1))      # numCol
                                                                      
        self.capsProblem.checkStatus(msg = "\twhile setting geometry variable - " + str(varname))

class _capsAnalysis(object):
    """
    Functions to interact with caps Analysis - intiataed within capsProblem (not a standalone class)
    """

    def __init__(self, name, offName, directory, index, problemParent):
        self.aimName = name
        self.officialName = offName
        self.analysisDir = directory
        self.aIndex = index
        self.capsProblem = problemParent

        # Check to see if directory exists
        if not os.path.isdir(self.analysisDir):
            print "Analysis directory does not currently exist - it will be made automatically"
            os.makedirs(self.analysisDir)

    def setAnalysisVal(self, varname=None, value=None):
        """
        Sets an ANALYSISIN variable
        """

        if varname is None or value is None:
            print "No variable name and/or value provided"
            return

        var_type = c_int()
        var_len  = c_int()
        #var_units= c_char_p()
        var_dim  = c_int()
        var_nrow = c_int()
        var_ncol = c_int()
        var_lfixed = c_int()
        var_sfixed = c_int()
        var_ntype  = c_int()

        # Get ANALYSIS IN varibale type
        self.capsProblem.status = self.capsProblem.libCAPS.getAnalysisInVarType(c_int(self.capsProblem.problemIndex),
                                                                                c_char_p(varname),
                                                                                byref(var_type),
                                                                                byref(var_len),
                                                                                byref(var_dim),
                                                                                byref(var_nrow),
                                                                                byref(var_ncol),
                                                                                byref(var_lfixed),
                                                                                byref(var_sfixed),
                                                                                byref(var_ntype),
                                                                                c_int(self.aIndex))

        self.capsProblem.checkStatus(msg = "\twhile getting variable - " + str(varname))

        # Parse variable types and cast python value(s) accordingly
        if var_type.value == 0: # Boolean
            if var_len.value != 1:
                print "Boolean type returned, but length doesn't equal 1"
                return
            valueLength = 1
            value = (c_int * valueLength)(*[int(value)]) # Convert to c_int array
            
        elif var_type.value == 1: # Integer

            if isinstance(value, list):
                valueLength = len(value)
            else:
                valueLength = 1
                value = [value] # Convert to list

            value = (c_int * valueLength)(*value) # Convert to c_int array

        elif var_type.value == 2: # Double

            if isinstance(value, list):
                valueLength = len(value)

            else:
                valueLength = 1
                value = [value] # Convert to list

            value = (c_double * valueLength)(*value) # Convert to c_double array

        elif var_type.value == 3: # String
            valueLength = 1
            value = c_char_p(value)
            
            #var_len.value is the length of the string

        elif var_type.value == 4: # Tuple Pair
            if isinstance(value, list):
                valueLength = len(value)
            else:
                valueLength = 1 
                value = [value] # Convert Tuple pair to list 
    
            temp = (_capsTuple *valueLength)()
        
            for i in range(valueLength):
                temp[i].name = c_char_p(str(value[i][0]))
                
                if (isinstance(value[i][1], str) or 
                    isinstance(value[i][1], int) or 
                    isinstance(value[i][1], float)):
                    
                    temp[i].value = c_char_p(str(value[i][1]))
                else:
                    temp[i].value = c_char_p(jsonDumps(value[i][1]))
            
            value = temp

        else:
            print "Unsupported variable type at this time"

        # If not a boolean
        if var_type.value != 0: # var_len.value == 0 ->it is a NULL
            if (valueLength != var_nrow.value*var_ncol.value and valueLength != var_len.value) and var_lfixed.value == 1:
                print "The length of the input value does not match the length of the CAPS " + \
                      "variable with name " + varname + " - variable length is FIXED!"
                return    
        
        # Set value
        self.capsProblem.status = self.capsProblem.libCAPS.setAnalysisInVar_VOID_P(c_int(self.capsProblem.problemIndex),
                                                                                   c_char_p(varname),
                                                                                   cast(value, c_void_p),
                                                                                   c_int(valueLength), # numRow
                                                                                   c_int(1),      # numCol
                                                                                   c_int(self.aIndex))

        self.capsProblem.checkStatus(msg = "\twhile setting variable - " + str(varname))

    def getAnalysisOutVal(self, varname=None):
        """
        Gets an ANALYSISOUT variable
        """

        if varname is None:
            print "No variable name provided"
            return None

        var_type = c_int()
        var_len  = c_int()
        #var_units= c_char_p()
        var_dim  = c_int()
        var_nrow = c_int()
        var_ncol = c_int()
        var_lfixed = c_int()
        var_sfixed = c_int()
        var_ntype  = c_int()

        val = c_void_p()
        # Get ANALYSIS Out varibale and type
        self.capsProblem.status = self.capsProblem.libCAPS.getAnalysisOutVar_VOID_P(c_int(self.capsProblem.problemIndex),
                                                                                    c_char_p(varname),
                                                                                    byref(var_type),
                                                                                    byref(var_len),
                                                                                    byref(var_dim),
                                                                                    byref(var_nrow),
                                                                                    byref(var_ncol),
                                                                                    byref(var_lfixed),
                                                                                    byref(var_sfixed),
                                                                                    byref(var_ntype),
                                                                                    byref(val),
                                                                                    c_int(self.aIndex))

        self.capsProblem.checkStatus(msg = "\twhile getting analysis out variable - " + str(varname))

        # Initialize empty list
        valOut = []

        if var_nrow.value != 1 and var_ncol.value != 1:
            print "Returning matrices is not supported.... right now!!!"

        #print hex(addressof(val))
        # Parse variable types and cast python value(s) accordingly
        if var_type.value == 0: # Boolean
            if var_len.value != 1:
                print "Boolean type returned, but length doesn't equal 1"
                return None

            val = cast(val, POINTER(c_int)).contents # Convert to c_int array
            valOut = bool(val.value)

        elif var_type.value == 1: # Integer

            for i in range(var_len.value):
                if var_len.value == 1:
                    valOut = int(cast(val, POINTER(c_int))[i])
                else:
                    valOut.append(int(cast(val, POINTER(c_int))[i]))

        elif var_type.value == 2: # Double

            for i in range(var_len.value):
                if var_len.value == 1:
                    valOut = float(cast(val, POINTER(c_double))[i])
                else:
                    valOut.append(float(cast(val,POINTER(c_double))[i]))

        elif var_type.value == 3: # String
            print "This is unchecked for return strings"
            val = cast(val, c_char_p())

            valOut = str(val.value)
        else:
            print "Unsupported variable type at this time"

        return valOut

    def getAnalysisInfo(self):
        """ 
        Get analysis information (print to screen)
        """

        if self.capsProblem.checkAIMStatus() and self.capsProblem.checkCAPSFileStaus():
            dirty = c_int()

            self.capsProblem.status = self.capsProblem.libCAPS.getAnalysisInfo(c_int(self.capsProblem.problemIndex),
                                                                               c_int(self.aIndex), byref(dirty))

            self.capsProblem.checkStatus(msg = "\twhile getting analysis information - " + str(self.aimName))

            print "Analysis returned a cleanliness state of", dirty.value

    def aimPreAnalysis(self):
        """
        Run AIM pre-analysis
        """

        if self.capsProblem.checkAIMStatus() and self.capsProblem.checkCAPSFileStaus():

            self.capsProblem.status = self.capsProblem.libCAPS.aimPreAnalysis(c_int(self.capsProblem.problemIndex),
                                                                              c_int(self.aIndex))

            self.capsProblem.checkStatus(msg = "\twhile getting aim pre-analysis - " + str(self.aimName))

    def aimPostAnalysis(self):
        """
        Run AIM post-analysis
        """

        if self.capsProblem.checkAIMStatus() and self.capsProblem.checkCAPSFileStaus():

            self.capsProblem.status = self.capsProblem.libCAPS.aimPostAnalysis(c_int(self.capsProblem.problemIndex),
                                                                               c_int(self.aIndex))

            self.capsProblem.checkStatus(msg = "\twhile getting aim post-analysis - " + str(self.aimName))

class _capsDataTransfer(object):
    """
    Functions to interact with caps Analysis - intiataed within capsProblem (not a standalone class)
    """

    def __init__(self, transferName, varList, index, problemParent):
        
        self.transferName = transferName
        self.variable = {}
        self.dIndex = index
        self.capsProblem = problemParent
        
        j = 0
        for i in varList:
            self.variable[i] = j 
            j = j+1
            
    def executeTransfer(self, variableName = None):
        
        if variableName is None:
            print "No variable name provided - defaulting to variable - " + self.variable.keys()[0]
            variableName = self.variable.keys()[0]
      
        elif variableName not in self.variable.keys():
            print "Unreconized varible name " + str(variableName)
            return
        
        variableIndex = self.variable[variableName]
            
        self.capsProblem.status = self.capsProblem.libCAPS.getDataTransfer(c_int(self.capsProblem.problemIndex),
                                                                           c_int(self.dIndex),
                                                                           c_char_p(variableName),
                                                                           c_int(variableIndex))
        self.capsProblem.checkStatus(msg = "\twhile getting data transfer - " + str(self.transferName) + " variable -" + str(variableName))