/*
 ************************************************************************
 *                                                                      *
 * Fitter -- best-fit cubic Bspline to cloud of points                  *
 *                                                                      *
 *              Written by John Dannenhoffer @ Syracuse University      *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2012/2016  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>

#define CINT    const int
#define CDOUBLE const double
#define CCHAR   const char

#define  SUCCESS 0
#include "common.h"
#include "Fitter.h"

#define  EPS04  0.0001

#ifdef GRAFIC
    #include "grafic.h"
#endif

/***********************************************************************/
/*                                                                     */
/* macros (including those that go along with common.h)                */
/*                                                                     */
/***********************************************************************/

#ifdef DEBUG
   #define DOPEN {if (dbg_fp == NULL) dbg_fp = fopen("Fitter.dbg", "w");}
   static  FILE *dbg_fp=NULL;
#endif

//static void *realloc_temp=NULL;            /* used by RALLOC macro */

/***********************************************************************/
/*                                                                     */
/* declarations                                                        */
/*                                                                     */
/***********************************************************************/

static int    eval1dBspline(double T, int n, double cp[], double XYZ[],
                            /*@null@*/double dXYZdT[], /*@null@*/double dXYZdP[]);
static int    eval2dBspline(double U, double V, int n, double P[], double XYZ[],
                            /*@null@*/double dXYZdU[], /*@null@*/double dXYZdV[], /*@null@*/double dXYZdP[]);
static int    cubicBsplineBases(int ncp, double t, double N[], double dN[]);
static int    solveSparse(double SAv[], int SAi[], double b[], double x[],
                          int itol, double *errmax, int *iter);
static int    findIndex(int irow, int icol, int SAi[]);
static double L2norm(double f[], int n);

#ifdef GRAFIC
    static void   plotCurve_image(int*, void*, void*, void*, void*, void*,
                                  void*, void*, void*, void*, void*, float*, char*, int);
    static void   plotSurface_image(int*, void*, void*, void*, void*, void*,
                                    void*, void*, void*, void*, void*, float*, char*, int);
#endif


/*
 ************************************************************************
 *                                                                      *
 *   fit1dCloud - find spline that best-fits the cloud of points        *
 *                                                                      *
 ************************************************************************
 */
int
fit1dCloud(int    m,                    /* (in)  number of points in cloud */
           double XYZcloud[],           /* (in)  array  of points in cloud */
           int    n,                    /* (in)  number of control points */
           double cp[],                 /* (in)  array  of control points (first and last are set) */
                                        /* (out) array  of control points (all set) */
           double Tcloud[],             /* (out) T-parameters of points in cloud */
           double *normf,               /* (out) RMS of distances between cloud and fit */
           double *dotmin,              /* (out) minimum normalized dot product of control polygon */
           int    *nmin,                /* (out) minimum number of cloud points in any interval */
           FILE   *fp)                  /* (in)  file for progress outputs (or NULL) */
{
    int    status = SUCCESS;            /* (out)  return status */

    int    np, nvar, ivar, jvar, nobj, iobj, i, j, k, next, ipass;
    int    niter, iter, count, itol, maxiter;
    double alen, XYZold[3], XYZnew[3], dlen;
    double frac, dbest, dtest, toler, lambda, normdelta, normfnew;
    double T, XYZ[3], dXYZdT[3], omega;
    double dx0, dy0, dz0, dx1, dy1, dz1;

    int    *MMi=NULL, *nper=NULL;
    double *MMd=NULL, errmax;
    double *dXYZdP=NULL, *cpnew=NULL;
    double *beta=NULL, *delta=NULL, *betanew=NULL;
    double *f=NULL,  *fnew=NULL;
    double *aa=NULL, *bb=NULL, *cc=NULL, *rhs=NULL;

    ROUTINE(fit1dCloud);

    /* --------------------------------------------------------------- */

    if (fp != NULL) {
        fprintf(fp, "enter fit1dCloud(m=%d, n=%d)\n", m, n);
    }

    assert(m > 1);                      // needed to avoid clang warning
    assert(n > 2);                      // needed to avoid clang warning

    /* default return */
    *normf  = 1e-12;
    *nmin   = m;

    /* number of design variables and objectives */
    np   = 3 * n - 6;
    nvar = m + np;
    nobj = 3 * m;

    /* if m < n, then assume that the linear spline is the best fit */
    if (m < n) {
        for (j = 1; j < n-1; j++) {
            frac = (double)(j) / (double)(n-1);

            cp[3*j  ] = (1-frac) * cp[0] + frac * cp[3*n-3];
            cp[3*j+1] = (1-frac) * cp[1] + frac * cp[3*n-2];
            cp[3*j+2] = (1-frac) * cp[2] + frac * cp[3*n-1];
        }

        if (fp != NULL) {
            fprintf(fp, "making linear fit because not enough points in cloud\n");
        }
        goto cleanup;
    }

    /* allocate all temporary arrays */
    MALLOC(dXYZdP,  double,   n);
    MALLOC(cpnew,   double, 3*n);
    MALLOC(nper,    int,      n);

    MALLOC(beta,    double, nvar);
    MALLOC(delta,   double, nvar);
    MALLOC(betanew, double, nvar);

    MALLOC(f,       double, nobj);
    MALLOC(fnew,    double, nobj);

    MALLOC(aa,      double, m    );
    MALLOC(bb,      double, m *np);
    MALLOC(cc,      double, np*np);
    MALLOC(rhs,     double, nvar );

#define AA(I)       aa[(I)]
#define BB(I,J)     bb[(J)+np*(I)]
#define CC(I,J)     cc[(J)+np*(I)]

    /* make two passes: (1) linear interpolate and (2) equi-arclength */
    for (ipass = 0; ipass < 2; ipass++) {
        if (fp != NULL) {
            fprintf(fp, "Starting pass %d ...\n", ipass+1);
        }

        /* linear interpolate control points from boundries */
        if (ipass == 0) {
            for (j = 1; j < n-1; j++) {
                frac = (double)(j) / (double)(n-1);

                cp[3*j  ] = (1-frac) * cp[0] + frac * cp[3*n-3];
                cp[3*j+1] = (1-frac) * cp[1] + frac * cp[3*n-2];
                cp[3*j+2] = (1-frac) * cp[2] + frac * cp[3*n-1];
            }

        /* equi-arclength spacing of control points */
        } else {

            for (j = 0; j < n-1; j++) {
                cpnew[3*j  ] = cp[3*j  ];
                cpnew[3*j+1] = cp[3*j+1];
                cpnew[3*j+2] = cp[3*j+2];
            }

            /* compute arc-length */
            alen      = 0;
            XYZold[0] = cp[0];
            XYZold[1] = cp[1];
            XYZold[2] = cp[2];

            for (i = 1; i < 101; i++) {
                T = (n - 3) * (double)(i) / (double)(101-1);

                status = eval1dBspline(T, n, cp, XYZnew, NULL, NULL);
                CHECK_STATUS(eval1dBspline);

                alen += sqrt( (XYZnew[0] - XYZold[0]) * (XYZnew[0] - XYZold[0])
                            + (XYZnew[1] - XYZold[1]) * (XYZnew[1] - XYZold[1])
                            + (XYZnew[2] - XYZold[2]) * (XYZnew[2] - XYZold[2]));

                XYZold[0] = XYZnew[0];
                XYZold[1] = XYZnew[1];
                XYZold[2] = XYZnew[2];
            }
            dlen = alen / (n-1);

            /* initialize the control points to be equally spaced (in arclength) */
            alen      = 0;
            T         = 0;
            XYZold[0] = cp[0];
            XYZold[1] = cp[1];
            XYZold[2] = cp[2];

            for (j = 1; j < n-1; j++) {
                while (alen < dlen*j) {
                    T += (double)(n-3) / (double)(101-1);

                    status = eval1dBspline(T, n, cp, XYZnew, NULL, NULL);
                    CHECK_STATUS(eval1dBspline);

                    alen += sqrt( (XYZnew[0] - XYZold[0]) * (XYZnew[0] - XYZold[0])
                                + (XYZnew[1] - XYZold[1]) * (XYZnew[1] - XYZold[1])
                                + (XYZnew[2] - XYZold[2]) * (XYZnew[2] - XYZold[2]));

                    XYZold[0] = XYZnew[0];
                    XYZold[1] = XYZnew[1];
                    XYZold[2] = XYZnew[2];
                }

                cp[3*j  ] = XYZold[0];
                cp[3*j+1] = XYZold[1];
                cp[3*j+2] = XYZold[2];
            }
        }

        /* for each point in the cloud, assign the value of "t"
           (which is stored in the first m betas) that is associated
           with the closest control point */
        for (k = 0; k < m; k++) {
            beta[k] = 0;
            dbest   = 1e20;

            for (j = 1; j < n-1; j++) {
                dtest = (XYZcloud[3*k  ] - cp[3*j  ]) * (XYZcloud[3*k  ] - cp[3*j  ])
                      + (XYZcloud[3*k+1] - cp[3*j+1]) * (XYZcloud[3*k+1] - cp[3*j+1])
                      + (XYZcloud[3*k+2] - cp[3*j+2]) * (XYZcloud[3*k+2] - cp[3*j+2]);
                if (dtest < dbest) {
                    beta[k] = (double)(j) / (double)(n-1) * (double)(n-3);
                    dbest   = dtest;
                }
            }
        }

    /* set the relaxation parameter for control points */
        dtest = sqrt( (cp[3*n-3] - cp[0]) * (cp[3*n-3] - cp[0])
                    + (cp[3*n-2] - cp[1]) * (cp[3*n-2] - cp[1])
                    + (cp[3*n-1] - cp[2]) * (cp[3*n-1] - cp[2]));
        omega = MIN(1/dtest, 0.25);

        /* insert the interior control points into the design variables */
        next = m;
        for (j = 1; j < n-1; j++) {
            beta[next++] = cp[3*j  ];
            beta[next++] = cp[3*j+1];
            beta[next++] = cp[3*j+2];
        }

        /* compute the initial objective function */
        for (k = 0; k < m; k++) {
            status = eval1dBspline(beta[k], n, cp, XYZ, NULL, NULL);
            CHECK_STATUS(eval1dBspline);

            f[3*k  ] = XYZcloud[3*k  ] - XYZ[0];
            f[3*k+1] = XYZcloud[3*k+1] - XYZ[1];
            f[3*k+2] = XYZcloud[3*k+2] - XYZ[2];
        }
        *normf = L2norm(f, nobj) / m;
        if (fp != NULL) {
            fprintf(fp, "initial   normf= %10.4e\n", *normf);
        }

        /* initialize the Levenberg-Marquardt algorithm */
        niter  = 501;
        toler  = 1.0e-12;
        lambda = 1;

        /* LM iterations */
        for (iter = 0; iter < niter; iter++) {

            /* initialize [AA  BB]
                          [      ] =  transpose(J) * J + lambda * diag(transpose(J) * J)
                          [BB' CC]

               and        rhs  = -transpose(J) * f
            */
            for (jvar = 0; jvar < np; jvar++) {
                for (ivar = 0; ivar < np; ivar++) {
                    CC(ivar,jvar) = 0;
                }
                CC(jvar,jvar) = 1e-6;
            }

            for (jvar = 0; jvar < nvar; jvar++) {
                rhs[jvar] = 0;
            }

            /* accumulate AA, BB, CC, and rhs by looping over points in cloud */
            for (k = 0; k < m; k++) {
                status = eval1dBspline(beta[k], n, cp, XYZ, dXYZdT, dXYZdP);
                CHECK_STATUS(eval1dBspline);

                AA(k) = dXYZdT[0] * dXYZdT[0] + dXYZdT[1] * dXYZdT[1] + dXYZdT[2] * dXYZdT[2];

                for (ivar = 1; ivar < n-1; ivar++) {
                    BB(k, 3*ivar-3) = dXYZdT[0] * dXYZdP[ivar];
                    BB(k, 3*ivar-2) = dXYZdT[1] * dXYZdP[ivar];
                    BB(k, 3*ivar-1) = dXYZdT[2] * dXYZdP[ivar];

                    for (jvar = 1; jvar < n-1; jvar++) {
#ifndef __clang_analyzer__
                        CC(3*ivar-3, 3*jvar-3) += dXYZdP[ivar] * dXYZdP[jvar];
                        CC(3*ivar-2, 3*jvar-2) += dXYZdP[ivar] * dXYZdP[jvar];
                        CC(3*ivar-1, 3*jvar-1) += dXYZdP[ivar] * dXYZdP[jvar];
#endif
                    }
                }

                rhs[k] = dXYZdT[0] * f[3*k] + dXYZdT[1] * f[3*k+1] + dXYZdT[2] * f[3*k+2];

                for (ivar = 1; ivar < n-1; ivar++) {
#ifndef __clang_analyzer__
                    rhs[m+3*ivar-3] += dXYZdP[ivar] * f[3*k  ];
                    rhs[m+3*ivar-2] += dXYZdP[ivar] * f[3*k+1];
                    rhs[m+3*ivar-1] += dXYZdP[ivar] * f[3*k+2];
#endif
                }
            }

            /* set up sparse-matrix arrays */
            count = m + 2 * m * np + np * np + 1;

            MALLOC(MMd, double, count);
            MALLOC(MMi, int,    count);

            /* store diagonal values (multiplied by (1+lambda)) */
            for (k = 0; k < m; k++) {
                MMd[k] = AA(k) * (1 + lambda);
            }
            for (ivar = 0; ivar < np; ivar++) {
                MMd[m+ivar] = CC(ivar,ivar) * (1 + lambda);
            }

            /* set up off-diagonal elements, including indices */
            MMi[0] = nvar + 1;
            count  = nvar;

            /* BB to the right of AA */
            for (k = 0; k < m; k++) {
                for (jvar = 0; jvar < np; jvar++) {
                    count++;
                    MMd[count] = BB(k,jvar);
                    MMi[count] = m + jvar;
                }
                MMi[k+1] = count + 1;
            }

            for (ivar = 0; ivar < np; ivar++) {
                /* transpose(BB) below A */
                for (k = 0; k < m; k++) {
                    count++;
                    MMd[count] = BB(k,ivar);
                    MMi[count] = k;
                }

                /* CC in bottom-right corner */
                for (jvar = 0; jvar < np; jvar++) {
                    if (ivar != jvar) {
                        count++;
                        MMd[count] = CC(ivar,jvar);
                        MMi[count] = m + jvar;
                    }
                }
                MMi[m+ivar+1] = count + 1;
            }

            /* arbitrary value (not used) */
            MMd[nvar] = 0;

            /* set up for sparse matrix solve (via biconjugate gradient technique) */
            itol    = 1;
            errmax  = 1.0e-12;
            maxiter = 2 * nvar;
            for (ivar = 0; ivar < nvar; ivar++) {
                delta[ivar] = 0;
            }

            status = solveSparse(MMd, MMi, rhs, delta, itol, &errmax, &maxiter);
            CHECK_STATUS(solveSparse);

            FREE(MMd);
            FREE(MMi);

            /* check for convergence on delta (which corresponds to a small
               change in beta) */
            normdelta = L2norm(delta, nvar);

            if (normdelta < toler) {
                if (fp != NULL) {
                    fprintf(fp, "converged with norm(delta)=%10.4e\n", normdelta);
                }
                break;
            }

            /* find the temporary new beta */
            for (ivar = 0; ivar < nvar; ivar++) {

                /* beta associated with Tcloud */
                if (ivar < m) {
                    betanew[ivar] = beta[ivar] + delta[ivar];

                    if (betanew[ivar] < 0  ) betanew[ivar] = 0;
                    if (betanew[ivar] > n-3) betanew[ivar] = n-3;

                /* beta associated with control points */
                } else {
                    betanew[ivar] = beta[ivar] + omega * delta[ivar];
                }
            }

            /* extract the temporary control points from betanew */
#ifndef __clang_analyzer__
            next = m;
            for (j = 0; j < n; j++) {
                if (j == 0 || j == n-1) {
                    cpnew[3*j  ] = cp[3*j  ];
                    cpnew[3*j+1] = cp[3*j+1];
                    cpnew[3*j+2] = cp[3*j+2];
                } else {
                    cpnew[3*j  ] = betanew[next++];
                    cpnew[3*j+1] = betanew[next++];
                    cpnew[3*j+2] = betanew[next++];
                }
            }
#endif

            /* compute the objective function based upon the new beta */
            for (k = 0; k < m; k++) {
                status = eval1dBspline(betanew[k], n, cp, XYZ, NULL, NULL);
                CHECK_STATUS(eval1dBspline);

                fnew[3*k  ] = XYZcloud[3*k  ] - XYZ[0];
                fnew[3*k+1] = XYZcloud[3*k+1] - XYZ[1];
                fnew[3*k+2] = XYZcloud[3*k+2] - XYZ[2];
            }
            normfnew = L2norm(fnew, nobj) / m;
            if (fp != NULL && iter%10 == 0) {
                fprintf(fp, "iter=%4d normf= %10.4e  ", iter, normfnew);
            }

            /* if this was a better step, accept it and decrease
               lambda (making it more Newton-like) */
            if (normfnew < *normf) {
                lambda /= 2;
                if (fp != NULL && iter%10 == 0) {
                    fprintf(fp, "ACCEPTED,  lambda=%10.3e\n", lambda);
                }

                /* save new design variables, control points, and
                   objective function */
                for (ivar = 0; ivar < nvar; ivar++) {
                    beta[ivar] = betanew[ivar];
                }
#ifndef __clang_analyzer__
                for (j = 0; j < n; j++) {
                    cp[3*j  ] = cpnew[3*j  ];
                    cp[3*j+1] = cpnew[3*j+1];
                    cp[3*j+2] = cpnew[3*j+2];
                }
#endif
                for (iobj = 0; iobj < nobj; iobj++) {
                    f[iobj] = fnew[iobj];
                }
                *normf = normfnew;

                /* otherwise do not take the step and increase lambda (making it
                   more steepest-descent-like) */
            } else {
                lambda *= 2;
                if (fp != NULL && iter %100 == 0) {
                    fprintf(fp, "rejected,  lambda=%10.3e\n", lambda);
                }
            }

            /* check for convergence (based upon a small value of
               objective function) */
            if (*normf < toler) {
                if (fp != NULL) {
                    fprintf(fp, "converged with norm(f)=%10.4e\n", *normf);
                }
                break;
            }
        }
    }

    /* find the minimum number of cloud points in each spline interval */
    for (j = 0; j < n-3; j++) {
        nper[j] = 0;
    }

    for (k = 0; k < m; k++) {
        j = MIN(MAX(0,floor(beta[k])), n-4);
        nper[j]++;
    }

    for (j = 0; j < n-3; j++) {
        if (nper[j] < *nmin) {
            *nmin = nper[j];
        }
    }

    /* find the smallest included angle in control points */
    *dotmin = 1;
    for (j = 1; j < n-1; j++) {
        dx0 = cp[3*j  ] - cp[3*j-3];
        dy0 = cp[3*j+1] - cp[3*j-2];
        dz0 = cp[3*j+2] - cp[3*j-1];

        dx1 = cp[3*j+3] - cp[3*j  ];
        dy1 = cp[3*j+4] - cp[3*j+1];
        dz1 = cp[3*j+5] - cp[3*j+2];

        dtest =   (dx0 * dx1 + dy0 * dy1 + dz0 * dz1)
            / sqrt(dx0 * dx0 + dy0 * dy0 + dz0 * dz0)
            / sqrt(dx1 * dx1 + dy1 * dy1 + dz1 * dz1);

        if (dtest < *dotmin) {
            *dotmin = dtest;
        }
    }

    /* copy the T parameters into output array */
    for (k = 0; k < m; k++) {
        Tcloud[k] = beta[k];
    }

cleanup:
    FREE(dXYZdP);
    FREE(cpnew );
    FREE(nper  );

    FREE(beta   );
    FREE(delta  );
    FREE(betanew);

    FREE(f   );
    FREE(fnew);

    FREE(aa );
    FREE(bb );
    FREE(cc );
    FREE(rhs);

    FREE(MMi);
    FREE(MMd);

#undef AA
#undef BB
#undef CC

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   fit2dCloud - find spline that best-fits the cloud of points        *
 *                                                                      *
 ************************************************************************
 */
int
fit2dCloud(int    m,                    /* (in)  number of points in cloud */
           double XYZcloud[],           /* (in)  array  of points in cloud */
           int    n,                    /* (in)  number of control points in each direction */
           double cp[],                 /* (in)  array  of control points (boundaries set) */
                                        /* (out) array  of control points (all set) */
           double UVcloud[],            /* (out) UV-parameters  of points in cloud */
           double *normf,               /* (out) RMS of distances between cloud and fit */
           int    *nmin,                /* (out) minimum number of cloud points in any interval */
           FILE   *fp)                  /* (in)  file for progress outputs (or NULL) */
{
    int    status = SUCCESS;            /* (out)  return status */

    double XYZ[3], dbest, dtest, lambda, normfinit, normfold, normfnew, errmax;
    double fraci, fracj, omega;
    int    i, j, k, ii, jj, kk, iter, nvar, ivar, niter, itol, maxiter;
    int    next, next1, next2, count, irow, icol;

    double *beta=NULL, *betanew=NULL, *delta=NULL, *MMd=NULL, *RR=NULL;
    double fx, fy, fz, dXYZdU[3], dXYZdV[3], *dXYZdP=NULL, *cpnew=NULL;

    int    *MMi=NULL, *nper=NULL;

    ROUTINE(fit2dCloud);

    /* --------------------------------------------------------------- */

    if (fp != NULL) {
        fprintf(fp, "enter fit2dCloud(m=%d, n=%d)\n", m, n);
    }

    assert(m > 1);                      // needed to avoid clang warning
    assert(n > 2);                      // needed to avoid clang warning

    /* default return */
    *normf  = EPS12;
    *nmin   = m;

    /* number of design variables and objectives */
    nvar = 2 * m + 3 * (n-2) * (n-2);

    /* number of elements needed in sparse matrix storage (actually a little
       larger than needed( */
    count = 3
          + 4 * m
          + 2 * (6 * m * (n-2) * (n-2))
          + 3 * (n-2) * (n-2) * (n-2) * (n-2);

    /* allocate all temporary arrays */
    MALLOC(dXYZdP, double,   n*n);
    MALLOC(cpnew,  double, 3*n*n);
    MALLOC(nper,   int,      n*n);

    MALLOC(beta,    double, nvar);
    MALLOC(betanew, double, nvar);
    MALLOC(delta,   double, nvar);
    MALLOC(RR,      double, nvar);

    MALLOC(MMd,   double, count);
    MALLOC(MMi,   int,    count);

    /* biliner interpolate from boundaries */
    for (j = 1; j < n-1; j++) {
        for (i =  1; i < n-1; i++) {
            fraci = (double)(i) / (double)(n-1);
            fracj = (double)(j) / (double)(n-1);

            cp[3*(i+n*j)  ]
                = (1-fraci)             * cp[3*((0  )+n*(j  ))  ]
                + (  fraci)             * cp[3*((n-1)+n*(j  ))  ]
                +             (1-fracj) * cp[3*((i  )+n*(0  ))  ]
                +             (  fracj) * cp[3*((i  )+n*(n-1))  ]
                - (1-fraci) * (1-fracj) * cp[3*((0  )+n*(0  ))  ]
                - (  fraci) * (1-fracj) * cp[3*((n-1)+n*(0  ))  ]
                - (1-fraci) * (  fracj) * cp[3*((0  )+n*(n-1))  ]
                - (  fraci) * (  fracj) * cp[3*((n-1)+n*(n-1))  ];
            cp[3*(i+n*j)+1]
                = (1-fraci)             * cp[3*((0  )+n*(j  ))+1]
                + (  fraci)             * cp[3*((n-1)+n*(j  ))+1]
                +             (1-fracj) * cp[3*((i  )+n*(0  ))+1]
                +             (  fracj) * cp[3*((i  )+n*(n-1))+1]
                - (1-fraci) * (1-fracj) * cp[3*((0  )+n*(0  ))+1]
                - (  fraci) * (1-fracj) * cp[3*((n-1)+n*(0  ))+1]
                - (1-fraci) * (  fracj) * cp[3*((0  )+n*(n-1))+1]
                - (  fraci) * (  fracj) * cp[3*((n-1)+n*(n-1))+1];
            cp[3*(i+n*j)+2]
                = (1-fraci)             * cp[3*((0  )+n*(j  ))+2]
                + (  fraci)             * cp[3*((n-1)+n*(j  ))+2]
                +             (1-fracj) * cp[3*((i  )+n*(0  ))+2]
                +             (  fracj) * cp[3*((i  )+n*(n-1))+2]
                - (1-fraci) * (1-fracj) * cp[3*((0  )+n*(0  ))+2]
                - (  fraci) * (1-fracj) * cp[3*((n-1)+n*(0  ))+2]
                - (1-fraci) * (  fracj) * cp[3*((0  )+n*(n-1))+2]
                - (  fraci) * (  fracj) * cp[3*((n-1)+n*(n-1))+2];
        }
    }

    /* for each point in the cloud, assign the values of "u" and "v"
       (which are stored in the first 2*m betas) that are associated
       with the closest interior control point.  note: only interior
       control points are used since corner points can cause problems */
    for (k = 0; k < m; k++) {
        beta[2*k  ] = 0;
        beta[2*k+1] = 0;
        dbest       = 1e20;

        for (j = 1; j < n-1; j++) {
            for (i = 1; i < n-1; i++) {
                dtest = (XYZcloud[3*k  ] - cp[3*((i)+n*(j))  ]) * (XYZcloud[3*k  ] - cp[3*((i)+n*(j))  ])
                      + (XYZcloud[3*k+1] - cp[3*((i)+n*(j))+1]) * (XYZcloud[3*k+1] - cp[3*((i)+n*(j))+1])
                      + (XYZcloud[3*k+2] - cp[3*((i)+n*(j))+2]) * (XYZcloud[3*k+2] - cp[3*((i)+n*(j))+2]);
                if (dtest < dbest) {
                    beta[2*k  ] = (double)(i) / (double)(n-1) * (double)(n-3);
                    beta[2*k+1] = (double)(j) / (double)(n-1) * (double)(n-3);
                    dbest       = dtest;
                }
            }
        }
    }

    /* set the relaxation parameter for control points */
    omega = 0.25;

    /* insert the interior control points into the design variables */
    next = 2 * m;
    for (j = 1; j < n-1; j++) {
        for (i = 1; i < n-1; i++) {
            beta[next++] = cp[3*((i  )+n*(j  ))  ];
            beta[next++] = cp[3*((i  )+n*(j  ))+1];
            beta[next++] = cp[3*((i  )+n*(j  ))+2];
        }
    }

    /* compute the initial objective function */
    for (k = 0; k < m; k++) {
        status = eval2dBspline(beta[2*k], beta[2*k+1], n, cp, XYZ, NULL, NULL, NULL);
        CHECK_STATUS(eval2dBspline);

        fx = XYZcloud[3*k  ] - XYZ[0];
        fy = XYZcloud[3*k+1] - XYZ[1];
        fz = XYZcloud[3*k+2] - XYZ[2];

        *normf += fx * fx + fy * fy + fz * fz;
    }
    *normf    = sqrt(*normf) / m;
    normfinit = *normf;
    normfold  = *normf;
    if (fp != NULL) {
        fprintf(fp, "initial   normf= %10.4e\n", *normf);
    }

    /* set up the index storage for the sparse matrix
       next = index of next available location in MMi
       irow = row index
       icol = column index */
    next = nvar + 1;

    /* top of matrix */
    irow = 0;                           // needed to avoid clang warning

    for (k = 0; k < m; k++) {
        irow = 2 * k;

        MMi[irow] = next;     /* first index for row */
        MMd[irow] = 0;

        /* top left (even rows) */
        MMi[next] = 2 * k + 1; /* superdiagonal */
        MMd[next] = 0;
        next++;

        /* top right (even rows) */
        icol = 2 * m;
        for (j = 1; j < n-1; j++) {
            for (i = 1; i < n-1; i++) {
                for (kk = 0; kk < 3; kk++) {
                    MMi[next] = icol;
                    MMd[next] = 0;
                    next++;
                    icol++;
                }
            }
        }
        irow++;

        MMi[irow] = next;      /* first index for row */
        MMd[irow] = 0;

        /* top left (odd rows) */
        MMi[next] = 2 * k;      /* subdiagonal */
        MMd[next] = 0;
        next++;

        /* top left (odd rows) */
        icol = 2 * m;
        for (j = 1; j < n-1; j++) {
            for (i = 1; i < n-1; i++) {
                for (kk = 0; kk < 3; kk++) {
                    MMi[next] = icol;
                    MMd[next] = 0;
                    next++;
                    icol++;
                }
            }
        }
        irow++;
    }

    /* bottom of matrix */
    for (j = 1; j < n-1; j++) {
        for (i = 1; i < n-1; i++) {
            for (kk = 0; kk < 3; kk++) {
                MMi[irow] = next;  /* first index for row */
                MMd[irow] = 0;

                /* bottom left */
                for (k = 0; k < m; k++) {
                    MMi[next] = 2 * k;
                    MMd[next] = 0;
                    next++;

                    MMi[next] = 2 * k + 1;
                    MMd[next] = 0;
                    next++;
                }

                /* bottom right */
                icol = 2 * m + kk;
                for (jj = 1; jj < n-1; jj++) {
                    for (ii = 1; ii < n-1; ii++) {
                        if (ii != i || jj != j) {
                            MMi[next] = icol;
                            MMd[next] = 0;
                            next++;
                        }
                        icol += 3;
                    }
                }

                irow++;
            }
        }
    }

    MMi[irow] = next;          /* first index for non-existant row */
    MMd[irow] = 0;

    /* initialize the Levenberg-Marquardt algorithm */
    niter  = 501;
    lambda = 1;

    /* LM iterations */
    for (iter = 0; iter <= niter; iter++) {

        /* initialize RR */
        for (j = 0; j < nvar; j++) {
            RR[j] = 0;
        }

        /* build MM =  transpose(J) * J   (MM is stored in MMd and MMi)
           and   RR = -transpose(j) * f  */
        *normf = 0;
        for (k = 0; k < m; k++) {
            status = eval2dBspline(beta[2*k], beta[2*k+1], n, cp, XYZ, dXYZdU, dXYZdV, dXYZdP);
            CHECK_STATUS(eval2dBspline);

            fx = XYZcloud[3*k  ] - XYZ[0];
            fy = XYZcloud[3*k+1] - XYZ[1];
            fz = XYZcloud[3*k+2] - XYZ[2];

            *normf += fx * fx + fy * fy + fz * fz;

            RR[           2*k               ] = fx        * dXYZdU[0] + fy        * dXYZdU[1] + fz        * dXYZdU[2];
            MMd[findIndex(2*k  , 2*k  , MMi)] = dXYZdU[0] * dXYZdU[0] + dXYZdU[1] * dXYZdU[1] + dXYZdU[2] * dXYZdU[2];
            MMd[findIndex(2*k  , 2*k+1, MMi)] = dXYZdV[0] * dXYZdU[0] + dXYZdV[1] * dXYZdU[1] + dXYZdV[2] * dXYZdU[2];

            RR[           2*k+1             ] = fx        * dXYZdV[0] + fy        * dXYZdV[1] + fz        * dXYZdV[2];
            MMd[findIndex(2*k+1, 2*k  , MMi)] = dXYZdU[0] * dXYZdV[0] + dXYZdU[1] * dXYZdV[1] + dXYZdU[2] * dXYZdV[2];
            MMd[findIndex(2*k+1, 2*k+1, MMi)] = dXYZdV[0] * dXYZdV[0] + dXYZdV[1] * dXYZdV[1] + dXYZdV[2] * dXYZdV[2];

            next1 = 2 * m;
            for (j = 1; j < n-1; j++) {
                for (i = 1; i < n-1; i++) {

                    RR[           next1             ] += fx        * dXYZdP[(i  )+n*(j  )];
                    MMd[findIndex(2*k,   next1, MMi)]  = dXYZdU[0] * dXYZdP[(i  )+n*(j  )];
                    MMd[findIndex(next1, 2*k,   MMi)]  = dXYZdU[0] * dXYZdP[(i  )+n*(j  )];
                    MMd[findIndex(2*k+1, next1, MMi)]  = dXYZdV[0] * dXYZdP[(i  )+n*(j  )];
                    MMd[findIndex(next1, 2*k+1, MMi)]  = dXYZdV[0] * dXYZdP[(i  )+n*(j  )];

                    next2 = 2 * m;
                    for (jj = 1; jj < n-1; jj++) {
                        for (ii = 1; ii < n-1; ii++) {
                            MMd[findIndex(next1, next2, MMi)] += dXYZdP[(i  )+n*(j  )] * dXYZdP[ii+jj*n];
                            next2 += 3;
                        }
                    }
                    next1++;

                    RR[           next1             ] += fy        * dXYZdP[(i  )+n*(j  )];
                    MMd[findIndex(2*k,   next1, MMi)]  = dXYZdU[1] * dXYZdP[(i  )+n*(j  )];
                    MMd[findIndex(next1, 2*k,   MMi)]  = dXYZdU[1] * dXYZdP[(i  )+n*(j  )];
                    MMd[findIndex(2*k+1, next1, MMi)]  = dXYZdV[1] * dXYZdP[(i  )+n*(j  )];
                    MMd[findIndex(next1, 2*k+1, MMi)]  = dXYZdV[1] * dXYZdP[(i  )+n*(j  )];

                    next2 = 2 * m + 1;
                    for (jj = 1; jj < n-1; jj++) {
                        for (ii = 1; ii < n-1; ii++) {
                            MMd[findIndex(next1, next2, MMi)] += dXYZdP[(i  )+n*(j  )] * dXYZdP[ii+jj*n];
                            next2 += 3;
                        }
                    }
                    next1++;

                    RR[           next1             ] += fz        * dXYZdP[(i  )+n*(j  )];
                    MMd[findIndex(2*k,   next1, MMi)]  = dXYZdU[2] * dXYZdP[(i  )+n*(j  )];
                    MMd[findIndex(next1, 2*k,   MMi)]  = dXYZdU[2] * dXYZdP[(i  )+n*(j  )];
                    MMd[findIndex(2*k+1, next1, MMi)]  = dXYZdV[2] * dXYZdP[(i  )+n*(j  )];
                    MMd[findIndex(next1, 2*k+1, MMi)]  = dXYZdV[2] * dXYZdP[(i  )+n*(j  )];

                    next2 = 2 * m + 2;
                    for (jj = 1; jj < n-1; jj++) {
                        for (ii = 1; ii < n-1; ii++) {
                            MMd[findIndex(next1, next2, MMi)] += dXYZdP[(i  )+n*(j  )] * dXYZdP[ii+jj*n];
                            next2 += 3;
                        }
                    }
                    next1++;
                }
            }
        }

        *normf = sqrt(*normf) / m;

        /* multiply diagonal elements of MM by (1 + lambda) */
        for (i = 0; i < nvar; i++) {
            MMd[i] *= (1 + lambda);
        }

        /* find the minimum number of cloud points in each spline interval */
        for (j = 0; j < n-3; j++) {
            for (i = 0; i < n-3; i++) {
                nper[i+n*j] = 0;
            }
        }

        for (k = 0; k < m; k++) {
            i = MIN(MAX(0,floor(beta[2*k+1])), n-4);
            j = MIN(MAX(0,floor(beta[2*k  ])), n-4);
            nper[i+n*j]++;
        }

        *nmin = m;
        for (j = 0; j < n-3; j++) {
            for ( i = 0; i < n-3; i++) {
                if (nper[i+n*j] < *nmin) {
                    *nmin = nper[i+n*j];
                }
            }
        }

        /* copy the UV parameters into output array */
        for (k = 0; k < m; k++) {
            UVcloud[2*k  ] = beta[2*k  ];
            UVcloud[2*k+1] = beta[2*k+1];
        }

        /* check convergence */
        if (*normf < EPS06) {
            if (fp != NULL) {
                fprintf(fp, "converged since *normf=%12.4e < %12.4e\n", *normf, EPS06);
            }
            break;
        }

        /* check that progress is being made (after the first 25 iterations) */
        if (fabs(*normf-normfold) < normfinit*EPS04 &&
            fabs(*normf-normfold) > 0               && iter >= 25) {
            if (fp != NULL) {
                fprintf(fp, "not making progress with delta(normf)=%12.4e\n",
                        fabs(*normf-normfold)/normfinit);
            }
            break;
        }
        normfold = *normf;

        /* check for running out of iterations */
        if (iter == niter) {
            if (fp != NULL) {
                fprintf(fp, "ran out of iterations with *normf=%12.4e\n", *normf);
            }
            break;
        }

        /* solve matrix equation */
        itol    = 1;
        errmax  = EPS12;
        maxiter = 2 * nvar;
        for (i = 0; i < nvar; i++) {
            delta[i] = 0;
        }

        status = solveSparse(MMd, MMi, RR, delta, itol, &errmax, &maxiter);
        CHECK_STATUS(solveSparse);

        if (L2norm(delta, nvar) < EPS12) {
            if (fp != NULL) {
                fprintf(fp, "L2norm(delta) = %12.4e < EPS12\n", L2norm(delta, nvar));
            }
            break;
        }

        /* find the temporary new beta */
        for (ivar = 0; ivar < nvar; ivar++) {

            /* beta associated with UVcloud */
            if (ivar < 2*m) {
                betanew[ivar] = beta[ivar] + delta[ivar];

                if (betanew[ivar] < 0  ) betanew[ivar] = 0;
                if (betanew[ivar] > n-3) betanew[ivar] = n-3;

            /* beta associated with control points */
            } else {
                betanew[ivar] = beta[ivar] + omega * delta[ivar];
            }
        }

        /* extract the temporary spline points from betanew */
        next = 2 * m;
        for (j = 0; j < n; j++) {
            for (i = 0; i < n; i++) {
                if (i == 0 || i == n-1 || j == 0 || j == n-1) {
                    cpnew[3*((i  )+n*(j  ))  ] = cp[3*((i  )+n*(j  ))  ];
                    cpnew[3*((i  )+n*(j  ))+1] = cp[3*((i  )+n*(j  ))+1];
                    cpnew[3*((i  )+n*(j  ))+2] = cp[3*((i  )+n*(j  ))+2];
                } else {
                    cpnew[3*((i  )+n*(j  ))  ] = betanew[next++];
                    cpnew[3*((i  )+n*(j  ))+1] = betanew[next++];
                    cpnew[3*((i  )+n*(j  ))+2] = betanew[next++];
                }
            }
        }

        /* compute the objective function based upon the new beta */
        normfnew = 0;
        for (k = 0; k < m; k++) {
            status = eval2dBspline(betanew[2*k], betanew[2*k+1], n, cpnew, XYZ, NULL, NULL, NULL);
            CHECK_STATUS(eval2dBspline);

            fx = XYZcloud[3*k  ] - XYZ[0];
            fy = XYZcloud[3*k+1] - XYZ[1];
            fz = XYZcloud[3*k+2] - XYZ[2];

            normfnew += fx * fx + fy * fy + fz * fz;
        }
        normfnew = sqrt(normfnew) / m;
        if (fp != NULL && iter%10 == 0) {
            fprintf(fp, "iter=%4d normf= %10.4e  ", iter, normfnew);
        }

        /* if this was a better step, accept it and decrease
           lambda (making it more Bewton-like) */
        if (normfnew < *normf) {
            lambda /= 2;
            if (fp != NULL && iter%10 == 0) {
                fprintf(fp, "ACCEPTED,  lambda=%10.3e\n", lambda);
            }

            /* save new design variables, spline points, and
               objective function */
            for (i = 0; i < nvar; i++) {
                beta[i] = betanew[i];
            }
            for (j = 0; j < n; j++) {
                for (i = 0; i < n; i++) {
                    cp[3*((i  )+n*(j  ))  ] = cpnew[3*((i  )+n*(j  ))  ];
                    cp[3*((i  )+n*(j  ))+1] = cpnew[3*((i  )+n*(j  ))+1];
                    cp[3*((i  )+n*(j  ))+2] = cpnew[3*((i  )+n*(j  ))+2];

                    cp[3*((i  )+n*(j  ))  ] = cpnew[3*((i  )+n*(j  ))  ];
                    cp[3*((i  )+n*(j  ))+1] = cpnew[3*((i  )+n*(j  ))+1];
                    cp[3*((i  )+n*(j  ))+2] = cpnew[3*((i  )+n*(j  ))+2];

                    cp[3*((i  )+n*(j  ))  ] = cpnew[3*((i  )+n*(j  ))  ];
                    cp[3*((i  )+n*(j  ))+1] = cpnew[3*((i  )+n*(j  ))+1];
                    cp[3*((i  )+n*(j  ))+2] = cpnew[3*((i  )+n*(j  ))+2];

                    cp[3*((i  )+n*(j  ))  ] = cpnew[3*((i  )+n*(j  ))  ];
                    cp[3*((i  )+n*(j  ))+1] = cpnew[3*((i  )+n*(j  ))+1];
                    cp[3*((i  )+n*(j  ))+2] = cpnew[3*((i  )+n*(j  ))+2];
                }
            }
            *normf = normfnew;

        /* otherwise do not take the step and increase lambda (making it
           more steepest-descent-like) */
        } else {
            lambda *= 2;
            if (fp != NULL && iter%10 == 0) {
                fprintf(fp, "rejected,  lambda=%10.3e\n", lambda);
            }

            if (lambda > 1000) break;
        }
    }

    /* copy the U and V parameters into output arrays */
    for (k = 0; k < m; k++) {
        UVcloud[2*k  ] = beta[2*k  ];
        UVcloud[2*k+1] = beta[2*k+1];
    }

cleanup:
    FREE(beta   );
    FREE(betanew);
    FREE(delta  );
    FREE(RR     );

    FREE(MMd);
    FREE(MMi);

    FREE(dXYZdP );
    FREE(cpnew);
    FREE(nper );

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   plotCurve - plot curve data                                        *
 *                                                                      *
 ************************************************************************
 */
#ifdef GRAFIC
int
plotCurve(int    m,                     /* (in)  number of points in cloud */
          double XYZcloud[],            /* (in)  array  of points in cloud */
/*@null@*/double Tcloud[],              /* (in)  T-parameters of points in cloud (or NULL) */
          int    n,                     /* (in)  number of control points */
          double cp[],                  /* (in)  array  of control points */
          double normf,                 /* (in)  RMS of distances between cloud and fit */
          double dotmin,                /* (in)  minimum normalized dot product of control polygon */
          int    nmin)                  /* (in)  minimum number of control points in any interval */
{
    int    status = SUCCESS;            /* (out) return status */

    int    indgr=1+2+4+16+64, itype=0;
    char   pltitl[255];

    ROUTINE(plotCurve);

    /* --------------------------------------------------------------- */

    sprintf(pltitl, "~x~y~ m=%d,  n=%d,  normf=%.7f,  dotmin=%.4f,  nmin=%d",
            m, n, normf, dotmin, nmin);

    grctrl_(plotCurve_image, &indgr, pltitl,
            (void*)(&itype),
            (void*)(&m),
            (void*)(XYZcloud),
            (void*)(Tcloud),
            (void*)(&n),
            (void*)(cp),
            (void*)NULL,
            (void*)NULL,
            (void*)NULL,
            (void*)NULL,
            strlen(pltitl));

//cleanup:
    return status;
}
#endif


/*
 ************************************************************************
 *                                                                      *
 *   plotSurface - plot surface data                                    *
 *                                                                      *
 ************************************************************************
 */
#ifdef GRAFIC
int
plotSurface(int    m,                   /* (in)  number of points in cloud */
            double XYZcloud[],          /* (in)  array  of points in cloud */
  /*@null@*/double UVcloud[],           /* (in)  UV-parameters of points in cloud (or NULL) */
            int    n,                   /* (in)  number of control points in each direction */
            double cp[],                /* (in)  array  of control points */
            double normf,               /* (in)  RMS of distances between cloud and fit */
            int    nmin)                /* (in)  minimum number of cloud points in any interval */
{
    int    status = SUCCESS;            /* (out) return status */

    int    indgr=1+4+16+64;
    char   pltitl[255];

    ROUTINE(plotSurface);

    /* --------------------------------------------------------------- */

    sprintf(pltitl, "~x~y~ m=%d,  n=%d,  normf=%.7f,  nmin=%d",
            m, n, normf, nmin);

    grctrl_(plotSurface_image, &indgr, pltitl,
            (void*)(&m),
            (void*)(XYZcloud),
            (void*)(UVcloud),
            (void*)(&n),
            (void*)(cp),
            (void*)NULL,
            (void*)NULL,
            (void*)NULL,
            (void*)NULL,
            (void*)NULL,
            strlen(pltitl));

//cleanup:
    return status;
}
#endif


/*
 ************************************************************************
 *                                                                      *
 *   eval1dBspline - evaluate cubic Bspline and its derivatives         *
 *                                                                      *
 ************************************************************************
 */
static int
eval1dBspline(double T,                 /* (in)  independent variable */
              int    n,                 /* (in)  number of control points */
              double cp[],              /* (in)  array  of control points */
              double XYZ[],             /* (out) dependent variables */
    /*@null@*/double dXYZdT[],          /* (out) derivative wrt T (or NULL) */
    /*@null@*/double dXYZdP[])          /* (out) derivative wrt P (or NULL) */
{
    int    status = SUCCESS;            /* (out) return status */

    int    i, span;
    double N[4], dN[4];

    ROUTINE(eval1dBspline);

    /* --------------------------------------------------------------- */

    assert (n > 3);

    XYZ[0] = 0;
    XYZ[1] = 0;
    XYZ[2] = 0;

    /* set up the Bspline bases */
    status = cubicBsplineBases(n, T, N, dN);
    CHECK_STATUS(cubicBsplineBases);

    span = MIN(floor(T), n-4);

    /* find the dependent variable */
    for (i = 0; i < 4; i++) {
        XYZ[0] += N[i] * cp[3*(i+span)  ];
        XYZ[1] += N[i] * cp[3*(i+span)+1];
        XYZ[2] += N[i] * cp[3*(i+span)+2];
    }

    /* find the deriviative wrt T */
    if (dXYZdT != NULL) {
        dXYZdT[0] = 0;
        dXYZdT[1] = 0;
        dXYZdT[2] = 0;

        for (i = 0; i < 4; i++) {
            dXYZdT[0] += dN[i] * cp[3*(i+span)  ];
            dXYZdT[1] += dN[i] * cp[3*(i+span)+1];
            dXYZdT[2] += dN[i] * cp[3*(i+span)+2];
        }
    }

    /* find the derivative wrt P */
    if (dXYZdP != NULL) {
        for (i = 0; i < n; i++) {
            dXYZdP[i] = 0;
        }
        for (i = 0; i < 4; i++) {
            dXYZdP[i+span] += N[i];
        }
    }

cleanup:
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   eval2dBspline - evaluate Spline and its derivative                 *
 *                                                                      *
 ************************************************************************
 */
static int
eval2dBspline(double U,                 /* (in)  first  independent variable */
              double V,                 /* (in)  second independent variable */
              int    n,                 /* (in)  number of control points */
              double cp[],              /* (in)  array  of control points */
              double XYZ[],             /* (out) dependent variable */
    /*@null@*/double dXYZdU[],          /* (out) derivative wrt U (or NULL) */
    /*@null@*/double dXYZdV[],          /* (out) derivative wrt V (or NULL) */
    /*@null@*/double dXYZdP[])          /* (out) derivative wrt P (or NULL) */
{
    int    status = SUCCESS;            /* (out) return status */

    int    i, j, spanu, spanv;
    double Nu[4], dNu[4], Nv[4], dNv[4];

    ROUTINE(eval2dBspline);

    /* --------------------------------------------------------------- */

    assert(n > 3);

    XYZ[0] = 0;
    XYZ[1] = 0;
    XYZ[2] = 0;

    /* set up the Bspline bases */
    status = cubicBsplineBases(n, U, Nu, dNu);
    CHECK_STATUS(cubicBsplineBases);

    status = cubicBsplineBases(n, V, Nv, dNv);
    CHECK_STATUS(cubicBsplineBases);

    spanu = MIN(floor(U), n-4);
    spanv = MIN(floor(V), n-4);

    /* find the dependent variable */
    for (j = 0; j < 4; j++) {
        for (i = 0; i < 4; i++) {
            XYZ[0] += Nu[i] * Nv[j] * cp[3*((i+spanu)+n*(j+spanv))  ];
            XYZ[1] += Nu[i] * Nv[j] * cp[3*((i+spanu)+n*(j+spanv))+1];
            XYZ[2] += Nu[i] * Nv[j] * cp[3*((i+spanu)+n*(j+spanv))+2];
        }
    }

    /* find the derivative wrt U */
    if (dXYZdU != NULL) {
        dXYZdU[0] = 0;
        dXYZdU[1] = 0;
        dXYZdU[2] = 0;

        for (j = 0; j < 4; j++) {
            for (i = 0; i < 4; i++) {
                dXYZdU[0] += dNu[i] * Nv[j] * cp[3*((i+spanu)+n*(j+spanv))  ];
                dXYZdU[1] += dNu[i] * Nv[j] * cp[3*((i+spanu)+n*(j+spanv))+1];
                dXYZdU[2] += dNu[i] * Nv[j] * cp[3*((i+spanu)+n*(j+spanv))+2];
            }
        }
    }

    /* find the derivative wrt V */
    if (dXYZdV != NULL) {
        dXYZdV[0] = 0;
        dXYZdV[1] = 0;
        dXYZdV[2] = 0;

        for (j = 0; j < 4; j++) {
            for (i = 0; i < 4; i++) {
                dXYZdV[0] += Nu[i] * dNv[j] * cp[3*((i+spanu)+n*(j+spanv))  ];
                dXYZdV[1] += Nu[i] * dNv[j] * cp[3*((i+spanu)+n*(j+spanv))+1];
                dXYZdV[2] += Nu[i] * dNv[j] * cp[3*((i+spanu)+n*(j+spanv))+2];
            }
        }
    }

    /* find the derivative wrt P */
    if (dXYZdP != NULL) {
        for (j = 0; j < n; j++) {
            for (i = 0; i < n; i++) {
                dXYZdP[(i)+n*(j)] = 0;
            }
        }

        for (j = 0; j < 4; j++) {
            for (i = 0; i < 4; i++) {
                dXYZdP[(i+spanu)+n*(j+spanv)] += Nu[i] * Nv[j];
            }
        }
    }

cleanup:
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   cubicBsplineBases - basis function values for cubic Bspline        *
 *                                                                      *
 ************************************************************************
 */
static int
cubicBsplineBases(int    ncp,           /* (in)  number of control points */
                  double T,             /* (in)  independent variable (0<=T<=(ncp-3) */
                  double N[],           /* (out) bases */
                  double dN[])          /* (out) d(bases)/d(T) */
{
    int       status = SUCCESS;         /* (out) return status */

    int      i, r, span;
    double   saved, dsaved, num, dnum, den, dden, temp, dtemp;
    double   left[4], dleft[4], rite[4], drite[4];

    ROUTINE(cubicBsplineBases);

    /* --------------------------------------------------------------- */

    span = MIN(floor(T)+3, ncp-1);

    N[ 0] = 1.0;
    dN[0] = 0;

    for (i = 1; i <= 3; i++) {
        left[ i] = T - MAX(0, span-2-i);
        dleft[i] = 1;

        rite[ i] = MIN(ncp-3,span-3+i) - T;
        drite[i] =                     - 1;

        saved  = 0;
        dsaved = 0;

        for (r = 0; r < i; r++) {
            num   = N[ r];
            dnum  = dN[r];

            den   = rite[ r+1] + left[ i-r];
            dden  = drite[r+1] + dleft[i-r];

            temp  = num / den;
            dtemp = (dnum * den - dden * num) / den / den;

            N[ r] = saved  + rite[ r+1] * temp;
            dN[r] = dsaved + drite[r+1] * temp + rite[r+1] * dtemp;

            saved  = left[ i-r] * temp;
            dsaved = dleft[i-r] * temp + left[i-r] * dtemp;
        }

        N[ i] = saved;
        dN[i] = dsaved;
    }

//cleanup:
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   solveSparse - solve: A * x = b  using biconjugate gradient method  *
 *                                                                      *
 ************************************************************************
 */
static int
solveSparse(double SAv[],               /* (in)  sparse array values */
            int    SAi[],               /* (in)  sparse array indices */
            double b[],                 /* (in)  rhs vector */
            double x[],                 /* (in)  guessed result vector */
                                        /* (out) result vector */
            int    itol,                /* (in)  stopping criterion */
            double *errmax,             /* (in)  convergence tolerance */
                                        /* (out) estimated error at convergence */
            int    *iter)               /* (in)  maximum number of iterations */
                                        /* (out) number of iterations taken */
{
    int    status = SUCCESS;            /* (out) return status */

    int    n, i, j, k, itmax;
    double tol, ak, akden, bk, bknum, bkden, bnorm, dxnorm, xnorm, znorm_old, znorm=0;
    double *p=NULL, *pp=NULL, *r=NULL, *rr=NULL, *z=NULL, *zz=NULL;

    ROUTINE(solveSparse);

    /* --------------------------------------------------------------- */

    tol   = *errmax;
    itmax = *iter;

    n = SAi[0] - 1;

    MALLOC(p,  double, n);
    MALLOC(pp, double, n);
    MALLOC(r,  double, n);
    MALLOC(rr, double, n);
    MALLOC(z,  double, n);
    MALLOC(zz, double, n);

    /* make sure none of the diagonals are very small */
    for (i = 0; i < n; i++) {
        if (fabs(SAv[i]) < 1.0e-14) {
            printf("ERROR:: cannot solve since SAv[%d]=%10.3e\n", i, SAv[i]);
            status = -1;
            goto cleanup;
        }
    }

    /* calculate initial residual */
    *iter = 0;

    /* r = A * x  */
    for (i = 0; i < n; i++) {
        r[i] = SAv[i] * x[i];

        for (k = SAi[i]; k < SAi[i+1]; k++) {
            r[i] += SAv[k] * x[SAi[k]];
        }
    }

    for (j = 0; j < n; j++) {
        r[ j] = b[j] - r[j];
        rr[j] = r[j];
    }

    if (itol == 1) {
        bnorm = L2norm(b, n);

        for (j = 0; j < n; j++) {
            z[j] = r[j] / SAv[j];
        }
    } else if (itol == 2) {
        for (j = 0; j < n; j++) {
            z[j] = b[j] / SAv[j];
        }

        bnorm = L2norm(z, n);

        for (j = 0; j < n; j++) {
            z[j] = r[j] / SAv[j];
        }
    } else {
        for (j = 0; j < n; j++) {
            z[j] = b[j] / SAv[j];
        }

        bnorm = L2norm(z, n);

        for (j = 0; j < n; j++) {
            z[j] = r[j] / SAv[j];
        }

        znorm = L2norm(z, n);
    }

    /* main iteration loop */
    for (*iter = 0; *iter < itmax; (*iter)++) {

        for (j = 0; j < n; j++) {
            zz[j] = rr[j] / SAv[j];
        }

        /* calculate coefficient bk and direction vectors p and pp */
        bknum = 0;
        for (j = 0; j < n; j++) {
            bknum += z[j] * rr[j];
        }

        if (*iter == 0) {
            for (j = 0; j < n; j++) {
                p[ j] = z[ j];
                pp[j] = zz[j];
            }
        } else {
            bk = bknum / bkden;

            for (j = 0; j < n; j++) {
                p[ j] = bk * p[ j] + z[ j];
                pp[j] = bk * pp[j] + zz[j];
            }
        }

        /* calculate coefficient ak, new iterate x, and new residuals r and rr */
        bkden = bknum;

        /* z = A * p  */
        for (i = 0; i < n; i++) {
            z[i] = SAv[i] * p[i];

            for (k = SAi[i]; k < SAi[i+1]; k++) {
                z[i] += SAv[k] * p[SAi[k]];
            }
        }

        akden = 0;
        for (j = 0; j < n; j++) {
            akden += z[j] * pp[j];
        }

        ak = bknum / akden;

        /* zz = transpose(A) * pp  */
        for (i = 0; i < n; i++) {
            zz[i] = SAv[i] * pp[i];
        }

        for (i = 0; i < n; i++) {
            for (k = SAi[i]; k < SAi[i+1]; k++) {
                j = SAi[k];
                zz[j] += SAv[k] * pp[i];
            }
        }

        for (j = 0; j < n; j++) {
            x[ j] += ak * p[ j];
            r[ j] -= ak * z[ j];
            rr[j] -= ak * zz[j];
        }

        /* solve Abar * z = r */
        for (j = 0; j < n; j++) {
            z[j] = r[j] / SAv[j];
        }

        /* compute and check stopping criterion */
        if (itol == 1) {
            *errmax = L2norm(r, n) / bnorm;
        } else if (itol == 2) {
            *errmax = L2norm(z, n) / bnorm;
        } else {
            znorm_old = znorm;
            znorm = L2norm(z, n);
            if (fabs(znorm_old-znorm) > (1.0e-14)*znorm) {
                dxnorm = fabs(ak) * L2norm(p, n);
                *errmax  = znorm / fabs(znorm_old-znorm) * dxnorm;
            } else {
                *errmax = znorm / bnorm;
                continue;
            }

            xnorm = L2norm(x, n);
            if (*errmax <= xnorm/2) {
                *errmax /= xnorm;
            } else {
                *errmax = znorm / bnorm;
                continue;
            }
        }

        /* exit if converged */
        if (*errmax <= tol) break;
    }

cleanup:
    FREE(p );
    FREE(pp);
    FREE(r );
    FREE(rr);
    FREE(z );
    FREE(zz);

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   findIndex - find index in sparse matrix storage (or -1)            *
 *                                                                      *
 ************************************************************************
 */
static int
findIndex(int    irow,                  /* (in)  row index */
          int    icol,                  /* (in)  column index */
          int    SAi[])                 /* (in)  index array */
{
    int index = -1;                     /* (out) index in SAv (or -1 if error) */

    int    nrow, ileft, irite, imid, count;

    ROUTINE(findIndex);

    /* --------------------------------------------------------------- */

    /* number of rows and columns */
    nrow = SAi[0] - 1;

    /* check for valid inputs */
    if (irow < 0 || irow >= nrow || icol < 0 || icol >= nrow) goto cleanup;

    /* diagonal element */
    if (irow == icol) {
        index = irow;
        goto cleanup;
    }

    /* off-diagonal element */
    ileft = SAi[irow  ];
    irite = SAi[irow+1]-1;
    imid  = (ileft + irite) / 2;        // needed to avoid clang warning

    if        (SAi[ileft] == icol) {
        index = ileft;
        goto cleanup;
    } else if (SAi[irite] == icol) {
        index = irite;
        goto cleanup;
    }

    /* binary search to find index between ileft and irite */
    for (count = 0; count < nrow; count++) {
        imid = (ileft + irite) / 2;
        if (SAi[imid] == icol) {
            index = imid;
            goto cleanup;
        } else if (SAi[imid] < icol) {
            ileft = imid;
        } else {
            irite = imid;
        }
    }

    /* not found */
    printf("irow=%3d, icol=%3d, ileft=%3d, irite=%3d, imid=%3d, SAi=%d\n",
           irow, icol, ileft, irite, imid, SAi[imid]);
    exit(0);

cleanup:
    if (index < 0) {
        printf("irow=%d, icol=%d, nrow=%d\n", irow, icol, nrow);
        exit(0);
    }

    return index;
}


/*
 ************************************************************************
 *                                                                      *
 *   L2norm - L2-norm of vector                                         *
 *                                                                      *
 ************************************************************************
 */
static double
L2norm(double f[],                      /* (in)  vector */
       int    n)                        /* (in)  length of vector */
{
    double L2norm;                      /* (out) L2-norm */

    int    i;

    ROUTINE(L2norm);

    /* --------------------------------------------------------------- */

    /* L2-norm */
    L2norm = 0;

    for (i = 0; i < n; i++) {
        L2norm += f[i] * f[i];
    }

    L2norm = sqrt(L2norm);

//cleanup:
    return L2norm;
}


/*
 ************************************************************************
 *                                                                      *
 *   plotCurve_image - plot curve data (level 3)                        *
 *                                                                      *
 ************************************************************************
 */
#ifdef GRAFIC
static void
plotCurve_image(int   *ifunct,
                void  *itypeP,
                void  *mP,
                void  *XYZcloudP,
                void  *TcloudP,
                void  *nP,
                void  *cpP,
    /*@unused@*/void  *a6,
    /*@unused@*/void  *a7,
    /*@unused@*/void  *a8,
    /*@unused@*/void  *a9,
                float *scale,
                char  *text,
                int   textlen)
{
    int    *itype    = (int    *)itypeP;
    int    *m        = (int    *)mP;
    double *XYZcloud = (double *)XYZcloudP;
    double *Tcloud   = (double *)TcloudP;
    int    *n        = (int    *)nP;
    double *cp       = (double *)cpP;

    int    i, k, status;
    float  x4, y4, z4;
    double xmin, xmax, ymin, ymax, zmin, zmax;
    double TT, XYZ[3];

    int    icircle = GR_CIRCLE;
    int    istar   = GR_STAR;
    int    isolid  = GR_SOLID;
    int    idotted = GR_DOTTED;
    int    igreen  = GR_GREEN;
    int    iblue   = GR_BLUE;
    int    ired    = GR_RED;
    int    iblack  = GR_BLACK;

    ROUTINE(plot1dBspine_image);

    /* --------------------------------------------------------------- */

    if (*ifunct == 0) {
        xmin = XYZcloud[0];
        xmax = XYZcloud[0];
        ymin = XYZcloud[1];
        ymax = XYZcloud[1];
        zmin = XYZcloud[2];
        zmax = XYZcloud[2];

        for (k = 1; k < *m; k++) {
            if (XYZcloud[3*k  ] < xmin) xmin = XYZcloud[3*k  ];
            if (XYZcloud[3*k  ] > xmax) xmax = XYZcloud[3*k  ];
            if (XYZcloud[3*k+1] < ymin) ymin = XYZcloud[3*k+1];
            if (XYZcloud[3*k+1] > ymax) ymax = XYZcloud[3*k+1];
            if (XYZcloud[3*k+2] < zmin) zmin = XYZcloud[3*k+2];
            if (XYZcloud[3*k+2] > zmax) zmax = XYZcloud[3*k+2];
        }

        if        (xmax-xmin >= zmax-zmin && ymax-ymin >= zmax-zmin) {
            *itype = 0;
            scale[0] = xmin - EPS06;
            scale[1] = xmax + EPS06;
            scale[2] = ymin - EPS06;
            scale[3] = ymax + EPS06;
        } else if (ymax-ymin >= xmax-xmin && zmax-zmin >= xmax-xmin) {
            *itype = 1;
            scale[0] = ymin - EPS06;
            scale[1] = ymax + EPS06;
            scale[2] = zmin - EPS06;
            scale[3] = zmax + EPS06;
        } else {
            *itype = 2;
            scale[0] = zmin - EPS06;
            scale[1] = zmax + EPS06;
            scale[2] = xmin - EPS06;
            scale[3] = xmax + EPS06;
        }

        strcpy(text, " ");

    } else if (*ifunct == 1) {

        /* cloud of points */
        grcolr_(&igreen);

        for (k = 0; k < *m; k++) {
            x4 = XYZcloud[3*k  ];
            y4 = XYZcloud[3*k+1];
            z4 = XYZcloud[3*k+2];
            if (*itype == 0) {
                grmov3_(&x4, &y4, &z4);
            } else if (*itype == 1) {
                grmov3_(&y4, &z4, &x4);
            } else {
                grmov3_(&z4, &x4, &y4);
            }
            grsymb_(&icircle);
        }

        /* control points */
        grcolr_(&iblue);
        grdash_(&idotted);

        x4 = cp[0];
        y4 = cp[1];
        z4 = cp[2];
        if (*itype == 0) {
            grmov3_(&x4, &y4, &z4);
        } else if (*itype == 1) {
            grmov3_(&y4, &z4, &x4);
        } else {
            grmov3_(&z4, &x4, &y4);
        }
        grsymb_(&istar);

        for (i = 1; i < *n; i++) {
            x4 = cp[3*i  ];
            y4 = cp[3*i+1];
            z4 = cp[3*i+2];
            if (*itype == 0) {
                grdrw3_(&x4, &y4, &z4);
            } else if (*itype == 1) {
                grdrw3_(&y4, &z4, &x4);
            } else {
                grdrw3_(&z4, &x4, &y4);
            }
            grsymb_(&istar);
        }

        /* Bspline curve */
        grcolr_(&iblack);
        grdash_(&isolid);

        x4 = cp[0];
        y4 = cp[1];
        z4 = cp[2];
        if (*itype == 0) {
            grmov3_(&x4, &y4, &z4);
        } else if (*itype == 1) {
            grmov3_(&y4, &z4, &x4);
        } else {
            grmov3_(&z4, &x4, &y4);
        }

        for (i = 1; i < 201; i++) {
            TT = (*n-3) * (double)(i) / (double)(201-1);

            status = eval1dBspline(TT, *n, cp, XYZ, NULL, NULL);
            if (status < SUCCESS) printf("eval1dBspline -> status=%d\n", status);

            x4 = XYZ[0];
            y4 = XYZ[1];
            z4 = XYZ[2];
            if (*itype == 0) {
                grdrw3_(&x4, &y4, &z4);
            } else if (*itype == 1) {
                grdrw3_(&y4, &z4, &x4);
            } else {
                grdrw3_(&z4, &x4, &y4);
            }
        }

        /* distance from point to curve */
        if (Tcloud != NULL) {
            grcolr_(&ired);

            for (k = 0; k < *m; k++) {
                x4 = XYZcloud[3*k  ];
                y4 = XYZcloud[3*k+1];
                z4 = XYZcloud[3*k+2];
                if (*itype == 0) {
                    grmov3_(&x4, &y4, &z4);
                } else if (*itype == 1) {
                    grmov3_(&y4, &z4, &x4);
                } else {
                    grmov3_(&z4, &x4, &y4);
                }

                status = eval1dBspline(Tcloud[k], *n, cp, XYZ, NULL, NULL);
                if (status < SUCCESS) printf("eval1dBspline -> status=%d\n", status);

                x4 = XYZ[0];
                y4 = XYZ[1];
                z4 = XYZ[2];
                if (*itype == 0) {
                    grdrw3_(&x4, &y4, &z4);
                } else if (*itype == 1) {
                    grdrw3_(&y4, &z4, &x4);
                } else {
                    grdrw3_(&z4, &x4, &y4);
                }
            }
        }

        grcolr_(&iblack);

    } else {
        printf("illegal option\n");
    }
}
#endif


/*
 ************************************************************************
 *                                                                      *
 *   plotSurface_image - plot surface data (level 3)                    *
 *                                                                      *
 ************************************************************************
 */
#ifdef GRAFIC
static void
plotSurface_image(int    *ifunct,
                  void   *mP,
                  void   *XYZcloudP,
                  void   *UVcloudP,
                  void   *nP,
                  void   *cpP,
                  void   *a5,
                  void   *a6,
                  void   *a7,
                  void   *a8,
                  void   *a9,
                  float  *scale,
                  char   *text,
                  int    textlen)
{
    int      *m        = (int    *) mP;
    double   *XYZcloud = (double *) XYZcloudP;
    double   *UVcloud  = (double *) UVcloudP;
    int      *n        = (int    *) nP;
    double   *cp       = (double *) cpP;

    int      status = SUCCESS;

    int      i, j, k;
    float    x4, y4, z4;
    double   xmin, xmax, ymin, ymax, UV[2], XYZ[3];

    int      icircle  = GR_CIRCLE;

    int      ired     = GR_RED;
    int      igreen   = GR_GREEN;
    int      iblue    = GR_BLUE;
    int      iyellow  = GR_YELLOW;
    int      iblack   = GR_BLACK;

    int      isolid   = GR_SOLID;
    int      idotted  = GR_DOTTED;

    ROUTINE(plotSurface_image);

    /* --------------------------------------------------------------- */

    /* return scales */
    if (*ifunct == 0) {
        xmin = XYZcloud[0];
        xmax = XYZcloud[0];
        ymin = XYZcloud[1];
        ymax = XYZcloud[1];

        for (k = 0; k < *m; k++) {
            if (XYZcloud[3*k  ] < xmin) xmin = XYZcloud[3*k  ];
            if (XYZcloud[3*k  ] > xmax) xmax = XYZcloud[3*k  ];
            if (XYZcloud[3*k+1] < ymin) ymin = XYZcloud[3*k+1];
            if (XYZcloud[3*k+1] > ymax) ymax = XYZcloud[3*k+1];
        }

        scale[0] = xmin;
        scale[1] = xmax;
        scale[2] = ymin;
        scale[3] = ymax;

        text[0] = '\0';


    /* plot image */
    } else if (*ifunct == 1) {

        /* points in cloud */
        grcolr_(&igreen);
        for (k = 0; k < *m; k++) {
            x4 = XYZcloud[3*k  ];
            y4 = XYZcloud[3*k+1];
            z4 = XYZcloud[3*k+2];

            grmov3_(&x4, &y4, &z4);
            grsymb_(&icircle);
        }

        /* draw the control net */
        grcolr_(&iblue);
        grdash_(&idotted);

        for (j = 0; j < *n; j++) {

            x4 = cp[3*((0)+(*n)*(j))  ];
            y4 = cp[3*((0)+(*n)*(j))+1];
            z4 = cp[3*((0)+(*n)*(j))+2];
            grmov3_(&x4, &y4, &z4);

            for (i = 1; i < *n; i++) {
                x4 = cp[3*((i)+(*n)*(j))  ];
                y4 = cp[3*((i)+(*n)*(j))+1];
                z4 = cp[3*((i)+(*n)*(j))+2];
                grdrw3_(&x4, &y4, &z4);
            }
        }

        for (i = 0; i < *n; i++) {
            grcolr_(&iblack);

            x4 = cp[3*((i)+(*n)*(0))  ];
            y4 = cp[3*((i)+(*n)*(0))+1];
            z4 = cp[3*((i)+(*n)*(0))+2];
            grmov3_(&x4, &y4, &z4);

            for (j = 1; j < *n; j++) {
                x4 = cp[3*((i)+(*n)*(j))  ];
                y4 = cp[3*((i)+(*n)*(j))+1];
                z4 = cp[3*((i)+(*n)*(j))+2];
                grdrw3_(&x4, &y4, &z4);
            }
        }

        /* Bspline surface */
        grcolr_(&iyellow);
        grdash_(&isolid);

        for (j = 0; j < 21; j++) {
            UV[1] = (*n-3) * (double)(j) / (double)(21-1);

            status= eval2dBspline(0, UV[1], *n, cp, XYZ, NULL, NULL, NULL);
            if (status < SUCCESS) printf("eval2dBspline -> status=%d\n", status);

            x4 = XYZ[0];
            y4 = XYZ[1];
            z4 = XYZ[2];
            grmov3_(&x4, &y4, &z4);

            for (i = 1; i < 21; i++) {
                UV[0] = (*n-3) * (double)(i) / (double)(21-1);

                status= eval2dBspline(UV[0], UV[1], *n, cp, XYZ, NULL, NULL, NULL);
                if (status < SUCCESS) printf("eval2dBspline -> status=%d\n", status);

                x4 = XYZ[0];
                y4 = XYZ[1];
                z4 = XYZ[2];
                grdrw3_(&x4, &y4, &z4);
            }
        }

        for (i = 0; i < 21; i++) {
            UV[0] = (*n-3) * (double)(i) / (double)(21-1);

            status= eval2dBspline(UV[0], 0, *n, cp, XYZ, NULL, NULL, NULL);
            if (status < SUCCESS) printf("eval2dBspline -> status=%d\n", status);

            x4 = XYZ[0];
            y4 = XYZ[1];
            z4 = XYZ[2];
            grmov3_(&x4, &y4, &z4);

            for (j = 1; j < 21; j++) {
                UV[1] = (*n-3) * (double)(j) / (double)(21-1);

                status= eval2dBspline(UV[0], UV[1], *n, cp, XYZ, NULL, NULL, NULL);
                if (status < SUCCESS) printf("eval2dBspline -> status=%d\n", status);

                x4 = XYZ[0];
                y4 = XYZ[1];
                z4 = XYZ[2];
                grdrw3_(&x4, &y4, &z4);
            }
        }

        /* distance from point to surface */
        if (UVcloud != NULL) {
            grcolr_(&ired);

            for (k = 0; k < *m; k++) {
                x4 = XYZcloud[3*k  ];
                y4 = XYZcloud[3*k+1];
                z4 = XYZcloud[3*k+2];
                grmov3_(&x4, &y4, &z4);

                status = eval2dBspline(UVcloud[2*k], UVcloud[2*k+1], *n, cp, XYZ, NULL, NULL, NULL);
                if (status < SUCCESS) printf("eval2dBspline -> status=%d\n", status);

                x4 = XYZ[0];
                y4 = XYZ[1];
                z4 = XYZ[2];
                grdrw3_(&x4, &y4, &z4);
            }
        }

        grcolr_(&iblack);
    } else {
        printf("illegal option\n");
    }

//cleanup:
}
#endif

