/*
 ************************************************************************
 *                                                                      *
 * Slugs.c -- server for Static Legacy Unstructured Geometry System     *
 *                                                                      *
 *              Written by John Dannenhoffer @ Syracuse University      *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2013/2016  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <unistd.h>
#include <assert.h>

#ifdef WIN32
    #define WIN32_LEAN_AND_MEAN
    #include <windows.h>
    #include <winsock2.h>
#endif

#define CINT    const int
#define CDOUBLE const double
#define CCHAR   const char

#ifdef GRAFIC
    #include "grafic.h"
#endif

#ifdef WIN32
    #define snprintf _snprintf
#endif

#define STRNCPY(A, B, LEN) strncpy(A, B, LEN); A[LEN-1] = '\0';

#include "egads.h"
#include "common.h"
#include "Fitter.h"
#include "Tessellate.h"
#include "RedBlackTree.h"

#include "wsserver.h"
#include "emp.h"

/*
 ***********************************************************************
 *                                                                     *
 * macros (including those that go along with common.h)                *
 *                                                                     *
 ***********************************************************************
 */
#ifdef DEBUG
   #define DOPEN {if (dbg_fp == NULL) dbg_fp = fopen("Slugs.dbg", "w");}
   static  FILE *dbg_fp=NULL;
#endif

static void *realloc_temp=NULL;            /* used by RALLOC macro */

#define  RED(COLOR)      (float)(COLOR / 0x10000        ) / (float)(255)
#define  GREEN(COLOR)    (float)(COLOR / 0x00100 % 0x100) / (float)(255)
#define  BLUE(COLOR)     (float)(COLOR           % 0x100) / (float)(255)

#define MAX_EXPR_LEN     128
#define MAX_STR_LEN    32767

#define MIN3(A,B,C)   (MIN(MIN(A,B),C))
#define MAX3(A,B,C)   (MAX(MAX(A,B),C))

#define UINT32 unsigned int
#define UINT16 unsigned short int
#define REAL32 float

/*
 ***********************************************************************
 *                                                                     *
 * structures                                                          *
 *                                                                     *
 ***********************************************************************
 */
#ifdef FOO
static float color_map[256*3] =
{ 0.0000,    0.0000,    1.0000,     0.0000,    0.0157,    1.0000,
  0.0000,    0.0314,    1.0000,     0.0000,    0.0471,    1.0000,
  0.0000,    0.0627,    1.0000,     0.0000,    0.0784,    1.0000,
  0.0000,    0.0941,    1.0000,     0.0000,    0.1098,    1.0000,
  0.0000,    0.1255,    1.0000,     0.0000,    0.1412,    1.0000,
  0.0000,    0.1569,    1.0000,     0.0000,    0.1725,    1.0000,
  0.0000,    0.1882,    1.0000,     0.0000,    0.2039,    1.0000,
  0.0000,    0.2196,    1.0000,     0.0000,    0.2353,    1.0000,
  0.0000,    0.2510,    1.0000,     0.0000,    0.2667,    1.0000,
  0.0000,    0.2824,    1.0000,     0.0000,    0.2980,    1.0000,
  0.0000,    0.3137,    1.0000,     0.0000,    0.3294,    1.0000,
  0.0000,    0.3451,    1.0000,     0.0000,    0.3608,    1.0000,
  0.0000,    0.3765,    1.0000,     0.0000,    0.3922,    1.0000,
  0.0000,    0.4078,    1.0000,     0.0000,    0.4235,    1.0000,
  0.0000,    0.4392,    1.0000,     0.0000,    0.4549,    1.0000,
  0.0000,    0.4706,    1.0000,     0.0000,    0.4863,    1.0000,
  0.0000,    0.5020,    1.0000,     0.0000,    0.5176,    1.0000,
  0.0000,    0.5333,    1.0000,     0.0000,    0.5490,    1.0000,
  0.0000,    0.5647,    1.0000,     0.0000,    0.5804,    1.0000,
  0.0000,    0.5961,    1.0000,     0.0000,    0.6118,    1.0000,
  0.0000,    0.6275,    1.0000,     0.0000,    0.6431,    1.0000,
  0.0000,    0.6588,    1.0000,     0.0000,    0.6745,    1.0000,
  0.0000,    0.6902,    1.0000,     0.0000,    0.7059,    1.0000,
  0.0000,    0.7216,    1.0000,     0.0000,    0.7373,    1.0000,
  0.0000,    0.7529,    1.0000,     0.0000,    0.7686,    1.0000,
  0.0000,    0.7843,    1.0000,     0.0000,    0.8000,    1.0000,
  0.0000,    0.8157,    1.0000,     0.0000,    0.8314,    1.0000,
  0.0000,    0.8471,    1.0000,     0.0000,    0.8627,    1.0000,
  0.0000,    0.8784,    1.0000,     0.0000,    0.8941,    1.0000,
  0.0000,    0.9098,    1.0000,     0.0000,    0.9255,    1.0000,
  0.0000,    0.9412,    1.0000,     0.0000,    0.9569,    1.0000,
  0.0000,    0.9725,    1.0000,     0.0000,    0.9882,    1.0000,
  0.0000,    1.0000,    0.9961,     0.0000,    1.0000,    0.9804,
  0.0000,    1.0000,    0.9647,     0.0000,    1.0000,    0.9490,
  0.0000,    1.0000,    0.9333,     0.0000,    1.0000,    0.9176,
  0.0000,    1.0000,    0.9020,     0.0000,    1.0000,    0.8863,
  0.0000,    1.0000,    0.8706,     0.0000,    1.0000,    0.8549,
  0.0000,    1.0000,    0.8392,     0.0000,    1.0000,    0.8235,
  0.0000,    1.0000,    0.8078,     0.0000,    1.0000,    0.7922,
  0.0000,    1.0000,    0.7765,     0.0000,    1.0000,    0.7608,
  0.0000,    1.0000,    0.7451,     0.0000,    1.0000,    0.7294,
  0.0000,    1.0000,    0.7137,     0.0000,    1.0000,    0.6980,
  0.0000,    1.0000,    0.6824,     0.0000,    1.0000,    0.6667,
  0.0000,    1.0000,    0.6510,     0.0000,    1.0000,    0.6353,
  0.0000,    1.0000,    0.6196,     0.0000,    1.0000,    0.6039,
  0.0000,    1.0000,    0.5882,     0.0000,    1.0000,    0.5725,
  0.0000,    1.0000,    0.5569,     0.0000,    1.0000,    0.5412,
  0.0000,    1.0000,    0.5255,     0.0000,    1.0000,    0.5098,
  0.0000,    1.0000,    0.4941,     0.0000,    1.0000,    0.4784,
  0.0000,    1.0000,    0.4627,     0.0000,    1.0000,    0.4471,
  0.0000,    1.0000,    0.4314,     0.0000,    1.0000,    0.4157,
  0.0000,    1.0000,    0.4000,     0.0000,    1.0000,    0.3843,
  0.0000,    1.0000,    0.3686,     0.0000,    1.0000,    0.3529,
  0.0000,    1.0000,    0.3373,     0.0000,    1.0000,    0.3216,
  0.0000,    1.0000,    0.3059,     0.0000,    1.0000,    0.2902,
  0.0000,    1.0000,    0.2745,     0.0000,    1.0000,    0.2588,
  0.0000,    1.0000,    0.2431,     0.0000,    1.0000,    0.2275,
  0.0000,    1.0000,    0.2118,     0.0000,    1.0000,    0.1961,
  0.0000,    1.0000,    0.1804,     0.0000,    1.0000,    0.1647,
  0.0000,    1.0000,    0.1490,     0.0000,    1.0000,    0.1333,
  0.0000,    1.0000,    0.1176,     0.0000,    1.0000,    0.1020,
  0.0000,    1.0000,    0.0863,     0.0000,    1.0000,    0.0706,
  0.0000,    1.0000,    0.0549,     0.0000,    1.0000,    0.0392,
  0.0000,    1.0000,    0.0235,     0.0000,    1.0000,    0.0078,
  0.0078,    1.0000,    0.0000,     0.0235,    1.0000,    0.0000,
  0.0392,    1.0000,    0.0000,     0.0549,    1.0000,    0.0000,
  0.0706,    1.0000,    0.0000,     0.0863,    1.0000,    0.0000,
  0.1020,    1.0000,    0.0000,     0.1176,    1.0000,    0.0000,
  0.1333,    1.0000,    0.0000,     0.1490,    1.0000,    0.0000,
  0.1647,    1.0000,    0.0000,     0.1804,    1.0000,    0.0000,
  0.1961,    1.0000,    0.0000,     0.2118,    1.0000,    0.0000,
  0.2275,    1.0000,    0.0000,     0.2431,    1.0000,    0.0000,
  0.2588,    1.0000,    0.0000,     0.2745,    1.0000,    0.0000,
  0.2902,    1.0000,    0.0000,     0.3059,    1.0000,    0.0000,
  0.3216,    1.0000,    0.0000,     0.3373,    1.0000,    0.0000,
  0.3529,    1.0000,    0.0000,     0.3686,    1.0000,    0.0000,
  0.3843,    1.0000,    0.0000,     0.4000,    1.0000,    0.0000,
  0.4157,    1.0000,    0.0000,     0.4314,    1.0000,    0.0000,
  0.4471,    1.0000,    0.0000,     0.4627,    1.0000,    0.0000,
  0.4784,    1.0000,    0.0000,     0.4941,    1.0000,    0.0000,
  0.5098,    1.0000,    0.0000,     0.5255,    1.0000,    0.0000,
  0.5412,    1.0000,    0.0000,     0.5569,    1.0000,    0.0000,
  0.5725,    1.0000,    0.0000,     0.5882,    1.0000,    0.0000,
  0.6039,    1.0000,    0.0000,     0.6196,    1.0000,    0.0000,
  0.6353,    1.0000,    0.0000,     0.6510,    1.0000,    0.0000,
  0.6667,    1.0000,    0.0000,     0.6824,    1.0000,    0.0000,
  0.6980,    1.0000,    0.0000,     0.7137,    1.0000,    0.0000,
  0.7294,    1.0000,    0.0000,     0.7451,    1.0000,    0.0000,
  0.7608,    1.0000,    0.0000,     0.7765,    1.0000,    0.0000,
  0.7922,    1.0000,    0.0000,     0.8078,    1.0000,    0.0000,
  0.8235,    1.0000,    0.0000,     0.8392,    1.0000,    0.0000,
  0.8549,    1.0000,    0.0000,     0.8706,    1.0000,    0.0000,
  0.8863,    1.0000,    0.0000,     0.9020,    1.0000,    0.0000,
  0.9176,    1.0000,    0.0000,     0.9333,    1.0000,    0.0000,
  0.9490,    1.0000,    0.0000,     0.9647,    1.0000,    0.0000,
  0.9804,    1.0000,    0.0000,     0.9961,    1.0000,    0.0000,
  1.0000,    0.9882,    0.0000,     1.0000,    0.9725,    0.0000,
  1.0000,    0.9569,    0.0000,     1.0000,    0.9412,    0.0000,
  1.0000,    0.9255,    0.0000,     1.0000,    0.9098,    0.0000,
  1.0000,    0.8941,    0.0000,     1.0000,    0.8784,    0.0000,
  1.0000,    0.8627,    0.0000,     1.0000,    0.8471,    0.0000,
  1.0000,    0.8314,    0.0000,     1.0000,    0.8157,    0.0000,
  1.0000,    0.8000,    0.0000,     1.0000,    0.7843,    0.0000,
  1.0000,    0.7686,    0.0000,     1.0000,    0.7529,    0.0000,
  1.0000,    0.7373,    0.0000,     1.0000,    0.7216,    0.0000,
  1.0000,    0.7059,    0.0000,     1.0000,    0.6902,    0.0000,
  1.0000,    0.6745,    0.0000,     1.0000,    0.6588,    0.0000,
  1.0000,    0.6431,    0.0000,     1.0000,    0.6275,    0.0000,
  1.0000,    0.6118,    0.0000,     1.0000,    0.5961,    0.0000,
  1.0000,    0.5804,    0.0000,     1.0000,    0.5647,    0.0000,
  1.0000,    0.5490,    0.0000,     1.0000,    0.5333,    0.0000,
  1.0000,    0.5176,    0.0000,     1.0000,    0.5020,    0.0000,
  1.0000,    0.4863,    0.0000,     1.0000,    0.4706,    0.0000,
  1.0000,    0.4549,    0.0000,     1.0000,    0.4392,    0.0000,
  1.0000,    0.4235,    0.0000,     1.0000,    0.4078,    0.0000,
  1.0000,    0.3922,    0.0000,     1.0000,    0.3765,    0.0000,
  1.0000,    0.3608,    0.0000,     1.0000,    0.3451,    0.0000,
  1.0000,    0.3294,    0.0000,     1.0000,    0.3137,    0.0000,
  1.0000,    0.2980,    0.0000,     1.0000,    0.2824,    0.0000,
  1.0000,    0.2667,    0.0000,     1.0000,    0.2510,    0.0000,
  1.0000,    0.2353,    0.0000,     1.0000,    0.2196,    0.0000,
  1.0000,    0.2039,    0.0000,     1.0000,    0.1882,    0.0000,
  1.0000,    0.1725,    0.0000,     1.0000,    0.1569,    0.0000,
  1.0000,    0.1412,    0.0000,     1.0000,    0.1255,    0.0000,
  1.0000,    0.1098,    0.0000,     1.0000,    0.0941,    0.0000,
  1.0000,    0.0784,    0.0000,     1.0000,    0.0627,    0.0000,
  1.0000,    0.0471,    0.0000,     1.0000,    0.0314,    0.0000,
  1.0000,    0.0157,    0.0000,     1.0000,    0.0000,    0.0000 };
#endif

static float lims[2]  = {-2.0, +2.0};

/* data structures */
typedef struct {
    int                  ipnt;               /* Point index (bias-0) */
    int                  nedg;               /* number of incident Edges */
    double               x;                  /* x-coordinate */
    double               y;                  /* y-coordinate */
    double               z;                  /* z-coordinate */
    ego                  enode;              /* ego object */
} node_T;

typedef struct {
    int                  ibeg;               /* Node at beginning (bias-1) */
    int                  iend;               /* Node at end       (bias-1)*/
    int                  ileft;              /* Face on left      (bias-1) */
    int                  irite;              /* Face on rite      (bias-1)*/
    int                  npnt;               /* number of Points          along Edge */
    int                  *pnt;               /* array  of Point indices   along Edge */
    double               *xyz;               /* array  of Point locations along Edge */
    int                  ncp;                /* number of control points */
    double               *cp;                /* array  of control points */
    ego                  eedge;              /* ego object */
} edge_T;

typedef struct {
    int                  icol;               /* color  of associated Triangles */
    tess_T               tess;               /* TESS object */
    int                  nedg;               /* number of associated Edges */
    int                  *edg;               /* array  of associated Edges (signed, bias-1) */
    int                  npnt;               /* number of interior Points in Face */
    double               *xyz;               /* array  of interior Points in Face */
    int                  ncp;                /* number of control points */
    double               *cp;                /* array  of control points */
    int                  done;               /* =1 when done */
    ego                  eface;              /* ego object */
} face_T;

typedef struct {
    int                  ibeg;               /* Point index at beg     (bias-0) */
    int                  iend;               /* Point index at end     (bias-0) */
    int                  prev;               /* previous Segment or -1 (bias-0) */
    int                  next;               /* next     Segment or -1 (bias-0) */
} sgmt_T;

typedef struct {
    void      *mutex;                        /* the mutex or NULL for single thread */
    long      master;                        /* master thread ID */
} emp_T;

/*
 ***********************************************************************
 *                                                                     *
 * global variables                                                    *
 *                                                                     *
 ***********************************************************************
 */
/* global variables associated with tessellation, ... */
static   tess_T          tess;               /* global TESS object */
static   tess_T          tess_undo;          /* undo copy of TESS object */

static   int             Mnode = 0;          /* maximum   Nodes */
static   int             Nnode = 0;          /* number of Nodes */
static   node_T          *node = NULL;       /* array  of Nodes */
static   int             Medge = 0;          /* maximum   Edges */
static   int             Nedge = 0;          /* number of Edges */
static   edge_T          *edge = NULL;       /* array  of Edges */
static   int             Mface = 0;          /* maximum   Faces */
static   int             Nface = 0;          /* number of Faces */
static   face_T          *face = NULL;       /* array  of Faces */

static int               outLevel = 1;       /* default output level */

/* global variables associated with graphical user interface (gui) */
static wvContext         *cntxt;             /* context for the WebViewer */
static int               port   = 7681;      /* port number */
static int               batch = 0;          /* =0 to enable visualization */
static float             focus[4];

/* global variables associated with scene graph meta-data */
#define MAX_METADATA_LEN 32000
static char              *sgFocusData = NULL;

/* global variables associated with scene graph */
static int               Tris_pend    =   0;  /* =1 if awaiting update */

/* global variables associated with CurPt */
static int               CurPt_index  = -1;  /* index (or -1 for none) */
static int               CurPt_gprim  = -1;  /* GPrim (or -1 for none) */
static int               CurPt_pend   =  0;  /* =1 if awaiting update */

/* global variables associated with Hangs */
static int               Hangs_gprim  = -1;  /* GPrim (or -1 for none) */
static int               Hangs_pend    = 0;  /* =1 if awaiting update */

/* global variables associated with Links */
static int               Links_gprim  = -1;  /* GPrim (or -1 for none) */
static int               Links_pend    = 0;  /* =1 if awaiting update */

/* global variables associated with the response buffer */
static int               max_resp_len = 0;
static int               response_len = 0;
static char              *response    = NULL;

/* global variables associated with journals */
static FILE              *jrnl_out = NULL;   /* output journal file */
static char              *filename = NULL;

/*
 ***********************************************************************
 *                                                                     *
 * declarations for routines defined below                             *
 *                                                                     *
 ***********************************************************************
 */

static int        buildCurPt();
static int        buildHangs();
static int        buildLinks();
static int        buildTriangles();

static int        generateBrep();
static int        generateEgads(char filename[]);
static int        generateFits(int ncp);
static void       empFit2dCloud(void *struc);

       void       browserMessage(void *wsi, char *text, /*@unused@*/ int lena);
static void       processMessage(char *text);
static int        getToken(char *text, int nskip, char *token);
static int        closestPoint(double xloc, double yloc, double zloc);
static int        closestTriangle(double xloc, double yloc, double zloc);
static int        storeUndo();
static void       setColor(int rgb, float color[]);
#ifdef FOO
static void       spec_col(float scalar, float lims[], float out[]);
#endif
       void       printEgo(ego obj);

/*
 ***********************************************************************
 *                                                                     *
 *   Main program                                                      *
 *                                                                     *
 ***********************************************************************
 */
int
main(int       argc,                /* (in)  number of arguments */
     char      *argv[])             /* (in)  array of arguments */
{

    int       status, status2, i, bias, showUsage=0, ihand;
    int       iedge, iface;
    float     fov, zNear, zFar;
    float     eye[3]    = {0.0, 0.0, 7.0};
    float     center[3] = {0.0, 0.0, 0.0};
    float     up[3]     = {0.0, 1.0, 0.0};
    char      *casename=NULL, *jrnlname=NULL, *tempname=NULL;
    char      *text=NULL, *wv_start, test[80];
    void      *temp;
    FILE      *jrnl_in=NULL, *fp=NULL;

#ifdef GRAFIC
    int      io_kbd=5, io_scr=6;
    char     pltitl[255];
#endif

    ROUTINE(MAIN);

    /* --------------------------------------------------------------- */

    /* dynamically allocated array so that everything is on heap (not stack) */
    MALLOC(sgFocusData, char, 256        );
    MALLOC(filename,    char, 257        );
    MALLOC(casename,    char, 257        );
    MALLOC(jrnlname,    char, 257        );
    MALLOC(tempname,    char, 257        );
    MALLOC(text,        char, MAX_STR_LEN);

    DPRINT0("starting Slugs");

    /* get the flags and casename(s) from the command line */
    casename[0] = '\0';
    jrnlname[0] = '\0';

    for (i = 1; i < argc; i++) {
        if        (strcmp(argv[i], "-port") == 0) {
            if (i < argc-1) {
                sscanf(argv[++i], "%d", &port);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-jrnl") == 0) {
            if (i < argc-1) {
                STRNCPY(jrnlname, argv[++i], 257);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-outLevel") == 0) {
            if (i < argc-1) {
                sscanf(argv[++i], "%d", &outLevel);
                if (outLevel < 0) outLevel = 0;
                if (outLevel > 3) outLevel = 3;
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-batch") == 0) {
            batch = 1;
        } else if (strlen(casename) == 0) {
            STRNCPY(casename, argv[i], 257);
        } else {
            SPRINT0(0, "two casenames given");
            showUsage = 1;
            break;
        }
    }

    if (showUsage) {
        SPRINT0(0, "proper usage: 'Slugs [-port X] [-jrnl jrnlname] [-outLevel X] [-batch] [casename[.stl]]'");
        SPRINT0(0, "STOPPING...\a");
        exit(0);
    }

    /* check size of various types used within .stl files */
    if        (sizeof(UINT16) != 2) {
        SPRINT0(0, "ERROR:: uint16 should have size 2");
        exit(0);
    } else if (sizeof(UINT32) != 4) {
        SPRINT0(0, "ERROR:: uint32 should have size 4");
        exit(0);
    } else if (sizeof(REAL32) != 4) {
        SPRINT0(0, "ERROR:: real32 should have size 4");
        exit(0);
    }

    /* welcome banner */
    SPRINT0(1, "**********************************************************");
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "*                    Program Slugs                       *");
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "*        written by John Dannenhoffer, 2013/2016         *");
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "**********************************************************");

#ifdef GRAFIC
    /* initialize the grafics */
    sprintf(pltitl, "Program Fitter2D");
    grinit_(&io_kbd, &io_scr, pltitl, strlen(pltitl));
#endif

    /* allocate nominal response buffer */
    max_resp_len = 4096;
    MALLOC(response, char, max_resp_len);

    /* add .stl to filename if not present */
    if (strlen(casename) > 0) {
        STRNCPY(filename, casename, 257);
        if (strstr(casename, ".stl") == NULL && strstr(casename, ".tri") == NULL) {
            strcat(filename, ".stl");
        }
    } else {
        SPRINT0(0, "ERROR:: a casename must be given");
        exit(0);
    }

    /* read the .stl or .tri file */
    if (strstr(filename, ".tri") != NULL) {
        status = readTriAscii(&tess, filename);
        SPRINT1(3, "--> readTriAscii 0> status=%d", status);

    } else if (strstr(filename, ".stl") != NULL) {
        fp = fopen(filename, "rb");
        if (fp == NULL) {
            SPRINT1(0, "ERROR:: \"%s\" does not exist", filename);
            exit(0);
        }

        (void) fread(test, sizeof(char), 5, fp);
        fclose(fp);

        if (strncmp(test, "solid", 5) == 0) {
            SPRINT1(1, "--> \"%s\" is an ASCII file", filename);
            status = readStlAscii(&tess, filename);
            SPRINT1(3, "--> readStlAscii -> status=%d", status);
        } else {
            SPRINT1(1, "--> \"%s\" is a binary file", filename);
            status = readStlBinary(&tess, filename);
            SPRINT1(3, "--> readStlBinary -> status=%d", status);
        }
    } else {
        SPRINT1(0, "ERROR:: \"%s\" is not a .stl or .tri file", filename);
        exit(0);
    }

    /* make links between the colors */
    if (tess.ncolr > 0) {
        status = makeLinks(&tess);
        SPRINT1(3, "--> makeLinks -> status=%d", status);
    }

    /* report initial statistics */
    SPRINT0(2, "==> initialization complete");
    SPRINT1(2, "    npnt  = %d", tess.npnt );
    SPRINT1(2, "    ntri  = %d", tess.ntri );
    SPRINT1(2, "    ncolr = %d", tess.ncolr);
    SPRINT1(2, "    nhang = %d", tess.nhang);
    SPRINT1(2, "    nlink = %d", tess.nlink);

    /* initialize the undo copy */
    status = initialTess(&tess_undo);
    SPRINT1(3, "initialTess -> status=%d", status);

    /* open the output journal file */
    snprintf(tempname, 257, "port%d.jrnl", port);
    jrnl_out = fopen(tempname, "w");

    /* initialize the scene graph meta data */
    if (batch == 0) {
        sgFocusData[0] = '\0';
    }

    /* create the WebViewer context */
    if (batch == 0) {
        bias  =  0;
        fov   = 30.0;
        zNear =  1.0;
        zFar  = 10.0;
        cntxt = wv_createContext(bias, fov, zNear, zFar, eye, center, up);
        if (cntxt == NULL) {
            SPRINT0(0, "ERROR:: failed to create wvContext");
            exit(0);
        }
    }

    /* build the initial Scene Graph */
    if (batch == 0) {
        status = buildTriangles();
        SPRINT1(3, "--> buildTriangles -> status=%d", status);

        status = buildHangs();
        SPRINT1(3, "--> buildHangs -> status=%d", status);

        status = buildLinks();
        SPRINT1(3, "--> buildLinks -> status=%d", status);
    }

    /* process the input journal file if jrnlname exists */
    if (strlen(jrnlname) > 0) {
        SPRINT1(0, "==> Opening input journal file \"%s\"", jrnlname);

        jrnl_in = fopen(jrnlname, "r");
        if (jrnl_in == NULL) {
            SPRINT0(0, "ERROR:: Journal file cannot be opened");
            exit(0);
        } else {
            while (1) {
                temp = fgets(text, MAX_STR_LEN, jrnl_in);
                if (temp == NULL) break;

                if (feof(jrnl_in)) break;

                /* if there is a \n, convert it to a \0 */
                text[MAX_STR_LEN-1] = '\0';

                for (i = 0; i < strlen(text); i++) {
                    if (text[i] == '\n') {
                        text[i] =  '\0';
                        break;
                    }
                }

                if (strncmp(text, "##end##", 7) == 0) break;

                processMessage(text);
            }

            fclose(jrnl_in);

            SPRINT0(0, "==> Closing input journal file");
        }
    }

    /* get the command to start the client (if any) */
    if (batch == 0) {
        wv_start = getenv("SLUGS_START");
    }

    /* start the server */
    if (batch == 0) {
        status2 = SUCCESS;
        if (wv_startServer(port, NULL, NULL, NULL, 0, cntxt) == 0) {

            /* stay alive a long as we have a client */
            while (wv_statusServer(0)) {
                usleep(100000);

                /* start the browser if the first time through this loop */
                if (status2 == SUCCESS) {
                    if (wv_start != NULL) {
                        status2 += system(wv_start);
                    }

                    status2++;
                }

                /* start hand-shaking */
                if (Tris_pend  != 0 ||
                    Hangs_pend != 0 ||
                    Links_pend != 0 ||
                    CurPt_pend != 0   ) {
                    ihand = 1;
                    if (wv_handShake(cntxt) != 1) {
                        SPRINT0(0, "ERROR:: handShake out of Sync 1");
                    }
                } else {
                    ihand = 0;
                }

                /* update Triangles if there are changes pending */
                if (Tris_pend != 0) {
                    Tris_pend = 0;

                    status = buildTriangles();
                    SPRINT1(3, "buildTriangles -> status=%d", status);
                }

                /* update Hangs if there are changes pending */
                if (Hangs_pend != 0) {
                    Hangs_pend = 0;

                    status = buildHangs();
                    SPRINT1(3, "buildHangs -> status=%d", status);
                }

                /* update Links if there are changes pending */
                if (Links_pend != 0) {
                    Links_pend = 0;

                    status = buildLinks();
                    SPRINT1(3, "buildLinks -> status=%d", status);
                }

                /* update CurPt if there are changes pending */
                if (CurPt_pend != 0) {
                    CurPt_pend = 0;

                    status = buildCurPt();
                    SPRINT1(3, "buildCurPt -> status=%d", status);
                }

                /* complete hand-shaking since all changes have been made */
                if (ihand == 1) {
                    if (wv_handShake(cntxt) != 0) {
                        SPRINT0(0, "ERROR:: handShake out of Sync 0");
                    }
                }
            }
        }
    }

    /* cleanup and exit */
    if (jrnl_out != NULL) fclose(jrnl_out);
    jrnl_out = NULL;

    for (iface = 1; iface <= Nface; iface++) {
        status = freeTess(&(face[iface].tess));
        SPRINT2(2, "freeTess(iface=%d) -> status=%d", iface, status);
 
        FREE(face[iface].edg);
        FREE(face[iface].xyz);
        FREE(face[iface].cp );
    }

    for (iedge = 1; iedge <= Nedge; iedge++) {
        FREE(edge[iedge].pnt);
        FREE(edge[iedge].xyz);
        FREE(edge[iedge].cp );
    }

    FREE(face);
    FREE(edge);
    FREE(node);

    status = freeTess(&tess);
    SPRINT1(2, "freeTess(tess) -> status=%d", status);

    status = freeTess(&tess_undo);
    SPRINT1(2, "freeTess(tess_undo) -> status=%d", status);

    wv_cleanupServers();
    status = SUCCESS;

    SPRINT0(1, "==> Slugs completed successfully");

cleanup:
    FREE(response   );
    FREE(text       );
    FREE(tempname   );
    FREE(jrnlname   );
    FREE(casename   );
    FREE(filename   );
    FREE(sgFocusData);

    return status;
}


/*
 ***********************************************************************
 *                                                                     *
 *   buildCurPt - make/update CurPt in scene graph                     *
 *                                                                     *
 ***********************************************************************
 */
static int
buildCurPt()
{
    int       status = SUCCESS;

    int       attrs;
    float     *pnt=NULL;

    float     col[3];
    char      gpname[33];

    wvData    items[7];

    ROUTINE(buildCurPt);

    /* --------------------------------------------------------------- */

    if (CurPt_index < 0) {
        goto cleanup;
    }

    /* note: we need to update more than one Point because of
       an undocumented "feature" in wv */
    MALLOC(pnt, float, 6);

    pnt[0] = tess.xyz[3*CurPt_index  ];   pnt[3] = pnt[0];
    pnt[1] = tess.xyz[3*CurPt_index+1];   pnt[4] = pnt[1];
    pnt[2] = tess.xyz[3*CurPt_index+2];   pnt[5] = pnt[2];
    status = wv_setData(WV_REAL32, 2, (void*)pnt, WV_VERTICES, &(items[0]));
    SPRINT4(3, "wv_setData(VERTICES, %f, %f, %f) -> status=%d", pnt[0], pnt[1], pnt[2], status);

    wv_adjustVerts(&(items[0]), focus);

    if (CurPt_gprim < 0) {
        snprintf(gpname, 32, "CurPt");
        attrs = WV_ON;

        setColor(0x000000, col);
        status = wv_setData(WV_REAL32, 1, (void*)col, WV_COLORS, &(items[1]));
        SPRINT4(3, "wv_setData(COLORS, %f, %f, %f) -> status=%d", col[0], col[1], col[2], status);

        CurPt_gprim = wv_addGPrim(cntxt, gpname, WV_POINT, attrs, 2, items);
        SPRINT1(3, "wv_addGPrim(WV_POINT) -> CurPt_gprim=%d", CurPt_gprim);

        cntxt->gPrims[CurPt_gprim].pSize = 8.0;
    } else {
        SPRINT1(3, "CurPt_gprim=%d", CurPt_gprim);
        status = wv_modGPrim(cntxt, CurPt_gprim, 1, items);
        SPRINT2(3, "wv_modGPrim(CurPt_gprim=%d) -> status=%d", CurPt_gprim, status);
    }

cleanup:
    FREE(pnt);

    return status;
}


/*
 ***********************************************************************
 *                                                                     *
 *   buildHangs - make/update Hangs in scene graph                     *
 *                                                                     *
 ***********************************************************************
 */
static int
buildHangs()
{
    int       status = SUCCESS;

    int       itri, jtri, nhang, ip0, ip1, ip2;
    int       attrs;
    float     *hang=NULL;

    float     color[3];
    char      gpname[33];

    wvData    items[7];

    ROUTINE(buildHangs);

    /* --------------------------------------------------------------- */

    /* determine the number of Hangs in the current tessellation */
    tess.nhang = 0;
    for (itri = 0; itri < tess.ntri; itri++) {
        if ((tess.ttyp[itri] & TRI_ACTIVE) == 0) continue;

        jtri = tess.trit[3*itri  ];
        if (jtri < 0 || (tess.ttyp[jtri] & TRI_ACTIVE) == 0) {
            tess.nhang++;
        }

        jtri = tess.trit[3*itri+1];
        if (jtri < 0 || (tess.ttyp[jtri] & TRI_ACTIVE) == 0) {
            tess.nhang++;
        }

        jtri = tess.trit[3*itri+2];
        if (jtri < 0 || (tess.ttyp[jtri] & TRI_ACTIVE) == 0) {
            tess.nhang++;
        }
    }

    /* plot Hangs (if there are any) */
    if (tess.nhang > 0) {
        snprintf(gpname, 32, "Hangs");
        attrs = WV_ON;

        MALLOC(hang, float, 6*tess.nhang);

        nhang = 0;
        for (itri = 0; itri < tess.ntri; itri++) {
            if ((tess.ttyp[itri] & TRI_ACTIVE) == 0) continue;

            jtri = tess.trit[3*itri  ];
            if (jtri < 0 || (tess.ttyp[jtri] & TRI_ACTIVE) == 0) {
                ip1 = tess.trip[3*itri+1];
                ip2 = tess.trip[3*itri+2];

                hang[6*nhang  ] = tess.xyz[3*ip1  ];
                hang[6*nhang+1] = tess.xyz[3*ip1+1];
                hang[6*nhang+2] = tess.xyz[3*ip1+2];
                hang[6*nhang+3] = tess.xyz[3*ip2  ];
                hang[6*nhang+4] = tess.xyz[3*ip2+1];
                hang[6*nhang+5] = tess.xyz[3*ip2+2];
                nhang++;
            }

            jtri = tess.trit[3*itri+1];
            if (jtri < 0 || (tess.ttyp[jtri] & TRI_ACTIVE) == 0) {
                ip2 = tess.trip[3*itri+2];
                ip0 = tess.trip[3*itri  ];

                hang[6*nhang  ] = tess.xyz[3*ip2  ];
                hang[6*nhang+1] = tess.xyz[3*ip2+1];
                hang[6*nhang+2] = tess.xyz[3*ip2+2];
                hang[6*nhang+3] = tess.xyz[3*ip0  ];
                hang[6*nhang+4] = tess.xyz[3*ip0+1];
                hang[6*nhang+5] = tess.xyz[3*ip0+2];
                nhang++;
            }

            jtri = tess.trit[3*itri+2];
            if (jtri < 0 || (tess.ttyp[jtri] & TRI_ACTIVE) == 0) {
                ip0 = tess.trip[3*itri  ];
                ip1 = tess.trip[3*itri+1];

                hang[6*nhang  ] = tess.xyz[3*ip0  ];
                hang[6*nhang+1] = tess.xyz[3*ip0+1];
                hang[6*nhang+2] = tess.xyz[3*ip0+2];
                hang[6*nhang+3] = tess.xyz[3*ip1  ];
                hang[6*nhang+4] = tess.xyz[3*ip1+1];
                hang[6*nhang+5] = tess.xyz[3*ip1+2];
                nhang++;
            }

            if (nhang > tess.nhang) {
                SPRINT2(0, "ERROR:: nhang=%d but tess.nhang=%d", nhang, tess.nhang);
                status = -999;
                goto cleanup;
            }
        }

        /* set up vertices for the Hangs */
        status = wv_setData(WV_REAL32, 2*nhang, (void*)hang, WV_VERTICES, &(items[0]));
        if (status != SUCCESS) {
            SPRINT1(3, "wv_setData -> status=%d", status);
        }

        wv_adjustVerts(&(items[0]), focus);

        /* hang color */
        setColor(0xff0000, color);
        status = wv_setData(WV_REAL32, 1, (void*)color, WV_COLORS, &(items[1]));
        if (status != SUCCESS) {
            SPRINT1(3, "wv_setData -> status=%d", status);
        }

        /* if a graphic primitive does not exist yet, make it */
        if (Hangs_gprim < 0) {
            Hangs_gprim = wv_addGPrim(cntxt, gpname, WV_LINE, attrs, 2, items);
            SPRINT1(3, "wv_addGPrim(WV_LINE) -> Hangs_gprim=%d", Hangs_gprim);
            if (Hangs_gprim >= 0) {

                /* make line width 5 */
                cntxt->gPrims[Hangs_gprim].lWidth = 5.0;
            }

        /* otherwise, simply modify the verticies in the previous one */
        } else {
            status = wv_modGPrim(cntxt, Hangs_gprim, 1, items);
            SPRINT2(3, "wv_modGPrim(Hangs_gprim=%d) -> status=%d", Hangs_gprim, status);
        }

    /* if no Hangs, remove the previous gprim */
    } else if (Hangs_gprim >= 0) {
        wv_removeGPrim(cntxt, Hangs_gprim);
        SPRINT0(3, "wv_removeGPrim");

        Hangs_gprim = -1;
    }

cleanup:
    FREE(hang);

    return status;
}


/*
 ***********************************************************************
 *                                                                     *
 *   buildLinks - make/update Links in scene graph                     *
 *                                                                     *
 ***********************************************************************
 */
static int
buildLinks()
{
    int       status = SUCCESS;

    int       itri, nlink, ip0, ip1, ip2;
    int       attrs;
    float     *link=NULL;

    float     color[3];
    char      gpname[33];

    wvData    items[7];

    ROUTINE(buildLinks);

    /* --------------------------------------------------------------- */

    /* plot Links (if there are any) */
    if (tess.nlink > 0) {
        snprintf(gpname, 32, "Links");
        attrs = WV_ON;

        MALLOC(link, float, 6*tess.nlink);

        nlink = 0;
        for (itri = 0; itri < tess.ntri; itri++) {
            if ((tess.ttyp[itri] & TRI_ACTIVE) == 0) continue;

            if (tess.ttyp[itri] & TRI_T0_LINK) {
                if (itri > tess.trit[3*itri  ]) {
                    ip1 = tess.trip[3*itri+1];
                    ip2 = tess.trip[3*itri+2];

                    link[6*nlink  ] = tess.xyz[3*ip1  ];
                    link[6*nlink+1] = tess.xyz[3*ip1+1];
                    link[6*nlink+2] = tess.xyz[3*ip1+2];
                    link[6*nlink+3] = tess.xyz[3*ip2  ];
                    link[6*nlink+4] = tess.xyz[3*ip2+1];
                    link[6*nlink+5] = tess.xyz[3*ip2+2];
                    nlink++;
                }
            }
            if (tess.ttyp[itri] & TRI_T1_LINK) {
                if (itri > tess.trit[3*itri+1]) {
                    ip2 = tess.trip[3*itri+2];
                    ip0 = tess.trip[3*itri  ];

                    link[6*nlink  ] = tess.xyz[3*ip2  ];
                    link[6*nlink+1] = tess.xyz[3*ip2+1];
                    link[6*nlink+2] = tess.xyz[3*ip2+2];
                    link[6*nlink+3] = tess.xyz[3*ip0  ];
                    link[6*nlink+4] = tess.xyz[3*ip0+1];
                    link[6*nlink+5] = tess.xyz[3*ip0+2];
                    nlink++;
                }
            }
            if (tess.ttyp[itri] & TRI_T2_LINK) {
                if (itri > tess.trit[3*itri+2]) {
                    ip0 = tess.trip[3*itri  ];
                    ip1 = tess.trip[3*itri+1];

                    link[6*nlink  ] = tess.xyz[3*ip0  ];
                    link[6*nlink+1] = tess.xyz[3*ip0+1];
                    link[6*nlink+2] = tess.xyz[3*ip0+2];
                    link[6*nlink+3] = tess.xyz[3*ip1  ];
                    link[6*nlink+4] = tess.xyz[3*ip1+1];
                    link[6*nlink+5] = tess.xyz[3*ip1+2];
                    nlink++;
                }
            }

            if (nlink > tess.nlink) {
                SPRINT2(0, "ERROR:: nlink=%d but tess.nlink=%d", nlink, tess.nlink);
                status = -999;
                goto cleanup;
            }
        }

        /* set up vertices for the Links */
        status = wv_setData(WV_REAL32, 2*nlink, (void*)link, WV_VERTICES, &(items[0]));
        if (status != SUCCESS) {
            SPRINT1(3, "wv_setData -> status=%d", status);
        }

        wv_adjustVerts(&(items[0]), focus);

        /* link color */
        setColor(0xffffff, color);
        status = wv_setData(WV_REAL32, 1, (void*)color, WV_COLORS, &(items[1]));
        if (status != SUCCESS) {
            SPRINT1(3, "wv_setData -> status=%d", status);
        }

        /* if a graphic primitive does not exist yet, make it */
        if (Links_gprim < 0) {
            Links_gprim = wv_addGPrim(cntxt, gpname, WV_LINE, attrs, 2, items);
            SPRINT1(3, "wv_addGPrim(WV_LINE) -> Links_gprim=%d", Links_gprim);
            if (Links_gprim >= 0) {

                /* make line width 5 */
                cntxt->gPrims[Links_gprim].lWidth = 5.0;
            }

        /* otherwise, simply modify the verticies in the previous one */
        } else {
            status = wv_modGPrim(cntxt, Links_gprim, 1, items);
            SPRINT2(3, "wv_modGPrim(Links_gprim=%d) -> status=%d", Links_gprim, status);
        }

    /* if no Links, remove the previous gprim */
    } else if (Links_gprim >= 0) {
        wv_removeGPrim(cntxt, Links_gprim);
        SPRINT0(3, "wv_removeGPrim");

        Links_gprim = -1;
    }

cleanup:
    FREE(link);

    return status;
}


/*
 ***********************************************************************
 *                                                                     *
 *   buildTriangles - make/update Triangles in scene graph             *
 *                                                                     *
 ***********************************************************************
 */
static int
buildTriangles()
{
    int       status = SUCCESS;

    int       icolr, itri, ntri, ipnt, npnt;
    int       igprim, attrs;
    int       *tri=NULL, *seg=NULL;
    float     *xyz=NULL;

    float     color[3];
    double    bigbox[6], size;
    char      gpname[33];

    wvData    items[7];

    ROUTINE(buildTriangles);

    /* --------------------------------------------------------------- */

    /* remove any graphic primitives that already exist */
    wv_removeAll(cntxt);

    /* reset all the GPrim values since they do not exist anymore */
    CurPt_gprim = -1;
    Hangs_gprim = -1;
    Links_gprim = -1;

    /* find the values needed to adjust the vertices */
    bigbox[0] = bigbox[1] = bigbox[2] = +HUGEQ;
    bigbox[3] = bigbox[4] = bigbox[5] = -HUGEQ;

    for (ipnt = 0; ipnt < tess.npnt; ipnt++) {
        if (tess.xyz[3*ipnt  ] < bigbox[0]) bigbox[0] = tess.xyz[3*ipnt  ];
        if (tess.xyz[3*ipnt+1] < bigbox[1]) bigbox[1] = tess.xyz[3*ipnt+1];
        if (tess.xyz[3*ipnt+2] < bigbox[2]) bigbox[2] = tess.xyz[3*ipnt+2];
        if (tess.xyz[3*ipnt  ] > bigbox[3]) bigbox[3] = tess.xyz[3*ipnt  ];
        if (tess.xyz[3*ipnt+1] > bigbox[4]) bigbox[4] = tess.xyz[3*ipnt+1];
        if (tess.xyz[3*ipnt+2] > bigbox[5]) bigbox[5] = tess.xyz[3*ipnt+2];
    }

                                    size = bigbox[3] - bigbox[0];
    if (size < bigbox[4]-bigbox[1]) size = bigbox[4] - bigbox[1];
    if (size < bigbox[5]-bigbox[2]) size = bigbox[5] - bigbox[2];

    focus[0] = (bigbox[0] + bigbox[3]) / 2;
    focus[1] = (bigbox[1] + bigbox[4]) / 2;
    focus[2] = (bigbox[2] + bigbox[5]) / 2;
    focus[3] = size;

    /* generate the scene graph focus data */
    snprintf(sgFocusData, 256, "sgFocus;[%20.12e,%20.12e,%20.12e,%20.12e]", focus[0], focus[1], focus[2], focus[3]);

    /* loop through the Colors */
    for (icolr = 0; icolr <= tess.ncolr; icolr++) {

        /* name and attributes */
        snprintf(gpname, 32, "Color %d", icolr);
        attrs = WV_ON | WV_ORIENTATION | WV_LINES;

        /* find number of Triangles with this color */
        ntri = 0;
        for (itri = 0; itri < tess.ntri; itri++) {
            if ((tess.ttyp[itri] & TRI_ACTIVE) == 0) continue;

            if ((tess.ttyp[itri] & TRI_COLOR) == icolr) {
                ntri++;
            }
        }

        /* if there are not Triangles with this color, skip the
           rest of the processing */
        if (ntri <= 0) continue;

        /* allocate storage for vertices and Triangles */
        MALLOC(xyz, float, 9*ntri);
        MALLOC(tri, int,   3*ntri);
        MALLOC(seg, int,   6*ntri);

        /* set up vertices and segments of this color */
        npnt = 0;
        ntri = 0;
        for (itri = 0; itri < tess.ntri; itri++) {
            if ((tess.ttyp[itri] & TRI_ACTIVE) == 0) continue;

            if ((tess.ttyp[itri] & TRI_COLOR) == icolr) {

                ipnt = tess.trip[3*itri  ];
                xyz[3*npnt  ] = tess.xyz[3*ipnt  ];
                xyz[3*npnt+1] = tess.xyz[3*ipnt+1];
                xyz[3*npnt+2] = tess.xyz[3*ipnt+2];
                npnt++;

                ipnt = tess.trip[3*itri+1];
                xyz[3*npnt  ] = tess.xyz[3*ipnt  ];
                xyz[3*npnt+1] = tess.xyz[3*ipnt+1];
                xyz[3*npnt+2] = tess.xyz[3*ipnt+2];
                npnt++;

                ipnt = tess.trip[3*itri+2];
                xyz[3*npnt  ] = tess.xyz[3*ipnt  ];
                xyz[3*npnt+1] = tess.xyz[3*ipnt+1];
                xyz[3*npnt+2] = tess.xyz[3*ipnt+2];
                npnt++;

                tri[3*ntri  ] = npnt - 3;
                tri[3*ntri+1] = npnt - 2;
                tri[3*ntri+2] = npnt - 1;

                seg[6*ntri  ] = npnt - 3;
                seg[6*ntri+1] = npnt - 2;
                seg[6*ntri+2] = npnt - 2;
                seg[6*ntri+3] = npnt - 1;
                seg[6*ntri+4] = npnt - 1;
                seg[6*ntri+5] = npnt - 3;
                ntri++;
            }
        }

        /* (indexed) verticies */
        status = wv_setData(WV_REAL32, npnt, (void*)xyz, WV_VERTICES, &(items[0]));
        if (status != SUCCESS) {
            SPRINT1(3, "wv_setData -> status=%d", status);
        }

        wv_adjustVerts(&(items[0]), focus);

        /* (indexed) Triangles */
        status = wv_setData(WV_INT32, 3*ntri, (void*)tri, WV_INDICES, &(items[1]));
        if (status != SUCCESS) {
            SPRINT1(3, "wv_setData -> status=%d", status);
        }

        /* constant Triangle colors */
        if         (icolr == 0) {
            setColor(0xcfcfcf, color);         /* light grey */
        } else if ((icolr-1)%6 == 0) {
            setColor(0xffbfbf, color);         /* light red */
        } else if ((icolr-1)%6 == 1) {
            setColor(0xbfffbf, color);         /* light green */
        } else if ((icolr-1)%6 == 2) {
            setColor(0xbfbfff, color);         /* light blue */
        } else if ((icolr-1)%6 == 3) {
            setColor(0xbfffff, color);         /* light cyan */
        } else if ((icolr-1)%6 == 4) {
            setColor(0xffbfff, color);         /* light magenta */
        } else {
            setColor(0xffffbf, color);         /* light yellow */
        }

        status = wv_setData(WV_REAL32, 1, (void*)color, WV_COLORS, &(items[2]));
        if (status != SUCCESS) {
            SPRINT1(3, "wv_setData -> status=%d", status);
        }

        /* (indexed) Triangle sides */
        status = wv_setData(WV_INT32, 6*ntri, (void*)seg, WV_LINDICES, &(items[3]));
        if (status != SUCCESS) {
            SPRINT1(3, "wv_setData -> status=%d", status);
        }

        /* segment colors */
        setColor(0x000000, color);
        status = wv_setData(WV_REAL32, 1, (void*)color, WV_LCOLOR, &(items[4]));
        if (status != SUCCESS) {
            SPRINT1(3, "wv_setData -> status=%d", status);
        }

        /* make graphic primitive */
        igprim = wv_addGPrim(cntxt, gpname, WV_TRIANGLE, attrs, 5, items);
        SPRINT1(3, "wv_addGPrim(WV_TRIANGLE) -> igprim=%d", igprim);
        if (igprim >= 0) {

            /* make line width 1 */
            cntxt->gPrims[igprim].lWidth = 1.0;
        }

        /* free storage associated with this color */
        FREE(xyz);
        FREE(tri);
        FREE(seg);
    }

cleanup:
    FREE(xyz);
    FREE(tri);
    FREE(seg);

    return status;
}


/*
 ***********************************************************************
 *                                                                     *
 *   generateBrep - generate Brep based upon Triangle colors           *
 *                                                                     *
 ***********************************************************************
 */
static int
generateBrep()
{
    int       status = SUCCESS;

    int       nchange;
    int       icolr, jcolr, itri, jtri, ntri;
    int       isgmt, nsgmt, ibeg, iend;
    int       ipnt, jpnt, count, done, i, swap;
    int       inode, iedge, jedge, iface;
    int       *nodnum=NULL;

    sgmt_T     *sgmt=NULL;

    ROUTINE(generateBrep);

    /* --------------------------------------------------------------- */

    SPRINT0(1, "Generating Brep...");


    /* determine if there are any coincident points */
    for (ipnt = 0; ipnt < tess.npnt; ipnt++) {
        for (jpnt = ipnt+1; jpnt < tess.npnt; jpnt++) {
            if (fabs(tess.xyz[3*ipnt  ]-tess.xyz[3*jpnt  ]) < EPS06 &&
                fabs(tess.xyz[3*ipnt+1]-tess.xyz[3*jpnt+1]) < EPS06 &&
                fabs(tess.xyz[3*ipnt+2]-tess.xyz[3*jpnt+2]) < EPS06   ) {
                printf("duplicate point found\n");
                printf("ipnt=%6d %20.10f %20.10f %20.10f\n", ipnt, tess.xyz[3*ipnt], tess.xyz[3*ipnt+1], tess.xyz[3*ipnt+2]);
                printf("jpnt=%6d %20.10f %20.10f %20.10f\n", jpnt, tess.xyz[3*jpnt], tess.xyz[3*jpnt+1], tess.xyz[3*jpnt+2]);
            }
        }
    }

    /* clear the Node, Edge, and Face tables */
    Mnode = 0;
    Nnode = 0;
    FREE(node);

    for (iedge = 1; iedge <= Nedge; iedge++) {
        FREE(edge[iedge].pnt);
        FREE(edge[iedge].xyz);
        FREE(edge[iedge].cp );
    }
    Medge = 0;
    Nedge = 0;
    FREE(edge);

    for (iface = 1; iface <= Nface; iface++) {
        FREE(face[iface].edg);
        FREE(face[iface].cp );
    }

    Mface = 0;
    Nface = 0;
    FREE(face);

    /* allocate Segment table (larger than needed) */
    MALLOC(sgmt, sgmt_T, tess.ntri);

    /* allocate a table which identifies the Node at any Point */
    MALLOC(nodnum, int, tess.npnt);
    for (ipnt = 0; ipnt < tess.npnt; ipnt++) {
        nodnum[ipnt] = -1;
    }

    /* determine the number of Faces */
    Mface = 0;
    for (icolr = 1; icolr <= tess.ncolr; icolr++) {
        for (itri = 0; itri < tess.ntri; itri++) {
            if ((tess.ttyp[itri] & TRI_COLOR) == icolr) {
                Mface++;
                break;
            }
        }
    }
    SPRINT1(1, "   there are %d Faces", Mface);

    /* preallocate Node, Edge, and Face tables */
    Mnode = 100;
    Medge = 100;
//  Mface set above

    MALLOC(node, node_T, Mnode+1);
    MALLOC(edge, edge_T, Medge+1);
    MALLOC(face, face_T, Mface+1);

    /* initialize the Faces */
    SPRINT0(1, "Initializing Faces...");
    for (icolr = 1; icolr <= tess.ncolr; icolr++) {
        ntri = 0;
        for (itri = 0; itri < tess.ntri; itri++) {
            if ((tess.ttyp[itri] & TRI_COLOR) == icolr) {
                ntri++;
            }
        }

        if (ntri > 0) {
            Nface++;

            face[Nface].icol  = icolr;
//          face[Nface].tess  = NULL;
            face[Nface].nedg  = 0;
            face[Nface].edg   = NULL;
            face[Nface].npnt  = 0;
            face[Nface].xyz   = NULL;
            face[Nface].ncp   = 0;
            face[Nface].cp    = NULL;
//          face[Nface].eface = NULL;

            status = extractColor(&tess, icolr, &(face[Nface].tess));
            CHECK_STATUS(extractColor);

            SPRINT4(1, "   created Face %3d .icol=%6d, .npnt=%6d, .ntri=%6d", Nface,
                   face[Nface].icol,
                   face[Nface].tess.npnt,
                   face[Nface].tess.ntri);
        }
    }

    /* loop through all color pairs to find possible Edges */
    SPRINT0(1, "Looking for possible Edges (and Nodes)...");
    for (icolr = 1; icolr <= tess.ncolr; icolr++) {
        for (jcolr = icolr+1; jcolr <= tess.ncolr; jcolr++) {

            /* find all posible Segments */
            nsgmt = 0;
            for (itri = 0; itri < tess.ntri; itri++) {
                if ((tess.ttyp[itri] & TRI_COLOR) == icolr) {
                    jtri = tess.trit[3*itri  ];
                    if (jtri >= 0) {
                        if ((tess.ttyp[jtri] & TRI_COLOR) == jcolr) {
                            sgmt[nsgmt].ibeg = tess.trip[3*itri+1];
                            sgmt[nsgmt].iend = tess.trip[3*itri+2];
                            sgmt[nsgmt].prev = -1;
                            sgmt[nsgmt].next = -1;
                            nsgmt++;
                        }
                    }

                    jtri = tess.trit[3*itri+1];
                    if (jtri >= 0) {
                        if ((tess.ttyp[jtri] & TRI_COLOR) == jcolr) {
                            sgmt[nsgmt].ibeg = tess.trip[3*itri+2];
                            sgmt[nsgmt].iend = tess.trip[3*itri  ];
                            sgmt[nsgmt].prev = -1;
                            sgmt[nsgmt].next = -1;
                            nsgmt++;
                        }
                    }

                    jtri = tess.trit[3*itri+2];
                    if (jtri >= 0) {
                        if ((tess.ttyp[jtri] & TRI_COLOR) == jcolr) {
                            sgmt[nsgmt].ibeg = tess.trip[3*itri  ];
                            sgmt[nsgmt].iend = tess.trip[3*itri+1];
                            sgmt[nsgmt].prev = -1;
                            sgmt[nsgmt].next = -1;
                            nsgmt++;
                        }
                    }
                }
            }
            SPRINT3(2, "icolr=%2d  jcolr=%2d  nsgmt=%d", icolr, jcolr, nsgmt);

            /* if there are no Segments, there is nothing to do */
            if (nsgmt <= 0) continue;

            /* arrange the Segments head to tail.  the while loop is needed
               because there may be more than one set of disjoint segments
               between two colors */
            while (1) {

                /* find the first segment that is not used */
                ibeg = -1;
                for (isgmt = 0; isgmt < nsgmt; isgmt++) {
                    if (sgmt[isgmt].prev == -1 && sgmt[isgmt].next == -1) {
                        ibeg = isgmt;         /* first Segment */
                        iend = isgmt;         /* last  Segment */
                        break;
                    }
                }

                /* if all the segments are used, we are done */
                if (ibeg < 0) break;

                /* keep adding Segments to end while possible */
                nchange = 1;
                while (nchange > 0) {
                    nchange = 0;

                    if (sgmt[ibeg].ibeg == sgmt[iend].iend) {
                        break;
                    }

                    for (isgmt = 1; isgmt < nsgmt; isgmt++) {
                        if (sgmt[isgmt].prev >= 0) {
                            continue;
                        } else if (sgmt[isgmt].ibeg == sgmt[iend].iend) {
                            sgmt[isgmt].prev = iend;
                            sgmt[iend ].next = isgmt;

                            iend = isgmt;
                            nchange++;
                        }
                    }
                }

                /* keep adding Segments to beginning while possible */
                nchange = 1;
                while (nchange > 0) {
                    nchange = 0;

                    if (sgmt[ibeg].ibeg == sgmt[iend].iend) {
                        break;
                    }

                    for (isgmt = 1; isgmt < nsgmt; isgmt++) {
                        if (sgmt[isgmt].next >= 0) {
                            continue;
                        } else if (sgmt[isgmt].iend == sgmt[ibeg].ibeg) {
                            sgmt[isgmt].next = ibeg;
                            sgmt[ibeg ].prev = isgmt;

                            ibeg = isgmt;
                            nchange++;
                        }
                    }
                }

                /* if only one Segment for this Edge, specially mark the Segment so that
                   it does not get used again */
                if (ibeg == iend) {
                    sgmt[ibeg].prev = -2;
                    sgmt[ibeg].next = -2;
                }

                /* create a Node at the beginning if there is not one there already */
                ipnt= sgmt[ibeg].ibeg;
                if (nodnum[ipnt] < 0) {
                    nodnum[ipnt] = Nnode + 1;

                    if (Nnode >= Mnode) {
                        Mnode += 100;
                        RALLOC(node, node_T, Mnode+1);
                    }

                    Nnode++;

                    node[Nnode].ipnt  = ipnt;
                    node[Nnode].nedg  = 0;
                    node[Nnode].x     = tess.xyz[3*ipnt  ];
                    node[Nnode].y     = tess.xyz[3*ipnt+1];
                    node[Nnode].z     = tess.xyz[3*ipnt+2];
                    node[Nnode].enode = NULL;

                    SPRINT6(1, "   created Node %3d .ipnt=%6d, .nedg=%6d, .x=%10.4f, .y=%10.4f, .z=%10.4f", Nnode,
                            node[Nnode].ipnt,
                            node[Nnode].nedg,
                            node[Nnode].x,
                            node[Nnode].y,
                            node[Nnode].z);
                }

                /* create a Node at the end if there is not one there already */
                ipnt= sgmt[iend].iend;
                if (nodnum[ipnt] < 0) {
                    nodnum[ipnt] = Nnode + 1;

                    if (Nnode >= Mnode) {
                        Mnode += 100;
                        RALLOC(node, node_T, Mnode+1);
                    }

                    Nnode++;

                    node[Nnode].ipnt  = ipnt;
                    node[Nnode].nedg  = 0;
                    node[Nnode].x     = tess.xyz[3*ipnt  ];
                    node[Nnode].y     = tess.xyz[3*ipnt+1];
                    node[Nnode].z     = tess.xyz[3*ipnt+2];
                    node[Nnode].enode = NULL;

                    SPRINT6(1, "   created Node %3d .ipnt=%6d, .nedg=%6d, .x=%10.4f, .y=%10.4f, .z=%10.4f", Nnode,
                            node[Nnode].ipnt,
                            node[Nnode].nedg,
                            node[Nnode].x,
                            node[Nnode].y,
                            node[Nnode].z);
                }

                /* create the Edge */
                if (Nedge >= Medge) {
                    Medge += 100;
                    RALLOC(edge, edge_T, Medge+1);
                }

                Nedge++;

                edge[Nedge].ibeg  = nodnum[sgmt[ibeg].ibeg];
                edge[Nedge].iend  = nodnum[sgmt[iend].iend];
                edge[Nedge].ileft = 0;
                edge[Nedge].irite = 0;
                edge[Nedge].npnt  = nsgmt + 1;
                edge[Nedge].pnt   = NULL;
                edge[Nedge].xyz   = NULL;
                edge[Nedge].ncp   = 0;
                edge[Nedge].cp    = NULL;
//              edge[Nedge].eface = NULL;

                node[edge[Nedge].ibeg].nedg += 1;
                node[edge[Nedge].iend].nedg += 1;

                for (iface = 1; iface <= Nface; iface++) {
                    if        (face[iface].icol == icolr) {
                        edge[Nedge].ileft = iface;
                        face[iface].nedg += 1;
                    } else if (face[iface].icol == jcolr) {
                        edge[Nedge].irite = iface;
                        face[iface].nedg += 1;
                    }
                }

                MALLOC(edge[Nedge].pnt, int, nsgmt+1);

                isgmt = ibeg;
                edge[Nedge].pnt[0] = sgmt[isgmt].ibeg;

                for (ipnt = 1; ipnt <= nsgmt; ipnt++) {
                    edge[Nedge].pnt[ipnt] = sgmt[isgmt].iend;
                    isgmt = sgmt[isgmt].next;
                    if (isgmt < 0) {
                        edge[Nedge].npnt = ipnt + 1;
                        break;
                    }
                }

                SPRINT6(1, "   created Edge %3d .ibeg=%6d, .iend=%6d, .ileft=%4d, .irite=%4d, .npnt=%6d", Nedge,
                        edge[Nedge].ibeg,
                        edge[Nedge].iend,
                        edge[Nedge].ileft,
                        edge[Nedge].irite,
                        edge[Nedge].npnt);
            }
        }
    }

    /* inform the Faces of their incident Edges */
    for (iface = 1; iface <= Nface; iface++) {

        MALLOC(face[iface].edg, int, face[iface].nedg);

        for (jedge = 0; jedge < face[iface].nedg; jedge++) {
            face[iface].edg[jedge] = 0;           // needed to avoid clang warning
        }

        jedge = 0;
        for (iedge = 1; iedge <= Nedge; iedge++) {
            if (edge[iedge].ileft == iface) {
                face[iface].edg[jedge++] = +iedge;
            }
            if (edge[iedge].irite == iface) {
                face[iface].edg[jedge++] = -iedge;
            }
        }
    }

    /* reorder the Edges in each Face to form loops */
    for (iface = 1; iface <= Nface; iface++) {
        printf("Face %6d\n", iface);

        if (face[iface].nedg <= 0) continue;

        done  = 0;
        iedge = face[iface].edg[done];
        if (iedge > 0) {
            ibeg = edge[+iedge].ibeg;
            iend = edge[+iedge].iend;
        } else {
            ibeg = edge[-iedge].iend;
            iend = edge[-iedge].ibeg;
        }
        done++;

        count=0;
        while (1) {
            for (i = done; i < face[iface].nedg; i++) {
                iedge = face[iface].edg[i];
                if (iedge > 0) {
                    if (edge[+iedge].ibeg == iend) {
                        if (i > done) {
                            swap                  = face[iface].edg[done];
                            face[iface].edg[done] = face[iface].edg[i   ];
                            face[iface].edg[i   ] = swap;
                        }
                        iend = edge[+iedge].iend;
                        done++;
                        break;
                    }
                } else {
                    if (edge[-iedge].iend == iend) {
                        if (i > done) {
                            swap                  = face[iface].edg[done];
                            face[iface].edg[done] = face[iface].edg[i   ];
                            face[iface].edg[i   ] = swap;
                    }
                        iend = edge[-iedge].ibeg;
                        done++;
                    break;
                    }
                }
            }

            if (iend == ibeg) break;

            count++;
            if (count > 100) {printf("rats too\n"); exit(0);}
        }

        for (i = 0; i < face[iface].nedg; i++) {
            iedge = face[iface].edg[i];
            if (iedge > 0) {
                printf("     Edge %6d, npnt=%6d, ibeg=%3d, iend=%3d\n",
                       iedge, edge[+iedge].npnt, edge[+iedge].ibeg, edge[+iedge].iend);
            } else {
                printf("     Edge %6d, npnt=%6d, iend=%3d, ibeg=%3d\n",
                       iedge, edge[-iedge].npnt, edge[-iedge].iend, edge[-iedge].ibeg);
            }
        }

        if (done != face[iface].nedg) {
            printf("more than one loop, so not face* file created\n");
            continue;
        }
    }

    /* set the coordinates for each Edge Point (including bounding Nodes) */
    for (iedge = 1; iedge <= Nedge; iedge++) {
        MALLOC(edge[iedge].xyz, double, 3*edge[iedge].npnt);

        for (i = 0; i < edge[iedge].npnt; i++) {
            ipnt = edge[iedge].pnt[i];
            edge[iedge].xyz[3*i  ] = tess.xyz[3*ipnt  ];
            edge[iedge].xyz[3*i+1] = tess.xyz[3*ipnt+1];
            edge[iedge].xyz[3*i+2] = tess.xyz[3*ipnt+2];
        }
    }

    /* set the ptyp for each Node, Edge, and Face Point */
    for (inode = 1; inode <= Nnode; inode++) {
        ipnt = node[inode].ipnt;
        tess.ptyp[ipnt] = PNT_NODE | inode;
    }

    for (iedge = 1; iedge <= Nedge; iedge++) {
        for (i = 1; i < edge[iedge].npnt-1; i++) {
            ipnt = edge[iedge].pnt[i];
            tess.ptyp[ipnt] = PNT_EDGE | iedge;
        }
    }

    for (itri = 0; itri < tess.ntri; itri++ ) {
        icolr = tess.ttyp[itri] & TRI_COLOR;

        for (iface = 1; iface <= Nface; iface++) {
            if (face[iface].icol == icolr) {
                ipnt = tess.trip[3*itri  ];
                if (tess.ptyp[ipnt] == 0) tess.ptyp[ipnt] = PNT_FACE | iface;

                ipnt = tess.trip[3*itri+1];
                if (tess.ptyp[ipnt] == 0) tess.ptyp[ipnt] = PNT_FACE | iface;

                ipnt = tess.trip[3*itri+2];
                if (tess.ptyp[ipnt] == 0) tess.ptyp[ipnt] = PNT_FACE | iface;

            }
        }
    }

    /* finally print out the whole structure */
    SPRINT0(1, " inode   ipnt   nedg          x          y          z");
    for (inode = 1; inode <= Nnode; inode++) {
        SPRINT6(1, "%6d %6d %6d %10.4f %10.4f %10.4f", inode,
                node[inode].ipnt,
                node[inode].nedg,
                node[inode].x,
                node[inode].y,
                node[inode].z);
    }

    SPRINT0(1, " iedge   ibeg   iend  ileft  irite   npnt");
    for (iedge = 1; iedge <= Nedge; iedge++) {
        SPRINT6(1, "%6d %6d %6d %6d %6d %6d", iedge,
                edge[iedge].ibeg,
                edge[iedge].iend,
                edge[iedge].ileft,
                edge[iedge].irite,
                edge[iedge].npnt);
    }

    SPRINT0(1, " iface   icol   nedg    edg...");
    for (iface = 1; iface <= Nface; iface++) {
        SPRINT1x(1,  "%6d",      iface      );
        SPRINT1x(1, " %6d", face[iface].icol);
        SPRINT1x(1, " %6d", face[iface].nedg);
        for (iedge = 0; iedge < face[iface].nedg; iedge++) {
            SPRINT1x(1, " %6d", face[iface].edg[iedge]);
        }
        SPRINT0(1, " ");
    }

cleanup:
    FREE(sgmt  );
    FREE(nodnum);

    return status;
}


/*
 ***********************************************************************
 *                                                                     *
 *   generateEgads - generate EGADS Brep                               *
 *                                                                     *
 ***********************************************************************
 */
static int
generateEgads(char   filename[])        /* (in)  name of file to write */
{
    int       status = SUCCESS;

    int       inode, iedge, iface, closed=1, nfaces, ncp, ndata;
    int       periodic;

    int       i, j, ij, header[7], senses[4];
    double    xyz[3], *cpdata=NULL, tdata[4], data[18];
    ego       context, ecurv, esurf, eloop, eshell, ebody, emodel;
    ego       enodes[2], *eedges=NULL, *efaces=NULL;

    ROUTINE(generateEgads);

    /* --------------------------------------------------------------- */

    SPRINT0(1, "Generating EGADS ...");

    status = EG_open(&context);
    CHECK_STATUS(EG_open);

    status = EG_setOutLevel(context, outLevel);
    CHECK_STATUS(EG_setOutLevel);

    /* make each of the Nodes */
    for (inode = 1; inode <= Nnode; inode++) {
        SPRINT1(1, "\n*********\nworking on inode=%d\n*********", inode);

        /* create an array of the point associated with the Node */
        xyz[0] = node[inode].x;
        xyz[1] = node[inode].y;
        xyz[2] = node[inode].z;

        /* make the Node */
        status = EG_makeTopology(context, NULL, NODE, 0,
                                 xyz, 0, NULL, NULL,
                                 &(node[inode].enode));
        SPRINT2(1, "EG_makeTopology(NODE=%2d) -> status=%d", inode, status);
        CHECK_STATUS(EG_makeTopology);
    }

    /* make each of the Edges */
    for (iedge = 1; iedge <= Nedge; iedge++) {
        SPRINT1(1, "\n*********\nworking on iedge=%d\n*********", iedge);

        ncp = edge[iedge].ncp;

        /* create the Curve */
        header[0] = 0;            // bitflag
        header[1] = 3;            // degree
        header[2] = ncp;          // number of control points
        header[3] = ncp+4;        // number of knots

        ndata = (ncp+4) + 3 * ncp;
        MALLOC(cpdata, double, ndata);

        ndata = 0;

        /* knot vector */
        cpdata[ndata++] = 0;
        cpdata[ndata++] = 0;
        cpdata[ndata++] = 0;
        cpdata[ndata++] = 0;

        for (j = 1; j < ncp-3; j++) {
            cpdata[ndata++] = j;
        }

        cpdata[ndata++] = ncp-3;
        cpdata[ndata++] = ncp-3;
        cpdata[ndata++] = ncp-3;
        cpdata[ndata++] = ncp-3;

        /* control points */
        for (j = 0; j < ncp; j++) {
            cpdata[ndata++] = edge[iedge].cp[3*j  ];
            cpdata[ndata++] = edge[iedge].cp[3*j+1];
            cpdata[ndata++] = edge[iedge].cp[3*j+2];
        }

        status = EG_makeGeometry(context, CURVE, BSPLINE, NULL,
                                 header, cpdata, &ecurv);
        CHECK_STATUS(EG_makeGeometry);
        printf("EG_makeGeometry(CURVE) -> status=%d\n", status);

        FREE(cpdata);

        /* create the Edge */
        status = EG_getRange(ecurv, tdata, &periodic);
        CHECK_STATUS(EG_getRange);

        enodes[0] = node[edge[iedge].ibeg].enode;
        enodes[1] = node[edge[iedge].iend].enode;

        status = EG_makeTopology(context, ecurv, EDGE, TWONODE,
                                 tdata, 2, enodes, NULL, &(edge[iedge].eedge));
        CHECK_STATUS(EG_makeTopology);

        SPRINT1(1, "EG_makeTopology(EDGE) -> status=%d", status);
    }

    /* make each of the Faces */
    MALLOC(efaces, ego, Nface);
    nfaces = 0;

    for (iface = 1; iface <= Nface; iface++) {
        SPRINT1(1, "\n*********\nworking on iface=%d\n*********", iface);

        ncp = face[iface].ncp;

        /* see if Points are co-planar */
        if (face[iface].cp == NULL) {

            /* make the list of Edges associated with this Face */
            MALLOC(eedges, ego, face[iface].nedg);

            for (i = 0; i < face[iface].nedg; i++) {
                iedge = abs(face[iface].edg[i]);
                eedges[i] = edge[iedge].eedge;
            }

            /* make the Loop */
            status = EG_makeLoop(face[iface].nedg, eedges, NULL, 0, &eloop);
            printf("EG_makeLoop -> status=%d\n", status);
            CHECK_STATUS(EG_makeLoop);

            /*  if there are any non-NULL entries in eedges, then there
                is more than one Loop */
            for (i = 0; i < face[iface].nedg; i++) {
                if (eedges[i] != NULL) {
                    printf("we have more than one Loop\n");
                    break;
                }
            }

            FREE(eedges);

            /* make the Face */
            status = EG_makeFace(eloop, SFORWARD, NULL, &(face[iface].eface));
            printf("EG_makeFace -> status=%d\n", status);
            CHECK_STATUS(EG_makeFace);

            efaces[nfaces] = face[iface].eface;
            nfaces++;

            /* no further processing for this Face */
            continue;
        }

        /* for now, this only works with Faces bounded by 4 Edges */
        if (face[iface].nedg != 4) {
            printf("Face %d has %d Edges (and is expecting 4)\n", iface, face[iface].nedg);
            closed = 0;
            continue;
        }

        /* create the Surface */
        header[0] = 0;            // bitflag
        header[1] = 3;            // u degree
        header[2] = ncp;          // u number of control points
        header[3] = ncp + 4;      // u number of knots
        header[4] = 3;            // v degree
        header[5] = ncp;          // v number of control points
        header[6] = ncp + 4;      // v number of knots

        ndata = 2 * (ncp+4) + 3 * ncp * ncp;
        MALLOC(cpdata, double, ndata);

        ndata = 0;

        /* knot vectors for u and v */
        for (i = 0; i < 2; i++) {
            cpdata[ndata++] = 0;
            cpdata[ndata++] = 0;
            cpdata[ndata++] = 0;
            cpdata[ndata++] = 0;

            for (j = 1; j < ncp-3; j++) {
                cpdata[ndata++] = j;
            }

            cpdata[ndata++] = ncp-3;
            cpdata[ndata++] = ncp-3;
            cpdata[ndata++] = ncp-3;
            cpdata[ndata++] = ncp-3;
        }

        /* control points */
        for (ij = 0; ij < ncp*ncp; ij++) {
            cpdata[ndata++] = face[iface].cp[3*ij  ];
            cpdata[ndata++] = face[iface].cp[3*ij+1];
            cpdata[ndata++] = face[iface].cp[3*ij+2];
        }

        status = EG_makeGeometry(context, SURFACE, BSPLINE, NULL,
                                 header, cpdata, &esurf);
        CHECK_STATUS(EG_makeGeometry);

        printf("EG_makeGeometry(SURFACE) -> status=%d\n", status);

        FREE(cpdata);

        MALLOC(eedges, ego, 8);
        eedges[0] = edge[abs(face[iface].edg[0])].eedge;
        eedges[1] = edge[abs(face[iface].edg[1])].eedge;
        eedges[2] = edge[abs(face[iface].edg[2])].eedge;
        eedges[3] = edge[abs(face[iface].edg[3])].eedge;

        /* create the PCurves associated with the Edges */
        if (face[iface].edg[0] > 0) {
            senses[0] = SFORWARD;
            data[0]   = 0;
            data[1]   = 0;
            data[2]   = ncp-3;
            data[3]   = 0;
        } else {
            senses[0] = SREVERSE;
            data[0]   = ncp-3;
            data[1]   = 0;
            data[2]   = 3-ncp;
            data[3]   = 0;
        }
        status = EG_makeGeometry(context, PCURVE, LINE, esurf, NULL, data, &(eedges[4]));
        CHECK_STATUS(EG_makeGeometry);
        printf("EG_makeGeometry(PCURVE 0) -> status=%d\n", status);

        if (face[iface].edg[1] > 0) {
            senses[1] = SFORWARD;
            data[0]   = ncp-3;
            data[1]   = 0;
            data[2]   = 0;
            data[3]   = ncp-3;
        } else {
            senses[1] = SREVERSE;
            data[0]   = ncp-3;
            data[1]   = ncp-3;
            data[2]   = 0;
            data[3]   = 3-ncp;
        }
        status = EG_makeGeometry(context, PCURVE, LINE, esurf, NULL, data, &(eedges[5]));
        CHECK_STATUS(EG_makeGeometry);
        printf("EG_makeGeometry(PCURVE 1) -> status=%d\n", status);

        if (face[iface].edg[2] > 0) {
            senses[2] = SFORWARD;
            data[0]   = ncp-3;
            data[1]   = ncp-3;
            data[2]   = 3-ncp;
            data[3]   = 0;
        } else {
            senses[2] = SREVERSE;
            data[0]   = 0;
            data[1]   = ncp-3;
            data[2]   = ncp-3;
            data[3]   = 0;
        }
        status = EG_makeGeometry(context, PCURVE, LINE, esurf, NULL, data, &(eedges[6]));
        CHECK_STATUS(EG_makeGeometry);
        printf("EG_makeGeometry(PCURVE 2) -> status=%d\n", status);

        if (face[iface].edg[3] > 0) {
            senses[3] = SFORWARD;
            data[0]   = 0;
            data[1]   = ncp-3;
            data[2]   = 0;
            data[3]   = 3-ncp;
        } else {
            senses[3] = SREVERSE;
            data[0]   = 0;
            data[1]   = 0;
            data[2]   = 0;
            data[3]   = ncp-3;
        }
        status = EG_makeGeometry(context, PCURVE, LINE, esurf, NULL, data, &(eedges[7]));
        CHECK_STATUS(EG_makeGeometry);
        printf("EG_makeGeometry(PCURVE 3) -> status=%d\n", status);

        /* create a Loop */
        status = EG_makeTopology(context, esurf, LOOP, CLOSED, NULL, 4, eedges, senses, &eloop);
        CHECK_STATUS(EG_makeTopology);
        printf("EG_makeTopology(LOOP) -> status=%d\n", status);

        FREE(eedges);

        /* create the Face */
        status = EG_makeTopology(context, esurf, FACE, SFORWARD, NULL, 1, &eloop, senses, &(face[iface].eface));
        CHECK_STATUS(EG_makeTopology);
        printf("EG_makeTopology(FACE) -> status=%d\n", status);

        efaces[nfaces] = face[iface].eface;
        nfaces++;
    }

    /* assemble the Faces into a Shell */
    printf("before makeTopology(SHELL, nfaces=%d, closed=%d)\n", nfaces, closed);
    if (closed == 1) {
        status = EG_makeTopology(context, NULL, SHELL, CLOSED,
                                 NULL, nfaces, efaces, NULL, &eshell);
    } else {
        status = EG_makeTopology(context, NULL, SHELL, OPEN,
                                 NULL, nfaces, efaces, NULL, &eshell);
    }
    SPRINT1(1, "EG_makeTopology(SHELL) -> status=%d", status);
    CHECK_STATUS(EG_makeTopology);

    SPRINT0(2, "\neshell");

    /* if Shell was closed, make a SolidBody from the Shell */
    if (closed == 1) {
        status = EG_makeTopology(context, NULL, BODY, SOLIDBODY,
                                 NULL, 1, &eshell, NULL, &ebody);
    } else {
        status = EG_makeTopology(context, NULL, BODY, SHEETBODY,
                                 NULL, 1, &eshell, NULL, &ebody);
    }
    SPRINT1(1, "EG_makeTopology(BODY) -> status=%d", status);
    CHECK_STATUS(EG_makeTopology);

    SPRINT0(2, "\nebody");
    if (outLevel > 1) printEgo(ebody);

    /* make a Model for the SolidBody */
    status = EG_makeTopology(context, NULL, MODEL, 0,
                             NULL, 1, &ebody, NULL, &emodel);
    SPRINT1(1, "EG_makeTopology(MODEL) -> status=%d", status);
    CHECK_STATUS(EG_makeTopology);

    SPRINT0(2, "\nemodel");
    if (outLevel > 1) printEgo(emodel);

    /* write out an egads file */
    status = remove(filename);
    if (status == 0) {
        SPRINT1(-1, "WARNING:: file \"%s\" is being overwritten", filename);
    } else {
        SPRINT1(-1, "File \"%s\" is being written", filename);
    }

    status = EG_saveModel(emodel, filename);
    CHECK_STATUS(EG_saveModel);

    /* clean up */
    status = EG_close(context);
    CHECK_STATUS(EG_close);

cleanup:
    FREE(eedges);
    FREE(efaces);

    return status;
}


/*
 ***********************************************************************
 *                                                                     *
 *   generateFits - generate Bspline fits for each Edge and Node       *
 *                                                                     *
 ***********************************************************************
 */
static int
generateFits(int    ncp)                /* (in)  number of control points */
{
    int       status = SUCCESS;

    int       npnt, ipnt, iedge, iface;
    int       i, j, ii, jj, nmin, count, planar;
    double    xx[4], yy[4], zz[4], prod, normf, dotmin;
    double    *Tcloud=NULL, *UVcloud=NULL;

    /* variables needed for multi-threading */
    int       nthread, ithread;
    long      start;
    void      **threads=NULL;
    emp_T     empFitter;

    ROUTINE(generateFits);

    /* --------------------------------------------------------------- */

    SPRINT1(1, "Generating Fits (ncp=%d) ...", ncp);

    /* make each of the Edges */
    for (iedge = 1; iedge <= Nedge; iedge++) {
        SPRINT1(1, "\n*********\nfitting iedge=%d\n*********", iedge);

        npnt = edge[iedge].npnt;

        /* allocate space for control points */
        edge[iedge].ncp = ncp;
        MALLOC(edge[iedge].cp, double, 3*ncp);

        /* set the control points at its boundaries */
        edge[iedge].cp[0      ] = edge[iedge].xyz[0       ];
        edge[iedge].cp[1      ] = edge[iedge].xyz[1       ];
        edge[iedge].cp[2      ] = edge[iedge].xyz[2       ];

        edge[iedge].cp[3*ncp-3] = edge[iedge].xyz[3*npnt-3];
        edge[iedge].cp[3*ncp-2] = edge[iedge].xyz[3*npnt-2];
        edge[iedge].cp[3*ncp-1] = edge[iedge].xyz[3*npnt-1];

        /* allocate space for the T-parameters associated with the cloud */
        MALLOC(Tcloud, double, npnt);

        /* fit the cloud of points  with ncp control points */
        status = fit1dCloud(npnt, edge[iedge].xyz, ncp, edge[iedge].cp,
                            Tcloud, &normf, &dotmin, &nmin, NULL);
        CHECK_STATUS(fit1dCloud);

        printf("fit1dCloud(npnt=%d, ncp=%d) -> status=%d,  normf=%12.4e,  dotmin=%.4f,  nmin=%d\n",
               npnt, ncp, status, normf, dotmin, nmin);

#ifdef GRAFIC
        /* plot the fit */
        status = plotCurve(npnt, edge[iedge].xyz, Tcloud,
                           ncp,  edge[iedge].cp, normf, dotmin, nmin);
        printf("plotCurve -> status=%d\n", status);
        CHECK_STATUS(plotCurve);
#endif

        FREE(Tcloud);
    }

    /* make each of the Faces */
    for (iface = 1; iface <= Nface; iface++) {
        SPRINT1(1, "\n*********\ninitializing iface=%d\n*********", iface);

        face[iface].done = 0;

        /* see if Points are co-planar by computing triple product
           by each set of 4 consecutive Points */
        count  = 0;
        planar = 1;
        for (ipnt = 0; ipnt < tess.npnt; ipnt++) {
            if (tess.ptyp[ipnt] == (PNT_FACE | iface)) {
                xx[count] = tess.xyz[3*ipnt  ];
                yy[count] = tess.xyz[3*ipnt+1];
                zz[count] = tess.xyz[3*ipnt+2];
                count++;

                if (count == 4) {
                    prod = (xx[1]-xx[0]) * ( (yy[2]-yy[0]) * (zz[3]-zz[0])
                                           - (zz[2]-zz[0]) * (yy[3]-yy[0]))
                         + (yy[1]-yy[0]) * ( (zz[2]-zz[0]) * (xx[3]-xx[0])
                                           - (xx[2]-xx[0]) * (zz[3]-zz[0]))
                         + (zz[1]-zz[0]) * ( (xx[2]-xx[0]) * (yy[3]-yy[0])
                                           - (yy[2]-yy[0]) * (xx[3]-xx[0]));

                    if (fabs(prod) > EPS06) {
                        planar = 0;
                        break;
                    }

                    xx[0] = xx[1];  yy[0] = yy[1];  zz[0] = zz[1];
                    xx[1] = xx[2];  yy[1] = yy[2];  zz[1] = zz[2];
                    xx[2] = xx[3];  yy[2] = yy[3];  zz[2] = zz[3];
                    count--;
                }
            }
        }

        if (planar == 1) {
            SPRINT0(1, "skipping (planar)");
            continue;
        }

        /* for now, this only works with Faces bounded by 4 Edges */
        if (face[iface].nedg != 4) {
            printf("Face %d has %d Edges (and is expecting 4)\n", iface, face[iface].nedg);
            continue;
        }

        /* count the number of points associated with this Face */
        npnt = 0;
        for (i = 0; i < face[iface].nedg; i++) {
            iedge = abs(face[iface].edg[i]);
            npnt += edge[iedge].npnt;
        }
        for (ipnt = 0; ipnt < tess.npnt; ipnt++) {
            if (tess.ptyp[ipnt] == (PNT_FACE | iface)) {
                npnt++;
            }
        }
        printf("npnt=%d\n", npnt);

        assert(npnt > 0);

        /* create an array of the discrete points associated with this Face */
        face[iface].npnt = npnt;
        MALLOC(face[iface].xyz, double, 3*npnt);

        npnt = 0;
        for (i = 0; i < face[iface].nedg; i++) {
            iedge = abs(face[iface].edg[i]);
            for (ipnt = 0; ipnt < edge[iedge].npnt; ipnt++) {
                face[iface].xyz[3*npnt  ] = edge[iedge].xyz[3*ipnt  ];
                face[iface].xyz[3*npnt+1] = edge[iedge].xyz[3*ipnt+1];
                face[iface].xyz[3*npnt+2] = edge[iedge].xyz[3*ipnt+2];
                npnt++;
            }
        }
        for (ipnt = 0; ipnt < tess.npnt; ipnt++) {
            if (tess.ptyp[ipnt] == (PNT_FACE | iface)) {
                face[iface].xyz[3*npnt  ] = tess.xyz[3*ipnt  ];
                face[iface].xyz[3*npnt+1] = tess.xyz[3*ipnt+1];
                face[iface].xyz[3*npnt+2] = tess.xyz[3*ipnt+2];
                npnt++;
            }
        }

        /* allocate space for control points */
        face[iface].ncp = ncp;
        MALLOC(face[iface].cp, double, 3*ncp*ncp);

        /* set the control points at its boundaries */
        // south
        iedge = face[iface].edg[0];
        printf("extracting south control points from iedge=%5d\n", iedge);
        if (edge[abs(iedge)].ncp != ncp) {
            printf("mismatch 0\n");
        } else if (iedge > 0) {
            j  = 0;
            for (i = 0; i < ncp; i++) {
                ii = i;
                face[iface].cp[3*(i+ncp*j)  ] = edge[iedge].cp[3*ii  ];
                face[iface].cp[3*(i+ncp*j)+1] = edge[iedge].cp[3*ii+1];
                face[iface].cp[3*(i+ncp*j)+2] = edge[iedge].cp[3*ii+2];
            }
        } else {
            iedge = -iedge;
            j  = 0;
            for (i = 0; i < ncp; i++) {
                ii = ncp - 1 - i;
                face[iface].cp[3*(i+ncp*j)  ] = edge[iedge].cp[3*ii  ];
                face[iface].cp[3*(i+ncp*j)+1] = edge[iedge].cp[3*ii+1];
                face[iface].cp[3*(i+ncp*j)+2] = edge[iedge].cp[3*ii+2];
            }
        }

        // east
        iedge = face[iface].edg[1];
        printf("extracting east  control points from iedge=%5d\n", iedge);
        if (edge[abs(iedge)].ncp != ncp) {
            printf("mismatch 1\n");
        } else if (iedge > 0) {
            i = ncp - 1;
            for (j = 0; j < ncp; j++) {
                jj = j;
                face[iface].cp[3*(i+ncp*j)  ] = edge[iedge].cp[3*jj  ];
                face[iface].cp[3*(i+ncp*j)+1] = edge[iedge].cp[3*jj+1];
                face[iface].cp[3*(i+ncp*j)+2] = edge[iedge].cp[3*jj+2];
            }
        } else {
            iedge = -iedge;
            i = ncp - 1;
            for (j = 0; j < ncp; j++) {
                jj = ncp - 1 - j;
                face[iface].cp[3*(i+ncp*j)  ] = edge[iedge].cp[3*jj  ];
                face[iface].cp[3*(i+ncp*j)+1] = edge[iedge].cp[3*jj+1];
                face[iface].cp[3*(i+ncp*j)+2] = edge[iedge].cp[3*jj+2];
            }
        }

        // north
        iedge = face[iface].edg[2];
        printf("extracting north control points from iedge=%5d\n", iedge);
        if (edge[abs(iedge)].ncp != ncp) {
            printf("mismatch 2\n");
        } else if (iedge > 0) {
            j = ncp - 1;
            for (i = 0; i < ncp; i++) {
                ii = ncp - 1 - i;
                face[iface].cp[3*(i+ncp*j)  ] = edge[iedge].cp[3*ii  ];
                face[iface].cp[3*(i+ncp*j)+1] = edge[iedge].cp[3*ii+1];
                face[iface].cp[3*(i+ncp*j)+2] = edge[iedge].cp[3*ii+2];
            }
        } else {
            iedge = -iedge;
            j = ncp - 1;
            for (i = 0; i < ncp; i++) {
                ii = i;
                face[iface].cp[3*(i+ncp*j)  ] = edge[iedge].cp[3*ii  ];
                face[iface].cp[3*(i+ncp*j)+1] = edge[iedge].cp[3*ii+1];
                face[iface].cp[3*(i+ncp*j)+2] = edge[iedge].cp[3*ii+2];
            }
        }

        // west
        iedge = face[iface].edg[3];
        printf("extracting west  control points from iedge=%5d\n", iedge);
        if (edge[abs(iedge)].ncp != ncp) {
            printf("mismatch 3\n");
        } else if (iedge > 0) {
            i = 0;
            for (j = 0; j < ncp; j++) {
                jj = ncp - 1 - j;
                face[iface].cp[3*(i+ncp*j)  ] = edge[iedge].cp[3*jj  ];
                face[iface].cp[3*(i+ncp*j)+1] = edge[iedge].cp[3*jj+1];
                face[iface].cp[3*(i+ncp*j)+2] = edge[iedge].cp[3*jj+2];
            }
        } else {
            iedge = -iedge;
            i = 0;
            for (j = 0; j < ncp; j++) {
                jj = j;
                face[iface].cp[3*(i+ncp*j)  ] = edge[iedge].cp[3*jj  ];
                face[iface].cp[3*(i+ncp*j)+1] = edge[iedge].cp[3*jj+1];
                face[iface].cp[3*(i+ncp*j)+2] = edge[iedge].cp[3*jj+2];
            }
        }
    }

    /* set up for multi-threading */
    empFitter.mutex  = NULL;
    empFitter.master = EMP_ThreadID();

    nthread = EMP_Init(&start);
    SPRINT1(1, "\n*********\nstarting multi-threaded fits with %d threads\n*********", nthread);

    if (nthread > 1) {

        /* create the mutex to handle list synchronization */
        empFitter.mutex = EMP_LockCreate();
        if (empFitter.mutex == NULL) {
            printf("EMPerror:: mutex creation = NULL\n");
            nthread = 1;
        } else {

            /* get storage for extra threads */
            threads = (void**) malloc((nthread-1)*sizeof(void*));
            if (threads == NULL) {
                EMP_LockDestroy(empFitter.mutex);
                nthread = 1;
            }
        }
    }

    /* create the threads and get going */
    if (threads != NULL) {
        for (ithread = 0; ithread < nthread-1; ithread++) {
            threads[ithread] = EMP_ThreadCreate(empFit2dCloud, &empFitter);
            if (threads[ithread] == NULL) {
                printf("EMPerror:: creating thread %d", ithread+1);
            }
        }
    }

    /* now run the fitter from the original thread */
    empFit2dCloud(&empFitter);

    /* wait for all others to return */
    if (threads != NULL) {
        for (ithread = 0; ithread < nthread-1; ithread++) {
            if (threads[ithread] != NULL) {
                EMP_ThreadWait(threads[ithread]);
            }
        }
    }

    /* cleanup the threads */
    if (threads != NULL) {
        for (ithread = 0; ithread < nthread-1; ithread++) {
            if (threads[ithread] != NULL) {
                EMP_ThreadDestroy(threads[ithread]);
            }
        }
    }

    if (empFitter.mutex != NULL) {
        EMP_LockDestroy(empFitter.mutex);
    }

    if (threads != NULL) {
        free(threads);
    }

cleanup:
    FREE(Tcloud );
    FREE(UVcloud);

    return status;
}


/*
 ***********************************************************************
 *                                                                     *
 *   empFit2dCloud - does fitting in a thread                          *
 *                                                                     *
 ***********************************************************************
 */
static void
empFit2dCloud(void *struc)              /* (both) emp structure */
{
    int    status, iface, jface, nmin, nmax;
    long   ID;
    double *UVcloud=NULL, normf;

    emp_T  *empFitter = (emp_T *)struc;

    ROUTINE(empFit2dCloud);

    /* --------------------------------------------------------------- */

    /* get our identifier */
    ID = EMP_ThreadID();
    if (ID == empFitter->master) {
        printf("ID %12lx: is master\n", ID);
    } else {
        printf("ID %12lx: start thread\n", ID);
    }

    /* look for work */
    while (1) {

        /* figure out which Face to do.  use a mutex to make sure
         that only one thread does one Face */
        if (empFitter->mutex != NULL) {
            EMP_LockSet(empFitter->mutex);
        }

        /* look for the Face with the most Points in the cloud */
        iface = -1;
        nmax  = -1;

        for (jface = 1; jface <= Nface; jface++) {
            if (face[jface].done == 0) {
                if (face[jface].npnt > nmax) {
                    iface = jface;
                    nmax  = face[jface].npnt;
                }
            }
        }

        /* release the mutex */
        if (empFitter->mutex != NULL) {
            EMP_LockRelease(empFitter->mutex);
        }

        /* if there are no Faces to process, break out of loop */
        if (iface < 1) break;

        /* mark this Face as being done (so that it does not get
           scheduled again) */
        face[iface].done = 1;
        printf("ID %12lx: iface %3d has %5d points\n", ID, iface, face[iface].npnt);

        /* do not process planar faces */
        if (face[iface].npnt == 0) {
            printf("ID %12lx: iface %3d skipped\n", ID, iface);
            continue;
        }

        /* fit the data */
        MALLOC(UVcloud, double, 2*face[iface].npnt);

        status = fit2dCloud(face[iface].npnt,
                            face[iface].xyz,
                            face[iface].ncp,
                            face[iface].cp,
                            UVcloud, &normf, &nmin, NULL);

        printf("ID %12lx: iface %3d complete with status=%d, normf=%12.4e, nmin=%d\n",
               ID, iface, status, normf, nmin);

#ifdef GRAFIC
        /* plot the fit */
        if (nthread == 1) {
            status = plotSurface(face[iface].npnt, face[iface].xyz, UVcloud,
                                 face[iface].ncp,  face[iface].cp, normf, nmin);
            printf("plotSurface -> status=%d\n", status);
            CHECK_STATUS(plotSurface);
        }
#endif

        FREE(UVcloud);
    }


    /* all work is done, so exit (if not the master) */
    if (ID != empFitter->master) {
        printf("ID %12lx: stop  thread\n", ID);
        EMP_ThreadExit();
    }

cleanup:
    return;
}


/*
 ***********************************************************************
 *                                                                     *
 *   browserMessage - called when client sends a message to the server *
 *                                                                     *
 ***********************************************************************
 */
void
browserMessage(void    *wsi,
               char    *text,
  /*@unused@*/ int     lena)
{
    int       status;

    ROUTINE(browserMessage);

    /* --------------------------------------------------------------- */

    /* process the Message */
    processMessage(text);

    /* send the response */
    SPRINT1(2, "response-> %s", response);
    wv_sendText(wsi, response);

    if (strlen(sgFocusData) > 0) {
        SPRINT1(3, "sgFocus-> %s", sgFocusData);
        wv_sendText(wsi, sgFocusData);

        /* nullify meta data so that it does not get sent again */
        sgFocusData[0] = '\0';
    }

    /* there is no key */
    status = wv_setKey(cntxt,   0, NULL,      lims[0], lims[1], NULL            );
    if (status != SUCCESS) {
        SPRINT1(3, "wv_setKet -> status=%d", status);
    }
}


/*
 ***********************************************************************
 *                                                                     *
 *   processMessage - process the message and create the response      *
 *                                                                     *
 ***********************************************************************
 */
static void
processMessage(char    *text)
{
    char      *arg1=NULL, *arg2=NULL, *arg3=NULL;
    char      *arg4=NULL, *arg5=NULL;

    int       status, ipnt, itri;

    ROUTINE(processMessage);

#define CHECK_FOR_ERROR(COMMAND,SUBCOMMAND) \
    if (status < 0) {                                                                             \
        snprintf(response, max_resp_len, "ERROR:: %s :: %s -> status=%d", #COMMAND, #SUBCOMMAND, status); \
        response_len = strlen(response);                                                          \
        goto cleanup;                                                                             \
    }

    /* --------------------------------------------------------------- */

    MALLOC(arg1, char, MAX_EXPR_LEN);
    MALLOC(arg2, char, MAX_EXPR_LEN);
    MALLOC(arg3, char, MAX_EXPR_LEN);
    MALLOC(arg4, char, MAX_EXPR_LEN);
    MALLOC(arg5, char, MAX_EXPR_LEN);

    SPRINT1(1, "==> processMessage(text=%s)", text);

    /* initialize the response */
    response_len = 0;
    response[0] = '\0';

    /* NO-OP */
    if (strlen(text) == 0) {

    /* "identify;" */
    } else if (strncmp(text, "identify;", 9) == 0) {

        /* build the response */
        snprintf(response, max_resp_len, "identify;Slugs;");
        response_len = strlen(response);

    /* "automaticLinks;" */
    } else if (strncmp(text, "automaticLinks;", 16) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* make an undo copy */
        status = storeUndo();
        CHECK_FOR_ERROR(automaticLinks, storeUndo);

        /* remove old Links and create new ones between colors */
        status = removeLinks(&tess);
        CHECK_FOR_ERROR(automaticLinks, removeLinks);

        status = makeLinks(&tess);
        CHECK_FOR_ERROR(automaticLinks, makeLinks);

        /* build the response */
        snprintf(response, max_resp_len, "automaticLinks;okay");
        response_len = strlen(response);

        /* set flags */
        Links_pend = 1;

    /* "bridgeToPoint;x;y;z;" */
    } else if (strncmp(text, "bridgeToPoint;", 14) == 0) {
        double xloc=0, yloc=0, zloc=0;

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* make an undo copy */
        status = storeUndo();
        CHECK_FOR_ERROR(bridgeToPoint, storeUndo);

        /* extract arguments */
        if (getToken(text, 1, arg1)) xloc  = strtod(arg1, NULL);
        if (getToken(text, 2, arg2)) yloc  = strtod(arg2, NULL);
        if (getToken(text, 3, arg3)) zloc  = strtod(arg3, NULL);
        SPRINT3(3, "xloc=%f  yloc=%f  zloc=%f", xloc, yloc, zloc);

        /* find the closest Triangle */
        itri = closestTriangle(xloc, yloc, zloc);
        SPRINT1(3, "closestTriangle -> itri=%d", itri);

        /* bridge between Triangle and current Point */
        status = bridgeToPoint(&tess, itri, CurPt_index);
        CHECK_FOR_ERROR(bridgeToPoint, bridgeToPoint);

        /* reset the current Point */
        CurPt_index = -1;

        /* build the response */
        snprintf(response, max_resp_len, "bridgeToPoint;okay");
        response_len = strlen(response);

        /* set flags */
        CurPt_index = -1;

        Tris_pend  = 1;
        Hangs_pend = 1;
        Links_pend = 1;
        CurPt_pend = 1;

    /* "colorTriangles;x;y;z;icolor;" */
    } else if (strncmp(text, "colorTriangles;", 15) == 0) {
        int    icolr=0;
        double xloc=0, yloc=0, zloc=0;

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* make an undo copy */
        status = storeUndo();
        CHECK_FOR_ERROR(colorTriangles, storeUndo);

        /* extract arguments */
        if (getToken(text, 1, arg1)) xloc  = strtod(arg1, NULL);
        if (getToken(text, 2, arg2)) yloc  = strtod(arg2, NULL);
        if (getToken(text, 3, arg3)) zloc  = strtod(arg3, NULL);
        if (getToken(text, 4, arg4)) icolr = strtol(arg4, NULL, 10);
        SPRINT4(3, "xloc=%f  yloc=%f  zloc=%f  icolr=%d", xloc, yloc, zloc, icolr);

        /* find the closest Triangle */
        itri = closestTriangle(xloc, yloc, zloc);
        SPRINT1(3, "closestTriangle -> itri=%d", itri);

        /* color the Triangle and its neighbors */
        status = colorTriangles(&tess, itri, icolr);
        CHECK_FOR_ERROR(colorTriangles, colorTriangles);

        /* update the latest color number */
        tess.ncolr = MAX(tess.ncolr, icolr);

        /* build the response */
        snprintf(response, max_resp_len, "colorTriangles;%d;okay", tess.ncolr);
        response_len = strlen(response);

        /* set flags */
        CurPt_index = -1;

        Tris_pend  = 1;
        Links_pend = 1;
        Hangs_pend = 1;
        CurPt_pend = 1;

    /* "cutTriangles;icolr;itype;xloc;yloc;zloc;" */
    } else if (strncmp(text, "cutTriangles;", 13) == 0) {
        int    jpnt, icolr=-1, itype=0;
        double xloc=0, yloc=0, zloc=0, data[4];

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* make an undo copy */
        status = storeUndo();
        CHECK_FOR_ERROR(cutTriangles, storeUndo);

        /* extract arguments */
        if (getToken(text, 1, arg1)) icolr = strtol(arg1, NULL, 10);
        if (getToken(text, 2, arg2)) itype = strtol(arg2, NULL, 10);
        if (getToken(text, 3, arg3)) xloc  = strtod(arg3, NULL);
        if (getToken(text, 4, arg4)) yloc  = strtod(arg4, NULL);
        if (getToken(text, 5, arg5)) zloc  = strtod(arg5, NULL);
        SPRINT5(3, "icolr=%d, type=%d, xloc=%f, yloc=%f, zloc=%f", icolr, itype, xloc, yloc, zloc);

        /* find the closest Point */
        ipnt = closestPoint(xloc, yloc, zloc);
        SPRINT1(3, "closestPoint -> ipnt=%d", ipnt);
        jpnt = CurPt_index;
        SPRINT1(3, "currentPoint -> jpnt=%d", jpnt);

        /* set up data matrix that defines cutting plane
           independent of Y and Z (constant X) */
        if        (itype == 0) {
            data[0] = tess.xyz[3*ipnt  ];
            data[1] = -1;
            data[2] =  0;
            data[3] =  0;
        /* independent of X and Z (constant Y) */
        } else if (itype == 1) {
            data[0] = tess.xyz[3*ipnt+1];
            data[1] =  0;
            data[2] = -1;
            data[3] =  0;
        /* independent of X and Y (constant Z) */
        } else if (itype == 2) {
            data[0] = tess.xyz[3*ipnt+2];
            data[1] =  0;
            data[2] =  0;
            data[3] = -1;
        /* independent of X */
        } else if (itype == 3 && jpnt >= 0) {
            data[0] = 1;
            data[1] = 0;
            data[2] = (tess.xyz[3*ipnt+2] - tess.xyz[3*jpnt+2])
                    / (tess.xyz[3*ipnt+1] * tess.xyz[3*jpnt+2]
                     - tess.xyz[3*jpnt+1] * tess.xyz[3*ipnt+2]);
            data[3] = (tess.xyz[3*jpnt+1] - tess.xyz[3*ipnt+1])
                    / (tess.xyz[3*ipnt+1] * tess.xyz[3*jpnt+2]
                     - tess.xyz[3*jpnt+1] * tess.xyz[3*ipnt+2]);
        /* independent of Y */
        } else if (itype == 4 && jpnt >= 0) {
            data[0] = 1;
            data[1] = (tess.xyz[3*jpnt+2] - tess.xyz[3*ipnt+2])
                    / (tess.xyz[3*ipnt+2] * tess.xyz[3*jpnt  ]
                     - tess.xyz[3*jpnt+2] * tess.xyz[3*ipnt  ]);
            data[2] = 0;
            data[3] = (tess.xyz[3*ipnt  ] - tess.xyz[3*jpnt  ])
                    / (tess.xyz[3*ipnt+2] * tess.xyz[3*jpnt  ]
                     - tess.xyz[3*jpnt+2] * tess.xyz[3*ipnt  ]);
        /* independent of Z */
        } else if (itype == 5 && jpnt >= 0) {
            data[0] = 1;
            data[1] = (tess.xyz[3*ipnt+1] - tess.xyz[3*jpnt+1])
                    / (tess.xyz[3*ipnt  ] * tess.xyz[3*jpnt+1]
                     - tess.xyz[3*jpnt  ] * tess.xyz[3*ipnt+1]);
            data[2] = (tess.xyz[3*jpnt  ] - tess.xyz[3*ipnt  ])
                    / (tess.xyz[3*ipnt  ] * tess.xyz[3*jpnt+1]
                     - tess.xyz[3*jpnt  ] * tess.xyz[3*ipnt+1]);
            data[3] = 0;
        /* data vector given */
        } else {
            printf("Enter data: "); fflush(stdout);
            fscanf(stdin, "%lf %lf %lf %lf", &(data[0]), &(data[1]), &(data[2]), &(data[3]));
        }

        /* cut the Triangles */
        status = cutTriangles(&tess, icolr, data);
        CHECK_FOR_ERROR(cutTriangles, cutTriangles);

        /* build the response */
        snprintf(response, max_resp_len, "cutTriangles;%d;okay", tess.ncolr);
        response_len = strlen(response);

        /* set flags */
        CurPt_index = -1;

        Tris_pend  = 1;
        Links_pend = 1;
        Hangs_pend = 1;
        CurPt_pend = 1;

    /* "deleteTriangle;x;y;z;" */
    } else if (strncmp(text, "deleteTriangle;", 15) == 0) {
        double xloc=0, yloc=0, zloc=0;

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* make an undo copy */
        status = storeUndo();
        CHECK_FOR_ERROR(deleteTriangle, storeUndo);

        /* extract arguments */
        if (getToken(text, 1, arg1)) xloc  = strtod(arg1, NULL);
        if (getToken(text, 2, arg2)) yloc  = strtod(arg2, NULL);
        if (getToken(text, 3, arg3)) zloc  = strtod(arg3, NULL);
        SPRINT3(3, "xloc=%f  yloc=%f  zloc=%f", xloc, yloc, zloc);

        /* find the closest Triangle */
        itri = closestTriangle(xloc, yloc, zloc);
        SPRINT1(3, "closestTriangle -> itri=%d", itri);

        /* delete the Triangle */
        status = deleteTriangle(&tess, itri);
        CHECK_FOR_ERROR(deleteTriangle, deleteTriangle);

        /* reset the current Point */
        CurPt_index = -1;

        /* build the response */
        snprintf(response, max_resp_len, "deleteTriangle;okay");
        response_len = strlen(response);

        /* set flags */
        CurPt_index = -1;

        Tris_pend  = 1;
        Hangs_pend = 1;
        CurPt_pend = 1;

    /* "fillHole;x;y;z;" */
    } else if (strncmp(text, "fillHole;", 9) == 0) {
        double xloc=0, yloc=0, zloc=0;

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* make an undo copy */
        status = storeUndo();
        CHECK_FOR_ERROR(fillHole, storeUndo);

        /* extract arguments */
        if (getToken(text, 1, arg1)) xloc  = strtod(arg1, NULL);
        if (getToken(text, 2, arg2)) yloc  = strtod(arg2, NULL);
        if (getToken(text, 3, arg3)) zloc  = strtod(arg3, NULL);
        SPRINT3(3, "xloc=%f  yloc=%f  zloc=%f", xloc, yloc, zloc);

        /* find the closest Point */
        ipnt = closestPoint(xloc, yloc, zloc);
        SPRINT1(3, "closestPoint -> ipnt=%d", ipnt);

        /* fill the hole adjacent to this Point */
        status = fillLoop(&tess, ipnt);
        CHECK_FOR_ERROR(fillHole, fillLoop);

        /* reset the current Point */
        CurPt_index = -1;

        /* build the response */
        snprintf(response, max_resp_len, "fillHole;okay");
        response_len = strlen(response);

        /* set flags */
        CurPt_index = -1;

        Tris_pend  = 1;
        Hangs_pend = 1;
        CurPt_pend = 1;

    /* "generateBrep;" */
    } else if (strncmp(text, "generateBrep;", 13) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        status = generateBrep();
        CHECK_FOR_ERROR(generateBrep, generateBrep);

        /* build the response */
        snprintf(response, max_resp_len, "generateBrep;okay");
        response_len = strlen(response);

        /* set flags */
        CurPt_index = -1;

        Tris_pend  = 1;
        Hangs_pend = 1;
        CurPt_pend = 1;

    /* "generateEgads;filename;ncp;" */
    } else if (strncmp(text, "generateEgads;", 14) == 0) {
        int  ncp = 5;

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract argument */
        if (getToken(text, 1, arg1)) strncpy(filename, arg1, 80);
        if (getToken(text, 2, arg2)) ncp = strtol(arg2, NULL, 10);

        /* generate the fits */
        status = generateFits(ncp);
        CHECK_FOR_ERROR(generateFits, generateFits);

        /* generate the egads file */
        status = generateEgads(filename);
        CHECK_FOR_ERROR(generateEgads, generateEgads);

        /* build the response */
        snprintf(response, max_resp_len, "generateEgads;okay");
        response_len = strlen(response);

        /* set flags */
        CurPt_index = -1;

        Tris_pend  = 1;
        Hangs_pend = 1;
        CurPt_pend = 1;

    /* "joinPoints;x;y;z;" */
    } else if (strncmp(text, "joinPoints;", 11) == 0) {
        double xloc=0, yloc=0, zloc=0;

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* make an undo copy */
        status = storeUndo();
        CHECK_FOR_ERROR(joinPoints, storeUndo);

        /* extract arguments */
        if (getToken(text, 1, arg1)) xloc = strtod(arg1, NULL);
        if (getToken(text, 2, arg2)) yloc = strtod(arg2, NULL);
        if (getToken(text, 3, arg3)) zloc = strtod(arg3, NULL);
        SPRINT3(3, "xloc=%f  yloc=%f  zloc=%f", xloc, yloc, zloc);

        /* find the closest Point */
        ipnt = closestPoint(xloc, yloc, zloc);
        SPRINT1(3, "closestPoint -> ipnt=%d", ipnt);

        /* join the Points */
        status = joinPoints(&tess, ipnt, CurPt_index);
        CHECK_FOR_ERROR(joinPoints, joinPoints);

        /* build the response */
        snprintf(response, max_resp_len, "joinPoints;okay");
        response_len = strlen(response);

        /* set flags */
        CurPt_index = ipnt;

        Tris_pend  = 1;
        Hangs_pend = 1;
        Links_pend = 1;
        CurPt_pend = 1;

    /* "linkToPoint;x;y;z;" */
    } else if (strncmp(text, "linkToPoint;", 11) == 0) {
        double xloc=0, yloc=0, zloc=0;

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* make an undo copy */
        if (CurPt_index > 0) {
            status = storeUndo();
            CHECK_FOR_ERROR(linkToPoint, storeUndo);
        }

        /* extract arguments */
        if (getToken(text, 1, arg1)) xloc = strtod(arg1, NULL);
        if (getToken(text, 2, arg2)) yloc = strtod(arg2, NULL);
        if (getToken(text, 3, arg3)) zloc = strtod(arg3, NULL);
        SPRINT3(3, "xloc=%f  yloc=%f  zloc=%f", xloc, yloc, zloc);

        /* find the closest Point */
        ipnt = closestPoint(xloc, yloc, zloc);
        SPRINT1(3, "closestPoint -> ipnt=%d", ipnt);

        /* build the Links */
        if (CurPt_index >= 0) {
            status = createLinks(&tess, CurPt_index, ipnt);
            CHECK_FOR_ERROR(linkToPoint, createLinks);

            /* build the response */
            snprintf(response, max_resp_len, "linkToPoint;%d;okay", CurPt_index);
            response_len = strlen(response);
        } else {
            CurPt_index = ipnt;

            snprintf(response, max_resp_len, "pickPoint;%d;okay", CurPt_index);
            response_len = strlen(response);
        }

        /* set flags */
        CurPt_index = ipnt;

        Links_pend = 1;
        CurPt_pend = 1;

    /* "markCreases;angdeg;" */
    } else if (strncmp(text, "markCreases;", 12) == 0) {
        double angdeg = 45;

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* make an undo copy */
        status = storeUndo();
        CHECK_FOR_ERROR(colorTriangles, storeUndo);

        /* extract argument */
        if (getToken(text, 1, arg1)) angdeg = strtod(arg1, NULL);

        /* mark the creases */
        status = detectCreases(&tess, angdeg);
        CHECK_FOR_ERROR(detectCreases, detectCreases);

        /* build the response */
        snprintf(response, max_resp_len, "markCreases;okay");
        response_len = strlen(response);

        /* set flags */
        CurPt_index = -1;

        Links_pend = 1;
        CurPt_pend = 1;

    /* "identifyPoint;x;y;z" */
    } else if (strncmp(text, "identifyPoint;", 14) == 0) {
        double xloc=0, yloc=0, zloc=0;

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        if (getToken(text, 1, arg1)) xloc = strtod(arg1, NULL);
        if (getToken(text, 2, arg2)) yloc = strtod(arg2, NULL);
        if (getToken(text, 3, arg3)) zloc = strtod(arg3, NULL);
        SPRINT3(3, "xloc=%f  yloc=%f  zloc=%f", xloc, yloc, zloc);

        /* find the closest Point */
        ipnt = closestPoint(xloc, yloc, zloc);
        SPRINT1(3, "closestPoint -> CurPt_index=%d", ipnt);

        /* print out information */
        printf("ipnt=%6d\n", ipnt);
        for (itri = 0; itri < tess.ntri; itri++) {
            if (tess.trip[3*itri  ] == ipnt ||
                tess.trip[3*itri+1] == ipnt ||
                tess.trip[3*itri+2] == ipnt   ) {
                printf("     itri=%6d  (%6d %6d %6d)  (%6d %6d %6d)\n", itri,
                    tess.trip[3*itri], tess.trip[3*itri+1], tess.trip[3*itri+2],
                    tess.trit[3*itri], tess.trit[3*itri+1], tess.trit[3*itri+2]);
            }
        }

        /* build the response */
        snprintf(response, max_resp_len, "identifyPoint;%d;okay", CurPt_index);
        response_len = strlen(response);

        /* set flags */

    /* "pickPoint;x;y;z;" */
    } else if (strncmp(text, "pickPoint;", 10) == 0) {
        double xloc=0, yloc=0, zloc=0;

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        if (getToken(text, 1, arg1)) xloc = strtod(arg1, NULL);
        if (getToken(text, 2, arg2)) yloc = strtod(arg2, NULL);
        if (getToken(text, 3, arg3)) zloc = strtod(arg3, NULL);
        SPRINT3(3, "xloc=%f  yloc=%f  zloc=%f", xloc, yloc, zloc);

        /* find the closest Point */
        CurPt_index = closestPoint(xloc, yloc, zloc);
        SPRINT1(3, "closestPoint -> CurPt_index=%d", CurPt_index);

        /* build the response */
        snprintf(response, max_resp_len, "pickPoint;%d;okay", CurPt_index);
        response_len = strlen(response);

        /* set flags */
        CurPt_pend = 1;

    /* "scribeToPoint;x;y;z;" */
    } else if (strncmp(text, "scribeToPoint;", 14) == 0) {
        double xloc=0, yloc=0, zloc=0;

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* make an undo copy */
        if (CurPt_index > 0) {
            status = storeUndo();
            CHECK_FOR_ERROR(linkToPoint, storeUndo);
        }

        /* extract arguments */
        if (getToken(text, 1, arg1)) xloc = strtod(arg1, NULL);
        if (getToken(text, 2, arg2)) yloc = strtod(arg2, NULL);
        if (getToken(text, 3, arg3)) zloc = strtod(arg3, NULL);
        SPRINT3(3, "xloc=%f  yloc=%f  zloc=%f", xloc, yloc, zloc);

        /* find the closest Point */
        ipnt = closestPoint(xloc, yloc, zloc);
        SPRINT1(3, "closestPoint -> ipnt=%d", ipnt);

        /* build the Links */
        if (CurPt_index >= 0) {
            status = scribe(&tess, CurPt_index, ipnt);
            CHECK_FOR_ERROR(scribePoints, scribePoints);

            status = createLinks(&tess, CurPt_index, ipnt);
            CHECK_FOR_ERROR(linkToPoint, createLinks);

            /* build the response */
            snprintf(response, max_resp_len, "scribeToPoint;%d;okay", CurPt_index);
            response_len = strlen(response);
        } else {
            CurPt_index = ipnt;

            snprintf(response, max_resp_len, "scribeToPoint;%d;okay", CurPt_index);
            response_len = strlen(response);
        }

        /* set flags */
        CurPt_index = ipnt;

        Tris_pend  = 1;
        Links_pend = 1;
        CurPt_pend = 1;

    /* "identifyTriangle;x;y;z;icolor;" */
    } else if (strncmp(text, "identifyTriangle;", 17) == 0) {
        double xloc=0, yloc=0, zloc=0;

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        if (getToken(text, 1, arg1)) xloc  = strtod(arg1, NULL);
        if (getToken(text, 2, arg2)) yloc  = strtod(arg2, NULL);
        if (getToken(text, 3, arg3)) zloc  = strtod(arg3, NULL);
        SPRINT3(3, "xloc=%f  yloc=%f  zloc=%f", xloc, yloc, zloc);

        /* find the closest Triangle */
        itri = closestTriangle(xloc, yloc, zloc);
        SPRINT1(3, "closestTriangle -> itri=%d", itri);

        /* build the response */
        snprintf(response, max_resp_len, "identifyTriangle;%d;okay", itri);
        response_len = strlen(response);

        /* set flags */
        CurPt_index = -1;

        Tris_pend  = 1;
        Links_pend = 1;
        Hangs_pend = 1;
        CurPt_pend = 1;

    /* "undo;" */
    } else if (strncmp(text, "undo;", 5) == 0) {
        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* if an undo exists, use it now */
        SPRINT1(3, "tess_undo.ntri=%d", tess_undo.ntri);
        if (tess_undo.ntri > 0) {
            status = freeTess(&tess);
            CHECK_FOR_ERROR(undo, freeTess);

            status = copyTess(&tess_undo, &tess);
            CHECK_FOR_ERROR(undo, copyTess);

            status = freeTess(&tess_undo);
            CHECK_FOR_ERROR(undo, freeTess);

            /* build the response */
            snprintf(response, max_resp_len, "undo;okay");
            response_len = strlen(response);

            /* set flags */
            CurPt_index = -1;

            Tris_pend  = 1;
            Hangs_pend = 1;
            Links_pend = 1;
            CurPt_pend = 1;

        } else {
            /* build the response */
            snprintf(response, max_resp_len, "ERROR:: nothing to undo");
            response_len = strlen(response);
        }

    /* "writeStlFile;filename;" */
    } else if (strncmp(text, "writeStlFile;", 13) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract argument */
        if (getToken(text, 1, arg1) == 0) arg1[0] = '\0';

        status = writeStlBinary(&tess, arg1);
        CHECK_FOR_ERROR(writeStlFile, writeStlBinary);

        /* build the response */
        snprintf(response, max_resp_len, "writeStlFile;okay");
        response_len = strlen(response);
    }

cleanup:
    FREE(arg5);
    FREE(arg4);
    FREE(arg3);
    FREE(arg2);
    FREE(arg1);
}


/*
 ***********************************************************************
 *                                                                     *
 *   getToken - get a token from a string                              *
 *                                                                     *
 ***********************************************************************
 */
static int
getToken(char   *text,                  /* (in)  full text */
         int    nskip,                  /* (in)  tokens to skip */
         char   *token)                 /* (out) token */
{
    int    lentok, i, count, iskip;

    ROUTINE(getToken);

    /* --------------------------------------------------------------- */

    token[0] = '\0';
    lentok   = 0;

    /* count the number of semi-colons */
    count = 0;
    for (i = 0; i < strlen(text); i++) {
        if (text[i] == ';') {
            count++;
        }
    }

    if (count < nskip+1) return 0;

    /* skip over nskip tokens */
    i = 0;
    for (iskip = 0; iskip < nskip; iskip++) {
        while (text[i] != ';') {
            i++;
        }
        i++;
    }

    /* extract the token we are looking for */
    while (text[i] != ';') {
        token[lentok++] = text[i++];
        token[lentok  ] = '\0';
    }

    return strlen(token);
}


/*
 ***********************************************************************
 *                                                                     *
 *   closestPoint - find closest Point to given coordinates            *
 *                                                                     *
 ***********************************************************************
 */
static int
closestPoint(double xloc,
             double yloc,
             double zloc)
{
    int status = 0;

    int    ipnt;
    double dbest, dtest;

    ROUTINE(closestPoint);

    /* --------------------------------------------------------------- */

    dbest = pow(xloc - tess.xyz[0], 2)
          + pow(yloc - tess.xyz[1], 2)
          + pow(zloc - tess.xyz[2], 2);

    for (ipnt = 1; ipnt < tess.npnt; ipnt++) {
        dtest = pow(xloc - tess.xyz[3*ipnt  ], 2)
              + pow(yloc - tess.xyz[3*ipnt+1], 2)
              + pow(zloc - tess.xyz[3*ipnt+2], 2);

        if (dtest < dbest) {
            dbest  = dtest;
            status = ipnt;
        }
    }

    return status;
}


/*
 ***********************************************************************
 *                                                                     *
 *   closestTriangle - find closest Triangle to given coordinates      *
 *                                                                     *
 ***********************************************************************
 */
static int
closestTriangle(double xloc,
                double yloc,
                double zloc)
{
    int status = 0;

    int    itri, ip0, ip1, ip2;
    double xcent, ycent, zcent, dbest, dtest;

    ROUTINE(closestTriangle);

    /* --------------------------------------------------------------- */

    ip0 = tess.trip[0];
    ip1 = tess.trip[1];
    ip2 = tess.trip[2];

    xcent = (tess.xyz[3*ip0  ] + tess.xyz[3*ip1  ] + tess.xyz[3*ip2  ]) / 3;
    ycent = (tess.xyz[3*ip0+1] + tess.xyz[3*ip1+1] + tess.xyz[3*ip2+1]) / 3;
    zcent = (tess.xyz[3*ip0+2] + tess.xyz[3*ip1+2] + tess.xyz[3*ip2+2]) / 3;

    dbest = pow(xloc - xcent, 2)
          + pow(yloc - ycent, 2)
          + pow(zloc - zcent, 2);

    for (itri = 0; itri < tess.ntri; itri++) {
        if ((tess.ttyp[itri] & TRI_ACTIVE) == 0) continue;

        ip0 = tess.trip[3*itri  ];
        ip1 = tess.trip[3*itri+1];
        ip2 = tess.trip[3*itri+2];

        xcent = (tess.xyz[3*ip0  ] + tess.xyz[3*ip1  ] + tess.xyz[3*ip2  ]) / 3;
        ycent = (tess.xyz[3*ip0+1] + tess.xyz[3*ip1+1] + tess.xyz[3*ip2+1]) / 3;
        zcent = (tess.xyz[3*ip0+2] + tess.xyz[3*ip1+2] + tess.xyz[3*ip2+2]) / 3;

        dtest = pow(xloc - xcent, 2)
              + pow(yloc - ycent, 2)
              + pow(zloc - zcent, 2);

        if (dtest < dbest) {
            dbest  = dtest;
            status = itri;
        }
    }

    return status;
}


/*
 ***********************************************************************
 *                                                                     *
 *   storeUndo - store an undo snapshot                                *
 *                                                                     *
 ***********************************************************************
 */
static int
storeUndo()
{
    int       status = SUCCESS;

    ROUTINE(storeUndo);

    /* --------------------------------------------------------------- */

    /* if there is previous undo information, free it now */
    if (tess_undo.ntri > 0) {
        status = freeTess(&tess_undo);
        if (status != SUCCESS) goto cleanup;
    }

    /* make a copy */
    status = copyTess(&tess, &tess_undo);

cleanup:
    return status;
}


/*
 ***********************************************************************
 *                                                                     *
 *   setColor - convert hex-code into a color                          *
 *                                                                     *
 ***********************************************************************
 */
static void
setColor(int    rgb,
         float  color[])
{
    ROUTINE(setColor);

    /* --------------------------------------------------------------- */

    color[0] = (float)((rgb & 0xff0000) / 0x10000) / 255.0;
    color[1] = (float)((rgb & 0x00ffff) / 0x00100) / 255.0;
    color[2] = (float)((rgb & 0x0000ff)          ) / 255.0;
}


#ifdef FOO
/*
 ***********************************************************************
 *                                                                     *
 *   spec_col - return color for a given scalar value                  *
 *                                                                     *
 ***********************************************************************
 */
static void
spec_col(float  scalar,
         float  lims[],
         float  color[])
{
    int   indx;
    float frac;

    ROUTINE(spec_col);

    /* --------------------------------------------------------------- */

    if (lims[0] == lims[1]) {
        color[0] = 0.0;
        color[1] = 1.0;
        color[2] = 0.0;
    } else if (scalar <= lims[0]) {
        color[0] = color_map[0];
        color[1] = color_map[1];
        color[2] = color_map[2];
    } else if (scalar >= lims[1]) {
        color[0] = color_map[3*255  ];
        color[1] = color_map[3*255+1];
        color[2] = color_map[3*255+2];
    } else {
        frac  = 255.0 * (scalar - lims[0]) / (lims[1] - lims[0]);
        if (frac < 0  ) frac = 0;
        if (frac > 255) frac = 255;
        indx  = frac;
        frac -= indx;
        if (indx == 255) {
            indx--;
            frac += 1.0;
        }

        color[0] = frac * color_map[3*(indx+1)  ] + (1.0-frac) * color_map[3*indx  ];
        color[1] = frac * color_map[3*(indx+1)+1] + (1.0-frac) * color_map[3*indx+1];
        color[2] = frac * color_map[3*(indx-1)+2] + (1.0-frac) * color_map[3*indx+2];
    }
}
#endif


/*
 ************************************************************************
 *                                                                      *
 *   printEgo - print EGADS topology for 5 levels                       *
 *                                                                      *
 ************************************************************************
 */
void
printEgo(ego    obj)                    /* (in)  ego to start */
{
    ego    eref0,    eref1,    eref2,    eref3,    eref4,   eref5;
    int    oclass0,  oclass1,  oclass2,  oclass3,  oclass4, oclass5;
    int    mtype0,   mtype1,   mtype2,   mtype3,   mtype4,  mtype5;
    int    nchild0,  nchild1,  nchild2,  nchild3,  nchild4, nchild5;
    int    ichild0,  ichild1,  ichild2,  ichild3,  ichild4, i;
    double data0[4], data1[4], data2[4], data3[4], data4[4], data5[4];
    ego    *ebodys0, *ebodys1, *ebodys2, *ebodys3, *ebodys4, *ebodys5;
    int    *senses0, *senses1, *senses2, *senses3, *senses4, *senses5;
    int    status, count=0;
    ego    topref, prev, next, refobj, context, prev1, next1;

    /* names take from egadsTypes.h */
    char *classname[27] = {"contxt",       "transform",  "tessellation",  "nil",
                           "empty",        "reference",  "ERROR 6",       "ERROR 7",
                           "ERROR 8",      "ERROR 9",    "pcurve",        "curve",
                           "surface",      "ERROR 13",   "ERROR 14",      "ERROR 15",
                           "ERROR 16",     "ERROR 17",   "ERROR 18",      "ERROR 19",
                           "node",         "edge",       "loop",          "face",
                           "shell",        "body",       "model"};

    /* pcurves and curves */
    char *mtypename1[10] = {"ERROR 0",     "line",       "circle",        "ellipse",
                            "parabola",    "hyperbola",  "trimmed",       "bezier",
                            "bspline",     "offset"};

    /* surfaces */
    char *mtypename2[12] = {"ERROR 0",     "plane",      "spherical",     "cylindrical",
                            "revolution",  "toroidal",   "trimmed",       "bezier",
                            "bspline",     "offset",    "conical",       "extrusion"};

    /* faces */
    char *mtypename3[ 3] = {"sreverse",    "nomtype",    "sforward"};

    /* other topology */
    char *mtypename4[10] = {"nomtype",     "onenode",    "twonode",       "open",
                            "closed",      "degenerate", "wirebody",      "facebody",
                            "sheetbody",   "solidbody"};

    char mtypename[32];

    /* --------------------------------------------------------------- */

    if (obj == NULL) {
        SPRINT0(0, "NULL");
        return;
    }

    status = EG_getContext(obj, &context);
    if (status < 0) {
        SPRINT1(0, "EG_getContext -> status=%d", status);
    }

    status = EG_getTopology(obj, &eref0, &oclass0, &mtype0,
                            data0, &nchild0, &ebodys0, &senses0);
    if (status != EGADS_SUCCESS) {
        EG_getInfo(obj, &oclass0, &mtype0, &topref, &prev, &next);
        nchild0 = -1;
        eref0   = topref;
    }

    if (oclass0 != NODE) {
        if (oclass0 == PCURVE || oclass0 == CURVE) {
            strcpy(mtypename, mtypename1[mtype0]);
        } else if (oclass0 == SURFACE) {
            strcpy(mtypename, mtypename2[mtype0]);
        } else if (oclass0 == FACE) {
            strcpy(mtypename, mtypename3[mtype0+1]);
        } else if (oclass0 == EDGE  || oclass0 == LOOP ||
                   oclass0 == SHELL || oclass0 == BODY   ) {
            strcpy(mtypename, mtypename4[mtype0]);
        } else {
            strcpy(mtypename, "");
        }
        SPRINT6(0, "oclass0=%3d (%s)  mtype0=%3d (%s)  obj=%llx,  eref0=%llx",
                oclass0, classname[oclass0], mtype0, mtypename,
                (long long)obj, (long long)eref0);
        if (oclass0 == LOOP) {
            SPRINT0x(0, "< senses=");
            for (i = 0; i < nchild0; i++) {
                SPRINT1x(0, "%2d ", senses0[i]);
            }
            SPRINT0(0, " ");
        }
    } else {
        SPRINT7(0,
                "oclass0=%3d (%s)  mtype0=%3d,  obj=%llx,  data0=%20.10e %20.10e %20.10e",
                oclass0, classname[oclass0], mtype0, (long long)obj, data0[0],
                data0[1], data0[2]);
    }

    next = obj->tref;
    prev = NULL;
    while (next != NULL) {
        refobj = next->attrs;
        if (refobj != context) {
            count++;
            EG_getInfo(refobj, &oclass1, &mtype1, &topref, &prev1, &next1);
            SPRINT5(0, "< refcount=%2d, refobj=%llx  (oclass=%2d (%s) mtype=%2d)",
                    count, (long long)refobj, oclass1, classname[oclass1], mtype1);
        }
        prev = next;
        next = prev->blind;
    }

    for (ichild0 = 0; ichild0 < nchild0; ichild0++) {
        status = EG_getTopology(ebodys0[ichild0], &eref1, &oclass1, &mtype1,
                                data1, &nchild1, &ebodys1, &senses1);
        if (status != EGADS_SUCCESS) {
            EG_getInfo(obj, &oclass1, &mtype1, &topref, &prev, &next);
            nchild1 = -1;
            eref1   = topref;
        }

        if (oclass1 != NODE) {
            if (oclass1 == PCURVE || oclass1 == CURVE) {
                strcpy(mtypename, mtypename1[mtype1]);
            } else if (oclass1 == SURFACE) {
                strcpy(mtypename, mtypename2[mtype1]);
            } else if (oclass1 == FACE) {
                strcpy(mtypename, mtypename3[mtype1+1]);
            } else if (oclass1 == EDGE  || oclass1 == LOOP ||
                       oclass1 == SHELL || oclass1 == BODY   ) {
                strcpy(mtypename, mtypename4[mtype1]);
            } else {
                strcpy(mtypename, "");
            }
            SPRINT6(0, ". oclass1=%3d (%s)  mtype1=%3d (%s)  obj=%llx,  eref1=%llx",
                    oclass1, classname[oclass1], mtype1, mtypename,
                    (long long)(ebodys0[ichild0]), (long long)eref1);
            if (oclass1 == LOOP) {
                SPRINT0x(0, ". < senses=");
                for (i = 0; i < nchild1; i++) {
                    SPRINT1x(0, "%2d ", senses1[i]);
                }
                SPRINT0(0, " ");
            }
        } else {
            SPRINT7(0, ". oclass1=%3d (%s)  mtype1=%3d,  obj=%llx,  data1=%20.10e %20.10e %20.10e",
                    oclass1, classname[oclass1], mtype1, (long long)(ebodys0[ichild0]),
                    data1[0], data1[1], data1[2]);
        }

        for (ichild1 = 0; ichild1 < nchild1; ichild1++) {
            status = EG_getTopology(ebodys1[ichild1], &eref2, &oclass2, &mtype2,
                                    data2, &nchild2, &ebodys2, &senses2);
            if (status != EGADS_SUCCESS) {
                EG_getInfo(obj, &oclass2, &mtype2, &topref, &prev, &next);
                nchild2 = -1;
                eref2   = topref;
            }

            if (oclass2 != NODE) {
                if (oclass2 == PCURVE || oclass2 == CURVE) {
                    strcpy(mtypename, mtypename1[mtype2]);
                } else if (oclass2 == SURFACE) {
                    strcpy(mtypename, mtypename2[mtype2]);
                } else if (oclass2 == FACE) {
                    strcpy(mtypename, mtypename3[mtype2+1]);
                } else if (oclass2 == EDGE  || oclass2 == LOOP ||
                           oclass2 == SHELL || oclass2 == BODY   ) {
                    strcpy(mtypename, mtypename4[mtype2]);
                } else {
                    strcpy(mtypename, "");
                }
                SPRINT6(0, ". . oclass2=%3d (%s)  mtype2=%3d (%s)  obj=%llx,  eref2=%llx",
                        oclass2, classname[oclass2], mtype2, mtypename,
                        (long long)(ebodys1[ichild1]), (long long)eref2);
                if (oclass2 == LOOP) {
                    SPRINT0x(0, ". . < senses=");
                    for (i=0; i < nchild2; i++) {
                        SPRINT1x(0, "%2d ", senses2[i]);
                    }
                    SPRINT0(0, " ");
                }
            } else {
                SPRINT7(0, ". . oclass2=%3d (%s)  mtype2=%3d,  obj=%llx,  data2=%20.10e %20.10e %20.10e",
                        oclass2, classname[oclass2], mtype2,
                        (long long)(ebodys1[ichild1]), data2[0], data2[1], data2[2]);
            }

            for (ichild2 = 0; ichild2 < nchild2; ichild2++) {
                status = EG_getTopology(ebodys2[ichild2], &eref3, &oclass3, &mtype3,
                               data3, &nchild3, &ebodys3, &senses3);
                if (status != EGADS_SUCCESS) {
                    EG_getInfo(obj, &oclass3, &mtype3, &topref, &prev, &next);
                    nchild3 = -1;
                    eref3   = topref;
                }

                if (oclass3 != NODE) {
                    if (oclass3 == PCURVE || oclass3 == CURVE) {
                        strcpy(mtypename, mtypename1[mtype3]);
                    } else if (oclass3 == SURFACE) {
                        strcpy(mtypename, mtypename2[mtype3]);
                    } else if (oclass3 == FACE) {
                        strcpy(mtypename, mtypename3[mtype3+1]);
                    } else if (oclass3 == EDGE  || oclass3 == LOOP ||
                               oclass3 == SHELL || oclass3 == BODY   ) {
                        strcpy(mtypename, mtypename4[mtype3]);
                    } else {
                        strcpy(mtypename, "");
                    }
                    SPRINT6(0, ". . . oclass3=%3d (%s)  mtype3=%3d (%s)  obj=%llx,  eref3=%llx",
                            oclass3, classname[oclass3], mtype3, mtypename,
                            (long long)(ebodys2[ichild2]), (long long)eref3);
                    if (oclass3 == LOOP) {
                        SPRINT0x(0, ". . . < senses=");
                        for (i=0; i < nchild3; i++) {
                            SPRINT1x(0, "%2d ", senses3[i]);
                        }
                        SPRINT0(0, " ");
                    }
                } else {
                    SPRINT7(0, ". . . oclass3=%3d (%s)  mtype3=%3d,  obj=%llx,  data3=%20.10e %20.10e %20.10e",
                            oclass3, classname[oclass3], mtype3,
                            (long long)(ebodys2[ichild2]), data3[0], data3[1], data3[2]);
                }

                for (ichild3 = 0; ichild3 < nchild3; ichild3++) {
                    status = EG_getTopology(ebodys3[ichild3], &eref4, &oclass4, &mtype4,
                                   data4, &nchild4, &ebodys4, &senses4);
                    if (status != EGADS_SUCCESS) {
                        EG_getInfo(obj, &oclass4, &mtype4, &topref, &prev, &next);
                        nchild4 = -1;
                        eref4   = topref;
                    }

                    if (oclass4 != NODE) {
                        if (oclass4 == PCURVE || oclass4 == CURVE) {
                            strcpy(mtypename, mtypename1[mtype4]);
                        } else if (oclass4 == SURFACE) {
                            strcpy(mtypename, mtypename2[mtype4]);
                        } else if (oclass4 == FACE) {
                            strcpy(mtypename, mtypename3[mtype4+1]);
                        } else if (oclass4 == EDGE  || oclass4 == LOOP ||
                                   oclass4 == SHELL || oclass4 == BODY   ) {
                            strcpy(mtypename, mtypename4[mtype4]);
                        } else {
                            strcpy(mtypename, "");
                        }
                        SPRINT6(0, ". . . . oclass4=%3d (%s)  mtype4=%3d (%s)  obj=%llx,  eref4=%llx",
                                oclass4, classname[oclass4], mtype4, mtypename,
                                (long long)(ebodys3[ichild3]), (long long)eref4);
                        if (oclass4 == LOOP) {
                            SPRINT0x(0, ". . . . < senses=");
                            for (i=0; i < nchild4; i++) {
                                SPRINT1x(0, "%2d ", senses4[i]);
                            }
                            SPRINT0(0, " ");
                        }
                    } else {
                        SPRINT7(0, ". . . . oclass4=%3d (%s)  mtype4=%3d,  obj=%llx,  data4=%20.10e %20.10e %20.10e",
                                oclass4, classname[oclass4], mtype4,
                                (long long)(ebodys3[ichild3]), data4[0], data4[1], data4[2]);
                    }

                    for (ichild4 = 0; ichild4 < nchild4; ichild4++) {
                        status = EG_getTopology(ebodys4[ichild4], &eref5, &oclass5, &mtype5,
                                                data5, &nchild5, &ebodys5, &senses5);
                        if (status != EGADS_SUCCESS) {
                            EG_getInfo(obj, &oclass5, &mtype5, &topref, &prev, &next);
                            nchild5 = -1;
                            eref5   = topref;
                        }

                        if (oclass5 != NODE) {
                            if (oclass5 == PCURVE || oclass5 == CURVE) {
                                strcpy(mtypename, mtypename1[mtype5]);
                            } else if (oclass5 == SURFACE) {
                                strcpy(mtypename, mtypename2[mtype5]);
                            } else if (oclass5 == FACE) {
                                strcpy(mtypename, mtypename3[mtype5+1]);
                            } else if (oclass5 == EDGE  || oclass5 == LOOP ||
                                       oclass5 == SHELL || oclass5 == BODY   ) {
                                strcpy(mtypename, mtypename4[mtype5]);
                            } else {
                                strcpy(mtypename, "");
                            }
                            SPRINT6(0, ". . . . . oclass5=%3d (%s)  mtype5=%3d (%s)  obj=%llx,  eref5=%llx",
                                    oclass5, classname[oclass5], mtype5, mtypename,
                                    (long long)(ebodys4[ichild4]), (long long)eref5);
                            if (oclass5 == LOOP) {
                                SPRINT0x(0, ". . . . . < senses=");
                                for (i=0; i < nchild5; i++) {
                                    SPRINT1x(0, "%2d ", senses5[i]);
                                }
                                SPRINT0(0, " ");
                            }
                        } else {
                            SPRINT7(0, ". . . . . oclass5=%3d (%s)  mtype5=%3d,  obj=%llx,  data5=%20.10e %20.10e %20.10e",
                                    oclass5, classname[oclass5], mtype5,
                                    (long long)(ebodys4[ichild4]), data5[0], data5[1], data5[2]);
                        }
                    }
                }
            }
        }
    }
}
