#include <string.h>
#include <math.h>
#include "miscUtils.h" //Bring in misc. utility functions
#include "meshUtils.h" // Bring in mesh utility functions

#include "feaTypes.h"  // Bring in FEA structures
#include "meshTypes.h" // Bring in mesh structures
#include "capsTypes.h" // Bring in CAPS types

#include "feaUtils.h"

#ifdef WIN32
#define strcasecmp  stricmp
#endif

//#define           MIN(A,B)        (((A) < (B)) ? (A) : (B))
//#define           MAX(A,B)        (((A) < (B)) ? (B) : (A))

// Convert an EGADS body to a boundary element model, modified by Ryan Durscher (AFRL)
// from code originally written by John Dannenhoffer @ Syracuse University, patterned after code
// written by Bob Haimes  @ MIT
int fea_bodyToBEM(ego    ebody,         		// (in)  EGADS Body
                  double paramTess[3],     		// (in)  Tessellation parameters
                  int    edgePointMin,  		// (in)  minimum points along any Edge
                  int    edgePointMax,  		// (in)  maximum points along any Edge
				  int    noQuadFaces,			// (in)  only do tris-for faces
				  mapAttrToIndexStruct *attrMap,// (in)  map from CAPSGroup names to indexes
				  mapAttrToIndexStruct *coordSystemMap,// (in)  map from CoordSystem names to indexes
				  genFEAMeshStruct *feaMesh)    // (out) FEA mesh structure
{

	// Note: This functions in really only setup to support a single body when
	// it comes to assigning PIDs

	int status = 0; // Function return status

	int i, j, face, edge, patch; // Indexing

	// Body entities
	int numNode = 0, numEdge = 0, numFace = 0;
	ego *enodes=NULL, *eedges=NULL, *efaces=NULL;

	// Meshing
	int numPoint = 0, numTri = 0;
	const int  *triConn = NULL, *triNeighbor = NULL; // Triangle

	int numPatch = 0; // Patching
	int n1, n2;

	int gID; // Global id

	const double *xyz, *uv;

	const int *ptype = NULL, *pindex = NULL, *pvindex = NULL, *pbounds = NULL;

    int       periodic, nchange, oclass, mtype, nchild, *senses;

    // Edge point distributions
    int   *points=NULL, *isouth=NULL, *ieast=NULL, *inorth=NULL, *iwest=NULL;
    double    params[3], bbox[6], size, range[2], arclen, data[4];
    double    *rpos=NULL;
    ego  eref, *echilds, eloop, tempBody;


    int 	  bodySubType = 0; // Body classification

    int pointType, pointTopoIndex;
    double xyzPoint[3];

    // Attributues
    const char *attrName;
    int         attrIndex, coordSystemIndex;

    int numElement = 0; // Number of elements
    int coordID = 0; // Default coordinate id for mesh

    /* --------------------------------------------------------------- */

    printf("Creating FEA BEM\n");

    // What type of BODY do we have?
    status = EG_getTopology(ebody, &eref, &oclass, &bodySubType, data, &nchild, &echilds, &senses);
    if (status < EGADS_SUCCESS) goto cleanup;

    // Get number of Nodes, Edges, and Faces in ebody
    status = EG_getBodyTopos(ebody, NULL, NODE, &numNode, &enodes);
    if (status < EGADS_SUCCESS) goto cleanup;

    status = EG_getBodyTopos(ebody, NULL, EDGE, &numEdge, &eedges);
    if (status != EGADS_SUCCESS) goto cleanup;

    status = EG_getBodyTopos(ebody, NULL, FACE, &numFace, &efaces);
    if (status < EGADS_SUCCESS) goto cleanup;

    /* determine the nominal number of points along each Edge */
    points = (int    *) EG_alloc((numEdge+1)     *sizeof(int   ));
    rpos  = (double *) EG_alloc((edgePointMax)*sizeof(double));
    if (points == NULL || rpos == NULL) {
        status = EGADS_MALLOC;
        printf("\tError in fea_bodyToBEM: EG_alloc\n");
        goto cleanup;
    }

    status = EG_getBoundingBox(ebody, bbox);
    if (status < EGADS_SUCCESS) {
    	printf("\tError in fea_bodyToBEM: EG_getBoundingBox\n");
    	goto cleanup;
    }

    size = sqrt( (bbox[3] - bbox[0]) * (bbox[3] - bbox[0])
                +(bbox[4] - bbox[1]) * (bbox[4] - bbox[1])
                +(bbox[5] - bbox[2]) * (bbox[5] - bbox[2]));

    params[0] = paramTess[0] * size;
    params[1] = paramTess[1] * size;
    params[2] = paramTess[2];

    status = EG_attributeAdd(ebody, ".tParams", ATTRREAL, 3,
                             NULL, params, NULL);
    if (status < EGADS_SUCCESS) {
    	printf("\tError in fea_bodyToBEM: EG_attributeAdd\n");
    	goto cleanup;
    }

    for (i = 1; i <= numEdge; i++) {
        status = EG_getRange(eedges[i-1], range, &periodic);
        if (status < EGADS_SUCCESS) {
        	printf("\tError in fea_bodyToBEM: EG_getRange\n");
        	goto cleanup;
        }

        status = EG_arcLength(eedges[i-1], range[0], range[1], &arclen);
        if (status < EGADS_SUCCESS) {
        	printf("\tError in fea_bodyToBEM: EG_arcLength\n");
        	goto cleanup;
        }

        //points[i] = MIN(MAX(MAX(edgePointMin,2), (int)(1+arclen/params[0])), edgePointMax);
        points[i] = (int) min_DoubleVal(
        					 	 max_DoubleVal(
        					 			 max_DoubleVal( (double) edgePointMin, 2.0),
										 (double) (1+arclen/params[0])),
								 (double) edgePointMax);
    }

    /* make arrays for "opposite" sides of four-sided Faces (with only one loop) */
    isouth = (int *) EG_alloc((numFace+1)*sizeof(int));
    ieast  = (int *) EG_alloc((numFace+1)*sizeof(int));
    inorth = (int *) EG_alloc((numFace+1)*sizeof(int));
    iwest  = (int *) EG_alloc((numFace+1)*sizeof(int));

    if (isouth == NULL ||
    	ieast == NULL  ||
        inorth == NULL ||
		iwest == NULL   ) {

    	status = EGADS_MALLOC;
        printf("\tError in fea_bodyToBEM: EG_alloc\n");
        goto cleanup;
    }

    for (i = 1; i <= numFace; i++) {
        isouth[i] = 0;
        ieast [i] = 0;
        inorth[i] = 0;
        iwest [i] = 0;

        status = EG_getTopology(efaces[i-1],
                                &eref, &oclass, &mtype, data,
                                &nchild, &echilds, &senses);
        if (status < EGADS_SUCCESS) goto cleanup;

        if (nchild != 1) continue;

        eloop = echilds[0];
        status = EG_getTopology(eloop, &eref, &oclass, &mtype, data,
                                &nchild, &echilds, &senses);
        if (status < EGADS_SUCCESS) goto cleanup;

        if (nchild != 4) continue;

        isouth[i] = status = EG_indexBodyTopo(ebody, echilds[0]);
        if (status < EGADS_SUCCESS) goto cleanup;

        ieast[i]  = status = EG_indexBodyTopo(ebody, echilds[1]);
        if (status < EGADS_SUCCESS) goto cleanup;

        inorth[i] = status = EG_indexBodyTopo(ebody, echilds[2]);
        if (status < EGADS_SUCCESS) goto cleanup;

        iwest[i]  = status = EG_indexBodyTopo(ebody, echilds[3]);
        if (status < EGADS_SUCCESS) goto cleanup;
    }

    /* make "opposite" sides of four-sided Faces (with only one loop) match */
    nchange = 1;
    for (i = 0; i < 20; i++) {
        nchange = 0;

        for (face = 1; face <= numFace; face++) {
            if (isouth[face] <= 0 || ieast[face] <= 0 ||
                inorth[face] <= 0 || iwest[face] <= 0   ) continue;

            if        (points[iwest[face]] < points[ieast[face]]) {
                points[iwest[face]] = points[ieast[face]];
                nchange++;

            } else if (points[ieast[face]] < points[iwest[face]]) {
                points[ieast[face]] = points[iwest[face]];
                nchange++;
            }

            if        (points[isouth[face]] < points[inorth[face]]) {
                points[isouth[face]] = points[inorth[face]];
                nchange++;
            } else if (points[inorth[face]] < points[isouth[face]]) {
                points[inorth[face]] = points[isouth[face]];
                nchange++;
            }
        }
        if (nchange == 0) break;
    }
    if (nchange > 0) {
        status = -999;
        goto cleanup;
    }

    /* mark the Edges with points[iedge] evenly-spaced points */
    for (edge = 1; edge <= numEdge; edge++) {
        for (i = 1; i < points[edge]-1; i++) {

        	rpos[i-1] = (double)(i) / (double)(points[edge]-1);
        }

        if (points[edge] == 2) {
            i = 0;
            status = EG_attributeAdd(eedges[edge-1],
                                     ".rPos", ATTRINT, 1, &i, NULL, NULL);
            if (status < EGADS_SUCCESS) goto cleanup;
        } else {
            status = EG_attributeAdd(eedges[edge-1],
                                     ".rPos", ATTRREAL, points[edge]-2, NULL, rpos, NULL);
            if (status < EGADS_SUCCESS) goto cleanup;
        }
    }

    // Make tessellation
    status = EG_makeTessBody(ebody, params, &feaMesh->bodyTessMapping.egadsTess);
    if (status != EGADS_SUCCESS) goto cleanup;


    // Make Quads on each four-sided Face
    params[0] = 0;
    params[1] = 0;
    params[2] = 0;

    for (face = 1; face <= numFace; face++) {
    	if (iwest[face] <= 0) continue;

    	if (noQuadFaces == (int) false) {

    		status = EG_attributeAdd(efaces[face-1], "_makeQuads", ATTRINT, 1, &j, NULL, NULL);
    		if (status < EGADS_SUCCESS) goto cleanup;

    		status = EG_makeQuads(feaMesh->bodyTessMapping.egadsTess, params, face);
    		if (status < EGADS_SUCCESS) goto cleanup;
    	}
    }

    // Get number of point in the tessellation
    status = EG_statusTessBody(feaMesh->bodyTessMapping.egadsTess, &tempBody, &i, &feaMesh->numNodes);
    if (status != EGADS_SUCCESS) goto cleanup;

    if (feaMesh->xyz          != NULL) EG_free(feaMesh->xyz);
    if (feaMesh->xyzCoordID   != NULL) EG_free(feaMesh->xyzCoordID);
    if (feaMesh->xyzAttrIndex != NULL) EG_free(feaMesh->xyzAttrIndex);

    feaMesh->xyz   	      = (double *) EG_alloc(3*feaMesh->numNodes*sizeof(double));
    feaMesh->xyzCoordID   = (int *)    EG_alloc(  feaMesh->numNodes*sizeof(int));
    feaMesh->xyzAttrIndex = (int *)    EG_alloc(  feaMesh->numNodes*sizeof(int));

    if (feaMesh->xyz == NULL ||
        feaMesh->xyzCoordID == NULL ||
	    feaMesh->xyzAttrIndex == NULL) {

    	status = EGADS_MALLOC;
    	goto cleanup;
    }

    // Fill up the coordinates
    for (i = 0; i < feaMesh->numNodes; i++) {

    	status = EG_getGlobal(feaMesh->bodyTessMapping.egadsTess, i+1, &pointType, &pointTopoIndex, xyzPoint);
		if (status != EGADS_SUCCESS) goto cleanup;

		//printf("Point (%d) : Type = %d, pointTopoIndex = %d\n", i, pointType, pointTopoIndex);

		feaMesh->xyz[3*i + 0] = xyzPoint[0];
    	feaMesh->xyz[3*i + 1] = xyzPoint[1];
    	feaMesh->xyz[3*i + 2] = xyzPoint[2];
    	feaMesh->xyzCoordID[i] = coordID;

    	// Get attribute index on entity
    	attrIndex = CAPSMAGIC;
    	if (pointType == 0) { // Node

    		status = retrieve_CAPSGroupAttr(enodes[pointTopoIndex-1], &attrName);
    		if (status == CAPS_SUCCESS) {
    			status = get_mapAttrToIndexIndex(attrMap, attrName, &attrIndex);
    			if (status != CAPS_SUCCESS && status != CAPS_NOTFOUND && status != EGADS_NOTFOUND) goto cleanup;
    		}

    	} else if (pointType > 0 ) { // Edge

    		status = retrieve_CAPSGroupAttr(eedges[pointTopoIndex-1], &attrName);
    		if (status == CAPS_SUCCESS) {
    			status = get_mapAttrToIndexIndex(attrMap, attrName, &attrIndex);
    			if (status != CAPS_SUCCESS && status != CAPS_NOTFOUND && status != EGADS_NOTFOUND) goto cleanup;
    		}

    	} else { // Face

    		status = retrieve_CAPSGroupAttr(efaces[pointTopoIndex-1], &attrName);
    		if (status == CAPS_SUCCESS) {
    			status = get_mapAttrToIndexIndex(attrMap, attrName, &attrIndex);
    			if (status != CAPS_SUCCESS && status != CAPS_NOTFOUND && status != EGADS_NOTFOUND) goto cleanup;
    		}
    	}

    	feaMesh->xyzAttrIndex[i] = attrIndex;
    }

    // Need to add a loop to fillup xyzAttrIndex if not filled by looping through faces

    // Fill element information
    feaMesh->numElement1DPID = 0;
    feaMesh->numElement2DPID = 0;
    feaMesh->numElement3DPID = 0;

    if (feaMesh->element1DPID != NULL) EG_free(feaMesh->element1DPID);
    if (feaMesh->element2DPID != NULL) EG_free(feaMesh->element2DPID);
    if (feaMesh->element3DPID != NULL) EG_free(feaMesh->element3DPID);

    // Can only have "free" edges in wire bodies - Don't want to count the edges of the faces
    //   as "free" edges
    if (bodySubType == WIREBODY) {

		feaMesh->numRod = numEdge;
		feaMesh->numElement1DPID = numEdge;

		if (feaMesh->crod    != NULL) EG_free(feaMesh->crod);
		if (feaMesh->crodPID != NULL) EG_free(feaMesh->crodPID);
		if (feaMesh->crodElementID != NULL) EG_free(feaMesh->crodElementID);
		if (feaMesh->crodCoordID != NULL) EG_free(feaMesh->crodCoordID);

		feaMesh->crod          = (int *) EG_alloc(2*feaMesh->numRod*sizeof(int));
		feaMesh->crodPID       = (int *) EG_alloc(  feaMesh->numRod*sizeof(int));
        feaMesh->crodElementID = (int *) EG_alloc(  feaMesh->numRod*sizeof(int));
        feaMesh->crodCoordID   = (int *) EG_alloc(  feaMesh->numRod*sizeof(int));

        feaMesh->element1DPID = (int *) EG_alloc(feaMesh->numElement1DPID*sizeof(int));

        if (feaMesh->crod   == NULL ||
        	feaMesh->crodPID== NULL ||
			feaMesh->crodElementID == NULL ||
			feaMesh->crodCoordID == NULL ||
			feaMesh->element1DPID  == NULL) {

        	status = EGADS_MALLOC;
        	goto cleanup;
        }

		for (i = 0; i < numEdge; i++) {

			numElement += 1;

			status = retrieve_CAPSGroupAttr(eedges[i], &attrName);
			if (status != CAPS_SUCCESS) {
			    printf("\tError: no capsGroup attribute found for edge - %d!!\n", i);
			    goto cleanup;
			}

			status = get_mapAttrToIndexIndex(attrMap, attrName, &attrIndex);
			if (status != CAPS_SUCCESS) {
			    printf("\tError: capsGroup name %s not found in attribute to index map\n", attrName);
			    goto cleanup;
			}

			status = EG_getTessEdge(feaMesh->bodyTessMapping.egadsTess, i, &numPoint, &xyz, &uv);
			if (status < EGADS_SUCCESS) goto cleanup;

			status = EG_localToGlobal(feaMesh->bodyTessMapping.egadsTess, -(i+1), 1, &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			feaMesh->crod[2*i + 0] = gID;

			status = EG_localToGlobal(feaMesh->bodyTessMapping.egadsTess, -(i+1), numPoint, &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			feaMesh->crod[2*i + 1] = gID;

			feaMesh->crodPID[i] = attrIndex;
			feaMesh->crodElementID[i] = numElement;

			feaMesh->element1DPID[i] = attrIndex;

			feaMesh->crodCoordID[i] = 0;

			/*
			status = retrieve_CoordSystemAttr(eedges[i], &attrName);
			if (status == CAPS_SUCCESS) {

				status = get_mapAttrToIndexIndex(coordSystemMap, attrName, &coordSystemIndex);
				if (status != CAPS_SUCCESS) {
					printf("\tError: CoordSystem name %s not found in CoordSystem attribute to index map\n", attrName);
					goto cleanup;
				}

				feaMesh->crodCoordID[i] = coordSystemIndex;
			}
			*/

			status = get_mapAttrToIndexIndex(coordSystemMap, attrName, &coordSystemIndex);
			if (status == CAPS_SUCCESS) feaMesh->crodCoordID[i] = coordSystemIndex;

		}
    } //WIREBODY IF

    feaMesh->numTria3 = 0;
    feaMesh->numQuad4 = 0;

    if (noQuadFaces == (int) false) printf("Getting quads for BEM!\n");

    // Get Tris and Quads from faces
    for (face = 0; face < numFace; face++) {

        feaMesh->numElement2DPID += 1;

        status = retrieve_CAPSGroupAttr(efaces[face], &attrName);
        if (status != CAPS_SUCCESS) {
            printf("Error: no capsGroup attribute found for face - %d!!\n", face);
            goto cleanup;
        }

        status = get_mapAttrToIndexIndex(attrMap, attrName, &attrIndex);
        if (status != CAPS_SUCCESS) {
            printf("Error: capsGroup name %s not found in attribute to index map\n", attrName);
            goto cleanup;
        }

		/*
		status = retrieve_CoordSystemAttr(efaces[face], &attrName);
		if (status == CAPS_SUCCESS) {

			status = get_mapAttrToIndexIndex(coordSystemMap, attrName, &coordSystemIndex);
			if (status != CAPS_SUCCESS) {
				printf("\tError: CoordSystem name %s not found in CoordSystem attribute to index map\n", attrName);
				goto cleanup;
			}
		} else {

			coordSystemIndex = 0;
		}
		*/

        status = get_mapAttrToIndexIndex(coordSystemMap, attrName, &coordSystemIndex);
        if (status != CAPS_SUCCESS) coordSystemIndex = 0;

        if (feaMesh->numElement2DPID == 1) {
            feaMesh->element2DPID = (int *) EG_alloc(feaMesh->numElement2DPID*sizeof(int));

        } else {
            feaMesh->element2DPID = (int *) EG_reall(feaMesh->element2DPID, feaMesh->numElement2DPID*sizeof(int));
        }

        if (feaMesh->element2DPID == NULL) {
        	status = EGADS_MALLOC;
        	goto cleanup;
        }

        feaMesh->element2DPID[feaMesh->numElement2DPID-1] = attrIndex;

        if (noQuadFaces == (int) false) {
        	status = EG_getQuads(feaMesh->bodyTessMapping.egadsTess, face+1, &numPoint, &xyz, &uv, &ptype, &pindex, &numPatch);
        	if (status < EGADS_SUCCESS) goto cleanup;

        } else numPatch = -1;

        if (numPatch > 0) {

            for (patch = 1; patch <= numPatch; patch++) {

                status = EG_getPatch(feaMesh->bodyTessMapping.egadsTess, face+1, patch, &n1, &n2, &pvindex, &pbounds);
                if (status < EGADS_SUCCESS) goto cleanup;

                for (j = 1; j < n2; j++) {
                    for (i = 1; i < n1; i++) {
                    	numElement += 1;

                    	feaMesh->numQuad4 += 1;
                        if (feaMesh->numQuad4 == 1) {

                        	if (feaMesh->cquad4    != NULL) EG_free(feaMesh->cquad4);
                        	if (feaMesh->cquad4PID != NULL) EG_free(feaMesh->cquad4PID);
                        	if (feaMesh->cquad4ElementID != NULL) EG_free(feaMesh->cquad4ElementID);
                        	if (feaMesh->cquad4CoordID != NULL) EG_free(feaMesh->cquad4CoordID);

                        	feaMesh->cquad4    = (int *) EG_alloc(4*feaMesh->numQuad4*sizeof(int));
                            feaMesh->cquad4PID = (int *) EG_alloc(  feaMesh->numQuad4*sizeof(int));
                        	feaMesh->cquad4ElementID = (int *) EG_alloc(  feaMesh->numQuad4*sizeof(int));
                        	feaMesh->cquad4CoordID   = (int *) EG_alloc(  feaMesh->numQuad4*sizeof(int));

                        } else {
                        	feaMesh->cquad4    = (int *) EG_reall(feaMesh->cquad4,    4*feaMesh->numQuad4*sizeof(int));
                        	feaMesh->cquad4PID = (int *) EG_reall(feaMesh->cquad4PID,   feaMesh->numQuad4*sizeof(int));
                        	feaMesh->cquad4ElementID = (int *) EG_reall(feaMesh->cquad4ElementID,   feaMesh->numQuad4*sizeof(int));
                        	feaMesh->cquad4CoordID = (int *) EG_reall(feaMesh->cquad4CoordID,   feaMesh->numQuad4*sizeof(int));
                        }

                        if (feaMesh->cquad4          == NULL ||
                        	feaMesh->cquad4PID 	     == NULL ||
							feaMesh->cquad4ElementID == NULL ||
							feaMesh->cquad4CoordID == NULL) {

                        	status = EGADS_MALLOC;
                        	goto cleanup;
                        }

                        status = EG_localToGlobal(feaMesh->bodyTessMapping.egadsTess, face+1, pvindex[(i-1)+n1*(j-1)], &gID);
						if (status != EGADS_SUCCESS) goto cleanup;

						feaMesh->cquad4[4*(feaMesh->numQuad4-1) + 0] = gID;

						status = EG_localToGlobal(feaMesh->bodyTessMapping.egadsTess, face+1, pvindex[(i  )+n1*(j-1)], &gID);
						if (status != EGADS_SUCCESS) goto cleanup;

						feaMesh->cquad4[4*(feaMesh->numQuad4-1) + 1] = gID;


						status = EG_localToGlobal(feaMesh->bodyTessMapping.egadsTess, face+1, pvindex[(i  )+n1*(j  )], &gID);
						if (status != EGADS_SUCCESS) goto cleanup;

						feaMesh->cquad4[4*(feaMesh->numQuad4-1) + 2] = gID;


						status = EG_localToGlobal(feaMesh->bodyTessMapping.egadsTess, face+1, pvindex[(i-1)+n1*(j  )], &gID);
						if (status != EGADS_SUCCESS) goto cleanup;

						feaMesh->cquad4[4*(feaMesh->numQuad4-1) + 3] = gID;

                        feaMesh->cquad4PID[(feaMesh->numQuad4-1)] = attrIndex;

                        feaMesh->cquad4ElementID[(feaMesh->numQuad4-1)] = numElement;

                        feaMesh->cquad4CoordID[(feaMesh->numQuad4-1)] = coordSystemIndex;

                        /*
                        feaMesh->cquad4[4*(feaMesh->numQuad4-1) + 0] = faceGID[iface][pvindex[(i1-1)+n1*(i2-1)]-1];
                        feaMesh->cquad4[4*(feaMesh->numQuad4-1) + 1] = faceGID[iface][pvindex[(i1  )+n1*(i2-1)]-1];
                        feaMesh->cquad4[4*(feaMesh->numQuad4-1) + 2] = faceGID[iface][pvindex[(i1  )+n1*(i2  )]-1];
                        feaMesh->cquad4[4*(feaMesh->numQuad4-1) + 3] = faceGID[iface][pvindex[(i1-1)+n1*(i2  )]-1];
                        */

                    }
                }

            }
        } else {
            status = EG_getTessFace(feaMesh->bodyTessMapping.egadsTess, face+1,
                                    &numPoint, &xyz, &uv, &ptype, &pindex,
                                    &numTri, &triConn, &triNeighbor);
            if (status < EGADS_SUCCESS) goto cleanup;

            for (i= 0; i < numTri; i++) {

            	numElement += 1;

                feaMesh->numTria3 += 1;
                if (feaMesh->numTria3 == 1) {

                	if (feaMesh->ctria3    != NULL) EG_free(feaMesh->ctria3);
                	if (feaMesh->ctria3PID != NULL) EG_free(feaMesh->ctria3PID);
                	if (feaMesh->ctria3ElementID != NULL) EG_free(feaMesh->ctria3ElementID);
                	if (feaMesh->ctria3CoordID != NULL) EG_free(feaMesh->ctria3CoordID);

                	feaMesh->ctria3    = (int *) EG_alloc(3*feaMesh->numTria3*sizeof(int));
                	feaMesh->ctria3PID = (int *) EG_alloc(  feaMesh->numTria3*sizeof(int));
                	feaMesh->ctria3ElementID = (int *) EG_alloc(  feaMesh->numTria3*sizeof(int));
                	feaMesh->ctria3CoordID   = (int *) EG_alloc(  feaMesh->numTria3*sizeof(int));

                } else {
                	feaMesh->ctria3    = (int *) EG_reall(feaMesh->ctria3,  4*feaMesh->numTria3*sizeof(int));
                	feaMesh->ctria3PID = (int *) EG_reall(feaMesh->ctria3PID, feaMesh->numTria3*sizeof(int));
                	feaMesh->ctria3ElementID = (int *) EG_reall(feaMesh->ctria3ElementID, feaMesh->numTria3*sizeof(int));
                	feaMesh->ctria3CoordID   = (int *) EG_reall(feaMesh->ctria3CoordID,   feaMesh->numTria3*sizeof(int));
                }

                if (feaMesh->ctria3          == NULL ||
					feaMesh->ctria3PID 	     == NULL ||
					feaMesh->ctria3ElementID == NULL ||
					feaMesh->ctria3CoordID   == NULL) {

					status = EGADS_MALLOC;
					goto cleanup;
				}

                status = EG_localToGlobal(feaMesh->bodyTessMapping.egadsTess, face+1, triConn[3*i + 0], &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				feaMesh->ctria3[3*(feaMesh->numTria3-1) + 0] = gID;

				status = EG_localToGlobal(feaMesh->bodyTessMapping.egadsTess, face+1, triConn[3*i + 1], &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				feaMesh->ctria3[3*(feaMesh->numTria3-1) + 1] = gID;

				status = EG_localToGlobal(feaMesh->bodyTessMapping.egadsTess, face+1, triConn[3*i + 2], &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				feaMesh->ctria3[3*(feaMesh->numTria3-1) + 2] = gID;

                feaMesh->ctria3PID[(feaMesh->numTria3-1)] = attrIndex;
                feaMesh->ctria3ElementID[(feaMesh->numTria3-1)] = numElement;
                feaMesh->ctria3CoordID[(feaMesh->numTria3-1)]   = coordSystemIndex;

            }
        }
    }

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:

		if (status != CAPS_SUCCESS) printf("Status in fea_bodyToBem = %d\n", status);

		if (iwest  != NULL) EG_free(iwest );
		if (inorth != NULL) EG_free(inorth);
		if (ieast  != NULL) EG_free(ieast );
		if (isouth != NULL) EG_free(isouth);
		if (rpos   != NULL) EG_free(rpos  );
		if (points  != NULL) EG_free(points );

		if (enodes != NULL) EG_free(enodes);
		if (eedges != NULL) EG_free(eedges);
		if (efaces != NULL) EG_free(efaces);

		enodes = NULL;
		eedges =NULL;
		efaces =NULL;
		//echilds = NULL;
		//EG_deleteObject(eref);
		//EG_deleteObject(eloop);

		return status;
}

// Initiate (0 out all values and NULL all pointers) of feaProperty in the feaPropertyStruct structure format
int initiate_feaPropertyStruct(feaPropertyStruct *feaProperty) {

    feaProperty->name = NULL;

    feaProperty->propertyType = UnknownProperty;

    feaProperty->propertyID = 0; // ID number of property

    feaProperty->materialID = 0; // ID number of material
    feaProperty->materialName = NULL; // Name of material associated with material ID

    // Rods
    feaProperty->crossSecArea           = 0.0; // Bar cross-sectional area
    feaProperty->torsionalConst         = 0.0; // Torsional constant
    feaProperty->torsionalStressReCoeff = 0.0; // Torsional stress recovery coefficient
    feaProperty->massPerLength          = 0.0; // Mass per unit length

    // Bar - see rod for additional variables
    feaProperty->zAxisInertia = 0.0; // Section moment of inertia about the z axis
    feaProperty->yAxisInertia = 0.0; // Section moment of inertia about the y axis

    feaProperty->yCoords[0] = 0.0; // Element y, z coordinates, in the bar cross-section, of
    feaProperty->yCoords[1] = 0.0; //    of four points at which to recover stresses
    feaProperty->yCoords[2] = 0.0;
    feaProperty->yCoords[3] = 0.0;

    feaProperty->zCoords[0] = 0.0;
    feaProperty->zCoords[1] = 0.0;
    feaProperty->zCoords[2] = 0.0;
    feaProperty->zCoords[3] = 0.0;

    feaProperty->areaShearFactors[0] = 0.0; // Area factors for shear
    feaProperty->areaShearFactors[1] = 0.0;

    feaProperty->crossProductInertia = 0.0; // Section cross-product of inertia

    // Shear
    feaProperty->shearPanelThickness  = 0.0; // Shear panel thickness
    feaProperty->nonStructMassPerArea = 0.0; // Nonstructural mass per unit area

    // Shell
    feaProperty->membraneThickness = 0.0;     // Membrane thickness
    feaProperty->materialBendingID = 0;     /// ID number of material for bending - if not specified and bendingInertiaRatio > 0 this value defaults to materialID
    feaProperty->bendingInertiaRatio = 1.0; // Ratio of actual bending moment inertia (I) to bending inertia of a solid
                                            //   plate of thickness TM Real  - default 1.0
    feaProperty->materialShearID = 0;       // ID number of material for shear - if not specified and shearMembraneRatio > 0 this value defaults to materialID
    feaProperty->shearMembraneRatio  = 5.0/6.0;     // Ratio of shear to membrane thickness  - default 5/6
    feaProperty->massPerArea = 0.0;           // Mass per unit area
    //feaProperty->neutralPlaneDist[0] = 0;   // Distances from the neutral plane of the plate to locations where
    //feaProperty->neutralPlaneDist[1] = 0;   //   stress is calculate

    feaProperty->compositeShearBondAllowable = 0.0;  // Shear stress limit for bonding between plies
    feaProperty->compositeFailureTheory = NULL;          // HILL, HOFF, TSAI, STRN
    feaProperty->compositeSymmetricLaminate = 0;      // 1- SYM only half the plies are specified, for odd number plies 1/2 thickness of center ply is specified
								// the first ply is the bottom ply in the stack, default (0) all plies specified
    feaProperty->numPly = 0;
    feaProperty->compositeMaterialID = NULL;      // Vector of material ID's from bottom to top for all plies
    feaProperty->compositeThickness = NULL;        // Vector of thicknesses from bottom to top for all plies
    feaProperty->compositeOrientation = NULL;      // Vector of orientations from bottom to top for all plies

    // Solid

    return 0;
}

// Destroy (0 out all values and NULL all pointers) of feaProperty in the feaPropertyStruct structure format
int destroy_feaPropertyStruct(feaPropertyStruct *feaProperty) {

    if (feaProperty->name != NULL) EG_free(feaProperty->name);

    feaProperty->propertyType = UnknownProperty;

    feaProperty->propertyID = 0; // ID number of property

    feaProperty->materialID = 0; // ID number of material

    if (feaProperty->materialName != NULL) EG_free(feaProperty->materialName); // Name of material associated with material ID
    feaProperty->materialName = NULL;

    // Rods
    feaProperty->crossSecArea           = 0; // Bar cross-sectional area
    feaProperty->torsionalConst         = 0; // Torsional constant
    feaProperty->torsionalStressReCoeff = 0; // Torsional stress recovery coefficient
    feaProperty->massPerLength          = 0; // Mass per unit length

    // Bar - see rod for additional variables
    feaProperty->zAxisInertia = 0; // Section moment of inertia about the z axis
    feaProperty->yAxisInertia = 0; // Section moment of inertia about the y axis

    feaProperty->yCoords[0] = 0; // Element y, z coordinates, in the bar cross-section, of
    feaProperty->yCoords[1] = 0; //    of four points at which to recover stresses
    feaProperty->yCoords[2] = 0;
    feaProperty->yCoords[3] = 0;

    feaProperty->zCoords[0] = 0;
    feaProperty->zCoords[1] = 0;
    feaProperty->zCoords[2] = 0;
    feaProperty->zCoords[3] = 0;

    feaProperty->areaShearFactors[0] = 0; // Area factors for shear
    feaProperty->areaShearFactors[1] = 0;

    feaProperty->crossProductInertia = 0; // Section cross-product of inertia

    // Shear
    feaProperty->shearPanelThickness  = 0; // Shear panel thickness
    feaProperty->nonStructMassPerArea = 0; // Nonstructural mass per unit area

    // Shell
    feaProperty->membraneThickness = 0; // Membrane thickness
    feaProperty->materialBendingID = 0;     // ID number of material for bending
    feaProperty->bendingInertiaRatio = 1.0; // Ratio of actual bending moment inertia (I) to bending inertia of a solid
                                            //   plate of thickness TM Real  - default 1.0
    feaProperty->materialShearID = 0;       // ID number of material for shear
    feaProperty->shearMembraneRatio  = 5.0/6.0;     // Ratio of shear to membrane thickness  - default 5/6
    feaProperty->massPerArea = 0;           // Mass per unit area
    //feaProperty->neutralPlaneDist[0] = 0;   // Distances from the neutral plane of the plate to locations where
    //feaProperty->neutralPlaneDist[1] = 0;   //   stress is calculate

    feaProperty->numPly = 0;
    feaProperty->compositeShearBondAllowable = 0.0;  // Shear stress limit for bonding between plies
    if (feaProperty->compositeFailureTheory != NULL) EG_free(feaProperty->compositeFailureTheory); // HILL, HOFF, TSAI, STRN
    feaProperty->compositeFailureTheory = NULL;
    feaProperty->compositeSymmetricLaminate = 0;      // 1- SYM only half the plies are specified, for odd number plies 1/2 thickness of center ply is specified
								// the first ply is the bottom ply in the stack, default (0) all plies specified
	if (feaProperty->compositeMaterialID != NULL) EG_free(feaProperty->compositeMaterialID);
	if (feaProperty->compositeThickness != NULL) EG_free(feaProperty->compositeThickness);
	if (feaProperty->compositeOrientation != NULL) EG_free(feaProperty->compositeOrientation);

	feaProperty->compositeMaterialID = NULL;
	feaProperty->compositeThickness = NULL;
	feaProperty->compositeOrientation = NULL;

    // Solid

    return 0;
}

// Initiate (0 out all values and NULL all pointers) of feaMaterial in the feaMaterialStruct structure format
int initiate_feaMaterialStruct(feaMaterialStruct *feaMaterial) {

	feaMaterial->name = NULL;  // Material name

	feaMaterial->materialType = UnknownMaterial; // Set

	feaMaterial->materialID = 0; // ID number of material

	feaMaterial->youngModulus = 0.0; // E - Young's Modulus
	feaMaterial->shearModulus = 0.0; // G - Shear Modulus
	feaMaterial->poissonRatio = 0.0; // Poisson's Ratio
	feaMaterial->density      = 0.0;    // Rho - material mass density
	feaMaterial->thermalExpCoeff = 0.0; //Coefficient of thermal expansion
	feaMaterial->temperatureRef  = 0.0; // Reference temperature
	feaMaterial->dampingCoeff    = 0.0; // Damping coefficient
	feaMaterial->tensionAllow    = 0.0; // Tension allowable for the material
	feaMaterial->compressAllow   = 0.0; // Compression allowable for the material
	feaMaterial->shearAllow      = 0.0; // Shear allowable for the material

	feaMaterial->youngModulusLateral = 0.0; // Young's Modulus in the lateral direction
	feaMaterial->shearModulusTrans1Z = 0.0; // Transverse shear modulus in the 1-Z plane
	feaMaterial->shearModulusTrans2Z = 0.0; // Transverse shear modulus in the 2-Z plane


	return 0;
}

// Destroy (0 out all values and NULL all pointers) of feaMaterial in the feaMaterialStruct structure format
int destroy_feaMaterialStruct(feaMaterialStruct *feaMaterial) {

	if (feaMaterial->name != NULL) EG_free(feaMaterial->name);
	feaMaterial->name = NULL;  // Material name

	feaMaterial->materialType = UnknownMaterial; // Material type

	feaMaterial->materialID = 0; // ID number of material

	feaMaterial->youngModulus = 0.0; // E - Young's Modulus
	feaMaterial->shearModulus = 0.0; // G - Shear Modulus
	feaMaterial->poissonRatio = 0.0; // Poisson's Ratio
	feaMaterial->density      = 0.0;    // Rho - material mass density
	feaMaterial->thermalExpCoeff = 0.0; //Coefficient of thermal expansion
	feaMaterial->temperatureRef  = 0.0; // Reference temperature
	feaMaterial->dampingCoeff    = 0.0; // Damping coefficient
	feaMaterial->tensionAllow    = 0.0; // Tension allowable for the material
	feaMaterial->compressAllow   = 0.0; // Compression allowable for the material
	feaMaterial->shearAllow      = 0.0; // Shear allowable for the material

	feaMaterial->youngModulusLateral = 0.0; // Young's Modulus in the lateral direction
	feaMaterial->shearModulusTrans1Z = 0.0; // Transverse shear modulus in the 1-Z plane
	feaMaterial->shearModulusTrans2Z = 0.0; // Transverse shear modulus in the 2-Z plane

	return 0;
}

// Initiate (0 out all values and NULL all pointers) of feaConstraint in the feaConstraintStruct structure format
int initiate_feaConstraintStruct(feaConstraintStruct *feaConstraint) {

	feaConstraint->name = NULL;  // Constraint name

	feaConstraint->constraintType = UnknownConstraint; // Constraint type

	feaConstraint->constraintID = 0; // ID number of constraint

	feaConstraint->numGridID = 0; // Component number of grid
	feaConstraint->gridIDSet = NULL; // List of component number of grids to apply constraint to

	feaConstraint->dofConstraint = 0; // Number to indicate DOF constraints
	feaConstraint->gridDisplacement = 0;

	return 0;
}

// Destroy (0 out all values and NULL all pointers) of feaConstraint in the feaConstraintStruct structure format
int destroy_feaConstraintStruct(feaConstraintStruct *feaConstraint) {

	if (feaConstraint->name != NULL) EG_free(feaConstraint->name);
	feaConstraint->name = NULL;  // Constraint name

	feaConstraint->constraintType = UnknownConstraint; // Constraint type

	feaConstraint->constraintID = 0; // ID number of constraint

	feaConstraint->numGridID = 0; // Number of grid IDs in grid ID set

	if (feaConstraint->gridIDSet != NULL) EG_free(feaConstraint->gridIDSet); // List of component number of grids to apply constraint to

	feaConstraint->dofConstraint = 0; // Number to indicate DOF constraints
	feaConstraint->gridDisplacement = 0; // The value for the displacement

	return 0;
}

// Initiate (0 out all values and NULL all pointers) of feaAnalysis in the feaAnalysisStruct structure format
int initiate_feaAnalysisStruct(feaAnalysisStruct *feaAnalysis) {

    feaAnalysis->name = NULL; // Analysis name

    feaAnalysis->analysisType = UnknownAnalysis; // Type of analysis

    feaAnalysis->analysisID = 0; // ID number of analysis

    // Loads for the analysis
    feaAnalysis->numLoad = 0; 	// Number of loads in the analysis
    feaAnalysis->loadSetID = NULL; // List of the load IDSs

	// Constraints for the analysis
    feaAnalysis->numConstraint = 0;   // Number of constraints in the analysis
    feaAnalysis->constraintSetID = NULL; // List of constraint IDs

    // Optimization constraints
    feaAnalysis->numDesignConstraint = 0; // Number of design constraints
    feaAnalysis->designConstraintSetID = NULL; // List of design constraint IDs

    // Optimization - objective
    //feaAnalysis->objectiveMinMax = NULL; // Minimize or maximize the objective
    //feaAnalysis->objectiveResponseType = NULL;  // Response type options for DRESP1 Entry

    // Eigenvalue
    feaAnalysis->extractionMethod = NULL;

    feaAnalysis->frequencyRange[0] = 0;
    feaAnalysis->frequencyRange[1] = 0;

    feaAnalysis->numEstEigenvalue = 0;
    feaAnalysis->numDesiredEigenvalue = 0;
    feaAnalysis->eigenNormaliztion = NULL;

    feaAnalysis->gridNormaliztion = 0;
    feaAnalysis->componentNormaliztion = 0;

    feaAnalysis->lanczosMode = 2; //Lanczos mode for calculating eigenvalues
    feaAnalysis->lanczosType = NULL; //Lanczos matrix type (DPB, DGB)
    return 0;
}

// Destroy (0 out all values and NULL all pointers) of feaAnalysis in the feaAnalysisStruct structure format
int destroy_feaAnalysisStruct(feaAnalysisStruct *feaAnalysis) {

    if (feaAnalysis->name != NULL) EG_free(feaAnalysis->name);
    feaAnalysis->name = NULL; // Analysis name

    feaAnalysis->analysisType = UnknownAnalysis; // Type of analysis

    feaAnalysis->analysisID = 0; // ID number of analysis

    // Loads for the analysis
    feaAnalysis->numLoad = 0; 	// Number of loads in the analysis
    if(feaAnalysis->loadSetID != NULL) EG_free(feaAnalysis->loadSetID);
    feaAnalysis->loadSetID = NULL; // List of the load IDSs

	// Constraints for the analysis
    feaAnalysis->numConstraint = 0;   // Number of constraints in the analysis
    if (feaAnalysis->constraintSetID != NULL) EG_free(feaAnalysis->constraintSetID);
    feaAnalysis->constraintSetID = NULL; // List of constraint IDs

    // Optimization constraints
    feaAnalysis->numDesignConstraint = 0; // Number of design constraints
    if (feaAnalysis->designConstraintSetID != NULL) EG_free(feaAnalysis->designConstraintSetID);
    feaAnalysis->designConstraintSetID = NULL; // List of design constraint IDs

    // Optimization - objective
    //if (feaAnalysis->objectiveMinMax != NULL) EG_free(feaAnalysis->objectiveMinMax);
    //feaAnalysis->objectiveMinMax = NULL; // Minimize or maximize the objective

    //if (feaAnalysis->objectiveResponseType != NULL) EG_free(feaAnalysis->objectiveResponseType);
    //feaAnalysis->objectiveResponseType = NULL;  // Response type options for DRESP1 Entry


    // Eigenvalue
    if (feaAnalysis->extractionMethod != NULL) EG_free(feaAnalysis->extractionMethod);
    feaAnalysis->extractionMethod = NULL;

    feaAnalysis->frequencyRange[0] = 0;
    feaAnalysis->frequencyRange[1] = 0;

    feaAnalysis->numEstEigenvalue = 0;
    feaAnalysis->numDesiredEigenvalue = 0;

    if (feaAnalysis->eigenNormaliztion != NULL) EG_free(feaAnalysis->eigenNormaliztion);
    feaAnalysis->eigenNormaliztion = NULL;

    feaAnalysis->gridNormaliztion = 0;
    feaAnalysis->componentNormaliztion = 0;

    feaAnalysis->lanczosMode = 2; //Lanczos mode for calculating eigenvalues

    if (feaAnalysis->lanczosType != NULL) EG_free(feaAnalysis->lanczosType);
    feaAnalysis->lanczosType = NULL; //Lanczos matrix type (DPB, DGB)

    return 0;
}

// Initiate (0 out all values and NULL all pointers) of feaLoad in the feaLoadStruct structure format
int initiate_feaLoadStruct(feaLoadStruct *feaLoad) {

	feaLoad->name = NULL;  // Load name

	feaLoad->loadType = UnknownLoad; // Load type

	feaLoad->loadID = 0; // ID number of load

	feaLoad->loadScaleFactor = 1; // Scale factor for when combining loads

	// Concentrated force at a grid point
	feaLoad->numGridID = 0; // Number of grid IDs in grid ID set
	feaLoad->gridIDSet = NULL; // List of grid IDs to apply the constraint to. size = [numGridID]
	feaLoad->coordSystemID= 0; // Component number of coordinate system in which force vector is specified

	feaLoad->forceScaleFactor= 0.0; // Overall scale factor for the force
	feaLoad->directionVector[0] = 0.0;   // [0]-x, [1]-y, [2]-z components of the force vector
	feaLoad->directionVector[1] = 0.0;
	feaLoad->directionVector[2] = 0.0;

	// Concentrated moment at a grid point (also uses coordSystemID and directionVector)
	feaLoad->momentScaleFactor= 0.0; // Overall scale factor for the moment

	// Gravitational load (also uses coordSystemID and directionVector)
	feaLoad->gravityAcceleration= 0.0; // Gravitational acceleration

	// Pressure load
	feaLoad->pressureForce= 0.0; // Pressure value
	feaLoad->pressureDistributeForce[0] = 0.0; // Pressure load at a specified grid location in the element
	feaLoad->pressureDistributeForce[1] = 0.0;
	feaLoad->pressureDistributeForce[2] = 0.0;
	feaLoad->pressureDistributeForce[3] = 0.0;

	feaLoad->pressureMultiDistributeForce = NULL; // Unique pressure load at a specified grid location for
											  	  // each element in elementIDSet size = [numElementID][4] - used in type PressureExternal
											  	  // where the pressure force is being provided by an external source (i.e. data transfer)

	feaLoad->numElementID = 0; // Number of elements IDs in element ID set
	feaLoad->elementIDSet = NULL; // List element IDs in which to apply the load. size = [numElementID]

	// Rotational velocity (also uses coordSystemID and directionVector)
	feaLoad->angularVelScaleFactor= 0; // Overall scale factor for the angular velocity
	feaLoad->angularAccScaleFactor= 0; // Overall scale factor for the angular acceleration

	// Thermal load - to be added


	return 0;
}

// Destroy (0 out all values and NULL all pointers) of feaLoad in the feaLoadStruct structure format
int destroy_feaLoadStruct(feaLoadStruct *feaLoad) {

	if (feaLoad->name != NULL) EG_free(feaLoad->name);
	feaLoad->name = NULL;  // Load name

	feaLoad->loadType = UnknownLoad; // Load type

	feaLoad->loadID = 0; // ID number of load

	feaLoad->loadScaleFactor = 1; // Scale factor for when combining loads

	// Concentrated force at a grid point
	feaLoad->numGridID = 0; // Number of grid IDs in grid ID set
	if (feaLoad->gridIDSet != NULL) EG_free(feaLoad->gridIDSet); // List of grid IDs to apply the constraint to
	feaLoad->gridIDSet = NULL;

	feaLoad->coordSystemID= 0; // Component number of coordinate system in which force vector is specified
	feaLoad->forceScaleFactor= 0; // Overall scale factor for the force
	feaLoad->directionVector[0] = 0;   // [0]-x, [1]-y, [2]-z components of the force vector
	feaLoad->directionVector[1] = 0;
	feaLoad->directionVector[2] = 0;

	// Concentrated moment at a grid pofeaLoad->(also uses coordSystemID and directionVector)
	feaLoad->momentScaleFactor= 0; // Overall scale factor for the moment

	// Gravitational load (also uses coordSystemID and directionVector)
	feaLoad->gravityAcceleration= 0; // Gravitational acceleration

	// Pressure load
	feaLoad->pressureForce= 0; // Pressure value

	feaLoad->pressureDistributeForce[0] = 0; // Pressure load at a specified grid location in the element
	feaLoad->pressureDistributeForce[1] = 0;
	feaLoad->pressureDistributeForce[2] = 0;
	feaLoad->pressureDistributeForce[3] = 0;

	if (feaLoad->pressureMultiDistributeForce != NULL) 	EG_free(feaLoad->pressureMultiDistributeForce); // Unique pressure load at a specified grid location for
	feaLoad->pressureMultiDistributeForce = NULL;						//each element in elementIDSet size = [4*numElementID]- used in type PressureExternal


	feaLoad->numElementID = 0; // Number of elements IDs in element ID set
	if (feaLoad->elementIDSet != NULL) EG_free(feaLoad->elementIDSet); // List element IDs in which to apply the load
	feaLoad->elementIDSet = NULL;

	// Rotational velocity (also uses coordSystemID and directionVector)
	feaLoad->angularVelScaleFactor= 0; // Overall scale factor for the angular velocity
	feaLoad->angularAccScaleFactor= 0; // Overall scale factor for the angular acceleration

	// Thermal load - to be added

	return 0;
}

// Initiate (0 out all values and NULL all pointers) of feaDesignVariable in the feaDesignVariableStruct structure format
int initiate_feaDesignVariableStruct(feaDesignVariableStruct *feaDesignVariable) {

	feaDesignVariable->name = NULL;

	feaDesignVariable->designVariableType = UnknownDesignVar;

	feaDesignVariable->designVariableID = 0; //  ID number of design variable


	feaDesignVariable->initialValue = 0.0; // Initial value of design variable
	feaDesignVariable->lowerBound = 0.0;   // Lower bounds of variable
	feaDesignVariable->upperBound = 0.0;   // Upper bounds of variable
	feaDesignVariable->maxDelta= 0.0;     // Move limit for design variable

    feaDesignVariable->numDiscreteValue = 0; // Number of discrete values that a design variable can assume

	feaDesignVariable->discreteValue = NULL; // List of discrete values that a design variable can assume;

	feaDesignVariable->numMaterialID = 0; // Number of materials to apply the design variable to
	feaDesignVariable->materialSetID = NULL; // List of materials IDs
	feaDesignVariable->materialSetType = NULL; // List of materials types corresponding to the materialSetID

	feaDesignVariable->numPropertyID = 0;   // Number of property ID to apply the design variable to
	feaDesignVariable->propertySetID = NULL; // List of property IDs
	feaDesignVariable->propertySetType = NULL; // List of property types corresponding to the propertySetID

	feaDesignVariable->fieldPosition = 0; //  Position in card to apply design variable to
	feaDesignVariable->fieldName = NULL; // Name of property/material to apply design variable to

	return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) of feaDesignVariable in the feaDesignVariableStruct structure format
int destroy_feaDesignVariableStruct(feaDesignVariableStruct *feaDesignVariable) {

	if (feaDesignVariable->name != NULL) EG_free(feaDesignVariable->name);
	feaDesignVariable->name = NULL;


	feaDesignVariable->designVariableType = UnknownDesignVar;

    feaDesignVariable->designVariableID = 0; //  ID number of design variable

	feaDesignVariable->initialValue = 0.0; // Initial value of design variable
	feaDesignVariable->lowerBound = 0.0;   // Lower bounds of variable
	feaDesignVariable->upperBound = 0.0;   // Upper bounds of variable
	feaDesignVariable->maxDelta= 0.0;     // Move limit for design variable

	feaDesignVariable->numDiscreteValue = 0; // Number of discrete values that a design variable can assume;
	if (feaDesignVariable->discreteValue != NULL) EG_free(feaDesignVariable->discreteValue);
	feaDesignVariable->discreteValue = NULL; // List of discrete values that a design variable can assume;

	feaDesignVariable->numMaterialID = 0; // Number of materials to apply the design variable to
	if (feaDesignVariable->materialSetID != NULL) EG_free(feaDesignVariable->materialSetID);
	feaDesignVariable->materialSetID = NULL; // List of materials IDs

	if (feaDesignVariable->materialSetType != NULL) EG_free(feaDesignVariable->materialSetType);
	feaDesignVariable->materialSetType = NULL; // List of materials types corresponding to the materialSetID

	feaDesignVariable->numPropertyID = 0;   // Number of property ID to apply the design variable to
	if (feaDesignVariable->propertySetID != NULL) EG_free(feaDesignVariable->propertySetID);
	feaDesignVariable->propertySetID = NULL; // List of property IDs

	if (feaDesignVariable->propertySetType != NULL) EG_free(feaDesignVariable->propertySetType);
	feaDesignVariable->propertySetType = NULL; // List of property types corresponding to the propertySetID

	feaDesignVariable->fieldPosition = 0; //  Position in card to apply design variable to
	if (feaDesignVariable->fieldName != NULL) EG_free(feaDesignVariable->fieldName);
	feaDesignVariable->fieldName = NULL; // Name of property/material to apply design variable to

	return CAPS_SUCCESS;

}

// Initiate (0 out all values and NULL all pointers) of feaDesignConstraint in the feaDesignConstraintStruct structure format
int initiate_feaDesignConstraintStruct(feaDesignConstraintStruct *feaDesignConstraint) {

	feaDesignConstraint->name = NULL;

	feaDesignConstraint->designConstraintID = 0; //  ID number of design constraint

	feaDesignConstraint->responseType = NULL;  // Response type options for DRESP1 Entry

	feaDesignConstraint->lowerBound = 0.0;   // Lower bounds of design response
	feaDesignConstraint->upperBound = 0.0;   // Upper bounds of design response

	feaDesignConstraint->numPropertyID = 0;   // Number of property ID to apply the design variable to
	feaDesignConstraint->propertySetID = NULL; // List of property IDs
    feaDesignConstraint->propertySetType = NULL; // List of property types corresponding to the propertySetID

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) of feaDesignConstraint in the feaDesignConstraintStruct structure format
int destroy_feaDesignConstraintStruct(feaDesignConstraintStruct *feaDesignConstraint) {

	if (feaDesignConstraint->name != NULL) EG_free(feaDesignConstraint->name);
	feaDesignConstraint->name = NULL;

    feaDesignConstraint->designConstraintID = 0; //  ID number of design constraint

	if (feaDesignConstraint->responseType != NULL) EG_free(feaDesignConstraint->responseType);
	feaDesignConstraint->responseType = NULL;  // Response type options for DRESP1 Entry

	feaDesignConstraint->lowerBound = 0.0;   // Lower bounds of design response
	feaDesignConstraint->upperBound = 0.0;   // Upper bounds of design response

	feaDesignConstraint->numPropertyID = 0;   // Number of property ID to apply the design variable to

	if (feaDesignConstraint->propertySetID != NULL) EG_free(feaDesignConstraint->propertySetID);
	feaDesignConstraint->propertySetID = NULL; // List of property IDs

    if (feaDesignConstraint->propertySetType != NULL) EG_free(feaDesignConstraint->propertySetType);
    feaDesignConstraint->propertySetType = NULL; // List of property types corresponding to the propertySetID

	return CAPS_SUCCESS;

}


// Initiate (0 out all values and NULL all pointers) of feaCoordSystem in the feaCoordSystemStruct structure format
int initiate_feaCoordSystemStruct(feaCoordSystemStruct *feaCoordSystem) {

	int i; // Indexing

	feaCoordSystem->name = NULL; // Coordinate system name

	feaCoordSystem->coordSystemType = UnknownCoordSystem;  // Coordinate system type

	feaCoordSystem->coordSystemID = 0; // ID number of coordinate system
    feaCoordSystem->refCoordSystemID = 0; // ID of reference coordinate system

    for (i = 0; i < 3; i++) {
		feaCoordSystem->origin[i] = 0; // x, y, and z coordinates for the origin

		feaCoordSystem->normal1[i] = 0; // First normal direction
		feaCoordSystem->normal2[i] = 0; // Second normal direction
		feaCoordSystem->normal3[i] = 0; // Third normal direction - found from normal1 x normal2
    }

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) of feaCoordSystem in the feaCoordSystemStruct structure format
int destroy_feaCoordSystemStruct(feaCoordSystemStruct *feaCoordSystem) {

	int i; // Indexing

	if (feaCoordSystem->name != NULL) EG_free(feaCoordSystem->name);
	feaCoordSystem->name = NULL; // Coordinate system name

	feaCoordSystem->coordSystemType = UnknownCoordSystem;  // Coordinate system type

	feaCoordSystem->coordSystemID = 0; // ID number of coordinate system
	feaCoordSystem->refCoordSystemID = 0; // ID of reference coordinate system

    for (i = 0; i < 3; i++) {
		feaCoordSystem->origin[i] = 0; // x, y, and z coordinates for the origin

		feaCoordSystem->normal1[i] = 0; // First normal direction
		feaCoordSystem->normal2[i] = 0; // Second normal direction
		feaCoordSystem->normal3[i] = 0; // Third normal direction - found from normal1 x normal2
    }

	return CAPS_SUCCESS;
}

// Get the material properties from a capsTuple
int fea_getMaterial(int numMaterialTuple,
                    capsTuple materialTuple[],
                    int *numMaterial,
                    feaMaterialStruct *feaMaterial[]) {

	/*! \page feaMaterial FEA Material
	 * Structure for the material tuple  = ("Material Name", "Value").
	 * "Material Name" defines the reference name for the material being specified.
	 *	The "Value" can either be a JSON String dictionary (see Section \ref jsonStringMaterial) or a single string keyword
     *	(see Section \ref keyStringMaterial).
     */

    int status; //Function return

    int i; // Indexing

    char *keyValue = NULL;
    char *keyWord = NULL;

    // Destroy our material structures coming in if aren't 0 and NULL already
    for (i = 0; i < *numMaterial; i++) {
    	status = destroy_feaMaterialStruct(&(*feaMaterial)[i]);
    	if (status != CAPS_SUCCESS) return status;
    }
    if (*feaMaterial != NULL) EG_free(*feaMaterial);
    *feaMaterial = NULL;
    *numMaterial = 0;

    printf("\nGetting FEA materials.......\n");

    *numMaterial = numMaterialTuple;
    printf("\tNumber of materials - %d\n", *numMaterial);

    if (*numMaterial > 0) {
        *feaMaterial = (feaMaterialStruct *) EG_alloc(*numMaterial * sizeof(feaMaterialStruct));
        if (*feaMaterial == NULL) return EGADS_MALLOC;

    } else {
        printf("\tNumber of material values in input tuple is 0\n");
        return CAPS_NOTFOUND;
    }

    for (i = 0; i < *numMaterial; i++) {
        status = initiate_feaMaterialStruct(&(*feaMaterial)[i]);
        if (status != CAPS_SUCCESS) return status;
    }

    for (i = 0; i < *numMaterial; i++) {

        printf("\tMaterial name - %s\n", materialTuple[i].name);

        (*feaMaterial)[i].name = (char *) EG_alloc(((strlen(materialTuple[i].name)) + 1)*sizeof(char));
        if ((*feaMaterial)[i].name == NULL) return EGADS_MALLOC;

        memcpy((*feaMaterial)[i].name, materialTuple[i].name, strlen(materialTuple[i].name)*sizeof(char));
        (*feaMaterial)[i].name[strlen(materialTuple[i].name)] = '\0';

        (*feaMaterial)[i].materialID = i + 1;

        // Do we have a json string?
        if (strncmp(materialTuple[i].value, "{", 1) == 0) {
            //printf("JSON String - %s\n", materialTuple[i].value);


        	/*! \page feaMaterial
        	 * \section jsonStringMaterial JSON String Dictionary
             *
             * If "Value" is JSON string dictionary
             * \if (MYSTRAN || NASTRAN)
			 *  (eg. "Value" = {"density": 7850, "youngModulus": 120000.0, "poissonRatio": 0.5, "materialType": "isotropic"})
			 * \endif
			 *  the following keywords ( = default values) may be used:
			 *
        	 * <ul>
        	 * <li> <B>materialType = "Isotropic"</B> </li> <br>
        	 *      Material property type. Options: Isotropic, Anisothotropic, Orthotropic, or Anisotropic.
        	 * </ul>
        	 */

            // Get material Type
            keyWord = "materialType";
            status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                //{UnknownMaterial, Isotropic, Anisothotropic, Orthotropic, Anisotropic}
                if      (strcasecmp(keyValue, "\"Isotropic\"")      == 0) (*feaMaterial)[i].materialType = Isotropic;
                else if (strcasecmp(keyValue, "\"Anisothotropic\"") == 0) (*feaMaterial)[i].materialType = Anisothotropic;
                else if (strcasecmp(keyValue, "\"Orthotropic\"")    == 0) (*feaMaterial)[i].materialType = Orthotropic;
                else if (strcasecmp(keyValue, "\"Anisotropic\"")    == 0) (*feaMaterial)[i].materialType = Anisotropic;
                else {

                    printf("\tUnrecognized \"%s\" specified (%s) for Material tuple %s, defaulting to \"Isotropic\"\n", keyWord,
                                                                                                                      keyValue,
                                                                                                                      materialTuple[i].name);
                    (*feaMaterial)[i].materialType = Isotropic;
                }

            } else {

                    printf("\tNo \"%s\" specified for Material tuple %s, defaulting to \"Isotropic\"\n", keyWord,
                                                                                                       materialTuple[i].name);
                    (*feaMaterial)[i].materialType = Isotropic;
            }

            if (keyValue != NULL) {
                EG_free(keyValue);
                keyValue = NULL;
            }

            //Fill up material properties

            /*! \page feaMaterial
             *
             * \if (MYSTRAN || NASTRAN)
             *  <ul>
             *	<li> <B>youngModulus = 0.0</B> </li> <br>
             *  Also know as the elastic modulus, defines the relationship between stress and strain.
             *  Default if `shearModulus' and `poissonRatio' != 0, youngModulus = 2*(1+poissonRatio)*shearModulus
             *  </ul>
             * \elseif ABAQUS
             *
             * Something else ....
             *
             * \else
             *  <ul>
             *  <li> <B>youngModulus = 0.0</B> </li> <br>
             *  Also know as the elastic modulus, defines the relationship between stress and strain.
             *  Default if `shearModulus' and `poissonRatio' != 0, youngModulus = 2*(1+poissonRatio)*shearModulus
             *  </ul>
             * \endif
             *
             */

            keyWord = "youngModulus";
            status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*feaMaterial)[i].youngModulus);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaMaterial
             *
             * \if (MYSTRAN || NASTRAN)
             *  <ul>
             *  <li> <B>shearModulus = 0.0</B> </li> <br>
             *  Also know as the modulus of rigidity is defined as the ratio of shear stress to the shear strain.
             *  Default if `youngModulus' and `poissonRatio' != 0, shearModulus = youngModulus/(2*(1+poissonRatio))
             *  </ul>
             * \elseif ABAQUS
             *
             * Something else ....
             *
             * \else
             *  <ul>
             *  <li> <B>shearModulus = 0.0</B> </li> <br>
             *  Also know as the modulus of rigidity is defined as the ratio of shear stress to the shear strain.
             *  Default if `youngModulus' and `poissonRatio' != 0, shearModulus = youngModulus/(2*(1+poissonRatio))
             *  </ul>
             * \endif
             */

            keyWord = "shearModulus";
            status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*feaMaterial)[i].shearModulus);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaMaterial
             *
             *
             * \if (MYSTRAN || NASTRAN)
             *  <ul>
             *  <li> <B>poissonRatio = 0.0</B> </li> <br>
             *  Is the fraction of expansion divided by the fraction of compression.
             *  Default if `youngModulus' and `shearModulus' != 0, poissonRatio = (2*youngModulus/shearModulus) - 1
             *  </ul>
             * \elseif ABAQUS
             *
             * Something else ....
             *
             * \else
             *  <ul>
             *  <li> <B>poissonRatio = 0.0</B> </li> <br>
             *  Is the fraction of expansion divided by the fraction of compression.
             *  Default if `youngModulus' and `shearModulus' != 0, poissonRatio = (2*youngModulus/shearModulus) - 1
             *  </ul>
             * \endif
             */

            keyWord = "poissonRatio";
            status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*feaMaterial)[i].poissonRatio);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            // Check Young's modulus, shear modulus, and Poisson's ratio
            if ( ((*feaMaterial)[i].youngModulus == 0 && (*feaMaterial)[i].poissonRatio == 0) ||
                 ((*feaMaterial)[i].shearModulus == 0 && (*feaMaterial)[i].poissonRatio == 0)) {
                // Do nothing

            } else if ((*feaMaterial)[i].youngModulus == 0) {
                (*feaMaterial)[i].youngModulus = 2*(1+(*feaMaterial)[i].poissonRatio)*(*feaMaterial)[i].shearModulus;
            } else if ((*feaMaterial)[i].shearModulus == 0) {
                (*feaMaterial)[i].shearModulus = (*feaMaterial)[i].youngModulus/(2*(1+(*feaMaterial)[i].poissonRatio));
            } else if ((*feaMaterial)[i].poissonRatio == 0) {
                (*feaMaterial)[i].poissonRatio = (*feaMaterial)[i].youngModulus/(2*(*feaMaterial)[i].shearModulus) -1;
            }

            /*! \page feaMaterial
             *
             *
             * \if (MYSTRAN || NASTRAN)
             *  <ul>
             *  <li> <B>density = 0.0</B> </li> <br>
             *  Density of the material.
             *  </ul>
             * \elseif ABAQUS
             *
             * Something else ....
             *
             * \else
             *  <ul>
             *  <li> <B>density = 0.0</B> </li> <br>
             *  Density of the material.
             *  </ul>
             * \endif
             */
            keyWord = "density";
            status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*feaMaterial)[i].density);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaMaterial
             *
             * \if (MYSTRAN || NASTRAN)
             *  <ul>
             *  <li> <B>thermalExpCoeff = 0.0</B> </li> <br>
             *  Thermal expansion coefficient of the material.
             *  </ul>
             * \elseif ABAQUS
             *
             * Something else ....
             *
             * \else
             *  <ul>
             *  <li> <B>thermalExpCoeff = 0.0</B> </li> <br>
             *  Thermal expansion coefficient of the material.
             *  </ul>
             * \endif
             */
            keyWord = "thermalExpCoeff";
            status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*feaMaterial)[i].thermalExpCoeff);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaMaterial
             *
             *
             * \if (MYSTRAN || NASTRAN)
             *  <ul>
             *  <li> <B>temperatureRef = 0.0</B> </li> <br>
             *  Reference temperature for material properties.
             *  </ul>
             * \elseif ABAQUS
             *
             * Something else ....
             *
             * \else
             *  <ul>
             *  <li> <B>temperatureRef = 0.0</B> </li> <br>
             *  Reference temperature for material properties.
             *  </ul>
             * \endif
             */
            keyWord = "temperatureRef";
            status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*feaMaterial)[i].temperatureRef);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaMaterial
             *
             * \if (MYSTRAN || NASTRAN)
             *  <ul>
             *  <li> <B>dampingCoeff = 0.0</B> </li> <br>
             *  Damping coefficient for the material.
             *  </ul>
             * \elseif ABAQUS
             *
             * Something else ....
             *
             * \else
             *  <ul>
             *  <li> <B>dampingCoeff = 0.0</B> </li> <br>
             *  Damping coefficient for the material.
             * </ul>
             * \endif
             */
            keyWord = "dampingCoeff";
            status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*feaMaterial)[i].dampingCoeff);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaMaterial
             *
             *
             * \if MYSTRAN
             * \elseif NASTRAN
             *  <ul>
             *  <li> <B>yieldAllow = 0.0</B> </li> <br>
             *  Yield strength/allowable for the material.
             *  </ul>
             * \elseif ABAQUS
             *
             * Something else ....
             *
             * \else
             *  <ul>
             *  <li> <B>yieldAllow = 0.0</B> </li> <br>
             *  Yield strength/allowable for the material.
             *  </ul>
             * \endif
             */
            keyWord = "yieldAllow";
			status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				status = string_toDouble(keyValue, &(*feaMaterial)[i].yieldAllow);
				if (keyValue != NULL) {
					EG_free(keyValue);
					keyValue = NULL;
				}
				if (status != CAPS_SUCCESS) return status;
			}

            /*! \page feaMaterial
             *
             *
             * \if MYSTRAN
             * \elseif NASTRAN
             *  <ul>
             *  <li> <B>tensionAllow = 0.0</B> </li> <br>
             *  Tension allowable for the material.
             *  </ul>
             * \elseif ABAQUS
             *
             * Something else ....
             *
             * \else
             *  <ul>
             *  <li> <B>tensionAllow = 0.0</B> </li> <br>
             *  Tension allowable for the material.
             *  </ul>
             * \endif
             */
            keyWord = "tensionAllow";
            status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*feaMaterial)[i].tensionAllow);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaMaterial
             *
             * \if MYSTRAN
             * \elseif NASTRAN
             *  <ul>
             *  <li> <B>compressAllow = 0.0</B> </li> <br>
             *  Compression allowable for the material.
             *  </ul>
             * \elseif ABAQUS
             *
             * Something else ....
             *
             * \else
             *  <ul>
             *  <li> <B>compressAllow = 0.0</B> </li> <br>
             *  Compression allowable for the material.
             *  </ul>
             * \endif
             */
            keyWord = "compressAllow";
            status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*feaMaterial)[i].compressAllow);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaMaterial
             *
             * \if MYSTRAN
             * \elseif NASTRAN
             *  <ul>
             *  <li> <B>shearAllow = 0.0</B> </li> <br>
             *  Shear allowable for the material.
             *  </ul>
             * \elseif ABAQUS
             *
             * Something else ....
             *
             * \else
             *  <ul>
             *  <li> <B>shearAllow = 0.0</B> </li> <br>
             *  Shear allowable for the material.
             *  </ul>
             * \endif
             */
            keyWord = "shearAllow";
            status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*feaMaterial)[i].shearAllow);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaMaterial
             *
             * \if MYSTRAN
             * <ul>
             *  <li> <B>youngModulusLateral = 0.0</B> </li> <br>
             *  Elastic modulus in lateral direction for an orthotropic material
             * </ul>
             * \elseif NASTRAN
             * \elseif ABAQUS
             *
             * Something else ....
             *
             * \else
             * <ul>
             *  <li> <B>youngModulusLateral = 0.0</B> </li> <br>
             *  Elastic modulus in lateral direction for an orthotropic material
             * </ul>
             * \endif
             */
            keyWord = "youngModulusLateral";
            status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*feaMaterial)[i].youngModulusLateral);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaMaterial
             *
             * \if MYSTRAN
             * <ul>
             *  <li> <B>shearModulusTrans1Z = 0.0</B> </li> <br>
             *  Transverse shear modulus in the 1-Z plane for an orthotropic material
             * </ul>
             * \elseif NASTRAN
             * \elseif ABAQUS
             * Something else ....
             *
             * \else
             *  <ul>
             *  <li> <B>shearModulusTrans1Z = 0.0</B> </li> <br>
             *  Transverse shear modulus in the 1-Z plane for an orthotropic material
             * </ul>
             * \endif
             */
            keyWord = "shearModulusTrans1Z";
            status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*feaMaterial)[i].shearModulusTrans1Z);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaMaterial
             *
             * \if MYSTRAN
             * <ul>
             *  <li> <B>shearModulusTrans2Z = 0.0</B> </li> <br>
             *  Transverse shear modulus in the 2-Z plane for an orthotropic material
             * </ul>
             * \elseif NASTRAN
             * \elseif ABAQUS
             *
             * Something else ....
             *
             * \else
             *  <ul>
             *  <li> <B>shearModulusTrans2Z = 0.0</B> </li> <br>
             *  Transverse shear modulus in the 2-Z plane for an orthotropic material
             *  </ul>
             * \endif
             */
            keyWord = "shearModulusTrans2Z";
            status = search_jsonDictionary( materialTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*feaMaterial)[i].shearModulusTrans2Z);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

        } else {

        	/*! \page feaMaterial
			 * \section keyStringMaterial Single Value String
			 *
			 * If "Value" is a string, the string value may correspond to an entry in a predefined material lookup
			 * table. NOT YET IMPLEMENTED!!!!
			 *
			 */
            // CALL material look up
        	printf("\tError: Material tuple value is expected to be a JSON string\n");
        	return CAPS_BADVALUE;
        }
    }

    if (keyValue != NULL) {
        EG_free(keyValue);
        keyValue = NULL;
    }

    printf("\tDone getting FEA materials\n");
    return CAPS_SUCCESS;
}

// Get the property properties from a capsTuple
int fea_getProperty(int numPropertyTuple,
                    capsTuple propertyTuple[],
                    mapAttrToIndexStruct *attrMap,
                    feaProblemStruct *feaProblem) {

	/*! \page feaProperty FEA Property
	 * Structure for the property tuple  = ("Property Name", "Value").
	 * "Property Name" defines the reference <em>capsGroup</em> for the property being specified.
	 *	The "Value" can either be a JSON String dictionary (see Section \ref jsonStringProperty) or a single string keyword
     *	(see Section \ref keyStringProperty).
	 */

    int status; //Function return

    int i, j,  matIndex; // Indexing
    int found; // Bool test function

    int pidIndex; // Property identification index found in attribute map

    int tempInteger;

    char *keyValue = NULL;
    char *keyWord = NULL;
    char *tempString = NULL;
    char **tempStringArray = NULL;

    // Destroy our property structures coming in if aren't 0 and NULL already
    for (i = 0; i < feaProblem->numProperty; i++) {
    	status = destroy_feaPropertyStruct(&feaProblem->feaProperty[i]);
    	if (status != CAPS_SUCCESS) return status;
    }
    if (feaProblem->feaProperty != NULL) EG_free(feaProblem->feaProperty);
    feaProblem->feaProperty = NULL;
    feaProblem->numProperty = 0;

    printf("\nGetting FEA properties.......\n");

    feaProblem->numProperty = numPropertyTuple;
    printf("\tNumber of properties - %d\n", feaProblem->numProperty);

    if (feaProblem->numProperty > 0) {

        feaProblem->feaProperty = (feaPropertyStruct *) EG_alloc(feaProblem->numProperty * sizeof(feaPropertyStruct));
        if (feaProblem->feaProperty == NULL) return EGADS_MALLOC;

    } else {
        printf("\tNumber of property values in input tuple is 0\n");
        return CAPS_NOTFOUND;
    }

    for (i = 0; i < feaProblem->numProperty; i++) {
        status = initiate_feaPropertyStruct(&feaProblem->feaProperty[i]);
        if (status != CAPS_SUCCESS) {
            printf("\tUnable to initiate feaProperty structure (number = %d\n",i);
            return status;
        }
    }

    for (i = 0; i < feaProblem->numProperty; i++) {

        printf("\tProperty name - %s\n", propertyTuple[i].name);

        // Set property name from tuple name
        feaProblem->feaProperty[i].name = (char *) EG_alloc(((strlen(propertyTuple[i].name)) + 1)*sizeof(char));
        if (feaProblem->feaProperty[i].name == NULL) return EGADS_MALLOC;

        memcpy(feaProblem->feaProperty[i].name, propertyTuple[i].name, strlen(propertyTuple[i].name)*sizeof(char));
        feaProblem->feaProperty[i].name[strlen(propertyTuple[i].name)] = '\0';

        // Get to property ID number from the attribute map
        status = get_mapAttrToIndexIndex(attrMap, (const char *) feaProblem->feaProperty[i].name, &pidIndex);

        if (status != CAPS_SUCCESS) {

            printf("\tTuple name %s not found in attribute map of PIDS!!!!\n", feaProblem->feaProperty[i].name);

            if (keyValue != NULL) {
                EG_free(keyValue);
                keyValue = NULL;
            }

            return status;

        } else {

            feaProblem->feaProperty[i].propertyID = pidIndex;
        }

        // Do we have a json string?
        if (strncmp(propertyTuple[i].value, "{", 1) == 0) {
            //printf("JSON String - %s\n",propertyTuple[i].value);

            /*! \page feaProperty
             * \section jsonStringProperty JSON String Dictionary
             *
             * If "Value" is JSON string dictionary
             * \if (MYSTRAN || NASTRAN)
			 *  (eg. "Value" = {"shearMembraneRatio": 0.83, "bendingInertiaRatio": 1.0, "membraneThickness": 0.2, "propertyType": "Shell"})
			 * \endif
			 *  the following keywords ( = default values) may be used:
			 *
             *
             * \if (MYSTRAN || NASTRAN)
             * <ul>
             *  <li> <B>propertyType = No Default value</B> </li> <br>
             *  Type of property to apply to a give capsGroup <em>Name</em>. Options: Rod, Bar, Shear, Shell, Composite, and Solid
             * </ul>
             * \elseif ABAQUS
             *
             * Something else ....
             *
             * \else
             * <ul>
             *  <li> <B>propertyType = No Default value</B> </li> <br>
             *  Type of property to apply to a give capsGroup <em>Name</em>. Options: Rod, Bar, Shear, Shell, Composite, and Solid
             * </ul>
             * \endif
             *
             */

            // Get property Type
            keyWord = "propertyType";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                //{UnknownProperty, Rod, Bar, Shear, Shell, Composite, Solid}
                if      (strcasecmp(keyValue, "\"Rod\"")      == 0) feaProblem->feaProperty[i].propertyType = Rod;
                else if (strcasecmp(keyValue, "\"Bar\"")      == 0) feaProblem->feaProperty[i].propertyType = Bar;
                else if (strcasecmp(keyValue, "\"Shear\"")    == 0) feaProblem->feaProperty[i].propertyType = Shear;
                else if (strcasecmp(keyValue, "\"Shell\"")    == 0) feaProblem->feaProperty[i].propertyType = Shell;
                else if (strcasecmp(keyValue, "\"Composite\"")== 0) feaProblem->feaProperty[i].propertyType = Composite;
                else if (strcasecmp(keyValue, "\"Solid\"")    == 0) feaProblem->feaProperty[i].propertyType = Solid;
                else {

                    printf("\tUnrecognized \"%s\" specified (%s) for Property tuple %s, current options are "
                           "\"Rod, Bar, Shear, Shell, Composite, and Solid\"\n", keyWord,
                                                                                 keyValue,
                                                                                 propertyTuple[i].name);
                    if (keyValue != NULL) {
                        EG_free(keyValue);
                        keyValue = NULL;
                    }

                    return CAPS_NOTFOUND;
                }

            } else {

                printf("\tNo \"%s\" specified for Property tuple %s, this mandatory! Current options options are "
                           "\"Rod, Bar, Shear, Shell, Composite, and Solid\"\n", keyWord,
                                                                                 propertyTuple[i].name);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }

                return CAPS_NOTFOUND;
            }

            if (keyValue != NULL) {
                EG_free(keyValue);
                keyValue = NULL;
            }

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 *  <ul>
			 *  <li> <B>material = "Material Name" (\ref feaMaterial) </B> </li> <br>
			 *  "Material Name" from \ref feaMaterial to use for property. If no material is set the first material
			 *   created will be used
			 *  </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 *  <ul>
			 *  <li> <B>material = `Material Name' (\ref feaMaterial) </B> </li> <br>
			 *  `Material Name' from \ref feaMaterial to use for property. If no material is set the first material
			 *   created will be used
			 * </ul>
			 * \endif
			 */
            keyWord = "material";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

               found = (int) false;
               for (matIndex = 0; matIndex < feaProblem->numMaterial; matIndex++ ) {

            	   if (tempString != NULL) EG_free(tempString);
                   tempString = string_removeQuotation(keyValue);

                   if (strcasecmp(feaProblem->feaMaterial[matIndex].name, tempString) == 0) {
                       feaProblem->feaProperty[i].materialID = feaProblem->feaMaterial[matIndex].materialID;

                       feaProblem->feaProperty[i].materialName = (char *) EG_alloc((strlen(feaProblem->feaMaterial[matIndex].name)+1) *sizeof(char));
                       if (feaProblem->feaProperty[i].name == NULL) return EGADS_MALLOC;

                       memcpy(feaProblem->feaProperty[i].materialName,
                              feaProblem->feaMaterial[matIndex].name,
                              strlen(feaProblem->feaMaterial[matIndex].name)*sizeof(char));

                       feaProblem->feaProperty[i].materialName[strlen(feaProblem->feaMaterial[matIndex].name)] = '\0';

                       found = (int) true;
                       break;
                   }
               }

               if (tempString != NULL) EG_free(tempString);
               tempString = NULL;

               if (found == (int) false) {
                   printf("\tUnrecognized \"%s\" specified (%s) for Property tuple %s. No match in Material tuple\n", keyWord,
                                                                                                                    keyValue,
                                                                                                                    propertyTuple[i].name);
                   if (keyValue != NULL) {
                       EG_free(keyValue);
                       keyValue = NULL;
                   }
                   return CAPS_NOTFOUND;
               }

            } else {

				if (feaProblem->feaProperty[i].propertyType != Composite) {
					printf("\tNo \"%s\" specified for Property tuple %s, defaulting to an index of 1\n", keyWord,
																									   propertyTuple[i].name);
            	}
                feaProblem->feaProperty[i].materialID = 1;
                if (feaProblem->numMaterial > 0) {
                	feaProblem->feaProperty[i].materialName = (char *) EG_alloc((strlen(feaProblem->feaMaterial[0].name)+1) *sizeof(char));
                    if (feaProblem->feaProperty[i].name == NULL) {
                        if (keyValue != NULL) EG_free(keyValue);
                        return EGADS_MALLOC;
                    }
                    memcpy(feaProblem->feaProperty[i].materialName,
                           feaProblem->feaMaterial[0].name,
                           strlen(feaProblem->feaMaterial[0].name)*sizeof(char));
                    feaProblem->feaProperty[i].materialName[strlen(feaProblem->feaMaterial[0].name)] = '\0';
                }
            }

            if (keyValue != NULL) {
                EG_free(keyValue);
                keyValue = NULL;
            }

            //  Fill up properties ///

            // Rods
            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 *  <ul>
			 *  <li> <B>crossSecArea = 0.0</B> </li> <br>
			 *  Cross sectional area.
			 *  </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 *  <ul>
		  	 *  <li> <B>crossSecArea = 0.0</B> </li> <br>
			 *  Cross sectional area.
			 *  </ul>
			 * \endif
			 */
            keyWord = "crossSecArea";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaProperty[i].crossSecArea);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 *  <ul>
			 *  <li> <B>torsionalConst = 0.0</B> </li> <br>
			 *  Torsional constant.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 *  <ul>
			 *  <li> <B>torsionalConst = 0.0</B> </li> <br>
			 *  Torsional constant.
			 * </ul>
			 * \endif
			 */
            keyWord = "torsionalConst";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaProperty[i].torsionalConst);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 *  <ul>
			 *  <li> <B>torsionalStressReCoeff = 0.0</B> </li> <br>
			 *  Torsional stress recovery coefficient.
			 *  </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>torsionalStressReCoeff = 0.0</B> </li> <br>
			 *  Torsional stress recovery coefficient.
			 * </ul>
			 * \endif
			 */
            keyWord = "torsionalStressReCoeff";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaProperty[i].torsionalStressReCoeff);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 *  <ul>
			 *  <li> <B>massPerLength = 0.0</B> </li> <br>
			 *  Mass per unit length.
			 *  </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 *  <ul>
			 *  <li> <B>massPerLength = 0.0</B> </li> <br>
			 *  Mass per unit length.
			 * </ul>
			 * \endif
			 */
            keyWord = "massPerLength";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaProperty[i].massPerLength);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            // Bar - see rod for additional variables

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>zAxisInertia = 0.0</B> </li> <br>
			 *  Section moment of inertia about the element z-axis.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>zAxisInertia = 0.0</B> </li> <br>
			 *  Section moment of inertia about the element z-axis.
			 * </ul>
			 * \endif
			 */
            keyWord = "zAxisInertia";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaProperty[i].zAxisInertia);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 *  <ul>
			 *  <li> <B>yAxisInertia = 0.0</B> </li> <br>
			 *  Section moment of inertia about the element y-axis.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>yAxisInertia = 0.0</B> </li> <br>
			 *  Section moment of inertia about the element y-axis.
			 * </ul>
			 * \endif
			 */
            keyWord = "yAxisInertia";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaProperty[i].yAxisInertia);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>yCoords[4] = [0.0, 0.0, 0.0, 0.0]</B> </li> <br>
			 *  Element y-coordinates, in the bar cross-section, of four points at which to recover stresses
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>yCoords[4] = [0.0, 0.0, 0.0, 0.0]</B> </li> <br>
			 *  Element y-coordinates, in the bar cross-section, of four points at which to recover stresses
			 * </ul>
			 * \endif
			 */
            keyWord = "yCoords";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {
                status = string_toDoubleArray(keyValue,
                                              (int) sizeof(feaProblem->feaProperty[i].yCoords)/sizeof(double),
                                              feaProblem->feaProperty[i].yCoords);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>zCoords[4] = [0.0, 0.0, 0.0, 0.0]</B> </li> <br>
			 *  Element z-coordinates, in the bar cross-section, of four points at which to recover stresses
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>zCoords[4] = [0.0, 0.0, 0.0, 0.0]</B> </li> <br>
			 *  Element z-coordinates, in the bar cross-section, of four points at which to recover stresses
			 * </ul>
			 * \endif
			 */
            keyWord = "zCoords";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {
                status = string_toDoubleArray(keyValue,
                                              (int) sizeof(feaProblem->feaProperty[i].zCoords)/sizeof(double),
                                              feaProblem->feaProperty[i].zCoords);

                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>areaShearFactors[2] = [0.0, 0.0]</B> </li> <br>
			 *  Area factors for shear.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>areaShearFactors[2] = [0.0, 0.0]</B> </li> <br>
			 *  Area factors for shear.
			 * </ul>
			 * \endif
			 */
            keyWord = "areaShearFactors";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {
                status = string_toDoubleArray(keyValue,
                                             (int) sizeof(feaProblem->feaProperty[i].areaShearFactors)/sizeof(double),
                                              feaProblem->feaProperty[i].areaShearFactors);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
 			 *  <li> <B>crossProductInertia = 0.0</B> </li> <br>
			 *  Section cross-product of inertia.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>crossProductInertia = 0.0</B> </li> <br>
			 *  Section cross-product of inertia.
			 * </ul>
			 * \endif
			 */
            keyWord = "crossProductInertia";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {
                status = string_toDouble(keyValue, &feaProblem->feaProperty[i].crossProductInertia);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }


            // Shear

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
 			 *  <li> <B>shearPanelThickness = 0.0</B> </li> <br>
			 *  Shear panel thickness.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>shearPanelThickness = 0.0</B> </li> <br>
			 *  Shear panel thickness.
			 * </ul>
			 * \endif
			 */
            keyWord = "shearPanelThickness";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {
                status = string_toDouble(keyValue, &feaProblem->feaProperty[i].shearPanelThickness);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>nonStructMassPerArea = 0.0</B> </li> <br>
			 *  Nonstructural mass per unit area.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>nonStructMassPerArea = 0.0</B> </li> <br>
			 *  Nonstructural mass per unit area.
			 * </ul>
			 * \endif
			 */
            keyWord = "nonStructMassPerArea";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {
                status = string_toDouble(keyValue, &feaProblem->feaProperty[i].nonStructMassPerArea);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            // Shell

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>membraneThickness = 0.0</B> </li> <br>
			 *  Membrane thickness.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>membraneThickness = 0.0</B> </li> <br>
			 *  Membrane thickness.
			 * </ul>
			 * \endif
			 */
            keyWord = "membraneThickness";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {
                status = string_toDouble(keyValue, &feaProblem->feaProperty[i].membraneThickness);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }


            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>bendingInertiaRatio = 1.0</B> </li> <br>
			 *  Ratio of actual bending moment inertia to the bending inertia of a solid plate of thickness "membranThickness"
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>bendingInertiaRatio = 1.0</B> </li> <br>
			 *  Ratio of actual bending moment inertia to the bending inertia of a solid plate of thickness "membranThickness"
			 * </ul>
			 * \endif
			 */
            keyWord = "bendingInertiaRatio";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {
            	status = string_toDouble(keyValue, &feaProblem->feaProperty[i].bendingInertiaRatio);
            	if (keyValue != NULL) {
            		EG_free(keyValue);
            		keyValue = NULL;
            	}
            	if (status != CAPS_SUCCESS) return status;
            }


            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>shearMembraneRatio = 5.0/6.0</B> </li> <br>
			 *  Ratio shear thickness to membrane thickness.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>shearMembraneRatio = 5.0/6.0</B> </li> <br>
			 *  Ratio shear thickness to membrane thickness.
			 * </ul>
			 * \endif
			 */
            keyWord = "shearMembraneRatio";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {
            	status = string_toDouble(keyValue, &feaProblem->feaProperty[i].shearMembraneRatio);
            	if (keyValue != NULL) {
            		EG_free(keyValue);
            		keyValue = NULL;
            	}
            	if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>materialBending = "Material Name" (\ref feaMaterial)</B> </li> <br>
			 *  "Material Name" from \ref feaMaterial to use for property bending. If no material is given and
			 *  "bendingIneriaRatio" is greater than 0, the material name provided in "material" is used.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>materialBending = "Material Name" (\ref feaMaterial)</B> </li> <br>
			 *  "Material Name" from \ref feaMaterial to use for property bending.
			 * </ul>
			 * \endif
			 */
            keyWord = "materialBending"; // Shell specific materials
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

            	found = (int) false;
            	for (matIndex = 0; matIndex < feaProblem->numMaterial; matIndex++ ) {

            		if (tempString != NULL) EG_free(tempString);
            		tempString = string_removeQuotation(keyValue);

            		if (strcasecmp(feaProblem->feaMaterial[matIndex].name, tempString) == 0) {
            			feaProblem->feaProperty[i].materialBendingID = feaProblem->feaMaterial[matIndex].materialID;
            			found = (int) true;

            			if (keyValue != NULL) {
            				EG_free(keyValue);
            				keyValue = NULL;
            			}
            			break;
            		}

            	}

            	if (tempString != NULL) EG_free(tempString);
            	tempString = NULL;

            	if (found == (int) false) {
            		printf("\tUnrecognized \"%s\" specified (%s) for Property tuple %s. No match in Material tuple\n", keyWord,
                                                                                                                     keyValue,
																													 propertyTuple[i].name);
            		if (keyValue != NULL) {
            			EG_free(keyValue);
            			keyValue = NULL;
            		}
            		return CAPS_NOTFOUND;
            	}

            } else { // Don't default to anything - yet

            	/*
                printf("No \"%s\" specified for Property tuple %s which is a shell element, "
                       " no bending material will be specified\n", keyWord,
                                                                   propertyTuple[i].name);

                feaProblem->feaProperty[i].materialBendingID = 0;
                */

            	if (feaProblem->feaProperty[i].bendingInertiaRatio > 0) {

            		feaProblem->feaProperty[i].materialBendingID = feaProblem->feaProperty[i].materialID;
            	}
            }



            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>materialShear = "Material Name" (\ref feaMaterial)</B> </li> <br>
			 *  "Material Name" from \ref feaMaterial to use for property shear. If no material is given and
			 *  "shearMembraneRatio" is greater than 0, the material name provided in "material" is used.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>materialShear = "Material Name" (\ref feaMaterial)</B> </li> <br>
			 *  "Material Name" from \ref feaMaterial to use for property shear.
			 * </ul>
			 * \endif
			 */
            keyWord = "materialShear"; // Shell specific materials
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

            	found = (int) false;
            	for (matIndex = 0; matIndex < feaProblem->numMaterial; matIndex++ ) {

            		if (tempString != NULL) EG_free(tempString);
            		tempString = string_removeQuotation(keyValue);

            		if (strcasecmp(feaProblem->feaMaterial[matIndex].name, tempString) == 0) {
            			feaProblem->feaProperty[i].materialShearID = feaProblem->feaMaterial[matIndex].materialID;
            			found = (int) true;

            			if (keyValue != NULL) {
            				EG_free(keyValue);
            				keyValue = NULL;
            			}
            			break;
            		}

            	}

            	if (tempString != NULL) EG_free(tempString);
            	tempString = NULL;

            	if (found == (int) false) {
            		printf("\tUnrecognized \"%s\" specified (%s) for Property tuple %s. No match in Material tuple\n", keyWord,
            				                                                                                         keyValue,
                                                                                                                     propertyTuple[i].name);
            		if (keyValue != NULL) {
            			EG_free(keyValue);
            			keyValue = NULL;
            		}
            		return CAPS_NOTFOUND;
            	}

            } else {
            	/*
                printf("No \"%s\" specified for Property tuple %s which is a shell element, "
                       " no shear material will be specified\n", keyWord,
						    	        						 propertyTuple[i].name);

                feaProblem->feaProperty[i].materialShearID = 0;
                */

            	if (feaProblem->feaProperty[i].shearMembraneRatio > 0) {
            		feaProblem->feaProperty[i].materialShearID = feaProblem->feaProperty[i].materialID;
            	}
            }

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>massPerArea = 0.0</B> </li> <br>
			 *  Mass per unit area.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>massPerArea = 0.0</B> </li> <br>
			 *  Mass per unit area.
			 * </ul>
			 * \endif
			 */
            keyWord = "massPerArea";
            status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {
                status = string_toDouble(keyValue, &feaProblem->feaProperty[i].massPerArea);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            // Composite

            /*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>compositeMaterial = "no default" </B> </li> <br>
			 *  List of "Material Name"s, ["Material Name -1", "Material Name -2", ...], from \ref feaMaterial to use for composites.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>compositeMaterial = "no default" </B> </li> <br>
			 *  List of "Material Name"s, ["Material Name -1", "Material Name -2", ...], from \ref feaMaterial to use for composites.
			 * </ul>
			 * \endif
			 */
            keyWord = "compositeMaterial";
			status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				status = string_toStringDynamicArray(keyValue, &feaProblem->feaProperty[i].numPly, &tempStringArray);
				if (status != CAPS_SUCCESS) return status;

				feaProblem->feaProperty[i].compositeMaterialID = (int *) EG_alloc(feaProblem->feaProperty[i].numPly*sizeof(int));
				if (feaProblem->feaProperty[i].compositeMaterialID == NULL) {
					(void) string_freeArray(feaProblem->feaProperty[i].numPly, &tempStringArray);
					if (keyValue != NULL) EG_free(keyValue);
					return EGADS_MALLOC;
				}

				for (j = 0; j < feaProblem->feaProperty[i].numPly; j++) {
					found = (int) false;
					for (matIndex = 0; matIndex < feaProblem->numMaterial; matIndex++ ) {

						if (tempString != NULL) EG_free(tempString);
						tempString = string_removeQuotation(tempStringArray[j]);

						if (strcasecmp(feaProblem->feaMaterial[matIndex].name, tempString) == 0) {
							feaProblem->feaProperty[i].compositeMaterialID[j] = feaProblem->feaMaterial[matIndex].materialID;
							found = (int) true;
							break;
						}

					}

					if (found == (int) false) {
						printf("\tUnrecognized \"%s\" specified (%s) for Property tuple %s. No match in Material tuple\n", tempString,
																														 keyWord,
																														 propertyTuple[i].name);

						if (keyValue != NULL) EG_free(keyValue);
						if (tempString != NULL) EG_free(tempString);

						(void) string_freeArray(feaProblem->feaProperty[i].numPly, &tempStringArray);

						return CAPS_NOTFOUND;
					 }

					if (tempString != NULL) EG_free(tempString);
					tempString = NULL;
				}
			}


			if (keyValue != NULL) {
				EG_free(keyValue);
				keyValue = NULL;
			}

			status = string_freeArray(feaProblem->feaProperty[i].numPly, &tempStringArray);
			if (status != CAPS_SUCCESS) return status;

	        /*! \page feaProperty
	         *
	         * \if (MYSTRAN || NASTRAN)
	         * <ul>
	         *  <li> <B>shearBondAllowable = 0.0 </B> </li> <br>
	         *  Allowable interlaminar shear stress.
	         * </ul>
	         * \elseif ABAQUS
	         *
	         * Something else ....
	         *
	         * \else
	         * <ul>
	         *  <li> <B>shearBondAllowable = 0.0 </B> </li> <br>
	         *  Allowable interlaminar shear stress.
	         * </ul>
	         * \endif
	         */
            keyWord = "shearBondAllowable";
			status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {
				status = string_toDouble(keyValue, &feaProblem->feaProperty[i].compositeShearBondAllowable);
				if (keyValue != NULL) {
					EG_free(keyValue);
					keyValue = NULL;
				}
				if (status != CAPS_SUCCESS) return status;
			}

	        /*! \page feaProperty
	         *
	         * \if (MYSTRAN || NASTRAN)
	         * <ul>
	         *  <li> <B>symmetricLaminate = 0 </B> </li> <br>
	         *  Symmetric lamination option. 1- SYM only half the plies are specified, for odd number plies 1/2 thickness
	         *   of center ply is specified the first ply is the bottom ply in the stack, default (0) all plies specified.
	         * </ul>
	         * \elseif ABAQUS
	         *
	         * Something else ....
	         *
	         * \else
	         * <ul>
	         *  <li> <B>symmetricLaminate = 0 </B> </li> <br>
	         *  Symmetric lamination option. 1- SYM only half the plies are specified, for odd number plies 1/2 thickness
	         *   of center ply is specified the first ply is the bottom ply in the stack, default (0) all plies specified
	         * </ul>
	         * \endif
	         */
			keyWord = "symmetricLaminate";
			status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {
				status = string_toInteger(keyValue, &feaProblem->feaProperty[i].compositeSymmetricLaminate);
				if (keyValue != NULL) {
					EG_free(keyValue);
					keyValue = NULL;
				}
				if (status != CAPS_SUCCESS) return status;
			}

	        /*! \page feaProperty
	         *
	         * \if (MYSTRAN || NASTRAN)
	         * <ul>
	         *  <li> <B>compositeFailureTheory = "(no default)" </B> </li> <br>
	         *  Composite failure theory. Options: "HILL", "HOFF", "TSAI", and "STRN"
	         * </ul>
	         * \elseif ABAQUS
	         *
	         * Something else ....
	         *
	         * \else
	         * <ul>
	         *  <li> <B>compositeFailureTheory = "(no default)" </B> </li> <br>
	         *  Composite failure theory.
	         * </ul>
	         * \endif
	         */
			keyWord = "compositeFailureTheory";
			status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {
				feaProblem->feaProperty[i].compositeFailureTheory = string_removeQuotation(keyValue);
				if (keyValue != NULL) {
					EG_free(keyValue);
					keyValue = NULL;
				}
				if (status != CAPS_SUCCESS) return status;
			}

	        /*! \page feaProperty
	         *
	         * \if (MYSTRAN || NASTRAN)
	         * <ul>
	         *  <li> <B>compositeThickness = (no default) </B> </li> <br>
	         *  List of composite thickness for each layer (eg. [1.2, 4.0, 3.0]). If the length of this list doesn't match the length
	         *  of the "compositeMaterial" list, the list is either truncated [ >length("compositeMaterial")] or expanded [ <length("compositeMaterial")]
	         *  in which case the last thickness provided is repeated.
	         * </ul>
	         * \elseif ABAQUS
	         *
	         * Something else ....
	         *
	         * \else
	         * <ul>
	         *  <li> <B>compositeThickness = (no default) </B> </li> <br>
	         *  List of composite thickness for each layer (eg. [1.2, 4.0, 3.0]). If the length of this list doesn't match the length
	         *  of the "compositeMaterial" list, the list is either truncated [ >length("compositeMaterial")] or expanded [ <length("compositeMaterial")]
	         *  in which case the last thickness provided is repeated.
	         * </ul>
	         * \endif
	         */
			keyWord = "compositeThickness";
			status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {
				status = string_toDoubleDynamicArray(keyValue, &tempInteger,
															   &feaProblem->feaProperty[i].compositeThickness);
				if (keyValue != NULL) {
					EG_free(keyValue);
					keyValue = NULL;
				}
				if (status != CAPS_SUCCESS) return status;

				if (tempInteger < feaProblem->feaProperty[i].numPly) {

					printf("\tThe number of thicknesses provided does not match the number of materials for the composite. "
							"The last thickness will be repeated %d times\n", feaProblem->feaProperty[i].numPly - tempInteger);

					feaProblem->feaProperty[i].compositeThickness = (double *) EG_reall(feaProblem->feaProperty[i].compositeThickness,
																					 feaProblem->feaProperty[i].numPly*sizeof(double));

					if (feaProblem->feaProperty[i].compositeThickness == NULL ) return EGADS_MALLOC;

					for (j = 0; j < feaProblem->feaProperty[i].numPly - tempInteger; j++) {

						feaProblem->feaProperty[i].compositeThickness[j+tempInteger] = feaProblem->feaProperty[i].compositeThickness[tempInteger-1];
					}
				}

				if (tempInteger > feaProblem->feaProperty[i].numPly) {

					printf("\tThe number of thicknesses provided does not match the number of materials for the composite. "
							"The last %d thicknesses will be not be used\n", tempInteger -feaProblem->feaProperty[i].numPly);

					feaProblem->feaProperty[i].compositeThickness = (double *) EG_reall(feaProblem->feaProperty[i].compositeThickness,
																					 feaProblem->feaProperty[i].numPly*sizeof(double));

					if (feaProblem->feaProperty[i].compositeThickness == NULL ) return EGADS_MALLOC;
				}
			} else {

				if (feaProblem->feaProperty[i].numPly != 0 &&
					feaProblem->feaProperty[i].propertyType == Composite) {

					printf("\"compositeMaterial\" have been set but no thicknesses (\"compositeThickness\") provided!!!\n");
					return CAPS_BADVALUE;

				}
			}

			/*! \page feaProperty
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>compositeOrientation = (no default) </B> </li> <br>
			 *  List of composite orientations (angle relative element material axis) for each layer (eg. [5.0, 10.0, 30.0]).
			 *  If the length of this list doesn't match the length of the "compositeMaterial" list, the list is either
			 *  truncated [ >length("compositeMaterial")] or expanded [ <length("compositeMaterial")] in which case
			 *  the last orientation provided is repeated.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>compositeOrientation = (no default) </B> </li> <br>
			 *  List of composite orientations (angle relative element material axis) for each layer (eg. [5.0, 10.0, 30.0]).
			 *  If the length of this list doesn't match the length of the "compositeMaterial" list, the list is either
			 *  truncated [ >length("compositeMaterial")] or expanded [ <length("compositeMaterial")] in which case
			 *  the last orientation provided is repeated.
			 * </ul>
			 * \endif
			 */
			keyWord = "compositeOrientation";
			status = search_jsonDictionary( propertyTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {
				status = string_toDoubleDynamicArray(keyValue, &tempInteger,
														       &feaProblem->feaProperty[i].compositeOrientation);
				if (keyValue != NULL) {
					EG_free(keyValue);
					keyValue = NULL;
				}
				if (status != CAPS_SUCCESS) return status;


				if (tempInteger < feaProblem->feaProperty[i].numPly) {

					printf("\tThe number of orientations provided does not match the number of materials for the composite. "
							"The last orientation will be repeated %d times\n", feaProblem->feaProperty[i].numPly - tempInteger);

					feaProblem->feaProperty[i].compositeOrientation = (double *) EG_reall(feaProblem->feaProperty[i].compositeOrientation,
																					 feaProblem->feaProperty[i].numPly*sizeof(double));

					if (feaProblem->feaProperty[i].compositeOrientation == NULL ) return EGADS_MALLOC;

					for (j = 0; j < feaProblem->feaProperty[i].numPly - tempInteger; j++) {

						feaProblem->feaProperty[i].compositeOrientation[j+tempInteger] = feaProblem->feaProperty[i].compositeOrientation[tempInteger-1];
					}
				}

				if (tempInteger > feaProblem->feaProperty[i].numPly) {

					printf("\tThe number of orientations provided does not match the number of materials for the composite. "
							"The last %d orientation will be not be used\n", tempInteger -feaProblem->feaProperty[i].numPly);

					feaProblem->feaProperty[i].compositeOrientation = (double *) EG_reall(feaProblem->feaProperty[i].compositeOrientation,
																					 feaProblem->feaProperty[i].numPly*sizeof(double));

					if (feaProblem->feaProperty[i].compositeOrientation == NULL ) return EGADS_MALLOC;
				}

			} else {

				if (feaProblem->feaProperty[i].numPly != 0 &&
					feaProblem->feaProperty[i].propertyType == Composite) {

					printf("\t\"compositeMaterial\" have been set but no Orientation  (\"compositeOrientation\") provided!!!\n");
					return CAPS_BADVALUE;
				}
			}

        } else {

        	/*! \page feaProperty
			 * \section keyStringProperty Single Value String
			 *
			 * If "Value" is a string, the string value may correspond to an entry in a predefined property lookup
			 * table. NOT YET IMPLEMENTED!!!!
			 *
			 */

            // CALL property look up
       	 printf("\tError: Property tuple value is expected to be a JSON string\n");
       	 return CAPS_BADVALUE;
        }
    }

    printf("\tDone getting FEA properties\n");

    if (keyValue != NULL) {
        EG_free(keyValue);
        keyValue = NULL;
    }

    return CAPS_SUCCESS;

}

// Get the constraint properties from a capsTuple
int fea_getConstraint(int numConstraintTuple,
                      capsTuple constraintTuple[],
                      mapAttrToIndexStruct *attrMap,
                      feaProblemStruct *feaProblem) {

	/*! \page feaConstraint FEA Constraint
	 * Structure for the constraint tuple  = ("Constraint Name", "Value").
	 * "Constraint Name" defines the reference name for the constraint being specified.
	 *	The "Value" can either be a JSON String dictionary (see Section \ref jsonStringConstraint) or a single string keyword
	 *	(see Section \ref keyStringConstraint).
	 */

    int status; //Function return

    int i, groupIndex, attrIndex, nodeIndex; // Indexing

    char *keyValue = NULL; // Key values from tuple searches
    char *keyWord = NULL; // Key words to find in the tuples

    int numGroupName = 0;
    char **groupName = NULL;

    // Destroy our constraint structures coming in if aren't 0 and NULL already
    for (i = 0; i < feaProblem->numConstraint; i++) {
    	status = destroy_feaConstraintStruct(&feaProblem->feaConstraint[i]);
    	if (status != CAPS_SUCCESS) return status;
    }
    if (feaProblem->feaConstraint != NULL) EG_free(feaProblem->feaConstraint);
    feaProblem->feaConstraint = NULL;
    feaProblem->numConstraint = 0;

    printf("\nGetting FEA constraints.......\n");

    feaProblem->numConstraint = numConstraintTuple;

    printf("\tNumber of constraints - %d\n", feaProblem->numConstraint);

    // Allocate constraints
    if (feaProblem->numConstraint > 0) {
        feaProblem->feaConstraint = (feaConstraintStruct *) EG_alloc(feaProblem->numConstraint * sizeof(feaConstraintStruct));

        if (feaProblem->feaConstraint == NULL ) return EGADS_MALLOC;
    }

    // Initiate constraints to default values
    for (i = 0; i < feaProblem->numConstraint; i++) {
        status = initiate_feaConstraintStruct(&feaProblem->feaConstraint[i]);
        if (status != CAPS_SUCCESS) return status;
    }

    // Loop through tuples and fill out the constraint structures
    for (i = 0; i < feaProblem->numConstraint; i++) {

        printf("\tConstraint name - %s\n", constraintTuple[i].name );

        // Set constraint name to tuple attribute name
        feaProblem->feaConstraint[i].name = (char *) EG_alloc((strlen(constraintTuple[i].name) + 1)*sizeof(char));
        if (feaProblem->feaConstraint[i].name == NULL) return EGADS_MALLOC;

        memcpy(feaProblem->feaConstraint[i].name, constraintTuple[i].name, strlen(constraintTuple[i].name)*sizeof(char));
        feaProblem->feaConstraint[i].name[strlen(constraintTuple[i].name)] = '\0';

        // Set constraint id -> 1 bias
        feaProblem->feaConstraint[i].constraintID = i+1;

        // Do we have a json string?
        if (strncmp(constraintTuple[i].value, "{", 1) == 0) {
            //printf("JSON String - %s\n", constraintTuple[i].value);

        	/*! \page feaConstraint
			 * \section jsonStringConstraint JSON String Dictionary
			 *
			 * If "Value" is JSON string dictionary
             * \if (MYSTRAN || NASTRAN)
			 *  (eg. "Value" = {"groupName": "plateEdge", "dofConstraint": 123456})
			 * \endif
			 *  the following keywords ( = default values) may be used:
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>constraintType = "ZeroDisplacement"</B> </li> <br>
			 *  Type of constraint. Options: "Displacement", "ZeroDisplacement".
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>constraintType = "ZeroDisplacement"</B> </li> <br>
			 *  Type of constraint. Options: "Displacement", "ZeroDisplacement".
			 * </ul>
			 * \endif
			 *
			 */

        	// Get constraint Type
            keyWord = "constraintType";
            status = search_jsonDictionary( constraintTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                //{UnknownConstraint, Displacement, ZeroDisplacement}
                if      (strcasecmp(keyValue, "\"Displacement\"")      == 0) feaProblem->feaConstraint[i].constraintType = Displacement;
                else if (strcasecmp(keyValue, "\"ZeroDisplacement\"")  == 0) feaProblem->feaConstraint[i].constraintType = ZeroDisplacement;
                else {

                    printf("\tUnrecognized \"%s\" specified (%s) for Constraint tuple %s, defaulting to \"ZeroDisplacement\"\n", keyWord,
                                                                                                                               keyValue,
                                                                                                                               constraintTuple[i].name);
                    feaProblem->feaConstraint[i].constraintType = ZeroDisplacement;
                }

            } else {

                printf("\tNo \"%s\" specified for Constraint tuple %s, defaulting to \"ZeroDisplacement\"\n", keyWord,
                                                                                                            constraintTuple[i].name);
                feaProblem->feaConstraint[i].constraintType = ZeroDisplacement;
            }

            if (keyValue != NULL) {
                EG_free(keyValue);
                keyValue = NULL;
            }

            // Get constraint node set

            /*! \page feaConstraint
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>groupName = "(no default)"</B> </li> <br>
			 *  Single or list of <em>capsGroup</em> names on which to apply the constraint (e.g. "Name1" or ["Name1","Name2",...].
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>groupName = "(no default)"</B> </li> <br>
			 *  Single or list of <em>capsGroup</em> names on which to apply the constraint (e.g. ["Name1","Name2",...].
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "groupName";
            status = search_jsonDictionary( constraintTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toStringDynamicArray(keyValue, &numGroupName, &groupName);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }

                if (status != CAPS_SUCCESS) return status;

                // Determine how many point constraints we have
                for (groupIndex = 0; groupIndex < numGroupName; groupIndex++) {

                    status = get_mapAttrToIndexIndex(attrMap, (const char *) groupName[groupIndex], &attrIndex);

                    if (status == CAPS_NOTFOUND) {
                        printf("\tgroupName name %s not found in attribute map of capsGroups!!!!\n", groupName[groupIndex]);
                        continue;
                    } else if (status != CAPS_SUCCESS) return status;

                    feaProblem->feaConstraint[i].numGridID = 0;
                    // Now lets loop through the grid to see how many grid points have the attrIndex
                    for (nodeIndex = 0; nodeIndex < feaProblem->feaMesh.numNodes; nodeIndex++ ) {
                        if (feaProblem->feaMesh.xyzAttrIndex[nodeIndex] == attrIndex) {

                            feaProblem->feaConstraint[i].numGridID += 1;

                            // Allocate/Re-allocate grid ID array
                            if (feaProblem->feaConstraint[i].numGridID == 1) {
                                feaProblem->feaConstraint[i].gridIDSet = (int *) EG_alloc(feaProblem->feaConstraint[i].numGridID*sizeof(int));
                            } else {
                                feaProblem->feaConstraint[i].gridIDSet = (int *) EG_reall(feaProblem->feaConstraint[i].gridIDSet,
                                                                                          feaProblem->feaConstraint[i].numGridID*sizeof(int));
                            }

                            if (feaProblem->feaConstraint[i].gridIDSet == NULL) {
                                status = string_freeArray(numGroupName, &groupName);
                                if (status == CAPS_SUCCESS) return status;
                                groupName = NULL;

                                return EGADS_MALLOC;
                            }

                            // Set grid ID value -> 1 bias
                            feaProblem->feaConstraint[i].gridIDSet[feaProblem->feaConstraint[i].numGridID-1] = nodeIndex+1;
                        }
                    }
                }

                status = string_freeArray(numGroupName, &groupName);
                if (status != CAPS_SUCCESS) return status;
                groupName = NULL;

            } else {

                printf("\tNo \"%s\" specified for Constraint tuple %s, where should these constraints be imposed?\n", keyWord,
                                                                                                                    constraintTuple[i].name);
                feaProblem->feaConstraint[i].numGridID = 0;
            }

            //Fill up constraint properties
            /*! \page feaConstraint
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>dofConstraint = 0 </B> </li> <br>
			 *  Component numbers / degrees of freedom that will be constrained (123 - zero translation in all three
			 *  directions).
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>dofConstraint = 0 </B> </li> <br>
			 *  Component numbers / degrees of freedom that will be constrained (123 - zero translation in all three
			 *  directions).
			 * </ul>
			 * \endif
			 *
			 */
			keyWord = "dofConstraint";
            status = search_jsonDictionary( constraintTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toInteger(keyValue, &feaProblem->feaConstraint[i].dofConstraint);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaConstraint
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>gridDisplacement = 0.0 </B> </li> <br>
			 *  Value of displacement for components defined in "dofConstraint".
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>gridDisplacement = 0.0 </B> </li> <br>
			 *  Value of displacement for components defined in "dofConstraint".
			 * </ul>
			 * \endif
			 */
            keyWord = "gridDisplacement";
            status = search_jsonDictionary( constraintTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaConstraint[i].gridDisplacement);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            

        } else {


        	/*! \page feaConstraint
			 * \section keyStringConstraint Single Value String
			 *
			 * If "Value" is a string, the string value may correspond to an entry in a predefined constraint lookup
			 * table. NOT YET IMPLEMENTED!!!!
			 *
			 */

            // Call some look up table maybe?
       	 printf("\tError: Constraint tuple value is expected to be a JSON string\n");
       	 return CAPS_BADVALUE;

        }
    }

    if (keyValue != NULL) {
        EG_free(keyValue);
        keyValue = NULL;
    }

    printf("\tDone getting FEA constraints\n");
    return CAPS_SUCCESS;
}

// Get the load properties from a capsTuple
int fea_getLoad(int numLoadTuple,
                capsTuple loadTuple[],
                mapAttrToIndexStruct *attrMap,
                feaProblemStruct *feaProblem) {

	/*! \page feaLoad FEA Load
	 * Structure for the load tuple  = ("Load Name", "Value").
	 * "Load Name" defines the reference name for the load being specified.
	 *	The "Value" can either be a JSON String dictionary (see Section \ref jsonStringLoad) or a single string keyword
	 *	(see Section \ref keyStringLoad).
	 */

    int status; //Function return

    int i, groupIndex, attrIndex, nodeIndex, elementIndex; // Indexing

    char *keyValue = NULL; // Key values from tuple searches
    char *keyWord = NULL; // Key words to find in the tuples

    int numGroupName = 0;
    char **groupName = NULL;

    char *tempString = NULL; // Temporary string holder

    // Destroy our load structures coming in if aren't 0 and NULL already
    for (i = 0; i < feaProblem->numLoad; i++) {
    	status = destroy_feaLoadStruct(&feaProblem->feaLoad[i]);
    	if (status != CAPS_SUCCESS) return status;
    }
    if (feaProblem->feaLoad != NULL) EG_free(feaProblem->feaLoad);
    feaProblem->feaLoad = NULL;
    feaProblem->numLoad = 0;

    printf("\nGetting FEA loads.......\n");

    feaProblem->numLoad = numLoadTuple;

    printf("\tNumber of loads - %d\n", feaProblem->numLoad);

    // Allocate loads
    if (feaProblem->numLoad > 0) {
        feaProblem->feaLoad = (feaLoadStruct *) EG_alloc(feaProblem->numLoad * sizeof(feaLoadStruct));

        if (feaProblem->feaLoad == NULL ) return EGADS_MALLOC;
    }

    // Initiate loads to default values
    for (i = 0; i < feaProblem->numLoad; i++) {
        status = initiate_feaLoadStruct(&feaProblem->feaLoad[i]);
        if (status != CAPS_SUCCESS) return status;
    }

    // Loop through tuples and fill out the load structures
    for (i = 0; i < feaProblem->numLoad; i++) {

        printf("\tLoad name - %s\n", loadTuple[i].name );

        // Set load name to tuple attribute name
        feaProblem->feaLoad[i].name = (char *) EG_alloc((strlen(loadTuple[i].name) + 1)*sizeof(char));
        if (feaProblem->feaLoad[i].name == NULL) return EGADS_MALLOC;

        memcpy(feaProblem->feaLoad[i].name, loadTuple[i].name, strlen(loadTuple[i].name)*sizeof(char));
        feaProblem->feaLoad[i].name[strlen(loadTuple[i].name)] = '\0';

        // Set load id -> 1 bias
        feaProblem->feaLoad[i].loadID = i+1;

        // Do we have a json string?
        if (strncmp(loadTuple[i].value, "{", 1) == 0) {
            //printf("JSON String - %s\n", constraintTuple[i].value);

        	/*! \page feaLoad
			 * \section jsonStringLoad JSON String Dictionary
			 *
			 * If "Value" is JSON string dictionary
             * \if (MYSTRAN || NASTRAN)
			 *  (eg. "Value" = {"groupName": "plate", "loadType": "Pressure", "pressureForce": 2000000.0})
			 * \endif
			 *  the following keywords ( = default values) may be used:
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>loadType = "(no default)"</B> </li> <br>
			 *  Type of load. Options: "GridForce", "GridMoment", "Rotational", "Thermal",
			 *  "Pressure", "PressureDistribute", "PressureExternal", "Gravity".
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>loadType = "(no default)"</B> </li> <br>
			 *  Type of load. Options: "GridForce", "GridMoment", "Rotational", "Thermal",
			 *  "Pressure", "PressureDistribute", "PressureExternal", "Gravity".
			 * </ul>
			 * \endif
			 *
			 */

            // Get load Type
            keyWord = "loadType";
            status = search_jsonDictionary(loadTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                //{UnknownLoad, GridForce, GridMoment, Gravity, Pressure, Rotational, Thermal}
                if      (strcasecmp(keyValue, "\"GridForce\"")   == 0) feaProblem->feaLoad[i].loadType = GridForce;
                else if (strcasecmp(keyValue, "\"GridMoment\"")  == 0) feaProblem->feaLoad[i].loadType = GridMoment;
                else if (strcasecmp(keyValue, "\"Rotational\"")  == 0) feaProblem->feaLoad[i].loadType = Rotational;
                else if (strcasecmp(keyValue, "\"Thermal\"")     == 0) feaProblem->feaLoad[i].loadType = Thermal;
                else if (strcasecmp(keyValue, "\"Pressure\"")    == 0) feaProblem->feaLoad[i].loadType = Pressure;
                else if (strcasecmp(keyValue, "\"PressureDistribute\"")== 0) feaProblem->feaLoad[i].loadType = PressureDistribute;
                else if (strcasecmp(keyValue, "\"PressureExternal\"")== 0) feaProblem->feaLoad[i].loadType = PressureExternal;
                else if (strcasecmp(keyValue, "\"Gravity\"")     == 0) feaProblem->feaLoad[i].loadType = Gravity;
                else {
                    printf("\tUnrecognized \"%s\" specified (%s) for Load tuple %s\n", keyWord,
                                                                                     keyValue,
                                                                                     loadTuple[i].name);
                    if (keyValue != NULL) EG_free(keyValue);
                    return CAPS_NOTFOUND;
                }

            } else {
                printf("\t\"loadType\" variable not found in tuple %s, this is required input!!\n", loadTuple[i].name);

                if (keyValue != NULL) EG_free(keyValue);
                return status;
            }

            if (keyValue != NULL) {
                EG_free(keyValue);
                keyValue = NULL;
            }

            // Get load node/element set
            /*! \page feaLoad
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>groupName = "(no default)"</B> </li> <br>
			 *  Single or list of <em>capsGroup</em> names on which to apply the load (e.g. "Name1" or ["Name1","Name2",...].
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>groupName = "(no default)"</B> </li> <br>
			 *  Single or list of <em>capsGroup</em> names on which to apply the load (e.g. ["Name1","Name2",...].
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "groupName";
            status = search_jsonDictionary( loadTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toStringDynamicArray(keyValue, &numGroupName, &groupName);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }

                if (status != CAPS_SUCCESS) return status;

                // Determine how many element/point loads we have
                for (groupIndex = 0; groupIndex < numGroupName; groupIndex++) {

                	status = get_mapAttrToIndexIndex(attrMap, (const char *) groupName[groupIndex], &attrIndex);

                    if (status == CAPS_NOTFOUND) {
                        printf("\tgroupName name %s not found in attribute map of capsGroups!!!!\n", groupName[groupIndex]);
                        continue;

                    } else if (status != CAPS_SUCCESS) {

                        status = string_freeArray(numGroupName, &groupName);
                        if (status != CAPS_SUCCESS) return status;
                        return status;
                    }

                    //{UnknownLoad, GridForce, GridMoment, Gravity, Pressure, Rotational, Thermal}
                    if (feaProblem->feaLoad[i].loadType == GridForce  ||
                        feaProblem->feaLoad[i].loadType == GridMoment ||
                        feaProblem->feaLoad[i].loadType == Rotational ||
                        feaProblem->feaLoad[i].loadType == Thermal) {

                        feaProblem->feaLoad[i].numGridID = 0;
                        // Now lets loop through the grid to see how many grid points have the attrIndex
                        for (nodeIndex = 0; nodeIndex < feaProblem->feaMesh.numNodes; nodeIndex++ ) {
                            if (feaProblem->feaMesh.xyzAttrIndex[nodeIndex] == attrIndex) {

                                feaProblem->feaLoad[i].numGridID += 1;

                                // Allocate/Re-allocate grid ID array
                                if (feaProblem->feaLoad[i].numGridID == 1) {

                                    feaProblem->feaLoad[i].gridIDSet = (int *) EG_alloc(feaProblem->feaLoad[i].numGridID*sizeof(int));

                                } else {

                                    feaProblem->feaLoad[i].gridIDSet = (int *) EG_reall(feaProblem->feaLoad[i].gridIDSet,
                                                                                        feaProblem->feaLoad[i].numGridID*sizeof(int));
                                }

                                if (feaProblem->feaLoad[i].gridIDSet == NULL) {
                                    status = string_freeArray(numGroupName, &groupName);
                                    if (status == CAPS_SUCCESS) return status;

                                    return EGADS_MALLOC;
                                }

                                // Set grid ID value -> 1 bias
                                feaProblem->feaLoad[i].gridIDSet[feaProblem->feaLoad[i].numGridID-1] = nodeIndex+1;
                            }
                        }

                    } else if (feaProblem->feaLoad[i].loadType == Pressure || feaProblem->feaLoad[i].loadType == PressureDistribute) {

                        feaProblem->feaLoad[i].numElementID = 0;
                        // Now lets loop through the elements to see how many elements have the attrIndex

                        // This is coming from write_Nastran mesh if other mesh types are added
                        //  this will need to be changed
                        //indexOffset = feaProblem->feaMesh.numRod;

                        // Element types - CTRIA3, CTRIA3K, CQUAD4, CQUAD4K
                        for (elementIndex = 0; elementIndex < feaProblem->feaMesh.numTria3; elementIndex++) {
                            if (feaProblem->feaMesh.ctria3PID[elementIndex] == attrIndex) {

                                feaProblem->feaLoad[i].numElementID += 1;

                                // Allocate/Re-allocate element ID array
                                if (feaProblem->feaLoad[i].numElementID == 1) {

                                    feaProblem->feaLoad[i].elementIDSet = (int *) EG_alloc(feaProblem->feaLoad[i].numElementID*sizeof(int));

                                } else {

                                    feaProblem->feaLoad[i].elementIDSet = (int *) EG_reall(feaProblem->feaLoad[i].elementIDSet,
                                                                                           feaProblem->feaLoad[i].numElementID*sizeof(int));
                                }

                                if (feaProblem->feaLoad[i].elementIDSet == NULL) {
                                    status = string_freeArray(numGroupName, &groupName);
                                    if (status == CAPS_SUCCESS) return status;

                                    return EGADS_MALLOC;
                                }

                                // Set element ID value -> 1 bias
                                //feaProblem->feaLoad[i].elementIDSet[feaProblem->feaLoad[i].numElementID-1] = elementIndex+1 + indexOffset;
                                feaProblem->feaLoad[i].elementIDSet[feaProblem->feaLoad[i].numElementID-1] = feaProblem->feaMesh.ctria3ElementID[elementIndex];
                            }
                        }

                        // This is coming from write_Nastran mesh if other mesh types are added
                        //  this will need to be changed
                        //indexOffset = feaProblem->feaMesh.numRod + feaProblem->feaMesh.numTria3;

                        for (elementIndex = 0; elementIndex < feaProblem->feaMesh.numQuad4; elementIndex++) {
                            if (feaProblem->feaMesh.cquad4PID[elementIndex] == attrIndex) {

                                feaProblem->feaLoad[i].numElementID += 1;

                                // Allocate/Re-allocate element ID array
                                if (feaProblem->feaLoad[i].numElementID == 1) {

                                    feaProblem->feaLoad[i].elementIDSet = (int *) EG_alloc(feaProblem->feaLoad[i].numElementID*sizeof(int));

                                } else {

                                    feaProblem->feaLoad[i].elementIDSet = (int *) EG_reall(feaProblem->feaLoad[i].elementIDSet,
                                                                                           feaProblem->feaLoad[i].numElementID*sizeof(int));
                                }

                                if (feaProblem->feaLoad[i].elementIDSet == NULL) {
                                    status = string_freeArray(numGroupName, &groupName);
                                    if (status == CAPS_SUCCESS) return status;

                                    return EGADS_MALLOC;
                                }

                                // Set element ID value -> 1 bias
                                //feaProblem->feaLoad[i].elementIDSet[feaProblem->feaLoad[i].numElementID-1] = elementIndex+1 + indexOffset;
                                feaProblem->feaLoad[i].elementIDSet[feaProblem->feaLoad[i].numElementID-1] = feaProblem->feaMesh.cquad4ElementID[elementIndex];
                            }
                        }
                    } else if (feaProblem->feaLoad[i].loadType == Gravity){
                        // Do nothing for
                    }

                } // End attr search loop

                status = string_freeArray(numGroupName, &groupName);
                if (status != CAPS_SUCCESS) return status;
                groupName = NULL;

            } else {
                // It is ok not to provide a groupName if we are expecting pressure from a external source
                if (feaProblem->feaLoad[i].loadType != PressureExternal) {
                    printf("\tNo \"%s\" specified for Load tuple %s, where should these loads be imposed?\n", keyWord,
                                                                                                                  loadTuple[i].name);
                }
            }
            // Free keyValue (just in case)
            if (keyValue != NULL) {
                EG_free(keyValue);
                keyValue = NULL;
            }

            //Fill up load properties

            /*! \page feaLoad
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>loadScaleFactor = 1.0 </B> </li> <br>
			 *  Scale factor to use when combining loads.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>loadScaleFactor = 1.0 </B> </li> <br>
			 *  Scale factor to use when combining loads.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "loadScaleFactor";
            status = search_jsonDictionary( loadTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaLoad[i].loadScaleFactor);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaLoad
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>forceScaleFactor = 0.0 </B> </li> <br>
			 *  Overall scale factor for the force for a "GridForce" load.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>forceScaleFactor = 0.0 </B> </li> <br>
			 *  Overall scale factor for the force for a "GridForce" load.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "forceScaleFactor";
            status = search_jsonDictionary( loadTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaLoad[i].forceScaleFactor);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaLoad
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>directionVector = [0.0, 0.0, 0.0] </B> </li> <br>
			 *  X-, y-, and z- components of the force vector for a "GridForce", "GridMoment", or "Gravity" load.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>directionVector = [0.0, 0.0, 0.0] </B> </li> <br>
			 *  X-, y-, and z- components of the force vector for a "GridForce", "GridMoment", "Rotational", or "Gravity" load.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "directionVector";
            status = search_jsonDictionary( loadTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDoubleArray(keyValue,
                                              (int) sizeof(feaProblem->feaLoad[i].directionVector)/sizeof(double),
                                              feaProblem->feaLoad[i].directionVector);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaLoad
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>momentScaleFactor = 0.0 </B> </li> <br>
			 *  Overall scale factor for the moment for a "GridMoment" load.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>momentScaleFactor = 0.0 </B> </li> <br>
			 *  Overall scale factor for the moment for a "GridMoment" load.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "momentScaleFactor";
            status = search_jsonDictionary( loadTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaLoad[i].momentScaleFactor);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaLoad
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>gravityAcceleration = 0.0 </B> </li> <br>
			 *  Acceleration value for a "Gravity" load.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>gravityAcceleration = 0.0 </B> </li> <br>
			 *  Acceleration value for a "Gravity" load.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "gravityAcceleration";
            status = search_jsonDictionary( loadTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaLoad[i].gravityAcceleration);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaLoad
			 *
			 * \if MYSTRAN
			 * <ul>
			 *  <li> <B>pressureForce = 0.0 </B> </li> <br>
			 *  Uniform pressure force for a "Pressure" load (only applicable to 2D elements).
			 * </ul>
			 * \elseif NASTRAN
			 *  <ul>
			 *  <li> <B>pressureForce = 0.0 </B> </li> <br>
			 *  Uniform pressure force for a "Pressure" load.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>pressureForce = 0.0 </B> </li> <br>
			 *  Uniform pressure force for a "Pressure" load.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "pressureForce";
            status = search_jsonDictionary( loadTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaLoad[i].pressureForce);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaLoad
			 *
			 * \if MYSTRAN
			 * <ul>
			 *  <li> <B>pressureDistributeForce = [0.0, 0.0, 0.0, 0.0] </B> </li> <br>
			 *  Distributed pressure force for a "PressureDistribute" load (only applicable to 2D elements). The four values
			 *  correspond to the 4 (quadrilateral elements) or 3 (triangle elements) node locations.
			 * </ul>
			 * \elseif NASTRAN
			 * <ul>
			 *  <li> <B>pressureDistributeForce = [0.0, 0.0, 0.0, 0.0] </B> </li> <br>
			 *  Distributed pressure force for a "PressureDistribute" load.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>pressureDistributeForce = [0.0, 0.0, 0.0, 0.0] </B> </li> <br>
			 *  Distributed pressure force for a "PressureDistribute" load.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "pressureDistributeForce";
            status = search_jsonDictionary( loadTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDoubleArray(keyValue,
                                              (int) sizeof(feaProblem->feaLoad[i].pressureDistributeForce)/sizeof(double),
                                              feaProblem->feaLoad[i].pressureDistributeForce);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaLoad
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>angularVelScaleFactor = 0.0 </B> </li> <br>
			 *  An overall scale factor for the angular velocity in revolutions per unit time for a "Rotational" load.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>angularVelScaleFactor = 0.0 </B> </li> <br>
			 *  An overall scale factor for the angular velocity in revolutions per unit time for a "Rotational" load.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "angularVelScaleFactor";
            status = search_jsonDictionary( loadTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaLoad[i].angularVelScaleFactor);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaLoad
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>angularAccScaleFactor = 0.0 </B> </li> <br>
			 *  An overall scale factor for the angular acceleration in revolutions per unit time squared for a "Rotational" load.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>angularAccScaleFactor = 0.0 </B> </li> <br>
			 *  An overall scale factor for the angular acceleration in revolutions per unit time squared for a "Rotational" load.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "angularAccScaleFactor";
            status = search_jsonDictionary( loadTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaLoad[i].angularAccScaleFactor);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaLoad
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>coordinateSystem = "(no default)" </B> </li> <br>
			 *  Name of coordinate system in which defined force components are in reference to. If no value
			 *  is provided the global system is assumed.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>coordinateSystem = "(no default)" </B> </li> <br>
			 *  Name of coordinate system in which defined force components are in reference to. If no value
			 *  is provided the global system is assumed.
			 * </ul>
			 * \endif
			 *
			 */
			keyWord = "coordinateSystem";
			status = search_jsonDictionary( loadTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				tempString = string_removeQuotation(keyValue);

				for (attrIndex = 0; attrIndex < feaProblem->numCoordSystem; attrIndex++) {

					if (strcasecmp(tempString, feaProblem->feaCoordSystem[attrIndex].name) == 0) {
						feaProblem->feaLoad[i].coordSystemID  = feaProblem->feaCoordSystem[attrIndex].coordSystemID;
						break;
					}
				}

				if (feaProblem->feaLoad[i].coordSystemID == 0) {
					printf("\tCoordinate system %s not found, defaulting to global system!!", keyValue);
				}

				if (keyValue != NULL) {
					EG_free(keyValue);
					keyValue = NULL;
				}

				if (tempString != NULL) {
					EG_free(tempString);
					tempString = NULL;
				}
			}


        } else { // if not a JSON string search

        	/*! \page feaLoad
			 * \section keyStringLoad Single Value String
			 *
			 * If "Value" is a string, the string value may correspond to an entry in a predefined load lookup
			 * table. NOT YET IMPLEMENTED!!!!
			 *
			 */

            // Call some look up table maybe?
        	printf("\tError: Load tuple value is expected to be a JSON string\n");
        	return CAPS_BADVALUE;
        }
    }

    /// Free keyValue and tempString (just in case)
    if (keyValue != NULL) EG_free(keyValue);

    if (tempString != NULL) EG_free(tempString);

    printf("\tDone getting FEA loads\n");

    return CAPS_SUCCESS;
}

// Get the analysis properties from a capsTuple
int fea_getAnalysis(int numAnalysisTuple,
                    capsTuple analysisTuple[],
					feaProblemStruct *feaProblem) {

	/*! \page feaAnalysis FEA Analysis
	 * Structure for the analysis tuple  = (`Analysis Name', `Value').
	 * 'Analysis Name' defines the reference name for the analysis being specified.
	 *	The "Value" can either be a JSON String dictionary (see Section \ref jsonStringAnalysis) or a single string keyword
	 *	(see Section \ref keyStringAnalysis).
	 */

    int status; //Function return

    int i, groupIndex, attrIndex; // Indexing

    char *keyValue = NULL;
    char *keyWord = NULL;

    char *tempString = NULL;

    char **groupName = NULL;
    int  numGroupName = 0;

    // Destroy our analysis structures coming in if aren't 0 and NULL already
    for (i = 0; i < feaProblem->numAnalysis; i++) {
    	status = destroy_feaAnalysisStruct(&feaProblem->feaAnalysis[i]);
    	if (status != CAPS_SUCCESS) return status;
    }
    if (feaProblem->feaAnalysis != NULL) EG_free(feaProblem->feaAnalysis);
    feaProblem->feaAnalysis = NULL;
    feaProblem->numAnalysis = 0;

    printf("\nGetting FEA analyses.......\n");

    feaProblem->numAnalysis = numAnalysisTuple;

    printf("\tNumber of analyses - %d\n", feaProblem->numAnalysis);

    if (feaProblem->numAnalysis > 0) {
    	feaProblem->feaAnalysis = (feaAnalysisStruct *) EG_alloc(feaProblem->numAnalysis * sizeof(feaAnalysisStruct));
    } else {
        printf("\tNumber of analysis values in input tuple is 0\n");
        return CAPS_NOTFOUND;
    }

    for (i = 0; i < feaProblem->numAnalysis; i++) {
        status = initiate_feaAnalysisStruct(&feaProblem->feaAnalysis[i]);
        if (status != CAPS_SUCCESS) return status;
    }

    for (i = 0; i < feaProblem->numAnalysis; i++) {

        printf("\tAnalysis name - %s\n", analysisTuple[i].name);

        feaProblem->feaAnalysis[i].name = (char *) EG_alloc(((strlen(analysisTuple[i].name)) + 1)*sizeof(char));
        if (feaProblem->feaAnalysis[i].name == NULL) return EGADS_MALLOC;

        memcpy(feaProblem->feaAnalysis[i].name, analysisTuple[i].name, strlen(analysisTuple[i].name)*sizeof(char));
        feaProblem->feaAnalysis[i].name[strlen(analysisTuple[i].name)] = '\0';

        feaProblem->feaAnalysis[i].analysisID = i + 1;

        // Do we have a json string?
        if (strncmp(analysisTuple[i].value, "{", 1) == 0) {
            //printf("JSON String - %s\n", analysisTuple[i].value);

        	/*! \page feaAnalysis
			 * \section jsonStringAnalysis JSON String Dictionary
			 *
			 * If "Value" is JSON string dictionary
			 * \if (MYSTRAN || NASTRAN)
			 *  (eg. "Value" = {"numDesiredEigenvalue": 10, "eigenNormaliztion": "MASS", "numEstEigenvalue": 1,
			 * "extractionMethod": "GIV", "frequencyRange": [0, 10000]})
			 * \endif
			 *  the following keywords ( = default values) may be used:
			 *
			 * \if MYSTRAN
			 * <ul>
			 *  <li> <B>analysisType = "Modal"</B> </li> <br>
			 *  Type of load. Options: "Modal", "Static".
			 * </ul>
			 * \elseif NASTRAN
			 * <ul>
			 *  <li> <B>analysisType = "Modal"</B> </li> <br>
			 *  Type of load. Options: "Modal", "Static", "StaticOpt".
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>analysisType = "Modal"</B> </li> <br>
			 *  Type of load. Options: "Modal", "Static", "StaticOpt".
			 * </ul>
			 * \endif
			 *
			 */

            // Get analysis Type
            keyWord = "analysisType";
            status = search_jsonDictionary( analysisTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                //{UnknownAnalysis, Modal, Static}
                if      (strcasecmp(keyValue, "\"Modal\"")  == 0) feaProblem->feaAnalysis[i].analysisType = Modal;
                else if (strcasecmp(keyValue, "\"Static\"") == 0) feaProblem->feaAnalysis[i].analysisType = Static;
                else if (strcasecmp(keyValue, "\"StaticOpt\"") == 0) feaProblem->feaAnalysis[i].analysisType = StaticOpt;
                else {

                    printf("\tUnrecognized \"%s\" specified (%s) for Analysis tuple %s, defaulting to \"Modal\"\n", keyWord,
                                                                                                                  keyValue,
                                                                                                                  analysisTuple[i].name);
                    feaProblem->feaAnalysis[i].analysisType = Modal;
                }

            } else {

                    printf("\tNo \"%s\" specified for Analysis tuple %s, defaulting to \"Modal\"\n", keyWord,
                                                                                                   analysisTuple[i].name);
                    feaProblem->feaAnalysis[i].analysisType = Modal;
            }

            if (keyValue != NULL) {
                EG_free(keyValue);
                keyValue = NULL;
            }

            // Get loads to be applied for a given analysis

           /*! \page feaAnalysis
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>analysisLoad = "(no default)"</B> </li> <br>
			 *  Single or list of "Load Name"s defined in \ref feaLoad in which to use for the analysis (e.g. "Name1" or ["Name1","Name2",...].
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>analysisLoad = "(no default)"</B> </li> <br>
			 *  Single or list of "Load Name"s defined in \ref feaLoad in which to use for the analysis (e.g. "Name1" or ["Name1","Name2",...].
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "analysisLoad";
            status = search_jsonDictionary( analysisTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

            	status = string_toStringDynamicArray(keyValue, &numGroupName, &groupName);
            	if (keyValue != NULL) {
            		EG_free(keyValue);
            		keyValue = NULL;
            	}

            	if (status != CAPS_SUCCESS) return status;

            	feaProblem->feaAnalysis[i].numLoad = 0;
            	for (groupIndex = 0; groupIndex < numGroupName; groupIndex++) {

            		for (attrIndex = 0; attrIndex < feaProblem->numLoad; attrIndex++) {

            			if (strcasecmp(feaProblem->feaLoad[attrIndex].name, groupName[groupIndex]) == 0) {

            				feaProblem->feaAnalysis[i].numLoad += 1;

            				if (feaProblem->feaAnalysis[i].numLoad == 1)  {

            					feaProblem->feaAnalysis[i].loadSetID = (int *) EG_alloc(feaProblem->feaAnalysis[i].numLoad *sizeof(int));

            				} else {

            					feaProblem->feaAnalysis[i].loadSetID = (int *) EG_reall(feaProblem->feaAnalysis[i].loadSetID,
            																			feaProblem->feaAnalysis[i].numLoad *sizeof(int));
            				}

            				if (feaProblem->feaAnalysis[i].loadSetID == NULL) {
            					status = string_freeArray(numGroupName, &groupName);
            					if (status != CAPS_SUCCESS) return status;
            					groupName = NULL;
            					return EGADS_MALLOC;
            				}

            				feaProblem->feaAnalysis[i].loadSetID[feaProblem->feaAnalysis[i].numLoad-1] = feaProblem->feaLoad[attrIndex].loadID;
            				break;
            			}
            		}

            		if (feaProblem->feaAnalysis[i].numLoad != groupIndex+1) {

            			printf("\tWarning: Analysis load name, %s, not found in feaLoad structure\n", groupName[groupIndex]);

            		}
            	}

				status = string_freeArray(numGroupName, &groupName);
				if (status != CAPS_SUCCESS) return status;
				groupName = NULL;

            }

            // Get constraints to be applied for a given analysis
            /*! \page feaAnalysis
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>analysisConstraint = "(no default)"</B> </li> <br>
			 *  Single or list of "Constraint Name"s defined in \ref feaConstraint in which to use for the analysis (e.g. "Name1" or ["Name1","Name2",...].
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>analysisConstraint = "(no default)"</B> </li> <br>
			 *  Single or list of "Constraint Name"s defined in \ref feaConstraint in which to use for the analysis (e.g. "Name1" or ["Name1","Name2",...].
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "analysisConstraint";
            status = search_jsonDictionary( analysisTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

            	status = string_toStringDynamicArray(keyValue, &numGroupName, &groupName);
            	if (keyValue != NULL) {
            		EG_free(keyValue);
            		keyValue = NULL;
            	}

            	if (status != CAPS_SUCCESS) return status;

            	feaProblem->feaAnalysis[i].numConstraint = 0;
            	for (groupIndex = 0; groupIndex < numGroupName; groupIndex++) {

            		for (attrIndex = 0; attrIndex < feaProblem->numConstraint; attrIndex++) {

            			if (strcasecmp(feaProblem->feaConstraint[attrIndex].name, groupName[groupIndex]) == 0) {

            				feaProblem->feaAnalysis[i].numConstraint += 1;

            				if (feaProblem->feaAnalysis[i].numConstraint == 1)  {

            					feaProblem->feaAnalysis[i].constraintSetID = (int *) EG_alloc(feaProblem->feaAnalysis[i].numConstraint *sizeof(int));

            				} else {

            					feaProblem->feaAnalysis[i].constraintSetID = (int *) EG_reall(feaProblem->feaAnalysis[i].constraintSetID,
            																			      feaProblem->feaAnalysis[i].numConstraint *sizeof(int));
            				}

            				if (feaProblem->feaAnalysis[i].constraintSetID == NULL) {
            					status = string_freeArray(numGroupName, &groupName);
            					if (status != CAPS_SUCCESS) return status;
            					groupName = NULL;
            					return EGADS_MALLOC;
            				}

            				feaProblem->feaAnalysis[i].constraintSetID[feaProblem->feaAnalysis[i].numConstraint-1] = feaProblem->feaConstraint[attrIndex].constraintID;
            				break;
            			}
            		}

            		if (feaProblem->feaAnalysis[i].numConstraint != groupIndex+1) {

            			printf("\tWarning: Analysis constraint name, %s, not found in feaConstraint structure\n", groupName[groupIndex]);

            		}
            	}

				status = string_freeArray(numGroupName, &groupName);
				if (status != CAPS_SUCCESS) return status;
				groupName = NULL;

            }

            // Get design constraints to be applied for a given analysis

            /*! \page feaAnalysis
			 *
			 * \if MYSTRAN
			 * \elseif NASTRAN
			 * <ul>
			 *  <li> <B>analysisDesignConstraint = "(no default)"</B> </li> <br>
			 *  Single or list of "Design Constraint Name"s defined in \ref feaDesignConstraint in which to use for the analysis (e.g. "Name1" or ["Name1","Name2",...].
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>analysisDesignConstraint = "(no default)"</B> </li> <br>
			 *  Single or list of "Design Constraint Name"s defined in \ref feaDesignConstraint in which to use for the analysis (e.g. "Name1" or ["Name1","Name2",...].
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "analysisDesignConstraint";
            status = search_jsonDictionary( analysisTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toStringDynamicArray(keyValue, &numGroupName, &groupName);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }

                if (status != CAPS_SUCCESS) return status;

                feaProblem->feaAnalysis[i].numDesignConstraint = 0;
                for (groupIndex = 0; groupIndex < numGroupName; groupIndex++) {

                    for (attrIndex = 0; attrIndex < feaProblem->numDesignConstraint; attrIndex++) {

                        if (strcasecmp(feaProblem->feaDesignConstraint[attrIndex].name, groupName[groupIndex]) == 0) {

                            feaProblem->feaAnalysis[i].numDesignConstraint += 1;

                            if (feaProblem->feaAnalysis[i].numDesignConstraint == 1)  {

                                feaProblem->feaAnalysis[i].designConstraintSetID = (int *) EG_alloc(feaProblem->feaAnalysis[i].numDesignConstraint *sizeof(int));

                            } else {

                                feaProblem->feaAnalysis[i].designConstraintSetID = (int *) EG_reall(feaProblem->feaAnalysis[i].designConstraintSetID,
                                                                                              	  	feaProblem->feaAnalysis[i].numDesignConstraint *sizeof(int));
                            }

                            if (feaProblem->feaAnalysis[i].designConstraintSetID == NULL) {
                                status = string_freeArray(numGroupName, &groupName);
                                if (status != CAPS_SUCCESS) return status;
                                groupName = NULL;
                                return EGADS_MALLOC;
                            }

                            feaProblem->feaAnalysis[i].designConstraintSetID[feaProblem->feaAnalysis[i].numDesignConstraint-1] = feaProblem->feaDesignConstraint[attrIndex].designConstraintID;
                            break;
                        }
                    }

                    if (feaProblem->feaAnalysis[i].numDesignConstraint != groupIndex+1) {

                        printf("\tWarning: Analysis design constraint name, %s, not found in feaDesignConstraint structure\n", groupName[groupIndex]);

                    }
                }

                status = string_freeArray(numGroupName, &groupName);
                if (status != CAPS_SUCCESS) {
                    if (tempString != NULL) EG_free(tempString);

                    return status;
                }
                groupName = NULL;

            }

            //Fill up analysis properties
            /*! \page feaAnalysis
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>extractionMethod = "(no default)"</B> </li> <br>
			 *  Extraction method for modal analysis.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>extractionMethod = "(no default)"</B> </li> <br>
			 *  Extraction method for modal analysis.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "extractionMethod";
            status = search_jsonDictionary( analysisTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                tempString = string_removeQuotation(keyValue);

                if (keyValue != NULL) {
                	EG_free(keyValue);
                	keyValue = NULL;
                }

                feaProblem->feaAnalysis[i].extractionMethod = EG_alloc((strlen(tempString) + 1)*sizeof(char));
                if(feaProblem->feaAnalysis[i].extractionMethod == NULL) {
                    if (tempString != NULL) EG_free(tempString);
                    return EGADS_MALLOC;
                }

                memcpy(feaProblem->feaAnalysis[i].extractionMethod, tempString, strlen(tempString)*sizeof(char));
                feaProblem->feaAnalysis[i].extractionMethod[strlen(tempString)] = '\0';

                if (tempString != NULL) {
                    EG_free(tempString);
                    tempString = NULL;
                }
            }


            /*! \page feaAnalysis
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>frequencyRange = [0.0, 0.0] </B> </li> <br>
			 *  Frequency range of interest for modal analysis.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>frequencyRange = [0.0, 0.0] </B> </li> <br>
			 *  Frequency range of interest for modal analysis.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "frequencyRange";
            status = search_jsonDictionary( analysisTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDoubleArray(keyValue,
                                             (int) sizeof(feaProblem->feaAnalysis[i].frequencyRange)/sizeof(double),
                                             feaProblem->feaAnalysis[i].frequencyRange);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaAnalysis
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>numEstEigenvalue = 0 </B> </li> <br>
			 *  Number of estimated eigenvalues for modal analysis.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>numEstEigenvalue = 0 </B> </li> <br>
			 *  Number of estimated eigenvalues for modal analysis.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "numEstEigenvalue";
            status = search_jsonDictionary( analysisTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toInteger(keyValue, &feaProblem->feaAnalysis[i].numEstEigenvalue);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaAnalysis
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>numDesiredEigenvalue = 0</B> </li> <br>
			 *  Number of desired eigenvalues for modal analysis.
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>numDesiredEigenvalue = 0</B> </li> <br>
			 *  Number of desired eigenvalues for modal analysis.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "numDesiredEigenvalue";
            status = search_jsonDictionary( analysisTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toInteger(keyValue, &feaProblem->feaAnalysis[i].numDesiredEigenvalue);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaAnalysis
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>eigenNormaliztion = "(no default)"</B> </li> <br>
			 *  Method of eigenvector renormilization. Options: "POINT", "MAX", "MASS"
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>eigenNormaliztion = "(no default)"</B> </li> <br>
			 *  Method of eigenvector renormilization.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "eigenNormaliztion";
            status = search_jsonDictionary( analysisTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                tempString = string_removeQuotation(keyValue);

                feaProblem->feaAnalysis[i].eigenNormaliztion = EG_alloc((strlen(tempString) + 1)*sizeof(char));
                if(feaProblem->feaAnalysis[i].eigenNormaliztion == NULL) {
                    if (tempString != NULL) EG_free(tempString);

                    return EGADS_MALLOC;
                }

                memcpy(feaProblem->feaAnalysis[i].eigenNormaliztion, tempString, strlen(tempString)*sizeof(char));
                feaProblem->feaAnalysis[i].eigenNormaliztion[strlen(tempString)] = '\0';

                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (tempString != NULL) {
                    EG_free(tempString);
                    tempString = NULL;
                }
            }

            /*! \page feaAnalysis
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>gridNormaliztion = 0 </B> </li> <br>
			 *  Grid point to be used in normalizing eigenvector to 1.0 when using eigenNormaliztion = "POINT"
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>gridNormaliztion = 0</B> </li> <br>
			 *  Grid point to be used in normalizing eigenvector to 1.0 when using eigenNormaliztion = "POINT"
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "gridNormaliztion";
            status = search_jsonDictionary( analysisTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toInteger(keyValue, &feaProblem->feaAnalysis[i].gridNormaliztion);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaAnalysis
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>componentNormaliztion = 0</B> </li> <br>
			 *  Degree of freedom about "gridNormalization" to be used in normalizing eigenvector to 1.0
			 *  when using eigenNormaliztion = "POINT"
			 * </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>componentNormaliztion = 0</B> </li> <br>
			 *  Degree of freedom about "gridNormalization" to be used in normalizing eigenvector to 1.0
			 *  when using eigenNormaliztion = "POINT"
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "componentNormaliztion";
            status = search_jsonDictionary( analysisTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toInteger(keyValue, &feaProblem->feaAnalysis[i].componentNormaliztion);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaAnalysis
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>lanczosMode = 2</B> </li> <br>
			 *  Mode refers to the Lanczos mode type to be used in the solution. In mode 3 the mass matrix, Maa,must
			 *  be nonsingular whereas in mode 2 the matrix K aa - sigma*Maa must be nonsingular
			 *  </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>lanczosMode = 2</B> </li> <br>
			 *  Mode refers to the Lanczos mode type to be used in the solution. In mode 3 the mass matrix, Maa,must
			 *  be nonsingular whereas in mode 2 the matrix K aa - sigma*Maa must be nonsingular
			 *  </ul>
			 * \endif
			 */
            keyWord = "lanczosMode";
            status = search_jsonDictionary( analysisTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toInteger(keyValue, &feaProblem->feaAnalysis[i].lanczosMode);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaAnalysis
			 *
			 * \if (MYSTRAN || NASTRAN)
			 * <ul>
			 *  <li> <B>lanczosType = "(no default)"</B> </li> <br>
			 *  Lanczos matrix type. Options: DPB, DGB.
			 *  </ul>
			 * \elseif ABAQUS
			 *
			 * Something else ....
			 *
			 * \else
			 * <ul>
			 *  <li> <B>lanczosType = "(no default)"</B> </li> <br>
			 *  Lanczos matrix type. Options: DPB, DGB.
			 *  </ul>
			 * \endif
			 */
            keyWord = "lanczosType";
            status = search_jsonDictionary( analysisTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                tempString = string_removeQuotation(keyValue);

                feaProblem->feaAnalysis[i].lanczosType = EG_alloc((strlen(tempString) + 1)*sizeof(char));
                if(feaProblem->feaAnalysis[i].lanczosType == NULL) {
                    if (tempString != NULL) EG_free(tempString);

                    return EGADS_MALLOC;
                }

                memcpy(feaProblem->feaAnalysis[i].lanczosType, tempString, strlen(tempString)*sizeof(char));
                feaProblem->feaAnalysis[i].lanczosType[strlen(tempString)] = '\0';

                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }

                if (tempString != NULL) {
                    EG_free(tempString);
                    tempString = NULL;
                }
            }

         } else {

        	 /*! \page feaAnalysis
			 * \section keyStringAnalysis Single Value String
			 *
			 * If "Value" is a string, the string value may correspond to an entry in a predefined analysis lookup
			 * table. NOT YET IMPLEMENTED!!!!
			 *
			 *
			 */

            // CALL analysis look up
        	 printf("\tError: Analysis tuple value is expected to be a JSON string\n");
        	 return CAPS_BADVALUE;
        }
    }

    if (keyValue != NULL) {
        EG_free(keyValue);
        keyValue = NULL;
    }

    printf("\tDone getting FEA analyses\n");
    return CAPS_SUCCESS;
}

// Get the design variables from a capsTuple
int fea_getDesignVariable(int numDesignVariableTuple,
                          capsTuple designVariableTuple[],
                          feaProblemStruct *feaProblem) {

	/*! \page feaDesignVariable FEA DesignVariable
	 * Structure for the design variable tuple  = (`DesignVariable Name', `Value').
	 * 'DesignVariable Name' defines the reference name for the design variable being specified.
	 *	The "Value" must be a JSON String dictionary (see Section \ref jsonStringDesignVariable).
	 */
    int status; //Function return

    int i, groupIndex, attrIndex; // Indexing

    char *keyValue = NULL;
    char *keyWord = NULL;

    char *tempString = NULL;

    char **groupName = NULL;
    int  numGroupName = 0;

    // Destroy our design variables structures coming in if aren't 0 and NULL already
    for (i = 0; i < feaProblem->numDesignVariable; i++) {
    	status = destroy_feaDesignVariableStruct(&feaProblem->feaDesignVariable[i]);
    	if (status != CAPS_SUCCESS) return status;
    }
    if (feaProblem->feaDesignVariable != NULL) EG_free(feaProblem->feaDesignVariable);
    feaProblem->feaDesignVariable = NULL;
    feaProblem->numDesignVariable = 0;

    printf("\nGetting FEA design variables.......\n");

    feaProblem->numDesignVariable = numDesignVariableTuple;

    printf("\tNumber of design variables - %d\n", feaProblem->numDesignVariable);

    if (feaProblem->numDesignVariable > 0) {
        feaProblem->feaDesignVariable = (feaDesignVariableStruct *) EG_alloc(feaProblem->numDesignVariable * sizeof(feaDesignVariableStruct));
    } else {
        printf("\tNumber of design variable values in input tuple is 0\n");
        return CAPS_NOTFOUND;
    }

    for (i = 0; i < feaProblem->numDesignVariable; i++) {
        status = initiate_feaDesignVariableStruct(&feaProblem->feaDesignVariable[i]);
        if (status != CAPS_SUCCESS) return status;
    }

    for (i = 0; i < feaProblem->numDesignVariable; i++) {

        printf("\tDesign_Variable name - %s\n", designVariableTuple[i].name);

        feaProblem->feaDesignVariable[i].name = (char *) EG_alloc(((strlen(designVariableTuple[i].name)) + 1)*sizeof(char));
        if (feaProblem->feaDesignVariable[i].name == NULL) return EGADS_MALLOC;

        memcpy(feaProblem->feaDesignVariable[i].name, designVariableTuple[i].name, strlen(designVariableTuple[i].name)*sizeof(char));
        feaProblem->feaDesignVariable[i].name[strlen(designVariableTuple[i].name)] = '\0';

        feaProblem->feaDesignVariable[i].designVariableID = i + 1;

        // Do we have a json string?
        if (strncmp(designVariableTuple[i].value, "{", 1) == 0) {
            //printf("JSON String - %s\n", designVariableTuple[i].value);

        	/*! \page feaDesignVariable
			 * \section jsonStringDesignVariable JSON String Dictionary
			 *
			 * If "Value" is JSON string dictionary
			 * \if NASTRAN
			 *  (eg. "Value" = {"designVariableType": "Property", "groupName": "plate", "upperBound": 10.0, "fieldName": "TM"})
			 * \endif
			 *  the following keywords ( = default values) may be used:
			 *
			 * \if NASTRAN
			 * <ul>
			 *  <li> <B>designVariableType = "Property"</B> </li> <br>
			 *  Type of design variable in an optimization problem. Options: "Material", "Property".
			 * </ul>
			 * \endif
			 *
			 */

            // Get designVariable Type
            keyWord = "designVariableType";
            status = search_jsonDictionary( designVariableTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                //{UnknownDesignVariable, Modal, Static}
                if      (strcasecmp(keyValue, "\"Material\"")  == 0) feaProblem->feaDesignVariable[i].designVariableType = MaterialDesignVar;
                else if (strcasecmp(keyValue, "\"Property\"") == 0) feaProblem->feaDesignVariable[i].designVariableType = PropertyDesignVar;
                else {

                    printf("\tUnrecognized \"%s\" specified (%s) for Design_Variable tuple %s, defaulting to \"Property\"\n", keyWord,
                                                                                                                            keyValue,
                                                                                                                            designVariableTuple[i].name);
                    feaProblem->feaDesignVariable[i].designVariableType = PropertyDesignVar;
                }

            } else {

                printf("\tNo \"%s\" specified for Design_Variable tuple %s, defaulting to \"Property\"\n", keyWord,
                                                                                                         designVariableTuple[i].name);

                feaProblem->feaDesignVariable[i].designVariableType = PropertyDesignVar;
            }

            if (keyValue != NULL) {
                EG_free(keyValue);
                keyValue = NULL;
            }

            // Get material/properties that the design variable should be applied to
            /*! \page feaDesignVariable
			 *
			 * \if NASTRAN
			 * <ul>
			 *  <li> <B>groupName = "(no default)"</B> </li> <br>
			 *  Single or list of <em>capsGroup</em> or "Material Name"s (\ref feaMaterial) names on which to add
			 *  to the design variable (e.g. "Name1" or ["Name1","Name2",...].
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "groupName";
            status = search_jsonDictionary( designVariableTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toStringDynamicArray(keyValue, &numGroupName, &groupName);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }

                if (status != CAPS_SUCCESS) return status;

                if (feaProblem->feaDesignVariable[i].designVariableType == MaterialDesignVar) {

                    feaProblem->feaDesignVariable[i].numMaterialID = 0;
                    for (groupIndex = 0; groupIndex < numGroupName; groupIndex++) {

                        for (attrIndex = 0; attrIndex < feaProblem->numMaterial; attrIndex++) {

                            if (strcasecmp(feaProblem->feaMaterial[attrIndex].name, groupName[groupIndex]) == 0) {

                                feaProblem->feaDesignVariable[i].numMaterialID += 1;

                                if (feaProblem->feaDesignVariable[i].numMaterialID == 1)  {

                                    feaProblem->feaDesignVariable[i].materialSetID   = (int *) EG_alloc(feaProblem->feaDesignVariable[i].numMaterialID *sizeof(int));
                                    feaProblem->feaDesignVariable[i].materialSetType = (int *) EG_alloc(feaProblem->feaDesignVariable[i].numMaterialID *sizeof(int));

                                } else {

                                    feaProblem->feaDesignVariable[i].materialSetID   = (int *) EG_reall(feaProblem->feaDesignVariable[i].materialSetID,
                                                                                                        feaProblem->feaDesignVariable[i].numMaterialID *sizeof(int));
                                    feaProblem->feaDesignVariable[i].materialSetType = (int *) EG_reall(feaProblem->feaDesignVariable[i].materialSetType,
                                    																    feaProblem->feaDesignVariable[i].numMaterialID *sizeof(int));

                                }

                                if (feaProblem->feaDesignVariable[i].materialSetID   == NULL ||
                                    feaProblem->feaDesignVariable[i].materialSetType == NULL   ) {

                                    status = string_freeArray(numGroupName, &groupName);
                                    if (status != CAPS_SUCCESS) return status;
                                    groupName = NULL;
                                    return EGADS_MALLOC;
                                }

                                feaProblem->feaDesignVariable[i].materialSetID[feaProblem->feaDesignVariable[i].numMaterialID-1] = feaProblem->feaMaterial[attrIndex].materialID;
                                feaProblem->feaDesignVariable[i].materialSetType[feaProblem->feaDesignVariable[i].numMaterialID-1] = feaProblem->feaMaterial[attrIndex].materialType;

                                break;
                            }
                        }

                        if (feaProblem->feaDesignVariable[i].numMaterialID != groupIndex+1) {

                            printf("\tWarning: DesignVariable material name, %s, not found in feaMaterial structure\n", groupName[groupIndex]);

                        }
                    }
                } else if (feaProblem->feaDesignVariable[i].designVariableType == PropertyDesignVar) {

                    feaProblem->feaDesignVariable[i].numPropertyID = 0;
                    for (groupIndex = 0; groupIndex < numGroupName; groupIndex++) {

                        for (attrIndex = 0; attrIndex < feaProblem->numProperty; attrIndex++) {

                            if (strcasecmp(feaProblem->feaProperty[attrIndex].name, groupName[groupIndex]) == 0) {

                                feaProblem->feaDesignVariable[i].numPropertyID += 1;

                                if (feaProblem->feaDesignVariable[i].numPropertyID == 1)  {

                                    feaProblem->feaDesignVariable[i].propertySetID   = (int *) EG_alloc(feaProblem->feaDesignVariable[i].numPropertyID *sizeof(int));
                                    feaProblem->feaDesignVariable[i].propertySetType = (int *) EG_alloc(feaProblem->feaDesignVariable[i].numPropertyID *sizeof(int));

                                } else {

                                    feaProblem->feaDesignVariable[i].propertySetID = (int *) EG_reall(feaProblem->feaDesignVariable[i].propertySetID,
                                                                                                      feaProblem->feaDesignVariable[i].numPropertyID *sizeof(int));

                                    feaProblem->feaDesignVariable[i].propertySetType = (int *) EG_reall(feaProblem->feaDesignVariable[i].propertySetType,
                                                                                                        feaProblem->feaDesignVariable[i].numPropertyID *sizeof(int));

                                }

                                if (feaProblem->feaDesignVariable[i].propertySetID   == NULL ||
                                    feaProblem->feaDesignVariable[i].propertySetType == NULL   ) {

                                    status = string_freeArray(numGroupName, &groupName);
                                    if (status != CAPS_SUCCESS) return status;
                                    groupName = NULL;
                                    return EGADS_MALLOC;
                                }

                                feaProblem->feaDesignVariable[i].propertySetID  [feaProblem->feaDesignVariable[i].numPropertyID-1] = feaProblem->feaProperty[attrIndex].propertyID;
                                feaProblem->feaDesignVariable[i].propertySetType[feaProblem->feaDesignVariable[i].numPropertyID-1] = feaProblem->feaProperty[attrIndex].propertyType;
                                break;
                            }
                        }

                        if (feaProblem->feaDesignVariable[i].numPropertyID != groupIndex+1) {

                            printf("\tWarning: DesignVariable property name, %s, not found in feaProperty structure\n", groupName[groupIndex]);

                        }
                    }
                }

                status = string_freeArray(numGroupName, &groupName);
                if (status != CAPS_SUCCESS) return status;
                groupName = NULL;
            }

            //Fill up designVariable properties

            /*! \page feaDesignVariable
			 *
			 * \if NASTRAN
			 * <ul>
			 *  <li> <B>initialValue = 0.0</B> </li> <br>
			 *  Initial value for the design variable.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "initialValue";
            status = search_jsonDictionary( designVariableTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaDesignVariable[i].initialValue);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaDesignVariable
			 *
			 * \if NASTRAN
			 * <ul>
			 *  <li> <B>lowerBound = 0.0</B> </li> <br>
			 *  Lower bound for the design variable.
			 * </ul>
			 * \endif
			 */
            keyWord = "lowerBound";
            status = search_jsonDictionary( designVariableTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaDesignVariable[i].lowerBound);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaDesignVariable
			 *
			 * \if NASTRAN
			 * <ul>
			 *  <li> <B>upperBound = 0.0</B> </li> <br>
			 *  Upper bound for the design variable.
			 * </ul>
			 * \endif
			 */
            keyWord = "upperBound";
            status = search_jsonDictionary( designVariableTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaDesignVariable[i].upperBound);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaDesignVariable
			 *
			 * \if NASTRAN
			 * <ul>
			 *  <li> <B>maxDelta = 0.0</B> </li> <br>
			 *  Move limit for the design variable.
			 * </ul>
			 * \endif
			 */
            keyWord = "maxDelta";
            status = search_jsonDictionary( designVariableTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaDesignVariable[i].maxDelta);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaDesignVariable
 			 *
 			 * \if NASTRAN
 			 * <ul>
 			 *  <li> <B>discreteValue = 0.0</B> </li> <br>
 			 *  List of discrete values do use for the design variable (e.g. [0.0,1.0,1.5,3.0].
 			 * </ul>
 			 * \endif
 			 */
            keyWord = "discreteValue";
            status = search_jsonDictionary( designVariableTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDoubleDynamicArray(keyValue,
                                                     &feaProblem->feaDesignVariable[i].numDiscreteValue,
                                                     &feaProblem->feaDesignVariable[i].discreteValue);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) {
                    feaProblem->feaDesignVariable[i].numDiscreteValue = 0;
                    feaProblem->feaDesignVariable[i].discreteValue = NULL;
                    return status;
                }
            }

            /*! \page feaDesignVariable
 			 *
 			 * \if NASTRAN
 			 * <ul>
 			 *  <li> <B>fieldName = "(no default)"</B> </li> <br>
 			 *  Fieldname of variable (e.g. "E" for Young's Modulus). See Nastran manual for options.
 			 * </ul>
 			 * \endif
 			 */
            keyWord = "fieldName";
            status = search_jsonDictionary( designVariableTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                tempString = string_removeQuotation(keyValue);

                feaProblem->feaDesignVariable[i].fieldName = EG_alloc((strlen(tempString) + 1)*sizeof(char));
                if(feaProblem->feaDesignVariable[i].fieldName == NULL) {
                    if (tempString != NULL) EG_free(tempString);

                    return EGADS_MALLOC;
                }

                memcpy(feaProblem->feaDesignVariable[i].fieldName, tempString, strlen(tempString)*sizeof(char));
                feaProblem->feaDesignVariable[i].fieldName[strlen(tempString)] = '\0';

                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }

                if (tempString != NULL) {
                    EG_free(tempString);
                    tempString = NULL;
                }
            }

        } else {

            printf("\tError: Design_Variable tuple value is expected to be a JSON string\n");
            return CAPS_BADVALUE;

        }
    }

    if (keyValue != NULL) {
        EG_free(keyValue);
        keyValue = NULL;
    }

    printf("\tDone getting FEA design variables\n");
    return CAPS_SUCCESS;
}

// Get the design constraints from a capsTuple
int fea_getDesignConstraint(int numDesignConstraintTuple,
                            capsTuple designConstraintTuple[],
                            feaProblemStruct *feaProblem) {

	/*! \page feaDesignConstraint FEA DesignConstraint
	 * Structure for the design constraint tuple  = (`DesignConstraint Name', `Value').
	 * 'DesignConstraint Name' defines the reference name for the design constraint being specified.
	 *	The "Value" must be a JSON String dictionary (see Section \ref jsonStringDesignConstraint).
	 */

    int status; //Function return

    int i, groupIndex, attrIndex; // Indexing

    char *keyValue = NULL;
    char *keyWord = NULL;

    char *tempString = NULL;

    char **groupName = NULL;
    int  numGroupName = 0;

    // Destroy our design constraints structures coming in if aren't 0 and NULL already
    for (i = 0; i < feaProblem->numDesignConstraint; i++) {
    	status = destroy_feaDesignConstraintStruct(&feaProblem->feaDesignConstraint[i]);
    	if (status != CAPS_SUCCESS) return status;
    }
    if (feaProblem->feaDesignConstraint != NULL) EG_free(feaProblem->feaDesignConstraint);
    feaProblem->feaDesignConstraint = NULL;
    feaProblem->numDesignConstraint = 0;

    printf("\nGetting FEA design constraints.......\n");

    feaProblem->numDesignConstraint = numDesignConstraintTuple;

    printf("\tNumber of design constraints - %d\n", feaProblem->numDesignConstraint);

    if (feaProblem->numDesignConstraint > 0) {
        feaProblem->feaDesignConstraint = (feaDesignConstraintStruct *) EG_alloc(feaProblem->numDesignConstraint * sizeof(feaDesignConstraintStruct));

        if (feaProblem->feaDesignConstraint == NULL) return EGADS_MALLOC;

    } else {
        printf("\tNumber of design constraint values in input tuple is 0\n");
        return CAPS_NOTFOUND;
    }

    for (i = 0; i < feaProblem->numDesignConstraint; i++) {
        status = initiate_feaDesignConstraintStruct(&feaProblem->feaDesignConstraint[i]);
        if (status != CAPS_SUCCESS) return status;
    }

    for (i = 0; i < feaProblem->numDesignConstraint; i++) {

        printf("\tDesign_Constraint name - %s\n", designConstraintTuple[i].name);

        feaProblem->feaDesignConstraint[i].name = (char *) EG_alloc(((strlen(designConstraintTuple[i].name)) + 1)*sizeof(char));
        if (feaProblem->feaDesignConstraint[i].name == NULL) return EGADS_MALLOC;

        memcpy(feaProblem->feaDesignConstraint[i].name, designConstraintTuple[i].name, strlen(designConstraintTuple[i].name)*sizeof(char));
        feaProblem->feaDesignConstraint[i].name[strlen(designConstraintTuple[i].name)] = '\0';

        feaProblem->feaDesignConstraint[i].designConstraintID = i + 1;

        // Do we have a json string?
        if (strncmp(designConstraintTuple[i].value, "{", 1) == 0) {
            //printf("JSON String - %s\n", designConstraintTuple[i].value);

        	/*! \page feaDesignConstraint
			 * \section jsonStringDesignConstraint JSON String Dictionary
			 *
			 * If "Value" is JSON string dictionary
			 * \if NASTRAN
			 *  (eg. "Value" = {"groupName": "plate", "upperBound": 10.0})
			 * \endif
			 *  the following keywords ( = default values) may be used:
			 *
			 * \if NASTRAN
			 * <ul>
			 *  <li> <B>groupName = "(no default)"</B> </li> <br>
			 *  Single or list of <em>capsGroup</em> or "Material Name"s (\ref feaMaterial) names on which to apply
			 *  the design constraint (e.g. "Name1" or ["Name1","Name2",...].
			 * </ul>
			 * \endif
			 *
			 */

            // Get material/properties that the design constraint should be applied to
            keyWord = "groupName";
            status = search_jsonDictionary( designConstraintTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toStringDynamicArray(keyValue, &numGroupName, &groupName);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }

                if (status != CAPS_SUCCESS) return status;

                feaProblem->feaDesignConstraint[i].numPropertyID = 0;
                for (groupIndex = 0; groupIndex < numGroupName; groupIndex++) {

                    for (attrIndex = 0; attrIndex < feaProblem->numProperty; attrIndex++) {

                        if (strcasecmp(feaProblem->feaProperty[attrIndex].name, groupName[groupIndex]) == 0) {

                            feaProblem->feaDesignConstraint[i].numPropertyID += 1;

                            if (feaProblem->feaDesignConstraint[i].numPropertyID == 1)  {

                                feaProblem->feaDesignConstraint[i].propertySetID   = (int *) EG_alloc(feaProblem->feaDesignConstraint[i].numPropertyID *sizeof(int));
                                feaProblem->feaDesignConstraint[i].propertySetType = (int *) EG_alloc(feaProblem->feaDesignConstraint[i].numPropertyID *sizeof(int));

                            } else {

                                feaProblem->feaDesignConstraint[i].propertySetID   = (int *) EG_reall(feaProblem->feaDesignConstraint[i].propertySetID,
                                                                                                      feaProblem->feaDesignConstraint[i].numPropertyID *sizeof(int));
                                feaProblem->feaDesignConstraint[i].propertySetType = (int *) EG_reall(feaProblem->feaDesignConstraint[i].propertySetType,
                                                                                                      feaProblem->feaDesignConstraint[i].numPropertyID *sizeof(int));

                            }

                            if (feaProblem->feaDesignConstraint[i].propertySetID   == NULL ||
                                feaProblem->feaDesignConstraint[i].propertySetType == NULL    ) {

                                status = string_freeArray(numGroupName, &groupName);
                                if (status != CAPS_SUCCESS) return status;
                                groupName = NULL;
                                return EGADS_MALLOC;
                            }

                            feaProblem->feaDesignConstraint[i].propertySetID[feaProblem->feaDesignConstraint[i].numPropertyID-1]   = feaProblem->feaProperty[attrIndex].propertyID;
                            feaProblem->feaDesignConstraint[i].propertySetType[feaProblem->feaDesignConstraint[i].numPropertyID-1] = feaProblem->feaProperty[attrIndex].propertyType;
                            break;
                        }
                    }

                    if (feaProblem->feaDesignConstraint[i].numPropertyID != groupIndex+1) {

                        printf("\tWarning: DesignConstraint property name, %s, not found in feaProperty structure\n", groupName[groupIndex]);

                    }
                }

                status = string_freeArray(numGroupName, &groupName);
                if (status != CAPS_SUCCESS) return status;
                groupName = NULL;
            }

            //Fill up designConstraint properties

            /*! \page feaDesignConstraint
			 *
			 * \if NASTRAN
			 * <ul>
			 *  <li> <B>lowerBound = 0.0</B> </li> <br>
			 *  Lower bound for the design constraint.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "lowerBound";
            status = search_jsonDictionary( designConstraintTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaDesignConstraint[i].lowerBound);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaDesignConstraint
			 *
			 * \if NASTRAN
			 * <ul>
			 *  <li> <B>upperBound = 0.0</B> </li> <br>
			 *  Upper bound for the design constraint.
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "upperBound";
            status = search_jsonDictionary( designConstraintTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &feaProblem->feaDesignConstraint[i].upperBound);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page feaDesignConstraint
			 *
			 * \if NASTRAN
			 * <ul>
			 *  <li> <B>responseType = "(no default)"</B> </li> <br>
			 *  Response type options for DRESP1 Entry (see Nastran manual).
			 * </ul>
			 * \endif
			 *
			 */
            keyWord = "responseType";
            status = search_jsonDictionary( designConstraintTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                tempString = string_removeQuotation(keyValue);

                feaProblem->feaDesignConstraint[i].responseType = EG_alloc((strlen(tempString) + 1)*sizeof(char));
                if(feaProblem->feaDesignConstraint[i].responseType == NULL) {
                    if (tempString != NULL) EG_free(tempString);

                    return EGADS_MALLOC;
                }

                memcpy(feaProblem->feaDesignConstraint[i].responseType, tempString, strlen(tempString)*sizeof(char));
                feaProblem->feaDesignConstraint[i].responseType[strlen(tempString)] = '\0';

                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }

                if (tempString != NULL) {
                    EG_free(tempString);
                    tempString = NULL;
                }
            }

        } else {

            printf("\tError: Design_Constraint tuple value is expected to be a JSON string\n");
            return CAPS_BADVALUE;
            // CALL designConstraint look up

        }
    }

    if (keyValue != NULL) {
        EG_free(keyValue);
        keyValue = NULL;
    }

    printf("Done getting FEA design constraints\n");
    return CAPS_SUCCESS;
}

// Get the coordinate system information from the bodies and an attribute map (of CoordSystem)
int fea_getCoordSystem(int numBody,
                ego bodies[],
				mapAttrToIndexStruct coordSystemMap,
				int *numCoordSystem,
				feaCoordSystemStruct *feaCoordSystem[]) {

	int status; // Function return

	int i, body, face, edge, node; // Indexing

	int stringLength = 0;

	int numFace = 0, numEdge = 0, numNode = 0; // Number of egos
	ego *faces = NULL, *edges = NULL, *nodes = NULL; // Geometry

	int atype, alen; // EGADS return variables
	const int    *ints;
	const double *reals;
	const char *string;

	int found = (int) false;

	// Destroy our CoordSystem structures coming in if aren't 0 and NULL already
	if (*feaCoordSystem != NULL) {
		for (i = 0; i < *numCoordSystem; i++) {
			status = destroy_feaCoordSystemStruct(&(*feaCoordSystem)[i]);
			if (status != CAPS_SUCCESS) return status;
		}
	}
	if (*feaCoordSystem != NULL) EG_free(*feaCoordSystem);
	*feaCoordSystem = NULL;
	*numCoordSystem = 0;

	printf("\nGetting FEA coordinate systems.......\n");

	*numCoordSystem = coordSystemMap.numAttribute;
	printf("\tNumber of coordinate systems - %d\n", *numCoordSystem);

	if (*numCoordSystem > 0) {
		*feaCoordSystem = (feaCoordSystemStruct *) EG_alloc((*numCoordSystem)*sizeof(feaCoordSystemStruct));
		if (*feaCoordSystem == NULL) return EGADS_MALLOC;

	} else {
		printf("\tNo coordinate systems found - defaulting to global\n");
		return CAPS_SUCCESS;
	}

	for (i = 0; i < *numCoordSystem; i++) {
		status = initiate_feaCoordSystemStruct(&(*feaCoordSystem)[i]);
		if (status != CAPS_SUCCESS) return status;
	}

	for (i = 0; i < *numCoordSystem; i++) {

		if (faces != NULL) EG_free(faces);
		if (edges != NULL) EG_free(edges);
		if (nodes != NULL) EG_free(nodes);

		faces = NULL;
		edges = NULL;
		nodes = NULL;

		printf("\tCoordinate system name - %s\n", coordSystemMap.attributeName[i]);

		stringLength = strlen(coordSystemMap.attributeName[i]);

		(*feaCoordSystem)[i].name = (char *) EG_alloc((stringLength + 1)*sizeof(char));
		if ((*feaCoordSystem)[i].name == NULL) return EGADS_MALLOC;

		strncpy((*feaCoordSystem)[i].name,
				coordSystemMap.attributeName[i],
				stringLength);

		(*feaCoordSystem)[i].name[stringLength] = '\0';

		(*feaCoordSystem)[i].coordSystemID = coordSystemMap.attributeIndex[i];

		(*feaCoordSystem)[i].refCoordSystemID = 0;

		(*feaCoordSystem)[i].coordSystemType = RectangularCoordSystem;

		found = (int) false;

		// Search through bodies
		for (body = 0; body < numBody; body++) {

			// Look at the body level
			status = EG_attributeRet(bodies[body], (*feaCoordSystem)[i].name, &atype, &alen, &ints, &reals, &string);
			if (status != EGADS_SUCCESS &&
				status != EGADS_NOTFOUND) goto cleanup;

			if (atype == ATTRCSYS) {

				// Save the origin
				(*feaCoordSystem)[i].origin[0] = reals[alen+0];
				(*feaCoordSystem)[i].origin[1] = reals[alen+1];
				(*feaCoordSystem)[i].origin[2] = reals[alen+2];

				(*feaCoordSystem)[i].normal1[0] = reals[alen+3];
				(*feaCoordSystem)[i].normal1[1] = reals[alen+4];
				(*feaCoordSystem)[i].normal1[2] = reals[alen+5];

				(*feaCoordSystem)[i].normal2[0] = reals[alen+6];
				(*feaCoordSystem)[i].normal2[1] = reals[alen+7];
				(*feaCoordSystem)[i].normal2[2] = reals[alen+8];

				(*feaCoordSystem)[i].normal3[0] = reals[alen+9];
				(*feaCoordSystem)[i].normal3[1] = reals[alen+10];
				(*feaCoordSystem)[i].normal3[2] = reals[alen+11];

				found = (int) true;
			}

			if (found == (int) true)  break;

			// Determine the number of faces, edges, and nodes
			status = EG_getBodyTopos(bodies[body], NULL, FACE, &numFace, &faces);
			if (status != EGADS_SUCCESS) goto cleanup;

			status = EG_getBodyTopos(bodies[body], NULL, EDGE, &numEdge, &edges);
			if (status != EGADS_SUCCESS) goto cleanup;

			status = EG_getBodyTopos(bodies[body], NULL, NODE, &numNode, &nodes);
			if (status != EGADS_SUCCESS) goto cleanup;

			// Loop through faces
			for (face = 0; face < numFace; face++) {

				status = EG_attributeRet(faces[face], (const char *) (*feaCoordSystem)[i].name, &atype, &alen, &ints, &reals, &string);

				if (status == EGADS_NOTFOUND) continue;
				if (status != EGADS_SUCCESS) goto cleanup;

				if (atype == ATTRCSYS) {

					// Save the origin
					(*feaCoordSystem)[i].origin[0] = reals[alen+0];
					(*feaCoordSystem)[i].origin[1] = reals[alen+1];
					(*feaCoordSystem)[i].origin[2] = reals[alen+2];

					(*feaCoordSystem)[i].normal1[0] = reals[alen+3];
					(*feaCoordSystem)[i].normal1[1] = reals[alen+4];
					(*feaCoordSystem)[i].normal1[2] = reals[alen+5];

					(*feaCoordSystem)[i].normal2[0] = reals[alen+6];
					(*feaCoordSystem)[i].normal2[1] = reals[alen+7];
					(*feaCoordSystem)[i].normal2[2] = reals[alen+8];

					(*feaCoordSystem)[i].normal3[0] = reals[alen+9];
					(*feaCoordSystem)[i].normal3[1] = reals[alen+10];
					(*feaCoordSystem)[i].normal3[2] = reals[alen+11];

					found = (int) true;

					break;
				}
			} // End face loop

			if (found == (int) true)  break;

			// Loop through edges
			for (edge = 0; edge < numEdge; edge++) {

				status = EG_attributeRet(edges[edge], (*feaCoordSystem)[i].name, &atype, &alen, &ints, &reals, &string);

				if (status == EGADS_NOTFOUND) continue;
				if (status != EGADS_SUCCESS) goto cleanup;

				if (atype == ATTRCSYS) {

					// Save the origin
					(*feaCoordSystem)[i].origin[0] = reals[alen+0];
					(*feaCoordSystem)[i].origin[1] = reals[alen+1];
					(*feaCoordSystem)[i].origin[2] = reals[alen+2];

					(*feaCoordSystem)[i].normal1[0] = reals[alen+3];
					(*feaCoordSystem)[i].normal1[1] = reals[alen+4];
					(*feaCoordSystem)[i].normal1[2] = reals[alen+5];

					(*feaCoordSystem)[i].normal2[0] = reals[alen+6];
					(*feaCoordSystem)[i].normal2[1] = reals[alen+7];
					(*feaCoordSystem)[i].normal2[2] = reals[alen+8];

					(*feaCoordSystem)[i].normal3[0] = reals[alen+9];
					(*feaCoordSystem)[i].normal3[1] = reals[alen+10];
					(*feaCoordSystem)[i].normal3[2] = reals[alen+11];

					found = (int) true;

					break;
				}

			} // End edge loop

			if (found == (int) true)  break;

			// Loop through nodes
			for (node = 0; node < numNode; node++) {

				status = EG_attributeRet(nodes[node], (*feaCoordSystem)[i].name, &atype, &alen, &ints, &reals, &string);

				if (status == EGADS_NOTFOUND) continue;
				if (status != EGADS_SUCCESS) goto cleanup;

				if (atype == ATTRCSYS) {

					// Save the origin
					(*feaCoordSystem)[i].origin[0] = reals[alen+0];
					(*feaCoordSystem)[i].origin[1] = reals[alen+1];
					(*feaCoordSystem)[i].origin[2] = reals[alen+2];

					(*feaCoordSystem)[i].normal1[0] = reals[alen+3];
					(*feaCoordSystem)[i].normal1[1] = reals[alen+4];
					(*feaCoordSystem)[i].normal1[2] = reals[alen+5];

					(*feaCoordSystem)[i].normal2[0] = reals[alen+6];
					(*feaCoordSystem)[i].normal2[1] = reals[alen+7];
					(*feaCoordSystem)[i].normal2[2] = reals[alen+8];

					(*feaCoordSystem)[i].normal3[0] = reals[alen+9];
					(*feaCoordSystem)[i].normal3[1] = reals[alen+10];
					(*feaCoordSystem)[i].normal3[2] = reals[alen+11];

					found = (int) true;

					break;
				}
			} // End node loop

			if (faces != NULL) EG_free(faces);
			if (edges != NULL) EG_free(edges);
			if (nodes != NULL) EG_free(nodes);

			faces = NULL;
			edges = NULL;
			nodes = NULL;

			if (found == (int) true)  break;

		} // End body loop
	}

	status = CAPS_SUCCESS;

	goto cleanup;

	cleanup:

		if (status != CAPS_SUCCESS) printf("\tError in fea_getCoordSystem = %d\n", status);

		if (faces != NULL) EG_free(faces);
		if (edges != NULL) EG_free(edges);
		if (nodes != NULL) EG_free(nodes);

		return status;
}


// Initiate (0 out all values and NULL all pointers) of feaProblem in the feaProblemStruct structure format
int initiate_feaProblemStruct(feaProblemStruct *feaProblem) {

	int status = 0;

	// Problem analysis
	feaProblem->numAnalysis = 0;
	feaProblem->feaAnalysis = NULL;

	// Materials
	feaProblem->numMaterial = 0;
	feaProblem->feaMaterial = NULL;

	// Properties
	feaProblem->numProperty = 0;
	feaProblem->feaProperty = NULL;

	// Constraints
	feaProblem->numConstraint = 0;
	feaProblem->feaConstraint = NULL;

	// Loads
	feaProblem->numLoad = 0;
	feaProblem->feaLoad = NULL;

	// Mesh
	status = initiate_genFEAMeshStruct(&feaProblem->feaMesh);

	// Output formatting
	status = initiate_feaFileFormatStruct(&feaProblem->feaFileFormat);

	// Optimization - design variables
	feaProblem->numDesignVariable = 0;
	feaProblem->feaDesignVariable = NULL;

	// Optimization - design constraint
	feaProblem->numDesignConstraint = 0;
	feaProblem->feaDesignConstraint = NULL;

	// Coordinate Systems
	feaProblem->numCoordSystem = 0;
	feaProblem->feaCoordSystem = NULL;

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) of feaProblem in the feaProblemStruct structure format
int destroy_feaProblemStruct(feaProblemStruct *feaProblem) {

	int i; // Indexing

	int status = 0; // Status return

	// Analysis
	for (i = 0; i < feaProblem->numAnalysis; i++) {
	    status = destroy_feaAnalysisStruct(&feaProblem->feaAnalysis[i]);
	    if (status != CAPS_SUCCESS) printf("Status %d during destroy_feaAnalysisStruct\n", status);
	}

	if (feaProblem->feaAnalysis != NULL) EG_free(feaProblem->feaAnalysis);
	feaProblem->feaAnalysis = NULL;

	feaProblem->numAnalysis = 0;


	// Materials
    for (i = 0; i < feaProblem->numMaterial; i++) {
        status = destroy_feaMaterialStruct(&feaProblem->feaMaterial[i]);
        if (status != CAPS_SUCCESS) printf("Status %d during destroy_feaMaterialStruct\n", status);
    }

    if (feaProblem->feaMaterial != NULL) EG_free(feaProblem->feaMaterial);
    feaProblem->feaMaterial = NULL;

    feaProblem->numMaterial = 0;

	// Properties
	for (i = 0; i < feaProblem->numProperty; i++) {
	    status = destroy_feaPropertyStruct(&feaProblem->feaProperty[i]);
	    if (status != CAPS_SUCCESS) printf("Status %d during destroy_feaPropertyStruct\n", status);
	}

	if (feaProblem->feaProperty != NULL)EG_free(feaProblem->feaProperty);
	feaProblem->feaProperty = NULL;

	feaProblem->numProperty = 0;

	// Constraints
    for (i = 0; i < feaProblem->numConstraint; i++) {
    	status = destroy_feaConstraintStruct(&feaProblem->feaConstraint[i]);
    	if (status != CAPS_SUCCESS) printf("Status %d during destroy_feaConstraintStruct\n", status);
    }

	feaProblem->numConstraint = 0;

	if (feaProblem->feaConstraint != NULL) EG_free(feaProblem->feaConstraint);
	feaProblem->feaConstraint = NULL;

	// Loads
    for (i = 0; i < feaProblem->numLoad; i++) {
    	status = destroy_feaLoadStruct(&feaProblem->feaLoad[i]);
    	if (status != CAPS_SUCCESS) printf("Status %d during destroy_feaLoadStruct\n", status);
    }

	feaProblem->numLoad = 0;

	if (feaProblem->feaLoad != NULL) EG_free(feaProblem->feaLoad);
	feaProblem->feaLoad = NULL;

	// Mesh
	status = destroy_genFEAMeshStruct(&feaProblem->feaMesh);
	if (status != CAPS_SUCCESS) printf("Status %d during destroy_genFEAMeshStruct\n", status);

	// Output formatting
	status = destroy_feaFileFormatStruct(&feaProblem->feaFileFormat);
	if (status != CAPS_SUCCESS) printf("Status %d during destroy_feaFileFormatStruct\n", status);

    // Optimization - design variables
    for (i = 0; i < feaProblem->numDesignVariable; i++) {
        status = destroy_feaDesignVariableStruct(&feaProblem->feaDesignVariable[i]);
        if (status != CAPS_SUCCESS) printf("Status %d during destroy_feaDesignVariableStruct\n", status);
    }

    feaProblem->numDesignVariable = 0;

    if (feaProblem->feaDesignVariable != NULL) EG_free(feaProblem->feaDesignVariable);
    feaProblem->feaDesignVariable = NULL;

    // Optimization - design constraint
    for (i = 0; i < feaProblem->numDesignConstraint; i++) {
        status = destroy_feaDesignConstraintStruct(&feaProblem->feaDesignConstraint[i]);
        if (status != CAPS_SUCCESS) printf("Status %d during destroy_feaDesignConstraintStruct\n", status);
    }

    feaProblem->numDesignConstraint = 0;

    if (feaProblem->feaDesignConstraint != NULL) EG_free(feaProblem->feaDesignConstraint);
    feaProblem->feaDesignConstraint = NULL;

    // Coordinate Systems
    if (feaProblem->feaCoordSystem != NULL) {

    	for (i = 0; i < feaProblem->numCoordSystem; i++) {
    		status = destroy_feaCoordSystemStruct(&feaProblem->feaCoordSystem[i]);
    		if (status != CAPS_SUCCESS) printf("Status %d during destroy_feaCoordSystemStruct\n", status);
    	}
    }

    if (feaProblem->feaCoordSystem != NULL) EG_free(feaProblem->feaCoordSystem);
    feaProblem->feaCoordSystem = NULL;

    feaProblem->numCoordSystem = 0;

    return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) of feaFileFormat in the feaFileFormatStruct structure format
int initiate_feaFileFormatStruct(feaFileFormatStruct *feaFileFormat) {

	feaFileFormat->fileType = SmallField;

	feaFileFormat->gridFileType = LargeField;

	return CAPS_SUCCESS;
}


// Destroy (0 out all values and NULL all pointers) of feaFileFormat in the feaFileFormatStruct structure format
int destroy_feaFileFormatStruct(feaFileFormatStruct *feaFileFormat) {

	feaFileFormat->fileType = SmallField;

	feaFileFormat->gridFileType = LargeField;

	return CAPS_SUCCESS;
}


